#ifndef RVG_DRIVER_BASE_H
#define RVG_DRIVER_BASE_H

#include "../constants.h"
#include "../xform/xform.h"
#include "../path/path.h"
#include "../paint/paint.h"
#include "../image/image.h"
#include "../surface/surface.h"
#include "../scene/scene.h"
#include "../quadtree/quadtree.h"
#include "../bbox.h"
#include "../rendersettings.h"

namespace rvg { namespace driver {

class Base {
public:
    virtual ~Base(void) {}

    // create a new output surface for use as a render target
    Surface
    new_surface(int width, int height) {
        return do_new_surface(width, height);
    }

    // create a new input image from PNG file for use as a texture
    Image
    new_image(FILE *file) {
        return do_new_image(file);
    }

    // create a new input image from PNG in memory for use as a texture
    Image
    new_image(const char *str, size_t len) {
        return do_new_image(str, len);
    }

    // creates a color ramp from initializer list
    paint::Ramp
    new_color_ramp(SpreadMethod spread,
                   const std::vector<paint::ramp::Stop> &stops) {
        return do_new_color_ramp(spread, stops);
    }

    // creates a new solid color paint
    Paint
    new_solid_paint(const Color &color, float opacity = 1.f) {
        return do_new_solid_paint(color, opacity);
    }

    // creates a new linear gradient paint
    Paint
    new_linear_gradient(paint::Ramp ramp,
        r2::Point p1, r2::Point p2,
        Transform xf = xform::Identity{}, float opacity = 1.f) {
        return do_new_linear_gradient(ramp, p1, p2, xf, opacity);
    }

    // creates a new radial graident paint
    Paint
    new_radial_gradient(paint::Ramp ramp,
        r2::Point c, r2::Point f, float r,
        Transform xf = xform::Identity{}, float opacity = 1.f) {
        return do_new_radial_gradient(ramp, c, f, r, xf, opacity);
    }

    // creates a new texture paint
    Paint
    new_texture(Image image, SpreadMethod spread_method,
        Transform xf = xform::Identity{}, float opacity = 1.f) {
        return do_new_texture(image, spread_method, xf, opacity);
    }

    // creates a new empty path
    Path
    new_path(void) {
        return do_new_path();
    }

    // creates a new path from initializer list
    Path
    new_path(const std::initializer_list<Command> commands) {
        return do_new_path().append(commands);
    }

    // creates a new empty scene
    Scene
    new_scene(void) {
        return do_new_scene();
    }

    // We should move some of these methods out of the
    // interface because they don't make sense for example
    // in the SVG driver. ??

#if 0
    // creates and builds a quadtree for a scene that is
    // valid inside a given viewport
    Quadtree
    new_quadtree(Scene scene, const Viewport &viewport) {
        return do_new_quadtree(scene, viewport);
    }

    // renders a tree into a viewport of a render_target
    // settings.sampling_rate samples are produced
    // according to settings.sampling_pattern.
    // each samples are transformed by the inverse of xf
    // then modified by settins.spread_method
    // a color is then obtained from the source_tree
    // samples are integrated as per settings.prefilter_kernel
    // then blended according to settings.blend_mode
    void
    render(Quadtree source_tree, Transform xf,
        Surface render_target, const Viewport &viewport,
        const RenderSettings &settings) {
        return do_render(source_tree, xf, render_target, viewport, settings);
    }

    // renders an image into a viewport of a render_target
    // settings.sampling_rate samples are produced
    // according to settings.sampling_pattern.
    // each samples are transformed by the inverse of xf
    // then modified by settins.spread_method
    // a color is then obtained from the source_tree
    // samples are integrated as per settings.prefilter_kernel
    // then blended according to settings.blend_mode
    void
    render(Surface source_image, Transform xf,
        Surface render_target, const Viewport &viewport,
        const RenderSettings &settings) {
        return do_render(source_image, xf, render_target, viewport, settings);
    }

    // clears a viewport from a render_target using a constant color
    void
    clear(const Color &color, Surface render_target,
        const Viewport &viewport) {
        return do_clear(color, render_target, viewport);
    }
#endif

protected:

    virtual Image
    do_new_image(FILE *file) = 0;

    virtual Image
    do_new_image(const char *str, size_t len) = 0;

    virtual Surface
    do_new_surface(int width, int height) = 0;

    virtual paint::Ramp
    do_new_color_ramp(SpreadMethod spread,
                      const std::vector<paint::ramp::Stop> &stops) = 0;

    virtual Scene
    do_new_scene(void) = 0;

    virtual Paint
    do_new_solid_paint(const Color &color, float opacity = 1.f) = 0;

    virtual Paint
    do_new_linear_gradient(const paint::Ramp ramp,
        r2::Point p1, r2::Point p2,
        Transform xf = xform::Identity{}, float opacity = 1.f) = 0;

    virtual Paint
    do_new_radial_gradient(const paint::Ramp ramp,
        r2::Point c, r2::Point f, float r,
        Transform xf = xform::Identity{}, float opacity = 1.f) = 0;

    virtual Paint
    do_new_texture(Image image, SpreadMethod spread_method,
        Transform xf = xform::Identity{}, float opacity = 1.f) = 0;

    virtual Path
    do_new_path(void) = 0;

}; // class Base

}} // namespace rvg::driver

#endif // RVG_DRIVER_BASE_H
