/* vim: set foldmethod=marker foldlevel=0: */
#include "cuda.h"
#include "../path/concrete.h"
#include "../math/util.h"
#include "../image/base.h"
#include "../image/concrete.h"
#include "../image/png.h"
#include <gpu/vector.h>
#include <gpu/value.h>
#include <gpu/timer.h>
#include <gpu/alloc.h>
#include "cuda/lib/fd/fontdump.h" // use gpuraster's version of fontdump header
#include "cuda/src/curve_arc.h"
#include "cuda/src/config.h"
#include "cuda/src/gradients.h"
#include "cuda/src/texture.h"
#include "cuda/src/implicit.h"
#include "cuda/src/icurve1.h"
#include "cuda/src/icurve2.h"
#include "cuda/src/icurve3.h"
#include "cuda/src/icurve_arc.h"
#include "cuda/src/quad.h"
#include "cuda/src/segments.h"
#include "cuda/src/qtree.h"
#include "cuda/src/paint.h"
#include "cuda/src/preprocess.h"
#include "cuda/src/raster.h"
#include "cuda/src/image_util.h"
#include <image/image.h>
#include <gpu/image.h>
#include "util.h"
#include <unordered_map>
#include "../util/chronos.h"

// warps
#include "cuda/src/warps.h"
//

namespace img = image;

namespace rvg { namespace driver {

namespace
{
    ::Transform to_gpu_transform(const Transform &xform)
    {/*{{{*/
        ::Transform out;

        xform::Projective proj = xform.projective();

        out.l.x = proj[0][0];
        out.l.y = proj[0][1];
        out.t.x = proj[0][2];
        out.l.z = proj[1][0];
        out.l.w = proj[1][1];
        out.t.y = proj[1][2];
        return out;
    }/*}}}*/

    ::Projective to_gpu_projective(const Transform &xform)
    {/*{{{*/
        ::Projective out;
        xform::Projective proj = xform.projective();
        out.l0.x = proj[0][0];
        out.l0.y = proj[0][1];
        out.l0.z = proj[0][2];
        out.l1.x = proj[1][0];
        out.l1.y = proj[1][1];
        out.l1.z = proj[1][2];
        out.l2.x = proj[2][0];
        out.l2.y = proj[2][1];
        out.l2.z = proj[2][2];
        return out;
    }/*}}}*/

    Transform to_transform(const ::Transform &xform)
    {/*{{{*/
        return xform::Affine(xform.l.x, xform.l.y, xform.t.x,
                             xform.l.z, xform.l.w, xform.t.y);
    }/*}}}*/
}


namespace cuda
{

class Surface: public surface::Base
{/*{{{*/
public:
    Surface(int width, int height)
        : m_width(width), m_height(height) {}

    int width(void) const { return m_width; }
    int height(void) const { return m_height; }

private:
    int m_width, m_height;
    gpu::vector<float4> m_surface;

    IMPLEMENT_CLONEABLE(Surface);
};/*}}}*/

class Ramp : public paint::ramp::Base
{/*{{{*/
public:
    using paint::ramp::Base::Base;

    cudaTextureObject_t tex() const { return texramp().tex; }
    size_t ramp_size() const { return texramp().ramp_size; }

private:
    const TexRamp &texramp() const
    {
        if(!is_dirty())
            return m_texramp;

        fd::Ramp ramp;
        ramp.reserve(stops().size());
        for(unsigned i=0; i<stops().size(); ++i)
        {
            const auto &s = stops()[i];
            ramp.push_back({{s.color.red, s.color.green, s.color.blue,
                             s.color.opacity},s.offset});
        }

        free(m_texramp);

        fd::Spread spread;
        switch(this->spread())
        {
        case SpreadMethod::pad:
            spread = fd::SPREAD_PAD;
            break;
        case SpreadMethod::repeat:
            spread = fd::SPREAD_REPEAT;
            break;
        case SpreadMethod::reflect:
            spread = fd::SPREAD_REFLECT;
            break;
        default:
            assert(false);
            spread = fd::SPREAD_PAD;
        }

        m_texramp = convert_ramp(spread, ramp);
        reset_dirty();
        return m_texramp;
    }

    mutable TexRamp m_texramp;
};/*}}}*/

} // namespace cuda

Image
CUDA::do_new_image(FILE *file) {
    image::Concrete *img = new image::Concrete;
    if (!image::png::load(file, *img))
        fprintf(stderr, "failed loading image\n");
    return Image(img);
}

Image
CUDA::do_new_image(const char *str, size_t len) {
    image::Concrete *img = new image::Concrete;
    if (!image::png::load(std::string(str, len), *img))
        fprintf(stderr, "failed loading image\n");
    return Image(img);
}

Surface
CUDA::do_new_surface(int width, int height) {
    return Surface(new cuda::Surface(width, height));
}

paint::Ramp
CUDA::do_new_color_ramp(SpreadMethod spread,
                      const std::vector<paint::ramp::Stop> &stops) {
    return paint::Ramp(new cuda::Ramp(spread, stops));
}

Scene
CUDA::do_new_scene(void) {
    return Scene();
}

Gradient create_gradient(const paint::Ramp &ramp, const Transform &xf)
{
    auto *cuda_ramp = dynamic_cast<const cuda::Ramp *>(ramp.ro());
    if(cuda_ramp == NULL)
        throw std::runtime_error("Cannot mix cuda and non cuda objects");

    ::Gradient grad;
    grad.inv_ramp_size = 1.f/cuda_ramp->ramp_size();
    grad.ramp = cuda_ramp->tex();

    grad.transform = to_gpu_transform(xf);

    return grad;
}

Path
CUDA::do_new_path(void) {
    return Path(new path::Concrete());
}

namespace {

void calculate_image_xform(const Window &wnd, const Viewport &vp,
                           float2 &out_scale, float2 &out_offset)
{/*{{{*/
    // input dimensions
    float W = wnd.max.x - wnd.min.x,
          H = wnd.max.y - wnd.min.y;

    // output dimensions
    unsigned output_width = vp.max.x-vp.min.x,
             output_height = vp.max.y-vp.min.y;

    // input width fits exactly in output width

    // output height was calculated to accomodate input height,
    // but there can be some gap (our pixels are square),
    // we'll center input vertically on output grid

    // this is the input height related to the actual integer output height
    float Hl = output_height*W/output_width;

    // center window on available space
    float dy = (Hl - H)*0.5f;

    out_scale.y = out_scale.x = W/output_width; // square pixels

    out_offset.x = out_scale.x*0.5f + wnd.min.x;
    out_offset.y = out_scale.y*0.5f + wnd.min.y - dy;

    // invert affine transform

    /* I = O*scale+offset
       O = (I-offset)/scale
       O = I/scale - offset/scale
    */

    out_scale.x = 1/out_scale.x;
    out_scale.y = 1/out_scale.y;

    out_offset.x = -out_offset.x*out_scale.x;
    out_offset.y = -out_offset.y*out_scale.y;
}/*}}}*/

using ::make_float2;

float2 make_float2(const r2::Point &p)
{/*{{{*/
    return make_float2(p.x, p.y);
}/*}}}*/

float2 make_float2(const r2::Vector &v)
{/*{{{*/
    return make_float2(v.x, v.y);
}/*}}}*/

auto create_cuda_paint(/*{{{*/
        const Transform& scenexf,
        const ::Transform xfimage,
                       const paint::Paint &paint)
    -> std::shared_ptr< ::Paint>
{
    ::Transform sxf = to_gpu_transform( scenexf );
    ::Transform xf = compose( xfimage, sxf );
    std::shared_ptr< ::Paint> cpaint;

    // TODO: type switch should be changed to a visitor
    if(auto *solid = dynamic_cast<const paint::Solid *>(paint.ro()))
    {
        float4 c = make_float4(solid->color().red,
                               solid->color().green,
                               solid->color().blue,
                               solid->color().opacity);

        if(paint.opacity()*solid->color().opacity == 1)
        {
            cpaint.reset(gpu::cuda_new< ::PaintSolid>(c),
                         &gpu::cuda_delete);
        }
        else
        {
            cpaint.reset(gpu::cuda_new< ::PaintSolidAlpha>(c,paint.opacity()),
                         &gpu::cuda_delete);
        }
    }
    else if(auto *linear = dynamic_cast<const paint::LinearGradient *>(paint.ro()))
    {
        Gradient grad = create_gradient(linear->ramp(), linear->xf());

        grad.transform = compose(xf, grad.transform);

        grad.type = Gradient::LINEAR;
        grad.linear.p1 = make_float2(linear->p1().x, linear->p1().y);
        grad.linear.p2 = make_float2(linear->p2().x, linear->p2().y);

        cpaint.reset(gpu::cuda_new< ::PaintLinearGrad>(grad, paint.opacity()),
                     &gpu::cuda_delete);
    }
    else if(auto *radial = dynamic_cast<const paint::RadialGradient *>(paint.ro()))
    {
        Gradient grad = create_gradient(radial->ramp(), radial->xf());
        grad.transform = compose(xf, grad.transform);

        grad.type = Gradient::RADIAL;
        grad.radial.c = make_float2(radial->c().x, radial->c().y);
        grad.radial.f = make_float2(radial->f().x, radial->f().y);
        grad.radial.r = radial->r();

        cpaint.reset(gpu::cuda_new< ::PaintRadialGrad>(grad, paint.opacity()),
                     &gpu::cuda_delete);
    }
    else if(auto *ptex = dynamic_cast<const paint::Texture *>(paint.ro()))
    {
        unsigned w = ptex->image().width(),
                 h = ptex->image().height();

        std::vector<float> texdata(w*h*4);
        ptex->image().store(w,h,
                    &texdata[0], &texdata[1], &texdata[2], &texdata[3],
                    4*w, 4);
        fd::Spread spread = fd::SPREAD_PAD;
        switch(ptex->spread())
        {
        case SpreadMethod::pad:
            spread = fd::SPREAD_PAD;
            break;
        case SpreadMethod::repeat:
            spread = fd::SPREAD_REPEAT;
            break;
        case SpreadMethod::reflect:
            spread = fd::SPREAD_REFLECT;
            break;
        default:
            assert(false);
            spread = fd::SPREAD_PAD;
        }
        ::Texture tex = convert_texture(texdata, w, h, spread);
        tex.m_transform = inverse(compose(xf, to_gpu_transform(ptex->xf())));

        cpaint.reset(gpu::cuda_new< ::PaintTexture>(tex, paint.opacity()),
                     &gpu::cuda_delete);
    }

    return cpaint;
}/*}}}*/

void convert_segments(const Scene &scene, const ::Transform &xfimage,
                      std::vector<Segment<> > &out_segs,
                      std::vector<char> &out_curve_data,
                      std::vector<std::shared_ptr<const ::Paint>> &out_paints)
{/*{{{*/
    gpu::scoped_timer_stop sts(gpu::timers.cpu_add("convert paths"));
#if FRONT_TO_BACK_RENDERING
    std::vector<scene::Element> elems;
    reverse_stream(scene, elems);
#else
    const std::vector<scene::Element> &elems = scene.elements();
#endif

    // maps api paints to cuda paints
    std::unordered_map<const paint::Base *, ::Paint *> cuda_paints;

    size_t data_size = 0;
    size_t num_paths = 0;
    for(const scene::Element &elem: elems)
    {
        data_size += sizeof(PathData);

        if(elem.is_stack())
            continue;

        ++num_paths;

        const path::Concrete *concrete = dynamic_cast<const path::Concrete *>(elem.fill.path.ro());
        for(const path::Concrete::Segment &seg: concrete->segments())
        {
            switch((SegmentType)seg.type)
            {
            case SegmentType::move_to:
            case SegmentType::close_path: // do nothing??
                break;
            case SegmentType::fline_to:
            case SegmentType::line_to:
            case SegmentType::vline_to:
            case SegmentType::hline_to:
                data_size += sizeof(Curve<1>);
                break;
            case SegmentType::quad_to:
            case SegmentType::squad_to:
                data_size += sizeof(Curve<2>);
                break;
            case SegmentType::cubic_to:
            case SegmentType::scubic_to:
                data_size += sizeof(Curve<3>);
                break;
            case SegmentType::rquad_to:
                data_size += sizeof(Curve<SEG_ARC>);
                break;
            default:
                break;
            }
        }
    }
//    num_paths += 2; // to cope with outer push/pop
    out_curve_data.reserve(data_size+num_paths*sizeof(Curve<1>));
    unsigned next_idpath = 0;
#if 0
    {
        Segment<> _seg;
        _seg.idpath = idpath++;

        auto &seg = _seg.as<SEG_PUSH>();
        seg.depth = 1;
        seg.type = SEG_PUSH;

        out_segs.push_back(_seg);
    }
#endif
    for(const scene::Element &elem: elems)
    {
        // fast track handling of stack segments
        if(elem.is_stack())
        {
            Segment<> _seg;
            switch(elem.type)
            {
            case scene::Element::Type::commit:
                {
                    auto &seg = _seg.as<SEG_COMMIT>();
                    seg.type = SEG_COMMIT;
                }
                break;
            case scene::Element::Type::pushfill:
            case scene::Element::Type::pushclip:
                {
                    auto &seg = _seg.as<SEG_PUSH>();
                    seg.type = SEG_PUSH;
                }
                break;
            case scene::Element::Type::popfill:
                {
                    auto &seg = _seg.as<SEG_POP_FILL>();
                    seg.type = SEG_POP_FILL;
                }
                break;
            case scene::Element::Type::popclip:
                {
                    auto &seg = _seg.as<SEG_POP_CLIP>();
                    seg.type = SEG_POP_CLIP;
                }
                break;
            default:
                break;
            }

            out_segs.push_back(_seg);
            continue;
        }


        Path path;

        if(elem.is_fill())
            path = elem.fill.path;
        else if(elem.is_clip())
            path = elem.clip.path;

        float2 last_pt = make_float2(0,0),
               first_pt = last_pt;
        bool is_closed = true;
        bool first_path_seg = true;

        Transform xf = scene.xf()*path.xf();
        const path::Concrete *concrete = dynamic_cast<const path::Concrete *>(path.ro());

        auto close_path = [&](bool cl) {
            if(!is_closed && first_pt != last_pt) {
                Segment<> seg;
                seg.type = SEG_LINEAR;
                seg.data = &out_curve_data.back()+1;
                out_curve_data.resize(out_curve_data.size()
                                      + sizeof(Curve<1>));
                Curve<1> &curve = *seg.as<1>().B;
                curve[0] = last_pt;
                curve[1] = first_pt;
                first_pt = make_float2(0,0);
                out_segs.push_back(seg);
                first_path_seg = false;
                is_closed = cl;
            }
        };

        for(const path::Concrete::Segment &seg: concrete->segments())
        {
            const float *args = &concrete->data()[seg.index]+2;
            switch((SegmentType)seg.type)
            {
            case SegmentType::close_path:
                close_path(true);
                break;
            case SegmentType::move_to:
                is_closed = false;
                first_pt = last_pt = make_float2(xf*r2::Point{args[-2], args[-1]});
                first_path_seg = true;
                break;
#if 0 // relative
            case SegmentType::move_to:
                last_pt += make_float2(xf*r2::Vector{args[0], args[1]});
                first_path_seg = true;
                break;
#endif
            case SegmentType::fline_to:
                {
                    close_path(true);
                }
                break;
            case SegmentType::line_to:
                {
                    Segment<> seg;
                    seg.type = SEG_LINEAR;

                    // seg.data must point to the first point added in the
                    // for loop
                    seg.data = &out_curve_data.back()+1;

                    out_curve_data.resize(out_curve_data.size()
                                          + sizeof(Curve<1>));

                    Curve<1> &curve = *seg.as<1>().B;

                    curve[0] = last_pt;

//                    if(cmd.type == SegmentType::line_to)
                        last_pt = make_float2(xf*r2::Point(args[0], args[1]));
                    /* relative
                    else
                        last_pt += make_float2(xf*r2::Vector(args[0], args[1]));
                    */
                    curve[1] = last_pt;

                    out_segs.push_back(seg);
                    first_path_seg = false;
                }
                break;
            case SegmentType::hline_to:
                {
                    Segment<> seg;
                    seg.type = SEG_LINEAR;

                    // seg.data must point to the first point added in the
                    // for loop
                    seg.data = &out_curve_data.back()+1;

                    out_curve_data.resize(out_curve_data.size()
                                          + sizeof(Curve<1>));

                    Curve<1> &curve = *seg.as<1>().B;

                    curve[0] = last_pt;
//                    if(cmd.type == SegmentType::hline_to)
                        last_pt = make_float2(xf*r2::Point(args[0],last_pt.y));
                    /* relative
                    else
                        last_pt += make_float2(xf*r2::Vector(args[0],0));
                    */
                    curve[1] = last_pt;

                    out_segs.push_back(seg);
                    first_path_seg = false;
                }
                break;
            case SegmentType::vline_to:
                {
                    Segment<> seg;
                    seg.type = SEG_LINEAR;

                    // seg.data must point to the first point added in the
                    // for loop
                    seg.data = &out_curve_data.back()+1;

                    out_curve_data.resize(out_curve_data.size()
                                          + sizeof(Curve<1>));

                    Curve<1> &curve = *seg.as<1>().B;

                    curve[0] = last_pt;
//                    if(cmd.type == SegmentType::vline_to)
                        last_pt = make_float2(xf*r2::Point(last_pt.x,args[0]));
                    /* relative
                    else
                        last_pt += make_float2(xf*r2::Vector(0,args[0]));
                    */
                    curve[1] = last_pt;

                    out_segs.push_back(seg);
                    first_path_seg = false;
                }
                break;

            case SegmentType::quad_to:
                {
                    Segment<> seg;
                    seg.type = SEG_QUADRATIC;

                    // seg.data must point to the first point added in the
                    // for loop
                    seg.data = &out_curve_data.back()+1;

                    out_curve_data.resize(out_curve_data.size()
                                          + sizeof(Curve<2>));

                    Curve<2> &curve = *seg.as<2>().B;

                    curve[0] = last_pt;

                    for(int i=0; i<2; ++i)
                    {
//                        if(cmd.type == SegmentType::quad_to)
                            last_pt = make_float2(xf*r2::Point(args[i*2], args[i*2+1]));
#if 0 //relative
                        else
                            last_pt += make_float2(xf*r2::Vector(args[i*2], args[i*2+1]));
#endif
                        curve[i+1] = last_pt;
                    }

                    out_segs.push_back(seg);
                    first_path_seg = false;
                }
                break;
            case SegmentType::squad_to:
                {
                    Segment<> seg;
                    seg.type = SEG_QUADRATIC;

                    // seg.data must point to the first point added in the
                    // for loop
                    seg.data = &out_curve_data.back()+1;

                    out_curve_data.resize(out_curve_data.size()
                                          + sizeof(Curve<2>));

                    Curve<2> &curve = *seg.as<2>().B;

                    curve[0] = last_pt;

                    if(!first_path_seg && out_segs.back().type == SEG_QUADRATIC)
                    {
                        const Curve<2> &last_curve = *out_segs.back().as<2>().B;
                        last_pt += last_curve[2] - last_curve[1];
                    }
                    else
                    {
                        // assume ctrl point is coincident with current pt
                    }
                    curve[1] = last_pt;

//                    if(cmd.type == SegmentType::squad_to)
                        last_pt = make_float2(xf*r2::Point(args[0], args[1]));
#if 0 // relative
                    else
                        last_pt += make_float2(xf*r2::Vector(args[0], args[1]));
#endif
                    curve[2] = last_pt;

                    out_segs.push_back(seg);
                    first_path_seg = false;
                }
                break;

            case SegmentType::cubic_to:
                {
                    Segment<> seg;
                    seg.type = SEG_CUBIC;

                    // seg.data must point to the first point added in the
                    // for loop
                    seg.data = &out_curve_data.back()+1;

                    out_curve_data.resize(out_curve_data.size()
                                          + sizeof(Curve<3>));

                    Curve<3> &curve = *seg.as<3>().B;

                    curve[0] = last_pt;

                    for(int i=0; i<3; ++i)
                    {
//                        if(cmd.type == SegmentType::cubic_to)
                            last_pt = make_float2(xf*r2::Point(args[i*2], args[i*2+1]));
#if 0 // relative
                        else
                            last_pt += make_float2(xf*r2::Vector(args[i*2], args[i*2+1]));
#endif
                        curve[i+1] = last_pt;
                    }

                    out_segs.push_back(seg);
                    first_path_seg = false;
                }
                break;
            case SegmentType::scubic_to:
                {
                    Segment<> seg;
                    seg.type = SEG_CUBIC;

                    // seg.data must point to the first point added in the
                    // for loop
                    seg.data = &out_curve_data.back()+1;

                    out_curve_data.resize(out_curve_data.size()
                                          + sizeof(Curve<3>));

                    Curve<3> &curve = *seg.as<3>().B;

                    curve[0] = last_pt;

                    if(!first_path_seg && out_segs.back().type == SEG_CUBIC)
                    {
                        const Curve<3> &last_curve = *out_segs.back().as<3>().B;
                        last_pt += last_curve[2] - last_curve[2];
                    }
                    else
                    {
                        // assume ctrl point is coincident with current pt
                    }
                    curve[1] = last_pt;

                    for(int i=0; i<2; ++i)
                    {
//                        if(cmd.type == SegmentType::scubic_to)
                            last_pt = make_float2(xf*r2::Point(args[i*2], args[i*2+1]));
#if 0 // relative
                        else
                            last_pt += make_float2(xf*r2::Vector(args[i*2], args[i*2+1]));
#endif
                        curve[i+2] = last_pt;
                    }

                    out_segs.push_back(seg);
                    first_path_seg = false;
                }
                break;

            case SegmentType::rquad_to:
                {
                    Segment<> seg;
                    seg.type = SEG_ARC;

                    // seg.data must point to the first point added in the
                    // for loop
                    seg.data = &out_curve_data.back()+1;

                    out_curve_data.resize(out_curve_data.size()
                                          + sizeof(Curve<SEG_ARC>));

                    Curve<SEG_ARC> &curve = *seg.as<SEG_ARC>().B;

                    curve[0] = last_pt;

                    r3::Point p;

//                    if(cmd.type == SegmentType::rquad_to)
                    {
                        //this transf must happen in homogeneous coordinates.
                        p = xf*r3::Point(args[0], args[1], args[2]);
                        last_pt = make_float2(p.x, p.y);
                    }
#if 0 // relative
                    else {
                        p = xf*r3::Point(args[0], args[1], args[2]);
                        last_pt += make_float2(p.x, p.y);
                    }
#endif
                    curve[1] = last_pt;

                    curve.w = p.z;

//                    if(cmd.type == SegmentType::rquad_to)
                        last_pt = make_float2(xf*r2::Point(args[3], args[4]));
#if 0 // relative
                    else
                        last_pt += make_float2(xf*r2::Vector(args[3], args[4]));
#endif
                    curve[2] = last_pt;

                    out_segs.push_back(seg);
                    first_path_seg = false;
                }
                break;
            default:
                break;
            }
        } // for each contour

        if(!is_closed)
            close_path(true);

        // it's a new path, add path
        {
            Segment<> _seg;
            _seg.type = SEG_PATH;

            _seg.data = &out_curve_data.back()+1;
            out_curve_data.resize(out_curve_data.size()
                                  + sizeof(PathData));

            auto &seg = _seg.as<SEG_PATH>();
            PathData *pathdata = seg.path;
            pathdata->idpath = next_idpath++;

            if(elem.is_fill())
            {
                auto *paint
                    = dynamic_cast<const paint::Base *>(elem.fill.paint.ro());
                if(paint == NULL)
                    throw std::runtime_error("Must use a cuda paint with path");
                else if(dynamic_cast<const paint::Solid *>(elem.fill.paint.ro()))
                    pathdata->paint_type = static_cast<int>(::Paint::SOLID);
                else if(dynamic_cast<const paint::LinearGradient *>(elem.fill.paint.ro()))
                    pathdata->paint_type = static_cast<int>(::Paint::GRAD_LINEAR);
                else if(dynamic_cast<const paint::RadialGradient *>(elem.fill.paint.ro()))
                    pathdata->paint_type = static_cast<int>(::Paint::GRAD_RADIAL);
                else if(dynamic_cast<const paint::Texture *>(elem.fill.paint.ro()))
                    pathdata->paint_type = static_cast<int>(::Paint::TEXTURE);

                auto it = cuda_paints.find(paint);
                if(it == cuda_paints.end())
                {
                    auto cuda_paint = create_cuda_paint(scene.xf(), xfimage, elem.fill.paint);

                    it = cuda_paints.insert({paint, cuda_paint.get()}).first;
                    out_paints.push_back(cuda_paint);
                }

                pathdata->paint = it->second;
            }
            else
                pathdata->paint = NULL;

            switch(elem.type)
            {
            case scene::Element::Type::ofill:
            case scene::Element::Type::oclip:
                pathdata->frule = FRULE_ODD;
                break;
            case scene::Element::Type::efill:
            case scene::Element::Type::eclip:
                pathdata->frule = FRULE_EVEN;
                break;
            case scene::Element::Type::nzfill:
            case scene::Element::Type::nzclip:
                pathdata->frule = FRULE_NONZERO;
                break;
            case scene::Element::Type::zfill:
            case scene::Element::Type::zclip:
                pathdata->frule = FRULE_ZERO;
                break;
            default:
                throw std::runtime_error("Bad element type");
            }

            out_segs.push_back(_seg);
        }
    } // for each path

#if 0
    {
        Segment<> _seg;
        _seg.idpath = idpath++;

        auto &seg = _seg.as<SEG_POP_FILL>();
        seg.depth = 1;
        seg.type = SEG_POP_FILL;

        out_segs.push_back(_seg);
    }
#endif
}/*}}}*/

void convert_segments(const Scene &scene, const Window &wnd,
                      const Viewport &vp,
                      Segments &outsegs,
                      std::vector<std::shared_ptr<const ::Paint>> &paints)
{/*{{{*/
    std::vector<Segment<> > h_segs;
    std::vector<char> h_curve_data;

    float2 scale, off;
    calculate_image_xform(wnd, vp, scale, off);

    ::Transform xfimage;
    xfimage.l.x = scale.x;
    xfimage.l.y = 0;
    xfimage.l.z = 0;
    xfimage.l.w = scale.y;
    xfimage.t.x = off.x;
    xfimage.t.y = off.y;
    xfimage.free_space.x = xfimage.free_space.y = 0;

    convert_segments(scene, xfimage, h_segs, h_curve_data, paints);


//    printf("# paths: %lu\n",fd.paths.size());
    //printf("# segments: %lu\n",h_segs.size());

    gpu::vector<Segment<> > segs(h_segs);
    gpu::vector<char> curve_data(h_curve_data);

    adjust_curve_data_pointer(segs, curve_data, h_curve_data);

    // convert segments to implicit curve segments
    preprocess(segs, outsegs, scale, off);

    //printf("# xy-monotonic segment: %lu\n",outsegs.size());

    /*
    printf("Sizes: B1:%lu, B2:%lu, B3:%lu, ARC:%lu, I:%lu\n",
           sizeof(ImplicitCurve<1>),
           sizeof(ImplicitCurve<2>),
           sizeof(ImplicitCurve<3>),
           sizeof(ImplicitCurve<SEG_ARC>),
           sizeof(ImplicitSegment<>));
           */
}/*}}}*/

} //namespace

struct ShortcutTree
{
    Quads qtree;
    Segments segs;
    std::vector<std::shared_ptr<const ::Paint>> paints;
};


template<bool do_benchmark = false>
ShortcutTree* create_shortcuttree(Scene scene,const Window& wnd,
                                    const Viewport& vp,
                                    const CUDAConfig &cfg, CUDABenchmark& bench)
{

    std::unique_ptr<ShortcutTree> stree(new ShortcutTree());

    convert_segments(scene, wnd, vp, stree->segs, stree->paints);

    size_t out_width = vp.max.x - vp.min.x,
           out_height = vp.max.y - vp.min.y;
    // Color _bg_color{1,1,1,1}; // never used

    // unsigned qtree_levels = 10; // never used
    // unsigned min_segs_split = 10;
    // unsigned nsamples = 32;

    if(stree->segs.size() != 0)
    {
        // create acceleration quadtree
        {
            //printf("mem used BEFORE qtree creation: %f MiB\n",gpu::mem_used()/1024.0/1024);
            //printf("# segments BEFORE quadtree split: %lu\n",stree->segs.size());

            unsigned dim = std::max(next_power_of_2(out_width),
                                    next_power_of_2(out_height));
            gpu::Box bbox;
            bbox.min = make_float2(0,0);
            bbox.max = make_float2(dim, dim);
            if ( do_benchmark ) {
                chronos clock;
                clock.reset();
                create_qtree(stree->qtree, stree->segs,
                             bbox, cfg.qtree_max_level,
                             cfg.min_segs_split, cfg.dump & 1);
                bench.qtree.ms = clock.elapsed();
                bench.qtree.mem = gpu::mem_used()/(1024.0*1024.0);
                bench.qtree.segs = stree->segs.size();
            }
            else {
                create_qtree(stree->qtree, stree->segs,
                             bbox, cfg.qtree_max_level,
                             cfg.min_segs_split, cfg.dump & 1);
            }

            //printf("mem used after qtree creation: %f MiB\n",gpu::mem_used()/1024.0/1024);
            //printf("# segments after quadtree split: %lu\n",stree->segs.size());

        }
        //printf("# segments after quadtree split: %lu\n",d_segs.size());
    }
    return stree.release();
}

ShortcutTree* CUDA::create_shortcut(Scene scene,const Window& wnd,
                                    const Viewport& vp,
                                    const CUDAConfig &cfg)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("shortcut"));
    CUDABenchmark bench;
    ShortcutTree * tree = create_shortcuttree( scene, wnd, vp, cfg, bench );
    sts.stop();
    gpu::timers.flush_results( tresult );
    return tree;
}

void CUDA::release(ShortcutTree *tree) const
{
    delete tree;
}

void
render_scene(const ShortcutTree &stree, Transform xf, const Viewport& vp,
        const std::string &fname, const CUDAConfig& config, gpu::image_ptr<float4> out )
{
    (void) fname;
    size_t out_width = vp.max.x - vp.min.x,
           out_height = vp.max.y - vp.min.y;
    Color _bg_color{1,1,1,1};
    unsigned nsamples = config.nsamples;
    float4 bg_color = make_float4(_bg_color.red, _bg_color.green,
                                  _bg_color.blue, _bg_color.opacity);
    switch(config.kernel)
    {
    case 42:
        rasterize_shared(bg_color, stree.qtree, stree.segs, nsamples, out, config.dump, config.use_bs3 );
        break;
    case 43:
        {
            if ( config.use_warp == 1 ) { //projective
                ::ProjectiveWarp fs(to_gpu_projective(xf));
                rasterize_warp<ProjectiveWarp>(fs, bg_color, stree.qtree, stree.segs, 32, out, config.dump, config.use_bs3 );
            }
            else if ( config.use_warp == 2 ) { //lens
                ::LensWarp lens( config.data.lens.cu,
                        config.data.lens.cv,
                        config.data.lens.outer,
                        config.data.lens.inner,
                        config.data.lens.zoom, out_width, out_height );
                rasterize_warp<LensWarp>(lens, bg_color, stree.qtree, stree.segs, 32, out, config.dump, config.use_bs3 );
            }
            else if ( config.use_warp == 3 ) { //swirl
                ::SwirlWarp swirl( config.data.swirl.amount, config.data.swirl.ang, out_width, out_height, config.data.swirl.scale );
                rasterize_warp<SwirlWarp>(swirl, bg_color, stree.qtree, stree.segs, 32, out, config.dump, config.use_bs3 );
            }
            else {
                ::NoWarp nowarp;
                rasterize_warp<NoWarp>(nowarp, bg_color, stree.qtree, stree.segs, 32, out, config.dump, config.use_bs3 );
            }
        }
        break;
    default:
        assert(0);
    }
}

void print_timer_result( const gpu::timer_result& result, int level )
{
    for (int i=0; i<level; i++) {
        std::cout << "-";
    }
    std::cout << result.name << " [" << result.time << "]" << std::endl;
    for( const auto& r: result.children ) {
        print_timer_result( r, level + 1 );
    }
}

void
CUDA::reset_timer( void )
{
    tresult.children.resize(0);
}

void
CUDA::render(const ShortcutTree &stree, Transform xf, const Viewport& vp,
        const std::string &fname, const CUDAConfig& config)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("rendering to file"));
    Color _bg_color{1,1,1,1};
    size_t out_width = vp.max.x - vp.min.x,
           out_height = vp.max.y - vp.min.y;
    img::rgba<> h_out;
    if(stree.segs.size() != 0) {
        gpu::image<float4> d_out;
        d_out.resize(out_width, out_height);
        render_scene( stree, xf, vp, fname, config, &d_out );
        h_out.resize(out_height, out_width);
        copy_to_host(h_out, d_out);
    }
    else {
        h_out.resize(out_height, out_width);
        for(int i=0; i<h_out.width()*h_out.height(); ++i) {
            h_out.r[i] = _bg_color.red;
            h_out.g[i] = _bg_color.green;
            h_out.b[i] = _bg_color.blue;
            h_out.a[i] = _bg_color.opacity;
        }
    }
    sts.stop();
    gpu::timers.flush_results( tresult );
    if ( fname.length() > 0 )
        store(fname.c_str(), h_out);
}

Paint CUDA::do_new_solid_paint(const Color &color, float opacity)
{
    return paint::Paint(new paint::Solid(color), opacity);
}

Paint CUDA::do_new_linear_gradient(const paint::Ramp ramp,
        r2::Point p1, r2::Point p2,
        Transform xf, float opacity)
{
    return Paint( new paint::LinearGradient(ramp, p1, p2, xf), opacity);
}

Paint CUDA::do_new_radial_gradient(const paint::Ramp ramp,
        r2::Point c, r2::Point f, float r,
        Transform xf, float opacity)
{
    return Paint( new paint::RadialGradient(ramp, c, f, r, xf), opacity);
}

Paint CUDA::do_new_texture(Image image, SpreadMethod spread_method,
        Transform xf, float opacity)
{
    return Paint(new paint::Texture(image, spread_method, xf), opacity);
}

CUDA::CUDA()
{ 
    tresult.time = 0;
    tresult.name = "total";
}

}} // namespace rvg::driver
