#ifndef RVG_DRIVER_CUDA_H
#define RVG_DRIVER_CUDA_H

#include <string>
#include <iosfwd>
#include "base.h"
#include "cuda/lib/gpu/timer_result.h"

struct cudaGraphicsResource;

namespace rvg { namespace driver {

struct CUDAConfig {
    int kernel;
    int min_segs_split;
    int dump;
    int qtree_max_level;
    int use_bs3;
    int use_warp;
    union {
        int foreshorten;
        struct {
            float cu, cv,
                  outer, inner,
                  zoom;
        } lens;
        struct {
            float amount;
            float ang;
            float scale;
        } swirl;
    } data;
    int nsamples;

    CUDAConfig( )
        : kernel(0), min_segs_split(8)
        , dump(0), qtree_max_level(12)
        , use_bs3(0), use_warp(0)
    {}

    CUDAConfig( int k, int mins, int dump, int qmax, int usebs3 )
        : kernel(k), min_segs_split(mins)
        , dump(dump), qtree_max_level(qmax)
        , use_bs3(usebs3)
    { }
};

struct CUDABenchmark {
    struct {
        double ms;
    } render;
    struct {
        double ms;
        double mem;
        int segs;
    } qtree;
};

struct ShortcutTree;

class CUDA final: public Base {
public:
    void render(const ShortcutTree &stree, Transform xf, const Viewport& vp,
            const std::string &fname, const CUDAConfig& config);

    ShortcutTree* create_shortcut(Scene scene,const Window& wnd,
                                  const Viewport& vp, const CUDAConfig &cfg);

    void release(ShortcutTree *tree) const;

    void reset_timer( void );

    gpu::timer_result tresult;

    CUDA();

private:
    Image
    do_new_image(FILE *file) override;

    Image
    do_new_image(const char *str, size_t len) override;

    Surface
    do_new_surface(int width, int height) override;

    paint::Ramp
    do_new_color_ramp(SpreadMethod spread,
                      const std::vector<paint::ramp::Stop> &stops) override;

    Paint
    do_new_solid_paint(const Color &color, float opacity) override;

    Paint
    do_new_linear_gradient(const paint::Ramp ramp,
        r2::Point p1, r2::Point p2,
        Transform xf = xform::Identity{}, float opacity=1.0f) override;

    Paint
    do_new_radial_gradient(const paint::Ramp ramp,
        r2::Point c, r2::Point f, float r,
        Transform xf = xform::Identity{}, float opacity=1.0f) override;

    Paint
    do_new_texture(Image image, SpreadMethod spread_method,
        Transform xf = xform::Identity{}, float opacity = 1.f) override;

    Scene
    do_new_scene(void) override;

    Path
    do_new_path(void) override;

}; // class rvg::svg::Driver

}} // namespace rvg::svg

#endif // RVG_DRIVER_CUDA_H
