#include <cstdio>
#include <sstream>
#include <cctype>
#include <cmath>
#include <cassert>
#include <cstring>
#include <algorithm>
#include <unordered_set>

#include "fontdump.h"
#include "error.h"

namespace fd
{

namespace
{

std::ostream &format_aux(std::ostream &out)
{
    return out;
}

template <class T, class...ARGS>
std::ostream &format_aux(std::ostream &out, const T &arg0, ARGS &&...args)
{
    out << arg0;
    return format_aux(out, std::forward<ARGS>(args)...);
}

template <class...ARGS>
std::string format(ARGS &&...args)
{
    std::ostringstream ss;
    format_aux(ss, std::forward<ARGS>(args)...);
    return ss.str();
}

// skips comments and empty lines
void skip(FILE *fin) {
    int c = ' ';
    while (1) {
        while (isspace(c))
            c = getc(fin);
        if (c == '#')
            while (c != '\n' && c != EOF)
                c = getc(fin);
        else
            break;
    }
    ungetc(c, fin);
}

void parse_point(FILE *fin, int &tag, Point &p) {
    int c = getc(fin);
    switch (c) {
    case 'i':
    case 'q':
    case 'c':
    case 'a':
        tag = c;
        break;
    default:
        throw std::runtime_error("invalid tag");
    }

    if (fscanf(fin, "%f %f", &p.x, &p.y) != 2)
        throw std::runtime_error("syntax error");
}

void add_point(Contour &contour, const Point &p0)
{
    contour.box.merge(p0);
    contour.points.push_back(p0);

    contour.segments.push_back(SEGMENT_POINT);
}

void add_linear(Contour &contour, const Point &p0)
{
    contour.box.merge(p0);
    contour.points.push_back(p0);

    contour.segments.push_back(SEGMENT_LINEAR);
}

void add_quadratic(Contour &contour,
                  const Point &p0, const Point &p1)
{
    contour.box.merge(p0);
    contour.points.push_back(p0);

    contour.box.merge(p1);
    contour.points.push_back(p1);

    contour.segments.push_back(SEGMENT_QUADRATIC);
}

void add_cubic(Contour &contour, const Point &p0,
              const Point &p1, Point &p2)
{
    contour.box.merge(p0);
    contour.points.push_back(p0);

    contour.box.merge(p1);
    contour.points.push_back(p1);

    contour.box.merge(p2);
    contour.points.push_back(p2);

    contour.segments.push_back(SEGMENT_CUBIC);
}

void add_arc(Contour &contour, const Point &p0,
              const Point &p1, Point &p2)
{
    contour.box.merge(p0);
    contour.points.push_back(p0);

    // we'll use p1.x only as a parameter, not a point
//    contour.box.merge(p1);
    contour.points.push_back(p1);

    contour.box.merge(p2);
    contour.points.push_back(p2);

    contour.segments.push_back(SEGMENT_ARC);
}


bool check_fontdump_magic(FILE *fin)
{
    char magic[5] = "fodu";
    if (fread(magic, 1, 4, fin) != 4)
       return false;
    if (strcmp("fodu", magic))
        return false;
    return true;
}

bool parse_linear_attr(FILE *fin, Linear &linear)
{
    skip(fin);

    if (fscanf(fin, "L %f %f %f %f",
                &linear.p1.x, &linear.p1.y,
                &linear.p2.x, &linear.p2.y) == 4)
    {
        return true;
    }
    else
        return false;
}

bool parse_radial_attr(FILE *fin, Radial &radial)
{
    skip(fin);

    if (fscanf(fin, "R %f %f %f %f %f",
                &radial.c.x, &radial.c.y,
                &radial.r, &radial.f.x, &radial.f.y) == 5)
    {
        return true;
    }
    else
        return false;
}

void parse_grad_attr(FILE *fin, Gradient &grad)
{
    Transform &t = grad.transform;
    // read transform
    if (fscanf(fin, "%f %f %f %f %f %f",
                &t.a, &t.b, &t.tx,
                &t.c, &t.d, &t.ty) == 6)
    {
        // try linear or radial gradients
        if (parse_linear_attr(fin, grad.linear))
            grad.type = GRADIENT_LINEAR;
        else if (parse_radial_attr(fin, grad.radial))
            grad.type = GRADIENT_RADIAL;
        else
            throw std::runtime_error("expected linear or radial parameters");
    }
    else
        throw std::runtime_error("expected transform");
}

bool parse_fill_attr(FILE *fin, std::vector<Gradient> &gradients, Fill &fill)
{
    skip(fin);
    // try gradient fill
    unsigned ramp, spread;
    if (fscanf(fin, "F %u %u", &ramp, &spread) == 2)
    {
        Gradient gradient;
        gradient.ramp = ramp;
        gradient.spread = (Spread) spread;

        // try to read opacity
        skip(fin);
        fill.opacity = 1;
        int ignore = fscanf(fin, "a %f", &fill.opacity); (void) ignore;

        parse_grad_attr(fin, gradient);

        fill.mode = COLORING_GRADIENT;
        fill.gradient = gradients.size();
        gradients.push_back(gradient);
        return true;
    // try solid fill
    }
    else if (fscanf(fin, "f %f %f %f",
                &fill.solid.r, &fill.solid.g, &fill.solid.b) == 3)
    {
        fill.solid.a = 1;
        fill.mode = COLORING_SOLID;
        fill.opacity = 1;
        int ignore = fscanf(fin, " a %f", &fill.opacity); (void) ignore;
        return true;
    }
    else
        return false;
}

bool parse_stroke_attr(FILE *fin, std::vector<Gradient> &gradients,
                       Stroke &stroke)
{
    skip(fin);
    // try gradient stroke
    unsigned ramp, spread;
    if (fscanf(fin, "S %u %u %f", &ramp, &spread, &stroke.width) == 3)
    {
        Gradient gradient;
        gradient.ramp = ramp;
        gradient.spread = (Spread)spread;

        // try to read opacity
        stroke.opacity = 1;
        skip(fin);
        int ignore = fscanf(fin, "a %f", &stroke.opacity); (void) ignore;

        // read remaining attributes
        parse_grad_attr(fin, gradient);

        stroke.mode = COLORING_GRADIENT;
        stroke.gradient = gradients.size();
        gradients.push_back(gradient);
        return true;
    // try solid stroke
    }
    else if (fscanf(fin, "s %f %f %f %f",
                &stroke.solid.r, &stroke.solid.g, &stroke.solid.b,
                &stroke.width) == 4)
    {
        stroke.solid.a = 1;
        stroke.mode = COLORING_SOLID;

        // try to read opacity
        skip(fin);
        stroke.opacity = 1;
        int ignore = fscanf(fin, "a %f", &stroke.opacity); (void) ignore;
        return true;
    }
    else
        return false;
}

bool parse_clip_path(FILE *fin, ClipPath &clippath)
{
    skip(fin);
    unsigned depth;
    if(fscanf(fin, "C %u", &depth) == 1)
    {
        clippath.depth = depth;
        return true;
    }
    else
        return false;
}

void parse_path_attr(FILE *fin, std::vector<Gradient> &gradients,
                     Fill &fill, Stroke &stroke, ClipPath &clippath)
{
    fill.mode = COLORING_NONE;
    stroke.mode = COLORING_NONE;
    clippath.depth = 0;

    bool f = parse_fill_attr(fin, gradients, fill),
         s = parse_stroke_attr(fin, gradients, stroke),
         c = parse_clip_path(fin, clippath);

    if (!s && !f && !c)
        throw std::runtime_error("expected stroke, fill or clip path");
}

Stop parse_stop_attr(FILE *fin)
{
    Stop stop;

    if(fscanf(fin, " %f %f %f %f a %f", &stop.offset,
            &stop.color.r, &stop.color.g, &stop.color.b,
            &stop.color.a) != 5)
    {
        throw std::runtime_error("syntax error");
    }

    return std::move(stop);
}

Contour parse_contour(FILE *fin)
{
    Contour contour;

    skip(fin);

    unsigned int npoints = 0;
    if (fscanf(fin, "%u\n", &npoints) != 1 || npoints < 1)
        throw std::runtime_error("expected number of points");

    skip(fin);

    Point p[4]{};
    int tag[4]{};

    parse_point(fin, tag[0], p[0]);
    if(tag[0] != 'i')
        throw std::runtime_error("expected i point");

    add_point(contour, p[0]);

    Point p0 = p[0];

    // go through all remaining points
    for (unsigned int i = 1; i < npoints; i++)
    {
        skip(fin);

        parse_point(fin, tag[i%4], p[i%4]);

        switch (tag[(i-1)%4])
        {
        case 'i':
            // linear segment
            if (tag[i%4] == 'i')
                add_linear(contour, p[i%4]);
            break;
        case 'q':
            switch (tag[i%4])
            {
            case 'i':
                // make sure we completed a quadratic segment
                if (tag[(i-2)%4] != 'i')
                    throw std::runtime_error("bad quadratic curve definition");
                add_quadratic(contour, p[(i-1)%4], p[i%4]);
                break;
            case 'q': {
                // double q implies midpoint i
                if (tag[(i-2)%4] != 'i')
                    throw std::runtime_error("bad quadratic curve definition");
                Point pc = .5f*(p[(i-1)%4] + p[i%4]);
                add_quadratic(contour, p[(i-1)%4], pc);
                p[(i-1)%4] = pc;
                tag[(i-1)%4] = 'i';
                break;
            }
            case 'c':
                throw std::runtime_error("mixed cubic/quadratic points");
            case 'a':
                throw std::runtime_error("mixed arc/quadratic points");
            }
            break;
        case 'c':
            switch (tag[i%4])
            {
            case 'i':
                // make sure we completed a cubic
                if (tag[(i-3)%4] != 'i')
                    throw std::runtime_error("bad cubic curve definition");
                if (tag[(i-2)%4] != 'c')
                    throw std::runtime_error("bad cubic curve definition");
                add_cubic(contour, p[(i-2)%4], p[(i-1)%4], p[i%4]);
                break;
            case 'q':
                throw std::runtime_error("mixed cubic/quadratic points");
            case 'a':
                throw std::runtime_error("mixed cubic/arc points");
            }
            break;
        case 'a':
            switch (tag[i%4])
            {
            case 'i':
                // make sure we completed an arc
                if (tag[(i-3)%4] != 'i')
                    throw std::runtime_error("bad arc curve definition");
                if (tag[(i-2)%4] != 'a')
                    throw std::runtime_error("bad arc curve definition");
                add_arc(contour, p[(i-2)%4], p[(i-1)%4], p[i%4]);
                break;
            case 'q':
                // error!
                throw std::runtime_error("mixed arc/quadratic points");
            case 'c':
                // error!
                throw std::runtime_error("mixed arc/cubic points");
            }
            break;
        } // tag switch
    } // for each point

    // check if first and last points are the same and close
    // contour
    if (p0 == p[(npoints-1)%4])
        contour.closed = true;
    else
        contour.closed = false;

    return std::move(contour);
}

Ramp parse_ramp(FILE *fin)
{
    Ramp ramp;

    unsigned nstops = 0;
    if (fscanf(fin, "%u", &nstops) != 1)
        throw std::runtime_error("expected number of stops");

    ramp.reserve(nstops);
    for (unsigned s = 0; s < nstops; s++)
    {
        try
        {
            ramp.emplace_back(parse_stop_attr(fin));
        }
        catch(...)
        {
            throw_with_nested(std::runtime_error(format("error parsing stop ", s)));
        }
    } // for each stop

    return std::move(ramp);
}


struct AttribIdx
{
    Attrib *attrib;

    bool operator==(const AttribIdx &that) const
    {
        return *attrib == *that.attrib;
    }

    mutable size_t idx;
};

size_t hash_value(const AttribIdx &idx)
{
    return hash_value(*idx.attrib);
}

} // local namespace

FontDump load(FILE *fin)
{
    FontDump fd;

    if (!check_fontdump_magic(fin))
        throw std::runtime_error("not a fontdump file");

    skip(fin);
    if (fscanf(fin, "b %f %f %f %f", &fd.bbox.min.x, &fd.bbox.max.x,
                &fd.bbox.min.y, &fd.bbox.max.y) != 4)
    {
        throw std::runtime_error("expecting fontdump bbox");
    }

    skip(fin);

    // try to read gradient ramps
    unsigned int nramps = 0;
    if(fscanf(fin, "r %u", &nramps) == 1)
    {
        fd.ramps.reserve(nramps);
        for(unsigned r = 0;  r < nramps; r++)
        {
            try
            {
                fd.ramps.emplace_back(parse_ramp(fin));
            }
            catch(...)
            {
                throw_with_nested(std::runtime_error(format("error ramp ", r)));
            }
        }
    }

    skip(fin);

    // read paths
    unsigned int npaths = 0;
    if(fscanf(fin, "p %u", &npaths) != 1 && fscanf(fin, "%u", &npaths) != 1)
        throw std::runtime_error("expected number of paths");

    std::vector<Path> &paths = fd.paths;
    paths.resize(npaths);

    std::vector<Attrib> &attribs = fd.attribs;
    attribs.reserve(npaths);

    typedef std::unordered_set<AttribIdx,boost::hash<AttribIdx>> AttribSet;
    AttribSet unique_attribs;

    for (unsigned int a = 0; a < npaths; a++)
    {
        try
        {
            unsigned int ncontours = 0;
            Path &path = paths[a];
            path.id = a;

            Attrib attrib;

            parse_path_attr(fin, fd.gradients,
                            attrib.fill, attrib.stroke, attrib.clippath);

            skip(fin);

            if (fscanf(fin, "%u", &ncontours) != 1 || ncontours < 1)
                throw std::runtime_error("expected number of contours");

            // try to read opacity
            skip(fin);

            attrib.opacity = 1;
            int ignore = fscanf(fin, "a %f", &attrib.opacity); (void) ignore;

            attribs.push_back(attrib);

            std::pair<AttribSet::iterator, bool> ins
                = unique_attribs.insert({&attribs.back(),attribs.size()-1});

            // new attribute?
            if(ins.second)
            {
                attribs.back().is_opaque =
                    ( attribs.back().fill.mode == COLORING_SOLID ) &&
                    ( attribs.back().fill.solid.a == 1.0f ) &&
                    ( attribs.back().fill.opacity == 1.0f ) &&
                    ( attribs.back().opacity == 1.0f );
            }
            else
                attribs.pop_back(); // rollback

            path.attrib = ins.first->idx;

            // read contours
            std::vector<Contour> &contours = paths[a].contours;

            for (unsigned int c = 0; c < ncontours; c++)
            {
                try
                {
                    contours.emplace_back(parse_contour(fin));
                }
                catch(...)
                {
                    throw_with_nested(std::runtime_error(format("error parsing contour ", c)));
                }
            } // for each contour
        }
        catch(...)
        {
            throw_with_nested(std::runtime_error(format("error parsing path ", a)));
        }
    } // for each path

    return std::move(fd);
}

FontDump load(const char *name)
{
    FILE *fin = fopen(name, "r");
    if (!fin)
        throw std::runtime_error(format(name, ": ", strerror(errno)));
    try
    {
        FontDump fd = load(fin);

        fclose(fin);

        return std::move(fd);
    }
    catch(...)
    {
        fclose(fin);
        throw;
    }
}

namespace
{

void save_linear_attr(FILE *fout, const Linear &linear)
{
    fprintf(fout, "L %g %g %g %g\n", linear.p1.x, linear.p1.y,
        linear.p2.x, linear.p2.y);
}

void save_radial_attr(FILE *fout, const Radial &radial)
{
    fprintf(fout, "R %g %g %g %g %g\n", radial.c.x, radial.c.y,
                radial.r, radial.f.x, radial.f.y);
}

void save_grad_attr(FILE *fout, const Gradient &g)
{
    const Transform &t = g.transform;

    // read transform
    fprintf(fout, "%g %g %g %g %g %g\n", t.a, t.b, t.tx, t.c, t.d, t.ty);
    switch (g.type)
    {
    case GRADIENT_LINEAR:
        save_linear_attr(fout, g.linear);
        break;
    case GRADIENT_RADIAL:
        save_radial_attr(fout, g.radial);
        break;
    }
}

} // local namespace

void store(const FontDump &fd, FILE *fout)
{
    fprintf(fout, "fodu\n");
    fprintf(fout, "b %g %g %g %g\n", fd.bbox.min.x, fd.bbox.max.x,
                fd.bbox.min.y, fd.bbox.max.y);

    const std::vector<Ramp> &ramps = fd.ramps;
    if(ramps.size() > 0)
    {
        fprintf(fout, "r %u\n", (unsigned) ramps.size());
        for (unsigned r = 0; r < ramps.size(); r++)
        {
            const std::vector<Stop> &stops = ramps[r];
            fprintf(fout, "%u\n", (unsigned) stops.size());
            for (unsigned s = 0; s < stops.size(); s++)
            {
                const Stop &stop = stops[s];
                fprintf(fout, "%g %.3g %.3g %.3g a %.3g\n",
                        stop.offset, stop.color.r, stop.color.g,
                        stop.color.b, stop.color.a);
            }
        }
    }

    // save paths
    const std::vector<Path> &paths = fd.paths;
    const std::vector<Attrib> &attribs = fd.attribs;

    fprintf(fout, "p %d\n", (int) paths.size());
    for (unsigned a = 0; a < paths.size(); a++)
    {
        const Path &path = paths[a];
        const Attrib &attrib = attribs[path.attrib];

        switch (attrib.fill.mode)
        {
        case COLORING_GRADIENT:
            {
                const Gradient &gradient = fd.gradients[attrib.fill.gradient];
                fprintf(fout, "F %d %d a %g\n",
                        gradient.ramp,
                        gradient.spread,
                        attrib.fill.opacity);
                save_grad_attr(fout, gradient);
                break;
            }
        case COLORING_SOLID:
            fprintf(fout, "f %g %g %g",
                attrib.fill.solid.r, attrib.fill.solid.g,
                attrib.fill.solid.b);
            if (attrib.fill.opacity != 1)
                fprintf(fout, " a %g\n", attrib.fill.opacity);
            else
                fprintf(fout, "\n");
            break;
        default:
            break;
        }

        switch (attrib.stroke.mode)
        {
        case COLORING_GRADIENT:
            {
                const Gradient &gradient = fd.gradients[attrib.stroke.gradient];
                fprintf(fout, "S %d %d %g a %g\n",
                    gradient.ramp,
                    gradient.spread,
                    attrib.stroke.width,
                    attrib.stroke.opacity);
                save_grad_attr(fout, gradient);
                break;
            }
        case COLORING_SOLID:
            fprintf(fout, "s %g %g %g %g",
                attrib.stroke.solid.r, attrib.stroke.solid.g,
                attrib.stroke.solid.b, attrib.stroke.width);
            if (attrib.stroke.opacity != 1)
                fprintf(fout, " a %g\n", attrib.stroke.opacity);
            else
                fprintf(fout, "\n");
            break;
        default:
            break;
        }

        const std::vector<Contour> &contours = path.contours;
        fprintf(fout, "%d\n", (int) contours.size());

        // save path opacity
        if (attrib.opacity != 1)
            fprintf(fout, "a %g\n", attrib.opacity);

        // save contours
        for (unsigned int c = 0; c < contours.size(); c++)
        {
            const std::vector<Segment> &segments = contours[c].segments;
            const std::vector<Point> &points = contours[c].points;

            fprintf(fout, "%d\n", (int) points.size());

            unsigned p = 0;
            for (unsigned s = 0; s < segments.size(); s++)
            {
                switch (segments[s])
                {
                case SEGMENT_POINT:
                case SEGMENT_LINEAR:
                    break;
                case SEGMENT_QUADRATIC:
                    fprintf(fout, "q %g %g\n", points[p].x, points[p].y);
                    p++;
                    break;
                case SEGMENT_CUBIC:
                    fprintf(fout, "c %g %g\n", points[p].x, points[p].y);
                    p++;
                    fprintf(fout, "c %g %g\n", points[p].x, points[p].y);
                    p++;
                    break;
                case SEGMENT_ARC:
                    fprintf(fout, "a %g %g\n", points[p].x, points[p].y);
                    p++;
                    fprintf(fout, "a %g %g\n", points[p].x, points[p].y);
                    p++;
                    break;
                default:
                    throw std::runtime_error("store() not implemented for types UP, DOWN, PLUS and MINUS");
                }
                fprintf(fout, "i %g %g\n", points[p].x, points[p].y);
                p++;
            }
        }
    }
}

void store(const FontDump &fd, const char *name)
{
    FILE *fout = fopen(name, "w");
    if (!fout)
        throw std::runtime_error(format(name, ": ", strerror(errno)));

    try
    {
        fd::store(fd, fout);
        fclose(fout);
    }
    catch(...)
    {
        fclose(fout);
        throw;
    }
}

void flip_inplace(FontDump fd)
{
    for (unsigned a = 0; a < fd.paths.size(); a++)
    {
        std::vector<Contour> &contours = fd.paths[a].contours;

        for (unsigned i = 0; i < contours.size(); i++)
        {
            std::vector<Point> &points = contours[i].points;
            for (unsigned j = 0; j < points.size(); j++)
                points[j].y = fd.bbox.max.y - points[j].y;
        }
    }

    Transform t;
    t.a = 1; t.b = 0; t.tx = 0;
    t.c = 0; t.d = -1; t.ty = fd.bbox.max.y;

    for (unsigned a = 0; a < fd.attribs.size(); a++)
    {
        Attrib &attrib = fd.attribs[a];
        if (attrib.fill.mode == COLORING_GRADIENT)
        {
            Gradient &gradient = fd.gradients[attrib.fill.gradient];
            gradient.transform = compose(t, gradient.transform);
        }
        if (attrib.stroke.mode == COLORING_GRADIENT)
        {
            Gradient &gradient = fd.gradients[attrib.stroke.gradient];
            gradient.transform = compose(t, gradient.transform);
        }
    }
    fd.bbox.max.y = fd.bbox.max.y - fd.bbox.min.y;
    fd.bbox.min.y = 0;
}

void close_contours(FontDump &fd)
{
    for(Path &path: fd.paths)
    {
        for(Contour &contour: path.contours)
        {
            if (contour.points.front() != contour.points.back())
            {
                contour.points.push_back(contour.points.front());
                contour.segments.push_back(SEGMENT_LINEAR);
            }
            contour.closed = true;
        }
    }
}

} // namespace fd
