#ifndef FONTDUMP_H
#define FONTDUMP_H

#include <cstdio>
#include <vector>
#include <cstring>
#include <cfloat>

#include <math/point.h>
#include <math/box.h>
#include <math/transform.h>
#include <image/color.h>
#include <boost/functional/hash.hpp>

namespace fd
{

using math::Point;
using math::Box;
using math::Transform;
using image::Color;

using boost::hash_value;
using boost::hash_combine;

struct Stop
{
    Color color;
    float offset;
};

typedef std::vector<Stop> Ramp;

struct Linear
{
    Point p1, p2;
};

struct Radial
{
    Point c, f;
    float r;
};

// in c++11 we should be using "enum class"
// but CUDA doesn't support it.
enum Spread
{
    SPREAD_PAD = 0,
    SPREAD_REFLECT = 1,
    SPREAD_REPEAT = 2
};

enum GradientType
{
    GRADIENT_LINEAR,
    GRADIENT_RADIAL
};


struct Gradient
{
    Gradient() : type(GRADIENT_LINEAR) {}

    GradientType type;

    // these two could be an in union in c++11
    Linear linear;
    Radial radial;

    unsigned ramp; // ramp index
    Spread spread;
    Transform transform;
};


enum Coloring
{
    COLORING_NONE,
    COLORING_SOLID,
    COLORING_GRADIENT
};

inline
size_t hash_value(const Coloring &c)
{
    return c;
}

struct Fill
{
    Coloring mode;
    Color solid;
    unsigned gradient; // gradient index
    float opacity;

    bool operator==(const Fill &that) const
    {
        if(mode == that.mode)
        {
            if(mode == COLORING_NONE)
                return true;
            else
            {
                return (mode==COLORING_SOLID ? solid == that.solid
                                         : gradient == that.gradient)
                        && opacity == that.opacity;
            }
        }
        else
            return false;
    }
};

inline
size_t hash_value(const Fill &f)
{
    size_t h = hash_value(f.mode);
    switch(f.mode)
    {
    case COLORING_NONE:
        return h;
    case COLORING_SOLID:
        hash_combine(h, f.solid);
        break;
    case COLORING_GRADIENT:
        hash_combine(h, f.gradient);
        break;
    }

    hash_combine(h, f.opacity);

    return h;
}

struct ClipPath
{
    unsigned depth;

    bool operator==(const ClipPath &that) const
    {
        return depth == that.depth;
    }

};

inline
size_t hash_value(const ClipPath &cp)
{
    return hash_value(cp.depth);
}

struct Stroke
{
    Coloring mode;
    Color solid;
    unsigned gradient; // gradient index
    float width;
    float opacity;

    bool operator==(const Stroke &that) const
    {
        if(mode == that.mode)
        {
            if(mode == COLORING_NONE)
                return true;
            else
            {
                return (mode==COLORING_SOLID ? solid == that.solid
                                             : gradient == that.gradient)
                    && width == that.width
                    && opacity == that.opacity;
            }
        }
        else
            return false;
    }
};

inline
size_t hash_value(const Stroke &s)
{
    size_t h = hash_value(s.mode);
    switch(s.mode)
    {
    case COLORING_NONE:
        return h;
    case COLORING_SOLID:
        hash_combine(h, s.solid);
        break;
    case COLORING_GRADIENT:
        hash_combine(h, s.gradient);
        break;
    }
    hash_combine(h, s.width);
    hash_combine(h, s.opacity);
    return h;
}

enum Segment
{
    SEGMENT_POINT,
    SEGMENT_LINEAR,
    SEGMENT_QUADRATIC,
    SEGMENT_CUBIC,
    SEGMENT_ARC
};

inline unsigned size(Segment type)
{
    switch( type )
    {
    case SEGMENT_POINT:
    case SEGMENT_LINEAR:
        return 1;
    case SEGMENT_QUADRATIC:
        return 2;
    case SEGMENT_CUBIC:
    case SEGMENT_ARC:
        return 3;
    default:
        return 0;
    }
}

struct Attrib
{
    Fill fill;
    Stroke stroke;
    ClipPath clippath;
    float opacity;
    bool is_opaque;

    bool operator==(const Attrib &that) const
    {
        return fill == that.fill
            && stroke == that.stroke
            && clippath == that.clippath
            && opacity == that.opacity;
    }
};

inline
size_t hash_value(const Attrib &a)
{
    size_t h = hash_value(a.fill);
    hash_combine(h, a.stroke);
    hash_combine(h, a.clippath);
    hash_combine(h, a.opacity);
    return  h;
}

struct Contour
{
    std::vector<Point> points;
    std::vector<Segment> segments;
    Box box;
    bool closed;
};

struct Path
{
    unsigned id;
    unsigned attrib;
    std::vector<Contour> contours;
};

inline int is_filled(const Attrib &attrib)
{
    return attrib.fill.mode != COLORING_NONE;
}

inline int is_stroked(const Attrib &attrib)
{
    return attrib.stroke.mode != COLORING_NONE;
}

inline int is_solid(const Stroke &stroke)
{
    return stroke.mode == COLORING_SOLID;
}

inline int is_solid(const Fill &fill)
{
    return fill.mode == COLORING_SOLID;
}

struct FontDump
{
    std::vector<Ramp> ramps;
    std::vector<Gradient> gradients;
    std::vector<Attrib> attribs;
    std::vector<Path> paths;
    Box bbox;
};

FontDump load(const char *name);
FontDump load(FILE *fin);
void store(const FontDump &fd, FILE *fout);
void store(const FontDump &fd, const char *name);

void flip_inplace(FontDump &fd);
void close_contours(FontDump &fd);

} // namespace fd

#endif
