#ifndef GPU_ARRAY_H
#define GPU_ARRAY_H

#include <util/util.h> // for enable_if, is_convertible
#include "hostdev.h"

namespace gpu
{

template <class T>
class array_iterator
{
public:
    HOSTDEV array_iterator() : m_ptr(NULL) {}
    HOSTDEV array_iterator(T *ptr) : m_ptr(ptr) {}

    template <class U>
    HOSTDEV array_iterator(U *ptr, typename rod::enable_if<rod::is_convertible<U*,T*>::value>::type * = NULL)
        : m_ptr(ptr)
    {
    }

    typedef T value_type;
    typedef T &reference_type;
    typedef T *pointer_type;

    HOSTDEV operator T*() const { return m_ptr; }
    HOSTDEV T &operator*() { return *m_ptr; }
    HOSTDEV const T &operator*() const { return *m_ptr; }

    HOSTDEV array_iterator &operator++()
    {
        ++m_ptr;
        return *this;
    }

    HOSTDEV array_iterator operator++(int)
    {
        array_iterator it = *this;
        operator++();
        return it;
    }

    HOSTDEV array_iterator &operator--()
    {
        --m_ptr;
        return *this;
    }

    HOSTDEV array_iterator operator--(int)
    {
        array_iterator it = *this;
        operator--();
        return it;
    }

    HOSTDEV array_iterator &operator+=(int offset)
    {
        m_ptr += offset;
        return *this;
    }

    HOSTDEV array_iterator &operator-=(int offset)
    {
        return operator+=(-offset);
    }

private:
    T *m_ptr;
};

template <class T>
HOSTDEV inline array_iterator<T> operator+(array_iterator<T> it, int offset)
{
    return it += offset;
}

template <class T>
HOSTDEV inline array_iterator<T> operator+(int offset, array_iterator<T> it)
{
    return it += offset;
}

template <class T>
HOSTDEV inline array_iterator<T> operator-(array_iterator<T> it, int offset)
{
    return it -= offset;
}

template <class T, int N>
struct array
{

    // public so that it can be initialized with {...}
    T data[N];

    typedef array_iterator<T> iterator;
    typedef array_iterator<const T> const_iterator;

    HOSTDEV const T &operator[](int i) const { return data[i]; }

    HOSTDEV T &operator[](int i) { return data[i]; }

    HOSTDEV size_t size() const { return N; }

    HOSTDEV iterator begin() { return &data[0]; }
    HOSTDEV iterator end() { return &data[N-1]; }

    HOSTDEV const_iterator begin() const { return &data[0]; }
    HOSTDEV const_iterator end() const { return &data[N-1]; }

    HOSTDEV const_iterator cbegin() const { return &data[0]; }
    HOSTDEV const_iterator cend() const { return &data[N-1]; }
};

// for mutable * -> const * conversion
template <class T, int N>
class array<const T *, N>
{
    const T *m_data[N];
public:
    typedef T **iterator;
    typedef const T **const_iterator;

    HOSTDEV
    array()
    {
    }

    HOSTDEV
    array(const array &that)
    {
        for(int i=0; i<N; ++i)
            m_data[i] = that[i];
    }

    HOSTDEV
    array(const array<T*,N> &that)
    {
        for(int i=0; i<N; ++i)
            m_data[i] = that[i];
    }

    HOSTDEV const T *&operator[](int i) { return m_data[i]; }
    HOSTDEV const T *const& operator[](int i) const { return m_data[i]; }

    HOSTDEV size_t size() const { return N; }

    HOSTDEV iterator begin() { return &m_data[0]; }
    HOSTDEV iterator end() { return &m_data[N-1]; }

    HOSTDEV const_iterator begin() const { return &m_data[0]; }
    HOSTDEV const_iterator end() const { return &m_data[N-1]; }

    HOSTDEV const_iterator cbegin() const { return &m_data[0]; }
    HOSTDEV const_iterator cend() const { return &m_data[N-1]; }
};

} // namespace gpu

#endif
