#ifndef GPU_BOX_H
#define GPU_BOX_H

#include <cfloat>
#include <gpu/cutil.h>
#include "vector.h"
#include "hostdev.h"

namespace gpu
{

struct Box
{
    HOSTDEV Box() {}
    HOSTDEV Box(float2 _min, float2 _max) : min(_min), max(_max) {}

    float2 min, max;

    HOSTDEV bool contains(float2 sample) const
    {
        return min.x <= sample.x && sample.x <= max.x &&
               min.y <= sample.y && sample.y <= max.y;
    }

    HOSTDEV void merge(float2 sample)
    {
        min.x = ::min(min.x, sample.x);
        max.x = ::max(max.x, sample.x);
        min.y = ::min(min.y, sample.y);
        max.y = ::max(max.y, sample.y);
    }

    HOSTDEV void reset()
    {
        min.x = min.y = FLT_MAX;
        max.x = max.y = -FLT_MAX;
    }

    HOSTDEV float2 mp() const
    {
        return (max+min)*0.5;
    }

    HOSTDEV int hit(const float2 &ray) const
    {
        // below bbox?
        if(ray.y < min.y)
            return 0;

        // above bbox?
        if(ray.y >= max.y)
            return 0;

        // after bbox?
        if(ray.x >= max.x)
            return 0;

        // before bbox?
        if(ray.x < min.x)
            return 1;

        return -1;
    }
};

struct BoxesPtr
{
    BoxesPtr() : size(0), min(NULL), max(NULL) {}

    unsigned size;
    float2 *min, *max;
};

struct const_BoxesPtr
{
    const_BoxesPtr() : size(0), min(NULL), max(NULL) {}
    const_BoxesPtr(const BoxesPtr &b)
        : size(b.size), min(b.min), max(b.max) {}

    unsigned size;
    const float2 *min, *max;
};

struct Boxes
{
    gpu::vector<float2> min, max;

    size_t size() const { return min.size(); }
    void clear()
    {
        min.clear();
        max.clear();
    }

    void resize(size_t s)
    {
        min.resize(s);
        max.resize(s);
    }

    BoxesPtr operator&()
    {
        BoxesPtr b;
        b.size = size();
        b.min = min.data();
        b.max = max.data();
        return b;
    }

    const_BoxesPtr operator&() const
    {
        const_BoxesPtr b;
        b.size = size();
        b.min = min.data();
        b.max = max.data();
        return b;
    }
};

} // namespace gpu

#endif
