#ifndef GPU_CUB_H
#define GPU_CUB_H

#include <cub/device/device_scan.cuh>
#include "vector.h"
#include "op.h"

namespace gpu
{

class CUB
{
public:

    template <class T, class OP>
    void inclusive_scan(const T *in, size_t size,
                        OP &op, T *out, cudaStream_t stream)
    {
        size_t bytes_needed;
        cub::DeviceScan::InclusiveScan(NULL, bytes_needed,
                                      in, out, op, size,
                                      stream);

        m_storage.resize(bytes_needed);

        cub::DeviceScan::InclusiveScan(&m_storage, bytes_needed,
                                      in, out, op, size,
                                      stream);
    }

    template <class T, class OP, class ID>
    void exclusive_scan(const T *in, size_t size,
                        OP &op, const ID &id, T *out, cudaStream_t stream)
    {
        size_t bytes_needed;
        cub::DeviceScan::ExclusiveScan(NULL, bytes_needed,
                                      in, out, op, id, size,
                                      stream);

        m_storage.resize(bytes_needed);

        cub::DeviceScan::ExclusiveScan(&m_storage, bytes_needed,
                                      in, out, op, id, size,
                                      stream);
    }

    template <class T, class OP>
    void inclusive_scan(const T *in, size_t size,
                        T *out, cudaStream_t stream)
    {
        OP op;
        inclusive_scan(in, size, op, out, stream);
    }

    template <class T, class OP, class ID>
    void exclusive_scan(const T *in, size_t size,
                        const ID &id, T *out, cudaStream_t stream)
    {
        OP op;
        exclusive_scan(in, size, op, out, stream);
    }

    template <class T>
    void inclusive_scan(const T *in, size_t size, T *out, Operation op=OP_ADD,
                        cudaStream_t stream=0)

    {
        size_t bytes_needed;
        switch(op)
        {
        case OP_ADD:
            cub::DeviceScan::InclusiveSum(NULL, bytes_needed,
                                          in, out, size,
                                          stream);

            m_storage.resize(bytes_needed);

            cub::DeviceScan::InclusiveSum(&m_storage, bytes_needed,
                                          in, out, size,
                                          stream);
            break;
        case OP_MAX:
            {
                cub::Max op;
                inclusive_scan(in, size, op, out, stream);
            }
            break;
        case OP_MIN:
            {
                cub::Min op;
                inclusive_scan(in, size, op, out, stream);
            }
            break;
        }
    }

    template <class T>
    void exclusive_scan(const T *in, size_t size, T *out, Operation op=OP_ADD,
                        cudaStream_t stream=0)
    {
        size_t bytes_needed;
        switch(op)
        {
        case OP_ADD:
            cub::DeviceScan::ExclusiveSum(NULL, bytes_needed,
                                          in, out, size,
                                          stream);

            m_storage.resize(bytes_needed);

            cub::DeviceScan::ExclusiveSum(&m_storage, bytes_needed,
                                          in, out, size,
                                          stream);
            break;
        case OP_MAX:
            {
                cub::Max op;
                exclusive_scan(in, size, op, -std::numeric_limits<T>::max(),
                               out, stream);
            }
            break;
        case OP_MIN:
            {
                cub::Min op;
                exclusive_scan(in, size, op, std::numeric_limits<T>::max(),
                               out, stream);
            }
            break;
        }
    }

    template <class T>
    void inclusive_scan(const vector<T> &in, vector<T> &out,Operation op=OP_ADD,
                        cudaStream_t stream=0)
    {
        out.resize(in.size());
        inclusive_scan(in.data(), in.size(), out.data(), op, stream);
    }

    template <class T>
    void exclusive_scan(const vector<T> &in, vector<T> &out,Operation op=OP_ADD,
                        cudaStream_t stream=0)
    {
        out.resize(in.size());
        exclusive_scan(in.data(), in.size(), out.data(), op, stream);
    }

private:
    vector<unsigned char> m_storage;
};

}

#endif
