#include <cudpp.h>
#include <stdexcept>
#include "vector.h"
#include "cudpp.h"

namespace gpu
{

class Plan
{
public:
    Plan();
    ~Plan();

    void update(const CUDPPConfiguration &cfg, unsigned nelements)
    {
        if(cfg.options != m_cfg.options || cfg.op != m_cfg.op
           || cfg.datatype != m_cfg.datatype || cfg.algorithm != m_cfg.algorithm
           || nelements > m_size)
        {
            CUDPPHandle plan;
            CUDPPResult res = cudppPlan(m_cudpp, &plan, cfg, nelements, 1, 0);

            if(res != CUDPP_SUCCESS)
                throw std::runtime_error("Error creating plan");

            if(m_plan)
                cudppDestroyPlan(m_plan);
            m_plan = plan;
            m_cfg = cfg;
            m_size = nelements;
        }
    }

    CUDPPHandle get() { return m_plan; }

private:
    CUDPPHandle m_cudpp, m_plan;
    CUDPPConfiguration m_cfg;
    unsigned m_size;

};

Plan::Plan()
{
    // CUDPPHandle cudpp; // was unused? 
    cudppCreate(&m_cudpp);

    m_plan = 0;
    m_cfg.algorithm = CUDPP_ALGORITHM_INVALID;
    m_cfg.op = CUDPP_OPERATOR_INVALID;
    m_cfg.options = 0;
    m_cfg.datatype = CUDPP_DATATYPE_INVALID;
    m_size = 0;
}

Plan::~Plan()
{
    if(m_plan)
        cudppDestroyPlan(m_plan);
    cudppDestroy(m_cudpp);
}

Cudpp::Cudpp()
{
    m_plan = new Plan();
}

Cudpp::~Cudpp()
{
    delete m_plan;
}

// scan -------------------------------------------------------------------
// unsigned
void Cudpp::inclusive_scan(const vector<unsigned> &in, vector<unsigned> &out,
                           Operation op)
{
    out.resize(in.size());
    inclusive_scan(in.data(), in.size(), out.data(), op);
}
void Cudpp::exclusive_scan(const vector<unsigned> &in, vector<unsigned> &out,
                           Operation op)
{
    out.resize(in.size());
    exclusive_scan(in.data(), in.size(), out.data(), op);
}

void Cudpp::segmented_inclusive_scan(const vector<unsigned> &in,
                         const vector<unsigned> &flags,
                         vector<unsigned> &out,
                         Operation op)
{
    if(flags.size() != in.size())
        throw std::runtime_error("flags size must be equal to input size");

    out.resize(in.size());
    segmented_inclusive_scan(in.data(), in.size(), flags.data(), out.data(), op);
}

void Cudpp::segmented_exclusive_scan(const vector<unsigned> &in,
                               const vector<unsigned> &flags,
                               vector<unsigned> &out, Operation op)
{
    if(flags.size() != in.size())
        throw std::runtime_error("flags size must be equal to input size");

    out.resize(in.size());
    segmented_exclusive_scan(in.data(), in.size(), flags.data(), out.data(), op);
}

void Cudpp::reverse_inclusive_scan(const vector<unsigned> &in, vector<unsigned> &out,
                           Operation op)
{
    out.resize(in.size());
    reverse_inclusive_scan(in.data(), in.size(), out.data(), op);
}
void Cudpp::reverse_exclusive_scan(const vector<unsigned> &in, vector<unsigned> &out,
                           Operation op)
{
    out.resize(in.size());
    reverse_exclusive_scan(in.data(), in.size(), out.data(), op);
}

void Cudpp::reverse_segmented_inclusive_scan(const vector<unsigned> &in,
                         const vector<unsigned> &flags,
                         vector<unsigned> &out,
                         Operation op)
{
    if(flags.size() != in.size())
        throw std::runtime_error("flags size must be equal to input size");

    out.resize(in.size());
    reverse_segmented_inclusive_scan(in.data(), in.size(), flags.data(), out.data(), op);
}

void Cudpp::reverse_segmented_exclusive_scan(const vector<unsigned> &in,
                               const vector<unsigned> &flags,
                               vector<unsigned> &out, Operation op)
{
    if(flags.size() != in.size())
        throw std::runtime_error("flags size must be equal to input size");

    out.resize(in.size());
    reverse_segmented_exclusive_scan(in.data(), in.size(), flags.data(), out.data(), op);
}

// int
void Cudpp::inclusive_scan(const vector<int> &in, vector<int> &out,
                           Operation op)
{
    out.resize(in.size());
    inclusive_scan(in.data(), in.size(), out.data(), op);
}
void Cudpp::exclusive_scan(const vector<int> &in, vector<int> &out,
                           Operation op)
{
    out.resize(in.size());
    exclusive_scan(in.data(), in.size(), out.data(), op);
}

void Cudpp::segmented_inclusive_scan(const vector<int> &in,
                         const vector<unsigned> &flags,
                         vector<int> &out,
                         Operation op)
{
    if(flags.size() != in.size())
        throw std::runtime_error("flags size must be equal to input size");

    out.resize(in.size());
    segmented_inclusive_scan(in.data(), in.size(), flags.data(), out.data(), op);
}

void Cudpp::segmented_exclusive_scan(const vector<int> &in,
                               const vector<unsigned> &flags,
                               vector<int> &out, Operation op)
{
    if(flags.size() != in.size())
        throw std::runtime_error("flags size must be equal to input size");

    out.resize(in.size());
    segmented_exclusive_scan(in.data(), in.size(), flags.data(), out.data(), op);
}

void Cudpp::reverse_inclusive_scan(const vector<int> &in, vector<int> &out,
                           Operation op)
{
    out.resize(in.size());
    reverse_inclusive_scan(in.data(), in.size(), out.data(), op);
}
void Cudpp::reverse_exclusive_scan(const vector<int> &in, vector<int> &out,
                           Operation op)
{
    out.resize(in.size());
    reverse_exclusive_scan(in.data(), in.size(), out.data(), op);
}

void Cudpp::reverse_segmented_inclusive_scan(const vector<int> &in,
                         const vector<unsigned> &flags,
                         vector<int> &out,
                         Operation op)
{
    if(flags.size() != in.size())
        throw std::runtime_error("flags size must be equal to input size");

    out.resize(in.size());
    reverse_segmented_inclusive_scan(in.data(), in.size(), flags.data(), out.data(), op);
}

void Cudpp::reverse_segmented_exclusive_scan(const vector<int> &in,
                               const vector<unsigned> &flags,
                               vector<int> &out, Operation op)
{
    if(flags.size() != in.size())
        throw std::runtime_error("flags size must be equal to input size");

    out.resize(in.size());
    reverse_segmented_exclusive_scan(in.data(), in.size(), flags.data(), out.data(), op);
}


namespace
{
    CUDPPOperator conv_op(Operation op)
    {
        switch(op)
        {
        case OP_ADD:
            return CUDPP_ADD;
        case OP_MULT:
            return CUDPP_MULTIPLY;
        case OP_MIN:
            return CUDPP_MIN;
        case OP_MAX:
            return CUDPP_MAX;
        default:
            return CUDPP_OPERATOR_INVALID;
        }
    }
}

// unsigned
void Cudpp::inclusive_scan(const unsigned *in, size_t size, unsigned *out,
                      Operation op, bool reverse)
{
    CUDPPConfiguration cfg;
    cfg.op = conv_op(op);
    cfg.datatype = CUDPP_UINT;
    cfg.algorithm = CUDPP_SCAN;
    cfg.options = CUDPP_OPTION_INCLUSIVE;

    if(reverse)
        cfg.options |= CUDPP_OPTION_BACKWARD;
    else
        cfg.options |= CUDPP_OPTION_FORWARD;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppScan(m_plan->get(), out, in, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing scan");
}

void Cudpp::exclusive_scan(const unsigned *in, size_t size, unsigned *out,
                      Operation op, bool reverse)
{
    CUDPPConfiguration cfg;
    cfg.op = conv_op(op);
    cfg.datatype = CUDPP_UINT;
    cfg.algorithm = CUDPP_SCAN;
    cfg.options = CUDPP_OPTION_EXCLUSIVE;

    if(reverse)
        cfg.options |= CUDPP_OPTION_BACKWARD;
    else
        cfg.options |= CUDPP_OPTION_FORWARD;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppScan(m_plan->get(), out, in, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing scan");
}

void Cudpp::segmented_exclusive_scan(const unsigned *in, size_t size,
                               const unsigned *flags, unsigned *out,
                               Operation op, bool reverse)
{
    CUDPPConfiguration cfg;
    cfg.op = conv_op(op);
    cfg.datatype = CUDPP_UINT;
    cfg.algorithm = CUDPP_SEGMENTED_SCAN;
    cfg.options = CUDPP_OPTION_EXCLUSIVE;

    if(reverse)
        cfg.options |= CUDPP_OPTION_BACKWARD;
    else
        cfg.options |= CUDPP_OPTION_FORWARD;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppSegmentedScan(m_plan->get(), out, in, flags, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing scan");
}

void Cudpp::segmented_inclusive_scan(const unsigned *in, size_t size,
                               const unsigned *flags, unsigned *out,
                               Operation op, bool reverse)
{
    CUDPPConfiguration cfg;
    cfg.op = conv_op(op);
    cfg.datatype = CUDPP_UINT;
    cfg.algorithm = CUDPP_SEGMENTED_SCAN;
    cfg.options = CUDPP_OPTION_INCLUSIVE;

    if(reverse)
        cfg.options |= CUDPP_OPTION_BACKWARD;
    else
        cfg.options |= CUDPP_OPTION_FORWARD;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppSegmentedScan(m_plan->get(), out, in, flags, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing scan");
}

// int
void Cudpp::inclusive_scan(const int *in, size_t size, int *out,
                      Operation op, bool reverse)
{
    CUDPPConfiguration cfg;
    cfg.op = conv_op(op);
    cfg.datatype = CUDPP_INT;
    cfg.algorithm = CUDPP_SCAN;
    cfg.options = CUDPP_OPTION_INCLUSIVE;

    if(reverse)
        cfg.options |= CUDPP_OPTION_BACKWARD;
    else
        cfg.options |= CUDPP_OPTION_FORWARD;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppScan(m_plan->get(), out, in, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing scan");
}

void Cudpp::exclusive_scan(const int *in, size_t size, int *out,
                      Operation op, bool reverse)
{
    CUDPPConfiguration cfg;
    cfg.op = conv_op(op);
    cfg.datatype = CUDPP_INT;
    cfg.algorithm = CUDPP_SCAN;
    cfg.options = CUDPP_OPTION_EXCLUSIVE;

    if(reverse)
        cfg.options |= CUDPP_OPTION_BACKWARD;
    else
        cfg.options |= CUDPP_OPTION_FORWARD;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppScan(m_plan->get(), out, in, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing scan");
}

void Cudpp::segmented_exclusive_scan(const int *in, size_t size,
                               const unsigned *flags, int *out,
                               Operation op, bool reverse)
{
    CUDPPConfiguration cfg;
    cfg.op = conv_op(op);
    cfg.datatype = CUDPP_INT;
    cfg.algorithm = CUDPP_SEGMENTED_SCAN;
    cfg.options = CUDPP_OPTION_EXCLUSIVE;

    if(reverse)
        cfg.options |= CUDPP_OPTION_BACKWARD;
    else
        cfg.options |= CUDPP_OPTION_FORWARD;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppSegmentedScan(m_plan->get(), out, in, flags, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing scan");
}

void Cudpp::segmented_inclusive_scan(const int *in, size_t size,
                               const unsigned *flags, int *out,
                               Operation op, bool reverse)
{
    CUDPPConfiguration cfg;
    cfg.op = conv_op(op);
    cfg.datatype = CUDPP_INT;
    cfg.algorithm = CUDPP_SEGMENTED_SCAN;
    cfg.options = CUDPP_OPTION_INCLUSIVE;

    if(reverse)
        cfg.options |= CUDPP_OPTION_BACKWARD;
    else
        cfg.options |= CUDPP_OPTION_FORWARD;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppSegmentedScan(m_plan->get(), out, in, flags, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing scan");
}

// radix sort -------------------------------------------------------------------

// unsigned
void Cudpp::radix_sort(vector<unsigned> &keys)
{
    radix_sort(keys.data(), keys.size());
}

void Cudpp::radix_sort(vector<unsigned> &keys, vector<unsigned> &values)
{
    radix_sort(keys.data(), keys.size(), values.data());
}

void Cudpp::radix_sort(unsigned *keys, size_t size, unsigned *values)
{
    CUDPPConfiguration cfg;
    cfg.op = CUDPP_OPERATOR_INVALID; // ?? just to initialize?
    cfg.datatype = CUDPP_UINT;
    cfg.algorithm = CUDPP_SORT_RADIX;
    cfg.options = CUDPP_OPTION_FORWARD;
    if(values == NULL)
        cfg.options |= CUDPP_OPTION_KEYS_ONLY;
    else
        cfg.options |= CUDPP_OPTION_KEY_VALUE_PAIRS;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppRadixSort(m_plan->get(), keys, values, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing radix sort");
}

// int
void Cudpp::radix_sort(vector<int> &keys)
{
    radix_sort(keys.data(), keys.size());
}

void Cudpp::radix_sort(vector<int> &keys, vector<int> &values)
{
    radix_sort(keys.data(), keys.size(), values.data());
}

void Cudpp::radix_sort(int *keys, size_t size, int *values)
{
    CUDPPConfiguration cfg;
    cfg.op = CUDPP_OPERATOR_INVALID; // ?? just to initialize?
    cfg.datatype = CUDPP_INT;
    cfg.algorithm = CUDPP_SORT_RADIX;
    cfg.options = CUDPP_OPTION_FORWARD;
    if(values == NULL)
        cfg.options |= CUDPP_OPTION_KEYS_ONLY;
    else
        cfg.options |= CUDPP_OPTION_KEY_VALUE_PAIRS;

    m_plan->update(cfg, size);

    CUDPPResult res = cudppRadixSort(m_plan->get(), keys, values, size);

    if(res != CUDPP_SUCCESS)
        throw std::runtime_error("Error performing radix sort");
}

}
