#ifndef GPU_IMAGE_H
#define GPU_IMAGE_H

#include <cassert>
#include "vector.h"
#include <util/util.h>
#include "pixel_traits.h"
#include "image_fwd.h"

namespace gpu
{

template <class T, int C>
class image
{
public:
    typedef typename pixel_traits<T,1>::texel_type texel_type;
    typedef typename rod::remove_const<texel_type>::type mutable_texel_type;

    image(texel_type *data, int width, int height, int rowstride);
    image(int width, int height, int rowstride=0);
    image();

    void reset(texel_type *data, int width, int height, int rowstride);
    void reset() { reset(NULL,0,0,0); }

    friend void set_zero(image &img) { img.m_data.fill(0); }

    void copy_to_host(mutable_texel_type *out, int rowstride=0) const;
    void copy_to_host(std::vector<mutable_texel_type> &out, int rowstride=0) const;

    void copy_from_host(const texel_type *in, int w, int h, int rs=0);
    void copy_from_host(const std::vector<mutable_texel_type> &in,
                        int width, int height, int rowstride=0);

    void resize(int width, int height, int rowstride=0);

    bool empty() const { return m_data.empty(); }

    image &operator=(const image_ptr<const texel_type, C> &img);
    image &operator=(const image &that);

    int width() const { return m_width; }
    int height() const { return m_height; }
    int rowstride() const { return m_rowstride; }
    int channelstride() const { return rowstride()*height(); }

    image_ptr<T,C> operator&();
    image_ptr<const T,C> operator&() const;

    texel_type *data() { return m_data.data(); }
    const texel_type *data() const { return m_data.data(); }

    operator texel_type*() { return m_data; }
    operator const texel_type*() const { return m_data; }

    int offset_at(int x, int y) const { return y*rowstride()+x; }
    int offset_at(int2 pos) const { return offset_at(pos.x, pos.y); }
    bool is_inside(int x, int y) const { return x < width() && y < height(); }
    bool is_inside(int2 pos) const { return is_inside(pos.x, pos.y); }

    image_ptr<T,1> operator[](int i);

    image_ptr<const T, 1> operator[](int i) const;

    friend void swap(image &orig, image &dest)
    {
        using std::swap;

        swap(orig.m_data, dest.m_data);
        swap(orig.m_width, dest.m_width);
        swap(orig.m_height, dest.m_height);
        swap(orig.m_rowstride, dest.m_rowstride);
    }

private:
    vector<texel_type> m_data;
    int m_width, m_height, m_rowstride;
};

template <class T, int C>
class image_ptr
{
public:
    // these should be defined in image.hpp but nvcc (cuda 4.2) doesn't
    // like outside definition of nested template class of a template class
    template <int D, class EN=void>
    class pixel_proxy/*{{{*/
    {
        image_ptr &m_img;

    public:
        HOSTDEV
        pixel_proxy(image_ptr &img): m_img(img) {}

        typedef typename pixel_traits<T,C>::pixel_type value_type;

        template <class V>
        HOSTDEV
        pixel_proxy &operator=(const V &p)
        {
            pixel_traits<texel_type,D>::assign(m_img.m_data,
                                               m_img.channelstride(),
                                           pixel_traits<T,C>::make_pixel(p));
            return *this;
        }

        HOSTDEV
        pixel_proxy &operator+=(const value_type &v)
        {
            value_type temp;
            pixel_traits<texel_type,D>::assign(temp, m_img.m_data, m_img.channelstride());
            temp += v;
            pixel_traits<texel_type,D>::assign(m_img.m_data, m_img.channelstride(), temp);
            return *this;
        }

        HOSTDEV
        pixel_proxy &operator-=(const value_type &v)
        {
            value_type temp;
            pixel_traits<texel_type,D>::assign(temp, m_img.m_data, m_img.channelstride());
            temp -= v;
            pixel_traits<texel_type,D>::assign(m_img.m_data, m_img.channelstride(), temp);
            return *this;
        }

        HOSTDEV
        operator value_type() const
        {
            value_type val;
            pixel_traits<texel_type,D>::assign(val, m_img.m_data, m_img.channelstride());
            return val;
        }

        T *operator&() { return m_img.m_data; }
        const T *operator&() const { return m_img.m_data; }
    };/*}}}*/

    template <int D>
    class pixel_proxy<D, typename rod::enable_if<(D<=0 || D>=5)>::type>/*{{{*/
    {
    public:
        HOSTDEV pixel_proxy(image_ptr<T,C> &img) {}
    };/*}}}*/

    template <int D, class EN=void>
    class const_pixel_proxy/*{{{*/
    {
        const image_ptr &m_img;

    public:
        HOSTDEV const_pixel_proxy(const image_ptr &img) : m_img(img) {}

        typedef typename pixel_traits<T,C>::pixel_type value_type;

        HOSTDEV operator value_type() const
        {
            value_type val;
            pixel_traits<T,D>::assign(val, m_img.m_data, m_img.channelstride());
            return val;
        }
        const T *operator&() const { return m_img.m_data; }
    };/*}}}*/

    template <int D>
    class const_pixel_proxy<D, typename rod::enable_if<(D<=0 || D>=5)>::type>/*{{{*/
    {
    public:
        HOSTDEV const_pixel_proxy(const image_ptr &img) {}
    };/*}}}*/

    typedef typename rod::copy_const<T,typename pixel_traits<T,1>::texel_type>::type
        texel_type;
    typedef typename rod::remove_const<texel_type>::type mutable_texel_type;

    HOSTDEV image_ptr(texel_type *data, int width, int height, int rowstride);

    template <class P>
    HOSTDEV image_ptr(const image_ptr<P,C> &that,
       typename rod::enable_if<rod::is_convertible<P,T>::value>::type* =NULL);

    HOSTDEV int width() const { return m_width; }
    HOSTDEV int height() const { return m_height; }
    HOSTDEV int rowstride() const { return m_rowstride; }
    HOSTDEV int channelstride() const { return rowstride()*height(); }

    HOSTDEV bool empty() const { return rowstride()==0 || height()==0; }

    HOSTDEV int offset_at(int x, int y) const { return y*rowstride()+x; }
    HOSTDEV int offset_at(int2 pos) const { return offset_at(pos.x, pos.y); }

    HOSTDEV bool is_inside(int x, int y) const
        { return x < width() && y < height(); }
    HOSTDEV bool is_inside(int2 pos) const { return is_inside(pos.x, pos.y); }

    void copy_to_host(mutable_texel_type *out, int rowstride=0) const;

    void copy_to_host(std::vector<mutable_texel_type> &out, int rowtride=0) const;

    void copy_from_host(const texel_type *in, int w, int h, int rs=0);
    void copy_from_host(const std::vector<mutable_texel_type> &in,
                        int width, int height, int rowstride=0);

    image_ptr &operator=(image_ptr<const T,C> img);

    template <class U>
    typename rod::enable_if<rod::is_same<T,U>::value
            && !rod::is_const<T>::value,image_ptr<U,C> &>::type
        operator=(const image_ptr<U,C> &img)
    {
        return operator=(static_cast<image_ptr<const T,C> >(img));
    }

    image_ptr &operator=(const image<T,C> &img) { return *this = &img; }

    HOSTDEV image_ptr<T,1> operator[](int i);

    HOSTDEV image_ptr<const T,1> operator[](int i) const;

    HOSTDEV operator texel_type*() { return m_data; }
    HOSTDEV operator const texel_type*() const { return m_data; }

    HOSTDEV texel_type *operator&() { return m_data; }
    HOSTDEV const texel_type *operator&() const { return m_data; }

    // These two here need to be inline to cope with VC++2008 non-conformance
    HOSTDEV pixel_proxy<C> operator*()
        { return pixel_proxy<C>(*this); }

    HOSTDEV const_pixel_proxy<C> operator*() const
        { return const_pixel_proxy<C>(*this); }

    HOSTDEV image_ptr &operator++();

    HOSTDEV image_ptr operator++(int);

    HOSTDEV image_ptr &operator--();

    HOSTDEV image_ptr operator--(int);

    HOSTDEV image_ptr &operator+=(int off);

    HOSTDEV image_ptr &operator-=(int off) { return operator+=(-off); }

private:
    template <class U, int D>
    friend class image_ptr;

    int m_width, m_height, m_rowstride;
    texel_type *m_data;
};

template <class T>
void copy_to_array(cudaArray *out, image_ptr<const T> in);

template <class T>
void copy_to_array(cudaArray *out, image_ptr<T> in)
{
    copy_to_array(out, image_ptr<const T>(in));
}

template <class T>
void copy_from_array(image_ptr<T> out, const cudaArray *in);

} // namespace gpu

#include "image.hpp"


#endif
