#include "cutil.h"
#include "image_util.h"

#define USE_LAUNCH_BOUNDS 1

namespace gpu
{

const int BW = 32, // cuda block width
          BH = 6, // cuda block height
#if USE_SM20
          NB = 8;
#else
          NB = 5;
#endif

// compose -----------------------------------------------------

__global__
#if USE_LAUNCH_BOUNDS
__launch_bounds__(BW*BH, NB)
#endif
void compose(float4 *output,
             const float *r, const float *g, const float *b,
             int width, int height, int rowstride)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int x = blockIdx.x*BW+tx, y = blockIdx.y*BH+ty;

    if(x >= width || y >= height)
        return;

    int idx = y*rowstride+x;

    output[idx] = make_float4(r[idx],g[idx],b[idx],0);
}

__global__
#if USE_LAUNCH_BOUNDS
__launch_bounds__(BW*BH, NB)
#endif
void compose(uchar4 *output,
             const float *r, const float *g, const float *b,
             int width, int height, int rowstride)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int x = blockIdx.x*BW+tx, y = blockIdx.y*BH+ty;

    if(x >= width || y >= height)
        return;

    int idx = y*rowstride+x;

    float3 v = saturate(make_float3(r[idx],g[idx],b[idx]))*255.0f;
    output[idx] = make_uchar4(v.x,v.y,v.z,1);
}

__global__
#if USE_LAUNCH_BOUNDS
__launch_bounds__(BW*BH, NB)
#endif
void compose(float4 *output,
             const float *r, const float *g,
             const float *b, const float *a,
             int width, int height, int rowstride)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int x = blockIdx.x*BW+tx, y = blockIdx.y*BH+ty;

    if(x >= width || y >= height)
        return;

    int idx = y*rowstride+x;

    output[idx] = make_float4(r[idx],g[idx],b[idx],a[idx]);
}

__global__
#if USE_LAUNCH_BOUNDS
__launch_bounds__(BW*BH, NB)
#endif
void compose(uchar4 *output,
             const float *r, const float *g,
             const float *b, const float *a,
             int width, int height, int rowstride)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int x = blockIdx.x*BW+tx, y = blockIdx.y*BH+ty;

    if(x >= width || y >= height)
        return;

    int idx = y*rowstride+x;

    float4 v = saturate(make_float4(r[idx],g[idx],b[idx],a[idx]))*255.0f;
    output[idx] = make_uchar4(v.x,v.y,v.z,v.w);
}

template <class T>
void compose(image_ptr<T> &out,
             const image<float> &r, const image<float> &g,
             const image<float> &b, const image<float> &a)
{
    if(r.width()!=g.width() || g.width()!=b.width() || b.width()!=a.width() ||
       r.height()!=g.height() || g.height()!=b.height() || b.height()!=a.height())
    {
        throw std::runtime_error("compose: non-conformant input image dimensions");
    }

//    out.resize(r.width(), r.height());

    dim3 bdim(BW,BH),
         gdim((out.width()+bdim.x-1)/bdim.x, (out.height()+bdim.y-1)/bdim.y);

    compose<<<gdim, bdim>>>(&out, &r,&g,&b,&a,
                            out.width(), out.height(), out.rowstride());
}

template <class T>
void compose(image_ptr<T> &out,
             const image<float> &r, const image<float> &g,
             const image<float> &b)
{
    if(r.width() != g.width() || g.width() != b.width() ||
       r.height() != g.height() || g.height() != b.height())
    {
        throw std::runtime_error("compose: non-conformant input image dimensions");
    }

//    out.resize(r.width(), r.height());

    dim3 bdim(BW,BH),
         gdim((out.width()+bdim.x-1)/bdim.x, (out.height()+bdim.y-1)/bdim.y);

    compose<<<gdim, bdim>>>(out, &r,&g,&b, out.width(), out.height(),
                            out.rowstride());
}

template
void compose(image_ptr<float4> &out,
             const image<float> &r, const image<float> &g,
             const image<float> &b);

template
void compose(image_ptr<uchar4> &out,
             const image<float> &r, const image<float> &g,
             const image<float> &b);

template
void compose(image_ptr<float4> &out,
             const image<float> &r, const image<float> &g,
             const image<float> &b, const image<float> &a);

template
void compose(image_ptr<uchar4> &out,
             const image<float> &r, const image<float> &g,
             const image<float> &b, const image<float> &a);


// decompose -----------------------------------------------------

__global__
#if USE_LAUNCH_BOUNDS
__launch_bounds__(BW*BH, NB)
#endif
void decompose(float *r, float *g, float *b, const float4 *input,
                          int width, int height, int rowstride)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int x = blockIdx.x*BW+tx, y = blockIdx.y*BH+ty;

    if(x >= width || y >= height)
        return;

    int idx = y*rowstride+x;

    float4 v = input[idx];
    r[idx] = v.x;
    g[idx] = v.y;
    b[idx] = v.z;
}

__global__
#if USE_LAUNCH_BOUNDS
__launch_bounds__(BW*BH, NB)
#endif
void decompose(float *r, float *g, float *b, const uchar4 *input,
                          int width, int height, int rowstride)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int x = blockIdx.x*BW+tx, y = blockIdx.y*BH+ty;

    if(x >= width || y >= height)
        return;

    int idx = y*rowstride+x;

    uchar4 v = input[idx];
    r[idx] = v.x/255.0f;
    g[idx] = v.y/255.0f;
    b[idx] = v.z/255.0f;
}

template <class T>
void decompose(vector<float> out[3], const vector<T> &in,
               int width, int height, int rowstride)
{
    for(int i=0; i<3; ++i)
        out[i].resize(rowstride*height);

    dim3 bdim(BW,BH),
         gdim((width+bdim.x-1)/bdim.x, (height+bdim.y-1)/bdim.y);

    decompose<<<gdim, bdim>>>(out[0],out[1],out[2], in, width, height, rowstride);
}

template
void decompose(vector<float> out[3], const vector<float4> &in,
               int width, int height, int rowstride);

template
void decompose(vector<float> out[3], const vector<uchar4> &in,
               int width, int height, int rowstride);

// grayscale ------------------------------------------------------------

__global__
#if USE_LAUNCH_BOUNDS
__launch_bounds__(BW*BH, NB)
#endif
void grayscale(float *output, const uchar4 *input,
               int width, int height, int rowstride)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int x = blockIdx.x*BW+tx, y = blockIdx.y*BH+ty;

    if(x >= width || y >= height)
        return;

    int idx = y*rowstride+x;

    uchar4 in = input[idx];
    output[idx] = in.x/255.0f*0.2126f + in.y/255.0f*0.7152f + in.z/255.0f*0.0722f;
}

__global__
#if USE_LAUNCH_BOUNDS
__launch_bounds__(BW*BH, NB)
#endif
void grayscale(float *output, const float4 *input,
               int width, int height, int rowstride)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int x = blockIdx.x*BW+tx, y = blockIdx.y*BH+ty;

    if(x >= width || y >= height)
        return;

    int idx = y*rowstride+x;

    float4 in = input[idx];
    output[idx] = in.x*0.2126f + in.y*0.7152f + in.z*0.0722f;
}

template <class T>
void grayscale(vector<float> &out, const vector<T> &in,
               int width, int height, int rowstride)
{
    out.resize(rowstride*height);

    dim3 bdim(BW,BH),
         gdim((width+bdim.x-1)/bdim.x, (height+bdim.y-1)/bdim.y);

    grayscale<<<gdim, bdim>>>(&out, &in, width, height, rowstride);
}


template
void grayscale(vector<float> &out, const vector<float4> &in,
               int width, int height, int rowstride);

template
void grayscale(vector<float> &out, const vector<uchar4> &in,
               int width, int height, int rowstride);

} // namespace gpu
