#ifndef GPU_RECURSIVE_FILTER_RECFILTER_H
#define GPU_RECURSIVE_FILTER_RECFILTER_H

// Disclaimer: this is optimized for usability, not speed
// Note: everything is row-major (i.e. math::Vector represents a row in a matrix)

#include <iostream>
#include <cassert>
#include "hostdev.h"
#include "linalg.h"
#include "vector.h"
#include "image.h"
#include "linalg.h"

#define ORDER 1

namespace gpu
{

#ifdef  __CUDA_ARCH__
#   ifdef assert
#       undef assert
#   endif
#   define assert (void)
#endif

template <class T>
HOSTDEV
T rec_op(const T &x, const T &a)
{
    return x - a;
}

// forward ---------------------------------------------------------

template <class T, int R>
HOSTDEV
T fwd(math::Vector<T,R> &p, T x, const math::Vector<T,R+1> &w)
{
    T acc = rec_op(x,p[R-1]*w[1]);
#pragma unroll
    for(int k=R-1; k>=1; --k)
    {
        acc = rec_op(acc,p[R-1-k]*w[k+1]);
        p[R-1-k] = p[R-1-k+1];
    }
    return p[R-1] = acc;
}


template <class T, int N, int R>
void fwd_inplace(const math::Vector<T,R> &_p, math::Vector<T,N> &b, const math::Vector<T,R+1> &w)
{
    math::Vector<T,R> p = _p;
    for(int j=0; j<b.size(); ++j)
        b[j] = fwd(p, w[0]*b[j], w);
}

template <class T, int M, int N, int R>
void fwdD_inplace(const math::Matrix<T,M,R> &p, math::Matrix<T,M,N> &b,
                 const math::Vector<T,R+1> &w)
{
    for(int i=0; i<b.rows(); ++i)
        fwd_inplace(p[i], b[i], w);
}

template <class T, int M, int N, int R>
void fwd_inplace(const math::Matrix<T,M,R> &p, math::Matrix<T,M,N> &b,
                 const math::Vector<T,R+1> &w)
{
    fwdD_inplace(p,b,w);
}

template <class T, int M, int N, int R>
math::Matrix<T,M,N> fwdD(const math::Matrix<T,M,R> &p, const math::Matrix<T,M,N> &b,
              const math::Vector<T,R+1> &w)
{
    math::Matrix<T,M,N> fb = b;
    fwdD_inplace(p, fb, w);
    return fb;
}

template <class T, int M, int N, int R>
math::Matrix<T,M,N> fwd(const math::Matrix<T,M,R> &p, const math::Matrix<T,M,N> &b,
              const math::Vector<T,R+1> &w)
{
    return fwdD(p, b, w);
}

// forward transposed  ----------------------------------------------------

template <class T, int M, int N, int R>
math::Matrix<T,M,N> fwdT(const math::Matrix<T,R,N> &pT, const math::Matrix<T,M,N> &b,
                  const math::Vector<T,R+1> &w)
{
    return transp(fwd(transp(pT), transp(b), w));
}

template <class T, int M, int N, int R>
math::Matrix<T,M,N> fwd(const math::Matrix<T,R,N> &pT, const math::Matrix<T,M,N> &b,
                  const math::Vector<T,R+1> &w)
{
    return fwdT(pT, b, w);
}

template <class T, int M, int N, int R>
void fwdT_inplace(const math::Matrix<T,R,N> &p, math::Matrix<T,M,N> &b,
                 const math::Vector<T,R+1> &w)
{
    b = fwdT(p, b, w);
}

template <class T, int M, int N, int R>
void fwd_inplace(const math::Matrix<T,R,N> &p, math::Matrix<T,M,N> &b,
                 const math::Vector<T,R+1> &w)
{
    fwdT_inplace(p, b, w);
}


// reverse ---------------------------------------------------------------

template <class T, int R>
HOSTDEV
T rev(T x, math::Vector<T,R> &e, const math::Vector<T,R+1> &w)
{
    T acc = rec_op(x,e[0]*w[1]);
#pragma unroll
    for(int k=R-1; k>=1; --k)
    {
        acc = rec_op(acc,e[k]*w[k+1]);
        e[k] = e[k-1];
    }
    return e[0] = acc;
}

template <class T, int N, int R>
void rev_inplace(math::Vector<T,N> &b, const math::Vector<T,R> &_e, const math::Vector<T,R+1> &w)
{
    math::Vector<T,R> e = _e;
    for(int j=b.size()-1; j>=0; --j)
        b[j] = rev(w[0]*b[j], e, w);
}

template <class T, int M, int N, int R>
void revD_inplace(math::Matrix<T,M,N> &b, const math::Matrix<T,M,R> &e,
              const math::Vector<T,R+1> &w)
{
    for(int i=0; i<b.rows(); ++i)
        rev_inplace(b[i], e[i], w);
}

template <class T, int M, int N, int R>
void rev_inplace(math::Matrix<T,M,N> &b, const math::Matrix<T,M,R> &e,
              const math::Vector<T,R+1> &w)
{
    revD_inplace(b, e, w);
}

template <class T, int M, int N, int R>
math::Matrix<T,M,N> revD(const math::Matrix<T,M,N> &b, const math::Matrix<T,M,R> &e,
              const math::Vector<T,R+1> &w)
{
    math::Matrix<T,M,N> rb = b;
    revD_inplace(rb, e, w);

    return rb;
}

template <class T, int M, int N, int R>
math::Matrix<T,M,N> rev(const math::Matrix<T,M,N> &b, const math::Matrix<T,M,R> &e,
              const math::Vector<T,R+1> &w)
{
    return revD(b, e, w);
}

// reverse transposed  ----------------------------------------------------

template <class T, int M, int N, int R>
math::Matrix<T,M,N> revT(const math::Matrix<T,M,N> &b, const math::Matrix<T,R,N> &eT,
                  const math::Vector<T,R+1> &w)
{
    return transp(rev(transp(b), transp(eT), w));
}

template <class T, int M, int N, int R>
math::Matrix<T,M,N> rev(const math::Matrix<T,M,N> &b, const math::Matrix<T,R,N> &eT,
                  const math::Vector<T,R+1> &w)
{
    return revT(b,eT,w);
}

template <class T, int M, int N, int R>
void revT_inplace(math::Matrix<T,M,N> &b, const math::Matrix<T,R,N> &p,
                 const math::Vector<T,R+1> &w)
{
    b = revT(b, p, w);
}

template <class T, int M, int N, int R>
void rev_inplace(math::Matrix<T,M,N> &b, const math::Matrix<T,R,N> &p,
                 const math::Vector<T,R+1> &w)
{
    revT_inplace(b, p, w);
}


// head ---------------------------------------------------------------

template <int R, int M, int N, class T>
math::Matrix<T,M,R> head(const math::Matrix<T,M,N> &mat)
{
    assert(mat.cols() >= R);

    math::Matrix<T,M,R> h;
    for(int j=0; j<R; ++j)
        for(int i=0; i<mat.rows(); ++i)
            h[i][j] = mat[i][j];

    return h;
}

// head transposed  ----------------------------------------------------

template <int R, int M, int N, class T>
math::Matrix<T,R,N> headT(const math::Matrix<T,M,N> &mat)
{
    return transp(head<R>(transp(mat)));
}


// tail ---------------------------------------------------------------

template <int R, int M, int N, class T>
math::Matrix<T,M,R> tail(const math::Matrix<T,M,N> &mat)
{
    assert(mat.cols() >= R);

    math::Matrix<T,M,R> t;
    for(int j=0; j<R; ++j)
        for(int i=0; i<mat.rows(); ++i)
            t[i][j] = mat[i][mat.cols()-R+j];

    return t;
}

// tail transposed  ----------------------------------------------------

template <int R, int M, int N, class T>
math::Matrix<T,R,N> tailT(const math::Matrix<T,M,N> &mat)
{
    return transp(tail<R>(transp(mat)));
}

enum BorderType
{
    CLAMP_TO_ZERO,
    CLAMP_TO_EDGE,
    REPEAT,
    REFLECT
};

typedef float pixel_type;
void recursive_filter_5_setup(int width, int height, int rowstride,
                              const math::Vector<float, ORDER+1> &w,
                              BorderType border_type=CLAMP_TO_ZERO,
                              int border=1);
void recursive_filter_5_free();

void recursive_filter_5(image<float> &d_inout, unsigned border=0);

void recursive_filter_5(image<float> &d_output,
                        const image<float> &d_input,
                        unsigned border=0);

} // namespace gpu

#endif

