#ifndef GPU_SOLVE_H
#define GPU_SOLVE_H

namespace gpu
{

__device__ inline int linear_solve(float a, float b, float r[1])
{
    if(a == 0)
    {
        if(b == 0)
        {
            // it has actually infinite solutions
            r[0] = 0;
            return 1;
        }
        else
            return 0; // no solutions
    }
    else
    {
        r[0] = -b/a;
        return 1;
    }
}

/* From "How to Solve a Quadratic Equation",
 * Jim Blinn's Corner, Nov/Dec 2005 */

// a >= 0 ? r[1] >= r[2] : r[1] < r[2];

__device__ inline int quadratic_solve(float a, float b, float c, float delta, float r[2])/*{{{*/
{
    if(a == 0)
        return linear_solve(b,c,r);

    if (delta < 0)
        return 0;

    if (b > 0.0)
    {
        b /= 2.0;

        float q = b+sqrt(delta);
        r[0] = -c/q;
        r[1] = -q/a;
        return 2;
    }
    else if (b < 0.0)
    {
        b /= 2.0;

        float q = -b+sqrt(delta);
        r[0] = q/a;
        r[1] = c/q;
        return 2;
    }
    else
    {
        float q = sqrt(-a*c);
        if (abs(a) >= abs(c))
        {
            r[0] = q/a;
            r[1] = -q/a;
        }
        else
        {
            r[0] = -c/q;
            r[1] = c/q;
        }
        return 2;
    }
}/*}}}*/

__device__ inline int quadratic_solve(float a, float b, float c, float r[2])/*{{{*/
{
    return quadratic_solve(a, b, c, b*b/4-a*c, r);
}/*}}}*/

// Computes the roots of the homogeneous polynomial
//
//     a s s + 2 b s t + c s s == 0.
//
// Please note the 2 b in the equation!!!
//
// Roots of a t t + 2 b t + c == 0 can be computed from
// t[i]/s[i], but s[i] can be very small or even zero.
// Never divide these numbers! We never really need the
// If s[i] is zero, what this means is that the quadratic
// degenerated to a linear equation.
// If both s[i] are zero, then the linear equation
// degenerated as well
template <class T>
__device__ inline int quadratic_solve(T a, T b, T c,/*{{{*/
                                      T delta, int e,
                                      T t[2], T s[2])
{
    if(delta >= 0)
    {
        T d = ldexp(T(1), e)*sqrt(delta);
        if(b > 0)
        {
            T e = b+d;
            t[0] = -c;
            t[1] = e;
            s[0] = e;
            s[1] = -a;
        }
        else if(b < 0)
        {
            T e = -b+d;
            t[0] = e;
            t[1] = c;
            s[0] = a;
            s[1] = e;
        }
        else if(abs(a) > abs(c))
        {
            t[0] = d;
            t[1] = -d;
            s[0] = s[1] = a;
        }
        else
        {
            t[0] = -c;
            t[1] = c;
            s[0] = s[1] = d;
        }
        return 2;
    }
    else
        return 0;
}/*}}}*/


__device__ inline float chop(float t) /*{{{*/
{
    if (t < 0.f && t > -1e-7f)
        return 0.f;
    else
        return t;
}/*}}}*/

__device__ inline double chop(double t) /*{{{*/
{
    if (t < 0.0 && t > -1e-15)
        return 0;
    else
        return t;
}/*}}}*/

// discriminant really is the return of the function, times 2^(2*e)
template <class T>
__device__ inline T discriminant(T a, T b, T c, int &e)/*{{{*/
{
    // Returns sqrt(b b - a c) avoiding overflows, underflows,
    // and tiny negative numbers
    // It may well be faster to do all computations in double
    // precision!

    // adjust exponent so the largest number doesn't overflow
    int ea, eb, ec;
    frexp(a, &ea);
    frexp(b, &eb);
    frexp(c, &ec);

    e = max(eb, (ea+ec)>>2);

    T s = ldexp(T(1), -e);
    a *= s;
    b *= s;
    c *= s;

    // compute b^2 - ac and clamp tiny negative numbers to zero
    return chop(b*b-a*c);
}/*}}}*/

template <class T>
__device__ inline int quadratic_solve(T a, T b, T c,/*{{{*/
                                      T t[2], T s[2])
{
    int e;
    T delta = discriminant(a, b, c, e);
    return quadratic_solve(a,b,c, delta, e, t, s);
}/*}}}*/

/* From "How to Solve a Cubic Equation, Part 5: Back to Numerics",
 * Jim Blinn's Corner, May/Jun 2007 */
namespace detail
{
    __device__ inline float curoot(float x)
    {
        float value;
        float absx;
        int neg;
        neg = 0;
        absx = x;
        if (x < 0.0)
        {
            absx = -x;
            neg = 1;
        }
        if (absx != 0.0)
            value = exp( log(absx)/3.0 );
        else
            value = 0.0;
        if (neg == 1)
            value = -value;
        return(value);
    }

    __device__ inline void single_helper(float tilde_a, float bar_c,
                                         float bar_d, float D, float *x)
    {
        const float sqrt_D = sqrt(-D);
        float p;
        if (bar_d < 0.)
            p = curoot(.5*(-bar_d+fabs(tilde_a)*sqrt_D));
        else
            p = curoot(.5*(-bar_d-fabs(tilde_a)*sqrt_D));
        *x = p - bar_c/p;
    }

    __device__ inline void triple_helper(float bar_c, float bar_d, float sqrt_D,
                                         float s, float r, float *x, float *o)
    {
        float theta = fabs(atan2(s*sqrt_D, -bar_d))/3.;
        float x1 = 2.*sqrt(fabs(bar_c))*cos(theta);
        float x3 = 2.*sqrt(fabs(bar_c))*(-.5*cos(theta)-(sqrt(3.)/2.)*sin(theta));
        if (x1+x3 > 2.*r)
        {
            *x = x1;
            *o = x3;
        }
        else
        {
            *x = x3;
            *o = x1;
        }
    }
} // namespace detail

__device__ inline int cubic_solve(float a, float b, float c, float d, float *r)
{
    if(a == 0)
        return quadratic_solve(b,c,d,r);

    b = b*(1.f/3);
    c = c*(1.f/3);
    const float d1 = a*c-b*b;
    const float d2 = a*d-b*c;
    const float d3 = b*d-c*c;
    const float D = 4.f*d1*d3-d2*d2;
    //fprintf(stderr, "D=%.10f\n", D);
    if (D <= 0.)
    {
        // triple root
        if(d1 == 0 && d2 == 0 && d3 == 0)
        {
            if(abs(a) > abs(b))
            {
                if(abs(a) > abs(c))
                    r[0] = r[1] = r[2] = -b/a;
                else
                    r[0] = r[1] = r[2] = -d/c;
            }
            else if(abs(b) > abs(c))
                r[0] = r[1] = r[2] = -c/b;
            else
                r[0] = r[1] = r[2] =  -d/c;
            return 3;
        }

        if (b*b*b*d >= a*c*c*c)
        {
            const float bar_c = d1;
            const float bar_d = -2*b*d1+a*d2;
            float r1;
            detail::single_helper(a, bar_c, bar_d, D, &r1);
            *r = (r1-b)/a;
            // double root
            if(D == 0)
            {
                r[1] = r[2] = (-0.5*r1-b)/a;
                return 3;
            }
            else
                return 1;
        }
        else
        {
            const float bar_c = d3;
            const float bar_d = -d*d2+2*c*d3;
            float r1;
            detail::single_helper(d, bar_c, bar_d, D, &r1);
            *r = -d/(r1+c);
            // double root
            if(D == 0)
            {
                r[1] = r[2] = -d/(-0.5*r1+b);
                return 3;
            }
            else
                return 1;
        }
    }
    else
    {
        float sqrt_D = sqrt(D);
        const float bar_c_a = d1;
        const float bar_d_a = -2*b*d1+a*d2;
        float xl, o;
        detail::triple_helper(bar_c_a, bar_d_a, sqrt_D, a, b, &xl, &o);
        xl = xl - b;
        float wl = a;
        const float bar_c_d = d3;
        const float bar_d_d = -d*d2+2*c*d3;
        float xs;
        detail::triple_helper(bar_c_d, bar_d_d, sqrt_D, d, c, &o, &xs);
        float ws = xs + c;
        xs = -d;
        float e = wl*ws;
        float f = -xl*ws-wl*xs;
        float g = xl*xs;
        float xm = c*f-b*g;
        float wm = -b*f+c*e;
        r[0] = xs/ws;
        r[1] = xm/wm;
        r[2] = xl/wl;
        return 3;
    }
}

} // namespace gpu

#endif
