#ifndef GPU_VECTOR_H
#define GPU_VECTOR_H

#include <vector>
#include <climits>
#include <util/util.h>
#include <thrust/iterator/detail/normal_iterator.h>
#include <thrust/iterator/reverse_iterator.h>
#include <thrust/device_ptr.h>
#include "alloc.h"
#include "hostdev.h"

namespace gpu
{

template <class T>
class vector_ptr
{
public:
    typedef thrust::detail::normal_iterator<thrust::device_ptr<T> > iterator;
    typedef thrust::detail::normal_iterator<thrust::device_ptr<const T> > const_iterator;

    typedef thrust::reverse_iterator<iterator> reverse_iterator;
    typedef thrust::reverse_iterator<const_iterator> const_reverse_iterator;

    HOSTDEV vector_ptr() : m_data(), m_size(0) {}
    HOSTDEV vector_ptr(T *data, size_t size);

    template <class U>
    HOSTDEV vector_ptr(const vector_ptr<U> &that,
                       typename rod::enable_if<rod::is_convertible<T*,U*>::value>::type *dummy = NULL)
        : m_data(that.m_data)
        , m_size(that.m_size)
    {
    }

    iterator begin() { return iterator(m_data); }
    iterator end() { return iterator(m_data+size()); }

    const_iterator begin() const { return const_iterator(m_data); }
    const_iterator end() const { return const_iterator(m_data+size()); }

    const_iterator cbegin() const { return const_iterator(m_data); }
    const_iterator cend() const { return const_iterator(m_data+size()); }

    reverse_iterator rbegin() { return make_reverse_iterator(end()); }
    reverse_iterator rend() { return make_reverse_iterator(begin()); }

    const_reverse_iterator rbegin() const { return make_reverse_iterator(end()); }
    const_reverse_iterator rend() const { return make_reverse_iterator(begin()); }

    const_reverse_iterator crbegin() const { return make_reverse_iterator(end()); }
    const_reverse_iterator crend() const { return make_reverse_iterator(begin()); }

    HOSTDEV void reset(T *data, size_t size);

    void fill(unsigned char byte);

    HOSTDEV const T &operator[](int idx) const;
    __device__ T &operator[](int idx);

    void copy_to_host(std::vector<T> &out) const;
    void copy_from_host(const std::vector<T> &in);

    size_t copy_to_host(T *data, size_t s=UINT_MAX) const;
    size_t copy_from_host(const T *data, size_t s=UINT_MAX);

    HOSTDEV bool empty() const { return size()==0; }
    HOSTDEV size_t size() const { return m_size; }

    HOSTDEV T *data() { return raw_pointer_cast(m_data); }
    HOSTDEV const T *data() const { return raw_pointer_cast(m_data); }

    HOSTDEV T &back() const { return operator[](size()-1); }

    HOSTDEV operator T*() { return data(); }
    HOSTDEV operator const T*() const { return data(); }

    HOSTDEV const T *operator&() const { return raw_pointer_cast(m_data); }
    HOSTDEV T *operator&() { return raw_pointer_cast(m_data); }

    template <class U>
    HOSTDEV friend void swap(vector_ptr<U> &a, vector_ptr<U> &b);

    HOSTDEV vector_ptr &operator+=(long long offset)
    {
        m_data = thrust::device_ptr<T>(data() + offset);
        return *this;
    }

    HOSTDEV vector_ptr &operator-=(long long offset)
    {
        m_data = thrust::device_ptr<T>(data() - offset);
        return *this;
    }

private:
    thrust::device_ptr<T> m_data;
    size_t m_size;
};


template <class T>
class vector
{
public:
    typedef thrust::detail::normal_iterator<thrust::device_ptr<T> > iterator;
    typedef thrust::detail::normal_iterator<thrust::device_ptr<const T> > const_iterator;
    typedef thrust::reverse_iterator<iterator> reverse_iterator;
    typedef thrust::reverse_iterator<const_iterator> const_reverse_iterator;

    vector(const T *data, size_t size);
    vector(const std::vector<T> &data);
    vector(const vector &that);
    vector(size_t size=0);

    ~vector();

    iterator begin() { return iterator(m_data); }
    iterator end() { return iterator(m_data+size()); }

    const_iterator begin() const { return const_iterator(m_data); }
    const_iterator end() const { return const_iterator(m_data+size()); }

    const_iterator cbegin() const { return const_iterator(m_data); }
    const_iterator cend() const { return const_iterator(m_data+size()); }

    reverse_iterator rbegin() { return make_reverse_iterator(end()); }
    reverse_iterator rend() { return make_reverse_iterator(begin()); }

    const_reverse_iterator rbegin() const { return make_reverse_iterator(end()); }
    const_reverse_iterator rend() const { return make_reverse_iterator(begin()); }

    const_reverse_iterator crbegin() const { return make_reverse_iterator(end()); }
    const_reverse_iterator crend() const { return make_reverse_iterator(begin()); }

    void resize(size_t size);
    void reserve(size_t size);

    void reset(T *data, size_t size);

    void clear() { m_size = 0; }
    void fill(unsigned char byte);

    const T &operator[](int idx) const;

    vector &operator=(const vector &that);

    void copy_to_host(std::vector<T> &out) const;
    void copy_from_host(const std::vector<T> &in);

    void copy_to_host(T *data, size_t s) const;
    void copy_from_host(const T *data, size_t s);

    void copy2D_to_host(T *out, size_t width, size_t height,
                        size_t rowstride) const;
    void copy2D_from_host(const T *in, size_t width, size_t height,
                          size_t rowstride);

    bool empty() const { return size()==0; }

    size_t size() const { return m_size; }
    size_t capacity() const { return m_capacity; }

    T *data() { return raw_pointer_cast(m_data); }
    const T *data() const { return raw_pointer_cast(m_data); }

    const T &back(cudaStream_t s=0) const;

    operator T*() { return data(); }
    operator const T*() const { return data(); }

    vector_ptr<const T> operator&() const
        { return vector_ptr<const T>(data(), size()); }
    vector_ptr<T> operator&()
        { return vector_ptr<T>(data(), size()); }

    template <class U>
    friend void swap(vector<U> &a, vector<U> &b);

private:
    thrust::device_ptr<T> m_data;
    size_t m_size, m_capacity;
};

} // namespace gpu

#include "vector.hpp"

#endif
//vi: ai sw=4 ts=4
