#include <cuda_runtime.h>
#include <stdio.h>
#include "alloc.h"

namespace gpu
{

// class vector<T> --------------------------------------------------------

template <class T>
vector<T>::vector(const T *data, size_t size)
    :m_data(), m_size(0), m_capacity(0)
{
    resize(size);
    cudaMemcpy(this->data(), const_cast<T *>(data), size*sizeof(T),
               cudaMemcpyHostToDevice);
    check_cuda_error("Error during memcpy from host to device");
}

template <class T>
vector<T>::vector(const std::vector<T> &data)
    :m_data(), m_size(0), m_capacity(0)
{
    resize(data.size());
    cudaMemcpy(this->data(), &data[0], data.size()*sizeof(T),
               cudaMemcpyHostToDevice);
    check_cuda_error("Error during memcpy from host to device");
}

template <class T>
vector<T>::vector(const vector &that)
    :m_data(), m_size(0), m_capacity(0)
{
    *this = that;
}

template <class T>
vector<T>::vector(size_t size)
    :m_data(), m_size(0), m_capacity(0)
{
    resize(size);
}

template <class T>
vector<T>::~vector()
{
    cuda_delete(raw_pointer_cast(m_data));
    update_memused(-sizeof(T)*m_capacity);
}

template <class T>
void vector<T>::resize(size_t size)
{
    if(size > m_capacity)
        reserve(size);

    m_size = size;
}

template <class T>
void vector<T>::reserve(size_t capacity)
{
    if(capacity > m_capacity)
    {
        thrust::device_ptr<T> newdata(cuda_new_array<T>(capacity));
        try
        {
            cudaMemcpy(raw_pointer_cast(newdata), 
                       raw_pointer_cast(m_data), size()*sizeof(T),
                       cudaMemcpyDeviceToDevice);

            cuda_delete(raw_pointer_cast(m_data));
            m_data = newdata;
        }
        catch(...)
        {
            cuda_delete(raw_pointer_cast(newdata));
            throw;
        }

        update_memused(-sizeof(T)*m_capacity);
        update_memused(sizeof(T)*capacity);
    }

    m_capacity = capacity;
}

template <class T>
void vector<T>::reset(T *data, size_t size)
{
    cuda_delete(m_data);

    update_memused(-sizeof(T)*m_capacity);

    m_data = NULL;

    m_capacity = m_size = size;
    m_data = data;
}

template <class T>
void vector<T>::fill(unsigned char byte)
{
    cudaMemset(raw_pointer_cast(m_data), byte, m_size*sizeof(T));
}

template <class T>
const T &vector<T>::operator[](int idx) const
{
    static T value;
    cudaMemcpy(&value, data()+idx, sizeof(T), cudaMemcpyDeviceToHost);
    return value;
}

template <class T>
const T &vector<T>::back(cudaStream_t s) const
{
    static T value;
    cudaMemcpyAsync(&value, data()+size()-1, sizeof(T), cudaMemcpyDeviceToHost, s);
    cudaStreamSynchronize(s);
    return value;
}

template <class T>
vector<T> &vector<T>::operator=(const vector &that)
{
    resize(that.size());
    cudaMemcpy(data(), that.data(), size()*sizeof(T), cudaMemcpyDeviceToDevice);
    check_cuda_error("Error during memcpy from device to device");
    return *this;
}

template <class T>
void vector<T>::copy_to_host(T *data, size_t s) const
{
    using std::min;

    cudaMemcpy(data, this->data(), min(size(),s)*sizeof(T),
               cudaMemcpyDeviceToHost);
    check_cuda_error("Error during memcpy from device to host");
}

template <class T>
void vector<T>::copy_from_host(const T *data, size_t s)
{
    using std::min;

    resize(s);

    cudaMemcpy(this->data(), data, s*sizeof(T), cudaMemcpyHostToDevice);
    check_cuda_error("Error during memcpy from device to host");
}

namespace detail
{
    // workaround due to std::vector<bool> being a completely different beast

    template <class T>
    void copy_to_host(std::vector<T> &out, const vector<T> &in)
    {
        out.resize(in.size());
        in.copy_to_host(&out[0], out.size());
    }

    template <class T>
    void copy_to_host(std::vector<T> &out, const vector_ptr<T> &in)
    {
        out.resize(in.size());
        in.copy_to_host(&out[0], out.size());
    }
    inline void copy_to_host(std::vector<bool> &out, const vector_ptr<bool> &in)
    {
        bool *temp = NULL;
        try
        {
            temp = new bool[in.size()];

            in.copy_to_host(temp, in.size());

            out.assign(temp, temp + in.size());
            delete[] temp;
        }
        catch(...)
        {
            delete[] temp;
            throw;
        }
    }

    template <class T>
    void copy_from_host(vector<T> &out, const std::vector<T> &in)
    {
        out.resize(in.size());
        out.copy_from_host(&in[0], in.size());
    }

    template <class T>
    void copy_from_host(vector_ptr<T> &out, const std::vector<T> &in)
    {
        out.copy_from_host(&in[0], in.size());
    }

    inline void copy_from_host(vector_ptr<bool> &out, const std::vector<bool> &in)
    {
        std::vector<unsigned char> temp;
        temp.assign(in.begin(), in.end());

        out.copy_from_host((const bool *)&temp[0], temp.size());
    }

}

template <class T>
void vector<T>::copy_to_host(std::vector<T> &out) const
{
    detail::copy_to_host(out, *this);
}

template <class T>
void vector<T>::copy_from_host(const std::vector<T> &in)
{
    detail::copy_from_host(*this, in);
}

template <class T>
void vector<T>::copy2D_to_host(T *out, size_t width, size_t height,
                                size_t rowstride) const
{
    cudaMemcpy2D(out, width*sizeof(T),
                 data(), rowstride*sizeof(T),
                 width*sizeof(T), height, cudaMemcpyDeviceToHost);

    check_cuda_error("Error during memcpy from device to host");
}

template <class T>
void vector<T>::copy2D_from_host(const T *in, size_t width, size_t height,
                                  size_t rowstride)
{
    resize(rowstride*height);

    cudaMemcpy2D(data(), rowstride*sizeof(T), in, width*sizeof(T),
                 width*sizeof(T), height, cudaMemcpyHostToDevice);

    check_cuda_error("Error during memcpy from host to device");
}

template <class T>
void swap(vector<T> &a, vector<T> &b)
{
    using std::swap;
    std::swap(a.m_data, b.m_data);
    swap(a.m_size, b.m_size);
    swap(a.m_capacity, b.m_capacity);
}


template <class T>
std::vector<T> to_host(const T *d_vec, unsigned len)
{
    std::vector<T> out;
    out.resize(len);

    cudaMemcpy(&out[0], d_vec, len*sizeof(T), cudaMemcpyDeviceToHost);
    check_cuda_error("Error during memcpy from device to host");

    return out;
}

template <class T>
std::vector<T> to_host(const vector<T> &v)
{
    return to_cpu(v.data(), v.size());
}

// class vector_ptr<T> -------------------------------------------------------

template <class T>
HOSTDEV
vector_ptr<T>::vector_ptr(T *data, size_t size)
    : m_size(size), m_data(data)
{
}

template <class T>
HOSTDEV
void vector_ptr<T>::reset(T *data, size_t size)
{
    m_size = size;
    m_data = thrust::device_ptr<T>(data);
}

template <class T>
void vector_ptr<T>::fill(unsigned char byte)
{
    cudaMemset(m_data, byte, m_size*sizeof(T));
}

#if !__CUDA_ARCH__
template <class T>
const T &vector_ptr<T>::operator[](int idx) const
{
    static T value;
    cudaMemcpy(&value, data()+idx, sizeof(T), cudaMemcpyDeviceToHost);
    return value;
}
#else
template <class T>
const T &vector_ptr<T>::operator[](int idx) const
{
    return data()[idx];
}
template <class T>
__device__ T &vector_ptr<T>::operator[](int idx)
{
    return data()[idx];
}
#endif

template <class T>
void vector_ptr<T>::copy_to_host(std::vector<T> &out) const
{
    detail::copy_to_host(out, *this);
}

template <class T>
void vector_ptr<T>::copy_from_host(const std::vector<T> &in)
{
    detail::copy_from_host(*this, in);
}


template <class T>
size_t vector_ptr<T>::copy_to_host(T *data, size_t s) const
{
    using std::min;

    size_t nwritten = min(size(),s);

    cudaMemcpy(data, this->data(), nwritten*sizeof(T), cudaMemcpyDeviceToHost);
    check_cuda_error("Error during memcpy from device to host");
    return nwritten;
}

template <class T>
size_t vector_ptr<T>::copy_from_host(const T *data, size_t s)
{
    using std::min;

    size_t nwritten = min(size(),s);

    cudaMemcpy(this->data(), data, nwritten*sizeof(T), cudaMemcpyHostToDevice);
    check_cuda_error("Error during memcpy from device to host");
    return nwritten;
}

template <class T>
HOSTDEV
void swap(vector_ptr<T> &a, vector_ptr<T> &b)
{
    using std::swap;

    swap(a.m_data, b.m_data);
    swap(a.m_size, b.m_size);
}

} // namespace gpu
