#include <cstring>
#include <cctype>
#include <stdexcept>
#include "config.h"

#include "image.h"

#if IMAGE_USE_PNG
#   include "pngimage.h"
#endif

#if IMAGE_USE_EXR
#   include "exrimage.h"
#endif

typedef enum _e_ext {
    EXT_PNG = 1,
    EXT_EXR = 2,
    EXT_UNKNOWN = 3
} e_ext;

static e_ext type(const char *name) {
    char ext[4] = "";
    size_t n = strlen(name);
    if (n < 3) return EXT_UNKNOWN;
    ext[3-3] = tolower(name[n-3]);
    ext[3-2] = tolower(name[n-2]);
    ext[3-1] = tolower(name[n-1]);
    ext[3-0] = '\0';
    if (strcmp(ext, "png") == 0) return EXT_PNG;
    else if (strcmp(ext, "exr") == 0) return EXT_EXR;
    else return EXT_UNKNOWN;
}

namespace image {

    void free_text(void) {
#if IMAGE_USE_PNG
        png::free_text();
#endif
#if IMAGE_USE_EXR
        exr::free_text();
#endif
    }

    void init_text(int argc, char **argv) {
#if IMAGE_USE_PNG
        png::init_text(argc, argv);
#endif
#if IMAGE_USE_EXR
        exr::init_text(argc, argv);
#endif
    }

    void push_text(const char *key, const char *text) {
#if IMAGE_USE_PNG
        png::push_text(key, text);
#endif
#if IMAGE_USE_EXR
        exr::push_text(key, text);
#endif
    }

    void pop_text(int n) {
#if IMAGE_USE_PNG
        png::pop_text(n);
#endif
#if IMAGE_USE_EXR
        exr::pop_text(n);
#endif
    }

    template <>
	int load(const char *name, image::rgba<float> *rgba) {
        switch (type(name)) {
#if IMAGE_USE_PNG
            case EXT_PNG: return png::load(name, rgba);
#endif
#if IMAGE_USE_EXR
            case EXT_EXR: return exr::load(name, rgba);
#endif
            default:
                throw std::runtime_error("unknown image format");
        }
    }

    template <>
	int store(const char *name, const image::rgba<float> &rgba) {
        switch (type(name)) {
#if IMAGE_USE_PNG
            case EXT_PNG: return png::store(name, rgba);
#endif
#if IMAGE_USE_EXR
            case EXT_EXR: return exr::store(name, rgba);
#endif
            default:
                throw std::runtime_error("unknown image format");
        }
    }
}
