#ifndef IMAGE_H
#define IMAGE_H

#include <cfloat>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "color.h"

namespace image {
    template <class T = float> class rgba {
    public:
        rgba(void) {
            r = g = b = a = NULL;
            m_w = m_h = 0;
        }
        ~rgba() {
            free(r); free(g); free(b); free(a);
            r = g = b = a = NULL;
            m_w = m_h = 0;
        }

#if __cplusplus >= 201103L
        rgba(rgba &&that)
        {
            r = that.r;
            g = that.g;
            b = that.b;
            m_w = that.m_w;
            m_h = that.m_h;

            that.r = that.g = that.b = NULL;
            that.m_w = that.m_h = 0;
        }

        rgba &operator=(rgba &&that)
        {
            free(r);
            free(g);
            free(b);

            r = that.r;
            g = that.g;
            b = that.b;
            m_w = that.m_w;
            m_h = that.m_h;

            that.r = that.g = that.b = NULL;
            that.m_w = that.m_h = 0;

            return *this;
        }

        rgba(const rgba &) = delete;
#endif

        int resize(int h, int w) {
            // try to allocate new buffers
            free(r); free(g); free(b); free(a);
            r = static_cast<T *>(calloc(w*h,sizeof(T)));
            g = static_cast<T *>(calloc(w*h,sizeof(T)));
            b = static_cast<T *>(calloc(w*h,sizeof(T)));
            a = static_cast<T *>(malloc(w*h*sizeof(T)));
            if (!r || !g || !b || !a) {
                free(r); free(g); free(b); free(a);
                r = g = b = a = NULL;
                w = h = 0;
                return 0;
            }
            // fill alpha with 1.0
            for (int i = 0; i < w*h; i++) a[i] = 1.f;
            m_h = h;
            m_w = w;
            return 1;
        }
        int load(const rgba<float> &other) {
            if (!resize(other.height(), other.width()))
                return 0;
            memcpy(r, other.r, m_w*m_h*sizeof(T));
            memcpy(g, other.g, m_w*m_h*sizeof(T));
            memcpy(b, other.b, m_w*m_h*sizeof(T));
            memcpy(a, other.a, m_w*m_h*sizeof(T));
            return 1;
        }
        T *r, *g, *b, *a;
        void size(int *h, int *w) const { *h = m_h; *w = m_w; }
        int width(void) const { return m_w; }
        int height(void) const { return m_h; }
        void lrgb2srgb(void) {
            for (int i = 0; i < m_w*m_h; i++)
                image::lrgb2srgb(r[i], g[i], b[i], r+i, g+i, b+i);
        }

        void normalize(void) {
            float min = FLT_MAX, max = -FLT_MAX;
            for (int i = 0; i < m_w*m_h; i++)  {
                float r = this->r[i];
                float g = this->g[i];
                float b = this->b[i];
                if (r < min) min = r;
                if (g < min) min = g;
                if (b < min) min = b;
                if (r > max) max = r;
                if (g > max) max = g;
                if (b > max) max = b;
            }
            if (max > min) {
                float inv = 1.f/(max-min);
                for (int i = 0; i < m_w*m_h; i++)  {
                    this->r[i] -= min;
                    this->r[i] *= inv;
                    this->g[i] -= min;
                    this->g[i] *= inv;
                    this->b[i] -= min;
                    this->b[i] *= inv;
                }
            }
        }

        void srgb2lrgb(void) {
            for (int i = 0; i < m_w*m_h; i++)
                image::srgb2lrgb(r[i], g[i], b[i], r+i, g+i, b+i);
        }
        void lrgb2lab(void) {
            for (int i = 0; i < m_w*m_h; i++)
                image::lrgb2lab(r[i], g[i], b[i], r+i, g+i, b+i);
        }
        void lab2lrgb(void) {
            for (int i = 0; i < m_w*m_h; i++)
                image::lab2lrgb(r[i], g[i], b[i], r+i, g+i, b+i);
        }
    private:
        int m_w, m_h;
#if __cplusplus < 201103L
        rgba(const rgba &);
#endif
    };

    // for exr comments
    void free_text(void);
    void init_text(int argc, char **argv);
    void push_text(const char *key, const char *text);
    void pop_text(int n = 1);
    // load and store
    template <class T> int load(const char *name, image::rgba<T> *rgba);
    template <class T> int store(const char *name, const image::rgba<T> &rgba);
}

#endif // IMAGE_H
