#include <cstdio>
#include <stdexcept>
#include <vector>
#include <zlib.h>
#include <png.h>
#include "pngimage.h"
#include "color.h"
#include "extension.h"

static std::vector<png_text> g_text;

static extension::clamp clamp;

void user_error_fn(png_structp png_ptr,
	png_const_charp error_msg) {
	(void) png_ptr;
	fprintf(stderr, "libpng error: %s\n", error_msg);
}

void user_warning_fn(png_structp png_ptr,
	png_const_charp warning_msg) {
	(void) png_ptr;
	fprintf(stderr, "libpng warning: %s\n", warning_msg);
}

namespace image { namespace png {

    void free_text(void) {
        for (unsigned i = 0; i < g_text.size(); i++) {
            free(g_text[i].key);
            free(g_text[i].text);
        }
        g_text.resize(0);
    }

    void init_text(int argc, char **argv) {
        free_text();
        g_text.resize(argc);
        for (int i = 0; i < argc; i++) {
            char key[256];
            sprintf(key, "argv:%02d", i);
            g_text[i].compression = PNG_TEXT_COMPRESSION_NONE;
            g_text[i].key = strdup(key);
            g_text[i].text = strdup(argv[i]);
        }
    }

    void push_text(const char *key, const char *text) {
        png_text entry;
        entry.compression = PNG_TEXT_COMPRESSION_NONE;
        entry.key = strdup(key);
        entry.text = strdup(text);
        g_text.push_back(entry);
    }

    void pop_text(int n) {
        for (int i = 0; i < n; i++) {
            if (!g_text.empty()) {
                free(g_text.back().key);
                free(g_text.back().text);
                g_text.pop_back();
            }
        }
    }

    template <>
    int load(const char *name, image::rgba<float> *rgba) {
        // temporary image storage
        png_uint_16 ** volatile row_pointers = NULL;
        png_uint_16 * volatile data = NULL;
        // libpng structures
        png_structp png_ptr = NULL;
        png_infop info_ptr = NULL;
        // open file and check if it is a PNG file
        FILE *fp = fopen(name, "rb");
        if (!fp)
            throw std::runtime_error(std::string(name)+": "+strerror(errno));
        unsigned char signature[8];
        if (fread(signature, 1, 8, fp) < 8) {
            fclose(fp);
            throw std::runtime_error("Error reading png header");
        }
        if (png_sig_cmp(signature, 0, 8)) {
            fclose(fp);
            throw std::runtime_error("Invalid png");
        }
        // allocate reading structures
        png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL,
            user_error_fn, user_warning_fn);
        if (png_ptr) {
            info_ptr = png_create_info_struct(png_ptr);
        }
        if (!png_ptr || !info_ptr) {
            png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
            fclose(fp);
            throw std::runtime_error("unable to allocate png structure");
        }
        // setup long jump for error return
        if (setjmp(png_jmpbuf(png_ptr))) {
            png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
            fclose(fp);
            free(row_pointers);
            free(data);
            throw std::runtime_error("error reading png");
        }
        // set png file handle
        png_init_io(png_ptr, fp);
        png_set_sig_bytes(png_ptr, 8); // already skept 8 bytes
        // do not premultiply alpha
        png_set_alpha_mode(png_ptr, PNG_ALPHA_PNG, PNG_DEFAULT_sRGB);
        // load image information
        png_read_info(png_ptr, info_ptr);
        // get dimensions
        int height = png_get_image_height(png_ptr, info_ptr);
        int width = png_get_image_width(png_ptr, info_ptr);
        // allocate temporary image buffer and row_pointers
        data = static_cast<png_uint_16 *>(
            malloc(height*width*4*sizeof(png_uint_16)));
        row_pointers = static_cast<png_uint_16 **>(
            malloc(height*sizeof(png_uint_16 *)));
        // try to allocate output image
        if (!data || !row_pointers || !rgba->resize(height, width)) {
            // might as well use the same error handling as libpng...
            longjmp(png_jmpbuf(png_ptr), 1);
        }
        for (int i = 0; i < height; i++) {
            row_pointers[i] = &data[i*width*4];
        }
        // set all transformations required to read from any
        // format into RGBA16
        int color_type = png_get_color_type(png_ptr, info_ptr);
        int bit_depth = png_get_bit_depth(png_ptr, info_ptr);
        if (color_type == PNG_COLOR_TYPE_PALETTE) {
            png_set_palette_to_rgb(png_ptr);
        }

        if (color_type == PNG_COLOR_TYPE_GRAY ||
            color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
            png_set_gray_to_rgb(png_ptr);
        }

        if (color_type == PNG_COLOR_TYPE_RGB ||
            color_type == PNG_COLOR_TYPE_GRAY ||
            color_type == PNG_COLOR_TYPE_PALETTE) {
            png_set_add_alpha(png_ptr, 0xFFFF, PNG_FILLER_AFTER);
        }

        if (bit_depth < 16) {
            png_set_expand_16(png_ptr);
        }

        if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) {
            png_set_tRNS_to_alpha(png_ptr);
        }

        // should we flip endianness?
        long int a = 1;
        int swap = (*((unsigned char *) &a) == 1);
        if (swap) {
            png_set_swap(png_ptr);
        }
        // read image
        png_read_image(png_ptr, (png_bytepp) row_pointers);
        // save to image object
        const float tof = (1.f/65535.f);
        for (int i = 0; i < height; i++) {
            for (int j = 0; j < width; j++) {
                int p = (height-1-i)*width+j; // flip image so y is up
                int d = 4*(i*width+j);
                rgba->r[p] = srgbuncurve(data[d+0]*tof);
                rgba->g[p] = srgbuncurve(data[d+1]*tof);
                rgba->b[p] = srgbuncurve(data[d+2]*tof);
                rgba->a[p] = data[d+3]*tof;
            }
        }
        // finish advancing file pointer to end of stream (useful?)
        png_read_end(png_ptr, NULL);
        // clean-up and we are done
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
        fclose(fp);
        free(data);
        free(row_pointers);
        return 1;
    }


    template <>
    int store(const char *name, const image::rgba<unsigned char> &rgba) {
        // temporary image storage
        png_byte ** volatile row_pointers = NULL;
        png_byte * volatile data = NULL;
        // libpng structures
        png_structp png_ptr = NULL;
        png_infop info_ptr = NULL;
        // open file and check if it is a PNG file
        FILE *fp = fopen(name, "wb");
        if (!fp)
            throw std::runtime_error(std::string(name)+": "+strerror(errno));
        // allocate reading structures
        png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL,
            user_error_fn, user_warning_fn);
        if (png_ptr) {
            info_ptr = png_create_info_struct(png_ptr);
        }
        if (!png_ptr || !info_ptr) {
            png_destroy_write_struct(&png_ptr, &info_ptr);
            fclose(fp);
            throw std::runtime_error("unable to allocate png structure");
        }
        // setup long jump for error return
        if (setjmp(png_jmpbuf(png_ptr))) {
            png_destroy_write_struct(&png_ptr, &info_ptr);
            fclose(fp);
            free(row_pointers);
            free(data);
            throw std::runtime_error("Error writing png");
        }
        // set png file handle
        png_init_io(png_ptr, fp);
        int height = rgba.height();
        int width = rgba.width();
        int color_type = PNG_COLOR_TYPE_RGB_ALPHA;
        int bit_depth = 8;
        // allocate temporary image buffer and row_pointers
        data = static_cast<png_byte *>(
            malloc(height*width*4*sizeof(png_byte)));
        row_pointers = static_cast<png_byte **>(
            malloc(height*sizeof(png_byte *)));
        if (g_text.size() > 0)
            png_set_text(png_ptr, info_ptr, &g_text[0], (int) g_text.size());
        // check output image allocation
        if (!data || !row_pointers) {
            // might as well use the same error handling as libpng...
            longjmp(png_jmpbuf(png_ptr), 1);
        }
        for (int i = 0; i < height; i++) {
            row_pointers[i] = &data[i*width*4];
        }
        // load image data
        for (int i = 0; i < height; i++) {
            for (int j = 0; j < width; j++) {
                int p = (height-1-i)*width+j; // flip image since y is up
                int d = 4*(i*width+j);
                data[d+0] = rgba.r[p];
                data[d+1] = rgba.g[p];
                data[d+2] = rgba.b[p];
                data[d+3] = rgba.a[p];
            }
        }
        // make sure we compress as much as possible
        png_set_filter(png_ptr, PNG_FILTER_TYPE_BASE, PNG_FILTER_NONE);
        png_set_compression_strategy(png_ptr, Z_DEFAULT_STRATEGY);
        png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
        // set basic image parameters
        png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth,
            color_type, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
            PNG_FILTER_TYPE_DEFAULT);
        png_set_sRGB_gAMA_and_cHRM(png_ptr, info_ptr,
            PNG_sRGB_INTENT_RELATIVE);
        // write image info
        png_write_info(png_ptr, info_ptr);
        // write image data
        png_write_image(png_ptr, (png_bytepp) row_pointers);
        // finish advancing file pointer to end of stream (useful?)
        png_write_end(png_ptr, NULL);
        // clean-up and we are done
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        free(data);
        free(row_pointers);
        return 1;
    }

    template <>
    int store(const char *name, const image::rgba<float> &rgba) {
        // temporary image storage
        png_uint_16 ** volatile row_pointers = NULL;
        png_uint_16 * volatile data = NULL;
        // libpng structures
        png_structp png_ptr = NULL;
        png_infop info_ptr = NULL;
        // open file and check if it is a PNG file
        FILE *fp = fopen(name, "wb");
        if (!fp)
            throw std::runtime_error(std::string(name)+": "+strerror(errno));
        // allocate reading structures
        png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL,
            user_error_fn, user_warning_fn);
        if (png_ptr) {
            info_ptr = png_create_info_struct(png_ptr);
        }
        if (!png_ptr || !info_ptr) {
            png_destroy_write_struct(&png_ptr, &info_ptr);
            fclose(fp);
            throw std::runtime_error("unable to allocate png structures");
            return 0;
        }
        // setup long jump for error return
        if (setjmp(png_jmpbuf(png_ptr))) {
            png_destroy_write_struct(&png_ptr, &info_ptr);
            fclose(fp);
            free(row_pointers);
            free(data);
            throw std::runtime_error("Error writing png");
        }
        // set png file handle
        png_init_io(png_ptr, fp);
        int height = rgba.height();
        int width = rgba.width();
        int color_type = PNG_COLOR_TYPE_RGB_ALPHA;
        int bit_depth = 16;
        // allocate temporary image buffer and row_pointers
        data = static_cast<png_uint_16 *>(
            malloc(height*width*4*sizeof(png_uint_16)));
        row_pointers = static_cast<png_uint_16 **>(
            malloc(height*sizeof(png_uint_16 *)));
        if (g_text.size() > 0)
            png_set_text(png_ptr, info_ptr, &g_text[0], (int) g_text.size());
        // try to allocate output image
        if (!data || !row_pointers) {
            // might as well use the same error handling as libpng...
            longjmp(png_jmpbuf(png_ptr), 1);
        }
        for (int i = 0; i < height; i++) {
            row_pointers[i] = &data[i*width*4];
        }
        // load image data
        for (int i = 0; i < height; i++) {
            for (int j = 0; j < width; j++) {
                int p = (height-1-i)*width+j; // flip image since y is up
                int d = 4*(i*width+j);
                //float r = clamp(rgba.r[p]);
                //float b = clamp(rgba.b[p]);
                //float g = clamp(rgba.g[p]);
                float r = srgbcurve(clamp(rgba.r[p]));
                float g = srgbcurve(clamp(rgba.g[p]));
                float b = srgbcurve(clamp(rgba.b[p]));
                data[d+0] = (unsigned short) (65535.f*r);
                data[d+1] = (unsigned short) (65535.f*g);
                data[d+2] = (unsigned short) (65535.f*b);
                data[d+3] = (unsigned short) (65535.f*clamp(rgba.a[p]));
            }
        }
        // set basic image parameters
        png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth,
            color_type, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
            PNG_FILTER_TYPE_DEFAULT);
        png_set_sRGB_gAMA_and_cHRM(png_ptr, info_ptr,
            PNG_sRGB_INTENT_RELATIVE);
        // write image info
        png_write_info(png_ptr, info_ptr);
        // should we flip endianness?
        long int a = 1;
        int swap = (*((unsigned char *) &a) == 1);
        if (swap) {
            png_set_swap(png_ptr);
        }
        // write image data
        png_write_image(png_ptr, (png_bytepp) row_pointers);
        // finish advancing file pointer to end of stream (useful?)
        png_write_end(png_ptr, NULL);
        // clean-up and we are done
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        free(data);
        free(row_pointers);
        return 1;
    }
} } // namespaces png and image
