#ifndef NEHAB_TRANSFORM_H
#define NEHAB_TRANSFORM_H

#include <cmath>
#include <vector>
#include <boost/functional/hash.hpp>
#include "math.h"
#include "point.h"

namespace math
{

// coordinate transformations
class Projective
{
public:
    Projective(void)
    {
        for(int i = 0; i < 3; i++)
            for(int j = 0; j < 3; j++)
                m_m[i][j] = 0.f;
    }

    Projective(float a, float b, float c,
               float d, float e, float f,
               float g, float h, float i)
    {
        m_m[0][0] = a; m_m[0][1] = b; m_m[0][2] = c;
        m_m[1][0] = d; m_m[1][1] = e; m_m[1][2] = f;
        m_m[2][0] = g; m_m[2][1] = h; m_m[2][2] = i;
    }

    static Projective identity(void)
    {
        return Projective(1.f, 0.f, 0.f,
                         0.f, 1.f, 0.f,
                         0.f, 0.f, 1.f);
    }

    static Projective translate(float tx, float ty)
    {
        return Projective(1.f, 0.f, tx,
                         0.f, 1.f, ty,
                         0.f, 0.f, 1.f);
    }

    static Projective rotate(float a)
    {
        a = (math::PI/180.f)*a;
        float ca = cosf(a);
        float sa = sinf(a);
        return Projective(ca, -sa, 0.f,
                         sa, ca, 0.f,
                         0.f, 0.f, 1.f);
    }

    static Projective scale(float sx, float sy)
    {
        return Projective(sx, 0.f, 0.f,
                         0.f, sy, 0.f,
                         0.f, 0.f, 1.f);
    }

    void apply(float *x, float *y) const
    {
        float u[3] = {*x, *y, 1.f};
        float v[3] = {0.f, 0.f, 0.f};
        for (int i = 0; i < 3; i++)
        {
            for (int k = 0; k < 3; k++)
                v[i] += m_m[i][k] * u[k];
        }
        *x = v[0]/v[2];
        *y = v[1]/v[2];
    }

    Projective operator*(const Projective &other) const
    {
        Projective result;
        for(int i = 0; i < 3; i++)
        {
            for(int j = 0; j < 3; j++)
                for(int k = 0; k < 3; k++)
                    result.m_m[i][j] += m_m[i][k] * other.m_m[k][j];
        }

        return result;
    }

    float *operator[](int i) { return m_m[i]; }
    const float *operator[](int i) const { return m_m[i]; }

private:
    float m_m[3][3];
};


struct Transform
{
    float a, b, c, d;
    float tx, ty;
};

using boost::hash_value;
using boost::hash_combine;

inline
size_t hash_value(const Transform &t)
{
    size_t h = hash_value(t.a);
    hash_combine(h, t.b);
    hash_combine(h, t.c);
    hash_combine(h, t.d);
    hash_combine(h, t.tx);
    hash_combine(h, t.ty);
    return h;
}

static inline Transform compose(const Transform &a, const Transform &b)
{
    Transform c;
    c.a = a.a*b.a + a.b*b.c;
    c.b = a.a*b.b + a.b*b.d;
    c.tx = a.tx + a.a*b.tx + a.b*b.ty;
    c.c = a.c*b.a + a.d*b.c;
    c.d = a.c*b.b + a.d*b.d;
    c.ty = a.ty + a.c*b.tx + a.d*b.ty;
    return c;
}

static inline Transform inverse(const Transform &t)
{
    float det = t.a*t.d - t.b*t.c;
    Transform i;
    i.a = t.d/det; i.b = -t.b/det;
    i.c = -t.c/det; i.d = t.a/det;
    i.tx = (t.b*t.ty - t.d*t.tx)/det;
    i.ty = (t.c*t.tx - t.a*t.ty)/det;
    return i;
}

static inline Point apply(const Transform &t, const Point &p)
{
    return Point(t.a*p.x + t.b*p.y + t.tx,
                 t.c*p.x + t.d*p.y + t.ty);
}

} // namespace transform

#endif
