#include <tr1/unordered_map>
#include <gpu/timer.h>
#include "gradients.h"
#include "paint.h"

float4 get_ramp_color(const fd::Ramp &ramp, float t)
{
    fd::Ramp::const_iterator stop;
    for(stop = ramp.begin(); stop!=ramp.end(); ++stop)
    {
        if(t < stop->offset)
            break;
    }

    // before the first stop?
    if(stop == ramp.begin())
        return to_float4(stop->color); // return first color
    // after last stop ?
    if(stop == ramp.end())
        return to_float4(stop[-1].color); // return last color

    fd::Ramp::const_iterator prev = stop-1;

    float ts = (t - prev->offset) / (stop->offset - prev->offset);

    return lerp(to_float4(prev->color), to_float4(stop->color), ts);
}

TexRamp convert_ramp(const fd::Spread &spread, const fd::Ramp &ramp)
{
    TexRamp out;
    float interval = 1.f, last_offset = 0.f;
    for ( int i=1; i<ramp.size(); ++i ) {
        const fd::Stop& stop = ramp[i];
        interval = std::min( interval, stop.offset - last_offset );
    }
    //
    int num_samples = std::max( 128, int(std::ceil(1.f/interval)) );
    float dx = 1.0f/num_samples;
    fd::Ramp::const_iterator si = ramp.begin();
    last_offset = 0.f;
    fd::Color last_color = si->color;
    std::vector<float4> data(num_samples);
    for (unsigned j=0; j<num_samples; ++j ) {
        float x = 0.5f*dx + dx*j;
        if ( x > si->offset ) {
            last_color = si->color;
            last_offset = si->offset;
            ++si;
        }
        const fd::Stop& stop = *si;
        float wx = (x - last_offset)/(stop.offset - last_offset);
        data[j] = lerp( to_float4(last_color), to_float4(stop.color), wx );
    }
    //
    out.ramp_size = data.size();
    cudaChannelFormatDesc channel_desc =
        cudaCreateChannelDesc(32, 32, 32, 32,
                              cudaChannelFormatKindFloat);
    cudaMallocArray(&out.data, &channel_desc, data.size());
    gpu::check_cuda_error("Allocating gradient cudaArray");

    cudaMemcpyToArray(out.data, 0, 0, &data[0],
                      data.size()*sizeof(data[0]),
                      cudaMemcpyHostToDevice);
    gpu::check_cuda_error("Copying gradient data to cudaArray");

    cudaResourceDesc res_desc;
    memset(&res_desc, 0, sizeof(res_desc));
    res_desc.resType = cudaResourceTypeArray;
    res_desc.res.array.array = out.data;

    cudaTextureDesc tex_desc;
    memset(&tex_desc, 0, sizeof(tex_desc));
    tex_desc.filterMode = cudaFilterModeLinear;
    tex_desc.readMode = cudaReadModeElementType;
    // must use normalized coords so that Mirror and Wrap address modes
    // work
    tex_desc.normalizedCoords = true;

    switch(spread)
    {
    case fd::SPREAD_PAD:
        tex_desc.addressMode[0] = cudaAddressModeClamp;
        break;
    case fd::SPREAD_REPEAT:
        tex_desc.addressMode[0] = cudaAddressModeWrap;
        break;
    case fd::SPREAD_REFLECT:
        tex_desc.addressMode[0] = cudaAddressModeMirror;
        break;
    default:
        throw std::runtime_error("Invalid spread type");
    }

    out.tex = 0;
    cudaCreateTextureObject(&out.tex, &res_desc, &tex_desc,
                            NULL);
    gpu::check_cuda_error("Creating gradient textures");

    return out;
}

void free(TexRamp &ramp)
{
    if(ramp.data)
        cudaFreeArray(ramp.data);
    ramp.ramp_size = 0;
    if(ramp.tex != 0)
        cudaDestroyTextureObject(ramp.tex);
}

struct invert_xforms/*{{{*/
{
    invert_xforms(float2 scale, float2 offset)
    {
        math::Projective xform
            = math::Projective::translate(offset.x,offset.y)
                * math::Projective::scale(scale.x,scale.y);

        m_xform.l.x = xform[0][0];
        m_xform.l.y = xform[0][1];
        m_xform.l.z = xform[1][0];
        m_xform.l.w = xform[1][1];
        m_xform.t.x = xform[0][2];
        m_xform.t.y = xform[1][2];
    }


    __device__ void operator()(Gradient &grad)
    {
        grad.transform = inverse(compose(m_xform,grad.transform));
    }

    Transform m_xform;
};/*}}}*/

void convert_gradients(const fd::FontDump &fd,/*{{{*/
                       std::vector<TexRamp> &out_texramp,
                       gpu::vector<Gradient> &out_grads,
                       float2 scale, float2 offset)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("invert grad xforms"));

    std::vector<Gradient> aux_grads;
    aux_grads.reserve(fd.gradients.size());

    std::tr1::unordered_map<int, TexRamp *> ramp_tex_map;

    out_texramp.clear();
    out_texramp.reserve(fd.gradients.size());

    cudaChannelFormatDesc channel_desc =
        cudaCreateChannelDesc(32, 32, 32, 32,
                              cudaChannelFormatKindFloat);

    for(unsigned i=0; i<fd.gradients.size(); ++i)
    {
        const fd::Gradient &in = fd.gradients[i];
        Gradient out;

        switch(in.type)
        {
        case fd::GRADIENT_LINEAR:
            out.type = Gradient::LINEAR;
            out.linear.p1 = to_float2(in.linear.p1);
            out.linear.p2 = to_float2(in.linear.p2);
            break;
        case fd::GRADIENT_RADIAL:
            out.type = Gradient::RADIAL;
            out.radial.c = to_float2(in.radial.c);
            out.radial.f = to_float2(in.radial.f);
            out.radial.r = in.radial.r;
            break;
        default:
            throw std::runtime_error("Invalid gradient type");
        }

        unsigned ramp_hash = in.ramp | (in.spread << 30);
        if(TexRamp *existing = ramp_tex_map[ramp_hash])
        {
            out.ramp = existing->tex;
            out.inv_ramp_size = 1.f/existing->ramp_size;
        }
        else
        {
            TexRamp texramp = convert_ramp(in.spread, fd.ramps[in.ramp]);

            out.inv_ramp_size = 1.f/texramp.ramp_size;
            out.ramp = texramp.tex;

            out_texramp.push_back(texramp);
            ramp_tex_map[ramp_hash] = &out_texramp.back();
        }

        out.transform.l.x = in.transform.a;
        out.transform.l.y = in.transform.b;
        out.transform.l.z = in.transform.c;
        out.transform.l.w = in.transform.d;
        out.transform.t.x = in.transform.tx;
        out.transform.t.y = in.transform.ty;

        aux_grads.push_back(out);
    }

    out_grads.copy_from_host(aux_grads);

    thrust::for_each(out_grads.begin(), out_grads.end(), invert_xforms(scale,offset));

    printf("# ramp textures created: %lu\n",out_texramp.size());
}/*}}}*/

__global__ void gpu_invert_paint_xforms(Paint **paints, int npaints,
                                        const Transform *orig_xforms,
                                        Transform xf)
{
    int id = blockIdx.x*blockDim.x + threadIdx.x;
    if(id >= npaints)
        return;

    Paint *_paint = paints[id];

    if(_paint == NULL)
        return;

    if(_paint->type == Paint::GRAD_LINEAR)
    {
        PaintLinearGrad &paint = _paint->as<PaintLinearGrad>();
        paint.set_xform(inverse(compose(xf, orig_xforms[id])));
    }
    else if(_paint->type == Paint::GRAD_RADIAL)

    {
        //PaintRadialGrad &paint = _paint->as<PaintRadialGrad>();
        //paint.set_xform(inverse(compose(xf, orig_xforms[id])));
    }
    else if(_paint->type == Paint::TEXTURE)

    {
        PaintTexture &paint = _paint->as<PaintTexture>();
        paint.set_xform(inverse(compose(xf, orig_xforms[id])));
    }
}

void invert_paint_xforms(gpu::vector<Paint *> &paints,
                         const gpu::vector<Transform> &base_xforms,
                         float2 scale, float2 off)
{
    math::Projective xform
        = math::Projective::translate(off.x,off.y)
            * math::Projective::scale(scale.x,scale.y);

    Transform xf;

    xf.l.x = xform[0][0];
    xf.l.y = xform[0][1];
    xf.l.z = xform[1][0];
    xf.l.w = xform[1][1];
    xf.t.x = xform[0][2];
    xf.t.y = xform[1][2];

    dim3 bdim(256),
         gdim((paints.size()+bdim.x-1)/bdim.x);

    gpu_invert_paint_xforms<<<gdim,bdim>>>(&paints, paints.size(), &base_xforms, xf);
}

