#include <gpu/timer.h>
#include <gpu/image.h>
#include "image_util.h"
#include "color_util.h"

const int RAST_BW = 32,
          RAST_BH = 8;

__global__ void gpu_separate_channels(gpu::image_ptr<const float4> img,
                                      float *red, float *green, float *blue,
                                      float *alpha)
{
    int x = blockIdx.x*RAST_BW + threadIdx.x,
        y = blockIdx.y*RAST_BH + threadIdx.y;

    if(x >= img.width() || y >= img.height())
        return;

    int idx = y*img.width()+x;

    img += img.offset_at(x,y);

    float4 pix = *img;

    red[idx] = pix.x;
    green[idx] = pix.y;
    blue[idx] = pix.z;
    alpha[idx] = pix.w;
}

void copy_to_host(image::rgba<> &h_out, const gpu::image<float4> &d_in)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("copy image to host"));

    gpu::vector<float> d_red, d_green, d_blue, d_alpha;

    d_red.resize(d_in.width()*d_in.height());
    d_green.resize(d_in.width()*d_in.height());
    d_blue.resize(d_in.width()*d_in.height());
    d_alpha.resize(d_in.width()*d_in.height());

    dim3 gdim((d_in.width()+RAST_BW-1)/RAST_BW,
              (d_in.height()+RAST_BH-1)/RAST_BH),
         bdim(RAST_BW, RAST_BH);

    gpu_separate_channels<<<gdim, bdim>>>(&d_in, d_red.data(), d_green.data(),
                                          d_blue.data(), d_alpha.data());

    d_red.copy_to_host(h_out.r, d_red.size());
    d_green.copy_to_host(h_out.g, d_green.size());
    d_blue.copy_to_host(h_out.b, d_blue.size());
    d_alpha.copy_to_host(h_out.a, d_alpha.size());
}

void copy_to_host(image::rgba<> &h_out, const gpu::image_ptr<float4> d_in)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("copy image to host"));

    gpu::vector<float> d_red, d_green, d_blue, d_alpha;

    d_red.resize(d_in.width()*d_in.height());
    d_green.resize(d_in.width()*d_in.height());
    d_blue.resize(d_in.width()*d_in.height());
    d_alpha.resize(d_in.width()*d_in.height());

    dim3 gdim((d_in.width()+RAST_BW-1)/RAST_BW,
              (d_in.height()+RAST_BH-1)/RAST_BH),
         bdim(RAST_BW, RAST_BH);

    gpu_separate_channels<<<gdim, bdim>>>(d_in, d_red.data(), d_green.data(),
                                          d_blue.data(), d_alpha.data());

    d_red.copy_to_host(h_out.r, d_red.size());
    d_green.copy_to_host(h_out.g, d_green.size());
    d_blue.copy_to_host(h_out.b, d_blue.size());
    d_alpha.copy_to_host(h_out.a, d_alpha.size());
}

__global__ void gpu_lrgb2srgb(float4* img, unsigned width, unsigned height, unsigned stride )
{
    int x = blockIdx.x*RAST_BW + threadIdx.x,
        y = blockIdx.y*RAST_BH + threadIdx.y;
    if(x >= width || y >= height)
        return;
    int idx = y*stride +x;
    float4 color = img[idx];
    img[idx] = lrgb2srgb(color);
}

void lrgb2srgb(gpu::image_ptr<float4>& d_img)
{
    dim3 gdim((d_img.width()+RAST_BW-1)/RAST_BW,
              (d_img.height()+RAST_BH-1)/RAST_BH),
         bdim(RAST_BW, RAST_BH);
    gpu_lrgb2srgb<<<gdim, bdim>>>(&d_img, d_img.width(), d_img.height(), d_img.rowstride());
}

