#ifndef GPURASTER_IMPLICIT_H
#define GPURASTER_IMPLICIT_H

#include <gpu/box.h>
#include <cassert>
#include "icurve.h"
#include "segtype.h"

class Paint;

template <int N=0> struct ImplicitSegment;
struct PathData;

enum SegFlags
{
    SEGFLAG_UP = 1,
    SEGFLAG_DOWN = 2,
    SEGFLAG_SHORTCUT = 3,

    SEGFLAG_SIZE = ( 0x07F << 2 ),

    SEGFLAG_PAINT_SIZE = ( 0x07F << 9 ),
};

#define SEG_GET_SIZE(f) ((f&SEGFLAG_SIZE) >> 2)
#define SEG_SET_SIZE(f) ((f&0x7F) << 2)
#define SEG_GET_PAINT_SIZE(f) ((f&SEGFLAG_PAINT_SIZE) >> 9)
#define SEG_SET_PAINT_SIZE(f) ((f&0x7F) << 9)

struct SegmentIndex {
    unsigned idx_data;
    int winding:8;
    unsigned type:8;
    unsigned geom_size:7;
    unsigned paint_size:7;
    unsigned flags:2;

    __device__
    SegType get_type() const {
        return static_cast<SegType>( type );
    }
};

struct ImplicitSegmentBase
{
    SegType type;
    unsigned short flags;
};

template<>
struct ImplicitSegment<SEG_PATH> : ImplicitSegmentBase
{
    const PathData *pdata;
};

struct ImplicitCurveSegment : ImplicitSegmentBase
{
    const ImplicitCurveBase *B;

    __device__ bool is_shortcut() const
    {
        return flags & (SEGFLAG_UP|SEGFLAG_DOWN);
    }

    __device__
    float2 shortcut_base() const
    {
        assert(is_shortcut());
        return (flags & SEGFLAG_UP) ? B->endpoint() : B->begpoint();
    }

    __device__
    int shortcut_hit(const float2 &pt) const
    {
        float2 p = shortcut_base();

        // below bbox?
        if(pt.y < p.y)
            return 0;

        //return pt.x < p().x; // hit if to the left
        if (pt.x < p.x) { // hit if to the left
            return (flags & SEGFLAG_UP) ? 1 : -1;
        }
        return 0;
    }
};

template<int N>
struct ImplicitSegment : ImplicitCurveSegment
{
    __device__
    const ImplicitCurve<N> *B() const
    {
        return static_cast<const ImplicitCurve<N> *>(ImplicitCurveSegment::B);
    }
};

struct StackSegment : ImplicitSegmentBase
{
};

template<>
struct ImplicitSegment<SEG_COMMIT> : StackSegment
{
};

template<>
struct ImplicitSegment<SEG_PUSH> : StackSegment
{
};

template<>
struct ImplicitSegment<SEG_POP_FILL> : StackSegment
{
};

template<>
struct ImplicitSegment<SEG_POP_CLIP> : StackSegment
{
};

template<>
struct ImplicitSegment<> : ImplicitSegmentBase
{
    const void *data;

    template<int N>
    HOSTDEV ImplicitSegment<N> &as()
    {
        return *reinterpret_cast<ImplicitSegment<N> *>(this);
    }

    template<int N>
    HOSTDEV const ImplicitSegment<N> &as() const
    {
        return *reinterpret_cast<const ImplicitSegment<N> *>(this);
    }

    HOSTDEV StackSegment &as_stack()
    {
        return *reinterpret_cast<StackSegment *>(this);
    }

    HOSTDEV const ImplicitCurveSegment &as_curve()
    {
        return *reinterpret_cast<const ImplicitCurveSegment *>(this);
    }
};
struct SegmentsPtr
{
    SegmentsPtr() 
        : size(0)
        , type(NULL)
        , winding(NULL)
        , flags(NULL)
        , idx_data(NULL)
        , _data(NULL)
    {
    }

    unsigned size;

    SegType *type;
    int *winding;
    unsigned short *flags;
    unsigned *idx_data;
    unsigned char *_data;

    template <class T>
    __device__ T *data(unsigned idseg) const
    {
        return reinterpret_cast<T *>(_data+idx_data[idseg]);
    }
};

struct const_SegmentsPtr
{
    __device__ __host__
    const_SegmentsPtr() 
        : size(0)
        , type(NULL)
        , idx_data(NULL)
        , _data(NULL)
        , winding(NULL)
        , flags(NULL)
    {
    }

    const_SegmentsPtr(const SegmentsPtr &that)
        : size(that.size)
        , type(that.type)
        , idx_data(that.idx_data)
        , _data(that._data)
        , winding(that.winding)
        , flags(that.flags)
    {
    }

    unsigned size;

    const SegType *type;
    const unsigned *idx_data;
    const unsigned char *_data;
    const int *winding;
    const unsigned short *flags;

    template <class T>
    __device__ const T *data(unsigned idseg) const
    {
        return reinterpret_cast<const T *>(_data+idx_data[idseg]);
    }
};

struct Segments
{
    gpu::vector<SegType> type;
    gpu::vector<unsigned> idx_data;
    gpu::vector<int> winding;
    gpu::vector<unsigned short> flags;

    gpu::vector<unsigned char> _data;

    size_t size() const { return type.size(); }

    void clear()
    {
        type.clear();
        idx_data.clear();
        winding.clear();
        flags.clear();
    }

    void resize(size_t s)
    {
        type.resize(s);
        idx_data.resize(s);
        winding.resize(s);
        flags.resize(s);
    }

    SegmentsPtr operator&()
    {
        SegmentsPtr ret;
        ret.size = size();

        ret.type = type.data();
        ret.winding = winding.data();
        ret.flags = flags.data();
        ret.idx_data = idx_data.data();
        ret._data = _data.data();

        return ret;
    }

    const const_SegmentsPtr operator&() const
    {
        const_SegmentsPtr ret;
        ret.size = size();

        ret.type = type.data();
        ret.winding = winding.data();
        ret.flags = flags.data();
        ret.idx_data = idx_data.data();
        ret._data = _data.data();

        return ret;
    }
};

inline void swap(Segments &a, Segments &b)
{
    swap(a.type, b.type);
    swap(a.winding, b.winding);
    swap(a.flags, b.flags);
    swap(a.idx_data, b.idx_data);
    swap(a._data, b._data);
}

#endif
