#ifndef GPURASTER_IMPLICITIZE3_H
#define GPURASTER_IMPLICITIZE3_H

#include <gpu/solve.h>
#include <gpu/util.h>
#include "icurve3.h"
#include "implicitize1.h" // for find_line_coefficients

enum CubicType
{
    CUBIC_UNKNOWN,
    CUBIC_SERPENTINE,
    CUBIC_LOOP,
    CUBIC_CUSP_AT_INFINITY,
    CUBIC_QUADRATIC,
    CUBIC_LINE,
    CUBIC_POINT
};

template<>
struct __align__(8) ImplicitizationContext<3>
{
    gpu::math::Matrix<float,3,3> M;
    gpu::array<float,4> d;
    float2 base;
    CubicType cubic_type;
};


/*
 * Create matrix of coeficients in the power basis:
 *  C = M3 * B, where
 *
 *     |  1  0  0  0 |
 * M3 =| -3  3  0  0 |
 *     |  3 -6  3  0 |
 *     | -1  3 -3  1 |
 *
 *     |  b.x  b.y 1 |   | a b 1 |
 * B = | c1.x c1.y 1 | = | c d 1 |,
 *     | c2.y c2.y 1 |   | e f 1 |
 *     |  e.x  e.y 1 |   | g h 1 |
 *
 *     |      a          b     1 |
 * C = |   -3a+3c     -3b+3d   0 |
 *     |  3a-6c+3e   3b-6d+3f  0 |
 *     | -a+3c-3e+g -b+3d-3f+h 0 |
 *  this notation follows Loop and Blinn 2005 (section 2.2).
 */
__device__ void coeficients_matrix(const Curve<3> &B,
                                   gpu::math::Matrix<float,4,3> &m)
{
    float a = B[0].x, b = B[0].y;
    float c = B[1].x, d = B[1].y;
    float e = B[2].x, f = B[2].y;
    float g = B[3].x, h = B[3].y;

    m[0][0] =                 a;  m[0][1] =                 b;  m[0][2] = 1;
    m[1][0] =       -3.0f*(a-c);  m[1][1] =       -3.0f*(b-d);  m[1][2] = 0;
    m[2][0] = 3.0f*(a-2.0f*c+e);  m[2][1] = 3.0f*(b-2.0f*d+f);  m[2][2] = 0;
    m[3][0] =    g+3.0f*(c-e)-a;  m[3][1] =    h+3.0f*(d-f)-b;  m[3][2] = 0;
}

/*              | a b 1 |
 * receives C = | c d 0 | and
 *              | e f 0 |
 *              | g h 0 |
 * computes:
 *         | c d 0 |               | a b 1 |
 * d0 = Det| e f 0 | = 0, d1 = -Det| e f 0 |,
 *         | g h 0 |               | g h 0 |
 *
 *         | a b 1 |           | a b 1 |
 * d2 = Det| c d 0 |, d3 = -Det| c d 0 |.
 *         | g h 0 |           | e f 0 |
 */
__device__ void find_discriminant_coeficients(const gpu::math::Matrix<float,4,3> &m,
                                          gpu::array<float,4> &dc)
{
    float c = m[1][0], d = m[1][1];
    float e = m[2][0], f = m[2][1];
    float g = m[3][0], h = m[3][1];

    dc[0] = 0;
    dc[1] = f*g - e*h;
    dc[2] = c*h - d*g;
    dc[3] = d*e - c*f;
}

/*
 * find the F matrix for the serpentine case. See Loop and Blinn section 4.4(1)
 * and the Mathematica implementation.
 * Also finds the sign for the implicit function.
 *
 * this case also handles "cusp with cusp at infinity"
 */
__device__ void find_serpentine_matrix(const gpu::array<float,4> &d,
                                       float tmtl[2],
                                       gpu::math::Matrix<float,4,4> &m,
                                       gpu::array<float2,2> &ts)
{
    float tl = tmtl[1],
          sl = 2*d[1];

    float inv_norm = 1 / sqrt(tl*tl + sl*sl);
    tl = tl * inv_norm;
    sl = sl * inv_norm;

    float tm = tmtl[0],
          sm = 2*d[1];

    inv_norm = 1 / sqrt(tm*tm + sm*sm);
    tm = tm * inv_norm;
    sm = sm * inv_norm;

    // now assemble the matrix
    // {
    // {tl*tm, tl^3, tm^3, 1},
    // {-sm*tl - sl*tm, -3 sl*tl^2, -3*sm*tm^2, 0},
    // {sl*sm, 3 sl^2 tl, 3 sm^2 tm, 0},
    // {0, -sl^3, -sm^3, 0}
    // }
    //
    m[0][0] =        tl*tm; m[0][1] =    tl*tl*tl; m[0][2] =    tm*tm*tm; m[0][3] = 1;
    m[1][0] = -sm*tl-sl*tm; m[1][1] = -3*sl*tl*tl; m[1][2] = -3*sm*tm*tm; m[1][3] = 0;
    m[2][0] =        sl*sm; m[2][1] =  3*sl*sl*tl; m[2][2] =  3*sm*sm*tm; m[2][3] = 0;
    m[3][0] =            0; m[3][1] =   -sl*sl*sl; m[3][2] =   -sm*sm*sm; m[3][3] = 0;

    // save the roots for later (tl,sl) and (tm,sm)
    ts[0] = make_float2(tl,sl);
    ts[1] = make_float2(tm,sm);
    // other inflection points:
    // tn == 1, sn == 0
}

/*
 * find the F matrix for the loop case. See Loop and Blinn section 4.4(1)
 * and the Mathematica implementation.
 */
__device__ void find_loop_matrix(const gpu::array<float,4> &d,
                                 float tetd[2],
                                 gpu::math::Matrix<float,4,4> &m,
                                 gpu::array<float2,2> &ts)
{
    float td = tetd[1],
          sd = 2*d[1];

    float inv_norm = 1 / sqrt(td*td + sd*sd);
    td = td * inv_norm;
    sd = sd * inv_norm;

    float te = tetd[0],
          se = 2*d[1];

    inv_norm = 1 / sqrt(te*te + se*se);
    te = te * inv_norm;
    se = se * inv_norm;


    // assemble the matrix
    // {
    // {td*te, td^2 te, td te^2, 1},
    // {-se*td - sd*te, -se*td^2 - 2 sd te td, -sd te^2 - 2 se td te, 0},
    // {sd*se, te sd^2 + 2 se td sd, td se^2 + 2 sd te se, 0},
    // {0, -sd^2 se, -sd se^2, 0}
    // }
    m[0][0] =        td*te; m[0][1] =             td*td*te; m[0][2] =             td*te*te; m[0][3] = 1;
    m[1][0] = -se*td-sd*te; m[1][1] = -se*td*td-2*sd*te*td; m[1][2] = -sd*te*te-2*se*td*te; m[1][3] = 0;
    m[2][0] =        sd*se; m[2][1] =  te*sd*sd+2*se*td*sd; m[2][2] =  td*se*se+2*sd*te*se; m[2][3] = 0;
    m[3][0] =            0; m[3][1] =            -sd*sd*se; m[3][2] =            -sd*se*se; m[3][3] = 0;

    // save the roots for later (td,sd) and (te,se)
    ts[0] = make_float2(td,sd);
    ts[1] = make_float2(te,se);
}

/*
 * find the F matrix for the cusp at infinity case. See Loop and Blinn section 4.4(1)
 * and the Mathematica implementation.
 * Also finds the sign for the implicit function.
 */
__device__ void find_cusp_at_infinity_matrix(const gpu::array<float,4> &d,
                                             gpu::math::Matrix<float,4,4> &m,
                                             gpu::array<float2,2> &ts)
{
    // compute and normalize the pair (tl, sl)
    float tl = d[3];
    float sl = 3*d[2];
    float inv_norm = 1 / sqrt(tl*tl + sl*sl);
    tl = tl * inv_norm;
    sl = sl * inv_norm;

    // assemble the matrix
    // {
    // {tl, tl^3, 1,1},
    // {-sl,-3sl tl^2, 0,0},
    // {0, 3sl^2 tl, 0,0},
    // {0,-sl^3,0,0}
    // }
    m[0][0] =  tl; m[0][1] =    tl*tl*tl; m[0][2] = 1; m[0][3] = 1;
    m[1][0] = -sl; m[1][1] = -3*sl*tl*tl; m[1][2] = 0; m[1][3] = 0;
    m[2][0] =   0; m[2][1] =  3*sl*sl*tl; m[2][2] = 0; m[2][3] = 0;
    m[3][0] =   0; m[3][1] =   -sl*sl*sl; m[3][2] = 0; m[3][3] = 0;

    // save the roots for later (tl,sl)
    ts[0] = make_float2(tl,sl);
    ts[1] = make_float2(2,1); // won't be used, enough to skip root creation
    // other inflection points are:
    // tm == 1, sm == 0
    // tn == 1, sn == 0
}

/*
 * Computes the final matrix for the cubic
 * M3^{-1}*F =
 * | 1  0   0  0 | | a b c 1 |
 * | 1 1/3  0  0 | | d e f 0 |
 * | 1 2/3 1/3 0 |*| g h i 0 |
 * | 1  1   1  1 | | 0 j k 0 |
 */
__device__ void premultiply_f_by_m3_inverse(const gpu::math::Matrix<float,4,4> &fm,
                                            gpu::math::Matrix<float,4,4> &m)
{
    float a = fm[0][0], b = fm[0][1], c = fm[0][2]; // skip 0,3
    float d = fm[1][0], e = fm[1][1], f = fm[1][2]; // skip 1,3
    float g = fm[2][0], h = fm[2][1], i = fm[2][2]; // skip 2,3
    float               j = fm[3][1], k = fm[3][2]; // skip 3,0 and 3,3

    //{
    //{a,b,c,1},
    //{a+d/3,b+e/3,c+f/3,1},
    //{a+(2 d)/3+g/3,b+(2 e)/3+h/3,c+(2 f)/3+i/3,1},
    //{a+d+g,b+e+h+j,c+f+i+k,1}
    //}
    m[0][0] =           a; m[0][1] =           b; m[0][2] =           c; m[0][3] = 1;
    m[1][0] =       a+d/3; m[1][1] =       b+e/3; m[1][2] =       c+f/3; m[1][3] = 1;
    m[2][0] = a+2*d/3+g/3; m[2][1] = b+2*e/3+h/3; m[2][2] = c+2*f/3+i/3; m[2][3] = 1;
    m[3][0] =       a+d+g; m[3][1] =     b+e+h+j; m[3][2] =     c+f+i+k; m[3][3] = 1;
}

/*
 * this function receives a degenerated cubic that is in fact a quadratic.
 */
__device__ void cubic_find_degenerated_as_quadratic_matrix(
    const Curve<3> &B, gpu::math::Matrix<float,4,4> &f)
{
    // since we have a quadratic, we begin by finding its control points.
    // the beggining and ending points are the same
    const float2 &b = B[0], &c1 = B[1], &c2 = B[2], &e = B[3];
    // we find the control point c for the quadratic
    float2 c = 0.25f*( -b + 3.0f*c1 + 3.0f*c2 - e );
    // linear interpolation for c1
    float t1 = length(c1-b)/length(c-b);
    // linear interpolation for c2
    float t2 = length(c2-e)/length(c-e);
    // the first line of f contains the coefficients for
    // the point b.
    f[0][0] = 0.0f; f[0][1] = 0.0f; f[0][2] = 0.0f; f[0][3] = 1.0f;
    // the second and third lines are the coefficients for
    // the control points c1 and c2. c1 is in the b->c line,
    // (1.0f-t1)*{ 0, 0, 0, 1 } + t1*{ 0.5, 0, 0.5, 1 }
    f[1][0] = 0.5f*t1; f[1][1] = 0.0f; f[1][2] = 0.5f*t1; f[1][3] = 1.0f;
    // c2 is in the e->c line
    // (1.0f-t2)*{1, 1, 1, 1} * t2*{ 0.5, 0, 0.5, 1 }
    float tmp1 = (1.0f-t2)*1.0f + t2*0.5f;
    float tmp2 = (1.0f-t2)*1.0f;
    f[2][0] = tmp1; f[2][1] = tmp2; f[2][2] = tmp1; f[2][3] = 1.0f;
    // coefficients for e
    f[3][0] = 1.0f; f[3][1] = 1.0f; f[3][2] = 1.0f; f[3][3] = 1.0f;

#if 0
    // find the sing for the segment, we use the same method we have used
    // for the "real" quadratic (see classify_quadratic function)
    sign = 1.0f;
    // by default the negative part of the quadratic is the convex one, but
    // for cubics the standard is to have the negative part to the right-side
    // of the curve in parametric space. we need to convert from the first to
    // the second case.
    if (ccw(c, b, e) > 0)
        sign = -1.0f;
#endif
}

__device__ CubicType classify(const Curve<3> &B,
                              gpu::math::Matrix<float,4,4> &m,
                              gpu::array<float2,2> &ts,
                              gpu::array<float,4> &d)
{
    // find the coeficients matrix in the power basis
    // we use this coeficients to classisfy the cubic
    // see Loop and Blin 2005 section 4.2
    gpu::math::Matrix<float,4,3> cm;
    coeficients_matrix(B, cm);

    // now find the coeficients of the discriminant
    // polynomial
    find_discriminant_coeficients(cm, d);

    // the classification follows section 4.4 of Loop and Blinn 2005.
    CubicType type = CUBIC_UNKNOWN;
    // space for the F matrix
    gpu::math::Matrix<float,4,4> f;
    //if ( d[1] != 0 ) {
    //TODO: we should find a better constant value
    if(abs(d[1]) > 1e-6)
    {
        // I've decided to change the original quadraticsolve to receive
        // the delta var as a argument. The reason is that the next two
        // quadratic have the same delta, except for a constant multiplication.
        // That is: delta_serp = -1/3 * delta_loop.
        // Calculating the deltas using the A,B and C coeficients from the
        // respective equations gives disagreeing results such as both deltas < 0.
        // The first equation is: t^2 -2*d2 + (4/3)*d1*d3,
        // the second is: t^2 - 2*d2 + 4*(d2^2 -d1*d3).
        // Using Blinn's convention we have delta = b*b - a*c,
        // where b = B/2.

        float delta = 3*d[2]*d[2] - 4*d[1]*d[3];
        if(delta >= 0)
        {
            float tmtl[2];
            unsigned n = gpu::quadratic_solve(1, -2*d[2], (4.f/3)*d[1]*d[3], (1.f/3)*delta, tmtl);

            if(tmtl[0] > tmtl[1])
                gpu::swap(tmtl[0], tmtl[1]);
            // serpentine and cusp with inflection at infinity
            find_serpentine_matrix(d, tmtl, f, ts);
            type = CUBIC_SERPENTINE;
        }
        else
        {
            // loop
            float tetd[2];
            gpu::quadratic_solve(1, -2*d[2], 4*(d[2]*d[2] - d[1]*d[3]), -delta, tetd);
            if(tetd[0] > tetd[1])
                gpu::swap(tetd[0], tetd[1]);

            find_loop_matrix(d, tetd, f, ts);
            type = CUBIC_LOOP;
        }
        premultiply_f_by_m3_inverse(f, m);
    }
    else
    {
        if(d[2] != 0)
        {
            // cusp with cusp at infinity
            find_cusp_at_infinity_matrix(d, f, ts);
            type = CUBIC_CUSP_AT_INFINITY;
            premultiply_f_by_m3_inverse(f, m);
        }
        // degenerated cubics...
        else if(d[3] != 0)
        {
            cubic_find_degenerated_as_quadratic_matrix(B, m);
            // this cubic is a quadratic
            type = CUBIC_QUADRATIC;
        }
        else
        {
            f = gpu::math::identity<float,4,4>();
            if((B[0] == B[1] && B[1] == B[3]) ||
                (B[0] == B[2] && B[2] == B[3]))
            {
//                printf("cubic degererated as point.\n");
                type = CUBIC_POINT;
            }
            else
            {
                //errorf(("cubic degererated as line.\n"));
                // a line or a point
                type = CUBIC_LINE;
            }
        }
    }
    return type;
}

__device__ inline
float2 cubic_find_E(const Curve<3> &B)
{
    float a11 = B[1].x - B[0].x, a12 = B[3].x - B[2].x;
    float a21 = B[1].y - B[0].y, a22 = B[3].y - B[2].y;
    float det = a11*a22 - a12*a21;
    //TODO: find a better constant
    float alpha;
    if (abs(det) < 1e-3 )
    {
        // A==B(and C!=D) or C==D(and A!=B).
        alpha = 1.0f;
        if (sqrnorm(B[0]-B[1]) < sqrnorm(B[2]-B[3]))
        {
            // A == B
            return B[2];
        }
        else
        {
            // C==D
            return B[1];
        }
    }
    float b1 = B[3].x - B[0].x, b2 = B[3].y - B[0].y;
    /* solve the system
     * | a11 a12 ||alpha|   | b1 |
     * | a21 a22 || beta| = | b2 | */
    alpha = - (a22*b1 - a12*b2)/(a12*a21 - a11*a22);
    return (1.0f-alpha)*B[0] + alpha * B[1];
}

/*
 * finds the determinant of the matrix
 * | p0.x p1.x p2.x |   | a b c |
 * | p0.y p1.y p2.y | = | d e f |
 * |   1    1    1  |   | 1 1 1 |
 */
__device__ float cubic_determinant(const float2 &p0, const float2 &p1, const float2 &p2)
{
    float a = p0.x, b = p1.x, c = p2.x;
    float d = p0.y, e = p1.y, f = p2.y;
    return -b*d + c*d + a*e - c*e - a*f + b*f;
}

/* receives the coefficient matrix from the implicitization
 *      | a b c 1 |
 * cm = | d e f 1 |,
 *      | g h i 1 |
 *      | j k l 1 |
 * and computes a 3x3 matrix selecting the lines i,j,k, ignoring
 * the 1's column and then traspose it.
 */
template <int i, int j, int k>
__device__
void find_klm_matrix_and_transpose(
    const gpu::math::Matrix<float,4,4> &cm,
    gpu::math::Matrix<float,3,3> &klm_matrix)
{
    // copy line i from cm into col 0 of klm_matrix
    for (unsigned l=0; l<3; l++)
        klm_matrix[l][0] = cm[i][l];
    // copy line j from cm into col 1 of klm_matrix
    for (unsigned l=0; l<3; l++)
        klm_matrix[l][1] = cm[j][l];
    // copy line k from cm into col 2 of klm_matrix
    for (unsigned l=0; l<3; l++)
        klm_matrix[l][2] = cm[k][l];
}

/*
 * receives a sequence of points defining a quadratic bezier:
 * b=pts[0], c=pts[1], e=pts[2] and calculates the inverse of the
 * following matrix
 *     | b.x c.x e.x |   | a b c |
 * M = | b.y c.y e.y | = | d e f |
 *     |  1   1   1  |   | 1 1 1 |
 * M^-1 transforms world coordinates [x y 1]^t into baricentric
 * coordinates [ u v (1-u-v) ]^t.
 * The parameter m should be allocated by the caller.
 */
__device__ void create_baricentric_matrix(const float2 &p0, const float2 &p1, const float2 &p2,
                                          gpu::math::Matrix<float,3,3> &m)
{
    float a = p0.x, b = p1.x, c = p2.x;
    float d = p0.y, e = p1.y, f = p2.y;
    // from Mathematica:
    // { {(e-f)/(-b d+c d+a e-c e-a f+b f),(-b+c)/(-b d+c d+a e-c e-a f+b f),(-c e+b f)/(-b d+c d+a e-c e-a f+b f)},
    //   {(-d+f)/(-b d+c d+a e-c e-a f+b f),(a-c)/(-b d+c d+a e-c e-a f+b f),(c d-a f)/(-b d+c d+a e-c e-a f+b f)},
    //   {(d-e)/(-b d+c d+a e-c e-a f+b f),(-a+b)/(-b d+c d+a e-c e-a f+b f),(-b d+a e)/(-b d+c d+a e-c e-a f+b f)}}
    float den = 1.0f/(-b*d + c*d + a*e - c*e - a*f + b*f);
    m[0][0] = (e-f)*den; m[0][1] = (c-b)*den; m[0][2] = (b*f-c*e)*den;
    m[1][0] = (f-d)*den; m[1][1] = (a-c)*den; m[1][2] = (c*d-a*f)*den;
    m[2][0] = (d-e)*den; m[2][1] = (b-a)*den; m[2][2] = (a*e-b*d)*den;
}

__device__
CubicType classify(const Curve<3> &B,
                   ImplicitizationContext<3> &ctx,
                   gpu::array<float2,2> &ts)
{
    gpu::math::Matrix<float,4,4> im;
    gpu::math::Matrix<float,3,3> bm, klm_matrix;
    const float2 &b = B[0], &c1 = B[1], &c2 = B[2], &e = B[3];

    // classify this cubic
    CubicType cubic_type = classify(B, im, ts, ctx.d);
    // now im contains the matrix with the coefficients for the cubic,
    // the first line for b, the second for c1, third for c2 and the
    // fourth for e.
    // Now we find wich subset of the control points (b, c1, c2, e)
    // forms the better triangle
    //
    // if b!=e we choose between c1 and c2. This may not be the optimal
    // choice because b and e may be close to each other.
    // TODO: find a better constant
    if(length(b-e) > 1e-5)
    {
        float v1 = cubic_determinant(b, c1, e);
        float v2 = cubic_determinant(b, c2, e);
        if (abs(v1) > abs(v2))
        {
            ctx.base = c1;
            create_baricentric_matrix(b-c1, c1-c1, e-c1, bm);
            find_klm_matrix_and_transpose<0,1,3>(im, klm_matrix);
        }
        else
        {
            ctx.base = c2;
            create_baricentric_matrix(b-c2, c2-c2, e-c2, bm);
            find_klm_matrix_and_transpose<0,2,3>(im, klm_matrix);
        }
    }
    else
    {
        float v1 = cubic_determinant(b, c1, c2);
        float v2 = cubic_determinant(e, c1, c2);
        if (abs(v1) > abs(v2))
        {
            ctx.base = c1;
            create_baricentric_matrix(b-c1, c1-c1, c2-c1, bm);
            find_klm_matrix_and_transpose<0,1,2>(im, klm_matrix);
        }
        else
        {
            ctx.base = c1;
            create_baricentric_matrix(e-c1, c1-c1, c2-c1, bm);
            find_klm_matrix_and_transpose<3,1,2>(im, klm_matrix);
        }
    }

    // multiply the klm_matrix by the baricentric transformation
    // this way we will have:
    // [k l m]^t = K*B[x y 1]^t, where K is klm_matrix and B is bm.
    // B will transform the vector [x y 1]^t into [a b c]^t which are
    // the baricentric coodinates relatives to the triagle chosen above.
    // K*[a b c]^t will make a convex combination of the K columns (given
    // by the 2005 paper) resulting in the desired values for [k l m]^t.

    ctx.M = klm_matrix * bm;
    return cubic_type;
}

__device__ void implicitize(const Curve<3> &B, ImplicitCurve<3> &iB,
                            float t0, float t1,
                            const ImplicitizationContext<3> &ctx)
{
    iB.bbox = get_xy_mono_bbox(B);

    assert(iB.flags == 0);

    if(B[0].x < B[3].x)
        iB.flags |= CFLAG_GOING_RIGHT;

    if(B[0].y < B[3].y)
        iB.flags |= CFLAG_GOING_UP;

    iB.M = ctx.M;
    iB.base = ctx.base;
    iB.E = cubic_find_E(B);

    if(ccw(iB.E, B[0], B[3]) < 0)
        iB.flags |= CFLAG_GEOM_TO_LEFT;

    if(iB.going_up() != iB.geom_to_left())
        iB.flags |= CFLAG_OUT_IS_LEFT;

    // find the sign
    if(ctx.cubic_type == CUBIC_LOOP)
    {
        // for loops we need to find the sign for each segment :-/
        // we compare the values of H(t/s,1) at each point of the
        // segment. see Loop and Blin 2005, section 4.4 The loop, last
        // paragraph.

        const gpu::array<float,4> &d = ctx.d;

        float h0 = d[3]*d[1] - d[2]*d[2] + d[1]*d[2]*t0 - d[1]*d[1]*t0*t0;
        float h1 = d[3]*d[1] - d[2]*d[2] + d[1]*d[2]*t1 - d[1]*d[1]*t1*t1;
        if (abs(h0) > abs(h1))
        {
            if(ctx.d[1]*h0 > 0)
                iB.flags |= CFLAG_NEGATIVE;
        }
        else
        {
            if(ctx.d[1]*h1 > 0)
                iB.flags |= CFLAG_NEGATIVE;
        }
    }
    // we must flip the sign if d[1] < 0
    else if(ctx.cubic_type == CUBIC_SERPENTINE)
    {
        if(ctx.d[1] < 0)
            iB.flags |= CFLAG_NEGATIVE;
    }

    // now we need to find the correct sign.
    // what we have, from the code above,
    // is the value of sign as defined by the paper
    // Loop and Blinn 2005 section 4.3 paragraph 6,
    // with the negative values to the right of the
    // direction of parametric travel.
    // We want the negative values to be at the left
    // side of the *curve*. For that we need the curve
    // to be monotonic.
    // If it's going up then we need to change the sign
    if (B[0].y < B[3].y)
    {
        // flip negative bit
        iB.flags ^= CFLAG_NEGATIVE;
    }
}

#endif
