#include <cstdio>
#include "packets.h"
#include <gpu/timer.h>
#include <gpu/cudpp.h>
#include "traversal.h"

const int BW = 256;

#define TRACE 0

__global__
void gpu_get_tinfo_head_flags(const unsigned *idquads, unsigned tinfo_count,
                              unsigned *hflags)
{
    int id = blockIdx.x*BW + threadIdx.x;
    if(id >= tinfo_count)
        return;

    if(id==0)
        hflags[id] = 1;
    else
        hflags[id] = idquads[id] != idquads[id-1];
}

__global__
void gpu_get_uint2_head_flags(const uint2 *q_ua, unsigned tinfo_count,
                              unsigned *hflags)
{
    int id = blockIdx.x*BW + threadIdx.x;
    if(id >= tinfo_count)
        return;

    if(id==0)
        hflags[id] = 1;
    else
        hflags[id] = q_ua[id].x != q_ua[id-1].x;
}

__global__
void gpu_compact_tinfo(const unsigned *scan_hflags, unsigned count,
                       unsigned *base)
{
    int id = blockIdx.x*BW + threadIdx.x;
    if(id >= count)
        return;

    if(scan_hflags[id] != scan_hflags[id+1])
        base[scan_hflags[id]] = id;
}

__global__
void gpu_compact_tinfo_calc_packets(const unsigned *base, unsigned count,
                                    unsigned tinfo_count,
                                    unsigned *numpackets,
                                    unsigned packet_size)
{
    int id = blockIdx.x*BW + threadIdx.x;
    if(id >= count)
        return;

    unsigned size;

    if(id == count-1)
        size = tinfo_count - base[id];
    else
        size = base[id+1] - base[id];

    //numpackets[id] = (size+PACKET_SIZE-1)/PACKET_SIZE;
    numpackets[id] = (size+packet_size-1)/packet_size;
}

__global__
void gpu_update_tinfo_skeleton(const unsigned *scan_numpackets, unsigned count,
                               const unsigned *base,
                               unsigned *skeleton,
                               unsigned *hflags)
{
    int id = blockIdx.x*BW + threadIdx.x;
    if(id >= count)
        return;

    unsigned pos = scan_numpackets[id];
    skeleton[pos] = base[id];
    hflags[pos] = 1;
}

__global__
void gpu_compact_tinfo_calc_size(const unsigned *base, unsigned count, 
                                 unsigned tinfo_count,
                                 unsigned *size)
{
    int id = blockIdx.x*BW + threadIdx.x;
    if(id >= count)
        return;

    if(id == count-1)
        size[id] = tinfo_count - base[id];
    else
        size[id] = base[id+1] - base[id];
}

void create_packets(gpu::vector<unsigned> &packets_base,
                    gpu::vector<unsigned> &packets_size,
                    const TraversalInfos &tinfo, int packet_size )
{
    gpu::scoped_timer_stop
        sts(gpu::timers.gpu_add("decompose pixels into packets"));

    gpu::vector<unsigned> hflags(tinfo.size()+1);

    dim3 bdim(BW),
         gdim((tinfo.size()+BW-1)/BW);

    // fg: tinfo.idquad is an array that contains the id of the cell
    // fg: for each pixel, and is sorted by quadid. This kernel marks
    // fg: with 1 the first pixel inside a cell and 0 for the other pixels.
    gpu_get_tinfo_head_flags<<<gdim,bdim>>>(&tinfo.idquad, tinfo.size(),
                                            &hflags);

    gpu::Cudpp cudpp;
    gpu::Cudpp cub;
    // fg: sums the marks on each pixel. hflags.back()
    // fg: has the number of cells used.
    cub.exclusive_scan(hflags, hflags);
    //
    //std::cerr << "----------------------- hflags.back():" << hflags.back() << std::endl;
    gpu::vector<unsigned> base(hflags.back());
    gpu_compact_tinfo<<<gdim, bdim>>>(&hflags, tinfo.size(), &base);

    gpu::vector<unsigned> numpackets(base.size()+1);

    gdim.x = (base.size()+BW-1)/BW;
    gpu_compact_tinfo_calc_packets<<<gdim, bdim>>>(&base, base.size(),
                                                   tinfo.size(),
                                                   &numpackets, packet_size);

    cub.exclusive_scan(numpackets, numpackets);

    unsigned packet_count = numpackets.back();
    numpackets.resize(numpackets.size()-1);

    gpu::vector<unsigned> skeleton(packet_count);
    hflags.resize(packet_count);
    hflags.fill(0);
    thrust::fill(skeleton.begin(), skeleton.end(), packet_size);

    gdim.x = (numpackets.size()+BW-1)/BW;

    gpu_update_tinfo_skeleton<<<gdim,bdim>>>(&numpackets, numpackets.size(), 
                                             &base, &skeleton, &hflags);

    cudpp.segmented_inclusive_scan(skeleton, hflags, skeleton);

    packets_size.resize(packet_count);

    gdim.x = (packet_count+BW-1)/BW;
    gpu_compact_tinfo_calc_size<<<gdim, bdim>>>(&skeleton, skeleton.size(),
                                                tinfo.size(),
                                                &packets_size);

    swap(skeleton, packets_base);

#if TRACE && false
    std::vector<TraversalInfo> _tinfo;
    std::vector<unsigned> base, size;
    packets_base.copy_to_host(base);
    packets_size.copy_to_host(size);
    tinfo.copy_to_host(_tinfo);
    for(unsigned i=0; i<base.size(); ++i)
        printf("packet #%d idquad %d: %d %d\n",i,_tinfo[base[i]].idquad,base[i], size[i]);
#endif
}

void create_packets_quad_unitarea(
                    gpu::vector<unsigned> &packets_base,
                    gpu::vector<unsigned> &packets_size,
                    const gpu::vector<uint2> &quad_ua, int packet_size )
{
    gpu::scoped_timer_stop
        sts(gpu::timers.gpu_add("decompose pixels into packets"));

    gpu::vector<unsigned> hflags(quad_ua.size()+1);

    dim3 bdim(BW),
         gdim((quad_ua.size()+BW-1)/BW);

    // fg: tinfo.idquad is an array that contains the id of the cell
    // fg: for each pixel, and is sorted by quadid. This kernel marks
    // fg: with 1 the first pixel inside a cell and 0 for the other pixels.
    gpu_get_uint2_head_flags<<<gdim,bdim>>>(&quad_ua, quad_ua.size(),
                                            &hflags);
    gpu::Cudpp cudpp;
    gpu::Cudpp cub;
    // fg: sums the marks on each pixel. hflags.back()
    // fg: has the number of cells used.
    cub.exclusive_scan(hflags, hflags);
    //
    //std::cerr << "----------------------- hflags.back():" << hflags.back() << std::endl;
    gpu::vector<unsigned> base(hflags.back());
    gpu_compact_tinfo<<<gdim, bdim>>>(&hflags, quad_ua.size(), &base);
    gpu::vector<unsigned> numpackets(base.size()+1);

    gdim.x = (base.size()+BW-1)/BW;
    gpu_compact_tinfo_calc_packets<<<gdim, bdim>>>(&base, base.size(),
                                                   quad_ua.size(),
                                                   &numpackets, packet_size);

    cub.exclusive_scan(numpackets, numpackets);
    unsigned packet_count = numpackets.back();
    numpackets.resize(numpackets.size()-1);

    gpu::vector<unsigned> skeleton(packet_count);
    hflags.resize(packet_count);
    hflags.fill(0);
    thrust::fill(skeleton.begin(), skeleton.end(), packet_size);

    gdim.x = (numpackets.size()+BW-1)/BW;

    gpu_update_tinfo_skeleton<<<gdim,bdim>>>(&numpackets, numpackets.size(), 
                                             &base, &skeleton, &hflags);

    cudpp.segmented_inclusive_scan(skeleton, hflags, skeleton);

    packets_size.resize(packet_count);

    gdim.x = (packet_count+BW-1)/BW;
    gpu_compact_tinfo_calc_size<<<gdim, bdim>>>(&skeleton, skeleton.size(),
                                                quad_ua.size(),
                                                &packets_size);

    swap(skeleton, packets_base);

    //printf("# packets: %u\n",packet_count);
}

