#include <fd/fontdump.h>
#include <gpu/cudpp.h>
#include <gpu/timer.h>
#include "util.h"
#include "paint.h"

const int BW = 256;

const int ALIGNMENT = 7; // 2^p

__device__ unsigned higher_multiple_of(unsigned v, unsigned p)
{
    unsigned r;
    unsigned rem = v & ((1<<p)-1);
    if(rem)
        r = (v & ~((1<<p)-1)) + (1<<p);
    else
        r = v;

    return r;
}

__global__ void gpu_calc_paint_data(const fd::Attrib *attribs, unsigned attrib_count,
                                    const Gradient *gradients,
                                    unsigned *paint_size, Paint::Type *paint_type)
{
    int idattrib = blockIdx.x*BW + threadIdx.x;
    if(idattrib >= attrib_count)
        return;

    const fd::Attrib &attrib = attribs[idattrib];

    if(attrib.is_opaque)
    {
        paint_size[idattrib] = higher_multiple_of(sizeof(PaintSolid),ALIGNMENT);
        paint_type[idattrib] = Paint::SOLID;
        return;
    }

    unsigned size = 0;
    Paint::Type type;

    const Gradient *grad;

    switch(attrib.fill.mode)
    {
    case fd::COLORING_SOLID:
        size = higher_multiple_of(sizeof(PaintSolidAlpha),ALIGNMENT);
        type = Paint::SOLID_ALPHA;
        break;

    case fd::COLORING_GRADIENT:
        grad = &gradients[attrib.fill.gradient];
        switch(grad->type)
        {
        case Gradient::LINEAR:
            size = higher_multiple_of(sizeof(PaintLinearGrad),ALIGNMENT);
            type = Paint::GRAD_LINEAR;
            break;
        case Gradient::RADIAL:
            size = higher_multiple_of(sizeof(PaintRadialGrad),ALIGNMENT);
            type = Paint::GRAD_RADIAL;
            break;
        }
        break;
    };

    if(size == 0)
    {
        size = higher_multiple_of(sizeof(PaintNone),ALIGNMENT);
        type = Paint::NONE;
    }

    paint_size[idattrib] = size;
    paint_type[idattrib] = type;
}

__global__
void gpu_create_paints(const fd::Attrib *attribs, unsigned attrib_count,
                       const Gradient *grads,
                       const Paint::Type *paint_types, const unsigned *paint_size,
                       unsigned char *paint_data, const Paint **paints)
{
    int idattrib = blockIdx.x*BW + threadIdx.x;
    if(idattrib >= attrib_count)
        return;

    const fd::Attrib &attrib = attribs[idattrib];

    paint_data += paint_size[idattrib];

    paints[idattrib] = reinterpret_cast<const Paint *>(paint_data);

    const Gradient *grad;
    if(attrib.fill.mode == fd::COLORING_GRADIENT)
        grad = grads + attrib.fill.gradient;

    switch(paint_types[idattrib])
    {
    case Paint::SOLID:
        new(paint_data) PaintSolid(to_float4(attrib.fill.solid));
        break;

    case Paint::SOLID_ALPHA:
        new(paint_data) PaintSolidAlpha(to_float4(attrib.fill.solid),
                                        attrib.fill.opacity*attrib.opacity);
        break;
    case Paint::GRAD_LINEAR:
        new(paint_data) PaintLinearGrad(*grad,
                                        attrib.fill.opacity*attrib.opacity);
        break;
    case Paint::GRAD_RADIAL:
        new(paint_data) PaintRadialGrad(*grad,
                                        attrib.fill.opacity*attrib.opacity);
        break;
    case Paint::NONE:
        new(paint_data) PaintNone();
        break;
    }
}

void convert_paints(const gpu::vector<fd::Attrib> &attribs,
                    const gpu::vector<Gradient> &grads,
                    gpu::vector<unsigned char> &paint_data,
                    gpu::vector<const Paint *> &paints)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("create paints"));

    dim3 bdim(BW),
         gdim((attribs.size()+BW-1)/BW);

    gpu::vector<unsigned> paint_size(attribs.size()+1); // +1 = total size
    gpu::vector<Paint::Type> paint_types(attribs.size());

    gpu_calc_paint_data<<<gdim,bdim>>>(&attribs, attribs.size(), &grads,
                                       &paint_size, &paint_types);

    gpu::Cudpp cub;
    cub.exclusive_scan(paint_size, paint_size);

    paint_data.resize(paint_size.back());
    paints.resize(attribs.size());

    gpu_create_paints<<<gdim, bdim>>>(&attribs, attribs.size(),
                                      &grads,
                                      &paint_types, &paint_size,
                                      &paint_data, &paints);
}
