#ifndef GPURASTER_PAINT_H
#define GPURASTER_PAINT_H

#include <cuda_runtime.h> // for tex1D
#include <fd/fontdump.h>
#include <gpu/vector.h>
#include "config.h"
#include "color_util.h"
#include "gradients.h"
#include "texture.h"

class Paint/*{{{*/
{
public:
    enum Type
    {
        NONE,

        SOLID,
        SOLID_ALPHA,

        GRAD_LINEAR,
        GRAD_RADIAL,

        TEXTURE,

        OVERDRAW
    } type;

    __device__
    Paint(Type _type) : type(_type) {}

    template <class P>
    __device__
    P &as() { return static_cast<P &>(*this); }

    template <class P>
    __device__
    const P &as() const { return static_cast<const P &>(*this); }
};/*}}}*/

class PaintNone : public Paint/*{{{*/
{
public:
    __device__ PaintNone() : Paint(NONE)
    {
    }

    __device__ const float4 &operator()(float2 coord, const float4 &sample) const
    {
        (void) coord;
        return sample;
    }
};/*}}}*/

class PaintOverdraw : public Paint/*{{{*/
{
public:
    __device__ PaintOverdraw() : Paint(OVERDRAW)
    {
    }

    __device__ float4 operator()(float2 coord, const float4 &sample) const
    {
        (void) coord;
        return make_float4(sample.x+0.1, sample.y+0.1, sample.z+0.1, 1);
    }
};/*}}}*/

class PaintSolid : public Paint/*{{{*/
{
public:
    __host__ __device__ PaintSolid(const float4 &color)
        : Paint(SOLID)
        , m_color(color)
    {
        assert(color.w == 1);
    }

#if FRONT_TO_BACK_RENDERING
    __device__ float4 operator()(float2 coord, float4 sample) const
    {
        (void) coord;
        return blend(sample,m_color);
    }
#else
    __device__ const float4 &operator()(float2 coord, float4 sample) const
    {
        (void) coord;
        return m_color;
    }
#endif

private:
    float4 m_color;
};/*}}}*/

class PaintSolidAlpha : public Paint/*{{{*/
{
public:

    __device__ PaintSolidAlpha(const float4 &color, float fill_opacity)
        : Paint(SOLID_ALPHA)
        , m_color(make_float4(color.x, color.y, color.z, color.w*fill_opacity))
    {
        m_color = premultiply(m_color);
    }

    __device__ float4 operator()(const float2 &coord,
                                 const float4 &sample) const
    {
        (void) coord;
#if FRONT_TO_BACK_RENDERING
        return blend(sample, m_color);
#else
        return blend(m_color, sample);
#endif
    }

private:
    float4 m_color;
};/*}}}*/

class PaintLinearGrad : public Paint/*{{{*/
{
public:
    __host__ __device__ PaintLinearGrad(const Gradient &grad,
                               float opacity)
        : Paint(GRAD_LINEAR)
        //, m_inv_ramp_size(grad.inv_ramp_size)
        //, m_linear(grad.linear)
        , m_ramp(grad.ramp)
        , m_opacity(opacity)
    {
        //set_xform(grad.transform);
        optimize( grad );
    }

    HOSTDEV void optimize( const Gradient& grad )
    {
        // is this the inverse already?
        const Transform& xform = inverse(grad.transform);
        const Linear& linear = grad.linear;

        float2 v = linear.p2-linear.p1;
        float inorm = 1.f/sqrt(dot(v, v));

        Transform s = scale(inorm,inorm),
                  t = translate(-linear.p1.x, -linear.p1.y);

        v = apply(s, v);

        Transform r;
        r.l.x = v.x,
        r.l.y = v.y,
        r.l.z = -v.y,
        r.l.w = v.x;
        r.t.x = r.t.y = 0;

        Transform xf = r*(s*t) * xform;

        m.x = xf.l.x;
        m.y = xf.l.y;
        m.z = xf.t.x;
    }

#ifdef __CUDACC__
    __device__ float4 operator()(const float2 &coord,
                                 const float4 &sample) const
    {
        float t = m.x*coord.x + m.y*coord.y + m.z;

        float4 fill_color = tex1D<float4>(m_ramp, t);

        fill_color.w *= m_opacity;
        fill_color = premultiply(fill_color);

#if FRONT_TO_BACK_RENDERING
        return blend(sample, fill_color);
#else
        return blend(fill_color, sample);
#endif
    }
#endif

    HOSTDEV void set_xform(const Transform &xform)
    {
        //apply the transpose of xform on m ( m * xf^t )
        float3 n;
        n.x = xform.l.x * m.x + xform.l.z * m.y;// + xform.t.x * m.z;
        n.y = xform.l.y * m.x + xform.l.w * m.y;// + xform.t.y * m.z;
        n.z = xform.t.x * m.x + xform.t.y * m.y + m.z;
        m = n;
    }
    __device__ const Transform xform() const
    {
        Transform t;
        t.l.x = m.x;
        t.l.y = m.y;
        t.t.x = m.z;
        return t;
    }

private:
    cudaTextureObject_t m_ramp;
    float3 m;
    float m_opacity;
};/*}}}*/

class PaintRadialGrad : public Paint/*{{{*/
{
public:
    __host__ __device__ PaintRadialGrad(const Gradient &grad,
                               float opacity)
        : Paint(GRAD_RADIAL)
        , m_ramp(grad.ramp)
        , m_xform(grad.transform)
        //, m_opacity(opacity)
    {
        m_xform.free_space.x = opacity;
        optimize( grad );
    }

    HOSTDEV void optimize( const Gradient& grad )
    {
        const Radial rad = grad.radial;
        Transform f0 = translate( -rad.f.x, -rad.f.y );
        float2 c = apply( f0, rad.c );
        Transform cy0 = rotate( -std::atan2( c.y, c.x ) );
        c = apply( cy0, c );
        Transform r1 = scale( 1.f/rad.r, 1.f/rad.r );
        c = apply( r1, c );
        float opacity = m_xform.free_space.x;
        m_xform = compose( r1, compose( cy0, compose( f0, inverse(grad.transform) ) ) );
        //m_cx = c.x;
        m_xform.free_space.x = opacity;
        m_xform.free_space.y = c.x;
    }

#ifdef __CUDACC__
    __device__ float4 operator()(const float2 &coord,
                                 const float4 &sample) const
    {
        float2 pg = apply( m_xform, coord );
        float a0 = dot( pg, pg );
        float c = m_xform.free_space.y;
        float a1 = -pg.x * c;
        float a2 = c*c -1.f;
        float t = a0/(-a1+sqrt(a1*a1-a0*a2));
        float4 fill_color = tex1D<float4>(m_ramp, t);
        //fill_color.w *= m_opacity;
        fill_color.w *= m_xform.free_space.x;
        fill_color = premultiply(fill_color);
#if FRONT_TO_BACK_RENDERING
        return blend(sample, fill_color);
#else
        return blend(fill_color, sample);
#endif
    }
#endif

    __device__ void set_xform(const Transform &xform)
    {
        (void) xform; // ?? why is the set_xform disabled like this??
        //m_xform = xform;
    }

    __device__ const Transform &xform() const { return m_xform; }

private:
    //float m_opacity; //m_xform.free_space.x
    cudaTextureObject_t m_ramp;
    Transform m_xform;
    //float m_cx; //m_xform.free_space.y
};/*}}}*/


#if 1
class PaintTexture : public Paint/*{{{*/
{
public:
    __host__ __device__ PaintTexture( const Texture& texture, float opacity)
        : Paint(TEXTURE)
        //, m_tex( texture )
        , m_opacity(opacity)
        , m_tex( texture.m_tex )
        , m_transform( texture.m_transform )
    {
    }

#ifdef __CUDACC__
    __device__ float4 operator()(const float2 &coord,
                                 const float4 &sample) const
    {
        //float2 uv = apply(m_tex.m_transform, coord);
        float2 uv = apply(m_transform, coord);
        //float4 fill_color = tex2D<float4>(m_tex.m_tex, uv.x, uv.y );
        float4 fill_color = tex2D<float4>(m_tex, uv.x, uv.y );
        fill_color.w *= m_opacity;
        fill_color = premultiply(fill_color);
        //return make_float4(0,0,1,1);
        //return make_float4(uv.x, uv.y, 0, 1);
        //fill_color.y = 0;
        //fill_color.z = 0;
        //fill_color.w = 1;
        //return fill_color;
        //*
#if FRONT_TO_BACK_RENDERING
        return blend(sample, fill_color);
#else
        return blend(fill_color, sample);
#endif
// */
    }
#endif

    __device__ void set_xform(const Transform &xform)
    {
        //m_tex.m_transform = xform;
        m_transform = xform;
    }

    __device__ const Transform &xform() const { return m_transform; }

private:
    float m_opacity;
    cudaTextureObject_t m_tex;
    Transform m_transform;
    //Texture m_tex;
};/*}}}*/
#endif

void convert_paints(const gpu::vector<fd::Attrib> &attribs,
                    const gpu::vector<Gradient> &grads,
                    gpu::vector<unsigned char> &paint_data,
                    gpu::vector<const Paint *> &paints);

#endif
