#include <gpu/box.h>
#include <gpu/timer.h>
#include <fd/fontdump.h>
#include <gpu/sort.h>
#include <gpu/cub.h>
#include "paint.h"
#include "segment.h"
#include "implicit.h"
#include "icurve1.h"
#include "icurve2.h"
#include "icurve3.h"
#include "curve1.h"
#include "curve2.h"
#include "curve3.h"
#include "curve_arc.h"
#include "implicitize1.h"
#include "implicitize2.h"
#include "implicitize3.h"
#include "implicitize_arc.h"
#include "monotonic2.h"
#include "monotonic3.h"
#include "monotonic_arc.h"

const int MONO_BW = 256;

template<int N>
__device__
Curve<N> illustration_to_image(const Curve<N> &B, float2 scale, float2 offset)
{
    Curve<N> R;

    for(int i=0; i<B.size(); ++i)
        R[i] = B[i]*scale+offset;

    return R;
}

__device__
Curve<SEG_ARC> illustration_to_image(const Curve<SEG_ARC> &B, float2 scale, float2 offset)
{
    Curve<SEG_ARC> R;

//    for(int i=0; i<B.size(); ++i)
    R[0] = B[0]*scale+offset;
    R[1] = B[1]*scale+B.w*offset;
    R[2] = B[2]*scale+offset;

    R.w = B.w;

    return R;
}

__global__ void gpu_calc_ctx_size(gpu::vector_ptr<const Segment<> > segs,
                                   unsigned *ctx_size)
{
    int idx = blockIdx.x*MONO_BW+threadIdx.x;

    if(idx > segs.size())
        return;

    if(idx == segs.size())
    {
        ctx_size[idx] = 0;
        return;
    }

    switch(segs[idx].type)
    {
    case SEG_LINEAR:
        ctx_size[idx] = sizeof(ImplicitizationContext<1>);
        break;

    case SEG_QUADRATIC:
        ctx_size[idx] = sizeof(ImplicitizationContext<2>);
        break;

    case SEG_CUBIC:
        ctx_size[idx] = sizeof(ImplicitizationContext<3>);
        break;

    case SEG_ARC:
        ctx_size[idx] = sizeof(ImplicitizationContext<SEG_ARC>);
        break;

    default:
        ctx_size[idx] = 0;
    }
}

__device__ void preprocess1(const Curve<1> &B,
                            ImplicitizationContext<1> &ctx,
                            Roots &roots,
                            unsigned &seg_count, unsigned &icurve_size)
{
    roots[0] = 1;
    roots[1] = FLT_MAX;
    icurve_size = sizeof(ImplicitCurve<1>);
    seg_count = 1;
}

__device__ void preprocess1(const Curve<2> &B,
                            ImplicitizationContext<2> &ctx,
                            Roots &roots,
                            unsigned &seg_count, unsigned &icurve_size)
{
    unsigned nroots = calc_roots(B, roots);
    roots[nroots++] = 1;
    roots[nroots] = FLT_MAX;

    int nsegs = nroots;
    icurve_size = nsegs*sizeof(ImplicitCurve<2>);
    seg_count = nsegs;
}

__device__ void preprocess1(const Curve<3> &B,
                            ImplicitizationContext<3> &ctx,
                            Roots &roots,
                            unsigned &seg_count, unsigned &icurve_size)
{
    gpu::array<float2,2> ts;
    ctx.cubic_type = classify(B, ctx, ts);

    switch(ctx.cubic_type)
    {
    case CUBIC_QUADRATIC:
    case CUBIC_LINE:
        {
            Curve<1> B1;
            B1[0] = B[0];
            B1[1] = B[3];
            ImplicitizationContext<1> ctx1;
            preprocess1(B1, ctx1, roots, seg_count, icurve_size);
        }
        return;
    case CUBIC_POINT:
        seg_count = 0;
        icurve_size = 0;
        roots[0] = FLT_MAX;
        return;
    }

    unsigned nroots = calc_roots(B, roots);

    {
        float r = ts[0].x / ts[0].y;
        if(0 < r && r < 1)
            roots[nroots++] = r;

        r = ts[1].x / ts[1].y;
        if(0 < r && r < 1)
            roots[nroots++] = r;
    }

    sort(roots.begin(), roots.begin()+nroots);

    roots[nroots++] = 1;

    if(nroots < MAX_ROOTS)
        roots[nroots] = FLT_MAX;

    int nsegs = nroots;

    icurve_size = nsegs*sizeof(ImplicitCurve<3>);
    seg_count = nsegs;
}

__device__ void preprocess1(const Curve<SEG_ARC> &B,
                            ImplicitizationContext<SEG_ARC> &ctx,
                            Roots &roots,
                            unsigned &seg_count, unsigned &icurve_size)
{
    unsigned nroots = calc_roots(B, roots);
    roots[nroots++] = 1;
    roots[nroots] = FLT_MAX;

    sort(roots.begin(), roots.begin()+nroots);

    int nsegs = nroots;
    icurve_size = nsegs*sizeof(ImplicitCurve<SEG_ARC>);
    seg_count = nsegs;
}

__global__ void gpu_preprocess1(
    gpu::vector_ptr<const Segment<> > segs,
    Roots *roots,
    unsigned *seg_count, unsigned *icurve_size,
    char *ctx_data, unsigned *ctx_idx,
    float2 scale, float2 offset)
{
    int idx = blockIdx.x*MONO_BW+threadIdx.x;

    if(idx > segs.size())
        return;

    if(idx == segs.size())
    {
        icurve_size[idx] = seg_count[idx] = 0;
        return;
    }

    switch(segs[idx].type)
    {
    case SEG_LINEAR:
        preprocess1(*segs[idx].as<1>().B,
                    *reinterpret_cast<ImplicitizationContext<1>*>(ctx_data+ctx_idx[idx]),
                    roots[idx],
                    seg_count[idx],
                    icurve_size[idx]);
        break;

    case SEG_QUADRATIC:
        preprocess1(*segs[idx].as<2>().B,
                    *reinterpret_cast<ImplicitizationContext<2>*>(ctx_data+ctx_idx[idx]),
                    roots[idx],
                    seg_count[idx],
                    icurve_size[idx]);
        break;

    case SEG_CUBIC:
        preprocess1(illustration_to_image(*segs[idx].as<3>().B, scale, offset),
                    *reinterpret_cast<ImplicitizationContext<3>*>(ctx_data+ctx_idx[idx]),
                    roots[idx],
                    seg_count[idx],
                    icurve_size[idx]);
        break;

    case SEG_ARC:
        preprocess1(*segs[idx].as<SEG_ARC>().B,
                    *reinterpret_cast<ImplicitizationContext<SEG_ARC>*>(ctx_data+ctx_idx[idx]),
                    roots[idx],
                    seg_count[idx],
                    icurve_size[idx]);
        break;
    case SEG_PATH:
        // enough space to copy the segment
        seg_count[idx] = 1;
        icurve_size[idx] = sizeof(PathData);
        break;
    default:
        // enough space to copy the segment
        seg_count[idx] = 1;
        icurve_size[idx] = 0;
        break;
    }
}

template <int N, class F>
__device__
unsigned split(const Curve<N> &B, const Roots &roots,
               F &process)
{
    float last_root = 0;
    int r;

    for(r=0; r<MAX_ROOTS && roots[r] < FLT_MAX; ++r)
    {
        process(subcurve(B, last_root, roots[r]), last_root, roots[r]);
        last_root = roots[r];
    }

    return r; // number of segments created
}

template <int N>
struct OutputImplicitSegment
{
    __device__ OutputImplicitSegment(const Segment<N> &seg,
                          SegmentsPtr &isegs, unsigned idiseg_base,
                          ImplicitCurve<N> *icurves,
                          ImplicitizationContext<N> &ctx)
        : m_seg(seg)
        , m_isegs(isegs)
        , m_idiseg_base(idiseg_base)
        , m_icurves(icurves)
        , m_ctx(ctx)
    {
    }

    __device__
    void operator()(const Curve<N> &B, float t0, float t1)
    {
        ImplicitCurve<N> &iB = *m_icurves;
        iB.flags = 0;

        m_isegs.idx_data[m_idiseg_base] = (unsigned)((unsigned char *)&iB - m_isegs._data);
        m_isegs.type[m_idiseg_base] = m_seg.type;
        m_isegs.flags[m_idiseg_base] = 0;

        implicitize(B, iB, t0, t1, m_ctx);

        ++m_idiseg_base;
        ++m_icurves;
    }

private:
    const Segment<N> &m_seg;
    SegmentsPtr &m_isegs;
    unsigned m_idiseg_base;
    ImplicitCurve<N> *m_icurves;
    ImplicitizationContext<N> &m_ctx;
};

template <int N>
__device__ void preprocess2(const Segment<N> &seg,
                            ImplicitizationContext<N> &ctx,
                            Roots &roots,
                            SegmentsPtr &isegs, unsigned idiseg_base,
                            ImplicitCurve<N> *icurves,
                            float2 scale, float2 offset)
{
    OutputImplicitSegment<N> output(seg, isegs, idiseg_base,icurves,ctx);

    split(illustration_to_image(*seg.B,scale,offset), roots, output);
}

__device__ void preprocess2(const Segment<3> &seg,
                            ImplicitizationContext<3> &ctx,
                            Roots &roots,
                            SegmentsPtr &isegs, unsigned idiseg_base,
                            ImplicitCurve<3> *icurves,
                            float2 scale, float2 offset)
{
    if(ctx.cubic_type == CUBIC_QUADRATIC ||
       ctx.cubic_type == CUBIC_LINE)
    {
        Segment<1> seg1
            = reinterpret_cast<const Segment<1> &>(seg);
        seg1.type = SEG_LINEAR;

        ImplicitCurve<1> *icurves1
            = reinterpret_cast<ImplicitCurve<1> *>(icurves);
        ImplicitizationContext<1> &ctx1
            = reinterpret_cast<ImplicitizationContext<1> &>(ctx);

        OutputImplicitSegment<1> output(seg1, isegs, idiseg_base,
                                        icurves1, ctx1);

        Curve<1> B1;
        B1[0] = seg[0];
        B1[1] = seg[3];

        split(illustration_to_image(B1,scale,offset), roots, output);
    }
    else
    {
        OutputImplicitSegment<3> output(seg, isegs, idiseg_base,
                                        icurves, ctx);

        split(illustration_to_image(*seg.B,scale,offset), roots, output);
    }
}

__global__ void gpu_preprocess2(
    gpu::vector_ptr<const Segment<> > segs,
    SegmentsPtr isegs,
    const unsigned *iseg_idx,
    const unsigned *icurve_idx,
    char *ctx_data,
    const unsigned *ctx_idx,
    Roots *roots,
    float2 scale, float2 offset)
{
    int idx = blockIdx.x*MONO_BW+threadIdx.x;

    if(idx >= segs.size())
        return;

    switch(segs[idx].type)
    {
    case SEG_LINEAR:
        preprocess2(segs[idx].as<1>(),
                    *reinterpret_cast<ImplicitizationContext<1>*>(ctx_data+ctx_idx[idx]),
                    roots[idx],
                    isegs, iseg_idx[idx],
                    reinterpret_cast<ImplicitCurve<1>*>(isegs._data+icurve_idx[idx]),
                    scale, offset);

        break;

    case SEG_QUADRATIC:
        preprocess2(segs[idx].as<2>(),
                    *reinterpret_cast<ImplicitizationContext<2>*>(ctx_data+ctx_idx[idx]),
                    roots[idx],
                    isegs, iseg_idx[idx],
                    reinterpret_cast<ImplicitCurve<2>*>(isegs._data+icurve_idx[idx]),
                    scale, offset);
        break;

    case SEG_CUBIC:
        preprocess2(segs[idx].as<3>(),
                    *reinterpret_cast<ImplicitizationContext<3>*>(ctx_data+ctx_idx[idx]),
                    roots[idx],
                    isegs, iseg_idx[idx],
                    reinterpret_cast<ImplicitCurve<3>*>(isegs._data+icurve_idx[idx]),
                    scale, offset);
        break;

    case SEG_ARC:
        preprocess2(segs[idx].as<SEG_ARC>(),
                    *reinterpret_cast<ImplicitizationContext<SEG_ARC>*>(ctx_data+ctx_idx[idx]),
                    roots[idx],
                    isegs, iseg_idx[idx],
                    reinterpret_cast<ImplicitCurve<SEG_ARC>*>(isegs._data+icurve_idx[idx]),
                    scale, offset);
        break;
    case SEG_PATH:
        {
            isegs.type[iseg_idx[idx]] = SEG_PATH;

            // deep copy the path data so that it'll be in implicit data array
            // the original path data is in input data array that will be
            // destroyed just after preprocessing (and before rasterization)
            isegs.idx_data[iseg_idx[idx]] = icurve_idx[idx];

            *isegs.data<PathData>(iseg_idx[idx]) = *(const PathData *)segs[idx].data;
            isegs.winding[iseg_idx[idx]] = 0;
            int paint_type = segs[idx].as<SEG_PATH>().path->paint_type;
            isegs.flags[iseg_idx[idx]] = paint_type << 8;
        }
        break;
    default:
        if(is_stack(segs[idx].type))
        {
            isegs.type[iseg_idx[idx]] = segs[idx].type;
            break;
        }
    }
}

void preprocess(const gpu::vector<Segment<> > &segs,
                Segments &out_segs,
                float2 scale, float2 offset)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("preprocessing"));

    if(segs.size() == 0)
    {
        out_segs.clear();
        out_segs._data.clear();
        return;
    }

    gpu::vector<unsigned> ctx_size(segs.size()+1);

    gpu::vector<char> ctx_data;

    gpu::CUB cub;

    // Calculate ImplicitizationContext index into out_ctx_data using
    // segment type
    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("calc ctx index"));
        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("calc ctx size"));

            dim3 bdim(MONO_BW),
                 gdim((segs.size()+1+bdim.x-1)/bdim.x);

            gpu_calc_ctx_size<<<gdim, bdim>>>(&segs, ctx_size.data());

        }

        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("exclusive scan ctx size"));
            cub.exclusive_scan(ctx_size, ctx_size);
        }

        ctx_data.resize(ctx_size.back());
    }

    // Now calculate split positions, output segment count and position into
    // array, and icurve index into out_icurve_data

    gpu::vector<Roots> roots;
    roots.resize(segs.size());

    gpu::vector<unsigned> seg_count, icurve_size;
    seg_count.resize(segs.size()+1);
    icurve_size.resize(segs.size()+1);

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("geom implicit. and root finding"));

        dim3 bdim(MONO_BW),
             gdim((segs.size()+1+bdim.x-1)/bdim.x);

        gpu_preprocess1<<<gdim, bdim>>>(&segs,
                                        roots.data(),
                                        seg_count.data(), icurve_size.data(),
                                        ctx_data.data(), ctx_size.data(), scale, offset);
    }

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("exclusive scan on segment size"));
        cub.exclusive_scan(seg_count, seg_count);

    }

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("exclusive scan on chull implicit. size"));
        cub.exclusive_scan(icurve_size, icurve_size);
    }

    out_segs.resize(seg_count.back());
    out_segs._data.resize(icurve_size.back() + 64);

    if(out_segs.size() == 0)
        return;

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("implicit. of chull"));

        dim3 bdim(MONO_BW),
             gdim((segs.size()+bdim.x-1)/bdim.x);

        gpu_preprocess2<<<gdim,bdim>>>(&segs,
                                       &out_segs, seg_count.data(),
                                       icurve_size.data(),
                                       &ctx_data, ctx_size.data(),
                                       roots.data(),
                                       scale, offset);
    }
}
