/* vim: set foldmethod=marker foldlevel=0: */
#include "config.h"
#include "qtree.h"
#include "implicit.h"
#include "raycast.h"
#include "paint.h"
#include <sstream>
#include <util/util.h> // for enable_if
#include <gpu/value.h>
#include <gpu/timer.h>
#include <gpu/symbol.h>
#include <gpu/cudpp.h>
#include <gpu/cub.h>
#include <thrust/count.h>
#include <thrust/remove.h>
#include <thrust/sort.h>
#include "util.h"
#include "dump.h"
#include "segment_util.h"

#define TRACE 0

__constant__ unsigned d_min_segs_split;

const int BW = 256;

template <int N>
__device__
void get_seg(const SegmentsPtr &segs, const unsigned char *segdata, 
             unsigned idseg, ImplicitSegment<N> &seg)
{
    seg.type = segs.type[idseg];
    seg.flags = segs.flags[idseg];
    seg.data = segdata + segs.idx_data[idseg];
}

HOSTDEV char4 operator+(const char4 &a, const char4 &b)
{
    return make_char4(a.x+b.x, a.y+b.y, a.z+b.z, a.w+b.w);
}

namespace cub
{

template<>
__host__ __device__ __forceinline__
uint4 Max::operator()(const uint4 &a, const uint4 &b) const
{
    return make_uint4(max(a.x,b.x),
                      max(a.y,b.y),
                      max(a.z,b.z),
                      max(a.w,b.w));
}

template<>
__host__ __device__ __forceinline__
uint4 Min::operator()(const uint4 &a, const uint4 &b) const
{
    return make_uint4(min(a.x,b.x),
                      min(a.y,b.y),
                      min(a.z,b.z),
                      min(a.w,b.w));
}


}

template <int N>
__device__
int eval(const ImplicitCurve<N> &B, float2 sample)/*{{{*/
{
    // above or below bbox?
    if(sample.y < B.bbox.min.y || sample.y >= B.bbox.max.y)
    {
        if(B.bbox.min.x <= sample.x && sample.x < B.bbox.max.x)
        {
            if(sample.y >= B.bbox.max.y)
                return B.going_right() == B.going_up() ? -1 : 1;
            else
                return B.going_right() == B.going_up() ? 1 : -1;
        }
        else
            return 0;
    }
    // after bbox?
    else if(sample.x >= B.bbox.max.x)
        return 1;
    // before bbox?
    else if(sample.x < B.bbox.min.x)
        return -1;
    else
    {
        int check = B.hit_chull(sample);
        if(check != -1)
            return check == 1 ? -1 : 1;

        return B.eval(sample) < 0 ? -1 : 1; // hit ? left : right
    }
}/*}}}*/

// nibble 0: FILL
// nibble 1: UP
// nibble 2: DOWN
// nibble 3: PATH
// nibble 4: PUSH
// nibble 5: POP_FILL
// nibble 6: POP_CLIP
// nibble 7: COMMIT

// bit 0: BOTTOM_LEFT
// bit 1: BOTTOM_RIGHT
// bit 2: TOP_LEFT
// bit 3: TOP_RIGHT

// all ones (255) -> don't split

#define DONT_SPLIT(x) (x == 255)

#define FILL(x) ((1ul<<(int)x)<<0)
#define UP(x)   ((1ul<<(int)x)<<4)
#define DOWN(x) ((1ul<<(int)x)<<8)
#define PATH(x) ((1ul<<(int)x)<<12)
#define PUSH(x) ((1ul<<(int)x)<<16)
#define POP_FILL(x) ((1ul<<(int)x)<<20)
#define POP_CLIP(x) ((1ul<<(int)x)<<24)
#define COMMIT(x) ((1ul<<(int)x)<<28)

#define ALLTYPES(x) (FILL(x)|UP(x)|DOWN(x)|PUSH(x)|POP_FILL(x)|POP_CLIP(x)|COMMIT(x)|PATH(x))

#define SEG_TYPES(x) (FILL(x)|PUSH(x)|POP_FILL(x)|POP_CLIP(x)|COMMIT(x)|PATH(x))

#define GEOM_TYPES(x) (FILL(x)|PATH(x))

HOSTDEV void print_sinfo(unsigned sinfo, bool clip=false)/*{{{*/
{
    for(int i=0; i<4; ++i)
    {
        int cnt = 0;
        if(clip)
        {
            if(sinfo & FILL(i))
                printf("f"), ++cnt;
            if(sinfo & PATH(i))
                printf("p"), ++cnt;
            if(sinfo & UP(i))
                printf("u"), ++cnt;
            if(sinfo & DOWN(i))
                printf("d"), ++cnt;
        }
        else
        {
            if(sinfo & FILL(i))
                printf("F"), ++cnt;
            if(sinfo & PATH(i))
                printf("P"), ++cnt;
            if(sinfo & UP(i))
                printf("U"), ++cnt;
            if(sinfo & DOWN(i))
                printf("D"), ++cnt;
        }
        if(sinfo & PUSH(i))
            printf("("), ++cnt;
        if(sinfo & COMMIT(i))
            printf("|"), ++cnt;
        if(sinfo & POP_CLIP(i))
            printf(")"), ++cnt;
        if(sinfo & POP_FILL(i))
            printf(")"), ++cnt;

        if(cnt < 3)
            printf("%*s",3-cnt,"");
        if(i != 3)
            printf(";");
    }
}/*}}}*/

struct SplitData/*{{{*/
{
    char winding[4];
    unsigned split_info;
};/*}}}*/
struct EdgeIntersectionInfo/*{{{*/
{
    /*     TL ---10 --- T ---11 --- TR -- 14 -- TI
           |            |           |
           7            8           9
           |            |           |
           L  --- 5 --- C --- 6 --- R  -- 13 -- I
           |            |           |
           2            3           4
           |            |           |
           BL --- 0 --- B --- 1 --- BR -- 12 -- BI

           11 and 14 -> 15
            6 and 13 -> 16
            1 and 12 -> 17
    */


    unsigned cross0 : 1,
             cross1 : 1,
             cross2 : 1,
             cross3 : 1,
             cross4 : 1,
             cross5 : 1,
             cross6 : 1,
             cross7 : 1,
             cross8 : 1,
             cross9 : 1,
             cross10 : 1,
             cross11 : 1,
             cross12 : 1,
             cross13 : 1,
             cross14 : 1,
             cross15 : 1,
             cross16 : 1,
             cross17 : 1;
};/*}}}*/

template <int N>
__device__
void get_edge_intersection_info(const ImplicitSegment<N> &seg, /*{{{*/
                                const Quad &quad, EdgeIntersectionInfo &info)
{
    const ImplicitCurve<N> &B = *seg.B();

    int signL = eval(B,make_float2(quad.min.x, quad.mp.y));
    int signC = eval(B,quad.mp);
    int signR = eval(B,make_float2(quad.max.x, quad.mp.y));
    int signI = eval(B,make_float2(FLT_MAX, quad.mp.y));

//    if(bbox.min.y < quad.max.y)
    {
        // L, C, R, I

        info.cross5  = (signL * signC < 0);
        info.cross13 = (signR * signI < 0);
        info.cross6  = (signC * signR < 0);
        info.cross16 = (signC * signI < 0);
    }

//    if(bbox.min.y < quad.mp.y)
    {
        // BL, B, BR, BI
        int signBL = eval(B,quad.min),
            signB = eval(B,make_float2(quad.mp.x, quad.min.y)),
            signBR = eval(B,make_float2(quad.max.x, quad.min.y)),
            signBI = eval(B,make_float2(FLT_MAX, quad.min.y));

        info.cross0  = (signBL * signB  < 0);
        info.cross12 = (signBR * signBI < 0);
        info.cross1  = (signB  * signBR < 0);
        info.cross17 = (signB  * signBI < 0);

        //if(bbox.min.y <= quad.max.y)
        {
            info.cross2 = (signBL * signL  < 0);
            info.cross3 = (signB  * signC  < 0);
            info.cross4 = (signBR * signR  < 0);
        }
    }

//    if(bbox.max.y > quad.mp.y)
    {
        int signTL = eval(B,make_float2(quad.min.x, quad.max.y)),
            signT = eval(B,make_float2(quad.mp.x, quad.max.y)),
            signTR = eval(B,quad.max),
            signTI = eval(B,make_float2(FLT_MAX, quad.max.y));

        // TL, T, TR, TI
        info.cross10 = (signTL * signT  < 0);
        info.cross14 = (signTR * signTI < 0);
        info.cross11 = (signT  * signTR < 0);
        info.cross15 = (signT  * signTI < 0);

//        if(bbox.min.y <= quad.max.y)
        {
            info.cross7 = (signTL * signL < 0);
            info.cross8 = (signT  * signC < 0);
            info.cross9 = (signTR * signR < 0);
        }
    }

}/*}}}*/

__device__
void calc_split_data(const ImplicitCurveSegment &seg,/*{{{*/
                     const EdgeIntersectionInfo &einfo, const Quad &quad,
                    SplitData &data)
{
    // if begin and endpoints are in the same quadrant, put segment
    // in there and call it a day

    float2 begpoint = seg.B->begpoint(),
           endpoint = seg.B->endpoint();

    int idxchild_beg = quad.classify(begpoint);
    int idxchild_end = quad.classify(endpoint);

#if TRACE
    printf("splitting seg%d #%d b:(%f,%f) e:(%f,%f)\n",N,idseg,b.x,b.y,e.x,e.y);
#endif

    int aux=0;

    if(quad.contains(begpoint))
    {
        data.split_info |= FILL(idxchild_beg);
#if TRACE
        printf("quad%d contains beg point\n",idxchild_beg);
#endif
        ++aux;
    }

    if(quad.contains(endpoint))
    {
        data.split_info |= FILL(idxchild_end);
#if TRACE
        printf("quad%d contains end point\n",idxchild_end);
#endif
        ++aux;
    }

    // segment is inside one of child quads?
//    if(idxchild_beg == idxchild_end && aux==2)
//        return; // nothing else to do, split_info is set to put it there

    const int going_up = seg.B->going_up() ? 1 : -1,
              going_right = seg.B->going_right() ? 1 : -1;

    // test edges ---------------------------------

    if(einfo.cross0)
        data.split_info |= FILL(Quad::BOTTOM_LEFT);

    if(einfo.cross1)
    {
        data.split_info |= FILL(Quad::BOTTOM_RIGHT);
        data.winding[Quad::BOTTOM_LEFT] += going_up;
    }

    if(einfo.cross2)
        data.split_info |= FILL(Quad::BOTTOM_LEFT);

    if(einfo.cross3)
    {
        data.split_info |= FILL(Quad::BOTTOM_LEFT) | FILL(Quad::BOTTOM_RIGHT);

        // only push a up/down shortcut if the segment does not
        // cross edges 6 or 13
        if(!einfo.cross16)
        {
            // doesn't cross edges 1 or 12?
            if(!einfo.cross17)
            {
                // in this case the segment ends at cell 1 and
                // we insert an up/down short cut
                if(going_right > 0)
                    data.split_info |= UP(Quad::BOTTOM_LEFT);
                else
                    data.split_info |= DOWN(Quad::BOTTOM_LEFT);
            }
            else
            {
                // the segment ends bellow the horizontal line
                // the shortcuts can be exchanged by increment/decrement
                // in the initial winding number
                data.winding[Quad::BOTTOM_LEFT] += going_right;
            }
        }
    }

    if(einfo.cross4)
    {
        data.split_info |= FILL(Quad::BOTTOM_RIGHT);
        // only push a up/down shortcut if the segment does not
        // cross the edge 13
        if(!einfo.cross13)
        {
            if(!einfo.cross12)
            {
                // in this case the segment ends at cell 1 and
                // we insert an up/down short cut
                if(going_right > 0)
                    data.split_info |= UP(Quad::BOTTOM_RIGHT);
                else
                    data.split_info |= DOWN(Quad::BOTTOM_RIGHT);
            }
            else
            {
                // the segment ends bellow the horizontal line
                // the shortcuts can be exchanged by increment/decrement
                // in the initial winding number
                data.winding[Quad::BOTTOM_RIGHT] += going_right;
            }
        }
    }

    if(einfo.cross5)
        data.split_info |= FILL(Quad::BOTTOM_LEFT) | FILL(Quad::TOP_LEFT);

    if(einfo.cross6)
    {
        data.split_info |= FILL(Quad::BOTTOM_RIGHT) | FILL(Quad::TOP_RIGHT);
        data.winding[Quad::TOP_LEFT] += going_up;
    }

    if(einfo.cross7)
        data.split_info |= FILL(Quad::TOP_LEFT);

    if(einfo.cross8)
    {
        data.split_info |= FILL(Quad::TOP_LEFT) | FILL(Quad::TOP_RIGHT);
        // only push a up/down shortcut if the segment does not
        // cross the edge 11 or edge 14
        if(!einfo.cross15)
        {
            if(!einfo.cross16)
            {
                // in this case the segment ends at cell 3 and
                // we insert an up/down short cut
                if(going_right > 0)
                    data.split_info |= UP(Quad::TOP_LEFT);
                else
                    data.split_info |= DOWN(Quad::TOP_LEFT);
            }
            else
            {
                // the segment ends bellow the horizontal line
                // the shortcuts can be exchanged by increment/decrement
                // in the initial winding number
                data.winding[Quad::TOP_LEFT] += going_right;
            }
        }
    }

    if(einfo.cross9)
    {
        data.split_info |= FILL(Quad::TOP_RIGHT);
        // only push a up/down shortcut if the segment does not
        // cross the edge 14
        if(!einfo.cross14)
        {
            if(!einfo.cross13)
            {
                // in this case the segment ends at cell 1 and
                // we insert an up/down short cut
                if(going_right > 0)
                    data.split_info |= UP(Quad::TOP_RIGHT);
                else
                    data.split_info |= DOWN(Quad::TOP_RIGHT);
            }
            else
            {
                // the segment ends bellow the horizontal line
                // the shortcuts can be exchanged by increment/decrement
                // in the initial winding number
                data.winding[Quad::TOP_RIGHT] += going_right;
            }
        }
    }

    if(einfo.cross10)
        data.split_info |= FILL(Quad::TOP_LEFT);

    if(einfo.cross11)
        data.split_info |= FILL(Quad::TOP_RIGHT);

    if(einfo.cross12)
    {
        // increment winding number of cells 0 and 1
        data.winding[Quad::BOTTOM_RIGHT] += going_up;
        data.winding[Quad::BOTTOM_LEFT] += going_up;
    }

    if(einfo.cross13)
    {
        // increment winding number of cells 2 and 3
        data.winding[Quad::TOP_RIGHT] += going_up;
        data.winding[Quad::TOP_LEFT] += going_up;
    }

    if(seg.is_shortcut())
    {
        float2 p = seg.shortcut_base();

        // segment is above quad?
        if(p.y >= quad.max.y)
            return;

        // segment is to the left?
        if(p.x < quad.min.x)
            return;

        // segment is to the right?
        if(p.x >= quad.max.x)
        {
            // cross 13?
            if(p.y <= quad.mp.y)
            {
                data.winding[Quad::TOP_LEFT] += (seg.flags & SEGFLAG_DOWN) ? -1 : 1;
                data.winding[Quad::TOP_RIGHT] += (seg.flags & SEGFLAG_DOWN) ? -1 : 1;

                // cross 12?
                if(p.y <= quad.min.y)
                {
                    data.winding[Quad::BOTTOM_LEFT] += (seg.flags & SEGFLAG_DOWN) ? -1 : 1;
                    data.winding[Quad::BOTTOM_RIGHT] += (seg.flags & SEGFLAG_DOWN) ? -1 : 1;
                }
            }
        }
    }
}/*}}}*/

__global__ void gpu_calc_segs_per_quad(const_SegmentsPtr segs,/*{{{*/
                                       const unsigned *segs_idquad,
                                       unsigned first_idquad_curlevel,
                                       unsigned *segs_per_quad)
{
    int idseg = blockIdx.x*BW + threadIdx.x;

    if(idseg >= segs.size)// || idseg!=3)
        return;

    unsigned idquad = segs_idquad[idseg];
    if(idquad < first_idquad_curlevel)
        return;

    atomicAdd(segs_per_quad+idquad-first_idquad_curlevel, 1);
}/*}}}*/

__launch_bounds__(BW,8)
__global__ void gpu_calc_split_qtree_level(/*{{{*/
    const_QuadsPtr quads,
    unsigned first_idquad_curlevel,
    SegmentsPtr segs,
    const unsigned char *segs_data,
    const unsigned *segs_idquad,
    unsigned *split_infos,
    char4 *winding_infos,
    unsigned *segs_per_quad)
{
    int idseg = blockIdx.x*BW + threadIdx.x;

    if(idseg >= segs.size)// || idseg!=3)
        return;

    ImplicitSegment<> seg;
    get_seg(segs, segs_data, idseg,seg);
    unsigned seg_idquad = segs_idquad[idseg];

    Quad quad;
    quads.get(seg_idquad, quad);

    bool should_split =
        seg_idquad >= first_idquad_curlevel &&
            (d_min_segs_split == 0 || segs_per_quad[seg_idquad-first_idquad_curlevel] >= d_min_segs_split);

    SplitData sdata;
    sdata.split_info = should_split ? 0 : 255;

    if(should_split)
    {
        (unsigned &)sdata.winding = 0;

        switch(seg.type)
        {
        case SEG_PATH:
            sdata.split_info = PATH(0) | PATH(1) | PATH(2) | PATH(3);
#pragma unroll
            for(int c=0; c<4; ++c)
                ((char *)&sdata.winding)[c] = segs.winding[idseg];
            break;
        case SEG_COMMIT:
            sdata.split_info = COMMIT(0) | COMMIT(1) | COMMIT(2) | COMMIT(3);
            break;
        case SEG_POP_CLIP:
            sdata.split_info = POP_CLIP(0) | POP_CLIP(1) | POP_CLIP(2) | POP_CLIP(3);
            break;
        case SEG_PUSH:
            sdata.split_info = PUSH(0) | PUSH(1) | PUSH(2) | PUSH(3);
            break;
        case SEG_POP_FILL:
            sdata.split_info = POP_FILL(0) | POP_FILL(1) | POP_FILL(2) | POP_FILL(3);
            break;
        default:
            assert(seg.type <= SEG_ARC);
            EdgeIntersectionInfo einfo;
            switch(seg.type)
            {
            case SEG_LINEAR:
                get_edge_intersection_info(seg.as<1>(), quad, einfo);
                break;
            case SEG_QUADRATIC:
                get_edge_intersection_info(seg.as<2>(), quad, einfo);
                break;
            case SEG_CUBIC:
                get_edge_intersection_info(seg.as<3>(), quad, einfo);
                break;
            case SEG_ARC:
                get_edge_intersection_info(seg.as<SEG_ARC>(), quad, einfo);
                break;
            }
            calc_split_data(seg.as_curve(),einfo, quad, sdata);
        }
    }

    split_infos[idseg] = sdata.split_info;
    winding_infos[idseg] = (const char4 &)sdata.winding;
}/*}}}*/

// returns the n'th bit set position (1-based)
__device__ int ffsn(unsigned v, unsigned n)/*{{{*/
{
    for(int i=0; i<n-1; ++i)
        v &= (v-1); // set rightmost set bit to zero
    return __ffs(v);
}/*}}}*/

__launch_bounds__(BW,8)
__global__ void gpu_split_qtree_level(SegmentsPtr segs,/*{{{*/
                                      const unsigned char *segs_data,
                                      const unsigned *segs_idquad,
                                      SegmentsPtr outsegs,
                                      unsigned *outsegs_idquad,
                                      unsigned first_idquad_curlevel,
                                      unsigned nquads_curlevel,
                                      unsigned newqtree_size,
                                      const unsigned *split_infos,
                                      const char4 *winding_infos,
                                      const unsigned *newseg_offset,
                                      const unsigned *idnewquads)
{
    unsigned idseg = blockIdx.x*BW + threadIdx.x;

    if(idseg >= segs.size)
        return;

    ImplicitSegment<> seg;
    get_seg(segs, segs_data, idseg, seg);
    unsigned seg_idquad = segs_idquad[idseg];

    unsigned splitinfo = split_infos[idseg];

    if(DONT_SPLIT(splitinfo))
    {
        unsigned idnewseg = newseg_offset[idseg];

        outsegs_idquad[idnewseg] = seg_idquad;
        outsegs.type[idnewseg] = seg.type;
        outsegs.idx_data[idnewseg] = (unsigned int)((const unsigned char *)seg.data - segs_data);
        outsegs.winding[idnewseg] = segs.winding[idseg];
        outsegs.flags[idnewseg] = seg.flags;
        return;
    }

    char4 winding_info = winding_infos[idseg];

    for(unsigned idchild=0; idchild<4; ++idchild)
    {
        unsigned idnewseg = newseg_offset[idchild*segs.size + idseg];
        unsigned idnewquad = first_idquad_curlevel + nquads_curlevel
            + idnewquads[nquads_curlevel*idchild + (seg_idquad-first_idquad_curlevel)];

        if(splitinfo & PATH(idchild))
        {
            // there must be only path segments set in splitinfo
            assert((splitinfo & ~(PATH(0)|PATH(1)|PATH(2)|PATH(3))) == 0);
            assert(segs.type[idseg] == SEG_PATH);

            outsegs_idquad[idnewseg] = idnewquad;
            outsegs.type[idnewseg] = SEG_PATH;
            outsegs.idx_data[idnewseg] = (unsigned int)((const unsigned char *)seg.data - segs_data);
            outsegs.winding[idnewseg] = ((char *)&winding_info)[idchild];
            outsegs.flags[idnewseg] = seg.flags;
            ++idnewseg;
            continue;
        }

        // up/down must have a FILL seg where it'll be stored
        assert((splitinfo & (UP(idchild)|DOWN(idchild))) == 0 || (splitinfo & FILL(idchild)));

        if(splitinfo & FILL(idchild))
        {
            assert((splitinfo & PATH(idchild)) == 0);
            assert(segs.type[idseg] != SEG_PATH);

            outsegs_idquad[idnewseg] = idnewquad;
            outsegs.type[idnewseg] = seg.type;
            outsegs.idx_data[idnewseg] = (unsigned int)((const unsigned char *)seg.data - segs_data);

            unsigned flags = 0;

            if(splitinfo & UP(idchild))
                flags = SEGFLAG_UP;
            else if(splitinfo & DOWN(idchild))
                flags = SEGFLAG_DOWN;

            outsegs.flags[idnewseg] = flags;

            ++idnewseg;
        }

        if(splitinfo & COMMIT(idchild))
        {
            assert((splitinfo & PATH(idchild)) == 0);
            assert(segs.type[idseg] != SEG_PATH);

            outsegs_idquad[idnewseg] = idnewquad;
            outsegs.type[idnewseg] = SEG_COMMIT;
            ++idnewseg;
        }
        else if(splitinfo & PUSH(idchild))
        {
            assert((splitinfo & PATH(idchild)) == 0);
            assert(segs.type[idseg] != SEG_PATH);

            outsegs_idquad[idnewseg] = idnewquad;
            outsegs.type[idnewseg] = SEG_PUSH;
            ++idnewseg;
        }
        else if(splitinfo & POP_FILL(idchild))
        {
            assert((splitinfo & PATH(idchild)) == 0);
            assert(segs.type[idseg] != SEG_PATH);

            outsegs_idquad[idnewseg] = idnewquad;
            outsegs.type[idnewseg] = SEG_POP_FILL;
            ++idnewseg;
        }
        else if(splitinfo & POP_CLIP(idchild))
        {
            assert((splitinfo & PATH(idchild)) == 0);
            assert(segs.type[idseg] != SEG_PATH);

            outsegs_idquad[idnewseg] = idnewquad;
            outsegs.type[idnewseg] = SEG_POP_CLIP;
            ++idnewseg;
        }
    }
}/*}}}*/

__global__ void gpu_grow_qtree(QuadsPtr quads,/*{{{*/
                               unsigned first_idquad_curlevel,
                               unsigned nquads_curlevel,
                               unsigned quad_count,
                               const unsigned *idnewquads)
{
    int idquad = blockIdx.x*BW + threadIdx.x + first_idquad_curlevel;

    if(idquad >= first_idquad_curlevel+nquads_curlevel)
        return;

    Quad quad;
    quads.get(idquad, quad);

    // children initially point to empty quad
    quad.children = make_int4(-1,-1,-1,-1);

    for(int i=0; i<4; ++i)
    {
        unsigned idxchild = i*nquads_curlevel + (idquad - first_idquad_curlevel);

        unsigned idrelchild = idnewquads[idxchild];

        // no segments in this child quad? don't grow it
        if(idrelchild == idnewquads[idxchild+1])
            continue;

        unsigned idnewquad = idrelchild + first_idquad_curlevel + nquads_curlevel;

        Quad child;

        switch(i)
        {
        case Quad::BOTTOM_LEFT:
            quad.children.x = idnewquad;
            child.min = quad.min;
            child.max = quad.mp;
            break;
        case Quad::BOTTOM_RIGHT:
            quad.children.y = idnewquad;
            child.min.x = quad.mp.x;
            child.min.y = quad.min.y;
            child.max.x = quad.max.x;
            child.max.y = quad.mp.y;
            break;
        case Quad::TOP_LEFT:
            quad.children.z = idnewquad;
            child.min.x = quad.min.x;
            child.min.y = quad.mp.y;
            child.max.x = quad.mp.x;
            child.max.y = quad.max.y;
            break;
        case Quad::TOP_RIGHT:
            quad.children.w = idnewquad;
            child.min = quad.mp;
            child.max = quad.max;
            break;
        }

        child.mp = (child.max + child.min)*0.5f;
        child.children = make_int4(-1,-1,-1,-1); // denotes leaf quad

        quads.put(idnewquad, child);
    }

    quads.put(idquad, quad);
}/*}}}*/

void print(const Quads &qtree, int first_idquad_curlevel, int nquads_curlevel)/*{{{*/
{
    std::vector<int4> children;
    qtree.children.copy_to_host(children);
    std::vector<float2> mp;
    qtree.mp.copy_to_host(mp);
    for(unsigned i=0; i<mp.size(); ++i)
    {
        if(i==first_idquad_curlevel)
            printf("---- curlevel\n");
        if(i==first_idquad_curlevel+nquads_curlevel)
            printf("---- nextlevel\n");
        printf("quad %d: %d %d %d %d - %f %f\n",i,children[i].x,children[i].y,children[i].z,children[i].w,
               mp[i].x, mp[i].y);
    }
}/*}}}*/
void print(const Segments &segs,/*{{{*/
           const gpu::vector<unsigned> &d_idquads,
           const gpu::vector<unsigned> &sinfos = gpu::vector<unsigned>())
{
    std::vector<SegType> types;
    segs.type.copy_to_host(types);
    std::vector<unsigned> idquads;
    d_idquads.copy_to_host(idquads);
    std::vector<unsigned> idx_datas;
    segs.idx_data.copy_to_host(idx_datas);
    std::vector<int> winding;
    segs.winding.copy_to_host(winding);
#if 0
    for(unsigned i=0; i<idquads.size(); ++i)
        printf("%2d %2d: %2d, type %d, split %8lx\n",idquads[i], idpaths[i], i, types[i],
               !sinfos.empty()?sinfos[i]:0);
    }
#endif
#if 0
    int depth = 0;
    for(unsigned i=0; i<idquads.size(); ++i)
    {
        if(types[i] == SEG_PUSH)
            ++depth;

        std::ostringstream ss;
        ss << types[i];
        std::string spc = std::string(depth*4-(is_stack(types[i])?2:0),' ');

        printf("%5d %5d %5d [%2d]: %s %s\n",i,idquads[i], depth, spc.c_str(), ss.str().c_str());

        if(types[i] == SEG_POP_FILL || types[i] == SEG_POP_CLIP)
            --depth;
    }
#endif

    for(unsigned i=0; i<types.size(); ++i)
    {
        std::ostringstream ss;
        ss << types[i];
        //printf("%d: %s ",idquads[i],ss.str().c_str());

        if(types[i] == SEG_PATH)
        {
            const PathData *dpdata = (const PathData *)(segs._data + idx_datas[i]);
            PathData pdata;
            cudaMemcpy(&pdata, dpdata, sizeof(pdata), cudaMemcpyDeviceToHost);

            printf(" idpath: %d, winding: %d ",pdata.idpath,winding[i]);
        }

        if(i < types.size()-1 && idquads[i]!=idquads[i+1]
           && types[i] != SEG_PATH)
        {
            printf("FAIL\n");
        }
        else
            printf("\n");
    }
#if 0
    printf("b1: %lu\n",thrust::count(segs.type.begin(), segs.type.end(), 1));
    printf("b2: %lu\n",thrust::count(segs.type.begin(), segs.type.end(), 2));
    printf("b3: %lu\n",thrust::count(segs.type.begin(), segs.type.end(), 3));
    printf("ac: %lu\n",thrust::count(segs.type.begin(), segs.type.end(), SEG_ARC));
    printf("up: %lu\n",thrust::count(segs.type.begin(), segs.type.end(), SEG_UP));
    printf("dn: %lu\n",thrust::count(segs.type.begin(), segs.type.end(), SEG_DOWN));
    printf("p : %lu\n",thrust::count(segs.type.begin(), segs.type.end(), SEG_PATH));
    printf("( : %lu\n",thrust::count(segs.type.begin(), segs.type.end(), SEG_PUSH));
    printf(")f: %lu\n",thrust::count(segs.type.begin(), segs.type.end(), SEG_POP_FILL));
    printf(")c: %lu\n",thrust::count(segs.type.begin(), segs.type.end(), SEG_POP_CLIP));
    printf("| : %lu\n",thrust::count(segs.type.begin(), segs.type.end(), SEG_COMMIT));
    printf("= : %lu\n",segs.type.size());
#endif
}/*}}}*/
void print(const Segments &segs,/*{{{*/
           const gpu::vector<int> &d_depth,
           const gpu::vector<int> &d_push,
           const gpu::vector<int> &d_commit,
           const gpu::vector<int> &d_pop,
           const gpu::vector<int4> &d_rem,
           const gpu::vector<uint4> &d_geomcount,
           const gpu::vector<unsigned> d_sinfo,
           const gpu::vector<unsigned> d_dist_to_next_cell)
{
    std::vector<uint4> geomcount;
    std::vector<int> push,commit,pop,depth;
    std::vector<unsigned> quad, path;
    std::vector<unsigned> sinfo;
    std::vector<SegType> type;
    std::vector<int4> rem;
    d_push.copy_to_host(push);
    d_commit.copy_to_host(commit);
    d_pop.copy_to_host(pop);
    segs.type.copy_to_host(type);
    d_depth.copy_to_host(depth);
    d_rem.copy_to_host(rem);
    d_sinfo.copy_to_host(sinfo);
    d_geomcount.copy_to_host(geomcount);
    std::vector<unsigned> dist_to_next_cell;
    d_dist_to_next_cell.copy_to_host(dist_to_next_cell);
    for(int i=0; i<segs.size(); ++i)
    {
        std::ostringstream ss;
        ss << type[i];
        std::string spc(depth[i]*4-(is_stack(type[i])?2:0),' ');
        printf("%5d q:%5d dq:%3d p:%5d [%2d]: %3d %3d %3d | ",
               i, quad[i], dist_to_next_cell[i], path[i], depth[i],
               push[i], commit[i], pop[i]);
//        print_sinfo(sinfo[i], paints[i]==NULL);
        printf(" | %3d %3d %3d %3d",
               geomcount[i].x, geomcount[i].y, geomcount[i].z, geomcount[i].w);
        printf(" | %2d %2d %2d %2d | %s %s\n",
               rem[i].x, rem[i].y, rem[i].z, rem[i].w,
               spc.c_str(), ss.str().c_str());
        if(i<segs.size()-1 && quad[i]!=quad[i+1])
            printf("----------------------------------\n");
    }
}/*}}}*/

__global__ void gpu_collect_sinfo_stats(const_SegmentsPtr segs,/*{{{*/
                                        const unsigned *segs_idquad,
                                        const unsigned *sinfos,
                                        unsigned first_idquad_curlevel,
                                        unsigned nquads_curlevel,
                                        unsigned *newsegs_in_child_count,
                                        unsigned *used_child_quads)
{
    unsigned idseg = blockIdx.x*BW + threadIdx.x;
    if(idseg >= segs.size)
        return;

    unsigned sinfo = sinfos[idseg];

    if(DONT_SPLIT(sinfo))
    {
        newsegs_in_child_count[idseg] = 1;
        for(int i=1; i<4; ++i)
            newsegs_in_child_count[i*segs.size+idseg] = 0;
        return;
    }

    unsigned idquad = segs_idquad[idseg];

    for(int i=0; i<4; ++i)
    {
        unsigned nsegs_child = __popc(sinfo & SEG_TYPES(i));

        // TODO: evaluate if changing this order to 4*idseg + i,
        // gives better memmory access pattern. Can gather this
        // 4 writes in one operation at the end...
        newsegs_in_child_count[i*segs.size+idseg] = nsegs_child;

        // mark the cell as used
        // if nsegs_child==0, this cell will be empty
        atomicCAS(used_child_quads + i*nquads_curlevel + (idquad-first_idquad_curlevel),
                  0, 1);
    }
}/*}}}*/

__global__ void gpu_mark_path_tail(const_SegmentsPtr segs, unsigned *tailflags)/*{{{*/
{
    int idseg = blockIdx.x*BW + threadIdx.x;
    if(idseg >= segs.size)
        return;

    tailflags[idseg] = segs.type[idseg] == SEG_PATH;
}/*}}}*/

__global__
void gpu_fill_seg_depth(const_SegmentsPtr segs, int *seg_depth)/*{{{*/
{
    unsigned idseg = blockIdx.x*BW+threadIdx.x;
    if(idseg >= segs.size)
        return;

    bool prev_is_pop = idseg > 0 && (segs.type[idseg-1] == SEG_POP_FILL ||
                                     segs.type[idseg-1] == SEG_POP_CLIP),
         cur_is_push = segs.type[idseg] == SEG_PUSH;

    int depth = 0;

    // if prev_is_pop and cur_is_push, we must write 1-1 = 0!

    if(prev_is_pop && !cur_is_push)
        depth = -1;
    else if(!prev_is_pop && cur_is_push)
        depth = 1;

    seg_depth[idseg] = depth;
}/*}}}*/

__global__ void gpu_mark_quad_head(const_SegmentsPtr segs,/*{{{*/
                                   const unsigned *segs_idquad,
                                   unsigned *headflags)
{
    int idseg = blockIdx.x*BW + threadIdx.x;
    if(idseg >= segs.size)
        return;

    if(idseg == 0)
        headflags[idseg] = 1;
    else
        headflags[idseg] = segs_idquad[idseg-1] != segs_idquad[idseg];
}/*}}}*/

__global__
void gpu_fill_push_headflag(const_SegmentsPtr segs, unsigned *head_flags,/*{{{*/
                            const int *seg_depth,
                            unsigned level)
{
    unsigned idseg = blockIdx.x*BW+threadIdx.x;
    if(idseg >= segs.size)
        return;

    if(seg_depth[idseg] == level)
        head_flags[idseg] = segs.type[idseg] == SEG_PUSH ? 1 : 0;
    else
        head_flags[idseg] = 0;
}/*}}}*/

__global__
void gpu_fill_push_links(const int *level_push_links, unsigned size,/*{{{*/
                         const int *seg_depths, unsigned level,
                         const SegType *seg_types,
                         int *seg_push_link,
                         int *seg_commit_link)
{
    unsigned idseg = blockIdx.x*BW+threadIdx.x;
    if(idseg >= size || seg_depths[idseg] != level)
        return;

    // push is *before* current segment, hence -dist
    seg_push_link[idseg] = idseg-level_push_links[idseg];

    // must initialize this in case of missing commit between push and pop
    seg_commit_link[idseg] = idseg;

    assert(seg_depths[idseg]==0 || seg_types[idseg] == SEG_POP_FILL ||
           seg_types[idseg] == SEG_POP_CLIP || idseg < size-1);
}/*}}}*/
__global__
void gpu_fill_commit_pop_links1(const_SegmentsPtr segs,/*{{{*/
                    const int *seg_depths,
                    const unsigned *dist_next_cell,
                    int *seg_link_push,
                    int *seg_link_commit,
                    int *seg_link_pop)
{
    unsigned idseg = blockIdx.x*BW+threadIdx.x;
    if(idseg >= segs.size)
        return;

    switch(segs.type[idseg])
    {
    case SEG_POP_FILL:
    case SEG_POP_CLIP:
        seg_link_pop[seg_link_push[idseg]] = idseg;
        break;
    case SEG_COMMIT:
        seg_link_commit[seg_link_push[idseg]] = idseg;
        break;
    default:
        // in segs with depth==0, idpop points to the where pop
        // would be
        if(seg_depths[idseg] == 0)
        {
            // set these to 0 just to make debug output look nice
            seg_link_push[idseg] = seg_link_commit[idseg] = 0;

            seg_link_pop[idseg] = idseg+dist_next_cell[idseg];
        }
        break;
    }
}/*}}}*/
__global__
void gpu_fill_commit_pop_links2(const_SegmentsPtr segs,/*{{{*/
                    const int *seg_depth,
                    const int *seg_link_push,
                    int *seg_link_commit,
                    int *seg_link_pop)
{
    unsigned idseg = blockIdx.x*BW+threadIdx.x;
    if(idseg >= segs.size || seg_depth[idseg] <= 0 || segs.type[idseg] == SEG_PUSH)
        return;

    int idpush = seg_link_push[idseg];

    seg_link_pop[idseg] = seg_link_pop[idpush];
    seg_link_commit[idseg] = seg_link_commit[idpush];

#if 0
    if(segs.type[seg_link_push[idseg]] != SEG_PUSH)
        printf("invalid push link: %d\n",idseg);

    if(segs.type[seg_link_commit[idseg]] != SEG_COMMIT)
        printf("invalid commit link: %d\n",idseg);

    if(segs.type[seg_link_pop[idseg]] != SEG_POP_CLIP &&
       segs.type[seg_link_pop[idseg]] != SEG_POP_FILL)
    {
        printf("invalid pop link: %d\n",idseg);
    }
#endif

}/*}}}*/

__global__ void gpu_mark_geom_segs(const_SegmentsPtr segs,/*{{{*/
                        uint *geom)
{
    unsigned idseg = blockIdx.x*BW+threadIdx.x;
    if(idseg >= segs.size)
        return;

    geom[idseg] = segs.type[idseg] <= SEG_LAST_GEOM ? 1 : 0;
}/*}}}*/

__global__
void gpu_mark_segs_to_be_removed(const_SegmentsPtr segs,/*{{{*/
                                 const unsigned char *segs_data,
                                 const int *seg_depth,
                                 const int *seg_links_push,
                                 const int *seg_links_commit,
                                 const int *seg_links_pop,
                                 const unsigned *geom_count,
                                 int *to_remove)
{
    unsigned idseg = blockIdx.x*BW+threadIdx.x;
    if(idseg >= segs.size)
        return;

    SegType type = segs.type[idseg];

#if ENABLE_PRUNING
    int idpush = seg_links_push[idseg],
        idcommit = seg_links_commit[idseg],
        idpop = seg_links_pop[idseg];
    const PathData *pdata = (const PathData *)(segs_data+segs.idx_data[idseg]);

    // is it a path segment (it has the winding information)
    if(type == SEG_PATH && pdata->intersects(segs.winding[idseg]))
    {
#if FRONT_TO_BACK_RENDERING
        // single segment in path?
        if(idseg==0 || segs.type[idseg-1] == SEG_PATH || is_stack(segs.type[idseg-1]))
        {
            // ( A | B 1F C ) -> ( A | B 1F )
            if(pdata->paint != NULL) // fill segment?
            {
                if(pdata->paint->type == Paint::SOLID)
                {
                    if(idseg+1 < segs.size)
                        atomicAdd(&to_remove[idseg+1], 1);

                    // for segments in depth==0 idpop points to the first
                    // segment of next cell
                    if(idpop < segs.size)
                        atomicAdd(&to_remove[idpop], -1);
                }
            }
            // ( A 1C B | C ) -> C
            else if(idcommit > idseg)
            {
                assert(idpush < segs.size);
                atomicAdd(&to_remove[idpush], 1);

                assert(idcommit+1 < segs.size);
                atomicAdd(&to_remove[idcommit+1], -1);

                assert(idpop < segs.size);
                atomicAdd(&to_remove[idpop], 1);

                if(idpop+1 < segs.size)
                    atomicAdd(&to_remove[idpop+1], -1);
            }
            // ( A | B 1C C ) -> ( A | 1C )
            else
            {
                assert(idcommit+1 < segs.size);
                atomicAdd(&to_remove[idcommit+1], 1);

                atomicAdd(&to_remove[idseg], -1);

                if(idseg+1 < segs.size)
                    atomicAdd(&to_remove[idseg+1], 1);

                assert(idpop < segs.size);
                atomicAdd(&to_remove[idpop], -1);
            }
        }
#endif
    }
    else
    {
        bool remove = false;

        int begrem = idpush, endrem = idpop;

        switch(type)
        {
        case SEG_PUSH:
            // ( | A ) -> ø
            // no paths between ( and |?
            assert(idcommit < segs.size);
            if(idseg != idcommit && geom_count[idseg] == geom_count[idcommit])
                remove = true;
            // ( ) -> ø
            else if(geom_count[idseg] == geom_count[idpop])
                remove = true;
            break;
        case SEG_COMMIT:
            // ( A | ) -> ø
            // no paths between | and )?
            assert(idpop < segs.size);
            if(idseg != idpop && geom_count[idseg] == geom_count[idpop])
                remove = true;
            break;
        case SEG_PATH:
            // empty path?
            if(segs.winding[idseg] == 0 &&
                (idseg == 0 || segs.type[idseg-1] == SEG_PATH ||
                   is_stack(segs.type[idseg-1])))
            {
                remove = true;
                begrem = endrem = idseg;
            }
            break;
        }

        if(remove)
        {
            assert(begrem < segs.size);
            atomicAdd(&to_remove[begrem], 1);

            if(endrem+1 < segs.size)
                atomicAdd(&to_remove[endrem+1], -1);
        }
    }
#else
    if(type == SEG_PATH)
    {
        // empty path?
        if(segs.winding[idseg] == 0 &&
            (idseg == 0 || segs.type[idseg-1] == SEG_PATH ||
               is_stack(segs.type[idseg-1])))
        {
            atomicAdd(&to_remove[idseg], 1);

            if(idseg+1 < segs.size)
                atomicAdd(&to_remove[idseg+1], -1);
        }
    }
#endif
}/*}}}*/

__global__ void gpu_create_head_flags(unsigned *hflags,/*{{{*/
                                      const unsigned *segs_idquad,
                                      size_t seg_count)
{
    int idseg = blockIdx.x*BW + threadIdx.x;

    if(idseg >= seg_count)
        return;

    if(idseg == 0)
        hflags[idseg] = 1;
    else
        hflags[idseg] = segs_idquad[idseg-1] != segs_idquad[idseg];
}/*}}}*/

__global__ void gpu_calc_quad_seg_ref(uint2 *seg_range,/*{{{*/
                                      const unsigned *seg_idquads,
                                      unsigned segs_size,
                                      const unsigned *hflags,
                                      const unsigned *scan)
{
    int idseg = blockIdx.x*BW + threadIdx.x+1;

    if(idseg > segs_size)
        return;

    if(idseg == segs_size || hflags[idseg])
    {
        int idquad = seg_idquads[idseg-1];

        seg_range[idquad] = make_uint2(idseg-scan[idseg-1],
                                       scan[idseg-1]);
    }
}/*}}}*/

void group_segs_by_quad(gpu::Cudpp &cudpp, gpu::vector<uint2> &seg_range,/*{{{*/
                        Segments &segs, const gpu::vector<unsigned> &segs_idquad)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("calc qtree -> seg"));

    /*
       2. create head flags (1 if prev_idquad != cur_idquad, 0 otherwise)
       3. inclusive scan on head flags, creating keys
       4. create vector with ones
       5. incl. segscan on ones vector using keys (called 'scan' henceforth)
       6. if head flag, then quad[idseg-1].base = idseg-scan[idseg-1],
                             quad[idseg-1].cnt = scan[idseg-1]
    */

    gpu::vector<unsigned> hflags(segs.size());
    {
        dim3 bdim(BW),
             gdim((segs.size()+bdim.x-1)/bdim.x);

        gpu_create_head_flags<<<gdim, bdim>>>(&hflags, &segs_idquad,
                                              segs_idquad.size());
    }

    gpu::vector<unsigned> scanned_hflags(segs.size());
    thrust::fill(scanned_hflags.begin(), scanned_hflags.end(), 1);

    cudpp.segmented_inclusive_scan(scanned_hflags, hflags, scanned_hflags);

    seg_range.fill(0);

    {
        dim3 bdim(BW),
             gdim((segs.size()+bdim.x-1)/bdim.x);

        gpu_calc_quad_seg_ref<<<gdim, bdim>>>(&seg_range,
                                              &segs_idquad, segs.size(),
                                              &hflags, &scanned_hflags);
    }

#if TRACE
    std::vector<uint2> h_segrange;
    seg_range.copy_to_host(h_segrange);
    for(unsigned i=0; i<h_segrange.size(); ++i)
        printf("Q%d: %d %d\n",i,h_segrange[i].x, h_segrange[i].y);

    std::vector<unsigned> h_idquads;
    segs.idquad.copy_to_host(h_idquads);
    for(unsigned i=0; i<h_idquads.size(); ++i)
        printf("S%d: %d\n",i,h_idquads[i]);
#endif
}/*}}}*/

struct greater_than_zero_predicate/*{{{*/
{
    __device__
    bool operator()(int v) const
    {
        return v > 0;
    }
};/*}}}*/

__global__
void gpu_gather_info( QuadsPtr qtree, const_SegmentsPtr segs )
{/*{{{*/
    int id = threadIdx.x + blockIdx.x*blockDim.x;
    if ( id >= qtree.size )
        return;
    uint2 range = qtree.seg_range[id];
    if ( range.y == 0 ) {
        qtree.mem_config[id] = make_short2( 0, 0 );
        return;
    }
    int num_segments = range.y;
    int data_size = 0;
    for (int k=0; k<num_segments; ++k) {
        int sid = range.x + k;
        SegType type = segs.type[sid];
        data_size += find_segment_size( type );
        if ( type == SEG_PATH ) {
            const PathData *pdata = segs.data<PathData>(sid);
            Paint::Type ptype = static_cast<Paint::Type>(pdata->paint_type);
            data_size += find_paint_size( ptype );
        }
    }
    int index_mem = round_to_m16(sizeof(SegmentIndex))*num_segments;
    int geom_mem = round_to_m16( data_size );
    qtree.mem_config[id] = make_short2( min(index_mem, (1<<15)-1), min(geom_mem, (1<<15)-1) );
    assert( qtree.mem_config[id].y > 0 );
    assert( qtree.mem_config[id].x > 0 );
}/*}}}*/

void create_qtree(Quads &qtree, Segments &segs,/*{{{*/
                  const gpu::Box &bbox, unsigned qtree_levels,
                  unsigned min_segs_split, bool dumpqtree)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("creating qtree"));

    qtree.clear();

    // no segments?
    if(segs.size() == 0)
        return;

    // limit qtree levels to the size of a pixel

    int _qtree_levels = std::min<int>(qtree_levels, log2(bbox.max.x-bbox.min.x));
    /*
    if(_qtree_levels != qtree_levels)
        std::cout << "Hard limiting qtree levels to " << _qtree_levels << std::endl;
        */

    qtree_levels = _qtree_levels;

    gpu::copy_to_symbol(d_min_segs_split, min_segs_split);

    gpu::vector<unsigned> segs_idquad;

    {
        std::vector<float2> mp, min, max;
        std::vector<int4> chld;

        // level 0 (root)
        chld.push_back(make_int4(-1,-1,-1,-1));
        mp.push_back(bbox.mp());
        min.push_back(bbox.min);
        max.push_back(bbox.max);

        // all segments belong to idquad 0 (root)
        segs_idquad.resize(segs.size());
        segs_idquad.fill(0);

        // copy to gpu
        qtree.mp.copy_from_host(mp);
        qtree.min.copy_from_host(min);
        qtree.max.copy_from_host(max);
        qtree.children.copy_from_host(chld);
    }

#if TRACE
    print(qtree, -1, 0);
#endif

    gpu::Cudpp cudpp;
    gpu::CUB cub;

    cudaStream_t stream[2];

    for(unsigned i=0; i<sizeof(stream)/sizeof(stream[0]); ++i)
    {
        cudaStreamCreate(&stream[i]);
        gpu::check_cuda_error("Error creating stream");
    }

    // segs will be rewritten in each level but seg data
    // must not be changed, we'll just use segdata from here
    // and set it back to segs at the end.
    gpu::vector<unsigned char> segs_data;
    swap(segs_data, segs._data);

    /*
       current level -> new level
    */

    // Index of the first quad cell at this level.
    // Updated for every level.
    unsigned first_idquad_curlevel = 0;

    for(unsigned level=1; level<qtree_levels; ++level)
    {
#if TRACE
        printf("LEVEL %d ---------------------------\n",level);
#endif

        std::string slevel;
        {
            std::ostringstream ss;
            ss << level;
            slevel = ss.str();
        }

        unsigned nquads_curlevel = qtree.size() - first_idquad_curlevel;
        unsigned newsegs_count, newquads_count;

        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("level #"+slevel));

        gpu::vector<unsigned> split_infos;
        gpu::vector<char4> winding_infos;
        gpu::vector<unsigned> newsegs_offset, used_child_quads, idnewquads;
        {
//            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("calc seg splits #"+slevel));
            /*
               1. calc where each segment will go. Output:
                   - split info per segment (split_infos)
                   - # of segments per child quad of current level (segs_in_children)
                   - # of new segments generated per segment (idnewsegs)
            */

            gpu::vector<unsigned> segs_per_quad;
            if(min_segs_split > 0)
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("calc segs per cell"));

                segs_per_quad.resize(nquads_curlevel);
                segs_per_quad.fill(0);

                dim3 bdim(BW),
                     gdim((segs.size()+BW-1)/BW);

                gpu_calc_segs_per_quad<<<gdim,bdim>>>(&segs, &segs_idquad,
                                                      first_idquad_curlevel,
                                                      &segs_per_quad);
            }

            split_infos.resize(segs.size());
            winding_infos.resize(segs.size());

            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("calc seg split"));

                dim3 bdim(BW),
                     gdim((segs.size()+BW-1)/BW);

                cudaFuncSetCacheConfig(gpu_calc_split_qtree_level, cudaFuncCachePreferL1);

                // for each segment in 'segs', find the split info.
                gpu_calc_split_qtree_level<<<gdim,bdim>>>(&qtree,
                                                          first_idquad_curlevel, //unused
                                                          &segs,
                                                          &segs_data,
                                                          &segs_idquad,
                                                          &split_infos,
                                                          &winding_infos,
                                                          &segs_per_quad);
            }

            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simplify windings"));

                gpu::vector<unsigned> path_tail(segs.size()),
                                      path_key(segs.size());

                dim3 bdim(BW),
                     gdim((segs.size()+BW-1)/BW);
                gpu_mark_path_tail<<<gdim,bdim>>>(&segs, &path_tail);

                cudpp.exclusive_scan(path_tail, path_key);

                // sum the winding numbers for each path in each cell.
                thrust::inclusive_scan_by_key(path_key.begin(),
                                              path_key.end(),
                                              winding_infos.begin(),
                                              winding_infos.begin());
            }

            // collect sinfo stats
            gpu::vector<unsigned> newsegs_count_in_child_quad(segs.size()*4+1);
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("collect sinfo stats"));

                used_child_quads.resize(nquads_curlevel*4+1); // +1 -> total, after scan
                used_child_quads.fill(0);

                // fills newsegs_count_... with the number of segments
                // in the new cell. Ordered by seg and then cell#: seg0/cell0,
                // seg1/cell0, ... segN-1/cell0, seg0/cell1, ..., segN-1/cell3.
                // Marks the used_child_quads with 1. Ordered by quad and cell:
                // quad0/cell0, quad1/cell0,..., quadN-1/cell0,...
                gpu_collect_sinfo_stats<<<
                    dim3((segs.size()+BW-1)/BW), dim3(BW)>>>(
                        &segs, &segs_idquad, &split_infos,
                        first_idquad_curlevel,
                        nquads_curlevel,
                        &newsegs_count_in_child_quad,
                        &used_child_quads);
#if 0
                std::vector<unsigned> sinfos, idquads, idpaths;
                split_infos.copy_to_host(sinfos);
                segs.idquad.copy_to_host(idquads);
                segs.idpath.copy_to_host(idpaths);
                for(unsigned i=0; i<sinfos.size(); ++i)
                    printf("%3d %3d: %8x\n",idquads[i], idpaths[i], sinfos[i]);
#endif
            }

            {
                cudaDeviceSynchronize();

                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("scan new segs count"));
                //finds the number of new segments.
                cub.exclusive_scan(newsegs_count_in_child_quad,
                                   newsegs_count_in_child_quad,
                                   gpu::OP_ADD, stream[0]);

            }

            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("calc new idquads"));
                //finds the number of new quads.
                cub.exclusive_scan(used_child_quads, used_child_quads,
                                   gpu::OP_ADD, stream[1]);

            }

            // previous kernel runs faster, so load its output first
            swap(used_child_quads, idnewquads);

            newquads_count = idnewquads.back(stream[1]);

#if TRACE
            printf("new quads: %d\n",newquads_count);
#endif
            if(newquads_count == 0)
                break;
            idnewquads.resize(idnewquads.size()-1);

            swap(newsegs_count_in_child_quad, newsegs_offset);

            newsegs_count = newsegs_offset.back(stream[0]);
            newsegs_offset.resize(newsegs_offset.size()-1);
#if TRACE
            printf("segs size: %lu\n",newsegs_count);
#endif
        }

        // create segments for this level {{{
        {

            // we won't resize seg_begin and seg_count right now
            qtree.mp.resize(qtree.size()+newquads_count);
            qtree.min.resize(qtree.mp.size());
            qtree.max.resize(qtree.mp.size());
            qtree.children.resize(qtree.mp.size());

            Segments newsegs;
            gpu::vector<unsigned> newsegs_idquad;
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("split segs"));

                newsegs.resize(newsegs_count);
                newsegs_idquad.resize(newsegs_count);

                dim3 bdim(BW),
                     gdim((segs.size()+BW-1)/BW);

                /* 8. use previous info to generate new segments for the next level */

                cudaFuncSetCacheConfig(gpu_split_qtree_level, cudaFuncCachePreferL1);

                gpu_split_qtree_level<<<gdim,bdim,0,stream[0]>>>(&segs, &segs_data, &segs_idquad,
                                                     &newsegs,
                                                     &newsegs_idquad,
                                                     first_idquad_curlevel,
                                                     nquads_curlevel,
                                                     qtree.size()+newquads_count,
                                                     &split_infos,
                                                     &winding_infos,
                                                     &newsegs_offset,
                                                     &idnewquads);
                swap(segs, newsegs);
                swap(segs_idquad, newsegs_idquad);
            }

#if 0
                {
                    printf("RESULT ------------------\n");
                    std::vector<unsigned> quad;
                    std::vector<SegType> type;
                    segs.type.copy_to_host(type);
                    segs.idquad.copy_to_host(quad);
                    unsigned depth = 0;
                    for(int i=0; i<segs.size(); ++i)
                    {
                        if(type[i] == SEG_PUSH)
                            ++depth;
                        std::ostringstream ss;
                        ss << type[i];
                        std::string spc(depth*4-(is_stack(type[i])?2:0),' ');
                        printf("%3d %3d [%2d]: %s %s\n",
                               i, quad[i], depth,
                               spc.c_str(), ss.str().c_str());

                        if(type[i] == SEG_POP_FILL || type[i] == SEG_POP_CLIP)
                            --depth;
                    }
                }
#endif

            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("grow qtree #"+slevel));

                dim3 bdim(BW),
                     gdim((nquads_curlevel+BW-1)/BW);

                cudaFuncSetCacheConfig(gpu_grow_qtree, cudaFuncCachePreferL1);

                gpu_grow_qtree<<<gdim,bdim,0,stream[1]>>>(&qtree,
                                              first_idquad_curlevel,
                                              nquads_curlevel,
                                              qtree.size(),
                                              idnewquads);
            }
        }/*}}}*/

        // segment pruning {{{
        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("segment pruning"));

#if ENABLE_PRUNING
            gpu::vector<unsigned> dist_next_cell(segs.size());
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("dist next cell"));

                thrust::fill(dist_next_cell.begin(), dist_next_cell.end(), 1);

                thrust::inclusive_scan_by_key(segs_idquad.rbegin(),
                                              segs_idquad.rend(),
                                              dist_next_cell.rbegin(),
                                              dist_next_cell.rbegin());
            }

            /*
               3. segment depth
                 - create int array with same size as segment array,
                   initialize with zero
                 - set 1 when pushing, -1 just after pop
                 - inclusive scan
                 - max-element = depth
            */
            gpu::vector<int> seg_depth(segs.size());
            unsigned depth;
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("calc seg depth"));
                dim3 bdim(BW),
                     gdim((seg_depth.size()+BW-1)/BW);
                gpu_fill_seg_depth<<<gdim,bdim>>>(&segs, &seg_depth);

                cub.inclusive_scan(seg_depth, seg_depth);

                gpu::vector<int>::const_iterator
                    itmax = thrust::max_element(seg_depth.begin(),
                                                seg_depth.end());
                depth = *itmax;
            }

            gpu::vector<int> seg_link_push(segs.size()),
                seg_link_commit(segs.size());

            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("link to push"));
                for(int slevel=1; slevel<=depth; ++slevel)
                {
                    /*
                       4. link to push
                         - for each level, excl-seg-scan a 1-filled int array
                           (size == parent segment array) boundary on push in
                           current level (since stack segments are copied to
                           all children, we can use the parent segment array to
                           create these links)
                    */
                    gpu::vector<int> level_link_push(segs.size());
                    thrust::fill(level_link_push.begin(),
                                 level_link_push.end(), 1);

                    gpu::vector<unsigned> head_flags(segs.size());

                    dim3 bdim(BW),
                         gdim((segs.size()+BW-1)/BW);

                    gpu_fill_push_headflag<<<gdim, bdim>>>(&segs,
                                                           &head_flags,
                                                           &seg_depth,
                                                           slevel);
                    cudpp.segmented_exclusive_scan(level_link_push,
                                                   head_flags,
                                                   level_link_push);
                    gpu_fill_push_links<<<gdim, bdim>>>(
                        &level_link_push, segs.size(),
                        &seg_depth, slevel, &segs.type,
                        &seg_link_push,
                        &seg_link_commit);
                }
            }
            /*
               5. links to pop and commit
                 - if is pop then
                     pop->push[pop_dist] = pop[push_dist]
                   elif is commit then
                     commit->push[commit_dist] = -commit[push_dist]
                   end
            */
            gpu::vector<int> seg_link_pop(segs.size());
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("link to commit and pop"));
                dim3 bdim(BW),
                     gdim((segs.size()+BW-1)/BW);
                gpu_fill_commit_pop_links1<<<gdim, bdim>>>(&segs,
                                               &seg_depth,
                                               &dist_next_cell,
                                               &seg_link_push,
                                               &seg_link_commit,
                                               &seg_link_pop);
                gpu_fill_commit_pop_links2<<<gdim, bdim>>>(&segs,
                                               &seg_depth,
                                               &seg_link_push,
                                               &seg_link_commit,
                                               &seg_link_pop);

            }

            gpu::vector<unsigned> geom_count(segs.size());
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("geom count"));

                dim3 bdim(BW),
                     gdim((segs.size()+BW-1)/BW);

                gpu_mark_geom_segs<<<gdim,bdim>>>(&segs, &geom_count);

                cub.inclusive_scan(geom_count, geom_count);
            }

            /*
           5. mark segments to be removed
             - ( A | B 1F C ) -> ( A | B 1F )
                next(1F) += 1
                1F->push->pop += -1

             - ( | A ) -> ø
                if next(push) == commit then
                    push += 1
                    next(push->pop) += -1
                end

             - ( A | ) -> ø
                if next(commit) == pop then
                    commit->push += 1
                    next(commit+1) += -1
                end

             - ( A 1C B | C ) -> C
                if 1C->push[commit_dist] > 1C[push_dist] then
                    next(1C->push) += 1
                    1C += -1
                    next(1C) += 1;
                    1C->commit += -1
                end

             - ( A | B 1C C ) -> ( A | 1C )
                if 1C->push[commit_dist] < 1C[push_dist] then
                    1C->push += 1
                    next(1C->push) += -1
                    1C->push->commit += 1
                    next(1C->push->pop += -1)
                end
            */
#endif // enable pruning

            gpu::vector<int> removed_segs(segs.size());
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("mark segs to be removed"));
                removed_segs.fill(0);

                dim3 bdim(BW),
                     gdim((segs.size()+BW-1)/BW);

#if ENABLE_PRUNING
                gpu_mark_segs_to_be_removed<<<gdim,bdim>>>(
                    &segs, &segs_data, &seg_depth,
                    &seg_link_push, &seg_link_commit, &seg_link_pop,
                    &geom_count, &removed_segs);
#else
                gpu_mark_segs_to_be_removed<<<gdim,bdim>>>(
                    &segs, &segs_data, NULL,
                    NULL, NULL, NULL,
                    NULL, &removed_segs);
#endif

                cub.inclusive_scan(removed_segs, removed_segs);
            }

#if 0
            {
                std::vector<int> rem;
                removed_segs.copy_to_host(rem);
                std::vector<unsigned> idquads;
                segs_idquad.copy_to_host(idquads);
                std::vector<SegType> types;
                segs.type.copy_to_host(types);
                std::vector<int> winding;
                segs.winding.copy_to_host(winding);
                std::vector<unsigned> idx_datas;
                segs.idx_data.copy_to_host(idx_datas);
                for(int i=0; i<types.size(); ++i)
                {
                    std::ostringstream ss;
                    ss << types[i];
                    printf("%d: %d %s",idquads[i], rem[i], ss.str().c_str());
                    if(types[i] == SEG_PATH)
                    {
                        const PathData *dpdata = (const PathData *)(segs_data + idx_datas[i]);
                        PathData pdata;
                        cudaMemcpy(&pdata, dpdata, sizeof(pdata), cudaMemcpyDeviceToHost);

                        printf(" idpath: %d, winding: %d ",pdata.idpath,winding[i]);
                    }

                    if(i < types.size()-1 && idquads[i]!=idquads[i+1]
                       && types[i] != SEG_PATH)
                    {
                        printf(" FAIL\n");
                    }
                    else
                        printf("\n");
                }
            }
#endif

            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("remove segs"));

                typedef thrust::zip_iterator<thrust::tuple<gpu::vector<SegType>::iterator,
                                                           gpu::vector<unsigned>::iterator,
                                                           gpu::vector<int>::iterator,
                                                           gpu::vector<unsigned short>::iterator,
                                                           gpu::vector<unsigned>::iterator> > zip_type;

                zip_type begin = make_zip_iterator(thrust::make_tuple(segs.type.begin(),
                                                                   segs.idx_data.begin(),
                                                                   segs.winding.begin(),
                                                                   segs.flags.begin(),
                                                                   segs_idquad.begin())),
                         end = make_zip_iterator(thrust::make_tuple(segs.type.end(),
                                                                   segs.idx_data.end(),
                                                                   segs.winding.end(),
                                                                   segs.flags.end(),
                                                                   segs_idquad.end()));

                zip_type itend = remove_if(begin, end, removed_segs.begin(), greater_than_zero_predicate());

                unsigned nremoved = distance(itend, end);

#if TRACE
                printf("PRUNED: %d\n",nremoved);
#endif

                segs.resize(segs.size()-nremoved);
                segs_idquad.resize(segs.size());
            }
        }/*}}}*/

        if(level+1 < qtree_levels)
            cudaDeviceSynchronize();
#if TRACE
        print(qtree, first_idquad_curlevel, nquads_curlevel);
#endif

        first_idquad_curlevel += nquads_curlevel;
//        print(segs);
    }

    // this won't stop the streams
    for(unsigned i=0; i<sizeof(stream)/sizeof(stream[0]); ++i)
        cudaStreamDestroy(stream[i]);

    swap(segs_data, segs._data);

    if ( dumpqtree ) {
        //dump qtree
        std::vector<float2> qmin, qmax;
        std::vector<int4> ch;
        qtree.min.copy_to_host(qmin);
        qtree.max.copy_to_host(qmax);
        qtree.children.copy_to_host(ch);
        dump_qtree( qmin, qmax, ch, stdout );
    }

#if TRACE
    print(segs, segs_idquad);
#endif

    qtree.seg_range.resize(qtree.size());

    group_segs_by_quad(cudpp, qtree.seg_range, segs, segs_idquad);
    

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("gather cell statistics"));
        qtree.mem_config.resize(qtree.size());
        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("find mem data"));
            //find qtree statistics
            int bsize = 64;
            int gsize = ( segs.size() + bsize - 1 )/bsize;
            dim3 grid( gsize ), block( bsize );
            cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);
            gpu_gather_info<<<grid, block>>>( &qtree, &segs ); 
            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
        }
    }

    //printf("# of qtree nodes: %lu\n",qtree.size());
    //printf("# of segs: %lu\n",segs.size());
}/*}}}*/

