#ifndef GPURASTER_QUAD_H
#define GPURASTER_QUAD_H

#if !defined(__CUDA_ARCH__) || __CUDA_ARCH__ <= 300
template <class T>
__device__ inline const T &__ldg(const T *p)
{
    return *p;
}
#endif

struct Quad
{
    enum
    {
        BOTTOM_LEFT,
        BOTTOM_RIGHT,
        TOP_LEFT,
        TOP_RIGHT
    };

    float2 mp, min, max;

    int4 children;

    __device__ __host__
    int classify(const float2 &p) const;

    __device__
    bool contains(const float2 &p) const
    {
        /*
        return p.x >= min.x && p.x < max.x && 
               p.y >= min.y && p.y < max.y;
               */
        return p.x >= min.x && p.x <= max.x &&
               p.y >= min.y && p.y <= max.y;
    }
};

__device__ __host__ inline
int classify(const float2 &mp, const float2 &p)
{
    if(p.x <= mp.x)
    {
        if(p.y <= mp.y)
            return Quad::BOTTOM_LEFT;
        else
            return Quad::TOP_LEFT;
    }
    else
    {
        if(p.y <= mp.y)
            return Quad::BOTTOM_RIGHT;
        else
            return Quad::TOP_RIGHT;
    }
}

__device__ __host__ inline
int Quad::classify(const float2 &p) const
{
    return ::classify(mp, p);
}

struct QuadsPtr
{
    QuadsPtr() : size(0), mp(NULL)
               , min(NULL), max(NULL)
               , children(NULL), seg_range(NULL)
               , mem_config(NULL)
    {
    }

    unsigned size;
    float2 *mp, *min, *max;
    int4 *children;

    uint2 *seg_range;
    short2 *mem_config;

    __device__
    void get(unsigned idx, Quad &q) const
    {
        q.mp = mp[idx];
        q.min = min[idx];
        q.max = max[idx];
        q.children = children[idx];
    }

    __device__
    void put(unsigned idx, const Quad &q)
    {
        mp[idx] = q.mp;
        min[idx] = q.min;
        max[idx] = q.max;
        children[idx] = q.children;
    }
};

struct const_QuadsPtr
{
    const_QuadsPtr() 
        : size(0), mp(NULL)
        , min(NULL), max(NULL)
        , children(NULL), seg_range(NULL)
    {
    }

    const_QuadsPtr(const QuadsPtr &that)
        : size(that.size)
        , mp(that.mp)
        , min(that.min)
        , max(that.max)
        , children(that.children)
        , seg_range(that.seg_range)
        , mem_config(that.mem_config)
    {
    }

    unsigned size;
    const float2 *mp, *min, *max;
    const int4 *children;

    const uint2 *seg_range; // x = begin, y = count;
    const short2 *mem_config;

    __device__
    void get(unsigned idx, Quad &q) const
    {
        q.mp = mp[idx];
        q.min = min[idx];
        q.max = max[idx];
        q.children =  children[idx];
    }

    __device__
    int classify(int idx, const float2 &p) const
    {
        return ::classify(__ldg(mp+idx), p);
    }

    __device__
    bool contains(int idx, const float2 &p) const
    {
        return p.x >= min[idx].x && p.x < max[idx].x && 
               p.y >= min[idx].y && p.y < max[idx].y;
    }
};

struct Quads
{
    // idquad==0: empty cell, qtree begins at idquad==1

    gpu::vector<float2> mp, min, max;
    gpu::vector<int4> children;

    gpu::vector<uint2> seg_range;

    gpu::vector<short2> mem_config;


    size_t size() const { return mp.size(); }
    void clear()
    {
        mp.clear();
        min.clear();
        max.clear();
        children.clear();
        seg_range.clear();
        mem_config.clear();
    }
    void resize(size_t s)
    {
        mp.resize(s);
        min.resize(s);
        max.resize(s);
        children.resize(s);
        seg_range.resize(s);
        mem_config.resize(s);
    }

    QuadsPtr operator&()
    {
        QuadsPtr q;
        q.size = size();
        q.mp = mp.data();
        q.min = min.data();
        q.max = max.data();
        q.children = children.data();
        q.seg_range = seg_range.data();
        q.mem_config = mem_config.data();
        return q;
    }

    const_QuadsPtr operator&() const
    {
        const_QuadsPtr q;
        q.size = size();
        q.mp = mp.data();
        q.min = min.data();
        q.max = max.data();
        q.children = children.data();
        q.seg_range = seg_range.data();
        q.mem_config = mem_config.data();
        return q;
    }
};

#endif

