#include <gpu/timer.h>
#include "image_util.h"
#include "segments.h"
#include "paint.h"
#include "raster.h"
#include "implicit.h"
#include "qtree.h"
#include "traversal.h"
#include "util.h"

namespace img = image;

void calculate_image_xform(const fd::Box &input_bbox,/*{{{*/
                           unsigned output_width, unsigned output_height,
                           float2 &scale, float2 &offset)
{
    // input dimensions
    float W = input_bbox.max.x - input_bbox.min.x,
          H = input_bbox.max.y - input_bbox.min.y;

    // input width fits exactly in output width

    // output height was calculated to accomodate input height,
    // but there can be some gap (our pixels are square),
    // we'll center input vertically on output grid

    // this is the input height related to the actual integer output height
    float Hl = output_height*W/output_width;

    // center window on available space
    float dy = (Hl - H)*0.5f;

    scale.y = scale.x = W/output_width; // square pixels

    offset.x = scale.x*0.5f + input_bbox.min.x;
    offset.y = scale.y*0.5f + input_bbox.min.y - dy;

    // invert affine transform

    /* I = O*scale+offset
       O = (I-offset)/scale
       O = I/scale - offset/scale
    */

    scale.x = 1/scale.x;
    scale.y = 1/scale.y;

    offset.x = -offset.x*scale.x;
    offset.y = -offset.y*scale.y;
}/*}}}*/

void rasterize(img::rgba<> &h_out, const fd::FontDump &fd, unsigned nsamples,
               unsigned qtree_levels, unsigned min_segs_split,
               const img::Color &_bg_color)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("processing"));

    float2 scale, off;
    calculate_image_xform(fd.bbox, h_out.width(), h_out.height(),
                          scale, off);

    gpu::vector<Gradient> d_gradients;
    std::vector<TexRamp> texramps;
    convert_gradients(fd, texramps, d_gradients, scale, off);

    gpu::vector<const Paint *> d_attrib_paints;
    gpu::vector<unsigned char> d_paint_data;
    convert_paints(fd.attribs, d_gradients,
                   d_paint_data, d_attrib_paints);

    Segments d_segs;
    convert_segments(fd, d_attrib_paints, scale, off, d_segs);

    if(d_segs.size() != 0)
    {
        // create acceleration quadtree

        Quads d_qtree;

        {
            unsigned dim = std::max(next_power_of_2(h_out.width()), next_power_of_2(h_out.height()));

            gpu::Box bbox;
            bbox.min = make_float2(0,0);
            bbox.max = make_float2(dim, dim);

            create_qtree(d_qtree, d_segs, bbox, qtree_levels,
                         min_segs_split);
        }

        printf("# segments after quadtree split: %lu\n",d_segs.size());

        // rasterize

        gpu::image<float4> d_out(h_out.width(), h_out.height());

        float4 bg_color = make_float4(_bg_color.r, _bg_color.g, _bg_color.b, _bg_color.a);

        if(nsamples == 1)
            rasterize(bg_color, d_qtree, d_segs, &d_out);
        else
            rasterize(bg_color, d_qtree, d_segs, nsamples, &d_out);

        copy_to_host(h_out, d_out);
    }
    else
    {
        for(int i=0; i<h_out.width()*h_out.height(); ++i)
        {
            h_out.r[i] = _bg_color.r;
            h_out.g[i] = _bg_color.g;
            h_out.b[i] = _bg_color.b;
            h_out.a[i] = _bg_color.a;
        }
    }
}
