/* Algoritmo do scheduler:

   cada warp processa um pixel
   cada thread processa um sample

   (OK)
   a thread desce a qtree e ve em qual celula o sample cai
   o warp calcula o numero de celulas diferentes, se for menor que N,
   indexamos o buffer de saida pelo indice do pixel e salvamos o indice das
   celulas. se for maior que N, salvamos QUAD_OVERFLOW (indica pixels complicados)

   (OK)
   criamos uma lista com elementos idcelula,idpixel, agrupados por celula

   cada warp processa um elemento da lista. cada thread transforma o sample e
   testa se ele está dentro da célula. marcamos num array em smem os samples
   que estão dentro e compactamos esta lista, agrupando no inicio os samples
   que estão dentro. dividimos este array em warps e cada warp processa no
   máximo 32 samples, que podem vir de pixels diferentes. fazemos uma soma
   parcial (por pixel) da cor do sample e somamos o resultado no pixel, na
   saida.

   compactamos a lista, criando uma só com os pixels complicados.

   calculamos a cor dos pixels complicados do jeito tradicional.
*/

#include <vector>
#include <gpu/vector.h>
#include <gpu/image.h>
#include <gpu/symbol.h>
#include <gpu/cub.h>
#include <gpu/timer.h>
#include <gpu/recfilter.h>
#include <gpu/image_util.h>
#include <cstdio>
#include "persistent.h"
#include "segments.h"
#include "sampling.h"
#include "implicit.h"
#include "blue_noise.h"
#include "quad.h"
#include "util.h"

const int MAX_BLOCKS_PER_SM = 8;
const int MAX_QUADS_PER_PIXEL = 4;
const int // BW = 32, // never referenced
          BH = 8;
const int KS = 4; // pre-filter kernel support size
const int IMG_BORDER = 32;

const unsigned QUAD_OVERFLOW = (unsigned)-2,
               NO_QUADS = (unsigned)-1;


__constant__ float2 d_sampling_pattern[MAX_SAMPDIM];
__device__ float d_sampling_pattern_x[MAX_SAMPDIM],
                 d_sampling_pattern_y[MAX_SAMPDIM];
__constant__ float d_bspline3_data[MAX_SAMPDIM*KS*KS];
__device__ float c_bspline3_sum[KS*KS];
__device__ float4 c_empty_samples[KS*KS];
__constant__ unsigned d_num_samples;
__constant__ float4 c_bgcolor;

__constant__ Projective c_inv_xform;

namespace {

/*{{{ blue noise initialization*/
__host__ __device__
float bspline3(float r)/*{{{*/
{
    r = std::abs(r);

    if (r < 1.f)
        return (4.f + r*r*(-6.f + 3.f*r))/6.f;
    else if (r < 2.f)
        return  (8.f + r*(-12.f + (6.f - r)*r))/6.f;
    else
        return 0.f;
}/*}}}*/

template <int NSAMP>
std::vector<float2> get_blue_noise_pattern(float adj)/*{{{*/
{
    std::vector<float2> blue_noise;
    blue_noise.reserve(NSAMP);
    for(int i=0; i<NSAMP; ++i)
    {
        float2 n = make_float2(BlueNoise<NSAMP>::x[i], BlueNoise<NSAMP>::y[i])+adj;
        blue_noise.push_back(n);
    }

    return blue_noise;
}/*}}}*/

std::vector<float> calc_bspline3_data(const std::vector<float2> &pattern)/*{{{*/
{
    std::vector<float> bspline3_data;
    bspline3_data.reserve(pattern.size()*KS*KS);
    for(int i=0; i<pattern.size(); ++i)
    {
        for(int y=0; y<KS; ++y)
        {
            for(int x=0; x<KS; ++x)
            {
                bspline3_data.push_back(bspline3(x+pattern[i].x)*
                                        bspline3(y+pattern[i].y)/pattern.size());
            }
        }
    }

    return bspline3_data;
}/*}}}*/

std::vector<float2> get_blue_noise_pattern(unsigned nsamples, float adj)/*{{{*/
{
    switch(nsamples)
    {
    case 8:
        return get_blue_noise_pattern<8>(adj);
    case 16:
        return get_blue_noise_pattern<16>(adj);
    case 32:
        return get_blue_noise_pattern<32>(adj);
    case 64:
        return get_blue_noise_pattern<64>(adj);
    case 128:
        return get_blue_noise_pattern<128>(adj);
    case 256:
        return get_blue_noise_pattern<256>(adj);
    default:
        throw std::runtime_error("Invalid sample count");
    }
}/*}}}*/
/*}}}*/

__device__
int traverse_qtree(const_QuadsPtr qtree, float2 coord)/*{{{*/
{
    unsigned idquad = 0;
    while(qtree.seg_range[idquad].y == 0)
    {
        int c = qtree.classify(idquad, coord);

        int idchildquad;

        int4 children = qtree.children[idquad];

        switch(c)
        {
        case 0:
            idchildquad = children.x;
            break;
        case 1:
            idchildquad = children.y;
            break;
        case 2:
            idchildquad = children.z;
            break;
        case 3:
            idchildquad = children.w;
            break;
        }

        if(idchildquad >= 0)
            idquad = idchildquad;
        else
        {
            idquad = (unsigned)-1;
            break;
        }
    }

    if(idquad != (unsigned)-1 && !qtree.contains(idquad, coord))
        idquad = (unsigned)-1;

    return idquad;
}/*}}}*/

__global__
void gpu_traverse_qtree32(const_QuadsPtr qtree, unsigned *pixel_quad, /*{{{*/
                          unsigned *quads_per_pixel,
                          unsigned w, unsigned h)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    int2 pix = make_int2(blockIdx.x,
                         blockIdx.y*BH+ty);

    if(pix.x >= w || pix.y >= h)
        return;

    unsigned idsample = tx;

    float2 coord
        = make_float2(pix.x, pix.y);
    coord.x += d_sampling_pattern[idsample].x;
    coord.y += d_sampling_pattern[idsample].y;

    coord = apply(c_inv_xform, coord);

    int idquad = traverse_qtree(qtree, coord);

    // now we must compact pixel's idquads
    __shared__ unsigned idquads[BH][32];
    idquads[ty][tx] = idquad;

    unsigned *data = &idquads[ty][0];

    unsigned idxout = (pix.y*w + pix.x)*MAX_QUADS_PER_PIXEL;

    if(__all(data[tx] != (unsigned)-1))
    {
        // bitonic sort
        for(uint size=2; size<=32; size *= 2)
        {
            for(uint stride = size/2; stride > 0; stride >>=1)
            {
                int ixj = tx ^ stride;
                int ddd = (tx & size) == 0;
                if(ixj > tx && (data[tx] > data[ixj]) == ddd)
                    gpu::swap(data[tx], data[ixj]);
            }
        }

        // mark duplicates
        int idx = tx;
        while(data[idx] == data[tx])
        {
            if(--idx < 0)
                break;
        }

        // remove marked elements
        __shared__ unsigned len[BH];
        if(tx == 0)
            len[ty] = 0;

        if(idx+1 == tx)
        {
            unsigned tmp = data[tx];
            unsigned pos = atomicAdd(len+ty, 1);
            data[pos] = tmp;
        }

        // write out result

        if(len[ty] <= MAX_QUADS_PER_PIXEL)
        {
            if(tx < len[ty])
                pixel_quad[idxout + tx] = data[tx];

            if(tx == 0)
                quads_per_pixel[pix.y*w+pix.x] = max(1,len[ty]);
            return;
        }
    }

    if(tx == 0)
    {
        pixel_quad[idxout] = QUAD_OVERFLOW;
        quads_per_pixel[pix.y*w+pix.x] = 1; // to hold problematic pixel
    }
}/*}}}*/

__global__
void gpu_create_quad_pixel_array(unsigned *pixel_quad, unsigned *out_index,/*{{{*/
                                 unsigned size, uint2 *quad_pixel)
{
    int idpixel = blockIdx.x*blockDim.x + threadIdx.x;
    if(idpixel >= size)
        return;

    int len = out_index[idpixel+1] - out_index[idpixel];
    if(len == 0)
        return;

    for(int i=0; i<len; ++i)
        quad_pixel[out_index[idpixel]+i] = make_uint2(pixel_quad[idpixel*MAX_QUADS_PER_PIXEL+i], idpixel);
}/*}}}*/

}

class RasterizerScheduler/*{{{*/
{
public:
    RasterizerScheduler(
              const Quads &qtree,
              const uint2 *quad_pixel,
              const Segments &segs,
              unsigned width, unsigned height,
              unsigned border,
              gpu::image<float> out[4])
        : m_qtree(&qtree)
        , m_quad_pixel(quad_pixel)
        , m_segs(&segs)
        , m_width(width)
        , m_height(height)
        , m_rowstride(out[0].rowstride())
    {
        for(unsigned i=0; i<4; ++i)
            m_out[i] = out[i].data() + m_rowstride*border + border;
    }

    __device__
    void operator()(int idwarp);

private:
    const_QuadsPtr m_qtree;
    const uint2 *m_quad_pixel;
    const uint2 *m_seg_range;
    const_SegmentsPtr m_segs;
    unsigned m_width, m_height, m_rowstride;
    float *m_out[4];
};/*}}}*/
__device__
void RasterizerScheduler::operator()(int idwarp)/*{{{*/
{
    int tx = threadIdx.x;

    // do not process NO_QUADS and QUAD_OVERFLOW
    int idquad = m_quad_pixel[idwarp].x,
        idpixel = m_quad_pixel[idwarp].y;
    if(idquad < 0)
        return;

    // it's faster to load x and y separately than to load float2 at once
    float2 pattern = make_float2(d_sampling_pattern_x[tx],
                                 d_sampling_pattern_y[tx]);

    uint2 range = m_qtree.seg_range[idquad];

    int2 pix = make_int2(idpixel%m_width,
                         idpixel/m_width);

    float2 coord;
    coord.x = pix.x + pattern.x;
    coord.y = pix.y + pattern.y;

    coord = apply(c_inv_xform, coord);

    float4 sample = make_float4(0,0,0,0);

    if(m_qtree.contains(idquad, coord))
    {
        unsigned idseg = range.x,
                 idseg_end = idseg + range.y;

        sample = vgsample(m_segs, idseg, idseg_end, coord, c_bgcolor);
    }

    // reload to save some registers across vgsample call
    pattern = make_float2(d_sampling_pattern_x[tx],
                          d_sampling_pattern_y[tx]);

    sample = srgb2lrgb(postdivide(sample));

    float bs3x[4];
#pragma unroll
    for(unsigned j=0; j<4; ++j)
        bs3x[j] = bspline3(j+pattern.x);

    int y = (int)pix.y;
    int x = (int)pix.x - tx;
    int idx_out = y*m_rowstride+x;

    // it's faster not to unroll loops here
    for(int i=0; i<KS; ++i)
    {
        float bs3y = bspline3(i+pattern.y)/32.f;

        float bs;

#pragma unroll
        for(int c=0; c<4; ++c)
        {
            float sy = ((float *)&sample)[c] * bs3y;
#pragma unroll
            for(int j=0; j<KS; ++j)
            {
                float aux = scan32(sy*bs3x[j]);
                if(tx == j)
                    bs = aux;
            }

            if(tx < KS)
                atomicAdd(m_out[c]+idx_out, bs);
        }

        idx_out -= m_rowstride;
    }
}/*}}}*/

class RasterizerSimple/*{{{*/
{
public:
    RasterizerSimple(const Quads &qtree,
              const uint2 *quad_pixel,
              const Segments &segs,
              unsigned width, unsigned height,
              unsigned border,
              gpu::image<float> out[4])
        : m_qtree(&qtree)
        , m_quad_pixel(quad_pixel)
        , m_segs(&segs)
        , m_width(width)
        , m_height(height)
        , m_rowstride(out[0].rowstride())
    {
        for(unsigned i=0; i<4; ++i)
            m_out[i] = out[i].data() + m_rowstride*border + border;
    }

    __device__
    void operator()(int idwarp);

private:
    const_QuadsPtr m_qtree;
    const uint2 *m_quad_pixel;
    const_SegmentsPtr m_segs;
    unsigned m_width, m_height, m_rowstride;
    float *m_out[4];
};/*}}}*/
__device__
void RasterizerSimple::operator()(int idwarp)/*{{{*/
{
    int tx = threadIdx.x;

    // do not process NO_QUADS and QUAD_OVERFLOW
    int idquad = m_quad_pixel[idwarp].x,
        idpixel = m_quad_pixel[idwarp].y;
    if(idquad != QUAD_OVERFLOW)
        return;

    // it's faster to load x and y separately than to load float2 at once
    float2 pattern = make_float2(d_sampling_pattern_x[tx],
                                 d_sampling_pattern_y[tx]);

    int2 pix = make_int2(idpixel%m_width,
                         idpixel/m_width);

    float2 coord;
    coord.x = pix.x + pattern.x;
    coord.y = pix.y + pattern.y;

    coord = apply(c_inv_xform, coord);

    idquad = traverse_qtree(m_qtree, coord);

    float4 sample = c_bgcolor;

    if(idquad != -1)
    {
        uint2 range = m_qtree.seg_range[idquad];

        unsigned idseg = range.x,
                 idseg_end = idseg + range.y;

        sample = vgsample(m_segs, idseg, idseg_end, coord, c_bgcolor);
    }

    // reload to save some registers across vgsample call
    pattern = make_float2(d_sampling_pattern_x[tx],
                          d_sampling_pattern_y[tx]);

    sample = srgb2lrgb(postdivide(sample));

    float bs3x[4];
#pragma unroll
    for(unsigned j=0; j<4; ++j)
        bs3x[j] = bspline3(j+pattern.x);

    int y = (int)pix.y;
    int x = (int)pix.x - tx;
    int idx_out = y*m_rowstride+x;

    // it's faster not to unroll loops here
    for(int i=0; i<KS; ++i)
    {
        float bs3y = bspline3(i+pattern.y)/32.f;

        float bs;

#pragma unroll
        for(int c=0; c<4; ++c)
        {
            float sy = ((float *)&sample)[c] * bs3y;
#pragma unroll
            for(int j=0; j<KS; ++j)
            {
                float aux = scan32(sy*bs3x[j]);
                if(tx == j)
                    bs = aux;
            }

            if(tx < KS)
                atomicAdd(m_out[c]+idx_out, bs);
        }

        idx_out -= m_rowstride;
    }
}/*}}}*/

class RasterizerEmpty/*{{{*/
{
public:
    RasterizerEmpty(
              const uint2 *quad_pixel,
              const uint2 *seg_range,
              const Segments &segs,
              unsigned width, unsigned height,
              unsigned border,
              gpu::image<float> out[4])
        : m_quad_pixel(quad_pixel)
        , m_seg_range(seg_range)
        , m_segs(&segs)
        , m_width(width)
        , m_height(height)
        , m_rowstride(out[0].rowstride())
    {
        for(unsigned i=0; i<4; ++i)
            m_out[i] = out[i].data() + m_rowstride*border + border;
    }

    __device__
    void operator()(int idwarp);

private:
    const uint2 *m_quad_pixel;
    const uint2 *m_seg_range;
    const_SegmentsPtr m_segs;
    unsigned m_width, m_height, m_rowstride;
    float *m_out[4];
};/*}}}*/
__device__
void RasterizerEmpty::operator()(int idwarp)/*{{{*/
{
    int tx = threadIdx.x;

    // do not process NO_QUADS and QUAD_OVERFLOW
    int idquad = m_quad_pixel[idwarp].x,
        idpixel = m_quad_pixel[idwarp].y;
    if(idquad != NO_QUADS)
        return;

    if(tx < KS*KS)
    {
        float4 out = c_empty_samples[tx];

        int2 pix = make_int2(idpixel%m_width,
                             idpixel/m_width);

        int x = (int)pix.x - tx%KS,
            y = (int)pix.y - tx/KS;

        int idx = y*m_rowstride + x;
        atomicAdd(m_out[0]+idx, out.x);
        atomicAdd(m_out[1]+idx, out.y);
        atomicAdd(m_out[2]+idx, out.z);
        atomicAdd(m_out[3]+idx, out.w);
    }
}/*}}}*/

void rasterize(const float4 &bg_color,
               const Projective &xform,
               const Quads &qtree,
               const Segments &segs,
               unsigned nsamples,
               gpu::image_ptr<float4> out)
{
    std::cerr << "raster xform" << std::endl;
    int w = out.width(), h = out.height();
    int W = out.width()+3, H = out.height()+3;

    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("rasterize"));

#if 0
    float ang = M_PI/8;
    Transform xform2 = translate(w/2.f, h/2.f)
                        *rotate(ang)
                        *translate(w/-2.f,h/-2.f)
                      ;
#endif

    gpu::copy_to_symbol(c_inv_xform, inverse(xform));

    gpu::copy_to_symbol(c_bgcolor, bg_color);
    gpu::copy_to_symbol(d_num_samples, nsamples);

    std::vector<float2> pattern = get_blue_noise_pattern(nsamples,-1.5);
    gpu::copy_to_symbol(d_sampling_pattern, pattern);
    {
        std::vector<float> patx, paty;
        for(unsigned i=0; i<pattern.size(); ++i)
        {
            patx.push_back(pattern[i].x);
            paty.push_back(pattern[i].y);
        }
        gpu::copy_to_symbol(d_sampling_pattern_x, patx);
        gpu::copy_to_symbol(d_sampling_pattern_y, paty);
    }

    std::vector<float> bspline3 = calc_bspline3_data(pattern);
    gpu::copy_to_symbol(d_bspline3_data, bspline3);

    gpu::vector<unsigned> pixel_quad;
    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("init pixel_quad"));

        std::vector<unsigned> init(W*H*MAX_QUADS_PER_PIXEL, NO_QUADS);
        pixel_quad.copy_from_host(init);
    }

    gpu::vector<unsigned> quads_per_pixel(W*H+1);
    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("qtree traversal"));

        dim3 bdim(32, BH),
             gdim(W, (H+BH-1)/BH);
        gpu_traverse_qtree32<<<gdim,bdim>>>(&qtree, &pixel_quad, &quads_per_pixel, W, H);
    }

#if 0
    {
        std::vector<unsigned> aux, aux2;
        pixel_quad.copy_to_host(aux);
        quads_per_pixel.copy_to_host(aux2);
        for(int i=0; i<aux2.size(); ++i)
        {
            printf("(%d,%d) %2d: ",i%w,i/w,aux2[i]);
            for(int j=0; j<aux2[i]; ++j)
                printf("%d ",aux[i*MAX_QUADS_PER_PIXEL+j]);
            printf("\n");
        }
    }
#endif

    // last element will hold total number of quads
    cudaMemset(&quads_per_pixel+W*H, 0, sizeof(unsigned));

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("quads_per_pixel scan"));

        gpu::CUB cub;
        cub.exclusive_scan(quads_per_pixel, quads_per_pixel);
    }

    gpu::vector<uint2> quad_pixel(quads_per_pixel.back());
    quads_per_pixel.resize(quads_per_pixel.size()-1);

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("create quad_pixel"));

        dim3 bdim(256),
             gdim((quads_per_pixel.size()+bdim.x-1)/bdim.x);

        gpu_create_quad_pixel_array<<<gdim,bdim>>>(&pixel_quad, 
                                                   &quads_per_pixel,
                                                   quads_per_pixel.size(), &quad_pixel);
    }

    gpu::image<float> rast_channels[4];
    gpu::image<float> channels[4];

    {
        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("collect samples"));

            {
                std::vector<float> bspline3_sum(KS*KS);
                std::vector<float4> empty_samples(KS*KS);

#pragma unroll
                for(int i=0; i<KS*KS; ++i)
                {
                    bspline3_sum[i] = 0;
                    for(int s=0; s<nsamples; ++s)
                        bspline3_sum[i] += bspline3[s*KS*KS + i];

                    empty_samples[i] = srgb2lrgb(bg_color)*bspline3_sum[i];
                }

                gpu::copy_to_symbol(c_empty_samples, empty_samples);
                gpu::copy_to_symbol(c_bspline3_sum, bspline3_sum);
            }

            for(int i=0; i<4; ++i)
            {
                rast_channels[i].resize(w+IMG_BORDER*2,H+IMG_BORDER*2, ((W+IMG_BORDER*2 + 256-1)/256)*256 );
                set_zero(rast_channels[i]);
            }

            cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("Rasterizer normal"));
                RasterizerScheduler rast(qtree, &quad_pixel,
                                         segs,
                                         W, H,
                                         IMG_BORDER, rast_channels);
                persistent_threads<MAX_BLOCKS_PER_SM>(rast, quad_pixel.size());
            }
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("Rasterizer simple"));
                RasterizerSimple rast(qtree, &quad_pixel,
                                      segs,
                                      W, H,
                                      IMG_BORDER, rast_channels);
                persistent_threads<MAX_BLOCKS_PER_SM>(rast, quad_pixel.size());
            }
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("Rasterizer empty"));
                RasterizerEmpty rast(&quad_pixel, &qtree.seg_range,
                                     segs,
                                     W, H,
                                     IMG_BORDER, rast_channels);
                persistent_threads<MAX_BLOCKS_PER_SM>(rast, quad_pixel.size());
            }

            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
        }
    }

    for(int i=0; i<4; ++i) {
        subimage(channels[i], &rast_channels[i], 0, 0, w+IMG_BORDER*2,h+IMG_BORDER*2);
        std::cerr << channels[i].width() << " " << channels[i].height() << " " << w+IMG_BORDER*2 << " " << h+IMG_BORDER*2 << std::endl;
    }

#if 1
    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("do cardinal bspline3"));

        // calculate cubic b-spline weights
        gpu::math::Vector<float,1+1> bspline3_weights;
        {
            float a = 2.f-std::sqrt(3.0f);

            bspline3_weights[0] = 1+a;
            bspline3_weights[1] = a;
        }

        recursive_filter_5_setup(out.width(),
                                 out.height(),
                                 out.rowstride(),
                                 bspline3_weights, gpu::CLAMP_TO_EDGE, 1);

        for(int i=0; i<4; ++i)
            recursive_filter_5(channels[i], IMG_BORDER);

        gpu::recursive_filter_5_free();
    }

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("compose channels into output image"));
        compose(out, channels[0], channels[1], channels[2], channels[3]);
    }
#endif
}
