#include "raster.h"
#include <gpu/timer.h>
#include <gpu/symbol.h>
#include "quad.h"
#include "segments.h"
#include "packets.h"
#include "traversal.h"
#include "persistent.h"
#include "sampling.h"

const int MAX_BLOCKS_PER_SM = 8;

__constant__ float4 c_bgcolor;

class RasterizerBox/*{{{*/
{
public:
    RasterizerBox(const unsigned *packet_base, const unsigned *packet_size,
              const TraversalInfos &tinfos,
              const uint2 *seg_range,
              const Segments &segs,
              unsigned width, unsigned height,
              float4 *out)
        : m_packet_base(packet_base)
        , m_packet_size(packet_size)
        , m_tinfos(&tinfos)
        , m_seg_range(seg_range)
        , m_segs(&segs)
        , m_width(width)
        , m_height(height)
        , m_out(out)
    {
    }

    __device__
    void operator()(int idwarp) const;

private:
    const unsigned *m_packet_base, *m_packet_size;
    const_TraversalInfosPtr m_tinfos;
    const uint2 *m_seg_range;
    const_SegmentsPtr m_segs;
    unsigned m_width, m_height;
    float4 *m_out;
};/*}}}*/

__device__
void RasterizerBox::operator()(int idwarp) const/*{{{*/
{
    int packet_size = m_packet_size[idwarp];
    if(threadIdx.x >= packet_size)
        return;

    int idinfo = m_packet_base[idwarp]+threadIdx.x;

    TraversalInfo tinfo;
    m_tinfos.get(idinfo, tinfo);

    unsigned idseg, idseg_end;

    if(tinfo.idquad != (unsigned)-1)
    {
        uint2 range = m_seg_range[tinfo.idquad];
        idseg = range.x;
        idseg_end = idseg + range.y;
    }
    else
        idseg = idseg_end = 0;

    float2 coord = make_float2(tinfo.pix.x, tinfo.pix.y);

    float4 sample = vgsample(m_segs, idseg, idseg_end, coord, c_bgcolor);

    sample = srgb2lrgb(postdivide(sample));

    m_out[tinfo.pix.y*m_width + tinfo.pix.x] = sample;
}/*}}}*/

void rasterize(const float4 &bg_color,
               const Quads &qtree,
               const Segments &segs,
               gpu::image_ptr<float4> out)
{
    std::cerr << "raster box!!!" << std::endl;
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("rasterize"));

    gpu::copy_to_symbol(c_bgcolor, bg_color);

    TraversalInfos tinfo;
    traverse_qtree(tinfo, qtree, out.width(), out.height(), 0);

    sort(tinfo);

    gpu::vector<unsigned> packet_base, packet_size;
    create_packets(packet_base, packet_size, tinfo);

    cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);

    RasterizerBox rast(&packet_base, &packet_size, tinfo,
                       &qtree.seg_range,
                       segs,
                       out.width(), out.height(), &out);

    persistent_threads<MAX_BLOCKS_PER_SM>(rast, packet_base.size());

    cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
}
