#include "raster.h"
#include "sampling.h"
#include "packets.h"
#include "segments.h"
#include "blue_noise.h"
#include "traversal.h"
#include <gpu/timer.h>
#include <gpu/symbol.h>
#include <gpu/recfilter.h>
#include <gpu/image_util.h>
#include "quad.h"
#include "persistent.h"

const int MAX_BLOCKS_PER_SM = 8;

const int KS = 4; // pre-filter kernel support size

// const int IMG_BORDER = 32; // never referenced

#define NB_MULTISAMPLE 8

__constant__ float2 d_sampling_pattern[MAX_SAMPDIM];
__device__ float d_sampling_pattern_x[MAX_SAMPDIM],
                 d_sampling_pattern_y[MAX_SAMPDIM];
__constant__ float d_bspline3_data[MAX_SAMPDIM*KS*KS];
__device__ float c_bspline3_sum[KS*KS];
__device__ float4 c_empty_samples[KS*KS];
__constant__ unsigned d_num_samples;
__constant__ float4 c_bgcolor;


namespace {
/*{{{ blue noise initialization*/
__host__ __device__
inline
float bspline3(float r)/*{{{*/
{
    r = std::abs(r);

    if (r < 1.f)
        return (4.f + r*r*(-6.f + 3.f*r))/6.f;
    else if (r < 2.f)
        return  (8.f + r*(-12.f + (6.f - r)*r))/6.f;
    else
        return 0.f;
}/*}}}*/

template <int NSAMP>
std::vector<float2> get_blue_noise_pattern(float adj)/*{{{*/
{
    std::vector<float2> blue_noise;
    blue_noise.reserve(NSAMP);
    for(int i=0; i<NSAMP; ++i)
    {
        float2 n = make_float2(BlueNoise<NSAMP>::x[i], BlueNoise<NSAMP>::y[i])+adj;
        blue_noise.push_back(n);
    }

    return blue_noise;
}/*}}}*/

std::vector<float> calc_bspline3_data(const std::vector<float2> &pattern)/*{{{*/
{
	(void) calc_bspline3_data; // avoid warnings
    std::vector<float> bspline3_data;
    bspline3_data.reserve(pattern.size()*KS*KS);
    for(int i=0; i<pattern.size(); ++i)
    {
        for(int y=0; y<KS; ++y)
        {
            for(int x=0; x<KS; ++x)
            {
                bspline3_data.push_back(bspline3(x+pattern[i].x)*
                                        bspline3(y+pattern[i].y)/pattern.size());
            }
        }
    }

    return bspline3_data;
}/*}}}*/

std::vector<float2> get_blue_noise_pattern(unsigned nsamples, float adj)/*{{{*/
{
    switch(nsamples)
    {
    case 8:
        return get_blue_noise_pattern<8>(adj);
    case 16:
        return get_blue_noise_pattern<16>(adj);
    case 32:
        return get_blue_noise_pattern<32>(adj);
    case 64:
        return get_blue_noise_pattern<64>(adj);
    case 128:
        return get_blue_noise_pattern<128>(adj);
    case 256:
        return get_blue_noise_pattern<256>(adj);
    default:
        throw std::runtime_error("Invalid sample count");
    }
}/*}}}*/
/*}}}*/
class RasterizerMulti32/*{{{*/
{
public:
    RasterizerMulti32(
              const TraversalInfos &tinfo,
              const uint2 *seg_range,
              const Segments &segs,
              unsigned width, unsigned height,
              unsigned border,
              gpu::image<float> out[4])
        : m_tinfos(&tinfo)
        , m_seg_range(seg_range)
        , m_segs(&segs)
        , m_width(width)
        , m_height(height)
        , m_rowstride(out[0].rowstride())
    {
        for(unsigned i=0; i<4; ++i)
            m_out[i] = out[i].data() + m_rowstride*border + border;
    }

    __device__
    void operator()(int idwarp);

private:
    const_TraversalInfosPtr m_tinfos;
    const uint2 *m_seg_range;
    const_SegmentsPtr m_segs;
    unsigned m_width, m_height, m_rowstride;
    float *m_out[4];
};/*}}}*/
__device__
void RasterizerMulti32::operator()(int idwarp)/*{{{*/
{
    int tx = threadIdx.x;

    TraversalInfo tinfo;
    m_tinfos.get(idwarp, tinfo);

    if(tinfo.idquad >= 0)
    {
        // it's faster to load x and y separately than to load float2 at once
        float2 pattern = make_float2(d_sampling_pattern_x[tx],
                                     d_sampling_pattern_y[tx]);

        uint2 range = m_seg_range[tinfo.idquad];

        float2 coord;
        coord.x = tinfo.pix.x + pattern.x;
        coord.y = tinfo.pix.y + pattern.y;

        unsigned idseg = range.x,
                 idseg_end = idseg + range.y;

        float4 sample = vgsample(m_segs, idseg, idseg_end, coord, c_bgcolor);

        // reload to save some registers across vgsample call
        pattern = make_float2(d_sampling_pattern_x[tx],
                              d_sampling_pattern_y[tx]);

        sample = srgb2lrgb(postdivide(sample));

        int y = (int)tinfo.pix.y;
        int x = (int)tinfo.pix.x - tx;
        int idx_out = y*m_rowstride+x;

#pragma unroll
        for(int c=0; c<4; ++c)
        {
            float aux = scan32(((const float *)&sample)[c]);

            if(tx==0)
                m_out[c][idx_out] = aux/32.0f;
        }

        idx_out -= m_rowstride;
    }
    else if(tx == 0)
    {
        int y = (int)tinfo.pix.y;
        int x = (int)tinfo.pix.x - tx;
        int idx = y*m_rowstride + x;
        for(int c=0; c<4; ++c)
            m_out[c][idx] = ((const float *)&c_bgcolor)[c];
    }
}/*}}}*/
}

void rasterize_box32(const float4 &bg_color,
               const Quads &qtree,
               const Segments &segs,
               unsigned nsamples,
               gpu::image_ptr<float4> out)
{
    std::cerr << "raster box32" << std::endl;
    int width = out.width(), height = out.height();

    gpu::copy_to_symbol(c_bgcolor, bg_color);
    gpu::copy_to_symbol(d_num_samples, nsamples);

    std::vector<float2> pattern = get_blue_noise_pattern(nsamples,.5);
    gpu::copy_to_symbol(d_sampling_pattern, pattern);
    {
        std::vector<float> patx, paty;
        for(unsigned i=0; i<pattern.size(); ++i)
        {
            patx.push_back(pattern[i].x);
            paty.push_back(pattern[i].y);
        }
        gpu::copy_to_symbol(d_sampling_pattern_x, patx);
        gpu::copy_to_symbol(d_sampling_pattern_y, paty);
    }

    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("rasterize"));

    gpu::image<float> channels[4];

    {
        TraversalInfos tinfo;
        traverse_qtree(tinfo, qtree, width, height, .5);

        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("collect samples"));

            for(int i=0; i<4; ++i)
            {
                channels[i].resize(width,height);
                set_zero(channels[i]);
            }

            cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);

            RasterizerMulti32 rast(tinfo, &qtree.seg_range,
                                   segs,
                                   width, height,
                                   0, channels);

            persistent_threads<MAX_BLOCKS_PER_SM>(rast, tinfo.size());

            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
        }

    }

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("compose channels into output image"));

        compose(out, channels[0], channels[1], channels[2], channels[3]);
    }
}
