#include "raster.h"
#include "sampling.h"
#include "packets.h"
#include "segments.h"
#include "blue_noise.h"
#include "traversal.h"
#include <gpu/timer.h>
#include <gpu/symbol.h>
#include <gpu/recfilter.h>
#include <gpu/image_util.h>
#include "quad.h"
#include "persistent.h"

const int MAX_BLOCKS_PER_SM = 8;

const int KS = 4; // pre-filter kernel support size

const int IMG_BORDER = 32;

#define NB_MULTISAMPLE 8

__constant__ float2 d_sampling_pattern[MAX_SAMPDIM];
__device__ float d_sampling_pattern_x[MAX_SAMPDIM],
                 d_sampling_pattern_y[MAX_SAMPDIM];
__constant__ float d_bspline3_data[MAX_SAMPDIM*KS*KS];
__device__ float c_bspline3_sum[KS*KS];
__device__ float4 c_empty_samples[KS*KS];
__constant__ unsigned d_num_samples;
__constant__ float4 c_bgcolor;


/*{{{ blue noise initialization*/
__host__ __device__
float bspline3(float r)/*{{{*/
{
    r = std::abs(r);

    if (r < 1.f)
        return (4.f + r*r*(-6.f + 3.f*r))/6.f;
    else if (r < 2.f)
        return  (8.f + r*(-12.f + (6.f - r)*r))/6.f;
    else
        return 0.f;
}/*}}}*/

template <int NSAMP>
std::vector<float2> get_blue_noise_pattern(float adj)/*{{{*/
{
    std::vector<float2> blue_noise;
    blue_noise.reserve(NSAMP);
    for(int i=0; i<NSAMP; ++i)
    {
        float2 n = make_float2(BlueNoise<NSAMP>::x[i], BlueNoise<NSAMP>::y[i])+adj;
        blue_noise.push_back(n);
    }

    return blue_noise;
}/*}}}*/

std::vector<float> calc_bspline3_data(const std::vector<float2> &pattern)/*{{{*/
{
    std::vector<float> bspline3_data;
    bspline3_data.reserve(pattern.size()*KS*KS);
    for(int i=0; i<pattern.size(); ++i)
    {
        for(int y=0; y<KS; ++y)
        {
            for(int x=0; x<KS; ++x)
            {
                bspline3_data.push_back(bspline3(x+pattern[i].x)*
                                        bspline3(y+pattern[i].y)/pattern.size());
            }
        }
    }

    return bspline3_data;
}/*}}}*/

std::vector<float2> get_blue_noise_pattern(unsigned nsamples, float adj)/*{{{*/
{
    switch(nsamples)
    {
    case 8:
        return get_blue_noise_pattern<8>(adj);
    case 16:
        return get_blue_noise_pattern<16>(adj);
    case 32:
        return get_blue_noise_pattern<32>(adj);
    case 64:
        return get_blue_noise_pattern<64>(adj);
    case 128:
        return get_blue_noise_pattern<128>(adj);
    case 256:
        return get_blue_noise_pattern<256>(adj);
    default:
        throw std::runtime_error("Invalid sample count");
    }
}/*}}}*/
/*}}}*/

class RasterizerMulti1/*{{{*/
{
public:
    RasterizerMulti1(const unsigned *packet_base, const unsigned *packet_size,
              const TraversalInfos &tinfos,
              const uint2 *seg_range,
              const Segments &segs,
              unsigned width, unsigned height, unsigned border,
              gpu::image<float> out[4])
        : m_packet_base(packet_base)
        , m_packet_size(packet_size)
        , m_tinfos(&tinfos)
        , m_seg_range(seg_range)
        , m_segs(&segs)
        , m_width(width)
        , m_height(height)
        , m_rowstride(out[0].rowstride())
    {
        for(unsigned i=0; i<4; ++i)
            m_out[i] = out[i].data() + m_rowstride*border + border;
    }

    __device__
    void operator()(int idinfo) const;

private:
    const unsigned *m_packet_base, *m_packet_size;
    const_TraversalInfosPtr m_tinfos;
    const uint2 *m_seg_range;
    const_SegmentsPtr m_segs;
    unsigned m_width, m_height, m_rowstride;
    float *m_out[4];
};/*}}}*/
__device__
void RasterizerMulti1::operator()(int idwarp) const/*{{{*/
{
    if(threadIdx.x >= m_packet_size[idwarp])
        return;

    TraversalInfo tinfo;
    m_tinfos.get(m_packet_base[idwarp]+threadIdx.x, tinfo);

    float4 sum[KS*KS];

    if(tinfo.idquad >= 0)
    {
        unsigned idseg = m_seg_range[tinfo.idquad].x,
                 idseg_end = idseg + m_seg_range[tinfo.idquad].y;

        float2 coord_base = make_float2(tinfo.pix.x, tinfo.pix.y);

#pragma unroll
        for(unsigned i=0; i<KS*KS; ++i)
            sum[i] = make_float4(0,0,0,0);

        for(int s=0; s<d_num_samples; ++s)
        {
            float2 coord = coord_base + d_sampling_pattern[s];

            float4 sample = vgsample(m_segs, idseg, idseg_end,
                                     coord, c_bgcolor);

            sample = srgb2lrgb(postdivide(sample));

#pragma unroll
            for(int i=0; i<KS*KS; ++i)
                sum[i] += sample*d_bspline3_data[s*KS*KS+i];
        }
    }
    else
    {
        for(unsigned i=0; i<KS*KS; ++i)
            sum[i] = c_empty_samples[i];
    }

#pragma unroll
    for(unsigned i=0; i<KS*KS; ++i)
    {
        int x = (int)tinfo.pix.x - i%KS,
            y = (int)tinfo.pix.y - i/KS;

        int idx = y*m_rowstride+x;
        atomicAdd(m_out[0]+idx, sum[i].x);
        atomicAdd(m_out[1]+idx, sum[i].y);
        atomicAdd(m_out[2]+idx, sum[i].z);
        atomicAdd(m_out[3]+idx, sum[i].w);
    }
}/*}}}*/

class RasterizerMulti32/*{{{*/
{
public:
    RasterizerMulti32(
              const TraversalInfos &tinfo,
              const uint2 *seg_range,
              const Segments &segs,
              unsigned width, unsigned height,
              unsigned border,
              gpu::image<float> out[4])
        : m_tinfos(&tinfo)
        , m_seg_range(seg_range)
        , m_segs(&segs)
        , m_width(width)
        , m_height(height)
        , m_rowstride(out[0].rowstride())
    {
        for(unsigned i=0; i<4; ++i)
            m_out[i] = out[i].data() + m_rowstride*border + border;
    }

    __device__
    void operator()(int idwarp);

private:
    const_TraversalInfosPtr m_tinfos;
    const uint2 *m_seg_range;
    const_SegmentsPtr m_segs;
    unsigned m_width, m_height, m_rowstride;
    float *m_out[4];
};/*}}}*/
__device__
void RasterizerMulti32::operator()(int idwarp)/*{{{*/
{
    int tx = threadIdx.x;

    TraversalInfo tinfo;
    m_tinfos.get(idwarp, tinfo);

    if(tinfo.idquad >= 0)
    {
        // it's faster to load x and y separately than to load float2 at once
        float2 pattern = make_float2(d_sampling_pattern_x[tx],
                                     d_sampling_pattern_y[tx]);

        uint2 range = m_seg_range[tinfo.idquad];

        float2 coord;
        coord.x = tinfo.pix.x + pattern.x;
        coord.y = tinfo.pix.y + pattern.y;

        unsigned idseg = range.x,
                 idseg_end = idseg + range.y;

        float4 sample = vgsample(m_segs, idseg, idseg_end, coord, c_bgcolor);

        // reload to save some registers across vgsample call
        pattern = make_float2(d_sampling_pattern_x[tx],
                              d_sampling_pattern_y[tx]);

        sample = srgb2lrgb(postdivide(sample));

        float bs3x[4];
#pragma unroll
        for(unsigned j=0; j<4; ++j)
            bs3x[j] = bspline3(j+pattern.x);

        int y = (int)tinfo.pix.y;
        int x = (int)tinfo.pix.x - tx;
        int idx_out = y*m_rowstride+x;

        // it's faster not to unroll loops here
        for(int i=0; i<KS; ++i)
        {
            float bs3y = bspline3(i+pattern.y)/32.f;

            float bs;

#pragma unroll
            for(int c=0; c<4; ++c)
            {
                float sy = ((float *)&sample)[c] * bs3y;
#pragma unroll
                for(int j=0; j<KS; ++j)
                {
                    float aux = scan32(sy*bs3x[j]);
                    if(tx == j)
                        bs = aux;
                }

                if(tx < KS)
                    atomicAdd(m_out[c]+idx_out, bs);
            }

            idx_out -= m_rowstride;
        }
    }
    else if(tx < KS*KS)
    {
        float4 out = c_empty_samples[tx];

        int x = (int)tinfo.pix.x - tx%KS,
            y = (int)tinfo.pix.y - tx/KS;

        int idx = y*m_rowstride + x;
        atomicAdd(m_out[0]+idx, out.x);
        atomicAdd(m_out[1]+idx, out.y);
        atomicAdd(m_out[2]+idx, out.z);
        atomicAdd(m_out[3]+idx, out.w);
    }
}/*}}}*/

void rasterize(const float4 &bg_color,
               const Quads &qtree,
               const Segments &segs,
               unsigned nsamples,
               gpu::image_ptr<float4> out)
{
    std::cerr << "raster multi" << std::endl;
    int width = out.width(), height = out.height();

    gpu::copy_to_symbol(c_bgcolor, bg_color);
    gpu::copy_to_symbol(d_num_samples, nsamples);

    std::vector<float2> pattern = get_blue_noise_pattern(nsamples,-1.5);
    gpu::copy_to_symbol(d_sampling_pattern, pattern);
    {
        std::vector<float> patx, paty;
        for(unsigned i=0; i<pattern.size(); ++i)
        {
            patx.push_back(pattern[i].x);
            paty.push_back(pattern[i].y);
        }
        gpu::copy_to_symbol(d_sampling_pattern_x, patx);
        gpu::copy_to_symbol(d_sampling_pattern_y, paty);
    }

    std::vector<float> bspline3 = calc_bspline3_data(pattern);
    gpu::copy_to_symbol(d_bspline3_data, bspline3);

    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("rasterize"));

    gpu::image<float> channels[4];

    {
        TraversalInfos tinfo;
        traverse_qtree(tinfo, qtree, width+3, height+3, -1.5);

        gpu::vector<unsigned> packet_base, packet_size;

        if(nsamples != 32)
        {
            sort(tinfo);
            create_packets(packet_base, packet_size, tinfo);
        }

        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("collect samples"));

            {
                std::vector<float> bspline3_sum(KS*KS);
                std::vector<float4> empty_samples(KS*KS);

#pragma unroll
                for(int i=0; i<KS*KS; ++i)
                {
                    bspline3_sum[i] = 0;
                    for(int s=0; s<nsamples; ++s)
                        bspline3_sum[i] += bspline3[s*KS*KS + i];

                    empty_samples[i] = srgb2lrgb(bg_color)*bspline3_sum[i];
                }

                gpu::copy_to_symbol(c_empty_samples, empty_samples);
                gpu::copy_to_symbol(c_bspline3_sum, bspline3_sum);
            }

            for(int i=0; i<4; ++i)
            {
                channels[i].resize(width+IMG_BORDER*2,height+IMG_BORDER*2);
                set_zero(channels[i]);
            }

            cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);

            if(nsamples == 32)
            {
                RasterizerMulti32 rast(tinfo, &qtree.seg_range,
                                       segs,
                                       width, height,
                                       IMG_BORDER, channels);

                persistent_threads<MAX_BLOCKS_PER_SM>(rast, tinfo.size());
            }
            else
            {
                RasterizerMulti1 rast(&packet_base, &packet_size,
                                      tinfo, &qtree.seg_range,
                                      segs,
                                      width, height, IMG_BORDER, channels);

                persistent_threads<MAX_BLOCKS_PER_SM>(rast, packet_base.size());
            }

            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
        }

    }

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("do cardinal bspline3"));

        // calculate cubic b-spline weights
        gpu::math::Vector<float,1+1> bspline3_weights;
        {
            float a = 2.f-std::sqrt(3.0f);

            bspline3_weights[0] = 1+a;
            bspline3_weights[1] = a;
        }

        recursive_filter_5_setup(out.width(),
                                 out.height(),
                                 out.rowstride(),
                                 bspline3_weights, gpu::CLAMP_TO_EDGE, 1);

        for(int i=0; i<4; ++i)
            recursive_filter_5(channels[i], IMG_BORDER);

        gpu::recursive_filter_5_free();
    }

    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("compose channels into output image"));

        compose(out, channels[0], channels[1], channels[2], channels[3]);
    }
}
