#include "raster.h"
#include "sampling.h"
#include "packets.h"
#include "segments.h"
#include "blue_noise.h"
#include "traversal.h"
#include <gpu/timer.h>
#include <gpu/symbol.h>
#include <gpu/recfilter.h>
#include <gpu/image_util.h>
#include <gpu/cub.h>
#include "quad.h"
#include "persistent.h"
#include "segment_util.h"

#include <thrust/sequence.h>
#include <thrust/sort.h>
#include <assert.h>

// const int MAX_BLOCKS_PER_SM = 8; // never referenced

const int KS = 4; // pre-filter kernel support size

const int IMG_BORDER = 32;

#define NB_MULTISAMPLE 8

__constant__ float2 d_sampling_pattern[MAX_SAMPDIM];
__device__ float d_sampling_pattern_x[MAX_SAMPDIM],
                 d_sampling_pattern_y[MAX_SAMPDIM];
__constant__ float d_bspline3_data[MAX_SAMPDIM*KS*KS];
__device__ float c_bspline3_sum[KS*KS];
__device__ float4 c_empty_samples[KS*KS];
__constant__ unsigned d_num_samples;
__constant__ float4 c_bgcolor;


namespace {

struct QuadInfo {
    int num_segments;
    int data_size;
    int data_area;
};


/* blue noise initialization*/
__host__ __device__
inline
float bspline3(float r)
{/*{{{*/
	(void) bspline3; // avoid "never used" warning
    r = std::abs(r);

    if (r < 1.f)
        return (4.f + r*r*(-6.f + 3.f*r))/6.f;
    else if (r < 2.f)
        return  (8.f + r*(-12.f + (6.f - r)*r))/6.f;
    else
        return 0.f;
}/*}}}*/


struct Context {
    Context(
                  const TraversalInfos &tinfo,
                  const uint2 *seg_range,
                  const Segments &segs,
                  const SegmentIndex *si,
                  unsigned width, unsigned height,
                  unsigned border,
                  gpu::image<float> out[4])
            : m_tinfos(&tinfo)
            , m_seg_range(seg_range)
            , m_segs(&segs)
            , m_si(si)
            , m_width(width)
            , m_height(height)
            , m_rowstride(out[0].rowstride())
    {
        for(unsigned i=0; i<4; ++i)
            m_out[i] = out[i].data() + m_rowstride*border + border;
    }

    const_TraversalInfosPtr m_tinfos;
    const uint2 *m_seg_range;
    const_SegmentsPtr m_segs;
    const SegmentIndex* m_si;
    unsigned m_width, m_height, m_rowstride;
    float *m_out[4];
};

struct Bs3 {
    const float4 * m_empty_samples;

    __device__ __host__
    Bs3( )
       : m_empty_samples( c_empty_samples ) 
    {
    }

    __device__ __host__
    float k0( float u ) { return (1.f/6.f)*(((u)*u)*u); }
    __device__ __host__
    float k1( float u ) { return (1.f/6.f)*(((-3.f*u+3.f)*u+3.f)*u+1.f); }
    __device__ __host__
    float k2( float u ) { return (1.f/6.f)*(((3.f*u-6.f)*u)*u+4.f); }
    __device__ __host__
    float k3( float u ) { return (1.f/6.f)*(((3.f-u)*u-3.f)*u+1.f); }

    __device__
    void write( const Context& ctx, const float4& sample, int ti, const float2& pattern, const ushort2& p )
    {/*{{{*/
        // find pixel color and save
        #pragma unroll
        for( int c = 0; c < 4; ++c ) {
            float s = ((const float *)&sample)[c];
            float vj0 = k3(pattern.x);
            float vj1 = k2(pattern.x);
            float vj2 = k1(pattern.x);
            float vj3 = k0(pattern.x);
            //
            float out = 0;
            // 0 -> 0,0
            float vi = scan32( s * vj0 * k3(pattern.y) );
            //*
            if ( ti == 0 ) out = vi;
            // 1 -> 1,0
            vi = scan32( s * vj0 * k2(pattern.y) );
            if ( ti == 1 ) out = vi;
            // 2 -> 2,0
            vi = scan32( s * vj0 * k1(pattern.y) );
            if ( ti == 2 ) out = vi;
            // 3 -> 3,0
            vi = scan32( s * vj0 * k0(pattern.y) );
            if ( ti == 3 ) out = vi;
            //
            // 4 -> 0,1
            vi = scan32( s * vj1 * k3(pattern.y) );
            if ( ti == 4 ) out = vi;
            // 5 -> 1,1
            vi = scan32( s * vj1 * k2(pattern.y) );
            if ( ti == 5 ) out = vi;
            // 6 -> 2,1
            vi = scan32( s * vj1 * k1(pattern.y) );
            if ( ti == 6 ) out = vi;
            // 7 -> 3,1
            vi = scan32( s * vj1 * k0(pattern.y) );
            if ( ti == 7 ) out = vi;
            //
            // 8 -> 0,2
            vi = scan32( s * vj2 * k3(pattern.y) );
            if ( ti == 8 ) out = vi;
            // 9 -> 1,2
            vi = scan32( s * vj2 * k2(pattern.y) );
            if ( ti == 9 ) out = vi;
            // 10 -> 2,2
            vi = scan32( s * vj2 * k1(pattern.y) );
            if ( ti == 10 ) out = vi;
            // 11 -> 3,2
            vi = scan32( s * vj2 * k0(pattern.y) );
            if ( ti == 11 ) out = vi;
            //
            // 12 -> 0,3
            vi = scan32( s * vj3 * k3(pattern.y) );
            if ( ti == 12 ) out = vi;
            // 13 -> 1,3
            vi = scan32( s * vj3 * k2(pattern.y) );
            if ( ti == 13 ) out = vi;
            // 14 -> 2,3
            vi = scan32( s * vj3 * k1(pattern.y) );
            if ( ti == 14 ) out = vi;
            // 15 -> 3,3
            vi = scan32( s * vj3 * k0(pattern.y) );
            if ( ti == 15 ) out = vi;
            // */
            //*
            if ( ti < 16 ) {
                int di = ti%4 - 1;
                int dj = ti/4 - 1;
                int idx_out = (p.y + di)*ctx.m_rowstride+(p.x+dj);
                atomicAdd( ctx.m_out[c] + idx_out, out/32.f );
            }
            // */
        }
    }/*}}}*/

    __device__
    void write_bg( const Context& ctx, int ti, const ushort2& p )
    {/*{{{*/
        for ( int i=0; i<4; ++i ) {
            for ( int j=0; j<4; ++j ) {
                int di = j - 1;
                int dj = i - 1;
                int idx_out = (p.y + dj)*ctx.m_rowstride+(p.x+di);
                float4 out = m_empty_samples[4*i +j];
                atomicAdd( ctx.m_out[0] + idx_out, out.x );
                atomicAdd( ctx.m_out[1] + idx_out, out.y );
                atomicAdd( ctx.m_out[2] + idx_out, out.z );
                atomicAdd( ctx.m_out[3] + idx_out, out.w );
            }
        }
    }/*}}}*/

};

struct Box {
    const float4& m_bgcolor;

    __device__ __host__
    Box() : m_bgcolor( c_bgcolor )
    { }

    __device__
    void write( const Context& ctx, const float4& sample, int ti, const float2& pattern, const ushort2& p )
    {
        int idx_out = p.y*ctx.m_rowstride+p.x;
        // find pixel color and save
        #pragma unroll
        for( int c = 0; c < 4; ++c ) {
            float aux = scan32(((const float *)&sample)[c]);
            if(ti == 0) {
                ctx.m_out[c][idx_out] = aux/32.0f;
            }
        }
    }

    __device__
    void write_bg( const Context& ctx, int ti, const ushort2& p )
    {
        int idx = p.y*ctx.m_rowstride + p.x;
        #pragma unroll
        for( int c = 0; c < 4; ++c ) {
            ctx.m_out[c][idx] = ((const float *)&m_bgcolor)[c];
        }
    }

};

template<class T>
__device__
inline T segscan_32( T val, int mindex, const unsigned id )
{
    T aux = __shfl_up( val, 1 );
    if ( id >= mindex +  1) val = aux + val;
    aux = __shfl_up( val, 2 );
    if ( id >= mindex +  2) val = aux + val;
    aux = __shfl_up( val, 4 );
    if ( id >= mindex +  4) val = aux + val;
    aux = __shfl_up( val, 8 );
    if ( id >= mindex +  8) val = aux + val;
    aux = __shfl_up( val, 16 );
    if ( id >= mindex +  16) val = aux + val;
    return val;
}

template<class T>
__device__
T scan_max_32( T v, const int ti )
{
    T tmp1 = v;
    for ( unsigned i=1; i<32; i<<=1) {
        T tmp2 = __shfl_up( tmp1, i );
        if ( ti >= i ) tmp1 = max(tmp1,tmp2);
    }
    return tmp1;
}

__device__
inline float4 sample_seg_scan_reg( const float4& sample, int head, int id )
{
    int hf = head;
    if ( hf ) hf = id;
    hf = scan_max_32( hf, id );
    float4 out;
    out.x = segscan_32( sample.x, hf, id );
    out.y = segscan_32( sample.y, hf, id );
    out.z = segscan_32( sample.z, hf, id );
    out.w = segscan_32( sample.w, hf, id );
    return out;
}

template<int nsamples>
struct BoxNSamples {
    const float4& m_bgcolor;

    __device__ __host__
    BoxNSamples() : m_bgcolor( c_bgcolor )
    { }

    __device__
    void write( const Context& ctx, const float4& sample, int ti, const float2& pattern, const ushort2& p )
    {
        int head = ti%nsamples == 0;
        int tail = (ti+1)%nsamples == 0;
        float4 s = sample_seg_scan_reg( sample, head, ti );
        int idx_out = p.y*ctx.m_rowstride+p.x;
        // find pixel color and save
        if(tail) {
            ctx.m_out[0][idx_out] = s.x/float(nsamples);
            ctx.m_out[1][idx_out] = s.y/float(nsamples);
            ctx.m_out[2][idx_out] = s.z/float(nsamples);
            ctx.m_out[3][idx_out] = s.w/float(nsamples);
        }
    }

    __device__
    void write_bg( const Context& ctx, int ti, const ushort2& p )
    {
        int idx = p.y*ctx.m_rowstride + p.x;
        #pragma unroll
        for( int c = 0; c < 4; ++c ) {
            ctx.m_out[c][idx] = ((const float *)&m_bgcolor)[c];
        }
    }

};

template<int nsamples>
struct Bs3NSamples {
    const float4 * m_empty_samples;

    __device__ __host__
    Bs3NSamples( )
       : m_empty_samples( c_empty_samples )
    {
    }

    __device__ __host__
    float k0( float u ) { return (1.f/6.f)*(((u)*u)*u); }
    __device__ __host__
    float k1( float u ) { return (1.f/6.f)*(((-3.f*u+3.f)*u+3.f)*u+1.f); }
    __device__ __host__
    float k2( float u ) { return (1.f/6.f)*(((3.f*u-6.f)*u)*u+4.f); }
    __device__ __host__
    float k3( float u ) { return (1.f/6.f)*(((3.f-u)*u-3.f)*u+1.f); }

    __device__
    void write( const Context& ctx, const float4& sample, int ti, const float2& pattern, const ushort2& p )
    {/*{{{*/
        // find pixel color and save
        int head = ti%nsamples == 0;
        int tail = (ti+1)%nsamples == 0;
        //
        float vj[4];
        vj[0] = k3( pattern.x ); vj[1] = k2( pattern.x ); vj[2] = k1( pattern.x ); vj[3] = k0( pattern.x );
        float vi[4];
        vi[0] = k3( pattern.y ); vi[1] = k2( pattern.y ); vi[2] = k1( pattern.y ); vi[3] = k0( pattern.y );
        for (int i=0; i<4; ++i) {
            #pragma unroll
            for (int j=0; j<4; ++j) {
                float4 s = sample_seg_scan_reg( vj[j]*vi[i]*sample, head, ti );
                if ( tail ) {
                    int idx_out = (p.y + i -1)*ctx.m_rowstride+(p.x+j-1);
                    atomicAdd( ctx.m_out[0] + idx_out, s.x/float(nsamples) );
                    atomicAdd( ctx.m_out[1] + idx_out, s.y/float(nsamples) );
                    atomicAdd( ctx.m_out[2] + idx_out, s.z/float(nsamples) );
                    atomicAdd( ctx.m_out[3] + idx_out, s.w/float(nsamples) );
                }
            }
        }
    }/*}}}*/

    __device__
    void write_bg( const Context& ctx, int ti, const ushort2& p )
    {/*{{{*/
        for ( int i=0; i<4; ++i ) {
            int di = i - 1;
            #pragma unroll
            for ( int j=0; j<4; ++j ) {
                int dj = j - 1;
                int idx_out = (p.y + di)*ctx.m_rowstride+(p.x+dj);
                float4 out = m_empty_samples[4*i + j];
                atomicAdd( ctx.m_out[0] + idx_out, out.x );
                atomicAdd( ctx.m_out[1] + idx_out, out.y );
                atomicAdd( ctx.m_out[2] + idx_out, out.z );
                atomicAdd( ctx.m_out[3] + idx_out, out.w );
            }
        }
    }/*}}}*/

};

template <int NSAMP>
std::vector<float2> get_blue_noise_pattern(float adj)
{/*{{{*/
    std::vector<float2> blue_noise;
    blue_noise.reserve(NSAMP);
    for(int i=0; i<NSAMP; ++i)
    {
        float2 n = make_float2(BlueNoise<NSAMP>::x[i], BlueNoise<NSAMP>::y[i])+adj;
        blue_noise.push_back(n);
    }

    return blue_noise;
}/*}}}*/

std::vector<float> calc_bspline3_data(const std::vector<float2> &pattern)
{/*{{{*/
    std::vector<float> bspline3_data;
    bspline3_data.resize(KS*KS, 0);
    Bs3 bs3;
    std::vector<float> px(4, 0), py(4, 0);
    for(int i=0; i<pattern.size(); ++i)
    {
        px[0] = bs3.k3( pattern[i].x ); 
        px[1] = bs3.k2( pattern[i].x ); 
        px[2] = bs3.k1( pattern[i].x ); 
        px[3] = bs3.k0( pattern[i].x ); 
        py[0] = bs3.k3( pattern[i].y ); 
        py[1] = bs3.k2( pattern[i].y ); 
        py[2] = bs3.k1( pattern[i].y ); 
        py[3] = bs3.k0( pattern[i].y ); 
        for ( int k=0; k<4; ++k ) {
            for ( int l=0; l<4; ++l ) {
                bspline3_data[k*4+l] += py[k]*px[l]/pattern.size();
            }
        }
    }
    return bspline3_data;
}/*}}}*/

std::vector<float2> get_blue_noise_pattern(unsigned nsamples, float adj)
{/*{{{*/
    switch(nsamples)
    {
    case 4:
    case 8:
        return get_blue_noise_pattern<8>(adj);
    case 16:
        return get_blue_noise_pattern<16>(adj);
    case 32:
        return get_blue_noise_pattern<32>(adj);
    case 64:
        return get_blue_noise_pattern<64>(adj);
    case 128:
        return get_blue_noise_pattern<128>(adj);
    case 256:
        return get_blue_noise_pattern<256>(adj);
    default:
        throw std::runtime_error("Invalid sample count");
    }
}/*}}}*/

template<class T>
__device__
T exscan_sum_32( T v, const int ti )
{/*{{{*/
    T tmp1 = __shfl_up( v, 1 );
    if ( ti == 0 ) tmp1 = 0;
    for ( unsigned i=1; i<32; i<<=1) {
        T tmp2 = __shfl_up( tmp1, i );
        if ( ti >= i ) tmp1 += tmp2;
    }
    return tmp1;
}/*}}}*/

__global__
void gpu_convert_segment_flags( const_SegmentsPtr segs, SegmentIndex* sindex )
{/*{{{*/
    int idseg = blockIdx.x*blockDim.x + threadIdx.x;
    if ( idseg >= segs.size )
        return;
    SegmentIndex ind;
    // loads segment
    ind.type = segs.type[idseg];
    ind.winding = segs.winding[idseg];
    ind.flags = segs.flags[idseg];
    //
    ind.geom_size = find_segment_size( static_cast<SegType>(ind.type) );
    //assert( ind.geom_size > 0 );
    ind.paint_size = 0;
    if ( ind.type == SEG_PATH ) {
        const PathData *pdata = segs.data<PathData>(idseg);
        Paint::Type ptype = static_cast<Paint::Type>(pdata->paint_type);
        ind.paint_size = find_paint_size( ptype );
    }
    /*
    flags |= SEG_SET_SIZE( seg_size );
    flags |= SEG_SET_PAINT_SIZE( paint_size );
    //
    segs.flags[idseg] = flags;
    //
    */
    ind.idx_data = segs.idx_data[idseg];
    sindex[idseg] = ind;
}/*}}}*/

template< class F, int bsize, int smsize >
__launch_bounds__(32*bsize,10)
__global__
void gpu_raster_shared( Context ctx, const unsigned *pkt_base, const unsigned* pkt_size,
                            unsigned size, const short2 * mem_cfg )
{/*{{{*/
    int ti = threadIdx.x; //for each sample inside a warp
    int tj = threadIdx.y; //one for each working warp
    //int pi = blockIdx.x * bsize + tj;
    int pi = blockIdx.x;

    if ( pi >= size )
        return;
    //
    // create local geometry
    //__shared__ bool proceed;
    //__shared__ SegmentIndex si[32];
    __shared__ short2 mem;
    //
    //__shared__ ushort2 pix[32]; //32 from packet size
    __shared__ int idquad;
    //

    // all threads load the same packet
    // loads the first tinfo
    if (ti == 0 && tj==0 ) {
        unsigned pbase = pkt_base[pi];
        TraversalInfo tinfo;
        // each thread loads one tinfo (for each unit in
        // the packet).
        ctx.m_tinfos.get(pbase, tinfo);
        //pix[ti] = tinfo.pix;
        idquad = tinfo.idquad;
        assert(idquad >= 0);
        mem = mem_cfg[idquad];
    }

    __shared__ float4 _buffer[smsize/sizeof(float4)];
    unsigned char *buffer = reinterpret_cast<unsigned char*>(_buffer);
    SegmentIndex * si = reinterpret_cast<SegmentIndex*>(_buffer);

    // rod: wait for warp 0 load idquad and mem
    __syncthreads();

    // exit if this quad if too big
    //if ( mem.x == -1)
    if ( int(mem.x) + int(mem.y) >= smsize )
        return;
    assert( int(mem.x) + int(mem.y) < smsize );

    uint2 range = ctx.m_seg_range[idquad];
    unsigned geom_begin = range.x;
    unsigned geom_size = range.y;
    if(geom_size == 0)
    {
        unsigned psize = pkt_size[pi];
        unsigned pbase = pkt_base[pi];
        F filter;
        for ( unsigned ua = 0; ua < psize; ua += 32 )
        {
            if ( tj == 0 )
            {
                if ( ua + ti < psize )
                {
                    ushort2 p = ctx.m_tinfos.pix[pbase + ua + ti];
                    filter.write_bg( ctx, ti, p );
                }
            }
    //        __syncthreads();
        }
        return;
    }

    // reconfigure memory
    // mem.y is the memory used by the indices
    buffer += mem.x;
    assert( (long long)buffer % 16 == 0 );

    // rod: warp loops through all segments
    // rod: 32 at a time, one for each thread
    for ( int l=0, mem_pos=0; l<geom_size; l+=32 )
    {
        __shared__ int pos_in_buf[32];
        __shared__ int total;

        if(tj == 0)
        {
            int size = 0;
            if(l + ti < geom_size)
            {
                si[l + ti] = ctx.m_si[geom_begin + l + ti];
                size = si[l + ti].geom_size + si[l + ti].paint_size;
            }
            pos_in_buf[ti] = exscan_sum_32(size, ti);
            // save the size and position of every segment the buffer
            assert( mem.x + mem_pos + pos_in_buf[ti] + size < smsize );
            buffer[mem_pos+pos_in_buf[ti]] = (unsigned char) size;

            int last_thread = min(32,geom_size-l)-1;
            total = pos_in_buf[last_thread]+__shfl(size,last_thread);
        }

        __syncthreads();

        // load each segment using all threads in the warp
        //assert( geom_size <= 32 );
        // rod: each thread loops through all loaded and
        // rod: the whole warp do a coalesced read of seg data
        for(unsigned _k=l; _k<min(geom_size,l+32); _k += bsize )
        {
            int k = _k+tj;

            if(k >= geom_size)
                break;

            SegmentIndex &lsi = si[k];
            unsigned char *lbuffer = buffer+mem_pos+pos_in_buf[k-l];

            assert( *lbuffer > 0 );
            if ( lsi.get_type() <= SEG_LAST_GEOM ) { // load data
                unsigned * dst = reinterpret_cast<unsigned*>(lbuffer);
                const unsigned * src = reinterpret_cast<const unsigned*>( ctx.m_segs._data + lsi.idx_data );
                // rod: warp loads up to 128 bytes worth of data (32*4)
                // rod: seg data is 16-byte aligned, so it's ok to load using
                // rod: one unsigned (4 bytes) per iteration
                if ( ti < lsi.geom_size/sizeof(unsigned) )
                    dst[ti] = __ldg(src + ti);

                // rod: if the loaded data is of a path segment,
                if(lsi.get_type() == SEG_PATH)
                {
                    // rod: load the paint data, if available
                    PathData *pdata = reinterpret_cast<PathData*>(lbuffer);
                    if(pdata->paint != NULL)
                    {
                        lbuffer += lsi.geom_size;

                        unsigned * dst = reinterpret_cast<unsigned*>(lbuffer);
                        const unsigned * src = reinterpret_cast<const unsigned*>(pdata->paint);
                        if(ti < lsi.paint_size/sizeof(unsigned))
                            dst[ti] = __ldg(src + ti);
                        pdata->paint = reinterpret_cast<const Paint *>(dst);
                    }
                }
            }
            if(ti == 0)
                lsi.idx_data = mem_pos + pos_in_buf[k-l];
        }

        mem_pos += total;

        __syncthreads();
    }

    // warps already synchronized

    unsigned psize = pkt_size[pi];

    // load the blue pattern
    float2 pattern = make_float2(d_sampling_pattern_x[ti],
                                 d_sampling_pattern_y[ti]);

    unsigned pbase = pkt_base[pi];
    assert(psize != 0);

#if 0
    if(tj==0 && ti==0)
    for(int i=0; i<smsize/sizeof(float4); ++i)
        *ctx.m_out[0] = _buffer[i].x+_buffer[i].y+_buffer[i].z+_buffer[i].w;
    return;
#endif

#if 1
    F filter;
    for ( unsigned ua = 0; ua < psize; ua+=bsize ) {
        if ( ua + tj < psize ) {
            ushort2 p = ctx.m_tinfos.pix[pbase + ua + tj];
            // sample coordinates
            float2 coord;
            coord.x = float(p.x) + pattern.x;
            coord.y = float(p.y) + pattern.y;
            // find the sample color
            #if ENABLE_CLIPPING
            float4 sample = vgsample_shared_clipping(si, buffer, 0, geom_size, coord, c_bgcolor);
            #else
            float4 sample = vgsample_shared(si, buffer, 0, geom_size, coord, c_bgcolor);
            #endif
            sample = srgb2lrgb(postdivide(sample));
            filter.write( ctx, sample, ti, pattern, p );
        }
    }
#endif
}/*}}}*/

/////////////////////////////////////////////////
template< class F, int bsize, int smsize, int nsamples >
__launch_bounds__(32*bsize,10)
__global__
void gpu_raster_shared( Context ctx, const unsigned *pkt_base, const unsigned* pkt_size,
                            unsigned size, const short2 * mem_cfg )
{/*{{{*/
    int ti = threadIdx.x; //for each sample inside a warp
    int tj = threadIdx.y; //one for each working warp
    //int pi = blockIdx.x * bsize + tj;
    int pi = blockIdx.x;
    //
    // create local geometry
    //__shared__ bool proceed;
    //__shared__ SegmentIndex si[32];
    __shared__ short2 mem;
    //
    //__shared__ ushort2 pix[32]; //32 from packet size
    __shared__ int idquad;
    //
    __shared__ float4 _buffer[smsize/sizeof(float4)];
    unsigned char *buffer = reinterpret_cast<unsigned char*>(_buffer);
    SegmentIndex * si = reinterpret_cast<SegmentIndex*>(_buffer);
    //
    if ( pi >= size )
        return;
    // all threads load the same packet
    unsigned pbase = pkt_base[pi];
    unsigned psize = pkt_size[pi];
    // load the blue pattern
    float2 pattern = make_float2(d_sampling_pattern_x[ti%nsamples],
                                     d_sampling_pattern_y[ti%nsamples]);
    // loads the first tinfo
    // TODO: later each thread loads a tinfo to identify
    // different quads.
    //int idquad = -1;
    if ( ti < psize && tj==0 ) {
        TraversalInfo tinfo;
        // each thread loads one tinfo (for each unit in
        // the packet).
        ctx.m_tinfos.get(pbase + ti, tinfo);
        //pix[ti] = tinfo.pix;
        if ( ti == 0 ) {
            idquad = tinfo.idquad;
            if ( idquad >= 0 )
                mem = mem_cfg[idquad];
        }
    }
    __syncthreads(); //wait for warp 0
    // if the packet is filled of bg unit areas
    if ( idquad < 0 ) {
        F filter;
        for ( unsigned ua = 0; ua < psize; ua += 32 ) {
            if ( tj == 0 ) {
                if ( ua + ti < psize ) {
                    TraversalInfo tinfo;
                    ctx.m_tinfos.get(pbase + ua + ti, tinfo);
                    ushort2 p = tinfo.pix;
                    filter.write_bg( ctx, ti, p );
                }
            }
        }
        return;
    }
    // exit if this quad if too big
    //if ( mem.x == -1 || mem.y == -1 )
    if ( mem.x + mem.y >= smsize )
        return;
    // reconfigure memory
    buffer += mem.x;
    assert( (long long)buffer % 16 == 0 );
    // exit if cell no good.
    uint2 range = ctx.m_seg_range[idquad];
    unsigned geom_begin = range.x;
    unsigned geom_size = range.y;
    // each thread load a segment
    // only the first warp load data.
    if ( tj == 0 ) {
        int size = 0;
        int mem_pos = 0;
        for ( int l=0; l<geom_size; l+=32 ) {
            size = 0;
            if ( l + ti < geom_size ) {
                si[l + ti] = ctx.m_si[geom_begin + l + ti];
                size = si[l + ti].geom_size + si[l + ti].paint_size;
            }
            int pos = exscan_sum_32( size, ti );
            // save the size and position of every segment the buffer
            assert( mem.x + mem_pos + pos + size < smsize );
            buffer[mem_pos + pos] = (unsigned char) size;
            // load each segment using all threads in the warp
            pos=0;
            //assert( geom_size <= 32 );
            for ( unsigned k=l; k<min(geom_size,l+32); ++k ) {
                int size = buffer[mem_pos + pos];
                assert( size > 0 );
                if ( si[k].get_type() <= SEG_LAST_GEOM ) { // load data
                    unsigned * dst = reinterpret_cast<unsigned*>( buffer + mem_pos + pos );
                    const unsigned * src = reinterpret_cast<const unsigned*>( ctx.m_segs._data + si[k].idx_data );
                    if ( ti < si[k].geom_size/sizeof(unsigned) ) {
                        dst[ti] = src[ti];
                    }
                }
                if ( si[k].get_type() == SEG_PATH ) {
                    PathData *pdata = reinterpret_cast<PathData*>( buffer + mem_pos + pos );
                    if ( pdata->paint > 0 ) {
                        unsigned * dst = reinterpret_cast<unsigned*>( buffer + mem_pos + pos + si[k].geom_size );
                        const unsigned * src = reinterpret_cast<const unsigned*>( pdata->paint );
                        if ( ti < si[k].paint_size/sizeof(unsigned) ) {
                            dst[ti] = src[ti];
                        }
                        pdata->paint = reinterpret_cast<const Paint *>(dst);
                    }
                    else {
                        pdata->paint = 0;
                    }
                }
                if ( ti == 0 ) {
                    //idx_data[k] = pos;
                    si[k].idx_data = mem_pos + pos;
                }
                pos += size;
            }
            mem_pos += pos;
        }
    }
    __syncthreads(); //other warps wait for warp 0.
    //return;
#if 1
    F filter;
    assert( 32%nsamples == 0 );
    int num_ua_per_warp = 32/nsamples;
    for ( unsigned ua = 0; ua < psize; ua+=(bsize*num_ua_per_warp) ) {
        if ( ua + tj*num_ua_per_warp < psize ) {
            int iua = ua + tj*num_ua_per_warp + ti/nsamples;
            // this if works because the tail of a block in the
            // seg_scan in filter matches nsamples.
            if ( iua < psize ) {
                TraversalInfo tinfo;
                ctx.m_tinfos.get(pbase + ua + tj*num_ua_per_warp + ti/nsamples, tinfo);
                ushort2 p = tinfo.pix;
                // sample coordinates
                float2 coord;
                coord.x = float(p.x) + pattern.x;
                coord.y = float(p.y) + pattern.y;
                // find the sample color
                #if ENABLE_CLIPPING
                float4 sample = vgsample_shared_clipping(si, buffer, 0, geom_size, coord, c_bgcolor);
                #else
                float4 sample = vgsample_shared(si, buffer, 0, geom_size, coord, c_bgcolor);
                #endif
                sample = srgb2lrgb(postdivide(sample));
                filter.write( ctx, sample, ti, pattern, p );
            }
        }
    }
#endif
}/*}}}*/

struct SampleState {
    uchar4 rgba;
    short winding;
#if ENABLE_CLIPPING
    ClipData clip;
#endif
};

__device__ inline
uchar4 to_uchar4( const float4& s )
{
    return make_uchar4( char(255.f*s.x), char(255.f*s.y),
            char(255.f*s.z), char(255.f*s.w) );
}

__device__ inline
float4 to_float4( const uchar4& s )
{
    return make_float4( float(s.x), float(s.y),
            float(s.z), float(s.w) )/255.f;
}

template< class F, int bsize, int smsize, int maxpktsize >
__launch_bounds__(32*bsize,10)
__global__
void gpu_raster_shared_geom( Context ctx, const unsigned *pkt_base, const unsigned* pkt_size,
        unsigned size, const short2 * mem_cfg )
{/*{{{*/
    int ti = threadIdx.x; //for each sample inside a warp
    int tj = threadIdx.y; //one for each working warp
    //int pi = blockIdx.x * bsize + tj;
    int pi = blockIdx.x;

    if ( pi >= size )
        return;

    // create local geometry
    __shared__ SegmentIndex si[32];

    __shared__ ushort2 pix[bsize]; //32 from packet size
    __shared__ int idquad;
    __shared__ int g_inc;
    __shared__ short2 mem;

    __shared__ float4 _buffer[smsize/sizeof(float4)];
    unsigned char *buffer = reinterpret_cast<unsigned char*>(_buffer);

    // 32 samples per ua, maxpktsize ua in the packet
    __shared__ SampleState samples[bsize][32];

    unsigned psize = pkt_size[pi];
    unsigned pbase = pkt_base[pi];

    // all threads load the same packet
    if(ti == 0 && tj == 0)
    {
        // each thread loads one tinfo (for each unit in
        // the packet).
        idquad = ctx.m_tinfos.idquad[pbase];
        assert(idquad >= 0);
        mem = mem_cfg[idquad];
    }

    __syncthreads(); //wait for warp 0

    uint2 range = ctx.m_seg_range[idquad];

    // return if this is a simple cell
    //if(mem.x != -1 || range.y == 0)
    if(mem.x + mem.y < smsize || range.y == 0)
        return;

    for(int p = 0; p < psize; p += bsize)
    {
        //Reseting sample context
#if FRONT_TO_BACK_RENDERING
        float4 sample = make_float4(0,0,0,0);
#else
        float4 sample = c_bgcolor;
#endif
        samples[tj][ti].rgba = to_uchar4(sample);
        samples[tj][ti].winding = 0;
#if ENABLE_CLIPPING
        samples[tj][ti].clip.data = 0;
        samples[tj][ti].clip.stop_depth = -1;
#endif

        // load the blue pattern
        float2 pattern = make_float2(d_sampling_pattern_x[ti],
                                     d_sampling_pattern_y[ti]);

        unsigned geom_begin = range.x;
        unsigned geom_size = range.y;

        for(int g=0; g<geom_size; g+=g_inc)/*{{{*/
        {
            // each thread load a segment
            // only the first warp load data.
            if(tj == 0)
            {
                int size = 0;
                int pos = 0;
                if(g + ti < geom_size)
                {
                    si[ti] = ctx.m_si[geom_begin + g + ti];
                    size = si[ti].geom_size + si[ti].paint_size;
                }
                pos = exscan_sum_32( size, ti);

                // rod: I think we can do something a tad faster with __shfl

                // g_inc will be the size of the data we'll process in this
                // iteration. It's enough to fill the shared mem we have.
                int mask = __ballot((pos + size) > smsize);
                if (ti == 0)
                    g_inc = (mask == 0) ? min(32, geom_size-g) : (__ffs( mask )-1);

                // save the size and position of every segment the buffer
                if (ti < g_inc)
                    buffer[pos] = (unsigned char) size;

                // load each segment using all threads in the warp
                pos=0;
                for(unsigned k=0; k<g_inc; ++k)/*{{{*/
                {
                    size = buffer[pos];
                    assert(size > 0);
                    if (si[k].get_type() <= SEG_LAST_GEOM ) // load data
                    {
                        unsigned * dst = reinterpret_cast<unsigned*>( buffer + pos );
                        const unsigned * src = reinterpret_cast<const unsigned*>( ctx.m_segs._data + si[k].idx_data );

                        if ( ti < si[k].geom_size/sizeof(unsigned) )
                            dst[ti] = __ldg(src+ti);

                        if(si[k].get_type() == SEG_PATH)
                        {
                            PathData *pdata = reinterpret_cast<PathData*>( buffer + pos );
                            if (pdata->paint != NULL)
                            {
                                unsigned * dst = reinterpret_cast<unsigned*>( buffer + pos + si[k].geom_size );
                                const unsigned * src = reinterpret_cast<const unsigned*>( pdata->paint );
                                if ( ti < si[k].paint_size/sizeof(unsigned) )
                                    dst[ti] = __ldg(src+ti);

                                pdata->paint = reinterpret_cast<const Paint *>(dst);
                            }
                        }
                    }
                    if (ti == 0)
                        si[k].idx_data = pos;

                    pos += size;
                }/*}}}*/
            }

            pix[tj] = ctx.m_tinfos.pix[pbase+p+tj];
            __syncthreads(); //other warps wait for warp 0.

            if ( p + tj < psize )/*{{{*/
            {
                int ua = tj;
                /*
                TraversalInfo tinfo;
                // all threads in the warp load the same tinfo
                ctx.m_tinfos.get(pbase + ua, tinfo);
                */
                ushort2 p = pix[ua];
                // sample coordinates
                float2 coord;
                coord.x = float(p.x) + pattern.x;
                coord.y = float(p.y) + pattern.y;
                // find the sample color
                float4 sample = to_float4(samples[ua][ti].rgba);
                int w = samples[ua][ti].winding;
#if ENABLE_CLIPPING
                ClipData clip = samples[ua][ti].clip;
                vgsample_shared_clipping(si, buffer, 0, g_inc, coord, c_bgcolor,
                        sample, w, clip);
                samples[ua][ti].clip = clip;
#else
                vgsample_shared(si, buffer, 0, g_inc, coord, c_bgcolor,
                        sample, w);
#endif
                //float4 sample = make_float4( 1.f, 0.f, 0.f, 1.f );
                samples[ua][ti].rgba = to_uchar4(sample);
                samples[ua][ti].winding = w;
            }/*}}}*/
            // all warps have to wait here, otherwise the first warp will write on
            // buffer while the other warps are still reading it.
            __syncthreads();
        }/*}}}*/

        if ( p + tj < psize )
        {
            int ua = tj;

            F filter;
            ushort2 p = pix[ua];
            // int idx_out = p.y*ctx.m_rowstride+p.x; // never referenced
            // find pixel color and save
            float4 sample = to_float4( samples[ua][ti].rgba );
            if(sample.w < 1.f)
                sample = blend(sample, premultiply(c_bgcolor));
            sample = srgb2lrgb(postdivide(sample));
            filter.write( ctx, sample, ti, pattern, p );
        }
    }
}/*}}}*/

template< class F, int bsize, int smsize, int maxpktsize, int nsamples >
__launch_bounds__(32*bsize,10)
__global__
void gpu_raster_shared_geom( Context ctx, const unsigned *pkt_base, const unsigned* pkt_size,
        unsigned size, const short2 * mem_cfg )
{/*{{{*/
    int ti = threadIdx.x; //for each sample inside a warp
    int tj = threadIdx.y; //one for each working warp
    //int pi = blockIdx.x * bsize + tj;
    int pi = blockIdx.x;
    // create local geometry
    __shared__ SegmentIndex si[32];
    //
    __shared__ ushort2 pix[32]; //32 from packet size
    __shared__ int idquad;
    __shared__ int g_inc;
    __shared__ short2 mem;
    //
    __shared__ float4 _buffer[smsize/sizeof(float4)];
    unsigned char *buffer = reinterpret_cast<unsigned char*>(_buffer);
    //
    // 32 samples per ua, maxpktsize ua in the packet
    __shared__ SampleState samples[maxpktsize][nsamples];
    //
    if ( pi >= size )
        return;
    // all threads load the same packet
    unsigned pbase = pkt_base[pi];
    unsigned psize = pkt_size[pi];
    // load the blue pattern
    float2 pattern = make_float2(d_sampling_pattern_x[ti%nsamples],
                                     d_sampling_pattern_y[ti%nsamples]);
    // loads the first tinfo
    // TODO: later each thread loads a tinfo to identify
    // different quads.
    //int idquad = -1;
    if ( ti < psize && tj==0 ) {
        TraversalInfo tinfo;
        // each thread loads one tinfo (for each unit in
        // the packet).
        ctx.m_tinfos.get(pbase + ti, tinfo);
        pix[ti] = tinfo.pix;
        if ( ti == 0 ) {
            idquad = tinfo.idquad;
            if ( idquad >= 0 )
                mem = mem_cfg[idquad];
        }
    }
    __syncthreads(); //wait for warp 0
    if( idquad < 0 || (mem.x + mem.y < smsize) )
        return;
    assert( mem.x + mem.y >= smsize );
    //Reseting sample context
    if ( tj==0 ) {
        for ( int k=0; k<maxpktsize; ++k ) {
            #if FRONT_TO_BACK_RENDERING
            float4 sample = make_float4(0,0,0,0);
            #else
            float4 sample = c_bgcolor;
            #endif
            if ( ti < nsamples ) {
                samples[k][ti].rgba = to_uchar4(sample);
                samples[k][ti].winding = 0;
                #if ENABLE_CLIPPING
                samples[k][ti].clip.data = 0;
                samples[k][ti].clip.stop_depth = -1;
                #endif
            }
        }
    }
    //
    uint2 range = ctx.m_seg_range[idquad];
    unsigned geom_begin = range.x;
    unsigned geom_size = range.y;
    for ( int g=0; g<geom_size; g+=g_inc ) {/*{{{*/
        // each thread load a segment
        // only the first warp load data.
        if ( tj == 0 ) {
            int size = 0;
            int pos = 0;
            if ( (g + ti < geom_size) && tj == 0 ) {
                si[ti] = ctx.m_si[geom_begin + g + ti];
                size = si[ti].geom_size + si[ti].paint_size;
            }
            pos = exscan_sum_32( size, ti );
            int mask = __ballot( (pos + size) > smsize );
            if ( ti == 0 ) {
                g_inc = (mask == 0) ? min(32, geom_size-g): (__ffs( mask )-1);
            }
            // save the size and position of every segment the buffer
            if ( ti < g_inc ) {
                buffer[pos] = (unsigned char) size;
            }
            // load each segment using all threads in the warp
            pos=0;
            for ( unsigned k=0; k<g_inc; ++k ) {/*{{{*/
                size = buffer[pos];
                assert(size > 0);
                if ( si[k].get_type() <= SEG_LAST_GEOM ) { // load data
                    unsigned * dst = reinterpret_cast<unsigned*>( buffer + pos );
                    const unsigned * src = reinterpret_cast<const unsigned*>( ctx.m_segs._data + si[k].idx_data );
                    //if ( pi == 16 )
                    //    printf("geosize:%d\n", si[k].geom_size/sizeof(unsigned));
                    if ( ti < si[k].geom_size/sizeof(unsigned) ) {
                        dst[ti] = src[ti];
                    }
                }
                if ( si[k].get_type() == SEG_PATH && tj == 0) {
                    PathData *pdata = reinterpret_cast<PathData*>( buffer + pos );
                    if ( pdata->paint > 0 ) {
                        unsigned * dst = reinterpret_cast<unsigned*>( buffer + pos + si[k].geom_size );
                        const unsigned * src = reinterpret_cast<const unsigned*>( pdata->paint );
                        if ( ti < si[k].paint_size/sizeof(unsigned) ) {
                            dst[ti] = src[ti];
                        }
                        pdata->paint = reinterpret_cast<const Paint *>(dst);
                    }
                    else {
                        pdata->paint = 0;
                    }
                }
                if ( tj == 0 && ti == 0 ) {
                    si[k].idx_data = pos;
                }
                pos += size;
            }/*}}}*/
        }
        __syncthreads(); //other warps wait for warp 0.
#if 1
        assert( 32%nsamples == 0 );
        int num_ua_per_warp = 32/nsamples;
        for ( unsigned ua = 0; ua < psize; ua+=(bsize*num_ua_per_warp) ) {
            if ( ua + tj*num_ua_per_warp < psize ) {
                int iua = ua + tj*num_ua_per_warp + ti/nsamples;
                if ( iua < psize ) {
                    int is = ti%nsamples;
                    TraversalInfo tinfo;
                    ctx.m_tinfos.get(pbase + iua, tinfo);
                    ushort2 p = tinfo.pix;
                    // sample coordinates
                    float2 coord;
                    coord.x = float(p.x) + pattern.x;
                    coord.y = float(p.y) + pattern.y;
                    // find the sample color
                    float4 sample = to_float4(samples[iua][is].rgba);
                    int w = samples[iua][is].winding;
                    #if ENABLE_CLIPPING
                    ClipData clip = samples[iua][is].clip;
                    vgsample_shared_clipping(si, buffer, 0, g_inc, coord, c_bgcolor,
                            sample, w, clip);
                    samples[iua][is].clip = clip;
                    #else
                    vgsample_shared(si, buffer, 0, g_inc, coord, c_bgcolor,
                            sample, w);
                    #endif
                    //float4 sample = make_float4( 1.f, 0.f, 0.f, 1.f );
                    samples[iua][is].rgba = to_uchar4(sample);
                    samples[iua][is].winding = w;
                }
            }
        }
        // all warps have to wait here, otherwise the first warp will write on
        // buffer while the other warps are still reading it.
        __syncthreads();
#endif
    }/*}}}*/
    if ( tj == 0 ) {
        F filter;
        int num_ua_per_warp = 32/nsamples;
        for ( int k=0; k<min(maxpktsize,psize); k+=num_ua_per_warp ) {
            int iua = k + ti/nsamples;
            if ( iua < psize ) {
                int is = ti%nsamples;
                ushort2 p = pix[iua];
                // int idx_out = p.y*ctx.m_rowstride+p.x; // never referenced
                // find pixel color and save
                float4 sample = to_float4( samples[iua][is].rgba );
                if(sample.w < 1.f)
                    sample = blend(sample, premultiply(c_bgcolor));
                sample = srgb2lrgb(postdivide(sample));
                filter.write( ctx, sample, ti, pattern, p );
            }
        }
    }
}/*}}}*/

template< class T >
void dump_vector( const gpu::vector<T>& vector )
{
    std::vector<T> host;
    vector.copy_to_host( host );
    for ( int i=0; i<host.size(); ++i ) {
        const T& e = host[i];
        std::cerr << e << " ";
    }
    std::cerr << std::endl;
}

void dump_packets( const gpu::vector<unsigned>& pkt_base, const gpu::vector<unsigned>& pkt_size, TraversalInfos& tinfos )
{
	(void) dump_packets; // avoid "never used" warning
    std::vector<unsigned> h_base, h_size;
    pkt_base.copy_to_host( h_base );
    pkt_size.copy_to_host( h_size );
    std::vector<unsigned> h_quad;
    std::vector<ushort2> h_pix;
    tinfos.idquad.copy_to_host( h_quad );
    tinfos.pix.copy_to_host( h_pix );
    for ( int i=0; i<pkt_base.size(); ++i ) {
        unsigned base = h_base[i];
        unsigned size = h_size[i];
        std::cerr << "base: " << h_base[i] << " size: " << h_size[i] << std::endl;
        for ( int j=0; j<size; ++j ) {
            ushort2 pix = h_pix[base + j];
            fprintf( stderr, "\t quad:%d pix:%d,%d\n", h_quad[base + j], pix.x, pix.y );
        }
    }
}

void dump_packets_size( const gpu::vector<unsigned>& pkt_size )
{
    for( int i=0; i<pkt_size.size(); ++i ) {
        std::cout << "PKTSIZE" << pkt_size[i] << std::endl;
    }
}

void sort_tinfos( TraversalInfos& tinfos )
{
    /*
    gpu::vector<unsigned> index( tinfos.size() );
    thrust::sequence( index.begin(), index.end() );
    thrust::sort_by_key( tinfos.idquad.begin(), tinfos.idquad.end(), index.begin() );
    */
    thrust::sort_by_key( tinfos.idquad.begin(), tinfos.idquad.end(), tinfos.pix.begin() );
}

} //namespace

void rasterize_shared(const float4 &bg_color,
               const Quads &qtree,
               const Segments &segs,
               unsigned nsamples,
               gpu::image_ptr<float4> out,
               unsigned flags,
               unsigned use_bs3)
{/*{{{*/
    cudaDeviceSetSharedMemConfig(cudaSharedMemBankSizeEightByte);

    //std::cerr << "raster shared" << std::endl;
    int width = out.width(), height = out.height();

    gpu::copy_to_symbol(c_bgcolor, bg_color);
    gpu::copy_to_symbol(d_num_samples, nsamples);

    std::vector<float2> pattern;
    if ( nsamples == 4 ) {
        std::vector<float2> tmp = get_blue_noise_pattern(4,.5);
        for (int i=0; i<8; i+=2 ) {
            pattern.push_back( tmp[i] );
        }
    }
    else {
        pattern = get_blue_noise_pattern(nsamples,.5);
    }
    gpu::copy_to_symbol(d_sampling_pattern, pattern);
    {
        std::vector<float> patx, paty;
        for(unsigned i=0; i<pattern.size(); ++i)
        {
            patx.push_back(pattern[i].x);
            paty.push_back(pattern[i].y);
        }
        gpu::copy_to_symbol(d_sampling_pattern_x, patx);
        gpu::copy_to_symbol(d_sampling_pattern_y, paty);
    }

    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("rasterize"));

    gpu::image<float> channels[4];
    //
    TraversalInfos tinfo;
    gpu::vector<unsigned> segs_per_quadpaint;
    { //traverse qtree
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("traverse qtree"));
        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("traversal"));
            traverse_qtree(tinfo, qtree, width, height, .5);
        }
        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("sorting tinfos"));
            sort_tinfos( tinfo );
        }
        //dump_tinfo( tinfo );
        //
    }
    //
    {
        const int maxpktsize = 32;
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("collect samples"));
        //
        gpu::vector<unsigned> pkt_base;
        gpu::vector<unsigned> pkt_size;
        //
        { //create packets
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("create_packets"));
            create_packets( pkt_base, pkt_size, tinfo, 32 );
            //create_packets( pkt_base, pkt_size, tinfo, 128 );
            if ( flags & 2 )
                dump_packets_size( pkt_size );
        }
        gpu::vector<SegmentIndex> segment_index;
        segment_index.resize( segs.size() );
        {//convert sizes into flags
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("convert sizes into flags"));
            cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);
            int bsize = 64;
            int gsize = ( segs.size() + bsize - 1 )/bsize;
            dim3 grid( gsize ), block( bsize );
            gpu_convert_segment_flags<<<grid, block>>>( &segs, &segment_index );
            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
        }
        //const int smsize = 4*1024;
        const int smsize = 2900;
        { //raster_shared
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("raster"));
            //cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);
            //cudaDeviceSetCacheConfig(cudaFuncCachePreferEqual);
            cudaDeviceSetCacheConfig(cudaFuncCachePreferShared);
            //std::cerr << "pkt size:" << pkt_base.size() << std::endl;
            const int bsize = 4;
            dim3 grid( pkt_base.size() ), block( 32, bsize );
            if ( use_bs3 ) {
                {
                    std::vector<float> bspline3 = calc_bspline3_data(pattern);
                    std::vector<float4> empty_samples(KS*KS);
                    #pragma unroll
                    for(int i=0; i<KS*KS; ++i) {
                        empty_samples[i] = srgb2lrgb(bg_color)*bspline3[i];
                    }
                    gpu::copy_to_symbol(c_empty_samples, empty_samples);
                }
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("resize channedls bs3"));
                    for(int i=0; i<4; ++i) {
                        channels[i].resize(width+IMG_BORDER*2,height+IMG_BORDER*2);
                        set_zero(channels[i]);
                    }
                }

                {
                    Context ctx( tinfo, &qtree.seg_range,
                                  segs, &segment_index,
                                  width, height, IMG_BORDER,
                                  channels );
                    if ( nsamples == 16 ) {
                        {
                            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple 16 samples bs3"));
                            gpu_raster_shared< Bs3NSamples<16>, bsize, smsize, 16 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                                pkt_base.size(), qtree.mem_config.data() );
                        }
                        {
                            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi 16 samples bs3"));
                            gpu_raster_shared_geom< Bs3NSamples<16>, bsize, smsize, maxpktsize, 16 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                                pkt_base.size(), qtree.mem_config.data() );
                        }
                    }
                    else if ( nsamples == 8 ) {
                        {
                            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple 8 samples bs3"));
                            gpu_raster_shared< Bs3NSamples<8>, bsize, smsize, 8 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                                pkt_base.size(), qtree.mem_config.data() );
                        }
                        {
                            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi 8 samples bs3"));
                            gpu_raster_shared_geom< Bs3NSamples<8>, bsize, smsize, maxpktsize, 8 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                                pkt_base.size(), qtree.mem_config.data() );
                        }
                    }
                    else if ( nsamples == 4 ) {
                        {
                            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple 4 samples bs3"));
                            gpu_raster_shared< Bs3NSamples<4>, bsize, smsize, 4 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                                pkt_base.size(), qtree.mem_config.data() );
                        }
                        {
                            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi 4 samples bs3"));
                            gpu_raster_shared_geom< Bs3NSamples<4>, bsize, smsize, maxpktsize, 4 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                                pkt_base.size(), qtree.mem_config.data() );
                        }
                    }
                    else {
//                        cudaFuncSetSharedMemConfig(gpu_raster_shared< Bs3, bsize, smsize>, cudaSharedMemBankSizeEightByte);
                        {
                            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple bs3"));
                            gpu_raster_shared< Bs3, bsize, smsize ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                                pkt_base.size(), qtree.mem_config.data() );
                        }
                        //cudaFuncSetSharedMemConfig(gpu_raster_shared_geom< Bs3, bsize, smsize, maxpktsize>, cudaSharedMemBankSizeEightByte);
                        {
                            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi bs3"));
                            gpu_raster_shared_geom< Bs3, bsize, smsize, maxpktsize ><<<grid, block>>>( ctx, pkt_base.data(),
                                pkt_size.data(), pkt_base.size(), qtree.mem_config.data() );
                        }
                    }
                }
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("do cardinal bspline3"));
                    // calculate cubic b-spline weights
                    gpu::math::Vector<float,1+1> bspline3_weights;
                    {
                        float a = 2.f-std::sqrt(3.0f);
                        bspline3_weights[0] = 1+a;
                        bspline3_weights[1] = a;
                    }
                    recursive_filter_5_setup(out.width(),
                                             out.height(),
                                             out.rowstride(),
                                             bspline3_weights, gpu::CLAMP_TO_EDGE, 1);
                    for(int i=0; i<4; ++i) {
                        recursive_filter_5(channels[i], IMG_BORDER);
                    }
                    gpu::recursive_filter_5_free();
                }
            }
            else {
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("resize channels"));
                    for(int i=0; i<4; ++i) {
                        channels[i].resize(width,height);
                        set_zero(channels[i]);
                    }
                }
                Context ctx( tinfo, &qtree.seg_range,
                              segs, &segment_index,
                              width, height, 0,
                              channels );
                if ( nsamples == 16 ) {
                    {
                        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple 16 samples box"));
                        gpu_raster_shared< BoxNSamples<16>, bsize, smsize, 16 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                            pkt_base.size(), qtree.mem_config.data() );
                    }
                    {
                        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi 16 samples box"));
                        gpu_raster_shared_geom< BoxNSamples<16>, bsize, smsize, maxpktsize, 16 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                            pkt_base.size(), qtree.mem_config.data() );
                    }
                }
                else if ( nsamples == 8 ) {
                    {
                        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple 8 samples box"));
                        gpu_raster_shared< BoxNSamples<8>, bsize, smsize, 8 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                            pkt_base.size(), qtree.mem_config.data() );
                    }
                    {
                        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi 8 samples box"));
                        gpu_raster_shared_geom< BoxNSamples<8>, bsize, smsize, maxpktsize, 8 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                            pkt_base.size(), qtree.mem_config.data() );
                    }
                }
                else if ( nsamples == 4 ) {
                    {
                        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple 4 samples box"));
                        gpu_raster_shared< BoxNSamples<4>, bsize, smsize, 4 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                            pkt_base.size(), qtree.mem_config.data() );
                    }
                    {
                        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi 4 samples box"));
                        gpu_raster_shared_geom< BoxNSamples<4>, bsize, smsize, maxpktsize, 4 ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                            pkt_base.size(), qtree.mem_config.data() );
                    }
                }
                else {
                    {
                        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple"));
                        gpu_raster_shared< Box, bsize, smsize ><<<grid, block>>>( ctx, pkt_base.data(), pkt_size.data(),
                            pkt_base.size(), qtree.mem_config.data() );
                    }
                    {
                        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi"));
                        gpu_raster_shared_geom< Box, bsize, smsize, maxpktsize ><<<grid, block>>>( ctx, pkt_base.data(),
                            pkt_size.data(), pkt_base.size(), qtree.mem_config.data() );
                    }
                }
            }
            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
        }
    }
    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("compose channels into output image"));
        compose(out, channels[0], channels[1], channels[2], channels[3]);
    }
}/*}}}*/
