/* Algoritmo do scheduler:

   cada warp processa um pixel
   cada thread processa um sample

   (OK)
   a thread desce a qtree e ve em qual celula o sample cai
   o warp calcula o numero de celulas diferentes, se for menor que N,
   indexamos o buffer de saida pelo indice do pixel e salvamos o indice das
   celulas. se for maior que N, salvamos QUAD_OVERFLOW (indica pixels complicados)

   (OK)
   criamos uma lista com elementos idcelula,idpixel, agrupados por celula

   cada warp processa um elemento da lista. cada thread transforma o sample e
   testa se ele está dentro da célula. marcamos num array em smem os samples
   que estão dentro e compactamos esta lista, agrupando no inicio os samples
   que estão dentro. dividimos este array em warps e cada warp processa no
   máximo 32 samples, que podem vir de pixels diferentes. fazemos uma soma
   parcial (por pixel) da cor do sample e somamos o resultado no pixel, na
   saida.

   compactamos a lista, criando uma só com os pixels complicados.

   calculamos a cor dos pixels complicados do jeito tradicional.
*/

#include <vector>
#include <gpu/vector.h>
#include <gpu/image.h>
#include <gpu/symbol.h>
#include <gpu/cub.h>
#include <gpu/timer.h>
#include <gpu/recfilter.h>
#include <gpu/image_util.h>
#include <gpu/value.h>
#include <thrust/sort.h>
#include <thrust/functional.h>
#include <cstdio>
#include "persistent.h"
#include "segments.h"
#include "sampling.h"
#include "implicit.h"
#include "blue_noise.h"
#include "quad.h"
#include "util.h"
#include "packets.h"
#include "warps.h"

const int MAX_BLOCKS_PER_SM = 8;
const int MAX_QUADS_PER_PIXEL = 16;
const int TW = 8, TH = 4;
const int BW = 32,
          BH = TW*TH;
const int KS = 4; // pre-filter kernel support size
const int IMG_BORDER = 32;

const unsigned QUAD_OVERFLOW = (unsigned)-2,
               NO_QUADS = (unsigned)-1;


__constant__ float2 d_sampling_pattern[MAX_SAMPDIM];
__device__ float d_sampling_pattern_x[MAX_SAMPDIM],
                 d_sampling_pattern_y[MAX_SAMPDIM];
__constant__ float d_bspline3_data[MAX_SAMPDIM*KS*KS];
__device__ float c_bspline3_sum[KS*KS];
__device__ float4 c_empty_samples[KS*KS];
__constant__ unsigned d_num_samples;
__constant__ float4 c_bgcolor;

namespace {

struct QuadInfo {
    short num_segments;
    short data_size;
};

/*{{{ blue noise initialization*/
__host__ __device__
float bspline3(float r)/*{{{*/
{
    r = std::abs(r);

    if (r < 1.f)
        return (4.f + r*r*(-6.f + 3.f*r))/6.f;
    else if (r < 2.f)
        return  (8.f + r*(-12.f + (6.f - r)*r))/6.f;
    else
        return 0.f;
}/*}}}*/

template <int NSAMP>
std::vector<float2> get_blue_noise_pattern(float adj)/*{{{*/
{
    std::vector<float2> blue_noise;
    blue_noise.reserve(NSAMP);
    for(int i=0; i<NSAMP; ++i)
    {
        float2 n = make_float2(BlueNoise<NSAMP>::x[i], BlueNoise<NSAMP>::y[i])+adj;
        blue_noise.push_back(n);
    }

    return blue_noise;
}/*}}}*/

template<class T>
__device__
T exscan_sum_32( T v, const int ti )
{
    T tmp1 = __shfl_up( v, 1 );
    if ( ti == 0 ) tmp1 = 0;
    for ( unsigned i=1; i<32; i<<=1) {
        T tmp2 = __shfl_up( tmp1, i );
        if ( ti >= i ) tmp1 += tmp2;
    }
    return tmp1;
}

template<class T>
__device__
T scan_max_32( T v, const int ti )
{
    T tmp1 = v;
    for ( unsigned i=1; i<32; i<<=1) {
        T tmp2 = __shfl_up( tmp1, i );
        if ( ti >= i ) tmp1 = max(tmp1,tmp2);
    }
    return tmp1;
}

template<class T>
__device__
T segscan_32( T *values, int mindex, const unsigned id )
{
    if ( id >= mindex +  1) values[id] = values[id -  1] + values[id];
    if ( id >= mindex +  2) values[id] = values[id -  2] + values[id];
    if ( id >= mindex +  4) values[id] = values[id -  4] + values[id];
    if ( id >= mindex +  8) values[id] = values[id -  8] + values[id];
    if ( id >= mindex + 16) values[id] = values[id - 16] + values[id];
    return values[id];
}

template<class T>
__device__
inline T segscan_32( T val, int mindex, const unsigned id )
{
    T aux = __shfl_up( val, 1 );
    if ( id >= mindex +  1) val = aux + val;
    aux = __shfl_up( val, 2 );
    if ( id >= mindex +  2) val = aux + val;
    aux = __shfl_up( val, 4 );
    if ( id >= mindex +  4) val = aux + val;
    aux = __shfl_up( val, 8 );
    if ( id >= mindex +  8) val = aux + val;
    aux = __shfl_up( val, 16 );
    if ( id >= mindex +  16) val = aux + val;
    return val;
}

__device__
inline int mark_head_flag( int idua, int id )
{
    int tmp1 = idua;
    int tmp2 = __shfl_up( tmp1, 1 );
    if ( id == 0 ) tmp2 = 1;
    else {
        tmp2 = ( idua != tmp2 ) ? 1 : 0;
    }
    return tmp2;
}

__device__
inline int mark_tail_flag( int idua, int id, int max)
{
    int tmp1 = idua;
    int tmp2 = __shfl_down( tmp1, 1 );
    if ( id == max-1 ) tmp2 = 1;
    else {
        tmp2 = ( idua != tmp2 ) ? 1 : 0;
    }
    return tmp2;
}

template<class T>
__device__
inline T sample_seg_scan( T* samples, int idua, int id )
{
    int hf = mark_head_flag( idua, id );
    if ( hf ) hf = id;
    hf = scan_max_32( hf, id );
    return segscan_32( samples, hf, id );
}

__device__
inline float4 sample_seg_scan_reg( const float4& samples, int idua, int id )
{
    int hf = mark_head_flag( idua, id );
    if ( hf ) hf = id;
    hf = scan_max_32( hf, id );
    float4 out;
    out.x = segscan_32( samples.x, hf, id );
    out.y = segscan_32( samples.y, hf, id );
    out.z = segscan_32( samples.z, hf, id );
    out.w = segscan_32( samples.w, hf, id );
    return out;
}

template<class T>
__device__
T scan_sum_32( T v, const int ti )
{
    T tmp1 = v;
    for ( unsigned i=1; i<32; i<<=1) {
        T tmp2 = __shfl_up( tmp1, i );
        if ( ti >= i ) tmp1 += tmp2;
    }
    return tmp1;
}

template<class T>
__device__
inline T scan_block( const T& v, int tj, int ti, T* tmp, int tmp_size )
{
    T nv = scan_sum_32( v, ti );
    if ( ti == 31 ) {
        tmp[tj] = nv;
    }
    __syncthreads();
    if ( tj == 0 ) {
        T v2 = 0;
        if ( ti < tmp_size ) {
            v2 = tmp[ti];
        }
        v2 = scan_sum_32( v2, ti );
        if ( ti < tmp_size ) {
            tmp[ti] = v2;
        }
    }
    __syncthreads();
    if ( tj > 0 ) {
        nv += tmp[tj-1];
    }
    return nv;
}

__device__
inline
int all_block( int v, int tj, int ti, int *tmp, int tmp_size )
{
    int all_used = __all(v?1:0);
    if ( ti == 0 ) {
        tmp[tj] = all_used;
    }
    __syncthreads();
    if ( tj == 0 ) {
        int v = 1;
        if ( ti < tmp_size ) {
            v = tmp[ti];
        }
        v = __all( v );
        if ( ti < tmp_size ) {
            tmp[ti] = v;
        }
    }
    __syncthreads();
    return tmp[tj];
}


struct Context {
    Context(
                  const uint2 *quad_ua,
                  const uint2 *seg_range,
                  const Segments &segs,
                  const Quads &qtree,
                  const SegmentIndex *si,
                  unsigned width, unsigned height,
                  unsigned border,
                  gpu::image<float> out[4])
            : m_quad_ua(quad_ua)
            , m_seg_range(seg_range)
            , m_segs(&segs)
            , m_qtree(&qtree)
            , m_si(si)
            , m_width(width)
            , m_height(height)
            , m_rowstride(out[0].rowstride())
    {
        for(unsigned i=0; i<4; ++i)
            m_out[i] = out[i].data() + m_rowstride*border + border;
    }

    const uint2 *m_quad_ua;
    const uint2 *m_seg_range;
    const_QuadsPtr m_qtree;
    const_SegmentsPtr m_segs;
    const SegmentIndex* m_si;
    unsigned m_width, m_height, m_rowstride;
    float *m_out[4];
};

struct Bs3 {
#if 0
    const float4 * m_empty_samples;

    __device__ __host__
    Bs3( )
       : m_empty_samples( c_empty_samples )
    {
    }
#endif

    __device__ __host__
    float k0( float u ) { return (1.f/6.f)*(((u)*u)*u); }
    __device__ __host__
    float k1( float u ) { return (1.f/6.f)*(((-3.f*u+3.f)*u+3.f)*u+1.f); }
    __device__ __host__
    float k2( float u ) { return (1.f/6.f)*(((3.f*u-6.f)*u)*u+4.f); }
    __device__ __host__
    float k3( float u ) { return (1.f/6.f)*(((3.f-u)*u-3.f)*u+1.f); }

    __device__
    void write( const Context& ctx, const float4& sample, int ti, const float2& pattern, const short2& p, int tail, int ua )
    {
        float vj[4];
        vj[0] = k3( pattern.x ); vj[1] = k2( pattern.x ); vj[2] = k1( pattern.x ); vj[3] = k0( pattern.x );
        float vi[4];
        vi[0] = k3( pattern.y ); vi[1] = k2( pattern.y ); vi[2] = k1( pattern.y ); vi[3] = k0( pattern.y );
        for (int i=0; i<4; ++i) {
            for (int j=0; j<4; ++j) {
                float4 s = sample_seg_scan_reg( vj[j]*vi[i]*sample, ua, ti );
                if ( tail ) {
                    int idx_out = (p.y + i -1)*ctx.m_rowstride+(p.x+j-1);
                    atomicAdd( ctx.m_out[0] + idx_out, s.x/32.0f );
                    atomicAdd( ctx.m_out[1] + idx_out, s.y/32.0f );
                    atomicAdd( ctx.m_out[2] + idx_out, s.z/32.0f );
                    atomicAdd( ctx.m_out[3] + idx_out, s.w/32.0f );
                }
            }
        }
    }

};

struct Box {

    __device__
    void write( const Context& ctx, const float4& sample, int ti, const float2& pattern, const short2& p, int tail, int ua )
    {
        float4 s = sample_seg_scan_reg( sample, ua, ti );
        int idx_out = p.y*ctx.m_rowstride+p.x;
        #pragma unroll
        for( int c = 0; c < 4; ++c ) {
            float aux = ((const float *)&s)[c];
            if(tail) {
                aux = atomicAdd( ctx.m_out[c] + idx_out, aux/32.0f );
            }
        }
    }

};

std::vector<float> calc_bspline3_data(const std::vector<float2> &pattern)
{/*{{{*/
    std::vector<float> bspline3_data;
    bspline3_data.resize(KS*KS, 0);
    Bs3 bs3;
    std::vector<float> px(4, 0), py(4, 0);
    for(int i=0; i<pattern.size(); ++i)
    {
        px[0] = bs3.k3( pattern[i].x );
        px[1] = bs3.k2( pattern[i].x );
        px[2] = bs3.k1( pattern[i].x );
        px[3] = bs3.k0( pattern[i].x );
        py[0] = bs3.k3( pattern[i].y );
        py[1] = bs3.k2( pattern[i].y );
        py[2] = bs3.k1( pattern[i].y );
        py[3] = bs3.k0( pattern[i].y );
        for ( int k=0; k<4; ++k ) {
            for ( int l=0; l<4; ++l ) {
                bspline3_data[k*4+l] += py[k]*px[l]/pattern.size();
            }
        }
    }
    return bspline3_data;
}/*}}}*/

std::vector<float2> get_blue_noise_pattern(unsigned nsamples, float adj)/*{{{*/
{
    switch(nsamples)
    {
    case 8:
        return get_blue_noise_pattern<8>(adj);
    case 16:
        return get_blue_noise_pattern<16>(adj);
    case 32:
        return get_blue_noise_pattern<32>(adj);
    case 64:
        return get_blue_noise_pattern<64>(adj);
    case 128:
        return get_blue_noise_pattern<128>(adj);
    case 256:
        return get_blue_noise_pattern<256>(adj);
    default:
        throw std::runtime_error("Invalid sample count");
    }
}/*}}}*/
/*}}}*/

__device__
int traverse_qtree(const_QuadsPtr qtree, float2 coord)/*{{{*/
{
    unsigned idquad = 0;
    while(__ldg(qtree.seg_range+idquad).y == 0)
    {
        int c = qtree.classify(idquad, coord);

        int idchildquad;

        int4 children = __ldg(qtree.children + idquad);

        switch(c)
        {
        case 0:
            idchildquad = children.x;
            break;
        case 1:
            idchildquad = children.y;
            break;
        case 2:
            idchildquad = children.z;
            break;
        case 3:
            idchildquad = children.w;
            break;
        }

        if(idchildquad >= 0)
            idquad = idchildquad;
        else
        {
            if ( !(children.x == -1 &&
                    children.y == -1 &&
                    children.z == -1 &&
                    children.w == -1 )) {
                idquad = (unsigned)-1;
            }
            else {
                //assert(false);
            }
            break;
        }
    }

    if(idquad != (unsigned)-1 && !qtree.contains(idquad, coord))
        idquad = (unsigned)-1;

    return idquad;
}/*}}}*/

#if 0
__device__
int traverse_qtree(const_QuadsPtr qtree, float2 coord)/*{{{*/
{
    unsigned idquad = 0;
    while(qtree.seg_range[idquad].y == 0)
    {
        int c = qtree.classify(idquad, coord);

        int idchildquad;

        int4 children = qtree.children[idquad];

        switch(c)
        {
        case 0:
            idchildquad = children.x;
            break;
        case 1:
            idchildquad = children.y;
            break;
        case 2:
            idchildquad = children.z;
            break;
        case 3:
            idchildquad = children.w;
            break;
        }

        if(idchildquad >= 0)
            idquad = idchildquad;
        else
        {
            idquad = (unsigned)-1;
            break;
        }
    }

    if(idquad != (unsigned)-1 && !qtree.contains(idquad, coord))
        idquad = (unsigned)-1;

    return idquad;
}/*}}}*/
#endif

template<class W>
__global__
void gpu_traverse_qtree32(const W warp, const_QuadsPtr qtree, uint2 *quad_pixel, /*{{{*/
                          unsigned *next_free_slot,
                          unsigned w, unsigned h)
{
    int tx = threadIdx.x, ty = threadIdx.y;
    //int2 pix = make_int2(blockIdx.x, blockIdx.y*BH+ty);
    int y = ty/TW;
    int x = ty%TW;
    int2 pix = make_int2(TW*blockIdx.x + x, TH*blockIdx.y + y);
    //
    if(pix.x >= w || pix.y >= h)
        return;
    //
    unsigned idsample = tx, idpixel = (pix.y*w)+pix.x;
    float2 coord = make_float2(pix.x, pix.y);
    coord.x += d_sampling_pattern[idsample].x;
    coord.y += d_sampling_pattern[idsample].y;
    //coord = apply(c_xform, coord);
    coord = warp.apply(coord);
    //
    int idquad = traverse_qtree(qtree, coord);
    //
    // now we must compact pixel's idquads
    __shared__ unsigned idquads[BH][32];
    idquads[ty][tx] = idquad;
    //
    unsigned *data = &idquads[ty][0];
    // Why not use __any ?
    //if(__all(data[tx] != (unsigned)-1)) {
    if(__any(data[tx] != (unsigned)-1)) {
        // bitonic sort
        for(uint size=2; size<=32; size *= 2) {
            for(uint stride = size/2; stride > 0; stride >>=1) {
                int ixj = tx ^ stride;
                int ddd = (tx & size) == 0;
                if(ixj > tx && (data[tx] > data[ixj]) == ddd)
                    gpu::swap(data[tx], data[ixj]);
            }
        }
        // mark duplicates
        int idx = tx;
        while(data[idx] == data[tx]) {
            if(--idx < 0)
                break;
        }
        // remove marked elements
        __shared__ unsigned len[BH];
        if(tx == 0)
            len[ty] = 0;
        if(idx+1 == tx) {
            unsigned tmp = data[tx];
            unsigned pos = atomicAdd(len+ty, 1);
            data[pos] = tmp;
        }

        __shared__ int idxout[BH];

        if(tx == 0)
            idxout[ty] = atomicAdd(next_free_slot, len[ty]);

        // each thread writes on cell index
        if(tx < len[ty])
            quad_pixel[idxout[ty]+tx] = make_uint2(data[tx], idpixel);
    }
    else {
        if ( tx == 0 ) {
            int idxout = atomicAdd(next_free_slot, 1);
            quad_pixel[idxout] = make_uint2( (unsigned) -1, idpixel );
        }
    }
}/*}}}*/

template<class W>
class TraverseQtree /*{{{*/
{
public:
    TraverseQtree( const W warp, const_QuadsPtr qtree, unsigned *pixel_quad,
                          unsigned *quads_per_pixel,
                          unsigned w, unsigned h ,
                          unsigned nw, unsigned nh )
        : m_warp( warp )
        , m_qtree( qtree )
        , m_pixel_quad( pixel_quad )
        , m_quads_per_pixel( quads_per_pixel )
        , m_w( w )
        , m_h( h )
        , m_nw( nw )
        , m_nh( nh )
    {
    }

    __device__
    void operator()(int idwarp);

private:
    W m_warp;
    const_QuadsPtr m_qtree;
    unsigned *m_pixel_quad;
    unsigned *m_quads_per_pixel;
    unsigned m_w, m_h;
    unsigned m_nw, m_nh;
};/*}}}*/

template< class W >
__device__
void TraverseQtree<W>::operator()(int idwarp)/*{{{*/
{
    int tx = threadIdx.x, ty = threadIdx.y;
    int x0 = idwarp%m_nw;
    int y0 = idwarp/m_nw;
    int x = ty%4;
    int y = ty/4;
    int2 pix = make_int2( 4*x0 + x, 4*y0 + y );
    //
    if(pix.x >= m_w || pix.y >= m_h)
        return;
    //
    unsigned idsample = tx;
    float2 coord = make_float2(pix.x, pix.y);
    coord.x += d_sampling_pattern[idsample].x;
    coord.y += d_sampling_pattern[idsample].y;
    coord = m_warp.apply(coord);
    //
    int idquad = traverse_qtree(m_qtree, coord);
    //
    // now we must compact pixel's idquads
    //__shared__ unsigned idquads[PERSISTENT_WARPS][32];
    __shared__ unsigned idquads[16][32];
    idquads[ty][tx] = idquad;
    //
    unsigned *data = &idquads[ty][0];
    unsigned idxout = (pix.y*m_w + pix.x)*MAX_QUADS_PER_PIXEL;
    if(__any(data[tx] != (unsigned)-1)) {
        // bitonic sort
        for(uint size=2; size<=32; size *= 2) {
            for(uint stride = size/2; stride > 0; stride >>=1) {
                int ixj = tx ^ stride;
                int ddd = (tx & size) == 0;
                if(ixj > tx && (data[tx] > data[ixj]) == ddd)
                    gpu::swap(data[tx], data[ixj]);
            }
        }
        // mark duplicates
        int idx = tx;
        while(data[idx] == data[tx]) {
            if(--idx < 0)
                break;
        }
        // remove marked elements
        //__shared__ unsigned len[PERSISTENT_WARPS];
        __shared__ unsigned len[16];
        if(tx == 0)
            len[ty] = 0;
        if(idx+1 == tx) {
            unsigned tmp = data[tx];
            unsigned pos = atomicAdd(len + ty, 1);
            data[pos] = tmp;
        }
        // if the number of cell is less than MAX..
        // write out result
        if(len[ty] <= MAX_QUADS_PER_PIXEL) {
            // each thread writes on cell index
            if(tx < len[ty]) {
                m_pixel_quad[idxout + tx] = data[tx];
            }
            // the first thread writes the number of cell
            if(tx == 0) {
                m_quads_per_pixel[pix.y*m_w+pix.x] = max(1,len[ty]);
            }
            //
            return;
        }
    }
    // These are the problematic cells
    if(tx == 0) {
        m_pixel_quad[idxout] = QUAD_OVERFLOW;
        m_quads_per_pixel[pix.y*m_w+pix.x] = 1; // to hold problematic pixel
    }
}/*}}}*/


__device__
int find_paint_size( const Paint::Type& type )
{
    int s = 0;
    switch ( type ) {
        case Paint::SOLID:
            s = sizeof(PaintSolid);
            break;
        case Paint::SOLID_ALPHA:
            s = sizeof(PaintSolidAlpha);
            break;
        case Paint::GRAD_LINEAR:
            s = sizeof(PaintLinearGrad);
            break;
        case Paint::GRAD_RADIAL:
            s = sizeof(PaintRadialGrad);
            break;
        case Paint::TEXTURE:
            s = sizeof(PaintTexture);
            break;
        default:
            break;
    }
    if ( s & 0x0F ) {
        s &= ~0x0F;
        s += 16;
    }
    return s;
}

__device__
int find_segment_size( const SegType& type )
{
    //return 64;
    int s = 0;
    switch( type ) {
        case SEG_LINEAR:
            s = sizeof(ImplicitCurve<1>);
            break;
        case SEG_QUADRATIC:
            s = sizeof(ImplicitCurve<2>);
            break;
        case SEG_CUBIC:
            s = sizeof(ImplicitCurve<3>);
            break;
        case SEG_ARC:
            s = sizeof(ImplicitCurve<SEG_ARC>);
            break;
        case SEG_PATH:
            s = sizeof(PathData);
            break;
        case SEG_PUSH:
        case SEG_POP_FILL:
        case SEG_POP_CLIP:
        case SEG_COMMIT:
        case SEG_POINT:
            //TODO: change this to 0...
            s = 16;
            break;
        default:
            break;
    }
    if ( s & 0x0F ) {
        s &= ~0x0F;
        s += 16;
    }
    return s;
}

__global__
void gpu_convert_segment_flags( const_SegmentsPtr segs, SegmentIndex* sindex )
{
    int idseg = blockIdx.x*blockDim.x + threadIdx.x;
    if ( idseg >= segs.size )
        return;
    SegmentIndex ind;
    // loads segment
    ind.type = segs.type[idseg];
    ind.winding = segs.winding[idseg];
    ind.flags = segs.flags[idseg];
    //
    ind.geom_size = find_segment_size( static_cast<SegType>(ind.type) );
    ind.paint_size = 0;
    if ( ind.type == SEG_PATH ) {
        const PathData *pdata = segs.data<PathData>(idseg);
        Paint::Type ptype = static_cast<Paint::Type>(pdata->paint_type);
        ind.paint_size = find_paint_size( ptype );
    }
    ind.idx_data = segs.idx_data[idseg];
    sindex[idseg] = ind;
}


#if 1
#endif


__device__ inline
uchar4 to_uchar4( const float4& s )
{
    return make_uchar4( char(255.f*s.x), char(255.f*s.y),
            char(255.f*s.z), char(255.f*s.w) );
}

__device__ inline
float4 to_float4( const uchar4& s )
{
    return make_float4( float(s.x), float(s.y),
            float(s.z), float(s.w) )/255.f;
}


}//namespace

inline __host__ __device__
bool operator<( const uint2& a, const uint2& b ) {
    return a.x < b.x;
}

inline __device__
bool qtree_contains( float2 p, float2 min, float2 max ) {
    return p.x >= min.x && p.x < max.x &&
           p.y >= min.y && p.y < max.y;
}

struct SampleContext {
    int idua; //index of the ua.
    int idsample; //index of the sample
};

enum SampleHitState {
    STATE_OUT,
    STATE_IN,
    STATE_IN_NOT_USED
};

template< class W, class F, int bsize, int smsize >
//__launch_bounds__(32*bsize,16)
__global__
void gpu_raster_warp( const W warp, Context ctx, const unsigned *pkt_base, const unsigned* pkt_size,
        unsigned size, const short2 * mem_cfg )
{/*{{{*/
    int ti = threadIdx.x; //for each sample inside a warp
    int tj = threadIdx.y; //one for each working warp
    int pi = blockIdx.x;
    //
    __shared__ int idquad, geom_begin, geom_size;
    __shared__ float2 quad_min, quad_max;
    __shared__ short2 mem;
    //TODO: one for each warp
    __shared__ int counter[bsize];
    __shared__ SampleContext samples[bsize][32];
    //__shared__ float4 samplecolor[32];
    //
    // local geometry
    //__shared__ SegmentIndex si[32];
    __shared__ float4 _buffer[smsize/sizeof(float4)];
    unsigned char *buffer = reinterpret_cast<unsigned char*>(_buffer);
    SegmentIndex * si = reinterpret_cast<SegmentIndex*>(_buffer);
    //
    if ( pi >= size )
        return;
    // all threads load the same packet
    unsigned pbase = pkt_base[pi];
    unsigned psize = pkt_size[pi];
    //*
    __shared__ float2 pattern[32];
    // */
    if ( tj == 0 ) {
        pattern[ti] = make_float2(d_sampling_pattern_x[ti],
                                  d_sampling_pattern_y[ti]);
        //first thread loads idquad
        if ( ti == 0 ) {
            uint2 pkt = ctx.m_quad_ua[pbase];
            idquad = (int)pkt.x;
            mem = make_short2(0,0);
            if ( idquad >= 0 ) {
                mem = mem_cfg[idquad];
                quad_min = ctx.m_qtree.min[idquad];
                quad_max = ctx.m_qtree.max[idquad];
                // load geometry range
                uint2 range = ctx.m_seg_range[idquad];
                geom_begin = range.x;
                geom_size = range.y;
            }
        }
    }
    __syncthreads();
#if 1
    //if ( idquad == -1 || idquad == -2) {
    if ( idquad < 0 ) {
        // do something and return
        return;
    }
#endif
    // exit if this quad if too big
    //if ( (mem.x == -1 || mem.y == -1 || idquad < 0))
    assert(mem.x>=0 && mem.y >= 0);
    if ( mem.x + mem.y >= smsize )
        return;
    buffer += mem.x;
    assert( (long long)buffer % 16 == 0 );
    //assert( geom_size > 0 );
    //
    if ( tj == 0 ) {
        if ( idquad >= 0 ) {
            int size = 0;
            int mem_pos = 0;
            for ( int l=0; l<geom_size; l+=32 ) {
                size = 0;
                if ( l + ti < geom_size ) {
                    si[l + ti] = ctx.m_si[geom_begin + l + ti];
                    size = si[l + ti].geom_size + si[l + ti].paint_size;
                }
                int pos = exscan_sum_32( size, ti );
                // save the size and position of every segment the buffer
                assert( mem.x + mem_pos + pos + size < smsize );
                buffer[mem_pos + pos] = (unsigned char) size;
                // load each segment using all threads in the warp
                pos=0;
                //assert( geom_size <= 32 );
                for ( unsigned k=l; k<min(geom_size,l+32); ++k ) {
                    int size = buffer[mem_pos + pos];
                    assert( size > 0 );
                    if ( si[k].get_type() <= SEG_LAST_GEOM ) { // load data
                        unsigned * dst = reinterpret_cast<unsigned*>( buffer + mem_pos + pos );
                        const unsigned * src = reinterpret_cast<const unsigned*>( ctx.m_segs._data + si[k].idx_data );
                        if ( ti < si[k].geom_size/sizeof(unsigned) ) {
                            dst[ti] = src[ti];
                        }
                    }
                    if ( si[k].get_type() == SEG_PATH ) {
                        PathData *pdata = reinterpret_cast<PathData*>( buffer + mem_pos + pos );
                        if ( pdata->paint > 0 ) {
                            unsigned * dst = reinterpret_cast<unsigned*>( buffer + mem_pos + pos + si[k].geom_size );
                            const unsigned * src = reinterpret_cast<const unsigned*>( pdata->paint );
                            if ( ti < si[k].paint_size/sizeof(unsigned) ) {
                                dst[ti] = src[ti];
                            }
                            pdata->paint = reinterpret_cast<const Paint *>(dst);
                        }
                        else {
                            pdata->paint = 0;
                        }
                    }
                    if ( ti == 0 ) {
                        //idx_data[k] = pos;
                        si[k].idx_data = mem_pos + pos;
                    }
                    pos += size;
                }
                mem_pos += pos;
            }
            // finish loading geometry
        }
    }
    __syncthreads();
#if 1
    // all warps are processing the same package so when
    // need to find begin and end index for each warp
    int num_ua_per_warp = ( psize + bsize - 1 )/bsize;
    // begin ua, tj is the warp index
    int bua = num_ua_per_warp*tj;
    // end, hope the compiler do something smart here...
    int eua = num_ua_per_warp*(tj+1);
    // if psize < 32 the num_us_per_warp == 1, and some
    // warps will be idle.
    // loop over the ua's in the packet
    //
    //printf("tj:%d ti:%d bua:%d eua:%d psize%d\n", tj, ti, bua, eua, psize);
    if ( bua <= psize ) {
        counter[tj] = 0; //how many samples in the buffer
        SampleHitState sstate = STATE_OUT; //state of each sample in the ua.
        for ( unsigned p = bua; p < eua && p < psize; ) {
        //for ( int p=0; p<psize; ) {
            // id of the ua
            // TODO: hold this in shared mem to avoid
            // unnecessary loads.
            unsigned idua = ctx.m_quad_ua[pbase + p].y;
            short2 ua_pix = make_short2(idua%ctx.m_width,
                                 idua/ctx.m_width);
            // create coordinates for the sample
            // in image space. using index displaced
            // by counter.
            float2 pat = pattern[ti];
            float2 coord;
            coord.x = float(ua_pix.x) + pat.x;
            coord.y = float(ua_pix.y) + pat.y;
            // find coordinates of the sample
            // in the world space.
            coord = warp.apply(coord);
            // test if sample is inside quad cell.
            if(qtree_contains(coord, quad_min, quad_max) && sstate != STATE_IN) {
                // TODO: try using scan.
                unsigned pos = atomicAdd(counter + tj, 1);
                // only writes if it fits into the buffer
                if ( pos < 32 ) {
                    samples[tj][pos].idua = idua;
                    samples[tj][pos].idsample = ti;
                    sstate = STATE_IN;
                }
                else {
                    sstate = STATE_IN_NOT_USED;
                }
            }
            bool all_used = __all(sstate != STATE_IN_NOT_USED);
            // assert that if all samples were used then
            // counter cannot be greater than 32
            assert ( !(all_used && counter[tj] > 32) );
            if ( all_used ) {
                // use next ua
                ++p;
                sstate = STATE_OUT;
                // if we have room on the buffer AND
                // this is not the last ua, skip to
                // the next one
                assert( counter[tj] <= 32 );
                if ( counter[tj] < 32 && p != min(eua,psize) ) {
                //if ( counter[tj] < 32 && p != psize ) {
                    continue;
                }
            }
            else {
                // in this case at least one sample was not used
                // we don't increment p so it uses the same ua.
                // Nor do we reset the sample's state.
                assert( counter[tj] >= 32 );
                // set counter to 32 because the buffer is full.
                //if ( ti == 0 ) {
                    counter[tj] = 32;
                //}
                assert( counter[tj] == 32 );
            }
            // now counter has the number of samples in the buffer,
            // this number is 32 or this is the last ua in the packet.
            assert( counter[tj] == 32 || p == min(eua,psize) );
            //
            // rasterization code
#if 1
            if ( ti < counter[tj] ) {
                SampleContext sc = samples[tj][ti];
                short2 pix = make_short2( sc.idua%ctx.m_width,
                                            sc.idua/ctx.m_width);
                //float2 pattern2 = make_float2(d_sampling_pattern_x[sc.idsample],
                //                             d_sampling_pattern_y[sc.idsample]);
                float2 pat = pattern[sc.idsample];
                float2 coord;
                coord.x = float(pix.x) + pat.x;
                coord.y = float(pix.y) + pat.y;
                coord = warp.apply(coord);
                #if ENABLE_CLIPPING
                float4 sample = vgsample_shared_clipping(si, buffer, 0, geom_size, coord, c_bgcolor);
                #else
                float4 sample = vgsample_shared(si, buffer, 0, geom_size, coord, c_bgcolor);
                #endif
                sample = srgb2lrgb(postdivide(sample));
                int tail = mark_tail_flag( sc.idua, ti, counter[tj] );
                //samplecolor[ti] = sample;
                //sample = sample_seg_scan( samplecolor, sc.idua, ti );
                F filter;
                filter.write( ctx, sample, ti, pat, pix, tail, sc.idua );
            }
            // find the sample color
#endif
            // end rasterization code
            //
            // this is the end of the loop.
            // if we are here then we need to
            // clear the buffer.
            //if ( ti == 0 ) {
                counter[tj] = 0;
            //}
        }
    }
#endif
}/*}}}*/

struct SampleState {
    uchar4 rgba;
    short winding;
#if ENABLE_CLIPPING
    ClipData clip;
#endif
};

template< class W, class F, int bsize, int smsize, int maxpktsize >
//__launch_bounds__(32*bsize,16)
__global__
void gpu_raster_warp_geom( const W warp, Context ctx, const unsigned *pkt_base, const unsigned* pkt_size,
        unsigned size, const short2 * mem_cfg )
{/*{{{*/
    int ti = threadIdx.x; //for each sample inside a warp
    int tj = threadIdx.y; //one for each working warp
    int pi = blockIdx.x;
    //
    __shared__ int idquad, geom_begin, geom_size, g_inc;
    __shared__ int ps;
    __shared__ short2 mem;
    __shared__ float2 quad_min, quad_max;
    //TODO: one for each warp
    __shared__ int counter;
    __shared__ SampleContext samples[bsize*32];
    //
    // local geometry
    __shared__ SegmentIndex si[32];
    __shared__ float4 _buffer[smsize/sizeof(float4)];
    unsigned char *buffer = reinterpret_cast<unsigned char*>(_buffer);
    //
    // TODO: could we use registers?
    __shared__ SampleState samples_state[bsize*32];
    //
    if ( pi >= size )
        return;
    // all threads load the same packet
    unsigned pbase = pkt_base[pi];
    unsigned psize = pkt_size[pi];
    //*
    __shared__ float2 pattern[32];
    // */
    if ( tj == 0 )
    {/*{{{*/
        pattern[ti] = make_float2(d_sampling_pattern_x[ti],
                                  d_sampling_pattern_y[ti]);
        //first thread loads idquad
        if ( ti == 0 ) {
            uint2 pkt = ctx.m_quad_ua[pbase];
            idquad = (int)pkt.x;
            if ( idquad >= 0 ) {
                mem = mem_cfg[idquad];
                quad_min = ctx.m_qtree.min[idquad];
                quad_max = ctx.m_qtree.max[idquad];
                // load geometry range
                uint2 range = ctx.m_seg_range[idquad];
                geom_begin = range.x;
                geom_size = range.y;
                counter = 0;
            }
        }
    }/*}}}*/
    __syncthreads();
#if 1
    //if ( idquad == (unsigned)-1 || idquad == (unsigned)-2) {
    if ( idquad < 0 ) {
        // do something and return
        return;
    }
    // test if simple cell
    //if ( mem.x != -1 && mem.y != -1 ) {
    if ( int(mem.x) + int(mem.y) < smsize ) {
        return;
    }
    {
        SampleHitState sstate = STATE_OUT; //state of each sample in the ua.
        for ( unsigned p = 0; p < psize; ) {
            // id of the ua
            // unnecessary loads.
            if ( tj == 0 ) {
                unsigned idua = ctx.m_quad_ua[pbase + p + tj].y;
                short2 ua_pix = make_short2(idua%ctx.m_width,
                                     idua/ctx.m_width);
                // create coordinates for the sample
                // in image space. using index displaced
                // by counter.
                float2 pat = pattern[ti];
                float2 coord;
                coord.x = float(ua_pix.x) + pat.x;
                coord.y = float(ua_pix.y) + pat.y;
                // find coordinates of the sample
                // in the world space.
                coord = warp.apply(coord);
                // test if sample is inside quad cell.
                if(qtree_contains(coord, quad_min, quad_max) && sstate != STATE_IN) {
                    // TODO: try using scan.
                    unsigned pos = atomicAdd(&counter, 1);
                    // only writes if it fits into the buffer
                    if ( pos < bsize*32 ) {
                        samples[pos].idua = idua;
                        samples[pos].idsample = ti;
                        sstate = STATE_IN;
                    }
                    else {
                        sstate = STATE_IN_NOT_USED;
                    }
                }
                //int all_used = all_block( sstate != STATE_IN_NOT_USED ? 1 : 0, tj, ti, tmp, bsize );
                bool all_used = __all(sstate != STATE_IN_NOT_USED);
                // assert that if all samples were used then
                // counter cannot be greater than 32
                assert ( !(all_used && counter > bsize*32) );
                if ( all_used ) {
                    // use next ua
                    ++p;
                    sstate = STATE_OUT;
                    // if we have room on the buffer AND
                    // this is not the last ua, skip to
                    // the next one
                    assert( counter <= bsize*32 );
                    if ( counter < bsize*32 && p != psize ) {
                        continue;
                    }
                }
                else {
                    // in this case at least one sample was not used
                    // we don't increment p so it uses the same ua.
                    // Nor do we reset the sample's state.
                    assert( counter >= bsize*32 );
                    // set counter to 32 because the buffer is full.
                    //if ( ti == 0 && tj == 0 ) {
                        counter = bsize * 32;
                    //}
                }
                //__syncthreads();
                // now counter has the number of samples in the buffer,
                // this number is 32 or this is the last ua in the packet.
                assert( (counter == (bsize*32)) || p >= psize );
                //
                // rasterization code
                //if ( ti == 0 )
                    ps = p;
            }
            __syncthreads();
            // set p so all warp can the loop together
            p = ps;
#if 1
            {
                #if FRONT_TO_BACK_RENDERING
                float4 sample = make_float4(0,0,0,0);
                #else
                float4 sample = c_bgcolor;
                #endif
                samples_state[tj*32 + ti].rgba = to_uchar4(sample);
                samples_state[tj*32 + ti].winding = 0;
                #if ENABLE_CLIPPING
                samples_state[tj*32 + ti].clip.data = 0;
                samples_state[tj*32 + ti].clip.stop_depth = -1;
                #endif
            }
            g_inc = 0;
            __syncthreads();
            for ( int g=0; g<geom_size; g+=g_inc ) {
                // each thread load a segment
                // only the first warp load data.
                if ( tj == 0 ) {
                    int size = 0;
                    int pos = 0;
                    if ( (g + ti < geom_size) && tj == 0 ) {
                        si[ti] = ctx.m_si[geom_begin + g + ti];
                        size = si[ti].geom_size + si[ti].paint_size;
                    }
                    pos = exscan_sum_32( size, ti );
                    int mask = __ballot( (pos + size) > smsize );
                    if ( ti == 0 ) {
                        g_inc = (mask == 0) ? min(32, geom_size-g): (__ffs( mask )-1);
                    }
                    // save the size and position of every segment the buffer
                    if ( ti < g_inc ) {
                        buffer[pos] = (unsigned char) size;
                    }
                    // load each segment using all threads in the warp
                    pos=0;
                    for ( unsigned k=0; k<g_inc; ++k ) {
                        size = buffer[pos];
                        assert(size > 0);
                        if ( si[k].get_type() <= SEG_LAST_GEOM ) { // load data
                            unsigned * dst = reinterpret_cast<unsigned*>( buffer + pos );
                            const unsigned * src = reinterpret_cast<const unsigned*>( ctx.m_segs._data + si[k].idx_data );
                            //if ( pi == 16 )
                            //    printf("geosize:%d\n", si[k].geom_size/sizeof(unsigned));
                            if ( ti < si[k].geom_size/sizeof(unsigned) ) {
                                dst[ti] = src[ti];
                            }
                        }
                        if ( si[k].get_type() == SEG_PATH && tj == 0) {
                            PathData *pdata = reinterpret_cast<PathData*>( buffer + pos );
                            if ( pdata->paint > 0 ) {
                                unsigned * dst = reinterpret_cast<unsigned*>( buffer + pos + si[k].geom_size );
                                const unsigned * src = reinterpret_cast<const unsigned*>( pdata->paint );
                                if ( ti < si[k].paint_size/sizeof(unsigned) ) {
                                    dst[ti] = src[ti];
                                }
                                pdata->paint = reinterpret_cast<const Paint *>(dst);
                            }
                            else {
                                pdata->paint = 0;
                            }
                        }
                        if ( tj == 0 && ti == 0 ) {
                            si[k].idx_data = pos;
                        }
                        pos += size;
                    }
                }
                __syncthreads(); //other warps wait for warp 0.
                if ( tj*32 + ti < counter ) {
                    SampleContext sc = samples[32*tj + ti];
                    short2 pix = make_short2( sc.idua%ctx.m_width,
                                                sc.idua/ctx.m_width);
                    //float2 pattern2 = make_float2(d_sampling_pattern_x[sc.idsample],
                    //                             d_sampling_pattern_y[sc.idsample]);
                    float2 pat = pattern[sc.idsample];
                    float2 coord;
                    coord.x = float(pix.x) + pat.x;
                    coord.y = float(pix.y) + pat.y;
                    coord = warp.apply(coord);
                    float4 sample = to_float4(samples_state[tj*32 + ti].rgba);
                    int w = samples_state[tj*32 + ti].winding;
                    #if ENABLE_CLIPPING
                    ClipData clip = samples_state[tj*32 + ti].clip;
                    vgsample_shared_clipping(si, buffer, 0, g_inc, coord, c_bgcolor,
                            sample, w, clip);
                    samples_state[tj*32 + ti].clip = clip;
                    #else
                    vgsample_shared(si, buffer, 0, g_inc, coord, c_bgcolor,
                            sample, w);
                    #endif
                    //float4 sample = make_float4( 1.f, 0.f, 0.f, 1.f );
                    samples_state[tj*32 + ti].rgba = to_uchar4(sample);
                    samples_state[tj*32 + ti].winding = w;
                }
                __syncthreads();
            }
            if ( tj*32 + ti < counter ) {
                SampleContext sc = samples[32*tj + ti];
                short2 pix = make_short2( sc.idua%ctx.m_width,
                                            sc.idua/ctx.m_width);
                float2 pat = pattern[sc.idsample];
                float4 sample = to_float4(samples_state[tj*32 + ti].rgba);
                if(sample.w < 1.f)
                    sample = blend(sample, premultiply(c_bgcolor));
                sample = srgb2lrgb(postdivide(sample));
                int maxpos = 32;
                if ( tj == counter/32 )  {
                    maxpos = counter%32;
                }
                int tail = mark_tail_flag( sc.idua, ti, maxpos );
                F filter;
                filter.write( ctx, sample, ti, pat, pix, tail, sc.idua );
            }
#endif
            // end rasterization code
            //
            // this is the end of the loop.
            // if we are here then we need to
            // clear the buffer.
            if ( tj == 0 ) {
                counter = 0;
            }
        }
    }
#endif
}/*}}}*/

template< class W, class F, int bsize, int smsize >
//__launch_bounds__(32*bsize,16)
__global__
void gpu_raster_warp_out_quad( const W warp, Context ctx, const unsigned *pkt_base, const unsigned* pkt_size,
        unsigned size, const short2 * mem_cfg )
{/*{{{*/
    int ti = threadIdx.x; //for each sample inside a warp
    int tj = threadIdx.y; //one for each working warp
    int pi = blockIdx.x;
    //
    __shared__ bool exit;
    //
    __shared__ int idquad;
    __shared__ float2 quad_min, quad_max;
    //TODO: one for each warp
    __shared__ int counter[bsize];
    __shared__ SampleContext samples[bsize][32];
    //
    // local geometry
    if ( pi >= size )
        return;
    // all threads load the same packet
    unsigned pbase = pkt_base[pi];
    unsigned psize = pkt_size[pi];
    //*
    __shared__ float2 pattern[32];
    // */
    if ( tj == 0 ) {
        pattern[ti] = make_float2(d_sampling_pattern_x[ti],
                                  d_sampling_pattern_y[ti]);
        //first thread loads idquad
        if ( ti == 0 ) {
            uint2 pkt = ctx.m_quad_ua[pbase];
            idquad = (int)pkt.x;
            exit = false;
            if ( idquad >= 0 ) {
                // normal cell
                exit = true;
            }
            else {
                // outside!
                quad_min = ctx.m_qtree.min[0];
                quad_max = ctx.m_qtree.max[0];
            }
        }
    }
    __syncthreads();
    // exit if not outside quad.
    if ( exit ) return;
    //
#if 1
    // all warps are processing the same package so when
    // need to find begin and end index for each warp
    int num_ua_per_warp = ( psize + bsize - 1 )/bsize;
    // begin ua, tj is the warp index
    int bua = num_ua_per_warp*tj;
    // end, hope the compiler do something smart here...
    int eua = num_ua_per_warp*(tj+1);
    // if psize < 32 the num_us_per_warp == 1, and some
    // warps will be idle.
    // loop over the ua's in the packet
    //
    //printf("tj:%d ti:%d bua:%d eua:%d psize%d\n", tj, ti, bua, eua, psize);
    if ( bua <= psize ) {
        counter[tj] = 0; //how many samples in the buffer
        SampleHitState sstate = STATE_OUT; //state of each sample in the ua.
        for ( unsigned p = bua; p < eua && p < psize; ) {
            // id of the ua
            // TODO: hold this in shared mem to avoid
            // unnecessary loads.
            unsigned idua = ctx.m_quad_ua[pbase + p].y;
            short2 ua_pix = make_short2(idua%ctx.m_width,
                                 idua/ctx.m_width);
            // create coordinates for the sample
            // in image space. using index displaced
            // by counter.
            float2 pat = pattern[ti];
            float2 coord;
            coord.x = float(ua_pix.x) + pat.x;
            coord.y = float(ua_pix.y) + pat.y;
            // find coordinates of the sample
            // in the world space.
            coord = warp.apply(coord);
            // test if sample is OUTSIDE root cell.
            if(!qtree_contains(coord, quad_min, quad_max) && sstate != STATE_IN) {
                // TODO: try using scan.
                unsigned pos = atomicAdd(counter + tj, 1);
                // only writes if it fits into the buffer
                if ( pos < 32 ) {
                    samples[tj][pos].idua = idua;
                    samples[tj][pos].idsample = ti;
                    sstate = STATE_IN;
                }
                else {
                    sstate = STATE_IN_NOT_USED;
                }
            }
            bool all_used = __all(sstate != STATE_IN_NOT_USED);
            // assert that if all samples were used then
            // counter cannot be greater than 32
            assert ( !(all_used && counter[tj] > 32) );
            if ( all_used ) {
                // use next ua
                ++p;
                sstate = STATE_OUT;
                // if we have room on the buffer AND
                // this is not the last ua, skip to
                // the next one
                assert( counter[tj] <= 32 );
                if ( counter[tj] < 32 && p != min(eua,psize) ) {
                //if ( counter[tj] < 32 && p != psize ) {
                    continue;
                }
            }
            else {
                // in this case at least one sample was not used
                // we don't increment p so it uses the same ua.
                // Nor do we reset the sample's state.
                assert( counter[tj] >= 32 );
                // set counter to 32 because the buffer is full.
                //if ( ti == 0 ) {
                    counter[tj] = 32;
                //}
                assert( counter[tj] == 32 );
            }
            // now counter has the number of samples in the buffer,
            // this number is 32 or this is the last ua in the packet.
            assert( counter[tj] == 32 || p == min(eua,psize) );
            //
            // rasterization code
#if 1
            if ( ti < counter[tj] ) {
                SampleContext sc = samples[tj][ti];
                short2 pix = make_short2( sc.idua%ctx.m_width,
                                            sc.idua/ctx.m_width);
                float4 sample = srgb2lrgb(postdivide(c_bgcolor));
                int tail = mark_tail_flag( sc.idua, ti, counter[tj] );
                F filter;
                filter.write( ctx, sample, ti, pat, pix, tail, sc.idua );
            }
            // find the sample color
#endif
            // end rasterization code
            //
            // this is the end of the loop.
            // if we are here then we need to
            // clear the buffer.
            //if ( ti == 0 ) {
                counter[tj] = 0;
            //}
        }
    }
#endif
}/*}}}*/

__global__
void gpu_mark_leaf_quad_cell( const_QuadsPtr qtree, int * hf )
{
    int id = threadIdx.x + blockIdx.x*blockDim.x;
    if ( id >= qtree.size )
        return;
    uint2 range = qtree.seg_range[id];
    hf[id] = range.y > 0;
}

__device__
inline
int round_to_m16( int a )
{
    if ( a & 0xF ) {
        a &= ~0xF;
        a += 16;
    }
    return a;
}

#if 0
__global__
void gpu_gather_info( const_QuadsPtr qtree, const_SegmentsPtr segs, QuadInfo *qinfo, int *hf, short2 *mem_config, int smsize,
                      bool *must_call_complex_kernel)
{
    int id = threadIdx.x + blockIdx.x*blockDim.x;
    if ( id >= qtree.size )
        return;
    uint2 range = qtree.seg_range[id];
    int4 ch = qtree.children[id];
    //if ( range.y == 0 ) return;
    if ( ch.x != -1 || ch.y != -1 || ch.z != -1 || ch.w != -1 )
        return;
    float2 min = qtree.min[id];
    float2 max = qtree.max[id];
    float2 tmp = max-min;
    QuadInfo qi;
    qi.num_segments = range.y;
    //qi.data_size = 0;
    int data_size = 0;
    for (int k=0; k<range.y; ++k) {
        int sid = range.x + k;
        SegType type = segs.type[sid];
        data_size += find_segment_size( type );
        if ( type == SEG_PATH ) {
            const PathData *pdata = segs.data<PathData>(sid);
            Paint::Type ptype = static_cast<Paint::Type>(pdata->paint_type);
            data_size += find_paint_size( ptype );
        }
    }
    int pos = hf[id];
    qi.data_size = data_size; //may explode if > 65535
    qinfo[pos] = qi;
    int index_mem = round_to_m16(sizeof(SegmentIndex))*qi.num_segments;
    int geom_mem = round_to_m16( data_size );
    if ( index_mem + geom_mem >= smsize )
    {
        mem_config[id] = make_short2( -1, -1 );
        *must_call_complex_kernel = true;
    }
    else 
        mem_config[id] = make_short2( index_mem, geom_mem );
}
#endif


template< class W >
void rasterize_warp_template(const W& warp, const float4 &bg_color,
               const Quads &qtree,
               const Segments &segs,
               unsigned nsamples,
               gpu::image_ptr<float4> out,
               unsigned flags,
               unsigned use_bs3)
{
    int width = out.width(), height = out.height();
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("rasterize"));
#if 0
    float ang = M_PI/8;
    Transform xform2 = translate(w/2.f, h/2.f)
                        *rotate(ang)
                        *translate(w/-2.f,h/-2.f)
                      ;
#endif
    //gpu::copy_to_symbol(c_xform, inverse(xform));
    gpu::copy_to_symbol(c_bgcolor, bg_color);
    gpu::copy_to_symbol(d_num_samples, nsamples);
    std::vector<float2> pattern = get_blue_noise_pattern(nsamples,.5);
    gpu::copy_to_symbol(d_sampling_pattern, pattern);
    {
        std::vector<float> patx, paty;
        for(unsigned i=0; i<pattern.size(); ++i)
        {
            patx.push_back(pattern[i].x);
            paty.push_back(pattern[i].y);
        }
        gpu::copy_to_symbol(d_sampling_pattern_x, patx);
        gpu::copy_to_symbol(d_sampling_pattern_y, paty);
    }
    //
    gpu::vector<uint2> quad_pixel(width*height*32);
    gpu::value<unsigned> next_empty_slot(0);
    //
    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("traverse qtree"));
        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("qtree traversal"));
            dim3 bdim(32, BH),
                 //gdim((width+3)/4, (height+BH-1)/BH);
                 gdim((width+TW-1)/TW, (height+TH-1)/TH);
            // This kernel fills pixel_quad with the index of the cells that are
            // touched by samples from some unit area.
            // quads_per_pixel holds the number of cell for each unit area.
            // TODO: change the naming from pixel to unit area.
            cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);
            gpu_traverse_qtree32<W><<<gdim,bdim>>>(warp, &qtree, &quad_pixel, &next_empty_slot, width, height);
            quad_pixel.resize(next_empty_slot);
            // persistent threads
            //int nw = (width+3)/4;
            //int nh = (height+3)/4;
            //TraverseQtree<W> traverse_qtree32(warp, &qtree, &pixel_quad, &quads_per_pixel, width, height, nw, nh);
            //persistent_threads<MAX_BLOCKS_PER_SM, 16>(traverse_qtree32, nw*nh);
            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
        }
#if 0
        {
            std::vector<unsigned> aux, aux2;
            pixel_quad.copy_to_host(aux);
            quads_per_pixel.copy_to_host(aux2);
            for(int i=0; i<aux2.size(); ++i)
            {
                printf("(%d,%d) %2d: ",i%w,i/w,aux2[i]);
                for(int j=0; j<aux2[i]; ++j)
                    printf("%d ",aux[i*MAX_QUADS_PER_PIXEL+j]);
                printf("\n");
            }
        }
#endif

        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("sort quad_pixel"));
            thrust::sort( quad_pixel.begin(), quad_pixel.end() );
        }
    }
    gpu::vector<unsigned> pkt_base, pkt_size;
    gpu::image<float> channels[4];
    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("collect samples"));
        {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("create quad_pixel packets"));
            create_packets_quad_unitarea( pkt_base, pkt_size, quad_pixel, 32);
            /*
            {
                std::vector<unsigned> aux_base, aux_size;
                pkt_base.copy_to_host(aux_base);
                pkt_size.copy_to_host(aux_size);
                for(int i=0; i<aux_size.size(); ++i) {
                    fprintf(stderr, "(%d,%d)\n: ", aux_base[i], aux_size[i]);
                }
            }
            // */
        }
        gpu::vector<SegmentIndex> segment_index;
        segment_index.resize( segs.size() );
        {//convert sizes into flags
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("convert sizes into flags"));
            cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);
            int bsize = 64;
            int gsize = ( segs.size() + bsize - 1 )/bsize;
            dim3 grid( gsize ), block( bsize );
            gpu_convert_segment_flags<<<grid, block>>>( &segs, &segment_index );
            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
        }
        const int bsize = 4;
        const int smsize = 2*1024;
        bool call_complex_kernel = true;
#if 0
        gpu::vector<QuadInfo> qinfo;
        //TODO: sort this vector to run the right kernel only on the
        // right packets.
        gpu::vector<short2> mem_config;
        gpu::vector<int> hf;
        {//gather info
            int bsize = 64;
            int gsize = ( segs.size() + bsize - 1 )/bsize;
            dim3 grid( gsize ), block( bsize );
            cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("gather info"));
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("find leaf cells"));
                hf.resize( qtree.size() + 1 );
                gpu_mark_leaf_quad_cell<<<grid, block>>>( &qtree, hf.data() );
                //
                gpu::CUB cub;
                cub.exclusive_scan( hf, hf );
                qinfo.resize( hf.back() );
            }
            gpu::value<bool> d_call_complex_kernel(false);
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("compute and store"));
                mem_config.resize( qtree.size() );
                gpu_gather_info<<<grid, block>>>( &qtree, &segs, qinfo.data(), hf.data(),
                        mem_config.data(), smsize, &d_call_complex_kernel);
                call_complex_kernel = d_call_complex_kernel;
            }
            cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
            /*
            if ( flags & 4 )
                dump_info( qinfo, flags );
            if ( flags & 8 )
                dump_info( qinfo, flags );
                */
        }
#endif
        dim3 grid( pkt_base.size() ), block( 32, bsize );
        if ( use_bs3 ) {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("raster bs3"));
            {
                std::vector<float> bspline3 = calc_bspline3_data(pattern);
                std::vector<float4> empty_samples(KS*KS);
                #pragma unroll
                for(int i=0; i<KS*KS; ++i) {
                    empty_samples[i] = srgb2lrgb(bg_color)*bspline3[i];
                }
                gpu::copy_to_symbol(c_empty_samples, empty_samples);
            }
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("resize channedls bs3"));
                for(int i=0; i<4; ++i) {
                    channels[i].resize(width+IMG_BORDER*2,height+IMG_BORDER*2);
                    set_zero(channels[i]);
                }
            }
            {
                Context ctx( &quad_pixel, &qtree.seg_range,
                              segs, qtree, &segment_index,
                              width, height, IMG_BORDER,
                              channels );
                const int maxpktsize = 32;
                cudaDeviceSetCacheConfig(cudaFuncCachePreferShared);
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple bs3"));
                    gpu_raster_warp<W, Bs3, bsize, smsize><<<grid, block>>>( warp, ctx, pkt_base.data(),
                            pkt_size.data(), pkt_base.size(), qtree.mem_config.data() );
                }
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi bs3"));
                    if(call_complex_kernel) {
                        gpu_raster_warp_geom<W, Bs3, bsize, smsize, maxpktsize><<<grid, block>>>( warp, ctx, pkt_base.data(),
                                pkt_size.data(), pkt_base.size(), qtree.mem_config.data() );
                    }
                }
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("outside bs3"));
                    gpu_raster_warp_out_quad<W, Bs3, bsize, smsize><<<grid, block>>>( warp, ctx, pkt_base.data(),
                            pkt_size.data(), pkt_base.size(), qtree.mem_config.data() );
                }
                cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
            }
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("do cardinal bspline3"));
                // calculate cubic b-spline weights
                gpu::math::Vector<float,1+1> bspline3_weights;
                {
                    float a = 2.f-std::sqrt(3.0f);
                    bspline3_weights[0] = 1+a;
                    bspline3_weights[1] = a;
                }
                recursive_filter_5_setup(out.width(),
                                         out.height(),
                                         out.rowstride(),
                                         bspline3_weights, gpu::CLAMP_TO_EDGE, 1);
                for(int i=0; i<4; ++i) {
                    recursive_filter_5(channels[i], IMG_BORDER);
                }
                gpu::recursive_filter_5_free();
            }
        }
        else {
            gpu::scoped_timer_stop sts(gpu::timers.gpu_add("raster normal"));
            {
                gpu::scoped_timer_stop sts(gpu::timers.gpu_add("resize channels"));
                for(int i=0; i<4; ++i) {
                    channels[i].resize(width,height);
                    set_zero(channels[i]);
                }
            }
            {
                Context ctx( &quad_pixel, &qtree.seg_range,
                              segs, qtree, &segment_index,
                              width, height, 0,
                              channels );
                const int maxpktsize = 32;
                cudaDeviceSetCacheConfig(cudaFuncCachePreferShared);
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("simple box"));
                    gpu_raster_warp<W, Box, bsize, smsize><<<grid, block>>>( warp, ctx, pkt_base.data(),
                            pkt_size.data(), pkt_base.size(), qtree.mem_config.data() );
                }
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("multi box"));
                    if(call_complex_kernel) {
                        gpu_raster_warp_geom<W, Box, bsize, smsize, maxpktsize><<<grid, block>>>( warp, ctx, pkt_base.data(),
                                pkt_size.data(), pkt_base.size(), qtree.mem_config.data() );
                    }
                }
                {
                    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("outside box"));
                    gpu_raster_warp_out_quad<W, Box, bsize, smsize><<<grid, block>>>( warp, ctx, pkt_base.data(),
                            pkt_size.data(), pkt_base.size(), qtree.mem_config.data() );
                }
                cudaDeviceSetCacheConfig(cudaFuncCachePreferNone);
            }
        }
    }
#if 1
    {
        gpu::scoped_timer_stop sts(gpu::timers.gpu_add("compose channels into output image"));
        compose(out, channels[0], channels[1], channels[2], channels[3]);
    }
#endif
}
