#ifndef GPURAST_RAYCAST_H
#define GPURAST_RAYCAST_H

#include <gpu/box.h>
#include "implicit.h"
#include <gpu/util.h>
#include "curve3.h"
#include "icurve1.h"
#include "icurve2.h"
#include "icurve3.h"
#include "icurve_arc.h"

__device__ inline bool hit_bbox(const Curve<1> &B, const float2 &ray)/*{{{*/
{
    // below bbox?
    if(ray.y < min(B[0].y, B[1].y))
        return false;

    // above bbox?
    if(ray.y > max(B[0].y, B[1].y))
        return false;

    // after bbox?
    if(ray.x > max(B[0].x, B[1].x))
        return false;

    return true;
}/*}}}*/

__device__ inline bool hit_bbox(const Curve<2> &B, const float2 &ray)/*{{{*/
{
    // below bbox?
    if(ray.y < min(B[0].y, min(B[1].y, B[2].y)))
        return false;

    // above bbox?
    if(ray.y > max(B[0].y, max(B[1].y, B[2].y)))
        return false;

    // after bbox?
    if(ray.x > max(B[0].x, max(B[1].x, B[2].x)))
        return false;

    return true;
}/*}}}*/

__device__ inline bool hit_bbox(const Curve<3> &B, const float2 &ray)/*{{{*/
{
    // below bbox?
    if(ray.y < min(B[0].y, min(B[1].y, min(B[2].y, B[3].y))))
        return false;

    // above bbox?
    if(ray.y > max(B[0].y, max(B[1].y, max(B[2].y, B[3].y))))
        return false;

    // after bbox?
    if(ray.x > max(B[0].x, max(B[1].x, max(B[2].x, B[3].x))))
        return false;

    return true;
}/*}}}*/

// ray intersection hit test ------------------------------------

__device__ inline int hit(Curve<1> B, float2 ray)/*{{{*/
{
    if(!hit_bbox(B, ray))
        return 0;

    // horizontal line? no intersection
    if(B[0].y == B[1].y)
        return 0;

    // order control points so segment goes up
    if(B[1].y < B[0].y) // going down?
        gpu::swap(B[0], B[1]);

    // if hit endpoint, ignore unless top endpoint
    if(ray.y == B[0].y)
        return 0;
    else if(ray.y == B[1].y && ray.x <= B[1].x)
        return 1;

    // calculate x coordinate of intersection

    float dy = B[1].y - B[0].y, // strictly > 0
          dx = B[1].x - B[0].x,
          ny = ray.y - B[0].y;

    // its a hit when ray.x <= x. To avoid
    // a division, we multiply both sides by dy.

    float xdy = B[0].x*dy + dx*ny,
          rdy = ray.x*dy;

    return xdy >= rdy;
}/*}}}*/

// curve must be monotonic
__device__ inline int hit(const Curve<2> &B, float2 ray)/*{{{*/
{
    if(!hit_bbox(B, ray))
        return 0;

    float a = B[2].y - 2*B[1].y + B[0].y,
          b = 2*(B[1].y - B[0].y),
          c = -ray.y + B[0].y;

    float r[2];
    int nroots = gpu::quadratic_solve(a,b,c,r);

    if(nroots == 2 && r[1] == r[0])
        --nroots;

    bool down = B[2].y < B[0].y;

    for(int i=0; i<nroots; ++i)
    {
        // if hit endpoint, ignore unless top endpoint
        if(r[i] == 0 && !down)
            continue;
        if(r[i] == 1 && down)
            continue;

        if(0 <= r[i] && r[i] <= 1)
        {
            float x = eval(B, r[i]).x;

            if(x >= ray.x)
                return 1;
        }
    }

    return 0;
}/*}}}*/

// seg must be monotonic
__device__ inline int hit(const Curve<3> &B, float2 ray)/*{{{*/
{
    if(!hit_bbox(B, ray))
        return 0;

    float a = -B[0].y + 3*B[1].y - 3*B[2].y + B[3].y,
          b = 3*(B[0].y - 2*B[1].y + B[2].y),
          c = 3*(-B[0].y + B[1].y),
          d = B[0].y;

    d -= ray.y;

    bool down = B[3].y < B[0].y;

    float t[3];

    int n = gpu::cubic_solve(a,b,c,d,t);

    if(n == 3)
    {
        if(t[0] == t[2])
            --n;
        else if(t[1] == t[2])
            --n;
        else if(t[0] == t[1])
        {
            t[1] = t[2];
            --n;
        }
    }

    if(n == 2)
    {
        if(t[0] == t[1])
            --n;
    }

    for(int i=0; i<n; ++i)
    {
        // if hit endpoint, ignore unless top endpoint
        if(t[i]==0 && !down)
            continue;
        if(t[i]==1 && down)
            continue;

        // since B is monotonic, we can bail out early on first hit
        if(0 <= t[i] && t[i] <= 1 && eval(B, t[i]).x >= ray.x)
            return 1;
    }

    return 0;
}/*}}}*/


#endif
