#ifndef GPURASTER_SAMPLING_H
#define GPURASTER_SAMPLING_H

#include "config.h"
#include "implicit.h"
#include "paint.h"
#include "raycast.h"

inline __device__
int get_intersections(const SegmentIndex *si,
                        const unsigned char *data,
                        int &idseg, int idseg_end,
                        float2 coord)
{
    int isects = 0;
    for(; idseg < idseg_end; ++idseg) {
        ImplicitSegment<> seg;
        seg.type = si[idseg].get_type();
        seg.data = reinterpret_cast<const void*>( data + si[idseg].idx_data );
        seg.flags = si[idseg].flags;
        //
        // finished processing this path?
        if(seg.type == SEG_PATH) {
            isects += si[idseg].winding;
            break;
        }
        const ImplicitCurveBase& icurve = *seg.as<1>().B();
        const gpu::Box &box = icurve.bbox;
        int val = icurve.going_up() ? 1 : -1;
        float2 b = icurve.begpoint(), e = icurve.endpoint();
        int opposite_side = ( ccw(coord, b, e) < 0 != icurve.geom_to_left() );
        int check = box.hit(coord);
        if ( check==0 ) { //above, below, to the right
            ;
        }
        else if ( check == 1) { //to the left
            isects += val;
        }
        else if ( opposite_side ) {
            if( icurve.out_is_left() ) isects += val;
        }
        else if ( seg.type == SEG_LINEAR ) {
            // sample is on the same side of geometry (left by default for linear)
            // so, if the curve is going up, the sample is on the left side (word).
            if ( icurve.going_up() ) isects += val;
        }
        else if ( seg.type == SEG_QUADRATIC ) {
            if ( seg.as<2>().B()->eval(coord) < 0 ) isects += val;
        }
        else if ( seg.type == SEG_ARC ) {
            if ( seg.as<SEG_ARC>().B()->eval(coord) < 0 ) isects += val;
        }
        else if ( seg.type == SEG_CUBIC ) {
            const ImplicitCurve<3>& ic3 = *seg.as<3>().B();
            bool sample_to_the_left = ccw(coord, b, e) < 0;
            if( (ccw(coord, e, ic3.E) < 0) == sample_to_the_left
                    && (ccw(coord, ic3.E, b) < 0) == sample_to_the_left) {
                if ( seg.as<3>().B()->eval(coord) < 0 ) isects += val;
            }
            else {
                // is outside
                if(!ic3.out_is_left())
                    isects += val;
            }
        }
        // we can call as ImplicitSegment<1> since all
        // ImplicitCurves have a bbox as first attribute.
        if(seg.as<1>().is_shortcut())
            isects += seg.as<1>().shortcut_hit(coord);
    }
#ifndef NDEBUG
    /*
    if(idseg == idseg_end) {
        assert(false);
        --idseg;
    }
    */
#endif
    //isects += si[idseg].winding;
    return isects;
}

inline __device__
int get_intersections(const const_SegmentsPtr &segs,
                      int &idseg, int idseg_end,
                      float2 coord)
{
    int isects = 0;

    for(; idseg < idseg_end; ++idseg)
    {
        ImplicitSegment<> seg;
        seg.type = segs.type[idseg];
        seg.data = segs.data<void>(idseg);
        seg.flags = segs.flags[idseg];

        // finished processing this path?
        if(seg.type == SEG_PATH)
            break;

        switch(seg.type)
        {
        case SEG_LINEAR:
            isects += seg.as<1>().B()->hit(coord);
            break;
        case SEG_QUADRATIC:
            isects += seg.as<2>().B()->hit(coord);
            break;
        case SEG_CUBIC:
            isects += seg.as<3>().B()->hit(coord);
            break;
        case SEG_ARC:
            isects += seg.as<SEG_ARC>().B()->hit(coord);
            break;
        default:
            assert(false);
            continue;
        }

        if(seg.as_curve().is_shortcut())
            isects += seg.as_curve().shortcut_hit(coord);
    }

#ifndef NDEBUG
    if(idseg == idseg_end)
    {
        assert(false);
        --idseg;
    }
#endif

    isects += segs.winding[idseg];

    return isects;
}

struct ClipData
{
    union
    {
        struct
        {
            bool skip_by_commit : 1,
                 skipping : 1;
            int cur_depth: 15,
                stop_depth : 15;
        };
        unsigned data;
    };
};

__device__
inline float4 vgsample_shared_clipping(const SegmentIndex *si, const unsigned char* data,
                       int idseg, int idseg_end,
                       float2 coord, float4 bgcolor)
{
#if ENABLE_CLIPPING
    ClipData clip;
    clip.data = 0;
    clip.stop_depth = -1;
#endif
#if FRONT_TO_BACK_RENDERING
    float4 sample = make_float4(0,0,0,0);
#else
    float4 sample = bgcolor;
#endif
    for(;idseg < idseg_end; ++idseg) {
#if ENABLE_CLIPPING
        SegType type = si[idseg].get_type();
        if(clip.skipping &&
                ( (type!=SEG_POP_CLIP && type!=SEG_POP_FILL && type!=SEG_COMMIT)
                  || clip.cur_depth > clip.stop_depth) ) {
            switch(type) {
                case SEG_PUSH:
                    ++clip.cur_depth;
                    break;
                case SEG_POP_FILL:
                case SEG_POP_CLIP:
                    --clip.cur_depth;
                    break;
                case SEG_COMMIT:
                    break;
                default:
                    for(; idseg < idseg_end; ++idseg) {
                        if(si[idseg].get_type() == SEG_PATH)
                            break;
                    }
            }
        }
        else if(type == SEG_PUSH) {
            ++clip.cur_depth;
        }
        else if(type == SEG_POP_FILL) {
            --clip.cur_depth;
            if(clip.cur_depth == 0)
                break;
        }
        else if(type == SEG_POP_CLIP) {
            --clip.cur_depth;
            if(clip.skipping) {
                if(clip.skip_by_commit)
                    clip.skipping = false;
                else
                    clip.stop_depth = clip.cur_depth;
            }
        }
        else if(type == SEG_COMMIT) {
            if(!clip.skipping) {
                clip.skipping = true;
                clip.skip_by_commit = true;
                clip.stop_depth = clip.cur_depth;
            }
            else
                clip.skipping = false;
        }
        else
#endif
        {
            int isects = get_intersections(si, data, idseg, idseg_end, coord);
            assert(idseg < idseg_end);
            const PathData *pdata = reinterpret_cast<const PathData*>( data + si[idseg].idx_data );
            if(pdata->intersects(isects)) {
                // it's a fill path?
                if(const Paint *paint = pdata->paint) {
                    switch(paint->type) {
                    case Paint::SOLID:
                        sample = paint->as<PaintSolid>()(coord, sample);
                        break;
                    case Paint::SOLID_ALPHA:
                        sample = paint->as<PaintSolidAlpha>()(coord, sample);
                        break;
                    case Paint::GRAD_LINEAR:
                        sample = paint->as<PaintLinearGrad>()(coord, sample);
                        break;
                    case Paint::GRAD_RADIAL:
                        sample = paint->as<PaintRadialGrad>()(coord, sample);
                        break;
                    case Paint::TEXTURE:
                        sample = paint->as<PaintTexture>()(coord, sample);
                        break;
                    case Paint::NONE:
                    default:
                        break;
                    }
#if FRONT_TO_BACK_RENDERING
                    if(sample.w > 0.99999f)
                        idseg = idseg_end; // skip to end of stream
#endif
                }
#if ENABLE_CLIPPING
                else { // it's a clip path
                    if(!clip.skipping) {
                        clip.skipping = true;
                        clip.skip_by_commit = false;
                        clip.stop_depth = clip.cur_depth;
                    }
                }
#endif
            } // intersects?
        }
    } // for each path
#if FRONT_TO_BACK_RENDERING
    if(sample.w < 1)
        sample = blend(sample, premultiply(bgcolor));
#endif
    return sample;
}

__device__
inline void vgsample_shared_clipping(const SegmentIndex *si, const unsigned char* data,
                       int idseg, int idseg_end,
                       float2 coord, float4 bgcolor,
                       float4& sample, int& winding, ClipData& clip)
{
#if FRONT_TO_BACK_RENDERING
    if ( sample.w > 0.9999f ) {
        return;
    }
#endif
    for(;idseg < idseg_end; ++idseg) {
#if ENABLE_CLIPPING
        SegType type = si[idseg].get_type();
        if(clip.skipping &&
                ( (type!=SEG_POP_CLIP && type!=SEG_POP_FILL && type!=SEG_COMMIT)
                  || clip.cur_depth > clip.stop_depth) ) {
            switch(type) {
                case SEG_PUSH:
                    ++clip.cur_depth;
                    break;
                case SEG_POP_FILL:
                case SEG_POP_CLIP:
                    --clip.cur_depth;
                    break;
                case SEG_COMMIT:
                    break;
                default:
                    for(; idseg < idseg_end; ++idseg) {
                        if(si[idseg].get_type() == SEG_PATH)
                            break;
                    }
            }
        }
        else if(type == SEG_PUSH) {
            ++clip.cur_depth;
        }
        else if(type == SEG_POP_FILL) {
            --clip.cur_depth;
            if(clip.cur_depth == 0)
                break;
        }
        else if(type == SEG_POP_CLIP) {
            --clip.cur_depth;
            if(clip.skipping) {
                if(clip.skip_by_commit)
                    clip.skipping = false;
                else
                    clip.stop_depth = clip.cur_depth;
            }
        }
        else if(type == SEG_COMMIT) {
            if(!clip.skipping) {
                clip.skipping = true;
                clip.skip_by_commit = true;
                clip.stop_depth = clip.cur_depth;
            }
            else
                clip.skipping = false;
        }
        else
#endif
        {
            winding += get_intersections(si, data, idseg, idseg_end, coord);
            if ( idseg < idseg_end && si[idseg].get_type() == SEG_PATH ) {
                assert(idseg < idseg_end);
                const PathData *pdata = reinterpret_cast<const PathData*>( data + si[idseg].idx_data );
                if(pdata->intersects(winding)) {
                    // it's a fill path?
                    if(const Paint *paint = pdata->paint) {
                        switch(paint->type) {
                        case Paint::SOLID:
                            sample = paint->as<PaintSolid>()(coord, sample);
                            break;
                        case Paint::SOLID_ALPHA:
                            sample = paint->as<PaintSolidAlpha>()(coord, sample);
                            break;
                        case Paint::GRAD_LINEAR:
                            sample = paint->as<PaintLinearGrad>()(coord, sample);
                            break;
                        case Paint::GRAD_RADIAL:
                            sample = paint->as<PaintRadialGrad>()(coord, sample);
                            break;
                        case Paint::TEXTURE:
                            sample = paint->as<PaintTexture>()(coord, sample);
                            break;
                        case Paint::NONE:
                        default:
                            break;
                        }
#if FRONT_TO_BACK_RENDERING
                        // TODO: make this make any sense...
                        // should jump to begining of the last
                        // path on the stream
                        if(sample.w > 0.99999f)
                            idseg = idseg_end; // skip to end of stream
#endif
                    }
#if ENABLE_CLIPPING
                    else { // it's a clip path
                        if(!clip.skipping) {
                            clip.skipping = true;
                            clip.skip_by_commit = false;
                            clip.stop_depth = clip.cur_depth;
                        }
                    }
#endif
                }
                winding = 0;
            } // intersects?
        }
    } // for each path
}

__device__
inline void vgsample_shared(const SegmentIndex *si, const unsigned char* data,
                       int idseg, int idseg_end,
                       float2 coord, float4 bgcolor,
                       float4& sample, int& winding)
{
#if FRONT_TO_BACK_RENDERING
    if ( sample.w > 0.9999f ) {
        return;
    }
#endif
    for(;idseg < idseg_end; ++idseg) {
        winding += get_intersections(si, data, idseg, idseg_end, coord);
        if ( idseg < idseg_end && si[idseg].get_type() == SEG_PATH ) {
            assert(idseg < idseg_end);
            const PathData *pdata = reinterpret_cast<const PathData*>( data + si[idseg].idx_data );
            if(pdata->intersects(winding)) {
                // it's a fill path?
                if(const Paint *paint = pdata->paint) {
                    switch(paint->type) {
                    case Paint::SOLID:
                        sample = paint->as<PaintSolid>()(coord, sample);
                        break;
                    case Paint::SOLID_ALPHA:
                        sample = paint->as<PaintSolidAlpha>()(coord, sample);
                        break;
                    case Paint::GRAD_LINEAR:
                        sample = paint->as<PaintLinearGrad>()(coord, sample);
                        break;
                    case Paint::GRAD_RADIAL:
                        sample = paint->as<PaintRadialGrad>()(coord, sample);
                        break;
                    case Paint::TEXTURE:
                        sample = paint->as<PaintTexture>()(coord, sample);
                        break;
                    case Paint::NONE:
                    default:
                        break;
                    }
#if FRONT_TO_BACK_RENDERING
                    // TODO: make this make any sense...
                    // should jump to begining of the last
                    // path on the stream
                    if(sample.w > 0.99999f)
                        idseg = idseg_end; // skip to end of stream
#endif
                }
            }
            winding = 0;
        } // intersects?
    } // for each path
}

__device__
inline float4 vgsample_shared(const SegmentIndex *si, const unsigned char* data,
                       int idseg, int idseg_end,
                       float2 coord, float4 bgcolor)
{
#if FRONT_TO_BACK_RENDERING
    float4 sample = make_float4(0,0,0,0);
#else
    float4 sample = bgcolor;
#endif
    for(;idseg < idseg_end; ++idseg) {
        int isects = get_intersections(si, data, idseg, idseg_end, coord);
        assert(idseg < idseg_end);
        const PathData *pdata = reinterpret_cast<const PathData*>( data + si[idseg].idx_data );
        if(pdata->intersects(isects)) {
            // it's a fill path?
            if(const Paint *paint = pdata->paint) {
                switch(paint->type) {
                case Paint::SOLID:
                    sample = paint->as<PaintSolid>()(coord, sample);
                    break;
                case Paint::SOLID_ALPHA:
                    sample = paint->as<PaintSolidAlpha>()(coord, sample);
                    break;
                case Paint::GRAD_LINEAR:
                    sample = paint->as<PaintLinearGrad>()(coord, sample);
                    break;
                case Paint::GRAD_RADIAL:
                    sample = paint->as<PaintRadialGrad>()(coord, sample);
                    break;
                case Paint::TEXTURE:
                    sample = paint->as<PaintTexture>()(coord, sample);
                    break;
                case Paint::NONE:
                default:
                    break;
                }
#if FRONT_TO_BACK_RENDERING
                if(sample.w > 0.99999f)
                    idseg = idseg_end; // skip to end of stream
#endif
            }
        } // intersects?
    } // for each path
#if FRONT_TO_BACK_RENDERING
    if(sample.w < 1)
        sample = blend(sample, premultiply(bgcolor));
#endif
    return sample;
}

/*
__device__
inline float4 vgsample_shared(const SegmentIndex *si, const unsigned char* data,
                       int idseg, int idseg_end,
                       float2 coord, float4 bgcolor)
{
#if FRONT_TO_BACK_RENDERING
    float4 sample = make_float4(0,0,0,0);
#else
    float4 sample = bgcolor;
#endif
    int winding = 0;
    return vgsample_shared( si, data, idseg, idseg_end, coord, bgcolor, sample, winding );
}
*/

__device__
inline float4 vgsample(const const_SegmentsPtr &segs,
                       int idseg, int idseg_end,
                       float2 coord, float4 bgcolor)
{
#if ENABLE_CLIPPING
    ClipData clip;
    clip.data = 0;
    clip.stop_depth = -1;
#endif

    float4 sample;

#if FRONT_TO_BACK_RENDERING
    sample = make_float4(0,0,0,0);
#else
    sample = bgcolor;
#endif

    for(;idseg < idseg_end; ++idseg)
    {
        SegType type = segs.type[idseg];

#if ENABLE_CLIPPING
        if(clip.skipping && ((type!=SEG_POP_CLIP && type!=SEG_POP_FILL && type!=SEG_COMMIT) || clip.cur_depth > clip.stop_depth))
        {
            switch(type)
            {
            case SEG_PUSH:
                ++clip.cur_depth;
                break;
            case SEG_POP_FILL:
            case SEG_POP_CLIP:
                --clip.cur_depth;
                break;
            case SEG_COMMIT:
                break;
            default:
                for(; idseg < idseg_end; ++idseg)
                {
                    if(segs.type[idseg] == SEG_PATH)
                        break;
                }
            }
        }
        else if(type == SEG_PUSH)
            ++clip.cur_depth;
        else if(type == SEG_POP_FILL)
        {
            --clip.cur_depth;
            if(clip.cur_depth == 0)
                break;
        }
        else if(type == SEG_POP_CLIP)
        {
            --clip.cur_depth;
            if(clip.skipping)
            {
                if(clip.skip_by_commit)
                    clip.skipping = false;
                else
                    clip.stop_depth = clip.cur_depth;
            }
        }
        else if(type == SEG_COMMIT)
        {
            if(!clip.skipping)
            {
                clip.skipping = true;
                clip.skip_by_commit = true;
                clip.stop_depth = clip.cur_depth;
            }
            else
                clip.skipping = false;
        }
        else
#endif
        {
            int isects = get_intersections(segs, idseg, idseg_end, coord);

            assert(idseg < idseg_end);
            assert(segs.type[idseg] == SEG_PATH);
            const PathData *pdata = segs.data<PathData>(idseg);
            assert(pdata != NULL);

            if(pdata->intersects(isects))
            {
                // it's a fill path?
                if(const Paint *paint = pdata->paint)
                {
                    switch(paint->type)
                    {
                    case Paint::SOLID:
                        sample = paint->as<PaintSolid>()(coord, sample);
                        break;
                    case Paint::SOLID_ALPHA:
                        sample = paint->as<PaintSolidAlpha>()(coord, sample);
                        break;
                    case Paint::GRAD_LINEAR:
                        sample = paint->as<PaintLinearGrad>()(coord, sample);
                        break;
                    case Paint::GRAD_RADIAL:
                        sample = paint->as<PaintRadialGrad>()(coord, sample);
                        break;
                    case Paint::TEXTURE:
                        sample = paint->as<PaintTexture>()(coord, sample);
                        break;
                    case Paint::NONE:
                    default:
                        break;
                    }

#if FRONT_TO_BACK_RENDERING
                    if(sample.w > 0.99999f)
                        idseg = idseg_end; // skip to end of stream
#endif
                }
#if ENABLE_CLIPPING
                else // it's a clip path
                {
                    if(!clip.skipping)
                    {
                        clip.skipping = true;
                        clip.skip_by_commit = false;
                        clip.stop_depth = clip.cur_depth;
                    }
                }
#endif
            } // intersects?
        } // seg type switch
    } // for each path

#if FRONT_TO_BACK_RENDERING
    if(sample.w < 1)
        sample = blend(sample, premultiply(bgcolor));
#endif

    return sample;
}

#endif
