#include <algorithm>
#include "config.h"
#include "segments.h"
#include "icurve1.h"
#include "icurve2.h"
#include "icurve3.h"
#include "icurve_arc.h"
#include "curve_arc.h"
#include "implicit.h"
#include "preprocess.h"
#include <gpu/timer.h>
#include <thrust/gather.h>
#include <thrust/scan.h>
#include <thrust/iterator/reverse_iterator.h>

void convert_segments(const fd::FontDump &fd,/*{{{*/
                      std::vector<Segment<> > &out_segs,
                      std::vector<unsigned> &out_path_attrib,
                      std::vector<char> &out_curve_data,
                      const gpu::vector<const Paint *> &attrib_paints)
{
    gpu::scoped_timer_stop sts(gpu::timers.cpu_add("convert paths"));

    size_t data_size = 0;

    out_path_attrib.resize(fd.paths.size());

    for(int p=0; p < fd.paths.size(); ++p)
    {
        const fd::Path &path = fd.paths[p];

        out_path_attrib[p] = path.attrib;

        // the first segment of a path is the path segment, reserve space for
        // its data structure.
        data_size += sizeof(PathData);

        for(int c=0; c < path.contours.size(); ++c)
        {
            const fd::Contour &contour = path.contours[c];

            for(int s=0; s<contour.segments.size(); ++s)
            {
                switch(contour.segments[s])
                {
                case fd::SEGMENT_LINEAR:
                    data_size += sizeof(Curve<1>);
                    break;
                case fd::SEGMENT_QUADRATIC:
                    data_size += sizeof(Curve<2>);
                    break;
                case fd::SEGMENT_CUBIC:
                    data_size += sizeof(Curve<3>);
                    break;
                case fd::SEGMENT_ARC:
                    data_size += sizeof(Curve<SEG_ARC>);
                    break;
                }
            }
        }
    }

    out_curve_data.reserve(data_size);

    for(int p=0; p < fd.paths.size(); ++p)
    {
        const fd::Path &path = fd.paths[p];

        for(int c=0; c < path.contours.size(); ++c)
        {
            const fd::Contour &contour = path.contours[c];

            const std::vector<math::Point> &cpts = contour.points;

            const math::Point *curpt = &cpts[0]-1;

            for(int s=0; s<contour.segments.size(); ++s)
            {
                switch(contour.segments[s])
                {
                case fd::SEGMENT_LINEAR:
                    {
                        Segment<> seg;
                        seg.type = SEG_LINEAR;

                        // seg.data must point to the first point added in the
                        // for loop
                        seg.data = &out_curve_data.back()+1;

                        out_curve_data.resize(out_curve_data.size() + sizeof(Curve<1>));

                        Curve<1> &curve = *seg.as<1>().B;

                        for(int i=0; i<seg.size(); ++i)
                            curve[i] = make_float2(curpt[i].x, curpt[i].y);

                        out_segs.push_back(seg);
                    }
                    break;
                case fd::SEGMENT_QUADRATIC:
                    {
                        Segment<> seg;
                        seg.type = SEG_QUADRATIC;

                        // seg.data must point to the first point added in the
                        // for loop
                        seg.data = &out_curve_data.back()+1;

                        out_curve_data.resize(out_curve_data.size() + sizeof(Curve<2>));

                        Curve<2> &curve = *seg.as<2>().B;

                        for(int i=0; i<seg.size(); ++i)
                            curve[i] = make_float2(curpt[i].x, curpt[i].y);

                        out_segs.push_back(seg);
                    }
                    break;
                case fd::SEGMENT_CUBIC:
                    {
                        Segment<> seg;
                        seg.type = SEG_CUBIC;

                        // seg.data must point to the first point added in the
                        // for loop
                        seg.data = &out_curve_data.back()+1;

                        out_curve_data.resize(out_curve_data.size() + sizeof(Curve<3>));

                        Curve<3> &curve = *seg.as<3>().B;

                        for(int i=0; i<seg.size(); ++i)
                            curve[i] = make_float2(curpt[i].x, curpt[i].y);

                        out_segs.push_back(seg);
                    }
                    break;

                case fd::SEGMENT_ARC:
                    {
                        Segment<> seg;
                        seg.type = SEG_ARC;

                        // seg.data must point to the first point added in the
                        // for loop
                        seg.data = &out_curve_data.back()+1;

                        out_curve_data.resize(out_curve_data.size() + sizeof(Curve<SEG_ARC>));

                        Curve<SEG_ARC> &curve = *seg.as<SEG_ARC>().B;

                        curve[0] = make_float2(curpt[0].x, curpt[0].y);
                        curve[1] = make_float2(curpt[1].x, curpt[1].y);
                        curve.w = curpt[2].x;
                        curve[2] = make_float2(curpt[3].x, curpt[3].y);

                        out_segs.push_back(seg);
                    }
                    break;
                }

                curpt += size(contour.segments[s]);
            } // for each segment
        } // for each contour

        // add the path segment
        {
            Segment<> seg;
            seg.type = SEG_PATH;

            // seg.data must point to the first point added in the
            // for loop
            seg.data = &out_curve_data.back()+1;

            out_curve_data.resize(out_curve_data.size() + sizeof(PathData));

            PathData &data = *seg.as<SEG_PATH>().path;

            data.idpath = path.id;
            data.frule = FRULE_NONZERO; // it's always nonzero for fd's
            data.paint = attrib_paints[path.attrib];

            out_segs.push_back(seg);
        }

    } // for each path

#if FRONT_TO_BACK_RENDERING
    std::reverse(out_segs.begin(), out_segs.end());
#endif
}/*}}}*/

void convert_segments(const fd::FontDump &fd,
                      const gpu::vector<const Paint *> &attrib_paints,
                      float2 scale, float2 off,
                      Segments &outsegs)
{
    std::vector<Segment<> > h_segs;
    std::vector<char> h_curve_data;
    std::vector<unsigned> h_path_attrib;

    convert_segments(fd, h_segs, h_path_attrib, h_curve_data, attrib_paints);

    printf("# paths: %lu\n",fd.paths.size());
    printf("# segments: %lu\n",h_segs.size());

    gpu::vector<Segment<> > segs(h_segs);
    gpu::vector<char> curve_data(h_curve_data);

    adjust_curve_data_pointer(segs, curve_data, h_curve_data);

    // convert segments to implicit curve segments
    preprocess(segs, outsegs, scale, off);

    printf("# xy-monotonic segment: %lu\n",outsegs.size());

    printf("Sizes: B1:%lu, B2:%lu, B3:%lu, ARC:%lu, P: %lu, I:%lu\n",
           sizeof(ImplicitCurve<1>),
           sizeof(ImplicitCurve<2>),
           sizeof(ImplicitCurve<3>),
           sizeof(ImplicitCurve<SEG_ARC>),
           sizeof(PathData),
           sizeof(ImplicitSegment<>));
}

class adj_coord/*{{{*/
{
public:
    adj_coord(long long off) : m_off(off) {}

    __device__ void operator()(Segment<> &seg) const
    {
        (char *&)seg.data += m_off;
    }

private:
    long long m_off;
};/*}}}*/

void adjust_curve_data_pointer(gpu::vector<Segment<> > &segs,
                               const gpu::vector<char> &d_curve_data,
                               const std::vector<char> &h_curve_data)
{
    long long addr_offset = d_curve_data.data() - h_curve_data.data();
    thrust::for_each(segs.begin(), segs.end(), adj_coord(addr_offset));
}
