#include "qtree.h"
#include <gpu/timer.h>
#include <thrust/sort.h>
#include "util.h"
#include "traversal.h"

const int BW = 256;

HOSTDEV
unsigned short demorton_odd(unsigned x)
{
    x = x & 0x55555555;
    x = (x | (x >> 1)) & 0x33333333;
    x = (x | (x >> 2)) & 0x0F0F0F0F;
    x = (x | (x >> 4)) & 0x00FF00FF;
    x = (x | (x >> 8)) & 0x0000FFFF;
    return x;
}

HOSTDEV
ushort2 demorton(unsigned z)
{
    return make_ushort2(demorton_odd(z), demorton_odd(z>>1));
}

__global__ void gpu_traverse_qtree(const_QuadsPtr qtree,/*{{{*/
                                   unsigned width, unsigned height,
                                   float adj_coord,
                                   TraversalInfosPtr tinfos)
{
    TraversalInfo tinfo;

    tinfo.pix = demorton(blockIdx.x*BW+threadIdx.x);

    if(tinfo.pix.x >= width || tinfo.pix.y >= height)
        return;

    float2 coord
        = make_float2(tinfo.pix.x, tinfo.pix.y)+adj_coord;

    tinfo.idquad = 0; // root cell
    while(__ldg(qtree.seg_range + tinfo.idquad).y == 0)
    {
        int c = qtree.classify(tinfo.idquad, coord);

        int idchildquad;

        int4 children = __ldg(qtree.children + tinfo.idquad);

        switch(c)
        {
        case 0:
            idchildquad = children.x;
            break;
        case 1:
            idchildquad = children.y;
            break;
        case 2:
            idchildquad = children.z;
            break;
        case 3:
            idchildquad = children.w;
            break;
        }

        if(idchildquad >= 0)
            tinfo.idquad = idchildquad;
        else
            break;
    }

    unsigned idx = tinfo.pix.y*width + tinfo.pix.x;

    tinfos.put(idx, tinfo);
}/*}}}*/

void traverse_qtree(TraversalInfos &tinfo,
                    const Quads &qtree,
                    unsigned imgwidth, unsigned imgheight,
                    float coord_adj)
{
    unsigned maxdim = std::max(next_power_of_2(imgwidth),
                               next_power_of_2(imgheight));
    dim3 bdim(BW),
         gdim((maxdim*maxdim+BW-1)/BW);

    {
        tinfo.clear();
        tinfo.resize(imgwidth*imgheight);

        assert(qtree.seg_range.size() == qtree.size());

        gpu_traverse_qtree<<<gdim, bdim>>>(&qtree,
                                           imgwidth, imgheight, coord_adj,
                                           &tinfo);
    }
}

void sort(TraversalInfos &tinfo)
{
    gpu::scoped_timer_stop sts(gpu::timers.gpu_add("sort traversal info"));

    thrust::stable_sort_by_key(tinfo.idquad.begin(), tinfo.idquad.end(),
                               tinfo.pix.begin());
}
