#ifndef GPURASTER_UTIL_H
#define GPURASTER_UTIL_H

#include <image/color.h>
#include <math/point.h>
#include <math/transform.h>
#include <gpu/linalg.h>
#include <gpu/cutil.h> // for dot
#include "segment.h"

__device__ __host__ inline float4 to_float4(const image::Color &c)
{
    return make_float4(c.r, c.g, c.b, c.a);
}

__device__ __host__ inline float2 to_float2(const math::Point &p)
{
    return make_float2(p.x, p.y);
}

struct Transform
{
    float4 l; // [x y; z w]
    float2 t;
    float2 free_space;

    HOSTDEV Transform &operator*=(const Transform &that);
    HOSTDEV Transform operator*(const Transform &that) const;
};

struct Projective
{
    float3 l0;
    float3 l1;
    float3 l2;
};

HOSTDEV  inline float2 apply(const Projective &t, const float2 &p)
{
    float x = t.l0.x*p.x + t.l0.y*p.y + t.l0.z;
    float y = t.l1.x*p.x + t.l1.y*p.y + t.l1.z;
    float z = t.l2.x*p.x + t.l2.y*p.y + t.l2.z;
    // TODO: != 0 is ugly... Provide a better numerical 
    // treatment.
    if ( z != 0 ) {
        return make_float2(x/z, y/z);
    }
    else {
        return make_float2(x, y);
    }
}

HOSTDEV inline Projective inverse(const Projective &t)
{
    float idet = 1.0f/(-t.l0.z*t.l1.y*t.l2.x + t.l0.y*t.l1.z*t.l2.x + t.l0.z*t.l1.x*t.l2.y 
                - t.l0.x*t.l1.z*t.l2.y - t.l0.y*t.l1.x*t.l2.z + t.l0.x*t.l1.y*t.l2.z);
    float a00 = -t.l1.z*t.l2.y + t.l1.y*t.l2.z,
        a01 = t.l0.z*t.l2.y - t.l0.y*t.l2.z,
        a02 = -t.l0.z*t.l1.y + t.l0.y*t.l1.z,
        a10 = t.l1.z*t.l2.x - t.l1.x*t.l2.z,
        a11 = -t.l0.z*t.l2.x + t.l0.x*t.l2.z,
        a12 =  t.l0.z*t.l1.x - t.l0.x*t.l1.z,
        a20 = -t.l1.y*t.l2.x + t.l1.x*t.l2.y,
        a21 = t.l0.y*t.l2.x - t.l0.x*t.l2.y,
        a22 = -t.l0.y*t.l1.x + t.l0.x*t.l1.y;
    Projective i;
    i.l0.x = a00*idet;
    i.l0.y = a01*idet;
    i.l0.z = a02*idet;
    i.l1.x = a10*idet;
    i.l1.y = a11*idet;
    i.l1.z = a12*idet;
    i.l2.x = a20*idet;
    i.l2.y = a21*idet;
    i.l2.z = a22*idet;
    return i;
}


HOSTDEV inline Transform inverse(const Transform &t)
{
    float a = t.l.x, b = t.l.y, c = t.l.z, d = t.l.w;

    float idet = 1.f/(a*d - b*c);
    Transform i;
    i.l.x = d*idet;
    i.l.y = -b*idet;
    i.l.z = -c*idet;
    i.l.w = a*idet;
    i.t.x = (b*t.t.y - d*t.t.x)*idet;
    i.t.y = (c*t.t.x - a*t.t.y)*idet;
    return i;
}

HOSTDEV inline Transform rotate(float ang)
{
    Transform t;
    t.l.x = cos(ang);
    t.l.y = -sin(ang);
    t.l.z = sin(ang);
    t.l.w = cos(ang);
    t.t.x = t.t.y = 0;
    return t;
}

HOSTDEV inline Transform translate(float x, float y)
{
    Transform t;
    t.l.x = 1;
    t.l.y = 0;
    t.l.z = 0;
    t.l.w = 1;
    t.t.x = x;
    t.t.y = y;
    return t;
}

HOSTDEV inline Transform scale(float x, float y)
{
    Transform t;
    t.l.x = x;
    t.l.y = 0;
    t.l.z = 0;
    t.l.w = y;
    t.t.x = t.t.y = 0;
    return t;
}

HOSTDEV inline Transform identity()
{
    Transform t;
    t.l.x = 1;
    t.l.y = 0;
    t.l.z = 0;
    t.l.w = 1;
    t.t.x = t.t.y = 0;
    return t;
}

HOSTDEV  inline float2 apply(const Transform &t, const float2 &p)
{
    return make_float2(t.l.x*p.x + t.l.y*p.y + t.t.x,
                       t.l.z*p.x + t.l.w*p.y + t.t.y);
}

HOSTDEV
inline void compose_into(Transform &a, const Transform &b)
{
    float Aa = a.l.x, Ab = a.l.y, Ac = a.l.z, Ad = a.l.w,
          Ba = b.l.x, Bb = b.l.y, Bc = b.l.z, Bd = b.l.w;

    a.l.x = Aa*Ba + Ab*Bc;
    a.l.y = Aa*Bb + Ab*Bd;
    a.t.x += Aa*b.t.x + Ab*b.t.y;
    a.l.z = Ac*Ba + Ad*Bc;
    a.l.w = Ac*Bb + Ad*Bd;
    a.t.y += Ac*b.t.x + Ad*b.t.y;
}

HOSTDEV
inline Transform compose(const Transform &a, const Transform &b)
{
    Transform c = a;
    compose_into(c, b);
    return c;
}

HOSTDEV
inline Transform &Transform::operator*=(const Transform &that)
{
    compose_into(*this, that);
    return *this;
}

HOSTDEV
inline Transform Transform::operator*(const Transform &that) const
{
    Transform xf(*this);
    return xf *= that;
}

/*
 * return <0 if p is to the left of ab
 */
__device__ inline
float ccw(float2 p, float2 a, float2 b)
{
    float d00 = p.x - a.x, d01 = p.y - a.y;
    float d10 = b.x - a.x, d11 = b.y - a.y;
    return d00*d11 - d01*d10;
}

// from bithacks
inline unsigned next_power_of_2(unsigned v)
{
    v--;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    v++;
    return v;
}


/*
 * this function is based on the ccw test:
 * ccw( p, b, e ) < 0 iff p is "to the left" (rotating counterclockwisely)
 * of the segment be. So
 *                     | px-bx py-by |
 * ccw( p, b, e ) = Det| ex-bx ey-ay | = (ey-by)*px + (bx-ex)*py + by*ex-bx*ey
 *                                     = A*px + B*py + C
 *  Updated to a more stable version
 */
__device__
inline void find_line_coefficients(float2 b, float2 e,
                                   gpu::math::Vector<float,3> &c)
{
    c[0] = e.y - b.y;
    c[1] = b.x - e.x;
    c[2] = -b.y*c[1] - b.x*c[0];
}

// returns c1, c2 and c3 in .x,.y,.z, det in .w
__device__
inline float4 baricentric_coordinates(float2 u, float2 v, float2 p)
{
    float a = u.x, b = u.y,
          c = v.x, d = v.y;

    float4 r;

    r.w = a*d - c*b;        // determinant
    r.y = d*p.x - c*p.y;    // c2
    r.z = -b*p.x + a*p.y;   // c3
    r.x = r.w - r.y - r.z;  // c1
    return r;
}

/*
hbarycoords[el : {{xA_, yA_, 1}, {wxB_, wyB_, wB_}, {xC_, yC_, 1}},
      p : {x_, y_}] :=
           {wyB (x + xA - xC) - (wxB - wB xC) (y +
                         yA) + (wxB - wB (x + xA)) yC,
             xA y - xC y + x (-yA + yC), -wyB x + wxB y - wB xA y + wB x yA}
*/
__device__ inline
float3 homo_baricentric_coordinates(float2 A,
                                    float2 B, float wB,
                                    float2 C,
                                    float2 p)
{
    float xA = A.x,
          yA = A.y;

    float wxB = B.x,
          wyB = B.y;

    float xC = C.x,
          yC = C.y;

    float3 bc;

    bc.x = wyB*(p.x + xA - xC) - (wxB - wB*xC)*(p.y + yA)
            + (wxB - wB*(p.x + xA))*yC;

    bc.y = xA*p.y - xC*p.y + p.x*(-yA + yC);

    bc.z = -wyB*p.x + wxB*p.y - wB*xA*p.y + wB*p.x*yA;

    return bc;
}

template <class T>
__device__
T scan32(T v)
{
#pragma unroll
    for (int i=16; i>=1; i/=2)
        v += __shfl_xor(v, i, 32);

    return v;
}

__device__ __host__ inline
float sqrnorm(const float2 &v)
{
    return dot(v,v);
}

__device__ __host__ inline
float norm(const float2 &v)
{
    return sqrt(sqrnorm(v));
}

#endif
