#ifndef WARPS_H
#define WARPS_H

#include "util.h"
#include <gpu/linalg.h>

struct NoWarp {

    __device__ __host__
    NoWarp( void )
    { }

    __device__
    float2 apply( const float2& p ) const
    {
        return p;
    }

};

struct ProjectiveWarp {
    ::Projective xf;

    __device__ __host__
    ProjectiveWarp( const ::Projective& xf )
        : xf( xf )
    { }

    __device__
    float2 apply( const float2& p ) const
    {
        return ::apply( xf, p );
    }
};

struct LensWarp {

    __device__ __host__
    LensWarp( float cu, float cv, float outer,
            float inner, float zoom, float w, float h )
            : m_center_u(cu), m_center_v(cv), m_outer_radius(outer)
            , m_inner_radius_frac(inner), m_zoom(zoom)
            , m_width(w), m_height(h)
    { }

    __device__
    float2 apply(const float2& p) const {
        float u, v;
        if ( m_width < m_height ) {
            u = 2.f*(p.x + (m_height-m_width)/2.f)/m_height -1.f;
            v = 2.f*p.y/m_height -1.f;
        }
        else {
            u = 2.f*p.x/m_width -1.f;
            v = 2.f*(p.y + (m_width-m_height)/2.f)/m_width -1.f;
        }
        float x = u-m_center_u;
        float y = v-m_center_v;
        float r = sqrtf(x*x+y*y)/m_outer_radius;
        r = (r-m_inner_radius_frac)/(1.f-m_inner_radius_frac);
        float z = rz(r,m_zoom);
        float2 wp;
        if ( m_width < m_height ) {
            wp.x = m_height*.5f*(m_center_u+x*z+1.f) - (m_height-m_width)/2.f;
            wp.y = m_height*.5f*(m_center_v+y*z+1.f);
        }
        else {
            wp.x = m_width*.5f*(m_center_u+x*z+1.f);
            wp.y = m_width*.5f*(m_center_v+y*z+1.f) - (-m_height+m_width)/2.f;
        }
        return wp;
    }

    __device__
    float rz(float r, float z) const {
        float s = (((3.f*(z-1.f)*r-8.f*(z-1.f))*r+6.f*(z-1.f))*r*r+1.f)/z;
        return r > 1.f? 1.f: (r < 0.f? 1.f/z: s);
    }

    float m_center_u, m_center_v, m_outer_radius, m_inner_radius_frac,
          m_zoom;
    float m_width, m_height;
};

struct StubWarp {
    int r;

    __device__ __host__
    StubWarp( int r )
        : r(r)
    { }

    __device__
    float2 apply( const float2& p ) const
    {
        float cr = cosf(r);
        float sr = sinf(r);
        return make_float2(cr*p.x - sr*p.y, sr*p.x + cr*p.y);
    }

};

struct SwirlWarp {

    __host__ __device__
    SwirlWarp(float amount, float ang, float w, float h, float scale = 1.f)
        : m_amount(amount)
        , m_ang(ang)
        , m_w(w)
        , m_h(h)
        , m_scale(scale)
    { }

    __device__
    float2 apply( const float2& p ) const {
        float x, y;
        if ( m_w < m_h ) {
            y = 2.f*(p.x + (m_h-m_w)/2.f)/m_h -1.f;
            x = 2.f*p.y/m_h -1.f;
        }
        else {
            y = 2.f*p.x/m_w -1.f;
            x = 2.f*(p.y + (m_w-m_h)/2.f)/m_w -1.f;
        }
        x *= m_scale;
        y *= m_scale;
        float c0 = cosf(m_ang*M_PI/180.f);
        float s0 = sinf(m_ang*M_PI/180.f);
        float t = c0*x - s0*y;
        y = s0*x + c0*y;
        x = t;
        float r = sqrtf(x*x+y*y);
        float ang = atan2f(y, x);
        ang += 2.f*3.141592f*r*m_amount;
        float2 wp;
        if ( m_w < m_h ) {
            wp.x = m_h*.5f*(r*sinf(ang)+1.f) - (m_h-m_w)/2.f;
            wp.y = m_h*.5f*(r*cosf(ang)+1.f);
        }
        else {
            wp.x = m_w*.5f*(r*sinf(ang)+1.f);
            wp.y = m_w*.5f*(r*cosf(ang)+1.f) - (-m_h+m_w)/2.f;
        }
        return wp;
    }

    float smoothstep(float t) const {
        t -= .5f;
        t *= 2.f;
        if (t < -1.f) t = -1.f;
        if (t > 1.f) t = 1.f;
        return (2 + 3*t - t*t*t)/4;
    }

    float m_amount, m_ang, m_w, m_h, m_scale;
};

#endif
