#include <cstdio>

#include "svg.h"

#include "../path/arc.h"
#include "../path/concrete.h"
#include "../path/stroke.h"
#include "../image/concrete.h"
#include "../image/png.h"
#include "../util/base64.h"
#include "../math/util.h"

static int touchar(float v) {
    int iv = static_cast<int>(255.f*v+0.5f);
    if (iv > 255) iv = 255;
    if (iv < 0) iv = 0;
    return iv;
}

namespace rvg { namespace driver {

const char *capname(CapStyle cap_style) {
    switch (cap_style) {
        case CapStyle::butt: return "butt";
        case CapStyle::round: return "round";
        case CapStyle::square: return "square";
        case CapStyle::triangle: return "triangle";
        default: return "invalid cap style!";
    }
}

const char *joinname(JoinStyle join_style) {
    switch (join_style) {
        case JoinStyle::miter: return "miter";
        case JoinStyle::round: return "round";
        case JoinStyle::bevel: return "bevel";
        default: return "invalid join style!";
    }
}

#ifdef SVGSTROKES
class SVGConcretePath: public path::Concrete {
public:
    SVGConcretePath(void): m_is_stroked(false) { }
    SVGConcretePath(const SVGConcretePath &that, const Transform &xf):
        path::Concrete(that, xf), m_style(that.m_style),
        m_is_stroked(that.m_is_stroked), m_ctol(that.m_ctol),
        m_swtol(that.m_swtol), m_maxdiv(that.m_maxdiv) {
    }
    SVGConcretePath(const path::Concrete &that, const Transform &xf
        = Transform{}):
        path::Concrete(that, xf), m_is_stroked(false) {
    }
    const stroke::Style &style(void) const { return m_style; }
    bool is_stroked(void) const { return m_is_stroked; }
protected:
    SVGConcretePath*
    do_stroke_clone(const Transform &xf, const stroke::Style &style,
        float atol, float swtol, int maxdiv) const {
        SVGConcretePath *cloned;
        if (m_is_stroked) {
            cloned = new SVGConcretePath(path::stroke(*this, xf, m_style, m_ctol, m_swtol, m_maxdiv));
        } else {
            cloned = new SVGConcretePath(*this, xf);
        }
        cloned->m_is_stroked = true;
        cloned->m_style = style;
        cloned->m_ctol = cos(atol*math::to_rad());
        cloned->m_swtol = swtol;
        cloned->m_maxdiv = maxdiv;
        return cloned;
    }

private:
    stroke::Style m_style;
    bool m_is_stroked;
    float m_ctol, m_swtol;
    int m_maxdiv;
};
#else
using SVGConcretePath = path::Concrete;
#endif

const image::Concrete &svg_get(const Image &image) {
    return *static_cast<const image::Concrete *>(image.ro());
}

// We should elmininate this hack ??
class SVGSurface: public rvg::surface::Base {
    IMPLEMENT_CLONEABLE(SVGSurface);
};

Surface
SVG::do_new_surface(int, int) {
    return Surface(new SVGSurface());
}

Image
SVG::do_new_image(FILE *file) {
    image::Concrete *img = new image::Concrete;
    if (!image::png::load(file, *img))
        fprintf(stderr, "failed loading image\n");
    return Image(img);
}

Image
SVG::do_new_image(const char *str, size_t len) {
    image::Concrete *img = new image::Concrete;
    if (!image::png::load(std::string(str, len), *img))
        fprintf(stderr, "failed loading image\n");
    return Image(img);
}

paint::Ramp
SVG::do_new_color_ramp(SpreadMethod spread,
                  const std::vector<paint::ramp::Stop> &stops)
{
    // spurious copy of 'stops' vector :(
    // but in most use cases we'll be using another representation
    // (textures in cuda, cairo's own representation, etc...) so
    // passing of stops by const ref is a good thing on average.
    return paint::Ramp(new paint::ramp::Base(spread, stops));
}

Scene
SVG::do_new_scene(void) {
    return Scene();
}

Paint
SVG::do_new_solid_paint(const Color &color, float opacity) {
    return Paint(new paint::Solid(color), opacity);
}

Paint
SVG::do_new_linear_gradient(const paint::Ramp ramp,
    r2::Point p1, r2::Point p2,
    Transform xf, float opacity) {
    return Paint(
        new paint::LinearGradient(ramp, p1, p2, xf), opacity);
}

Paint
SVG::do_new_radial_gradient(const paint::Ramp ramp,
    r2::Point c, r2::Point f, float r,
    Transform xf, float opacity) {
    return Paint(
        new paint::RadialGradient(ramp, c, f, r, xf), opacity);
}

Paint
SVG::do_new_texture(Image image, SpreadMethod spread_method,
    Transform xf, float opacity) {
    return Paint(new paint::Texture(image, spread_method, xf), opacity);
}

Path
SVG::do_new_path(void) {
    return Path(new SVGConcretePath);
}

void printxf(FILE *out, const Transform &obj) {

    class PrintVisitor: public xform::Visitor {
    public:
        explicit PrintVisitor(FILE *out): m_out(out) {}

    private:
        void do_visit(const xform::Identity &) override { }

        void do_visit(const xform::Translation &xf) override {
            fprintf(m_out, "translate(%g", xf.dx);
            if (xf.dy != 0.f) fprintf(m_out, ", %g", xf.dy);
            fputc(')', m_out);
        }

        void do_visit(const xform::Rotation &xf) override {
            fprintf(m_out, "rotate(%g)", xf.ang() * math::to_deg());
        }

        void do_visit(const xform::Scale &xf) override {
            fprintf(m_out, "scale(%g", xf.sx);
            if (xf.sx != xf.sy) fprintf(m_out, ", %g", xf.sy);
            fputc(')', m_out);
        }

        void do_visit(const xform::Linear &xf) override {
            fprintf(m_out, "matrix(%g,%g,%g,%g,0,0)",
                xf[0][0], xf[1][0], xf[0][1], xf[1][1]);
        }

        void do_visit(const xform::Affine &xf) override {
            fprintf(m_out, "matrix(%g,%g,%g,%g,%g,%g)",
                xf[0][0], xf[1][0], xf[0][1], xf[1][1], xf[0][2], xf[1][2]);
        }

        void do_visit(const xform::Projective &xf) override {
            (void) xf;
            assert(0); // shouldn't be here
        }

        FILE *m_out;
    };

    PrintVisitor p{out};
    obj.accept(p);
}

void printpath(FILE *out, const SVGConcretePath &obj,
    const Transform &xf = xform::Identity{}) {
    const auto &segments = obj.segments();
    const auto &data = obj.data();
    fputs("d=\"", out);
    SegmentType last_s = SegmentType::move_to;
    for (const auto segment: segments) {
        const int i = segment.index;
        SegmentType s = static_cast<SegmentType>(segment.type);
        switch (s) {
            case SegmentType::close_path:
                fputs("Z ", out);
                break;
            case SegmentType::move_to: {
                r2::Point p0 = xf * r2::Point{data[i], data[i+1]};
                fprintf(out, "M %g,%g ", p0.x, p0.y);
                break;
            }
            case SegmentType::line_to: {
                if (last_s != s) fputs("L ", out);
                r2::Point p1 = xf * r2::Point{data[i+2], data[i+3]};
                fprintf(out, "%g,%g ", p1.x, p1.y);
                break;
            }
            case SegmentType::quad_to: {
                if (last_s != s) fputs("Q ", out);
                r2::Point p1 = xf * r2::Point{data[i+2], data[i+3]};
                r2::Point p2 = xf * r2::Point{data[i+4], data[i+5]};
                fprintf(out,  "%g,%g %g,%g ", p1.x, p1.y, p2.x, p2.y);
                break;
            }
            case SegmentType::cubic_to: {
                r2::Point p1 = xf * r2::Point{data[i+2], data[i+3]};
                r2::Point p2 = xf * r2::Point{data[i+4], data[i+5]};
                r2::Point p3 = xf * r2::Point{data[i+6], data[i+7]};
                if (last_s != s) fputs("C ", out);
                fprintf(out,  "%g,%g %g,%g %g,%g ",
                    p1.x, p1.y, p2.x, p2.y, p3.x, p3.y);
                break;
            }
            case SegmentType::rquad_to: {
                r2::Point p0 = xf * r2::Point{data[i+0], data[i+1]};
                r3::Point p1 = xf * r3::Point{data[i+2], data[i+3], data[i+4]};
                r2::Point p2 = xf * r2::Point{data[i+5], data[i+6]};
                path::RationalArc rarc{p0, p1, p2};
                if (rarc.is_parabola()) {
                    s = SegmentType::quad_to;
                    if (last_s != s) fputs("Q ", out);
                    fprintf(out, "%g,%g %g,%g ",
                        p1.x, p1.y, p2.x, p2.y);
                } else {
                    path::VGArc vg(rarc);
                    if (last_s != s) fputs("A ", out);
                    fprintf(out, "%g,%g %g %d %d %g,%g ",
                        vg.rx(), vg.ry(), vg.ang()*math::to_deg(),
                        vg.fa(), vg.fs(), p2.x, p2.y);
                    break;
                }
            }
            case SegmentType::fline_to:
                break; // ignore internal segments
            default:
                // should never be here!
                assert(0);
                break;
        }
        last_s = s;
    }
    fputc('"', out);
}

void printrampdef(FILE *out, const paint::Ramp &obj) {
    for (const auto &stop: obj.stops()) {
        fprintf(out, "<stop offset=\"%g\" stop-color=\"rgb(%d,%d,%d)\" "
           "stop-opacity=\"%g\"/>\n", stop.offset,
            touchar(stop.color.red), touchar(stop.color.green),
            touchar(stop.color.blue),  stop.color.opacity);
    }
}

const char *spreadname(SpreadMethod spread_method) {
    switch(spread_method) {
        case SpreadMethod::pad: return "pad";
        case SpreadMethod::repeat: return "repeat";
        case SpreadMethod::reflect: return "reflect";
        default: return "invalid spread method!";
    }
}


void printstroke(FILE *out, const stroke::Style style) {
    fprintf(out, "stroke-width=\"%g\" ", style.width);
    fprintf(out, "stroke-linecap=\"%s\" ", capname(style.cap));
    fprintf(out, "stroke-linejoin=\"%s\" ", joinname(style.join));
    if (style.join == JoinStyle::miter)
        fprintf(out, "stroke-miterlimit=\"%g\" ", style.miter_limit);
    if (!style.dash.array.empty()) {
        fprintf(out, "stroke-dasharray=\"");
        for (auto s: style.dash.array)
            fprintf(out, "%g ", s);
        fprintf(out, "\" ");
        fprintf(out, "stroke-dashoffset=\"%g\" ", style.dash.initial_phase);
    }
}

void printpaintdef(FILE *out, const Paint &obj, int &id) {

    class PrintVisitor: public paint::Visitor {
    public:
        explicit PrintVisitor(FILE *out, int &id):
            m_out(out), m_id(id) {}

    private:
        void do_visit(const paint::Solid &) override {
        }

        void do_visit(const paint::LinearGradient &linear) override {
            fprintf(m_out, "<linearGradient id=\"linear%d\" x1=\"%g\" "
                "y1=\"%g\" x2=\"%g\" y2=\"%g\" spreadMethod=\"%s\" "
                "gradientUnits=\"userSpaceOnUse\" gradientTransform=\"",
                m_id++, linear.p1().x, linear.p1().y, linear.p2().x,
                linear.p2().y, spreadname(linear.ramp().spread()));
                printxf(m_out, linear.xf());
            fputs("\">\n", m_out);
            printrampdef(m_out, linear.ramp());
            fputs("</linearGradient>\n", m_out);
        }

        void do_visit(const paint::RadialGradient &radial) override {
            fprintf(m_out, "<radialGradient id=\"radial%d\" cx=\"%g\" "
               "cy=\"%g\" fx=\"%g\" fy=\"%g\" r=\"%g\" spreadMethod=\"%s\" "
               "gradientUnits=\"userSpaceOnUse\" gradientTransform=\"", m_id++,
                radial.c().x, radial.c().y, radial.f().x, radial.f().y,
                radial.r(), spreadname(radial.ramp().spread()));
            printxf(m_out, radial.xf());
            fputs("\">\n", m_out);
            printrampdef(m_out, radial.ramp());
            fputs("</radialGradient>\n", m_out);
        }

        void do_visit(const paint::Texture &texture) override {
            fprintf(m_out, "<image id=\"image%d\" width=\"1\" height=\"1\" "
                "preserveAspectRatio=\"none\" transform=\"", m_id++);
            printxf(m_out, texture.xf());
            fputs(" scale(1,-1) translate(0,-1)\" xlink:href=\" "
                "data:image/png;base64,\n", m_out);
            std::string encoded;
            const image::Concrete &img = svg_get(texture.image());
            image::png::store8(encoded, img);
            encoded = encode_base64(encoded);
            fwrite(&encoded[0], 1, encoded.size(), m_out);
            fputs("\"/>\n", m_out);
        }

        FILE *m_out;
        int &m_id;
    };

    PrintVisitor p(out, id);
    obj.accept(p);
}

bool is_texture(const Paint &obj) {
    class IsTextureVisitor: public paint::Visitor {
    public:
        explicit IsTextureVisitor(bool &is_it): m_is_it(is_it) {}
    private:
        void do_visit(const paint::Solid &) override { }
        void do_visit(const paint::LinearGradient &) override { }
        void do_visit(const paint::RadialGradient &) override { }
        void do_visit(const paint::Texture &) override { m_is_it = true; }
        bool &m_is_it;
    };
    bool is_it = false;
    IsTextureVisitor p{is_it};
    obj.accept(p);
    return is_it;
}

void printpaint(FILE *out, const Paint &obj,
    const char *mode, int &grad_id, int clip_id = 0) {

    class PrintVisitor: public paint::Visitor {
    public:
        explicit PrintVisitor(FILE *out, float opacity,
            const char *mode, int &grad_id, int clip_id):
            m_out(out), m_opacity(opacity), m_mode(mode),
            m_grad_id(grad_id), m_clip_id(clip_id) {}

    private:
        void do_visit(const paint::Solid &solid) override {
            const Color &c = solid.color();
            fprintf(m_out, "%s=\"rgb(%d,%d,%d)\"", m_mode,
                touchar(c.red), touchar(c.green), touchar(c.blue));
            if (m_opacity != 1.f || c.opacity != 1.f)
                fprintf(m_out, " %s-opacity=\"%g\"", m_mode,
                    m_opacity*c.opacity);
        }

        void do_visit(const paint::LinearGradient &) override {
            fprintf(m_out, "%s=\"url(#linear%d)\"", m_mode, m_grad_id++);
            if (m_opacity != 1.f)
                fprintf(m_out, " %s-opacity=\"%g\"", m_mode, m_opacity);
        }

        void do_visit(const paint::RadialGradient &) override {
            fprintf(m_out,"%s=\"url(#radial%d)\"", m_mode, m_grad_id++);
            if (m_opacity != 1.f)
                fprintf(m_out, " %s-opacity=\"%g\"", m_mode, m_opacity);

        }

        void do_visit(const paint::Texture &) override {
            fprintf(m_out,"<use xlink:href=\"#image%d\" "
                "clip-path=\"url(#clip%d)\" opacity=\"%g\" />\n",
                m_grad_id++, m_clip_id, m_opacity);
        }

        int touchar(float v) {
            int iv = static_cast<int>(255.f*v+0.5f);
            if (iv > 255) iv = 255;
            if (iv < 0) iv = 0;
            return iv;
        }

        FILE *m_out;
        float m_opacity;
        const char *m_mode;
        int &m_grad_id;
        int m_clip_id;
    };

    PrintVisitor p{out, obj.opacity(), mode, grad_id, clip_id};
    obj.accept(p);
}

const SVGSurface &svg_get(const Surface &surface) {
    return *static_cast<const SVGSurface *>(surface.ro());
}

const SVGConcretePath &svg_get(const Path &path) {
    return *static_cast<const SVGConcretePath *>(path.ro());
}

void
SVG::render(Scene scene, const Viewport &viewport, FILE *out) {
    using namespace scene;
    r2::Vectori d = viewport.max - viewport.min;
    fprintf(out, "<?xml version=\"1.0\" standalone=\"no\"?>\n"
       "<svg width=\"%d\" height=\"%d\"\n"
       "xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n"
       "xmlns:dc=\"http://purl.org/dc/elements/1.1/\"\n"
       "xmlns:cc=\"http://creativecommons.org/ns#\"\n"
       "xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"\n"
       "xmlns:svg=\"http://www.w3.org/2000/svg\"\n"
       "xmlns=\"http://www.w3.org/2000/svg\"\n"
       "version=\"1.1\">\n", d.x, d.y);
    fputs("<defs>\n", out);
    int grad_id = 0;
    for (const auto &el: scene.elements()) {
        if (el.is_fill()) {
            printpaintdef(out, el.fill.paint, grad_id);
        }
    }
    fputs("</defs>\n", out);
    // change coordinate system. SVG uses Y going down.
    fprintf(out, "<g transform=\"scale(1,-1) translate(%d,%d) ",
        -viewport.min.x, -viewport.max.y);
    printxf(out, scene.xf());
    fputs("\">\n", out);

    int path_id = 0;
    grad_id = 0;
    int clippath_id = 0;
    std::vector<int> pushed;
    std::vector<int> committed;
    for (const auto &el: scene.elements()) {
        switch (el.type) {
        case Element::Type::ofill:
        case Element::Type::nzfill: {
            if (is_texture(el.fill.paint)) {
                fprintf(out, "<clipPath id=\"clip%d\">\n", clippath_id);
                if (el.type == Element::Type::nzfill)
                    fputs("<path clip-rule=\"nonzero\" ", out);
                else
                    fputs("<path clip-rule=\"evenodd\" ", out);
                if (!committed.empty()) {
                    fprintf(out, "clip-path=\"url(#clip%d)\" ",
                            committed.back());
                }
                printpath(out, svg_get(el.fill.path), el.fill.path.xf());
                fputs("/>\n", out);
                fprintf(out, "</clipPath>\n");
                printpaint(out, el.fill.paint, "fill", grad_id, clippath_id);
                clippath_id++;
            } else {
                fprintf(out, "<path id=\"path%d\" ", path_id++);
                if (el.type == Element::Type::nzfill) {
                    fputs("fill-rule=\"nonzero\" ", out);
                } else if (el.type == Element::Type::ofill) {
                    fputs("fill-rule=\"evenodd\" ", out);
                } else {
                    assert(0); // unsupported by SVG
                }
                const SVGConcretePath &svgpath = svg_get(el.fill.path);
#ifdef SVGSTROKES
                if (svgpath.is_stroked()) {
                    fputs("fill=\"none\" ", out);
                    printstroke(out, svgpath.style());
                    printpaint(out, el.fill.paint, "stroke", grad_id);
                } else {
                    fputs("stroke=\"none\" ", out);
                    printpaint(out, el.fill.paint, "fill", grad_id);
                }
#else
                fputs("stroke=\"none\" ", out);
                printpaint(out, el.fill.paint, "fill", grad_id);
#endif
                fputc(' ', out);
                if (!committed.empty()) {
                    fprintf(out, "clip-path=\"url(#clip%d)\" ", committed.back());
                }
                printpath(out, svgpath, el.fill.path.xf());
                fputs("/>\n", out);
            }
            break;
        }
        case Element::Type::nzclip:
        case Element::Type::oclip:
            if (el.type == Element::Type::nzclip)
                fputs("<path clip-rule=\"nonzero\" ", out);
            else
                fputs("<path clip-rule=\"evenodd\" ", out);
            if (!committed.empty()) {
                fprintf(out, "clip-path=\"url(#clip%d)\" ", committed.back());
            }
            printpath(out, svg_get(el.clip.path), el.clip.path.xf());
            fputs("/>\n", out);
            break;
        case Element::Type::pushfill:
        case Element::Type::pushclip:
            fprintf(out, "<clipPath id=\"clip%d\">\n", clippath_id);
            pushed.push_back(clippath_id);
            clippath_id++;
            break;
        case Element::Type::commit:
            fputs("</clipPath>\n", out);
            committed.push_back(pushed.back());
            pushed.pop_back();
            break;
        case Element::Type::popfill:
        case Element::Type::popclip:
            committed.pop_back();
            break;
        default:
            assert(0); // unsupported by SVG
            break;
        }
    }

    // close however many open clip-paths are still active
    for (unsigned i = 0; i < pushed.size(); i++)
        fputs("</clipPath>\n", out);

    fputs("</g></svg>\n", out);
}

}} // namespace rvg::driver
