#include <lua.hpp>
#define GLFW_INCLUDE_GLCOREARB
//#define GLFW_INCLUDE_GLU
#define GL_GLEXT_PROTOTYPES 1
#include <GLFW/glfw3.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <memory>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_gl_interop.h>
#include <cassert>

#include "glwindow.h"
#include "callback.h"

void check_cuda_error(const std::string &msg)
{
    cudaError_t err = cudaGetLastError();
    if(err != cudaSuccess)
    {
        if(msg.empty())
            throw std::runtime_error(cudaGetErrorString(err));
        else
        {
            std::stringstream ss;
            ss << msg << ": " << cudaGetErrorString(err);
            throw std::runtime_error(ss.str().c_str());
        }
    }
}

class LuaObjectViewer : public GlWindow {
public:
    LuaObjectViewer(void)
    { ; }
    ~LuaObjectViewer(void) { ; }

    void register_buffer( void );

    void do_draw( void )
    {/*{{{*/
        auto e = glGetError();
	//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        e = glGetError();
        if ( e != GL_NO_ERROR )
            std::cerr << "error glClear" << std::endl;
        if ( m_draw_cb ) {
            auto& cb = *m_draw_cb;
            cb( );
            glRasterPos2i(0, 0);
            glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, m_pbo);
            //glDrawPixels(m_width, m_height, GL_SRGB_ALPHA, GL_FLOAT, 0);
            glDrawPixels(m_width, m_height, GL_RGBA, GL_FLOAT, 0);
            glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
        }
        else {
            assert(0);
        }
    }/*}}}*/

    void do_key_event( int key, int scancode, int action, int mods )
    {
        (void) scancode;
        (void) action;
        (void) mods;
        if ( m_key_cb && action == 0 ) { //key release
            auto& cb = *m_key_cb;
            cb( key );
        }
    }
    void do_framebuffer_resize( int width, int height ) { (void) width; (void) height; }
    void do_window_resize( int width, int height ) { (void) width; (void) height; }
    void do_mouse_button( int button, int action, int mods ) { (void)button; (void)action; (void)mods; }
    void do_scroll( double dx, double dy ) { (void)dx; (void)dy; }
    void do_cursor( double x, double y )
    {
        if ( m_mouse_move_cb ) {
            auto& cb = *m_mouse_move_cb;
            cb( x, y );
        }
    }
    void do_cursor_enter( int enter ) { (void) enter; }

    void set_key_cb( LuaCallback *cb ) {
        m_key_cb.reset( cb );
    }

    void set_mouse_move_cb( LuaCallback *cb ) {
        m_mouse_move_cb.reset( cb );
    }

    void set_draw_cb( LuaCallback *cb ) {
        m_draw_cb.reset( cb );
    }

    void * init( int w, int h, std::string name )
    {
        m_width = w;
        m_height = h;
        GlWindow::init( w, h, name );
        //
        glViewport(0, 0, m_width, m_height);
        //
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        //
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
        //
        info();
        register_buffer( );
        return (void*) &m_cuda_res;
    }

    void show( void )
    {/*{{{*/
        //
        glClearColor( 0.9f, 0.9f, 0.9f, 1.0f );
        //
        //main_loop(false);
        main_loop(true);
        //
    }/*}}}*/

private:
    std::unique_ptr<LuaCallback> m_key_cb;
    std::unique_ptr<LuaCallback> m_mouse_move_cb;
    std::unique_ptr<LuaCallback> m_draw_cb;
    cudaGraphicsResource *m_cuda_res;
    GLuint m_pbo;
    int m_width, m_height;
};

void LuaObjectViewer::register_buffer( void )
{
    glGenBuffers(1, &m_pbo);
    std::cout << "pbo: " << m_pbo << std::endl;
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, m_pbo);
    glBufferData(GL_PIXEL_UNPACK_BUFFER, m_width*m_height*4*sizeof(GLfloat),
                 NULL, GL_STREAM_DRAW);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
    cudaGraphicsGLRegisterBuffer(&m_cuda_res, m_pbo, cudaGraphicsMapFlagsWriteDiscard);
    check_cuda_error("buffer register");
}

static int object_show(lua_State *L) {
    auto *o = static_cast<LuaObjectViewer*>( lua_touserdata(L, lua_upvalueindex(1)) );
    o->show( );
    return 0;
}

static int object_init(lua_State *L) {
    auto *o = static_cast<LuaObjectViewer*>( lua_touserdata(L, lua_upvalueindex(1)) );
    int w = luaL_checkinteger(L, 1);
    int h = luaL_checkinteger(L, 2);
    std::string name = luaL_checkstring(L, 3);
    lua_pushlightuserdata(L, o->init( w, h, name ));
    return 1;
}

static int object_delete(lua_State *L) {
    auto *o = static_cast<LuaObjectViewer*>( lua_touserdata(L,1) );
    o->~LuaObjectViewer();
    return 0;
}

static int set_key_cb(lua_State *L) {
    if ( lua_gettop(L) == 1 ) {
        if ( lua_isfunction(L,-1) ) {
            //TODO: use shared_ptr to avoid mem leak
            int index = luaL_ref(L, LUA_REGISTRYINDEX);
            auto *lbc = new LuaCallback(L, index);
            LuaObjectViewer *canvas = static_cast<LuaObjectViewer *>( lua_touserdata(L, lua_upvalueindex(1)) );
            canvas->set_key_cb( lbc );
        }
        return 0;
    }
    return 0;
}

static int set_mouse_move_cb(lua_State *L) {
    if ( lua_gettop(L) == 1 ) {
        if ( lua_isfunction(L,-1) ) {
            //TODO: use shared_ptr to avoid mem leak
            int index = luaL_ref(L, LUA_REGISTRYINDEX);
            auto *lbc = new LuaCallback(L, index);
            LuaObjectViewer *canvas = static_cast<LuaObjectViewer *>( lua_touserdata(L, lua_upvalueindex(1)) );
            canvas->set_mouse_move_cb( lbc );
        }
        return 0;
    }
    return 0;
}

static int set_draw_cb(lua_State *L) {
    if ( lua_gettop(L) == 1 ) {
        if ( lua_isfunction(L,-1) ) {
            //TODO: use shared_ptr to avoid mem leak
            int index = luaL_ref(L, LUA_REGISTRYINDEX);
            auto *lbc = new LuaCallback(L, index);
            LuaObjectViewer *canvas = static_cast<LuaObjectViewer *>( lua_touserdata(L, lua_upvalueindex(1)) );
            canvas->set_draw_cb( lbc );
        }
        return 0;
    }
    return 0;
}

static const struct luaL_Reg cmethods[] = {
    {"show", object_show},
    {"init", object_init},
    {"key_cb", set_key_cb},
    {"mouse_move_cb", set_mouse_move_cb},
    {"draw_cb", set_draw_cb},
    {NULL, NULL}
};

extern "C"
int luaopen_rvg10_glcanvas(lua_State *L) {
    luaL_newlibtable(L, cmethods); //modtable
    LuaObjectViewer *o = static_cast<LuaObjectViewer*>( lua_newuserdata(L, sizeof(LuaObjectViewer)) );
    new (o) LuaObjectViewer;
    //
    lua_newtable(L); // modtable, ptr, table
    lua_pushliteral(L, "__gc"); // modtable, ptr, table, "__gc"
    lua_pushcfunction(L, object_delete); // modtable, ptr, table cpu, "__gc", delete
    lua_rawset(L, -3); // modtable, ptr, table
    lua_setmetatable(L, -2); // modtable, ptr
    //
    luaL_setfuncs(L, cmethods, 1);
    return 1;
}
