#include <cstdio>
#include <iostream>
#include <sstream>
#include <cstring>
#include <cfloat>

#include "../bbox.h"
#include "../math/util.h"
#include "../util/base64.h"
#include "base.h"
#include "compat.h"

namespace rvg { namespace lua {

int meta_tostring_const(lua_State *L) {
    lua_pushvalue(L, 1); // const,const
    lua_rawget(L, lua_upvalueindex(1)); // const,constname
    lua_getmetatable(L, 1); // const,constname,meta
    lua_rawget(L, lua_upvalueindex(1)); // const,constname,typename
    lua_pushfstring(L, "%s.%s",
        Base::name(static_cast<Base::Type>(lua_tonumber(L, -1))),
        lua_tostring(L, -2));
    return 1;
}

int meta_index_const(lua_State *L) {
    const char *field = lua_tostring(L, 2);
    if (field) {
        luaL_error(L, "'%s' is not a '%s'",
            field, lua_tostring(L, lua_upvalueindex(1)));
    } else {
        luaL_error(L, "index is not a '%s'",
            lua_tostring(L, lua_upvalueindex(1)));
    }
    return 0;
}

int meta_newindex_const(lua_State *L) {
    luaL_error(L, "'%s' is protected", lua_tostring(L, lua_upvalueindex(1)));
    return 0;
}

void get_type_metatable(lua_State *L, Base::Type type) {
    lua_rawgeti(L, lua_upvalueindex(1), static_cast<int>(type));
}

bool is_type(lua_State *L, int idx, Base::Type type) {
    // this is tricky. getmetatable pushes nothing when the
    // ojbect does not have a metatable. I'd expect it to
    // push nil, but it doesn't. It returns zero. ??
    if (!lua_getmetatable(L, idx)) return false;
    get_type_metatable(L, type);
    bool v = is_equal(L, -1, -2);
    lua_pop(L, 2);
    return v;
}

void set_type(lua_State *L, Base::Type type) {
    lua_rawgeti(L, lua_upvalueindex(1), static_cast<int>(type));
    lua_setmetatable(L, -2);
}

driver::Base *get_driver(lua_State *L) {
    return static_cast<driver::Base *>(lua_touserdata(L, lua_upvalueindex(2)));
}

void push_color(lua_State *L, float r, float g, float b, float a) {
    float f[4] = {r, g, b, a};
    // get color cache from upvalue
    lua_pushlightuserdata(L, reinterpret_cast<void*>(Base::Type::color));
    lua_rawget(L, lua_upvalueindex(1));
    // try to find color in cache
    // we could squeeze the color spec in the size of pointer and
    // use a lightuserdata rather than a string to index the cache.
    // would be faster. ??
    lua_pushlstring(L, reinterpret_cast<char *>(f), sizeof(f));
    lua_rawget(L, -2);
    // on a miss, create color and set it into cache
    if (lua_isnil(L, -1)) {
        lua_pop(L, 1);
        void *p = lua_newuserdata(L, sizeof(Color));
        new (p) Color{r, g, b, a};
        set_type(L, Base::Type::color);
        lua_pushlstring(L, reinterpret_cast<char *>(f), sizeof(f));
        lua_pushvalue(L, -2);
        lua_rawset(L, -4);
    }
    // at this point we have cache,color
    lua_insert(L, -2);
    lua_pop(L, 1);
}

void push_paint_solid(lua_State *L, const Color *c, float o) {
    float f[5] = {c->red, c->green, c->blue, c->opacity, o};
    // get solid paint cache from upvalue
    lua_pushlightuserdata(L, reinterpret_cast<void*>(Base::Type::paint));
    lua_rawget(L, lua_upvalueindex(1));
    // try to find solid paint in cache
    // we could squeeze the paint spec in the size of pointer and
    // use a lightuserdata rather than a string to index the cache.
    // would be faster. ??
    lua_pushlstring(L, reinterpret_cast<char *>(f), sizeof(f));
    lua_rawget(L, -2);
    // on a miss, create solid paint and set it into cache
    if (lua_isnil(L, -1)) {
        lua_pop(L, 1);
        void *p = lua_newuserdata(L, sizeof(Paint));
        new (p) Paint(get_driver(L)->new_solid_paint(*c, o));
        set_type(L, Base::Type::paint);
        lua_pushlstring(L, reinterpret_cast<char *>(f), sizeof(f));
        lua_pushvalue(L, -2);
        lua_rawset(L, -4);
    }
    else {

    }
    // at this point we have cache,solid
    lua_insert(L, -2);
    lua_pop(L, 1);
}

using PathIterator = path::Base::const_iterator;

static stroke::Dash check_dash(lua_State *L, int idx) {
    idx = abs_index(L, idx);
    stroke::Dash dash;
    if (!lua_istable(L, idx)) luaL_error(L, "invalid dash specification");
    int n = table_len(L, idx);
    for (int i = 1; i <= n; i++) {
        lua_rawgeti(L, idx, i);
        if (!lua_isnumber(L, -1))
            luaL_error(L, "dash array entry %d should be a number", i);
        dash.array.push_back(lua_tonumber(L, -1));
        lua_pop(L, 1);
    }
    lua_pushliteral(L, "initial_phase");
    lua_rawget(L, idx);
    if (!lua_isnil(L, -1)) {
        if (!lua_isnumber(L, -1)) luaL_error(L, "invalid initial_phase");
        dash.initial_phase = lua_tonumber(L, -1);
    }
    lua_pop(L, 1);
    lua_pushliteral(L, "phase_reset");
    lua_rawget(L, idx);
    dash.phase_reset = lua_toboolean(L, -1);
    lua_pop(L, 1);
    return dash;
}

static PathIterator *check_path_iterators(lua_State *L, int idx) {
    if (!is_type(L, idx, Base::Type::path_iterator))
        luaL_argerror(L, idx, "expected path iterator");
    return static_cast<PathIterator *>(lua_touserdata(L, idx));
}

using SceneIterator = scene::Scene::const_iterator;

static SceneIterator *check_scene_iterators(lua_State *L, int idx) {
    if (!is_type(L, idx, Base::Type::scene_iterator))
        luaL_argerror(L, idx, "expected scene iterator");
    return static_cast<SceneIterator *>(lua_touserdata(L, idx));
}

SpreadMethod check_spread(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::spread_method)) luaL_argerror(L, idx, msg);
    return *static_cast<SpreadMethod *>(lua_touserdata(L, idx));
}

CapStyle check_cap_style(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::cap_style)) luaL_argerror(L, idx, msg);
    return *static_cast<CapStyle *>(lua_touserdata(L, idx));
}

CapStyle to_cap_style(lua_State *L, int idx) {
    CapStyle *c = static_cast<CapStyle *>(lua_touserdata(L, idx));
    if (c) return *c;
    else return CapStyle::butt;
}

CapStyle cap_style_or_none(lua_State *L, int idx, CapStyle def) {
    if (lua_isnil(L, idx) || lua_isnone(L, idx)) return def;
    else return check_cap_style(L, idx);
}

JoinStyle check_join_style(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::join_style)) luaL_argerror(L, idx, msg);
    return *static_cast<JoinStyle *>(lua_touserdata(L, idx));
}

JoinStyle to_join_style(lua_State *L, int idx) {
    JoinStyle *j = static_cast<JoinStyle *>(lua_touserdata(L, idx));
    if (j) return *j;
    else return JoinStyle::miter;
}

JoinStyle join_style_or_none(lua_State *L, int idx, JoinStyle def) {
    if (lua_isnil(L, idx) || lua_isnone(L, idx)) return def;
    else return check_join_style(L, idx);
}

stroke::Style check_stroke_style(lua_State *L, int idx, const char *msg) {
    idx = abs_index(L, idx);
    if (lua_istable(L, idx)) {
        stroke::Style s;
        // width
        lua_pushliteral(L, "width");
        lua_rawget(L, idx);
        if (lua_isnumber(L, -1))
            s.width = lua_tonumber(L, -1);
        lua_pop(L, 1);
        // cap
        lua_pushliteral(L, "cap");
        lua_rawget(L, idx);
        if (is_type(L, -1, Base::Type::cap_style))
            s.cap = to_cap_style(L, -1);
        lua_pop(L, 1);
        // join
        lua_pushliteral(L, "join");
        lua_rawget(L, idx);
        if (is_type(L, -1, Base::Type::join_style))
            s.join = to_join_style(L, -1);
        lua_pop(L, 1);
        // miter_limit
        lua_pushliteral(L, "miter_limit");
        lua_rawget(L, idx);
        if (lua_isnumber(L, -1))
            s.miter_limit = lua_tonumber(L, -1);
        lua_pop(L, 1);
        // dash
        lua_pushliteral(L, "dash");
        lua_rawget(L, idx);
        if (lua_istable(L, -1))
            s.dash = check_dash(L, -1);
        lua_pop(L, 1);
        return s;
    } else if (lua_isnumber(L, idx)) {
        return stroke::Style(lua_tonumber(L, idx));
    } else if (!is_type(L, idx, Base::Type::stroke_style))
        luaL_argerror(L, idx, msg);
    return *static_cast<stroke::Style *>(lua_touserdata(L, idx));
}

r2::Point *check_p2(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::p2)) luaL_argerror(L, idx, msg);
    return static_cast<r2::Point *>(lua_touserdata(L, idx));
}

r3::Point *check_p3(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::p3)) luaL_argerror(L, idx, msg);
    return static_cast<r3::Point *>(lua_touserdata(L, idx));
}

paint::Ramp *check_ramp(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::ramp)) luaL_argerror(L, idx, msg);
    return static_cast<paint::Ramp *>(lua_touserdata(L, idx));
}

Transform *check_xform(lua_State *L, int idx, const char *msg) {
    if (lua_isnone(L, idx) || lua_isnil(L, idx)) {
        // default to identity
        lua_rawgeti(L, lua_upvalueindex(1),
            static_cast<int>(Base::Type::identity));
        Transform *id = static_cast<Transform *>(lua_touserdata(L, -1));
        lua_pop(L, -1);
        return id;
    }
    if (!is_type(L, idx, Base::Type::xform)) luaL_argerror(L, idx, msg);
    return static_cast<Transform *>(lua_touserdata(L, idx));
}

Path *check_path(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::path)) luaL_argerror(L, idx, msg);
    return static_cast<Path *>(lua_touserdata(L, idx));
}

Image *check_image(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::image)) luaL_argerror(L, idx, msg);
    return static_cast<Image *>(lua_touserdata(L, idx));
}

Viewport *check_viewport(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::viewport)) luaL_argerror(L, idx, msg);
    return static_cast<Viewport *>(lua_touserdata(L, idx));
}

Window *check_window(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::window)) luaL_argerror(L, idx, msg);
    return static_cast<Window *>(lua_touserdata(L, idx));
}

Paint *check_paint_or_color(lua_State *L, int idx) {
    if (is_type(L, idx, Base::Type::color)) {
        // promote color to solid opaque paint, populate cache, return
        idx = abs_index(L, idx);
        push_paint_solid(L, check_color(L, idx), 1.f);
        lua_replace(L, idx);
        return static_cast<Paint *>(lua_touserdata(L, idx));
    } else {
        if (!is_type(L, idx, Base::Type::paint))
            luaL_argerror(L, idx, "expected paint or color");
        return static_cast<Paint *>(lua_touserdata(L, idx));
    }
}

Paint *check_paint(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::paint)) luaL_argerror(L, idx, msg);
    return static_cast<Paint *>(lua_touserdata(L, idx));
}

Color *check_color(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::color)) luaL_argerror(L, idx, msg);
    return static_cast<Color *>(lua_touserdata(L, idx));
}

scene::Element *check_element(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::element)) luaL_argerror(L, idx, msg);
    return static_cast<scene::Element *>(lua_touserdata(L, idx));
}

Scene *check_scene(lua_State *L, int idx, const char *msg) {
    if (!is_type(L, idx, Base::Type::scene)) luaL_argerror(L, idx, msg);
    return static_cast<Scene *>(lua_touserdata(L, idx));
}

void push_paint(lua_State *L, const Paint &paint) {
    void *p = lua_newuserdata(L, sizeof(Paint));
    new (p) Paint(paint);
    set_type(L, Base::Type::paint);
}

static void push_image(lua_State *L, const Image &image) {
    void *i = lua_newuserdata(L, sizeof(Image));
    new (i) Image(image);
    set_type(L, Base::Type::image);
}

static void push_path(lua_State *L, const Path &path) {
    void *p = lua_newuserdata(L, sizeof(Path));
    new (p) Path(path);
    set_type(L, Base::Type::path);
}

static void push_ramp(lua_State *L, const paint::Ramp &ramp) {
    void *p = lua_newuserdata(L, sizeof(paint::Ramp));
    new (p) paint::Ramp(ramp);
    set_type(L, Base::Type::ramp);
}

static void push_scene(lua_State *L, const Scene &scene) {
    void *p = lua_newuserdata(L, sizeof(Scene));
    new (p) Scene(scene);
    set_type(L, Base::Type::scene);
}

void open_type(lua_State *L, driver::Base *drv, Base::Type type,
    const luaL_Reg *create, const luaL_Reg *meta, const luaL_Reg *index) {
    // starts with package,consts
    lua_pushinteger(L, static_cast<int>(type)); // package,consts,type
    lua_newtable(L); // package,consts,type,meta
    if (index) {
        lua_pushliteral(L, "__index"); // package,consts,type,meta,"idx"
        lua_newtable(L); // package,consts,type,meta,"idx",idx
        lua_pushvalue(L, -5); // package,consts,type,meta,"idx",idx,consts
        lua_pushlightuserdata(L, drv);
        // package,consts,type,meta,"idx",idx,consts,drv
        luaL_setfuncs(L, index, 2); // package,consts,type,meta,"idx",idx
        lua_rawset(L, -3); // package,consts,type,meta
    }
    if (meta) {
        lua_pushvalue(L,-3); // package,consts,type,meta,consts
        lua_pushlightuserdata(L, drv); // package,consts,type,meta,consts,drv
        // get potentiall __index table and sets as last upvalue
        lua_pushstring(L, "__index");
        // package,consts,type,meta,consts,drv,"__index"
        lua_rawget(L, -4);
        if (lua_isnil(L, -1)) { // create empty table if not found
            lua_pop(L, 1);
            lua_newtable(L);
        }
        luaL_setfuncs(L, meta, 3); // package,consts,type,meta
    }
    lua_pushvalue(L, -1); // package,consts,type,meta,meta
    lua_pushvalue(L, -3); // package,consts,type,meta,meta,type
    lua_rawset(L, -5); // package,consts,type,meta
    lua_rawset(L, -3); // package,consts
    if (create) {
        lua_pushvalue(L, -2); // package,consts,package
        lua_pushvalue(L, -2); // package,consts,package,consts
        lua_pushlightuserdata(L, drv);// package,consts,package,consts,drv
        luaL_setfuncs(L, create, 2); // package,consts,package
        lua_pop(L, 1); // package,consts
    }
}

static const_Reg<scene::Element::Type> element_types[] = {
    {"nzfill", scene::Element::Type::nzfill},
    {"zfill", scene::Element::Type::zfill},
    {"ofill", scene::Element::Type::ofill},
    {"efill", scene::Element::Type::efill},
    {"nzclip", scene::Element::Type::nzclip},
    {"zclip", scene::Element::Type::zclip},
    {"oclip", scene::Element::Type::oclip},
    {"eclip", scene::Element::Type::eclip},
    {"pushclip", scene::Element::Type::pushclip},
    {"popclip", scene::Element::Type::popclip},
    {"pushfill", scene::Element::Type::pushfill},
    {"popfill", scene::Element::Type::popfill},
    {"commit", scene::Element::Type::commit}
};

static const_Reg<JoinStyle> join_styles[] = {
    {"miter", JoinStyle::miter},
    {"round", JoinStyle::round},
    {"bevel", JoinStyle::bevel}
};

static const_Reg<CapStyle> cap_styles[] = {
    {"butt", CapStyle::butt},
    {"round", CapStyle::round},
    {"square", CapStyle::square},
    {"triangle", CapStyle::triangle}
};

static const_Reg<BlendMode> blend_modes[] = {
    {"src", BlendMode::src},
    {"src_over", BlendMode::src_over},
    {"dst_over", BlendMode::dst_over},
    {"src_in", BlendMode::src_in},
    {"dst_in", BlendMode::dst_in},
    {"multiply", BlendMode::multiply},
    {"screen", BlendMode::screen},
    {"darken", BlendMode::darken},
    {"lighten", BlendMode::lighten},
    {"additive", BlendMode::additive}
};

static const_Reg<SpreadMethod> spread_methods[] = {
    {"clamp", SpreadMethod::pad},
    {"pad", SpreadMethod::pad},
    {"mod", SpreadMethod::repeat},
    {"repeat", SpreadMethod::repeat},
    {"mirror", SpreadMethod::reflect},
    {"reflect", SpreadMethod::reflect},
    {"transparent", SpreadMethod::transparent},
    {nullptr, SpreadMethod::pad}
};

static const_Reg<CommandType> command_types[] = {
    {"close_path_abs", CommandType::close_path_abs},
    {"close_path_rel", CommandType::close_path_rel},
    {"hline_to_abs",   CommandType::hline_to_abs},
    {"hline_to_rel",   CommandType::hline_to_rel},
    {"vline_to_abs",   CommandType::vline_to_abs},
    {"vline_to_rel",   CommandType::vline_to_rel},
    {"move_to_abs",    CommandType::move_to_abs},
    {"move_to_rel",    CommandType::move_to_rel},
    {"line_to_abs",    CommandType::line_to_abs},
    {"line_to_rel",    CommandType::line_to_rel},
    {"quad_to_abs",    CommandType::quad_to_abs},
    {"quad_to_rel",    CommandType::quad_to_rel},
    {"squad_to_abs",   CommandType::squad_to_abs},
    {"squad_to_rel",   CommandType::squad_to_rel},
    {"rquad_to_abs",   CommandType::rquad_to_abs},
    {"rquad_to_rel",   CommandType::rquad_to_rel},
    {"cubic_to_abs",   CommandType::cubic_to_abs},
    {"cubic_to_rel",   CommandType::cubic_to_rel},
    {"scubic_to_abs",  CommandType::scubic_to_abs},
    {"scubic_to_rel",  CommandType::scubic_to_rel},
    {"svgarc_to_abs",  CommandType::svgarc_to_abs},
    {"svgarc_to_rel",  CommandType::svgarc_to_rel},
    {nullptr,          CommandType::close_path_abs}
};

const char *Base::name(Type type) {
    switch (type) {
        case Type::quadtree: return "quadtree";
        case Type::path: return "path";
        case Type::path_iterator: return "path_iterator";
        case Type::xform: return "xform";
        case Type::ramp: return "ramp";
        case Type::surface: return "surface";
        case Type::viewport: return "viewport";
        case Type::window: return "window";
        case Type::scene: return "scene";
        case Type::scene_iterator: return "scene_iterator";
        case Type::color: return "color";
        case Type::paint: return "paint";
        case Type::p2: return "p2";
        case Type::p3: return "p2";
        case Type::stroke_style: return "stroke_style";
        case Type::dash: return "dash";
        case Type::identity: return "identity";
        case Type::command_type: return "command";
        case Type::spread_method: return "spread";
        case Type::cap_style: return "cap";
        case Type::join_style: return "join";
        case Type::fill_rule: return "fill";
        case Type::blend_mode: return "blend";
        case Type::element_type: return "element_type";
        case Type::element: return "element";
        case Type::last: return "last";
        default: assert(0); return nullptr;
    }
}

void Base::do_close(lua_State *L) {
    (void) L;
}

static int package_stroke_style(lua_State *L) {
    stroke::Style s = check_stroke_style(L, 1);
    void *p = lua_newuserdata(L, sizeof(stroke::Style));
    new (p) stroke::Style(s);
    set_type(L, Base::Type::stroke_style);
    return 1;
}

static const luaL_Reg stroke_style_create[] = {
  {"stroke_style", package_stroke_style},
  {nullptr, nullptr}
};

static int meta_tostring_stroke_style(lua_State *L) {
    stroke::Style *s = static_cast<stroke::Style *>(lua_touserdata(L, 1));
    std::ostringstream sout;
    sout << *s;
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static int meta_gc_stroke_style(lua_State *L) {
    stroke::Style *s = static_cast<stroke::Style *>(lua_touserdata(L, 1));
    s->~Style();
    return 0;
}

static void push_dash(lua_State *L, const stroke::Dash &dash) {
    lua_createtable(L, dash.array.size(), 2);
    for (unsigned i = 0; i < dash.array.size(); i++) {
        lua_pushnumber(L, dash.array[i]);
        lua_rawseti(L, -2, i+1);
    }
    lua_pushliteral(L, "initial_phase");
    lua_pushnumber(L, dash.initial_phase);
    lua_rawset(L, -3);
    lua_pushliteral(L, "phase_reset");
    lua_pushboolean(L, dash.phase_reset);
    lua_rawset(L, -3);
}

static int meta_index_stroke_style(lua_State *L) {
    stroke::Style *s = static_cast<stroke::Style *>(lua_touserdata(L, 1));
    const char *lst[] = {"width", "dash", "cap", "join",
        "miter_limit", nullptr};
    switch (luaL_checkoption(L, 2, nullptr, lst)) {
        case 0:
            lua_pushnumber(L, s->width);
            return 1;
        case 1:
            push_dash(L, s->dash);
            return 1;
        case 2:
            push_const(L, Base::Type::cap_style, s->cap);
            return 1;
        case 3:
            push_const(L, Base::Type::join_style, s->join);
            return 1;
        case 4:
            lua_pushnumber(L, s->miter_limit);
            return 1;
        case 5:
            lua_pushnumber(L, s->dash.initial_phase);
            return 1;
        case 6:
            lua_pushboolean(L, s->dash.phase_reset);
            return 1;
        default:
            return 0;
    }
}

static int meta_newindex_stroke_style(lua_State *L) {
    stroke::Style *s = static_cast<stroke::Style *>(lua_touserdata(L, 1));
    const char *lst[] = {"width", "dash", "cap", "join",
        "miter_limit", nullptr};
    switch (luaL_checkoption(L, 2, nullptr, lst)) {
        case 0:
            s->width = luaL_checknumber(L, 3);
            return 0;
        case 1:
            s->dash = check_dash(L, 3);
            return 0;
        case 2:
            s->cap = check_cap_style(L, 3);
            return 0;
        case 3:
            s->join = check_join_style(L, 3);
            return 0;
        case 4:
            s->miter_limit = luaL_checknumber(L, 3);
            return 0;
        default:
            return 0;
    }
}

static const luaL_Reg stroke_style_meta[] = {
  {"__gc", meta_gc_stroke_style},
  {"__tostring", meta_tostring_stroke_style},
  {"__index", meta_index_stroke_style},
  {"__newindex", meta_newindex_stroke_style},
  {nullptr, nullptr}
};

void Base::do_open_stroke_style(lua_State *L) {
    open_type(L, driver(), Base::Type::stroke_style, stroke_style_create,
        stroke_style_meta);
}

void Base::do_open(lua_State *L) {
    // We should change this so all open*.* functions are
    // called with lua_pcall so the enviroment is the same
    // for all functions dealing with Lua. Right now, we
    // distinguish between these open*.* functions and the
    // remaining functions. This is awkeard, error-prone,
    // and leads to duplication of code. ??
    open_consts(L, Base::Type::command_type, command_types);
    open_consts(L, Base::Type::spread_method, spread_methods);
    open_consts(L, Base::Type::cap_style, cap_styles);
    open_consts(L, Base::Type::join_style, join_styles);
    open_consts(L, Base::Type::blend_mode, blend_modes);
    open_consts(L, Base::Type::element_type, element_types);
    // all consts open before objects that may depend on them
    do_open_path(L);
    do_open_xform(L);
    do_open_quadtree(L);
    do_open_ramp(L);
    do_open_color(L);
    do_open_paint(L);
    do_open_image(L);
    do_open_scene(L);
    do_open_element(L);
    do_open_surface(L);
    do_open_point(L);
    do_open_viewport(L);
    do_open_window(L);
    do_open_stroke_style(L);
}

static int meta_tostring_xform(lua_State *L) {
    Transform *p = check_xform(L, 1);
    std::ostringstream sout;
    sout << *p;
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static int meta_gc_xform(lua_State *L) {
    Transform *p = check_xform(L, 1);
    p->~Transform();
    return 0;
}

static int meta_mul_xform(lua_State *L) {
    Transform *p = check_xform(L, 1);
    Transform *q = check_xform(L, 2);
    void *r = lua_newuserdata(L, sizeof(Transform));
    new (r) Transform((*p) * (*q));
    set_type(L, Base::Type::xform);
    return 1;
}

static Transform *new_xform_translation(lua_State *L, int base) {
    float tx = luaL_checknumber(L, base);
    float ty = luaL_optnumber(L, base+1, 0.f);
    void *p = lua_newuserdata(L, sizeof(Transform));
    new (p) Transform(xform::Translation(tx, ty));
    set_type(L, Base::Type::xform);
    return static_cast<Transform *>(p);
}

static int package_xform_translation(lua_State *L) {
    new_xform_translation(L, 1);
    return 1;
}

static Transform *new_xform_rotation(lua_State *L, int base) {
    void *p = nullptr;
    if (lua_gettop(L) > base) {
        float c = luaL_checknumber(L, base);
        float s = luaL_checknumber(L, base+1);
        p = lua_newuserdata(L, sizeof(Transform));
        new (p) Transform(xform::Rotation(c, s));
    } else {
        float a = luaL_checknumber(L, base);
        p = lua_newuserdata(L, sizeof(Transform));
        new (p) Transform(xform::Rotation(math::to_rad()*a));
    }
    set_type(L, Base::Type::xform);
    return static_cast<Transform *>(p);
}

static int package_xform_rotation(lua_State *L) {
    new_xform_rotation(L, 1);
    return 1;
}

static Transform *new_xform_scale(lua_State *L, int base) {
    void *p = nullptr;
    if (lua_gettop(L) > base) {
        float sx = luaL_checknumber(L, base);
        float sy = luaL_checknumber(L, base+1);
        p = lua_newuserdata(L, sizeof(Transform));
        new (p) Transform(xform::Scale(sx, sy));
    } else {
        float s = luaL_checknumber(L, base);
        p = lua_newuserdata(L, sizeof(Transform));
        new (p) Transform(xform::Scale(s, s));
    }
    set_type(L, Base::Type::xform);
    return static_cast<Transform *>(p);
}

static int package_xform_scale(lua_State *L) {
    new_xform_scale(L, 1);
    return 1;
}

static int package_xform_identity(lua_State *L) {
    // return identity from constants table
    lua_rawgeti(L, lua_upvalueindex(1), static_cast<int>(Base::Type::identity));
    return 1;
}

static Transform *new_xform_linear(lua_State *L, int base) {
    float a11 = luaL_checknumber(L, base++);
    float a12 = luaL_checknumber(L, base++);
    float a21 = luaL_checknumber(L, base++);
    float a22 = luaL_checknumber(L, base);
    void *p = lua_newuserdata(L, sizeof(Transform));
    new (p) Transform(xform::Linear(a11, a12, a21, a22));
    set_type(L, Base::Type::xform);
    return static_cast<Transform *>(p);
}

static int package_xform_linear(lua_State *L) {
    new_xform_linear(L, 1);
    return 1;
}

static Transform *new_xform_affine(lua_State *L, int base) {
    float a11 = luaL_checknumber(L, base++);
    float a12 = luaL_checknumber(L, base++);
    float a13 = luaL_checknumber(L, base++);
    float a21 = luaL_checknumber(L, base++);
    float a22 = luaL_checknumber(L, base++);
    float a23 = luaL_checknumber(L, base);
    void *p = lua_newuserdata(L, sizeof(Transform));
    new (p) Transform(xform::Affine(a11, a12, a13,
        a21, a22, a23));
    set_type(L, Base::Type::xform);
    return static_cast<Transform *>(p);
}

static int package_xform_affine(lua_State *L) {
    new_xform_affine(L, 1);
    return 1;
}

static Transform *new_xform_windowviewport(lua_State *L, float xw, float yw, float Xw, float Yw,
    float xv, float yv, float Xv, float Yv) {
    void *p = lua_newuserdata(L, sizeof(Transform));
    float inv_dwx = (xw-Xw);
    if (std::abs(inv_dwx) < FLT_MIN) luaL_error(L, "degenerate window");
    inv_dwx = 1.f/inv_dwx;
    float inv_dwy = (yw-Yw);
    if (std::abs(inv_dwy) < FLT_MIN) luaL_error(L, "degenerate window");
    inv_dwy = 1.f/inv_dwy;
    new (p) Transform(xform::Affine(
        (xv-Xv)*inv_dwx, 0.f,              (Xv*xw - xv*Xw)*inv_dwx,
        0.f,             (yv-Yv)*inv_dwy,  (Yv*yw - yv*Yw)*inv_dwy
    ));
    set_type(L, Base::Type::xform);
    return static_cast<Transform *>(p);
}

static Transform *new_xform_windowviewport(lua_State *L, int base) {
    Window *w = check_window(L, base++);
    Viewport *v = check_viewport(L, base);
    return new_xform_windowviewport(L, w->min.x, w->min.y, w->max.x, w->max.y,
        v->min.x, v->min.y, v->max.x, v->max.y);
}

static int package_xform_windowviewport(lua_State *L) {
    new_xform_windowviewport(L, 1);
    return 1;
}

static Transform *new_xform_foreshorten(lua_State *L, float xm, float ym, float xM, float yM, float d) {
    void *p = lua_newuserdata(L, sizeof(Transform));
    new (p) Transform(xform::Projective(
        (-1.f + 2.f*d - xm + xM)*(1.f + ym - yM),
        -d*(-1.f + xm + xM),
        d*(-1.f + xm + xM)*(-1.f + yM),
        0.f,
        -(1.f + xm - xM)*(1.f + ym - yM) - 2.f*d*(-1.f + yM), 2.f*d*ym*(-1.f + yM),
        0.f,
        -2.f*d,
        -1.f - ym + 2*d*ym + xM*(1.f + ym - yM) + yM + xm*(-1.f - ym + yM))
    );
    set_type(L, Base::Type::xform);
    return static_cast<Transform *>(p);
}

static Transform *new_xform_foreshorten(lua_State *L, int base) {
    Viewport *v = check_viewport(L, base++);
    float d = luaL_checknumber(L, base);
    float xm = v->min.x;
    float ym = v->min.y;
    float xM = v->max.x;
    float yM = v->max.y;
    return new_xform_foreshorten(L, xm, ym, xM, yM, d);
}

static int package_xform_foreshorten(lua_State *L) {
    new_xform_foreshorten(L, 1);
    return 1;
}

static Transform *new_xform_projective(lua_State *L, int base) {
    float a11 = luaL_checknumber(L, base++);
    float a12 = luaL_checknumber(L, base++);
    float a13 = luaL_checknumber(L, base++);
    float a21 = luaL_checknumber(L, base++);
    float a22 = luaL_checknumber(L, base++);
    float a23 = luaL_checknumber(L, base++);
    float a31 = luaL_checknumber(L, base++);
    float a32 = luaL_checknumber(L, base++);
    float a33 = luaL_checknumber(L, base);
    void *p = lua_newuserdata(L, sizeof(Transform));
    new (p) Transform(xform::Projective(a11, a12, a13,
        a21, a22, a23, a31, a32, a33));
    set_type(L, Base::Type::xform);
    return static_cast<Transform *>(p);
}

static int package_xform_projective(lua_State *L) {
    new_xform_projective(L, 1);
    return 1;
}

static const luaL_Reg xform_create[] = {
  {"identity", package_xform_identity},
  {"translation", package_xform_translation},
  {"rotation", package_xform_rotation},
  {"scale", package_xform_scale},
  {"linear", package_xform_linear},
  {"affine", package_xform_affine},
  {"foreshorten", package_xform_foreshorten},
  {"windowviewport", package_xform_windowviewport},
  {"projective", package_xform_projective},
  {nullptr, nullptr}
};

static const luaL_Reg xform_meta[] = {
  {"__gc", meta_gc_xform},
  {"__tostring", meta_tostring_xform},
  {"__mul", meta_mul_xform},
  {nullptr, nullptr}
};

void Base::do_open_xform(lua_State *L) {
    open_type(L, driver(), Base::Type::xform, xform_create, xform_meta);
    // hang an identity in the constants table
    // note that we don't have upvalues here, so we can't
    // use the set_type function.
    void *p = lua_newuserdata(L, sizeof(Transform));
    new (p) Transform;
    lua_rawgeti(L, -2, static_cast<int>(Base::Type::xform));
    lua_setmetatable(L, -2);
    lua_rawseti(L, -2, static_cast<int>(Base::Type::identity));
}

static int meta_tostring_path(lua_State *L) {
    Path *p = check_path(L, 1);
    std::ostringstream sout;
    sout << *p;
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static int meta_gc_path(lua_State *L) {
    Path *p = check_path(L, 1);
    p->~Path();
    return 0;
}

static float get_number(lua_State *L, int tab, int i) {
    lua_rawgeti(L, tab, i);
    if (!lua_isnumber(L, -1))
        luaL_error(L, "path entry %d should be a number", i);
    float v = static_cast<float>(lua_tonumber(L, -1));
    lua_pop(L, 1);
    return v;
}

static bool get_boolean(lua_State *L, int tab, int i) {
    lua_rawgeti(L, tab, i);
    if (lua_isboolean(L, -1)) {
        bool v = lua_toboolean(L, -1);
        lua_pop(L, 1);
        return v;
    } else if (lua_isnumber(L, -1)) {
        bool v = (lua_tonumber(L, -1) != 0.f);
        lua_pop(L, 1);
        return v;
    } else {
        luaL_error(L, "path entry %d should be a boolean", i);
        return false; // never reaches here
    }
}

static bool is_number(lua_State *L, int tab, int i) {
    lua_rawgeti(L, tab, i);
    bool ret = lua_type(L, -1) == LUA_TNUMBER;
    lua_pop(L, 1);
    return ret;
}

static Color *get_color(lua_State *L, int tab, int i) {
    lua_rawgeti(L, tab, i); // meta, obj
    lua_getmetatable(L, -1); // meta, obj, meta
    if (!is_equal(L, -1, -3))
        luaL_error(L, "ramp entry %d should be a color", i);
    Color *c = static_cast<Color *>(lua_touserdata(L, -2));
    lua_pop(L, 2); // meta
    return c;
}

static CommandType get_command(lua_State *L, int tab, int i) {
    lua_rawgeti(L, tab, i); // meta, obj
    lua_getmetatable(L, -1); // meta, obj, meta
    if (!is_equal(L, -1, -3))
        luaL_error(L, "path entry %d should be a command", i);
    CommandType ct = *static_cast<CommandType *>(lua_touserdata(L, -2));
    lua_pop(L, 2); // meta
    return ct;
}

static int method_transform_path(lua_State *L) {
    Path p = *check_path(L, 1);
    Transform *t = check_xform(L, 2);
    push_path(L, p.transform(*t));
    return 1;
}

static int method_stroke_path(lua_State *L) {
    Path p = *check_path(L, 1);
    stroke::Style s = check_stroke_style(L, 2);
    float atol = luaL_optnumber(L, 3, 5.f);
    float swtol = luaL_optnumber(L, 4, .5f);
    int maxdiv = luaL_optint(L, 5, 10);
    push_path(L, p.stroke(s, atol, swtol, maxdiv));
    return 1;
}

static int method_translate_path(lua_State *L) {
    Path p = *check_path(L, 1);
    Transform *t = new_xform_translation(L, 2);
    push_path(L, p.transform(*t));
    return 1;
}

static int method_scale_path(lua_State *L) {
    Path p = *check_path(L, 1);
    Transform *t = new_xform_scale(L, 2);
    push_path(L, p.transform(*t));
    return 1;
}

static int method_affine_path(lua_State *L) {
    Path p = *check_path(L, 1);
    Transform *t = new_xform_affine(L, 2);
    push_path(L, p.transform(*t));
    return 1;
}

static int method_linear_path(lua_State *L) {
    Path p = *check_path(L, 1);
    Transform *t = new_xform_linear(L, 2);
    push_path(L, p.transform(*t));
    return 1;
}

static int method_rotate_path(lua_State *L) {
    Path p = *check_path(L, 1);
    Transform *t = new_xform_rotation(L, 2);
    push_path(L, p.transform(*t));
    return 1;
}

static int method_windowviewport_path(lua_State *L) {
    Path p = *check_path(L, 1);
    Transform *t = new_xform_windowviewport(L, 2);
    push_path(L, p.transform(*t));
    return 1;
}

// we could speed this function a little by merging the
// checking in get_number and is_number, but I think it is
// overkill
static void append_path(lua_State *L, int base, Path *p) {
    using namespace path;
    if (lua_istable(L, base)) {
        int i = 1;
        int n = table_len(L, -1);
        // put command metatable on top of stack
        // get_command depends on it
        get_type_metatable(L, Base::Type::command_type); // input,output,meta
        while (i <= n) {
            CommandType ct = get_command(L, base, i);
            switch (ct) {
                case CommandType::close_path_abs:
                case CommandType::close_path_rel:
                    i++;
                    p->append(ct);
                    break;
                case CommandType::hline_to_abs:
                case CommandType::hline_to_rel:
                case CommandType::vline_to_abs:
                case CommandType::vline_to_rel:
                    i++;
                    do {
                        p->append(Command{ct, get_number(L, base, i)});
                        i++;
                    } while (i <= n && is_number(L, base, i));
                    break;
                case CommandType::move_to_abs: {
                    i++;
                    float x0 = get_number(L, base, i++);
                    float y0 = get_number(L, base, i++);
                    p->append(Command{ct, x0, y0});
                    while (is_number(L, base, i) && i <= n) {
                       x0 = get_number(L, base, i++);
                       y0 = get_number(L, base, i++);
                       p->append(Command{CommandType::line_to_abs, x0, y0});
                    }
                    break;
                }
                case CommandType::move_to_rel: {
                    i++;
                    float x0 = get_number(L, base, i++);
                    float y0 = get_number(L, base, i++);
                    p->append(Command{ct, x0, y0});
                    while (is_number(L, base, i) && i <= n) {
                       x0 = get_number(L, base, i++);
                       y0 = get_number(L, base, i++);
                       p->append(Command{CommandType::line_to_rel, x0, y0});
                    }
                    break;
                }
                case CommandType::line_to_abs:
                case CommandType::line_to_rel:
                case CommandType::squad_to_abs:
                case CommandType::squad_to_rel:
                    i++;
                    do {
                        float x0 = get_number(L, base, i++);
                        float y0 = get_number(L, base, i++);
                        p->append(Command{ct, x0, y0});
                    } while (i <= n && is_number(L, base, i));
                    break;
                case CommandType::quad_to_abs:
                case CommandType::quad_to_rel:
                case CommandType::scubic_to_abs:
                case CommandType::scubic_to_rel:
                    i++;
                    do {
                        float x0 = get_number(L, base, i++);
                        float y0 = get_number(L, base, i++);
                        float x1 = get_number(L, base, i++);
                        float y1 = get_number(L, base, i++);
                        p->append(Command{ct, x0, y0, x1, y1});
                    } while (i <= n && is_number(L, base, i));
                    break;
                case CommandType::rquad_to_abs:
                case CommandType::rquad_to_rel:
                    i++;
                    do {
                        float x0 = get_number(L, base, i++);
                        float y0 = get_number(L, base, i++);
                        float w0 = get_number(L, base, i++);
                        float x1 = get_number(L, base, i++);
                        float y1 = get_number(L, base, i++);
                        p->append(Command{ct, x0, y0, w0, x1, y1});
                    } while (i <= n && is_number(L, base, i));
                    break;
                case CommandType::svgarc_to_abs:
                case CommandType::svgarc_to_rel:
                    i++;
                    do {
                        float rx = get_number(L, base, i++);
                        float ry = get_number(L, base, i++);
                        float alpha_deg = get_number(L, base, i++);
                        bool fa = get_boolean(L, base, i++);
                        bool fs = get_boolean(L, base, i++);
                        float fa_fs = float((int(fa) << 1) | int(fs));
                        float x1 = get_number(L, base, i++);
                        float y1 = get_number(L, base, i++);
                        p->append(Command{ct, rx, ry, alpha_deg, fa_fs, x1, y1});
                    } while (i <= n && is_number(L, base, i));
                    break;
                case CommandType::cubic_to_abs:
                case CommandType::cubic_to_rel:
                    i++;
                    do {
                        float x0 = get_number(L, base, i++);
                        float y0 = get_number(L, base, i++);
                        float x1 = get_number(L, base, i++);
                        float y1 = get_number(L, base, i++);
                        float x2 = get_number(L, base, i++);
                        float y2 = get_number(L, base, i++);
                        p->append(Command{ct, x0, y0, x1, y1, x2, y2});
                    } while (i <= n && is_number(L, base, i));
                    break;
                default:
                    luaL_error(L, "unknown command type at %d", i);
                    break;
            }
        }
    } else {
        Path *q = check_path(L, base, "expected path or command list");
        p->append(*q);
    }
}

static int method_append_path(lua_State *L) {
    Path p = *check_path(L, 1);
    append_path(L, 2, &p);
    push_path(L, p);
    return 1;
}

static int package_path(lua_State *L) {
    Path p = get_driver(L)->new_path();
    append_path(L, 1, &p);
    push_path(L, p);
    return 1;
}

static int package_viewport(lua_State *L) {
    int xmin = luaL_checkinteger(L, 1);
    int ymin = luaL_checkinteger(L, 2);
    int xmax = luaL_checkinteger(L, 3);
    int ymax = luaL_checkinteger(L, 4);
    void *p = lua_newuserdata(L, sizeof(Viewport));
    new (p) Viewport(xmin, ymin, xmax, ymax);
    set_type(L, Base::Type::viewport);
    return 1;
}

static int package_window(lua_State *L) {
    int xmin = luaL_checknumber(L, 1);
    int ymin = luaL_checknumber(L, 2);
    int xmax = luaL_checknumber(L, 3);
    int ymax = luaL_checknumber(L, 4);
    void *p = lua_newuserdata(L, sizeof(Window));
    new (p) Window(xmin, ymin, xmax, ymax);
    set_type(L, Base::Type::window);
    return 1;
}

static int meta_index_viewport(lua_State *L) {
    Viewport *v = check_viewport(L, 1);
    const char *lst[] = {"min", "max", nullptr};
    void *p = lua_newuserdata(L, sizeof(r2::Point));
    set_type(L, Base::Type::p2);
    switch (luaL_checkoption(L, 2, nullptr, lst)) {
        case 0: {
            new (p) r2::Point(v->min.x, v->min.y);
            return 1;
        }
        case 1:
            new (p) r2::Point(v->max.x, v->max.y);
            return 1;
        default:
            return 0;
    }
}

static int meta_gc_viewport(lua_State *L) {
    Viewport *p = check_viewport(L, 1);
    p->~Viewport();
    return 0;
}

static int meta_gc_window(lua_State *L) {
    Window *p = check_window(L, 1);
    p->~Window();
    return 0;
}

static int meta_tostring_viewport(lua_State *L) {
    Viewport *p = check_viewport(L, 1);
    lua_pushfstring(L, "viewport %d,%d %d,%d", p->min.x, p->min.y,
        p->max.x, p->max.y);
    return 1;
}

static int meta_tostring_window(lua_State *L) {
    Window *p = check_window(L, 1);
    lua_pushfstring(L, "window %f,%f %f,%f", p->min.x, p->min.y,
        p->max.x, p->max.y);
    return 1;
}

static const luaL_Reg viewport_meta[] = {
  {"__gc", meta_gc_viewport},
  {"__tostring", meta_tostring_viewport},
  {"__index", meta_index_viewport},
  {nullptr, nullptr}
};

static const luaL_Reg viewport_create[] = {
  {"viewport", package_viewport},
  {nullptr, nullptr}
};

void Base::do_open_viewport(lua_State *L) {
    open_type(L, driver(), Base::Type::viewport, viewport_create, viewport_meta);
}

static int meta_index_window(lua_State *L) {
    Window *v = check_window(L, 1);
    const char *lst[] = {"min", "max", nullptr};
    void *p = lua_newuserdata(L, sizeof(r2::Point));
    set_type(L, Base::Type::p2);
    switch (luaL_checkoption(L, 2, nullptr, lst)) {
        case 0: {
            new (p) r2::Point(v->min.x, v->min.y);
            return 1;
        }
        case 1:
            new (p) r2::Point(v->max.x, v->max.y);
            return 1;
        default:
            return 0;
    }
}

static const luaL_Reg window_meta[] = {
  {"__gc", meta_gc_window},
  {"__tostring", meta_tostring_window},
  {"__index", meta_index_window},
  {nullptr, nullptr}
};

static const luaL_Reg window_create[] = {
  {"window", package_window},
  {nullptr, nullptr}
};

void Base::do_open_window(lua_State *L) {
    open_type(L, driver(), Base::Type::window, window_create, window_meta);
}

void Base::do_open_quadtree(lua_State *L) {
    (void) L;
}

static int meta_gc_ramp(lua_State *L) {
    paint::Ramp *r = check_ramp(L, 1);
    r->~Ramp();
    return 0;
}

static void append_ramp(lua_State *L, int base, paint::Ramp *r) {
    if (lua_gettop(L) < base) return;
    luaL_checktype(L, base, LUA_TTABLE);
    int i = 1;
    int n = table_len(L, -1);
    float last_t = r->stops().empty()? 0.f: r->stops().back().offset;
    // put color metatable on top of stack
    // get_color depends on it
    get_type_metatable(L, Base::Type::color); // input,output,meta
    while (i <= n) {
        float t = get_number(L, base, i);
        if (t < last_t || t < 0.f || t > 1.f)
            luaL_error(L, "invalid stop offset %f at entry %d", t, i);
        Color *c = get_color(L, base, i+1);
        r->append(paint::ramp::Stop(t, c->red, c->green, c->blue, c->opacity));
        last_t = t;
        i += 2;
    }
    // check spread and set accordingly
    lua_pushliteral(L, "spread");
    lua_rawget(L, base);
    if (is_type(L, -1, Base::Type::spread_method)) {
        r->set_spread(*static_cast<SpreadMethod *>(lua_touserdata(L, -1)));
    } else if (!lua_isnil(L, -1)) {
        luaL_error(L, "invalid spread method");
    }
}

static int method_append_ramp(lua_State *L) {
    paint::Ramp r = *check_ramp(L, 1);
    append_ramp(L, 2, &r);
    push_ramp(L, r);
    return 1;
}

static int package_ramp(lua_State *L) {
    paint::Ramp r = get_driver(L)->new_color_ramp(SpreadMethod::pad,{});
    append_ramp(L, 1, &r);
    push_ramp(L, r);
    return 1;
}

static int meta_tostring_ramp(lua_State *L) {
    paint::Ramp *r = check_ramp(L, 1);
    std::ostringstream sout;
    sout << *r;
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static const luaL_Reg ramp_create[] = {
  {"ramp", package_ramp},
  {nullptr, nullptr}
};

static const luaL_Reg ramp_meta[] = {
  {"__gc", meta_gc_ramp},
  {"__tostring", meta_tostring_ramp},
  {nullptr, nullptr}
};

static const luaL_Reg ramp_meta_index[] = {
  {"append", method_append_ramp},
  {nullptr, nullptr}
};

void Base::do_open_ramp(lua_State *L) {
    open_type(L, driver(), Base::Type::ramp, ramp_create,
        ramp_meta, ramp_meta_index);
}

static int push_command(lua_State *L, CommandType type, float *args) {
    push_const(L, Base::Type::command_type, type);
    switch (type) {
        case CommandType::close_path_abs:
        case CommandType::close_path_rel:
            return 1;
        case CommandType::hline_to_abs:
        case CommandType::hline_to_rel:
        case CommandType::vline_to_abs:
        case CommandType::vline_to_rel:
            lua_pushnumber(L, args[0]);
            return 2;
        case CommandType::move_to_abs:
        case CommandType::move_to_rel:
        case CommandType::line_to_abs:
        case CommandType::line_to_rel:
        case CommandType::squad_to_abs:
        case CommandType::squad_to_rel:
            lua_pushnumber(L, args[0]);
            lua_pushnumber(L, args[1]);
            return 3;
        case CommandType::quad_to_abs:
        case CommandType::quad_to_rel:
        case CommandType::scubic_to_abs:
        case CommandType::scubic_to_rel:
            lua_pushnumber(L, args[0]);
            lua_pushnumber(L, args[1]);
            lua_pushnumber(L, args[2]);
            lua_pushnumber(L, args[3]);
            return 5;
        case CommandType::rquad_to_abs:
        case CommandType::rquad_to_rel:
            lua_pushnumber(L, args[0]);
            lua_pushnumber(L, args[1]);
            lua_pushnumber(L, args[2]);
            lua_pushnumber(L, args[3]);
            lua_pushnumber(L, args[4]);
            return 6;
        case CommandType::cubic_to_abs:
        case CommandType::cubic_to_rel:
            lua_pushnumber(L, args[0]);
            lua_pushnumber(L, args[1]);
            lua_pushnumber(L, args[2]);
            lua_pushnumber(L, args[3]);
            lua_pushnumber(L, args[4]);
            lua_pushnumber(L, args[5]);
            return 7;
        default:
            // reaching here is a runtime error...
            assert(0);
            return 0;
    }
}

// first call receives path only
// remaining calls receive path, followed by nil, followed by iterator pair
// while there are still items,
//     returns the iterator pair, followed by the command and arguments
// when no items are left
//     returns nil to end iteration
int meta_call_path(lua_State *L) {
    Path *path = check_path(L, 1);
    PathIterator *it = nullptr;
    // first call
    if (lua_isnil(L, 3)) {
        lua_pop(L, 1);
        it = static_cast<PathIterator*>(
            lua_newuserdata(L, 2*sizeof(PathIterator)));
        new (it) PathIterator(path->begin());
        new (it+1) PathIterator(path->end());
        set_type(L, Base::Type::path_iterator);
    // remaining calls
    } else {
        it = check_path_iterators(L, 3);
    }
    // still loop to go
    if (it[0] != it[1]) {
        // get command
        Command c = *it[0];
        // move to next element
        it[0]++;
        // push iterator pair, command, and arguments
        lua_pushvalue(L, 3);
        return push_command(L, c.type, c.arguments) + 1;
    // terminate loop
    } else {
        lua_pushnil(L);
        return 1;
    }
}

static int meta_gc_path_iterator(lua_State *L) {
    PathIterator *it = check_path_iterators(L, 1);
    it[0].~Iterator();
    it[1].~Iterator();
    return 0;
}

static int meta_tostring_path_iterator(lua_State *L) {
    PathIterator *it = check_path_iterators(L, 1);
    lua_pushfstring(L, "[%d,%d)", it[0].index(), it[1].index());
    return 1;
}

static int method_flatten_path(lua_State *L) {
    Path p = *check_path(L, 1);
    float atol = luaL_optnumber(L, 2, 5.f);
    float ptol = luaL_optnumber(L, 3, FLT_MAX);
    int maxdiv = luaL_optint(L, 4, 10);
    push_path(L, p.flatten(atol, ptol, maxdiv));
    return 1;
}

static int method_subdiv_path(lua_State *L) {
    Path p = *check_path(L, 1);
    float atol = luaL_optnumber(L, 2, 5.f);
    float ptol = luaL_optnumber(L, 3, FLT_MAX);
    int maxdiv = luaL_optint(L, 4, 10);
    push_path(L, p.subdiv(atol, ptol, maxdiv));
    return 1;
}

static const luaL_Reg path_create[] = {
  {"path", package_path},
  {nullptr, nullptr}
};

static const luaL_Reg path_meta[] = {
  {"__gc", meta_gc_path},
  {"__tostring", meta_tostring_path},
  {"__call", meta_call_path},
  {nullptr, nullptr}
};

static const luaL_Reg path_meta_index[] = {
  {"stroke", method_stroke_path},
  {"append", method_append_path},
  {"transform", method_transform_path},
  {"translate", method_translate_path},
  {"scale", method_scale_path},
  {"rotate", method_rotate_path},
  {"affine", method_affine_path},
  {"linear", method_linear_path},
  {"windowviewport", method_windowviewport_path},
  {"flatten", method_flatten_path},
  {"subdiv", method_subdiv_path},
  {nullptr, nullptr}
};

static const luaL_Reg path_iterator_meta[] = {
  {"__gc", meta_gc_path_iterator},
  {"__tostring", meta_tostring_path_iterator},
  {nullptr, nullptr}
};

void Base::do_open_path(lua_State *L) {
    open_type(L, driver(), Base::Type::path, path_create,
        path_meta, path_meta_index);
    open_type(L, driver(), Base::Type::path_iterator, nullptr,
        path_iterator_meta);
    // promote M,L,Q,C,A,Z to package table
    static const_Reg<CommandType> promoted[] = {
        {"T", CommandType::squad_to_abs},
        {"t", CommandType::squad_to_rel},
        {"R", CommandType::rquad_to_abs},
        {"r", CommandType::rquad_to_rel},
        {"A", CommandType::svgarc_to_abs},
        {"a", CommandType::svgarc_to_rel},
        {"C", CommandType::cubic_to_abs},
        {"c", CommandType::cubic_to_rel},
        {"H", CommandType::hline_to_abs},
        {"h", CommandType::hline_to_rel},
        {"L", CommandType::line_to_abs},
        {"l", CommandType::line_to_rel},
        {"M", CommandType::move_to_abs},
        {"m", CommandType::move_to_rel},
        {"Q", CommandType::quad_to_abs},
        {"q", CommandType::quad_to_rel},
        {"S", CommandType::scubic_to_abs},
        {"s", CommandType::scubic_to_rel},
        {"V", CommandType::vline_to_abs},
        {"v", CommandType::vline_to_rel},
        {"Z", CommandType::close_path_abs},
        {"z", CommandType::close_path_rel},
        {nullptr, CommandType::close_path_abs}
    };
    // package,consts
    // get command table from consts
    lua_pushlightuserdata(L,
        reinterpret_cast<void *>(Base::Type::command_type));
    lua_rawget(L, -2); // package,consts,commandt
    for (auto c = &promoted[0]; c->name; c++) {
        lua_pushstring(L, c->name); // package,consts,commandt,name
        lua_rawgeti(L, -2, static_cast<int>(c->value)); 
        // package,consts,commandt,name,value
        lua_rawset(L, -5);
    }
    lua_pop(L, 1); // package,consts
}

void push_element(lua_State *L, const scene::Element &element) {
    void *e = lua_newuserdata(L, sizeof(scene::Element));
    new (e) scene::Element(element);
    set_type(L, Base::Type::element);
}

int meta_call_scene(lua_State *L) {
    Scene *s = check_scene(L, 1);
    SceneIterator *it = nullptr;
    // first call
    if (lua_isnil(L, 3)) {
        lua_pop(L, 1);
        it = static_cast<SceneIterator*>(
            lua_newuserdata(L, 2*sizeof(SceneIterator)));
        new (it) SceneIterator(s->begin());
        new (it+1) SceneIterator(s->end());
        set_type(L, Base::Type::scene_iterator);
    // remaining calls
    } else {
        it = check_scene_iterators(L, 3);
    }
    // still loop to go
    if (it[0] != it[1]) {
        // push iterator pair and element
        lua_pushvalue(L, 3);
        push_element(L, *it[0]);
        ++it[0];
        return 2;
    // terminate loop
    } else {
        lua_pushnil(L);
        return 1;
    }
}

static void append_scene(lua_State *L, int base, Scene *s) {
    if (lua_istable(L, base)) {
        int n = table_len(L, -1);
        for (int i = 1; i <= n; i++) {
            lua_rawgeti(L, base, i);
            if (!is_type(L, -1, Base::Type::element))
                luaL_error(L, "entry %d not an element", i);
            s->append(*static_cast<scene::Element *>(lua_touserdata(L, -1)));
            lua_pop(L, 1);
        }
    } else if (!lua_isnone(L, base)) {
        scene::Element *p = check_element(L, base, "expected element or list");
        s->append(*p);
    }
}

static int method_append_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    append_scene(L, 2, &s);
    push_scene(L, s);
    return 1;
}

static int package_scene(lua_State *L) {
    Scene s = get_driver(L)->new_scene();
    append_scene(L, 1, &s);
    push_scene(L, s);
    return 1;
}

static int meta_gc_scene(lua_State *L) {
    Scene *s = check_scene(L, 1);
    s->~Scene();
    return 0;
}

static int meta_tostring_scene(lua_State *L) {
    Scene *s = static_cast<Scene *>(lua_touserdata(L, 1));
    std::ostringstream sout;
    int i = 1;
    sout << "scene " << s->xf() << '\n';
    for (const auto &e: s->elements()) {
        sout << "  " << i << ' ' << e << '\n';
    }
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static int method_nzclip_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Path *p = check_path(L, 2);
    push_scene(L, s.nzclip(*p));
    return 1;
}

static int method_zclip_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Path *p = check_path(L, 2);
    push_scene(L, s.zclip(*p));
    return 1;
}

static int method_oclip_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Path *p = check_path(L, 2);
    push_scene(L, s.oclip(*p));
    return 1;
}

static int method_eclip_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Path *p = check_path(L, 2);
    push_scene(L, s.eclip(*p));
    return 1;
}

static int method_pop_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    push_scene(L, s.pop());
    return 1;
}

static int method_push_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    push_scene(L, s.push());
    return 1;
}

static int method_commit_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    push_scene(L, s.commit());
    return 1;
}

static int method_nzfill_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Path *p = check_path(L, 2);
    Paint *q = check_paint_or_color(L, 3);
    push_scene(L, s.nzfill(*p, *q));
    return 1;
}

static int method_zfill_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Path *p = check_path(L, 2);
    Paint *q = check_paint_or_color(L, 3);
    push_scene(L, s.zfill(*p, *q));
    return 1;
}

static int method_ofill_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Path *p = check_path(L, 2);
    Paint *q = check_paint_or_color(L, 3);
    push_scene(L, s.ofill(*p, *q));
    return 1;
}

static int method_efill_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Path *p = check_path(L, 2);
    Paint *q = check_paint_or_color(L, 3);
    push_scene(L, s.efill(*p, *q));
    return 1;
}

static int meta_gc_scene_iterator(lua_State *L) {
    SceneIterator *it = check_scene_iterators(L, 1);
    it[0].~Iterator();
    it[1].~Iterator();
    return 0;
}

static int meta_tostring_scene_iterator(lua_State *L) {
    SceneIterator *it = check_scene_iterators(L, 1);
    lua_pushfstring(L, "[%d,%d)", it[0].index(), it[1].index());
    return 1;
}

static int method_transform_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Transform *t = check_xform(L, 2);
    push_scene(L, s.transform(*t));
    return 1;
}

static int method_translate_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Transform *t = new_xform_translation(L, 2);
    push_scene(L, s.transform(*t));
    return 1;
}

static int method_scale_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Transform *t = new_xform_scale(L, 2);
    push_scene(L, s.transform(*t));
    return 1;
}

static int method_affine_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Transform *t = new_xform_affine(L, 2);
    push_scene(L, s.transform(*t));
    return 1;
}

static int method_linear_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Transform *t = new_xform_linear(L, 2);
    push_scene(L, s.transform(*t));
    return 1;
}

static int method_rotate_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Transform *t = new_xform_rotation(L, 2);
    push_scene(L, s.transform(*t));
    return 1;
}

static int method_windowviewport_scene(lua_State *L) {
    Scene s = *check_scene(L, 1);
    Transform *t = new_xform_windowviewport(L, 2);
    push_scene(L, s.transform(*t));
    return 1;
}

static const luaL_Reg scene_create[] = {
  {"scene", package_scene},
  {nullptr, nullptr}
};

static const luaL_Reg scene_meta[] = {
  {"__gc", meta_gc_scene},
  {"__call", meta_call_scene},
  {"__tostring", meta_tostring_scene},
  {nullptr, nullptr}
};

static const luaL_Reg scene_meta_index[] = {
  {"fill", method_nzfill_scene},
  {"nzfill", method_nzfill_scene},
  {"zfill", method_zfill_scene},
  {"eofill", method_ofill_scene},
  {"ofill", method_ofill_scene},
  {"efill", method_efill_scene},

  {"clip", method_nzclip_scene},
  {"nzclip", method_nzclip_scene},
  {"zclip", method_zclip_scene},
  {"eoclip", method_oclip_scene},
  {"oclip", method_oclip_scene},
  {"eclip", method_eclip_scene},

  {"pop", method_pop_scene},
  {"push", method_push_scene},
  {"commit", method_commit_scene},
  {"activate", method_commit_scene},

  {"append", method_append_scene},

  {"transform", method_transform_scene},
  {"translate", method_translate_scene},
  {"scale", method_scale_scene},
  {"rotate", method_rotate_scene},
  {"affine", method_affine_scene},
  {"linear", method_linear_scene},
  {"windowviewport", method_windowviewport_scene},

  {nullptr, nullptr}
};

static const luaL_Reg scene_iterator_meta[] = {
  {"__gc", meta_gc_scene_iterator},
  {"__tostring", meta_tostring_scene_iterator},
  {nullptr, nullptr}
};

void Base::do_open_scene(lua_State *L) {
    open_type(L, driver(), Base::Type::scene, scene_create,
        scene_meta, scene_meta_index);
    open_type(L, driver(), Base::Type::scene_iterator, nullptr,
        scene_iterator_meta);
}

static int package_nzfill_element(lua_State *L) {
    using namespace scene;
    Path *p = check_path(L, 1);
    Paint *q = check_paint_or_color(L, 2);
    push_element(L, Element(Element::Type::nzfill, *p, *q));
    return 1;
}

static int package_zfill_element(lua_State *L) {
    using namespace scene;
    Path *p = check_path(L, 1);
    Paint *q = check_paint_or_color(L, 2);
    push_element(L, Element(Element::Type::zfill, *p, *q));
    return 1;
}

static int package_ofill_element(lua_State *L) {
    using namespace scene;
    Path *p = check_path(L, 1);
    Paint *q = check_paint_or_color(L, 2);
    push_element(L, Element(Element::Type::ofill, *p, *q));
    return 1;
}

static int package_efill_element(lua_State *L) {
    using namespace scene;
    Path *p = check_path(L, 1);
    Paint *q = check_paint_or_color(L, 2);
    push_element(L, Element(Element::Type::efill, *p, *q));
    return 1;
}

static int package_nzclip_element(lua_State *L) {
    using namespace scene;
    Path *p = check_path(L, 1);
    // we don't know the depth. so we leave this as depth 0 for now.
    // will know when this is added to a scene
    push_element(L, Element(Element::Type::nzclip, *p, 0));
    return 1;
}

static int package_zclip_element(lua_State *L) {
    using namespace scene;
    Path *p = check_path(L, 1);
    // we don't know the depth. so we leave this as depth 0 for now.
    // will know when this is added to a scene
    push_element(L, Element(Element::Type::zclip, *p, 0));
    return 1;
}

static int package_oclip_element(lua_State *L) {
    using namespace scene;
    Path *p = check_path(L, 1);
    // we don't know the depth. so we leave this as depth 0 for now.
    // will know when this is added to a scene
    push_element(L, Element(Element::Type::oclip, *p, 0));
    return 1;
}

static int package_eclip_element(lua_State *L) {
    using namespace scene;
    Path *p = check_path(L, 1);
    // we don't know the depth. so we leave this as depth 0 for now.
    // will know when this is added to a scene
    push_element(L, Element(Element::Type::eclip, *p, 0));
    return 1;
}

static int package_pushfill_element(lua_State *L) {
    using namespace scene;
    // this could be a pushclip. we can't tell until it's in a scene.
    // we also don't know the depth. so we leave this as depth 0 and
    // pushfill for now.
    push_element(L, Element(Element::Type::pushfill, 0));
    return 1;
}

static int package_popfill_element(lua_State *L) {
    using namespace scene;
    // this could be a popclip. we can't tell until it's in a scene.
    // we also don't know the depth. so we leave this as depth 0 and
    // popfill for now.
    push_element(L, Element(Element::Type::popfill, 0));
    return 1;
}

static int package_commit_element(lua_State *L) {
    using namespace scene;
    // we don't know the depth. so we leave this as depth 0 for now.
    // will know when this is added to a scene
    push_element(L, Element(Element::Type::commit, 0));
    return 1;
}

static const luaL_Reg element_create[] = {
  {"fill", package_nzfill_element},
  {"nzfill", package_nzfill_element},
  {"zfill", package_zfill_element},
  {"eofill", package_ofill_element},
  {"ofill", package_ofill_element},
  {"efill", package_efill_element},

  {"clip", package_nzclip_element},
  {"nzclip", package_nzclip_element},
  {"zclip", package_zclip_element},
  {"eoclip", package_oclip_element},
  {"oclip", package_oclip_element},
  {"eclip", package_eclip_element},

  {"pop", package_popfill_element},
  {"push", package_pushfill_element},
  {"commit", package_commit_element},
  {"activate", package_commit_element},
  {nullptr, nullptr}
};

static int method_transform_element(lua_State *L) {
    scene::Element e = *check_element(L, 1);
    Transform *t = check_xform(L, 2);
    push_element(L, e.transform(*t));
    return 1;
}

static int method_translate_element(lua_State *L) {
    scene::Element e = *check_element(L, 1);
    Transform *t = new_xform_translation(L, 2);
    push_element(L, e.transform(*t));
    return 1;
}

static int method_scale_element(lua_State *L) {
    scene::Element e = *check_element(L, 1);
    Transform *t = new_xform_scale(L, 2);
    push_element(L, e.transform(*t));
    return 1;
}

static int method_affine_element(lua_State *L) {
    scene::Element e = *check_element(L, 1);
    Transform *t = new_xform_affine(L, 2);
    push_element(L, e.transform(*t));
    return 1;
}

static int method_linear_element(lua_State *L) {
    scene::Element e = *check_element(L, 1);
    Transform *t = new_xform_linear(L, 2);
    push_element(L, e.transform(*t));
    return 1;
}

static int method_rotate_element(lua_State *L) {
    scene::Element e = *check_element(L, 1);
    Transform *t = new_xform_rotation(L, 2);
    push_element(L, e.transform(*t));
    return 1;
}

static int method_windowviewport_element(lua_State *L) {
    scene::Element e = *check_element(L, 1);
    Transform *t = new_xform_windowviewport(L, 2);
    push_element(L, e.transform(*t));
    return 1;
}

static const luaL_Reg element_meta_index[] = {
  {"transform", method_transform_element},
  {"translate", method_translate_element},
  {"scale", method_scale_element},
  {"rotate", method_rotate_element},
  {"affine", method_affine_element},
  {"linear", method_linear_element},
  {"windowviewport", method_windowviewport_element},
  {nullptr, nullptr}
};

static int meta_gc_element(lua_State *L) {
    scene::Element *p = check_element(L, 1);
    p->~Element();
    return 0;
}

static int meta_tostring_element(lua_State *L) {
    scene::Element *e = check_element(L, 1);
    std::ostringstream sout;
    sout << *e;
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static int meta_index_element(lua_State *L) {
    scene::Element *e = check_element(L, 1);
    const char *opt = luaL_checkstring(L, 2);
    if (strcmp(opt, "type") == 0) {
        push_const(L, Base::Type::element_type, e->type);
        lua_rawget(L, lua_upvalueindex(1));
    } else if (strcmp(opt, "paint") == 0) {
        if (e->is_fill()) {
            push_paint(L, e->fill.paint);
        } else {
            lua_pushnil(L);
        }
    } else if (strcmp(opt, "path") == 0) {
        if (e->is_fill()) {
            push_path(L, e->fill.path);
        } else if (e->is_clip()) {
            push_path(L, e->clip.path);
        } else {
            lua_pushnil(L);
        }
    } else {
        lua_pushvalue(L, 2); // fallback to potential methods table
        lua_rawget(L, lua_upvalueindex(3));
    }
    return 1;
}

static const luaL_Reg element_meta[] = {
  {"__index", meta_index_element},
  {"__gc", meta_gc_element},
  {"__tostring", meta_tostring_element},
  {nullptr, nullptr}
};

void Base::do_open_element(lua_State *L) {
    open_type(L, driver(), Base::Type::element, element_create, 
        element_meta, element_meta_index);
}


static int package_color_rgb(lua_State *L) {
    float red = luaL_checknumber(L, 1);
    float green = luaL_checknumber(L, 2);
    float blue = luaL_checknumber(L, 3);
    float opacity = luaL_optnumber(L, 4, 1.f);
    if (red < 0.f || red > 1.f)
        luaL_argerror(L, 1, "red component out of range");
    if (green < 0.f || green > 1.f)
        luaL_argerror(L, 1, "green component out of range");
    if (blue < 0.f || blue > 1.f)
        luaL_argerror(L, 1, "blue component out of range");
    if (opacity < 0.f || opacity > 1.f)
        luaL_argerror(L, 1, "opacity component out of range");
    push_color(L, red, green, blue, opacity);
    return 1;
}

static int meta_gc_color(lua_State *L) {
    Color *c = check_color(L, 1);
    c->~Color();
    return 0;
}

static int package_color_gray(lua_State *L) {
    float value = luaL_checknumber(L, 1);
    float opacity = luaL_optnumber(L, 2, 1.f);
    if (value < 0.f || value > 1.f)
        luaL_argerror(L, 1, "value out of range");
    if (opacity < 0.f || opacity > 1.f)
        luaL_argerror(L, 1, "opacity out of range");
    void *p = lua_newuserdata(L, sizeof(Color));
    new (p) Color{value, value, value, opacity};
    set_type(L, Base::Type::color);
    return 1;
}

static int package_color_hsv(lua_State *L) {
    float h = luaL_checknumber(L, 1);
    float s = luaL_checknumber(L, 2);
    float v = luaL_checknumber(L, 3);
    float opacity = luaL_optnumber(L, 4, 1.f);
    if (h < 0.f || h > 1.f) luaL_argerror(L, 1, "hue out of range");
    if (s < 0.f || s > 1.f) luaL_argerror(L, 1, "saturation out of range");
    if (v < 0.f || v > 1.f) luaL_argerror(L, 1, "value out of range");
    if (opacity < 0.f || opacity > 1.f)
        luaL_argerror(L, 4, "opacity out of range");
    void *p = lua_newuserdata(L, sizeof(Color));
    new (p) Color(color::hsv(h,s,v, opacity));
    set_type(L, Base::Type::color);
    return 1;
}

static int meta_tostring_color(lua_State *L) {
    Color *c = check_color(L, 1);
    lua_pushfstring(L, "color %f,%f,%f,%f", c->red, c->green,
        c->blue, c->opacity);
    return 1;
}

static const luaL_Reg color_create[] = {
  {"rgb", package_color_rgb},
  {"rgba", package_color_rgb},
  {"gray", package_color_gray},
  {"hsv", package_color_hsv},
  {nullptr, nullptr}
};

static const luaL_Reg color_meta[] = {
  {"__gc", meta_gc_color},
  {"__tostring", meta_tostring_color},
  {nullptr, nullptr}
};

void Base::do_open_color(lua_State *L) {
    open_type(L, driver(), Base::Type::color, color_create, color_meta);
    // create empty color cache
    // we should probably set this to be a weak table ??
    lua_pushlightuserdata(L, reinterpret_cast<void *>(Base::Type::color));
    lua_newtable(L);
    lua_rawset(L, -3);
    // can't use set_type because we don't have upvalues here
    // package, const
    lua_pushliteral(L, "color");
    lua_newtable(L);
    // package,const,"color",colort
    lua_rawgeti(L, -3, static_cast<int>(Base::Type::color));
    // package,const,"color",colort,meta
    for (auto c = &color::List[0]; c->name; c++) {
        lua_pushstring(L, c->name); // package,const,"color",colort,meta,name
        void *p = lua_newuserdata(L, sizeof(Color));
        // package,const,"color",colort,meta,name,color
        new (p) Color(c->color);
        lua_pushvalue(L, -3);
        // package,const,"color",colort,meta,name,color,meta
        lua_setmetatable(L, -2); // package,const,"color",colort,meta,name,color
        lua_rawset(L, -4); // package,const,"color",colort,meta
    }
    lua_pop(L, 1); // package,const,"color",colort
    lua_rawset(L, -4);
}

static int package_paint_solid(lua_State *L) {
    Color *c = check_color(L, 1);
    float o = luaL_optnumber(L, 2, 1.f);
    push_paint_solid(L, c, o);
    return 1;
}

static int package_paint_linear_gradient(lua_State *L) {
    paint::Ramp *r = check_ramp(L, 1);
    r2::Point *p1 = check_p2(L, 2);
    r2::Point *p2 = check_p2(L, 3);
    Transform *xf = check_xform(L, 4);
    float o = luaL_optnumber(L, 5, 1.f);
    void *p = lua_newuserdata(L, sizeof(Paint));
    new (p) Paint(get_driver(L)->new_linear_gradient(*r, *p1, *p2, *xf, o));
    set_type(L, Base::Type::paint);
    return 1;
}

static int package_paint_radial_gradient(lua_State *L) {
    paint::Ramp *r = check_ramp(L, 1);
    r2::Point *c = check_p2(L, 2);
    r2::Point *f = check_p2(L, 3);
    float rd = luaL_checknumber(L, 4);
    Transform *xf = check_xform(L, 5);
    float o = luaL_optnumber(L, 6, 1.f);
    void *p = lua_newuserdata(L, sizeof(Paint));
    new (p) Paint(get_driver(L)->new_radial_gradient(*r, *c, *f, rd, *xf, o));
    set_type(L, Base::Type::paint);
    return 1;
}

static int package_paint_texture(lua_State *L) {
    Image *i = check_image(L, 1);
    SpreadMethod s = check_spread(L, 2);
    Transform *xf = check_xform(L, 3);
    float o = luaL_optnumber(L, 4, 1.f);
    void *p = lua_newuserdata(L, sizeof(Paint));
    new (p) Paint(get_driver(L)->new_texture(*i, s, *xf, o));
    set_type(L, Base::Type::paint);
    return 1;
}

static int meta_gc_paint(lua_State *L) {
    Paint *p = check_paint(L, 1);
    p->~Paint();
    return 0;
}

static int meta_tostring_paint(lua_State *L) {
    Paint *p = check_paint(L, 1);
    std::ostringstream sout;
    sout << *p;
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static const luaL_Reg paint_create[] = {
  {"solid", package_paint_solid},
  {"lineargradient", package_paint_linear_gradient},
  {"radialgradient", package_paint_radial_gradient},
  {"texture", package_paint_texture},
  {nullptr, nullptr}
};

static int method_transform_paint(lua_State *L) {
    Paint p = *check_paint(L, 1);
    Transform *t = check_xform(L, 2);
    push_paint(L, p.transform(*t));
    return 1;
}

static int method_translate_paint(lua_State *L) {
    Paint p = *check_paint(L, 1);
    Transform *t = new_xform_translation(L, 2);
    push_paint(L, p.transform(*t));
    return 1;
}

static int method_scale_paint(lua_State *L) {
    Paint p = *check_paint(L, 1);
    Transform *t = new_xform_scale(L, 2);
    push_paint(L, p.transform(*t));
    return 1;
}

static int method_affine_paint(lua_State *L) {
    Paint p = *check_paint(L, 1);
    Transform *t = new_xform_affine(L, 2);
    push_paint(L, p.transform(*t));
    return 1;
}

static int method_linear_paint(lua_State *L) {
    Paint p = *check_paint(L, 1);
    Transform *t = new_xform_linear(L, 2);
    push_paint(L, p.transform(*t));
    return 1;
}

static int method_rotate_paint(lua_State *L) {
    Paint p = *check_paint(L, 1);
    Transform *t = new_xform_rotation(L, 2);
    push_paint(L, p.transform(*t));
    return 1;
}

static int method_windowviewport_paint(lua_State *L) {
    Paint p = *check_paint(L, 1);
    Transform *t = new_xform_windowviewport(L, 2);
    push_paint(L, p.transform(*t));
    return 1;
}

static int method_hash_paint(lua_State *L) {
    lua_pushlightuserdata(L, const_cast<paint::Base*>(check_paint(L, 1)->ro()));
    return 1;
}

static const luaL_Reg paint_meta_index[] = {
  {"transform", method_transform_paint},
  {"translate", method_translate_paint},
  {"scale", method_scale_paint},
  {"rotate", method_rotate_paint},
  {"affine", method_affine_paint},
  {"linear", method_linear_paint},
  {"windowviewport", method_windowviewport_paint},
  {"hash", method_hash_paint},
  {nullptr, nullptr}
};

static const luaL_Reg paint_meta[] = {
  {"__gc", meta_gc_paint},
  //{"__tostring", meta_tostring_paint},
  {nullptr, nullptr}
};



void Base::do_open_paint(lua_State *L) {
    open_type(L, driver(), Base::Type::paint, paint_create, paint_meta, 
        paint_meta_index);
    // create empty paint cache
    // we should probably set this to be a weak table ??
    lua_pushlightuserdata(L, reinterpret_cast<void *>(Base::Type::paint));
    lua_newtable(L);
    lua_rawset(L, -3);
}

static void push_p2(lua_State *L, const r2::Point &p2) {
    void *p = lua_newuserdata(L, sizeof(r2::Point));
    new (p) r2::Point(p2);
    set_type(L, Base::Type::p2);
}

static int package_p2(lua_State *L) {
    float x = luaL_optnumber(L, 1, 0.f);
    float y = luaL_optnumber(L, 2, 0.f);
    push_p2(L, r2::Point(x, y));
    return 1;
}

static int meta_index_p2(lua_State *L) {
    r2::Point *p = check_p2(L, 1);
    const char *opt = luaL_checkstring(L, 2);
    if (strcmp(opt, "x") == 0) {
        lua_pushnumber(L, p->x);
    } else if (strcmp(opt, "y") == 0) {
        lua_pushnumber(L, p->y);
    } else {
        lua_pushvalue(L, 2); // fallback to potential methods table
        lua_rawget(L, lua_upvalueindex(3));
    }
    return 1;
}

static int meta_tostring_p2(lua_State *L) {
    r2::Point *p = check_p2(L, 1);
    std::ostringstream sout;
    sout << *p;
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static int meta_gc_p2(lua_State *L) {
    r2::Point *p = check_p2(L, 1);
    p->~Point();
    return 0;
}

static const luaL_Reg p2_create[] = {
  {"p2", package_p2},
  {nullptr, nullptr}
};

static const luaL_Reg p2_meta[] = {
  {"__gc", meta_gc_p2},
  {"__tostring", meta_tostring_p2},
  {"__index", meta_index_p2},
  {nullptr, nullptr}
};

static int method_transform_p2(lua_State *L) {
    r2::Point p = *check_p2(L, 1);
    Transform *t = check_xform(L, 2);
    push_p2(L, (*t)*p);
    return 1;
}

static int method_translate_p2(lua_State *L) {
    r2::Point p = *check_p2(L, 1);
    Transform *t = new_xform_translation(L, 2);
    push_p2(L, (*t)*p);
    return 1;
}

static int method_scale_p2(lua_State *L) {
    r2::Point p = *check_p2(L, 1);
    Transform *t = new_xform_scale(L, 2);
    push_p2(L, (*t)*p);
    return 1;
}

static int method_affine_p2(lua_State *L) {
    r2::Point p = *check_p2(L, 1);
    Transform *t = new_xform_affine(L, 2);
    push_p2(L, (*t)*p);
    return 1;
}

static int method_linear_p2(lua_State *L) {
    r2::Point p = *check_p2(L, 1);
    Transform *t = new_xform_linear(L, 2);
    push_p2(L, (*t)*p);
    return 1;
}

static int method_rotate_p2(lua_State *L) {
    r2::Point p = *check_p2(L, 1);
    Transform *t = new_xform_rotation(L, 2);
    push_p2(L, (*t)*p);
    return 1;
}

static int method_windowviewport_p2(lua_State *L) {
    r2::Point p = *check_p2(L, 1);
    Transform *t = new_xform_windowviewport(L, 2);
    push_p2(L, (*t)*p);
    return 1;
}

static const luaL_Reg p2_meta_index[] = {
  {"transform", method_transform_p2},
  {"translate", method_translate_p2},
  {"scale", method_scale_p2},
  {"rotate", method_rotate_p2},
  {"affine", method_affine_p2},
  {"linear", method_linear_p2},
  {"windowviewport", method_windowviewport_p2},
  {nullptr, nullptr}
};

static int package_p3(lua_State *L) {
    float x = luaL_optnumber(L, 1, 0.f);
    float y = luaL_optnumber(L, 2, 0.f);
    float w = luaL_optnumber(L, 3, 1.f);
    void *p = lua_newuserdata(L, sizeof(r3::Point));
    new (p) r3::Point(x,y,w);
    set_type(L, Base::Type::p3);
    return 1;
}

static int meta_index_p3(lua_State *L) {
    r3::Point *p = check_p3(L, 1);
    const char *lst[] = {"x", "y", "z", nullptr};
    switch (luaL_checkoption(L, 2, nullptr, lst)) {
        case 0:
            lua_pushnumber(L, p->x);
            return 1;
        case 1:
            lua_pushnumber(L, p->y);
            return 1;
        case 2:
            lua_pushnumber(L, p->z);
            return 1;
        default:
            return 0;
    }
}

static int meta_tostring_p3(lua_State *L) {
    r3::Point *p = check_p3(L, 1);
    std::ostringstream sout;
    sout << *p;
    const std::string &str = sout.str();
    lua_pushlstring(L, str.data(), str.length());
    return 1;
}

static int meta_gc_p3(lua_State *L) {
    r3::Point *p = check_p3(L, 1);
    p->~Point();
    return 0;
}

static const luaL_Reg p3_create[] = {
  {"p3", package_p3},
  {nullptr, nullptr}
};

static const luaL_Reg p3_meta[] = {
  {"__gc", meta_gc_p3},
  {"__tostring", meta_tostring_p3},
  {"__index", meta_index_p3},
  {nullptr, nullptr}
};

void Base::do_open_point(lua_State *L) {
    open_type(L, driver(), Base::Type::p2, p2_create, p2_meta, p2_meta_index);
    open_type(L, driver(), Base::Type::p3, p3_create, p3_meta);
}

static int package_image(lua_State *L) {
    if (lua_isstring(L, 1)) {
        const char *str = nullptr;
        size_t len = 0;
        str = lua_tolstring(L, 1, &len);
        std::string decoded = decode_base64(std::string(str, len));
        push_image(L, get_driver(L)->new_image(&decoded[0], decoded.size()));
        return 1;
    } else {
        FILE *f = check_file(L, 1);
        push_image(L, get_driver(L)->new_image(f));
        return 1;
    }
}

static int meta_gc_image(lua_State *L) {
    Image *i = check_image(L, 1);
    i->~Image();
    return 0;
}

static int meta_tostring_image(lua_State *L) {
    Image *i = check_image(L, 1);
    lua_pushfstring(L, "image %dx%d", i->width(), i->height());
    return 1;
}

static const luaL_Reg image_create[] = {
  {"image", package_image},
  {nullptr, nullptr}
};

static const luaL_Reg image_meta[] = {
  {"__gc", meta_gc_image},
  {"__tostring", meta_tostring_image},
  {nullptr, nullptr}
};

void Base::do_open_image(lua_State *L) {
    open_type(L, driver(), Base::Type::image, image_create, image_meta);
}

void Base::do_open_surface(lua_State *L) {
    (void) L;
}

}}

