#ifndef RVG_LUA_BASE_H
#define RVG_LUA_BASE_H

#include <lua.hpp>
#include "../driver/base.h"

namespace rvg { namespace lua {

class Base {
public:
    enum class Type {quadtree, command_type, path, path_iterator, xform, ramp,
        surface, viewport, window, scene, scene_iterator, color, paint,
        p2, p3, spread_method, identity, blend_mode, fill_rule,
        cap_style, join_style, element, element_type, stroke_style, dash, 
        texture, image, last};
    static const char *name(Type type);
    virtual ~Base() { }
    void open(lua_State *L) { return do_open(L); }
    void close(lua_State *L) { return do_close(L); }
    driver::Base *driver(void) { return do_driver(); }
    const driver::Base *driver(void) const { return do_driver(); }
protected:
    virtual void do_open(lua_State *L);
    virtual void do_close(lua_State *L);
    virtual void do_open_xform(lua_State *L);
    virtual void do_open_path(lua_State *L);
    virtual void do_open_point(lua_State *L);
    virtual void do_open_quadtree(lua_State *L);
    virtual void do_open_color(lua_State *L);
    virtual void do_open_element(lua_State *L);
    virtual void do_open_scene(lua_State *L);
    virtual void do_open_image(lua_State *L);
    virtual void do_open_ramp(lua_State *L);
    virtual void do_open_paint(lua_State *L);
    virtual void do_open_surface(lua_State *L);
    virtual void do_open_viewport(lua_State *L);
    virtual void do_open_window(lua_State *L);
    virtual void do_open_stroke_style(lua_State *L);
    virtual driver::Base *do_driver(void) = 0;
    virtual const driver::Base *do_driver(void) const = 0;

};

template <typename T> struct const_Reg {
    const char *name;
    T value;
};

void open_type(lua_State *L, driver::Base *drv,
    Base::Type type, const luaL_Reg *create = nullptr,
    const luaL_Reg *meta = nullptr, const luaL_Reg *index = nullptr);

void get_type_metatable(lua_State *L, Base::Type type);

bool is_type(lua_State *L, int idx, Base::Type type);
void set_type(lua_State *L, Base::Type type);
driver::Base *get_driver(lua_State *L);

Transform *check_xform(lua_State *L, int idx,
    const char *msg = "expected xform");
Path *check_path(lua_State *L, int idx,
    const char *msg = "expected path");
Image *check_image(lua_State *L, int idx,
    const char *msg = "expected image");
Viewport *check_viewport(lua_State *L, int idx,
    const char *msg = "expected viewport");
Window *check_window(lua_State *L, int idx,
    const char *msg = "expected window");
Paint *check_paint(lua_State *L, int idx,
    const char *msg = "expected paint");
Color *check_color(lua_State *L, int idx,
    const char *msg = "expected color");
Scene *check_scene(lua_State *L, int idx,
    const char *msg = "expected scene");
scene::Element *check_element(lua_State *L, int idx, 
    const char *msg = "expected scene element");
r2::Point *check_p2(lua_State *L, int idx,
    const char *msg = "expected p2");
r3::Point *check_p3(lua_State *L, int idx,
    const char *msg = "expected p3");
paint::Ramp *check_ramp(lua_State *L, int idx,
    const char *msg = "expected color ramp");
SpreadMethod check_spread(lua_State *L, int idx,
    const char *msg = "expected spread method");
CapStyle check_cap_style(lua_State *L, int idx,
    const char *msg = "expected cap style");
JoinStyle check_join_style(lua_State *L, int idx,
    const char *msg = "expected join style");
stroke::Style check_stroke_style(lua_State *L, int idx, 
    const char *msg = "expected stroke style");

int meta_tostring_const(lua_State *L);
int meta_index_const(lua_State *L);
int meta_newindex_const(lua_State *L);

}}

#include "base.hpp"

#endif // RVG_LUA_BASE_H
