#include <cstdio>
#include "cuda.h"
#include "../driver/cuda.h"

namespace rvg { namespace lua {

void
CUDA::do_close(lua_State *L) {
    // clear lua::CUDA from registry to allow for collection
    lua_pushlightuserdata(L, driver());
    lua_pushnil(L);
    lua_settable(L, LUA_REGISTRYINDEX);
}

// expects table at the top of the stack.
bool get_field( lua_State *L, const char *key, int& result )
{
    result = 0;
    lua_pushstring( L, key );
    lua_gettable( L, -2 );
    if ( !lua_isnumber(L, -1) ) {
        lua_pop(L, 1);
        return false;
    }
    result = lua_tointeger(L,-1);
    lua_pop(L, 1);
    return true;
}

// expects table at the top of the stack.
bool get_field_float( lua_State *L, const char *key, float& result )
{
    result = 0;
    lua_pushstring( L, key );
    lua_gettable( L, -2 );
    if ( !lua_isnumber(L, -1) ) {
        lua_pop(L, 1);
        return false;
    }
    result = lua_tonumber(L,-1);
    lua_pop(L, 1);
    return true;
}

static driver::CUDAConfig parse_config(lua_State *L)
{
    driver::CUDAConfig config;
    int f=0;
    float ff=0;
    if ( get_field(L, "kernel", f) )
        config.kernel = f;
    if ( get_field(L, "min_segs_split", f) )
        config.min_segs_split = f;
    if ( get_field(L, "dump", f) )
        config.dump = f;
    if ( get_field(L, "qtree_max_level", f) )
        config.qtree_max_level = f;
    if ( get_field(L, "bs3", f) )
        config.use_bs3 = f;
    if ( get_field(L, "warp", f) )
        config.use_warp = f;
    if ( get_field(L, "nsamples", f) )
        config.nsamples = f;
    if ( config.use_warp == 1 ) {
        if ( get_field(L, "foreshorten", f) )
            config.data.foreshorten = f;
    }
    else if ( config.use_warp == 2 ) {
        if ( get_field_float(L, "lens_cu", ff) )
            config.data.lens.cu = ff;
        if ( get_field_float(L, "lens_cv", ff) )
            config.data.lens.cv = ff;
        if ( get_field_float(L, "lens_outer", ff) )
            config.data.lens.outer = ff;
        if ( get_field_float(L, "lens_inner", ff) )
            config.data.lens.inner = ff;
        if ( get_field_float(L, "lens_zoom", ff) )
            config.data.lens.zoom = ff;
    }
    else if ( config.use_warp == 3 ) {
        if ( get_field_float(L, "swirl", ff) )
            config.data.swirl.amount = ff;
        if ( get_field_float(L, "swirl_ang", ff) )
            config.data.swirl.ang = ff;
        if ( get_field_float(L, "swirl_scale", ff) )
            config.data.swirl.scale = ff;
    }
    return config;
}

static int shortcut(lua_State *L) {
    Scene *s = check_scene(L, 1);
    Window *w = check_window(L, 2);
    Viewport *v = check_viewport(L, 3);

    driver::CUDAConfig cfg = parse_config(L);

    driver::CUDA *cuda = static_cast<driver::CUDA *>(get_driver(L));
    driver::ShortcutTree *tree = cuda->create_shortcut(*s, *w, *v, cfg);
    lua_pushlightuserdata(L, tree);
    return 1;
}

static int release_shortcut(lua_State *L) {
    driver::ShortcutTree *tree = reinterpret_cast<driver::ShortcutTree*>(lua_touserdata(L, 1));
    driver::CUDA *cuda = static_cast<driver::CUDA *>(get_driver(L));

    cuda->release(tree);
    return 0;
}

static int render(lua_State *L) {
    driver::ShortcutTree *tree
        = reinterpret_cast<driver::ShortcutTree*>(lua_touserdata(L, 1));

    Transform *xf = check_xform(L, 2);

    const char *fname = lua_tostring(L, 3);
    if(fname == NULL)
        throw std::runtime_error("Invalid parameter 3 to cuda::render");
    Viewport *v = check_viewport(L, 4);
    driver::CUDAConfig cfg = parse_config(L);
    /*
    int kernel = luaL_optint(L,6,0);
    int min_segs_split = luaL_optint(L,7,8);
    int dumpqtree = luaL_optint(L,8,0);
    */
    driver::CUDA *cuda = static_cast<driver::CUDA *>(get_driver(L));
    cuda->render(*tree, *xf, *v, fname, cfg);
    return 0;
}

void add_timer_result( lua_State *L, const gpu::timer_result& result, int level )
{
    //std::cerr << "--[" << result.children.size() << "]" << result.name << "  " << result.time << std::endl;
    lua_newtable(L);
    lua_pushliteral(L, "level");
    lua_pushnumber(L, level );
    lua_settable(L,-3);
    lua_pushliteral(L, "time");
    lua_pushnumber(L, result.time );
    lua_settable(L,-3);
    lua_pushliteral(L, "name");
    lua_pushstring(L, result.name.c_str() );
    lua_settable(L,-3);
    lua_pushliteral(L, "children");
    lua_newtable(L);
    int i = 1;
    for( const auto& r: result.children ) {
        add_timer_result( L, r, level + 1 );
        lua_rawseti( L, -2, i++ );
    }
    lua_settable(L,-3);
}

static int get_timer_results(lua_State *L) {
    driver::CUDA *cuda = static_cast<driver::CUDA *>(get_driver(L));
    add_timer_result(L, cuda->tresult, 0);
    return 1;
}

static int reset_timer(lua_State *L) {
    driver::CUDA *cuda = static_cast<driver::CUDA *>(get_driver(L));
    cuda->reset_timer();
    return 0;
}

void
CUDA::do_open(lua_State *L) {
    Base::do_open(L); // package,consts
    lua_pushliteral(L, "render"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, render, 2);//package,consts,"render",render
    lua_rawset(L, -4);

    lua_pushliteral(L, "shortcut"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, shortcut, 2);//package,consts,"render",render
    lua_rawset(L, -4);

    lua_pushliteral(L, "release_shortcut"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, release_shortcut, 2);//package,consts,"render",render
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "get_timer_results"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, get_timer_results, 2);//package,consts,"render",render
    lua_rawset(L, -4);

    lua_pushliteral(L, "reset_timer"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, reset_timer, 2);//package,consts,"render",render
    lua_rawset(L, -4);
}

}}

static int delete_cuda(lua_State *L) {
    rvg::lua::CUDA *p = static_cast<rvg::lua::CUDA *>(lua_touserdata(L,1));
    p->~CUDA();
    return 0;
}

extern "C"
__attribute__((visibility("default")))
int luaopen_rvg10_cuda(lua_State *L) {
    // create userdata holding lua::CUDA object
    rvg::lua::CUDA *p = static_cast<rvg::lua::CUDA *>(
        lua_newuserdata(L, sizeof(rvg::lua::CUDA))); // cuda
    new (p) rvg::lua::CUDA;
    // set metatable with method for garbage collection
    lua_newtable(L); // cuda,meta
    lua_pushliteral(L, "__gc"); // cuda,meta,"gc"
    lua_pushcfunction(L, delete_cuda); // cuda,meta,"gc",delete
    lua_rawset(L, -3); // cuda,meta
    lua_setmetatable(L, -2); // cuda
    // store it in registry keyed by driver address
    lua_pushlightuserdata(L, p->driver()); // cuda,driver
    lua_pushvalue(L, -2); // cuda,driver,cuda
    lua_settable(L, LUA_REGISTRYINDEX); // cuda
    lua_pop(L, 1); //
    // initialize Lua binding
    lua_newtable(L); // package
    lua_newtable(L); // package,consts
    p->open(L); // package,consts
    lua_pop(L, 1);
    return 1;
}
