#include <cstdio>
#include "cudagl.h"

namespace rvg { namespace lua {

void
CUDAGL::do_close(lua_State *L) {
    // clear lua::CUDAGL from registry to allow for collection
    lua_pushlightuserdata(L, driver());
    lua_pushnil(L);
    lua_settable(L, LUA_REGISTRYINDEX);
}

static int init(lua_State *L) {
    Viewport *v = check_viewport(L, 1);
    driver::CUDAGL *cuda = static_cast<driver::CUDAGL *>(get_driver(L));
    cuda->init(*v);
    return 0;
}

static int shortcut(lua_State *L) {
    Scene *s = check_scene(L, 1);
    Window *w = check_window(L, 2);
    Viewport *v = check_viewport(L, 3);
    driver::CUDAGL *cuda = static_cast<driver::CUDAGL *>(get_driver(L));
    driver::ShortcutTree *tree = cuda->shortcut(*s, *w, *v);
    lua_pushlightuserdata(L, tree);
    return 1;
}

static int render(lua_State *L) {
    //if(fname == NULL)
    //    throw std::runtime_error("Invalid parameter 3 to cuda::render");
    driver::CUDAGL *cuda = static_cast<driver::CUDAGL *>(get_driver(L));
    cuda->render();
    return 0;
}

static int show(lua_State *L) {
    driver::ShortcutTree *tree = reinterpret_cast<driver::ShortcutTree*>(lua_touserdata(L, 1));
    Transform *xf = check_xform(L, 2);
    Viewport *v = check_viewport(L, 3);
    driver::CUDAGL *cuda = static_cast<driver::CUDAGL *>(get_driver(L));
    cuda->show(tree, *xf, *v);
    return 0;
}

static int release_shortcut(lua_State *L) {
    driver::ShortcutTree *tree = reinterpret_cast<driver::ShortcutTree*>(lua_touserdata(L, 1));
    driver::CUDAGL *cuda = static_cast<driver::CUDAGL *>(get_driver(L));
    cuda->release_shortcut(tree);
    return 0;
}

class LuaCallback : public rvg::callback::Base {
    public:
    LuaCallback( lua_State* L, int index )
        : m_L(L), m_index(index)
    { ; }

    void operator()(void) {
        lua_rawgeti(m_L, LUA_REGISTRYINDEX, m_index);
        if ( lua_pcall(m_L, 0, 0, 0) == 0 ) {
            std::cerr << "error: " << lua_tostring(m_L,-1) << std::endl;
        }
    }

    void operator()( int key ) {
        lua_rawgeti(m_L, LUA_REGISTRYINDEX, m_index);
        lua_pushnumber(m_L, key);
        if ( lua_pcall(m_L, 1, 0, 0) == 0 ) {
            std::cerr << "error: " << lua_tostring(m_L,-1) << std::endl;
        }
    }

    void operator()( int button, int state, int x, int y ) {
        lua_rawgeti(m_L, LUA_REGISTRYINDEX, m_index);
        lua_pushnumber(m_L, button);
        lua_pushnumber(m_L, state);
        lua_pushnumber(m_L, x);
        lua_pushnumber(m_L, y);
        if ( lua_pcall(m_L, 4, 0, 0) == 0 ) {
            std::cerr << "error: " << lua_tostring(m_L,-1) << std::endl;
        }
    }

    void operator()( int x, int y ) {
        lua_rawgeti(m_L, LUA_REGISTRYINDEX, m_index);
        lua_pushnumber(m_L, x);
        lua_pushnumber(m_L, y);
        if ( lua_pcall(m_L, 2, 0, 0) == 0 ) {
            std::cerr << "error: " << lua_tostring(m_L,-1) << std::endl;
        }
    }

    private:
    lua_State* m_L;
    int m_index;
};

static int set_display_cb(lua_State *L) {
    if ( lua_gettop(L) == 1 ) {
        if ( lua_isfunction(L,-1) ) {
            //TODO: use shared_ptr to avoid mem leak
            int index = luaL_ref(L, LUA_REGISTRYINDEX);
            auto *lbc = new LuaCallback(L, index);
            driver::CUDAGL *cudagl = static_cast<driver::CUDAGL *>(get_driver(L));
            cudagl->set_display_cb( lbc );
        }
        return 0;
    }
    return 0;
}

static int set_keyboard_cb(lua_State *L) {
    if ( lua_gettop(L) == 1 ) {
        if ( lua_isfunction(L,-1) ) {
            //TODO: use shared_ptr to avoid mem leak
            int index = luaL_ref(L, LUA_REGISTRYINDEX);
            auto *lbc = new LuaCallback(L, index);
            driver::CUDAGL *cudagl = static_cast<driver::CUDAGL *>(get_driver(L));
            cudagl->set_keyboard_cb( lbc );
        }
        return 0;
    }
    return 0;
}

static int set_mouse_cb(lua_State *L) {
    if ( lua_gettop(L) == 1 ) {
        if ( lua_isfunction(L,-1) ) {
            //TODO: use shared_ptr to avoid mem leak
            int index = luaL_ref(L, LUA_REGISTRYINDEX);
            auto *lbc = new LuaCallback(L, index);
            driver::CUDAGL *cudagl = static_cast<driver::CUDAGL *>(get_driver(L));
            cudagl->set_mouse_cb( lbc );
        }
        return 0;
    }
    return 0;
}

static int set_mouse_move_cb(lua_State *L) {
    if ( lua_gettop(L) == 1 ) {
        if ( lua_isfunction(L,-1) ) {
            //TODO: use shared_ptr to avoid mem leak
            int index = luaL_ref(L, LUA_REGISTRYINDEX);
            auto *lbc = new LuaCallback(L, index);
            driver::CUDAGL *cudagl = static_cast<driver::CUDAGL *>(get_driver(L));
            cudagl->set_mouse_move_cb( lbc );
        }
        return 0;
    }
    return 0;
}

void
CUDAGL::do_open(lua_State *L) {
    Base::do_open(L); // package,consts
    lua_pushliteral(L, "init"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, init, 2);//package,consts,"render",render
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "shortcut"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, shortcut, 2);//package,consts,"render",render
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "release_shortcut"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, release_shortcut, 2);//package,consts,"render",render
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "render"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, render, 2);//package,consts,"render",render
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "show"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, show, 2);//package,consts,"render",render
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "display_cb"); // package,consts,"set_callback"
    lua_pushvalue(L, -2); //
    lua_pushlightuserdata(L, driver());//package,consts,"set_callback",consts,drv
    lua_pushcclosure(L, set_display_cb, 2);//package,consts,"set_callback",set_callback
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "keyboard_cb"); // package,consts,"set_callback"
    lua_pushvalue(L, -2); //
    lua_pushlightuserdata(L, driver());//package,consts,"set_callback",consts,drv
    lua_pushcclosure(L, set_keyboard_cb, 2);//package,consts,"set_callback",set_callback
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "mouse_cb"); // package,consts,"set_callback"
    lua_pushvalue(L, -2); //
    lua_pushlightuserdata(L, driver());//package,consts,"set_callback",consts,drv
    lua_pushcclosure(L, set_mouse_cb, 2);//package,consts,"set_callback",set_callback
    lua_rawset(L, -4);
    //
    lua_pushliteral(L, "mouse_move_cb"); // package,consts,"set_callback" lua_pushvalue(L, -2); //
    lua_pushvalue(L, -2); //
    lua_pushlightuserdata(L, driver());//package,consts,"set_callback",consts,drv
    lua_pushcclosure(L, set_mouse_move_cb, 2);//package,consts,"set_callback",set_callback
    lua_rawset(L, -4);
}

}}

static int delete_cuda(lua_State *L) {
    rvg::lua::CUDAGL *p = static_cast<rvg::lua::CUDAGL *>(lua_touserdata(L,1));
    p->~CUDAGL();
    return 0;
}

extern "C"
__attribute__((visibility("default")))
int luaopen_rvg10_cudagl(lua_State *L) {
    // create userdata holding lua::CUDAGL object
    rvg::lua::CUDAGL *p = static_cast<rvg::lua::CUDAGL *>(
        lua_newuserdata(L, sizeof(rvg::lua::CUDAGL))); // cuda
    new (p) rvg::lua::CUDAGL();
    // set metatable with method for garbage collection
    lua_newtable(L); // cuda,meta
    lua_pushliteral(L, "__gc"); // cuda,meta,"gc"
    lua_pushcfunction(L, delete_cuda); // cuda,meta,"gc",delete
    lua_rawset(L, -3); // cuda,meta
    lua_setmetatable(L, -2); // cuda
    // store it in registry keyed by driver address
    lua_pushlightuserdata(L, p->driver()); // cuda,driver
    lua_pushvalue(L, -2); // cuda,driver,cuda
    lua_settable(L, LUA_REGISTRYINDEX); // cuda
    lua_pop(L, 1); //
    // initialize Lua binding
    lua_newtable(L); // package
    lua_newtable(L); // package,consts
    p->open(L); // package,consts
    lua_pop(L, 1);
    return 1;
}
