#include <cstdio>
#include "nvpr.h"
#include "../callback/base.h"

namespace rvg { namespace lua {

void
NVPR::do_close(lua_State *L) {
    // clear lua::NVPR from registry to allow for collection
    lua_pushlightuserdata(L, driver());
    lua_pushnil(L);
    lua_settable(L, LUA_REGISTRYINDEX);
}

static int render(lua_State *L) {
    Scene *s = check_scene(L, 1);
    Transform *xf = check_xform(L, 2);
    const char *fname = lua_tostring(L, 3);
    if(fname == NULL)
        throw std::runtime_error("Invalid parameter 3 to nvpr::render");

    Viewport *v = check_viewport(L, 4);
    driver::NVPR *nvpr = static_cast<driver::NVPR *>(get_driver(L));
    nvpr->render(*s, *xf, fname, *v);
    return 0;
}


class LuaCallback : public rvg::callback::Base {
    public:
    LuaCallback( lua_State* L, int index )
        : m_L(L), m_index(index)
    { ; }

    void operator()(void) {
        lua_rawgeti(m_L, LUA_REGISTRYINDEX, m_index);
        if ( lua_pcall(m_L, 0, 0, 0) == 0 ) {
            std::cerr << "error: " << lua_tostring(m_L,-1) << std::endl;
        }
    }

    private:
    lua_State* m_L;
    int m_index;
};

static int set_callback(lua_State *L) {
    std::cerr << "set_callback" << std::endl;
    if ( lua_gettop(L) == 1 ) {
        if ( lua_isfunction(L,-1) ) {
            //TODO: use shared_ptr to avoid mem leak
            int index = luaL_ref(L, LUA_REGISTRYINDEX);
            auto *lbc = new LuaCallback(L, index);
            driver::NVPR *nvpr = static_cast<driver::NVPR *>(get_driver(L));
            nvpr->set_callback( lbc );
        }
        return 0;
    }
    return 0;
}

void
NVPR::do_open(lua_State *L) {
    Base::do_open(L); // package,consts
    std::cerr << "do_open: " << lua_gettop(L) << std::endl;
    lua_pushliteral(L, "render"); // package,consts,"render"
    lua_pushvalue(L, -2); //package,consts,"render",consts
    lua_pushlightuserdata(L, driver());//package,consts,"render",consts,drv
    lua_pushcclosure(L, render, 2);//package,consts,"render",render
    lua_rawset(L, -4);
    //
    std::cerr << "do_open: " << lua_gettop(L) << std::endl;
    lua_pushliteral(L, "set_callback"); // package,consts,"set_callback"
    lua_pushvalue(L, -2); //
    lua_pushlightuserdata(L, driver());//package,consts,"set_callback",consts,drv
    lua_pushcclosure(L, set_callback, 2);//package,consts,"set_callback",set_callback
    lua_rawset(L, -4);
}

}}

static int delete_nvpr(lua_State *L) {
    rvg::lua::NVPR *p = static_cast<rvg::lua::NVPR *>(lua_touserdata(L,1));
    p->~NVPR();
    return 0;
}

extern "C"
__attribute__((visibility("default")))
int luaopen_rvg10_nvpr(lua_State *L) {
    // create userdata holding lua::NVPR object
    rvg::lua::NVPR *p = static_cast<rvg::lua::NVPR *>(
        lua_newuserdata(L, sizeof(rvg::lua::NVPR))); // nvpr
    new (p) rvg::lua::NVPR;
    // set metatable with method for garbage collection
    lua_newtable(L); // nvpr,meta
    lua_pushliteral(L, "__gc"); // nvpr,meta,"gc"
    lua_pushcfunction(L, delete_nvpr); // nvpr,meta,"gc",delete
    lua_rawset(L, -3); // nvpr,meta
    lua_setmetatable(L, -2); // nvpr
    // store it in registry keyed by driver address
    lua_pushlightuserdata(L, p->driver()); // nvpr,driver
    lua_pushvalue(L, -2); // nvpr,driver,nvpr
    lua_settable(L, LUA_REGISTRYINDEX); // cpu
    lua_pop(L, 1); //
    // initialize Lua binding
    lua_newtable(L); // package
    lua_newtable(L); // package,consts
    p->open(L); // package,consts
    lua_pop(L, 1);
    return 1;
}
