#ifndef MATH_INTERSECT_H
#define MATH_INTERSECT_H

#include "euclidean/point.h"
#include "euclidean/vector.h"
#include "../xform/xform.h"

namespace rvg { namespace math {

// test if two line segments intersect each other
static inline bool intersects(const r2::Point &p0, const r2::Point &p1,
    const r2::Point &q0, const r2::Point &q1) {
    // ostensibly solve p0 + (s/u) (p1-p0) == q0 + (t/u) (q1-q0) for
    // (s/u) and (t/u). then test whether both solutions are within [0,1]
    // in practice, do that without any risky or slow divisions
    const r2::Vector dp = p1-p0;
    const r2::Vector dq = q1-q0;
    const r2::Vector dqp = q0-p0;
    // star of matrix (inverse times the determinant) of linear system
    const xform::Linear sM =
        transpose(cofactors(xform::Linear{dp.x, -dq.x, dp.y, -dq.y}));
    const float u = det(sM);
    const r2::Vector v = sM*dqp;
    const float sus = copysign(v.x, v.x*u); // t times sign of u
    const float sut = copysign(v.y, v.y*u); // s times sign of u
    const float au = fabs(u);
    return u == 0.f? // parallel segments?
        p1 == q1: // segments coincide?
        0.f <= sus && sus <= au && 0.f <= sut && sut <= au;
}

// intersects the two line segments
static inline r3::Point intersect(const r2::Point &p0, const r2::Point &p1,
    const r2::Point &q0, const r2::Point &q1) {
    const r2::Point zero{0.f,0.f};
    // solve p0 + (s/u) (p1-p0) == q0 + (t/u) (q1-q0) for (s/u) and (t/u). 
    const r2::Vector dp = p1-p0;
    const r2::Vector dq = q1-q0;
    const r2::Vector dqp = q0-p0;
    // star of matrix (inverse times the determinant) of linear system
    const xform::Linear sM =
        transpose(cofactors(xform::Linear{dp.x, -dq.x, dp.y, -dq.y}));
    const float u = det(sM);
    const r2::Vector v = sM*dqp;
    return r3::Point{zero+(p0-zero)*u+v.x*dp, u};
}

}} // namespace rvg::math

#endif
