#include <cmath>
#include <algorithm>

#include "quadratic.h"

namespace rvg { namespace math { namespace quadratic {

static inline float chop(float t) {
    if (t < 0.f && t > -1e-7f) return 0.f;
    else return t;
}

static inline double chop(double t) {
    if (t < 0. && t > -1e-15) return 0.;
    else return t;
}

// Returns sqrt(b b - a c) avoiding overflows, underflows,
// and tiny negative numbers
// It may well be faster to do all computations in double
// precision!
template <typename T>
T auxiliardiscriminant(T a, T b, T c, int &e) {
    int ea, eb, ec;
    // adjust exponent so the largest number doesn't overflow
    frexp(a, &ea); frexp(b, &eb); frexp(c, &ec);
    e = std::max(eb, (ea+ec) >> 2);
    T s = ldexp(T(1), -e);
    a *= s; b *= s; c *= s;
    // compute b^2 - ac using fused multiply-add
    // (will it make a difference in the future?)
    // and clamp tiny negative numbers to zero
    return chop(std::fma(b,b,-a*c));
}

float discriminant(float a, float b, float c, int &e) {
    return auxiliardiscriminant(a, b, c, e);
}

double discriminant(double a, double b, double c, int &e) {
    return auxiliardiscriminant(a, b, c, e);
}

template <typename T>
int auxiliarsolve(T a, T b, T c, T delta, int e,
    std::array<T,2> &t, std::array<T,2> &s) {
    if (delta >= T(0)) {
        T d = ldexp(T(1), e)*sqrt(delta);
        if (b > T(0)) {
            T e = b+d;
            t[0] = -c; s[0] = e;
            t[1] = e; s[1] = -a;
        } else if (b < T(0)) {
            T e = -b+d;
            t[0] = e; s[0] = a;
            t[1] = c; s[1] = e;
        } else if (abs(a) > abs(c)) {
            t[0] = d; s[0] = a;
            t[1] = -d; s[1] = a;
        } else {
            t[0] = -c; s[0] = d;
            t[1] = c; s[1] = d;
        }
        return 2;
    } else
        return 0;
}

int solve(float a, float b, float c, float delta, int e,
    std::array<float,2> &t,
    std::array<float,2> &s) {
    return auxiliarsolve(a, b, c, delta, e, t, s);
}

int solve(double a, double b, double c, double delta, int e,
    std::array<double,2> &t,
    std::array<double,2> &s) {
    return auxiliarsolve(a, b, c, delta, e, t, s);
}

int solve(float a, float b, float c,
    std::array<float,2> &t,
    std::array<float,2> &s) {
    int e;
    float delta = auxiliardiscriminant(a, b, c, e);
    return auxiliarsolve(a, b, c, delta, e, t, s);
}

int solve(double a, double b, double c,
    std::array<double,2> &t,
    std::array<double,2> &s) {
    int e;
    double delta = auxiliardiscriminant(a, b, c, e);
    return auxiliarsolve(a, b, c, delta, e, t, s);
}

}}} // rvg::math::quadratic
