#include <cfloat>
#include "svd.h"

namespace rvg { namespace math {

// build an elementary projector from one of the
// vectors in the nullspace of symmetric matrix
// {{r, s}, {s,t}}, which is known to be rank defficient
static xform::Rotation projector(float r, float s, float t) {
    if (abs(r) > abs(t)) {
        float h = hypot(r,s);
        if (h > FLT_MIN) {
            float inv_h = 1.f/h;
            return xform::Rotation{s*inv_h,-r*inv_h};
        } else return xform::Rotation{1.f,0.f};
    } else {
        float h = hypot(t,s);
        if (h > FLT_MIN) {
            float inv_h = 1.f/h;
            return xform::Rotation{t*inv_h,-s*inv_h};
        } else return xform::Rotation{1.f,0.f};
    }
}

#if 0
// newton on t^2 + b*t + c == 0
static inline
float newton(float b, float c, float x) {
    const int n = 2; // two or three should be enough.
    for (int i = 0; i < n; i++)
        x -= (x*(x+b)+c)/(2.f*x+b);
    return x;
}
#endif

void svd(const xform::Linear &A, xform::Rotation &U,
    xform::Scale &S) {
    auto sq = [](float a) { return a*a; };
    // we start computing the two roots of the characteristic
    // polynomial of AAt
    const float ac = A[0][0]*A[1][0];
    const float bd = A[0][1]*A[1][1];
    const float a2 = sq(A[0][0]);
    const float b2 = sq(A[0][1]);
    const float c2 = sq(A[1][0]);
    const float d2 = sq(A[1][1]);
    // we have expressed things in a way that the
    // discriminant is certainly non-negative even in the
    // presence of numerical errors
    const float D = hypot(.5f*(a2+b2-c2-d2), ac+bd);
    const float m = -.5f*(a2+b2+c2+d2);
    const float p = b2*c2+a2*d2-2.f*ac*bd;
    float el0, el1;
    if (m < 0.f) {
        el0 = -m+D;
        el1 = p/el0;
    } else {
        el0 = el1 = 0.f;
    }
#if 0
    // we can try to improve the solutions using Newton
    // but there doesn't seem to be any need
    if (el0*1e-2f > el1) {
        el0 = newton(2.f*m, p, el0);
        el1 = newton(2.f*m, p, el1);
    }
#endif
    // so now we have the
    const float s0 = sqrt(el0); // largest singular value
    if (s0 > FLT_MIN) { // at least 1 singular value above threshold
        // get projector from AAt - el0*I
        U = projector(a2+b2-el0, ac+bd, c2+d2-el0);
        // we will also use the
        const float s1 = sqrt(el1); // smallest singular value
        if (s1 > FLT_MIN) { // both singular values are above threshold
            S = xform::Scale{s0, s1};
        } else { // only largest is above threshold
            S = xform::Scale{s0, 0.f};
        }
    } else { // zero matrix
        U = xform::Identity{};
        S = xform::Scale{0.f,0.f};
    }
}

// analytic singular value decomposition of 2D matrix
void svd(const xform::Linear &A, xform::Rotation &U,
    xform::Scale &S, xform::Linear &V) {
    // compute the U and S matrices of the SVD
    svd(A, U, S);
    // use results to find the corresponding V matrix
    if (S.sy > FLT_MIN) { // both singular values are above threshold
        V = transpose(A)*U*inv(S);
    } else if (S.sx > FLT_MIN) { // only largest is above threshold
        auto v0 = (1.f/S.sx)*transpose(A)*r2::Vector{U.c, U.s};
        V = xform::Linear{v0.x, -v0.y, v0.y, v0.x};
    } else { // zero matrix
        V = xform::Identity{};
    }
}

}} // rvg::math

