#ifndef RVG_RAMP_H
#define RVG_RAMP_H

#include <iosfwd>
#include <initializer_list>
#include <vector>
#include "../color.h"
#include "../util/refcnt.h"
#include "../util/cloneable.h"

namespace rvg { namespace paint { namespace ramp {
class Stop {
public:
    Stop(float _offset = 0.f, float _red = 0.f, float _green = 0.f,
        float _blue = 0.f, float _opacity = 1.f): offset(_offset) {
        color.red = _red;
        color.green = _green;
        color.blue = _blue;
        color.opacity = _opacity;
    }
    float offset;
    Color color;
};

class Base : public Cloneable<Base> {
public:
    Base(SpreadMethod spread = SpreadMethod::pad)
        : m_spread(spread), m_dirty(true) {}
    Base(SpreadMethod spread, std::vector<Stop> stops)
        : m_spread(spread)
        , m_stops(std::move(stops))
        , m_dirty(true) {}

    const std::vector<Stop> &stops(void) const { return m_stops; }

    void append(const Stop &stop)
    { 
        m_stops.push_back(stop);
        m_dirty = true;
    }
    void append(const std::initializer_list<Stop> stops)
    {
        m_stops.insert(m_stops.end(), stops.begin(), stops.end());
        m_dirty = true;
    }

    SpreadMethod spread() const { return m_spread; }
    void set_spread(SpreadMethod spread)
    { 
        m_spread = spread;
        m_dirty = true;
    }

    friend std::ostream &operator<<(std::ostream &out, const Base &stops);

    IMPLEMENT_CLONEABLE(Base);

protected:
    void reset_dirty() const { m_dirty = false; }
    bool is_dirty() const { return m_dirty; }
private:
    SpreadMethod m_spread;
    std::vector<Stop> m_stops;
    mutable bool m_dirty;
};

class Ramp {
public:
    explicit Ramp(Base *base)
        : m_cow(base) {}

    const std::vector<Stop> &stops(void) const {
        return ro()->stops();
    }

    SpreadMethod spread(void) const {
        return ro()->spread();
    }

    void set_spread(SpreadMethod spread) {
        rw()->set_spread(spread);
    }

    Ramp &append(const Stop &stop) {
        rw()->append(stop);
        return *this;
    }

    Ramp &append(const std::initializer_list<Stop> stops) {
        rw()->append(stops);
        return *this;
    }

    friend std::ostream &operator<<(std::ostream &out, const Ramp &ramp);

    IMPLEMENT_COW(Base, m_cow);
private:
    COW<Base> m_cow;
};

} // namespace rvg::paint::ramp

using ramp::Ramp;

} // namespace rvg::paint

} // namespace rvg

#endif // RVG_RAMP_H
