#ifndef RVG_PATH_ARC_H
#define RVG_PATH_ARC_H

#include <cassert>
#include <cfloat>

#include "../xform/xform.h"
#include "../math/point.h"
#include "../math/vector.h"
#include "../math/svd.h"
#include "../math/util.h"
#include "../constants.h"

namespace rvg { namespace path {

// Arc as defined in SVG and OpenVG
class VGArc {
public:

    enum class Type { lccwarc_to, lcwarc_to, sccwarc_to, scwarc_to };

    // Using SVG parameters
    VGArc(const r2::Point &p0, float rx, float ry, float c, float s,
            bool fa, bool fs, const r2::Point &p2): m_p0(p0),
        m_R(c, s), m_S(rx, ry), m_fa(fa), m_fs(fs), m_p2(p2) { }

    VGArc(const r2::Point &p0, float rx, float ry, float c, float s, Type t,
            const r2::Point &p2): m_p0(p0), m_R(c, s),
        m_S(rx, ry), m_p2(p2) {
        flags(t, m_fa, m_fs);
    }

    const xform::Scale &scale(void) const { return m_S; }
    const xform::Rotation &rotation(void) const { return m_R; }
    float ang(void) const { return m_R.ang(); }

    float rx(void) const { return m_S.sx; }
    float ry(void) const { return m_S.sy; }

    bool fa(void) const { return m_fa; }
    bool fs(void) const { return m_fs; }

    static void flags(Type t, bool &fa, bool &fs) {
        switch (t) {
            case Type::lccwarc_to:
                fa = true; fs = true;
                break;
            case Type::lcwarc_to:
                fa = true; fs = false;
                break;
            case Type::sccwarc_to:
                fa = false; fs = true;
                break;
            case Type::scwarc_to:
                fa = false; fs = false;
                break;
        }
    }

    static constexpr Type type(bool fa, bool fs) {
        return fa? (fs? Type::lccwarc_to: Type::lcwarc_to):
                   (fs? Type::sccwarc_to: Type::scwarc_to);
    }

    void transform(const Transform &xf) {

        class XFormVisitor: public xform::Visitor {
        public:
            explicit XFormVisitor(VGArc &that): m_that(that) {}
        private:
            void do_visit(const xform::Identity &) override { }
            void do_visit(const xform::Translation &) override { }
            void do_visit(const xform::Rotation &xf) override {
                m_that.m_R = xf*m_that.m_R;
            }
            void do_visit(const xform::Scale &xf) override { transform(xf); }
            void do_visit(const xform::Linear &xf) override { transform(xf); }
            void do_visit(const xform::Affine &xf) override {
                transform(xf.linear());
            }
            void do_visit(const xform::Projective &xf) override {
                assert(0); // shouldn't be here
                transform(xf.linear());
            }
        private:
            void transform(const xform::Linear &L) {
                math::svd(L*m_that.m_R*m_that.m_S, m_that.m_R, m_that.m_S);
                m_that.m_fs = (det(L) < 0.f) ? !m_that.m_fs: m_that.m_fs;
            }

            VGArc &m_that;
        };

        XFormVisitor v(*this);
        xf.accept(v);
    }

    const r2::Point &p0(void) const { return m_p0; }
    const r2::Point &p2(void) const { return m_p2; }

private:
    r2::Point m_p0;
    xform::Rotation m_R;
    xform::Scale m_S;
    bool m_fa, m_fs;
    r2::Point m_p2;

};

class RationalArc {
public:
    RationalArc(const r2::Point &p0, const r3::Point &p1, const r2::Point &p2):
        m_p0(p0), m_p1(p1), m_p2(p2) {
        // assert(std::abs(p1.z) < 1.f); // elipse
        // the pipeline *should* work evern when the thing
        // is not an elipse...
    }

    RationalArc(const r3::Point &p0, const r3::Point &p1, const r3::Point &p2) {
        //assert(std::abs(p1.z) < 1.f); // elipse
        // the pipeline *should* work evern when the thing
        // is not an elipse...
        float inv_z0 = 1.f/p0.z;
        m_p0 = r2::Point{inv_z0*p0.x, inv_z0*p0.y};
        float inv_z2 = 1.f/p2.z;
        m_p2 = r2::Point{inv_z2*p2.x, inv_z2*p2.y};
        float s = sqrt(inv_z0*inv_z2);
        m_p1 = r3::Point{s*p1.x, s*p1.y, s*p1.z};
    }

    // Construct from VG arc
    // It would be much easier to have a simple function
    // that tries to compute the radius. If it succeeds,
    // return the projective transformation that maps the
    // unit circle to this arc. Then we simply multiply the
    // central point of the rational Bézier by this
    // projection matrix. (?????)
    RationalArc(const r2::Point &p0, float rx, float ry, float c, float s,
            bool fa, bool fs, const r2::Point &p2): m_p0(p0), m_p2(p2) {
        rx = std::abs(rx);
        ry = std::abs(ry);
        const r2::Point zero(0.f,0.f);
        const auto S = xform::Scale(rx, ry);
        const auto R = xform::Rotation(c, s);
        const auto iM = inv(S)*inv(R);
        r2::Point q0 = iM*p0;
        r2::Point q2 = iM*p2;
        r2::Vector dq = q2-q0;
        float dqnorm2 = dot(dq, dq);
        float dqnorm = sqrt(dqnorm2);
        // if arc is not degenerate...
        if (dqnorm > FLT_MIN && rx > FLT_MIN && ry > FLT_MIN) {
            r2::Point qm = q0+.5f*dq;
            float r = std::max(1.f, .5f*dqnorm);
            r2::Vector ndqperp = (1.f/dqnorm)*cross(dq);
            float t = (dqnorm2 < 4.f)? .5f*sqrt(4.f-dqnorm2): 0.f;
            float d = fs != fa? -1.f: 1.f;
            r2::Point cq = qm-ndqperp*t*d; // center of circle
            // z1 is the cosine of half the sector angle
            float z1 = sqrt(.5f*std::max((r*r + dot(q0-cq, q2-cq))/r, 0.f));
            // This business of using zero is awkward??
            r2::Point mq = (zero+(cq-zero)*z1)+ndqperp*r*d;
            m_p1 = r3::Point(R*S*mq, z1);
            if (fa) {
                m_p1 *= -1.f;
            }
        } else {
            // otherwise, use degree elevation to represent
            // the line connecting p0 and p2 as a quadratic Bézier
            m_p1 = r3::Point{.5f*(p0.x+p2.x), .5f*(p0.y+p2.y), 1.f};
        }
    }

    // Returns the projective transformation that maps the unit circle
    // to our conic section.
    // The center is the third column of the matrix, and may be in infinity
    // if the conic is a parabola
    // Otherwise, the transformation is actually affine.
    xform::Projective xf(void) const {
        const auto &p0 = m_p0;
        const auto &p1 = m_p1;
        const auto &p2 = m_p2;
        const float c = m_p1.z; // in elipse, this is cos of the half arc
        const float s2 = 1.f-c*c; // s2 with sign
        float s = std::copysign(std::sqrt(std::abs(s2)), s2);
        return xform::Projective{
            2.f*p1.x-c*(p0.x+p2.x),  s*(p2.x-p0.x), p0.x+p2.x-2.f*c*p1.x,
            2.f*p1.y-c*(p0.y+p2.y),  s*(p2.y-p0.y), p0.y+p2.y-2.f*c*p1.y,
            0.f,                     0.f,             2.f*s2
        };
    }

    bool is_parabola(void) {
        const float c = m_p1.z;
        const float s2 = 1.f-c*c;
        const float scale = 2.f*s2;
        return std::abs(scale) <= FLT_MIN;
    }

    operator VGArc() const {
        const xform::Projective P = xf();
        const float scale = P[2][2];
        assert(std::abs(scale) > FLT_MIN); // conversion from parabola!
        const xform::Linear M = P.linear();
        xform::Rotation R;
        xform::Scale S;
        math::svd(M, R, S);
        bool fa = m_p1.z < 0.f;
        bool fs = det(r3::Point{m_p0, 1.f}, m_p1, r3::Point{m_p2, 1.f}) > 0.f;
        return VGArc(m_p0, S.sx/scale, S.sy/scale, R.c, R.s, fa, fs, m_p2);
    }

    const r2::Point p0(void) const { return m_p0; }
    const r3::Point p1(void) const { return m_p1; }
    const r2::Point p2(void) const { return m_p2; }

private:
    r2::Point m_p0;
    r3::Point m_p1;
    r2::Point m_p2;
};

}} // namespace rvg::path

#endif
