#ifndef RVG_PATH_BASE_H
#define RVG_PATH_BASE_H

#include <iosfwd>

#include "../util/refcnt.h"
#include "../xform/xform.h"
#include "../util/cloneable.h"
#include "../iterator/base.h"
#include "../stroke/style.h"
#include "command.h"

namespace rvg { namespace path {

// abstract class containing path data
class Base: public Cloneable<Base> {
public:

    using internal_const_iterator =
        iterator::Internal<const Command, const Command>;
    using const_iterator = iterator::Iterator<internal_const_iterator>;

    virtual ~Base() {}

    friend std::ostream &operator<<(std::ostream &out, const Base &base);

    void close_path(void) {
        return do_close_path();
    }

    void move_to_abs(float x0, float y0) {
        return do_move_to_abs(x0, y0);
    }

    void move_to_rel(float x0, float y0) {
        return do_move_to_rel(x0, y0);
    }

    void line_to_abs(float x0, float y0) {
        return do_line_to_abs(x0, y0);
    }

    void line_to_rel(float x0, float y0) {
        return do_line_to_rel(x0, y0);
    }

    void hline_to_abs(float x0) {
        return do_hline_to_abs(x0);
    }

    void hline_to_rel(float x0) {
        return do_hline_to_rel(x0);
    }

    void vline_to_abs(float y0) {
        return do_vline_to_abs(y0);
    }

    void vline_to_rel(float y0) {
        return do_vline_to_rel(y0);
    }

    void quad_to_abs(float x0, float y0, float x1, float y1) {
        return do_quad_to_abs(x0, y0, x1, y1);
    }

    void quad_to_rel(float x0, float y0, float x1, float y1) {
        return do_quad_to_rel(x0, y0, x1, y1);
    }

    void squad_to_abs(float x1, float y1) {
        return do_squad_to_abs(x1, y1);
    }

    void squad_to_rel(float x1, float y1) {
        return do_squad_to_rel(x1, y1);
    }

    void cubic_to_abs(float x0, float y0, float x1, float y1,
        float x2, float y2) {
        return do_cubic_to_abs(x0, y0, x1, y1, x2, y2);
    }

    void cubic_to_rel(float x0, float y0, float x1, float y1,
        float x2, float y2) {
        return do_cubic_to_rel(x0, y0, x1, y1, x2, y2);
    }

    void scubic_to_abs(float x1, float y1, float x2, float y2) {
        return do_scubic_to_abs(x1, y1, x2, y2);
    }

    void scubic_to_rel(float x1, float y1, float x2, float y2) {
        return do_scubic_to_rel(x1, y1, x2, y2);
    }

    void svgarc_to_abs(float rx, float ry, float rot_deg, bool fa, bool fs,
        float x2, float y2) {
        return do_svgarc_to_abs(rx, ry, rot_deg, fa, fs, x2, y2);
    }

    void svgarc_to_rel(float rx, float ry, float rot_deg, bool fa, bool fs,
        float x2, float y2) {
        return do_svgarc_to_rel(rx, ry, rot_deg, fa, fs, x2, y2);
    }

    void rquad_to_abs(float x0, float y0, float w0, float x1, float y1) {
        return do_rquad_to_abs(x0, y0, w0, x1, y1);
    }

    void rquad_to_rel(float x0, float y0, float w0, float x1, float y1) {
        return do_rquad_to_rel(x0, y0, w0, x1, y1);
    }

    void transform(const Transform &xf) {
        return do_transform(xf);
    }

    void print(std::ostream &out) const {
        return do_print(out);
    }

    void append_into(Base &that) const {
        return do_append_into(that);
    }

    void append_into(Base &that, const Transform &xf) const {
        return do_append_into(that, xf);
    }

    void append(const Base &that, const Transform &xf) {
        return do_append(that, xf);
    }

    void append(const Base &that) {
        return do_append(that);
    }

    void append(int ncommands, const char *types, const float *arguments) {
        return do_append(ncommands, types, arguments);
    }

    void append(const std::initializer_list<Command> commands) {
        return do_append(commands);
    }

    void append(const Command &cmd) {
        return do_append(cmd);
    }

    RefPtr<Base> xform_clone(const Transform &xf) const {
        return RefPtr<Base>(do_xform_clone(xf));
    }

    RefPtr<Base> stroke_clone(const Transform &xf, const stroke::Style &style, 
        float atol, float swtol, int maxdiv) const  {
        return RefPtr<Base>(do_stroke_clone(xf, style, atol, swtol, maxdiv));

    }

    RefPtr<Base> subdiv_clone(const Transform &xf, 
        float atol, float ptol, int maxdiv) const  {
        return RefPtr<Base>(do_subdiv_clone(xf, atol, ptol, maxdiv));
    }

    RefPtr<Base> flatten_clone(const Transform &xf, 
        float atol, float ptol, int maxdiv) const {
        return RefPtr<Base>(do_flatten_clone(xf, atol, ptol, maxdiv));
    }

    const_iterator begin(void) const {
        return do_begin();
    }

    const_iterator end(void) const {
        return do_end();
    }

protected:

    virtual void do_close_path(void) = 0;
    virtual void do_move_to_abs(float x0, float y0) = 0;
    virtual void do_move_to_rel(float x0, float y0) = 0;
    virtual void do_line_to_abs(float x0, float y0) = 0;
    virtual void do_line_to_rel(float x0, float y0) = 0;
    virtual void do_hline_to_abs(float x0) = 0;
    virtual void do_hline_to_rel(float x0) = 0;
    virtual void do_vline_to_abs(float y0) = 0;
    virtual void do_vline_to_rel(float y0) = 0;
    virtual void do_quad_to_abs(float x0, float y0, float x1, float y1) = 0;
    virtual void do_quad_to_rel(float x0, float y0, float x1, float y1) = 0;
    virtual void do_cubic_to_abs(float x0, float y0, float x1, float y1,
        float x2, float y2) = 0;
    virtual void do_cubic_to_rel(float x0, float y0, float x1, float y1,
        float x2, float y2) = 0;
    virtual void do_scubic_to_abs(float x1, float y1, float x2, float y2) = 0;
    virtual void do_scubic_to_rel(float x1, float y1, float x2, float y2) = 0;
    virtual void do_squad_to_abs(float x1, float y1) = 0;
    virtual void do_squad_to_rel(float x1, float y1) = 0;
    virtual void do_rquad_to_abs(float x0, float y0, float w0, float x1, float y1) = 0;
    virtual void do_rquad_to_rel(float x0, float y0, float w0, float x1, float y1) = 0;
    virtual void do_svgarc_to_abs(float rx, float ry, float rot_deg, bool fa,
        bool fs, float x2, float y2) = 0;
    virtual void do_svgarc_to_rel(float rx, float ry, float rot_deg, bool fa,
            bool fs, float x2, float y2) = 0;

    // Classes derived from Base do not know of the internal
    // representation of other classes derived from it. So it is
    // impossible for any of them to receive a Base and append
    // it into itself. We would need an iterator to the other
    // paths's contents, which is awkward. Instead, we pass the job
    // to the object to be appended, passing ourselves so it
    // can use our Base interface to append itself into us.
    virtual void do_append_into(Base &that) const = 0;
    virtual void do_append_into(Base &that, const Transform &xf) const = 0;

    virtual void do_append(const Base &that) = 0;
    virtual void do_append(const Base &that, const Transform &xf) = 0;

    virtual void do_append(int ncommands, const char *types,
        const float *arguments) = 0;

    virtual void do_append(const std::initializer_list<Command> commands) = 0;
    virtual void do_append(const Command &cmd) = 0;

    virtual Base *do_xform_clone(const Transform &xf) const = 0;

    virtual Base *do_stroke_clone(const Transform &xf, 
        const stroke::Style &style, 
        float atol, float swtol, int maxdiv) const = 0; 

    virtual Base *do_subdiv_clone(const Transform &xf, 
        float atol, float ptol, int maxdiv) const = 0; 

    virtual Base *do_flatten_clone(const Transform &xf, 
        float atol, float ptol, int maxdiv) const = 0; 

    virtual void do_transform(const Transform &xf) = 0;

    virtual void do_print(std::ostream &out) const = 0;

    // iterator interface
    virtual const_iterator do_begin(void) const = 0;

    virtual const_iterator do_end(void) const = 0;

};

}} // namespace rvg::path

#include "base.hpp"

#endif // RVG_PATH_BASE
