#include <cfloat>
#include "../math/point.h"
#include "../math/util.h"
#include "concrete.h"
#include "flatten.h"
#include "stroke.h"
#include "arc.h"

namespace rvg { namespace path {

Concrete*
Concrete::do_xform_clone(const Transform &xf) const {
    return new Concrete(*this, xf);
}

Concrete*
Concrete::do_stroke_clone(const Transform &xf, const stroke::Style &style,
    float atol, float swtol, int maxdiv) const {
    float ctol = cos(atol*math::to_rad());
    return new Concrete(path::stroke(*this, xf, style, ctol, swtol, maxdiv));
}

Concrete*
Concrete::do_subdiv_clone(const Transform &xf,
    float atol, float ptol, int maxdiv) const {
    float ctol = cos(atol*math::to_rad());
    return new Concrete(path::subdiv(*this, xf, ctol, ptol, maxdiv));
}

Concrete*
Concrete::do_flatten_clone(const Transform &xf,
    float atol, float ptol, int maxdiv) const {
    float ctol = cos(atol*math::to_rad());
    return new Concrete(path::flatten(*this, xf, ctol, ptol, maxdiv));
}

void
Concrete::do_svgarc_to_abs(float rx, float ry, float rot_deg, bool fa, bool fs,
        float x2, float y2) {
    // Guard against degenerate arcs right here or somewhere else??
    rx = std::abs(rx); ry = std::abs(ry);
    r2::Point p0{m_current_x, m_current_y};
    r2::Point p2{x2, y2};
    // Endpoints too close or radii too small? Convert to line segment
    if (dot(p2-p0, p2-p0) > 64*FLT_MIN && rx > FLT_MIN && ry > FLT_MIN) {
        float rot_rad = rot_deg*math::to_rad();
        float c = cos(rot_rad), s = sin(rot_rad);
        path::RationalArc arc(p0, rx, ry, c, s, fa, fs, p2);
        const r3::Point &p1 = arc.p1();
        rquad_to_abs(p1.x, p1.y, p1.z, x2, y2);
    } else {
        line_to_abs(x2, y2);
    }
}

void
Concrete::do_svgarc_to_rel(float rx, float ry, float rot_deg, bool fa,
    bool fs, float x2, float y2) {
    svgarc_to_abs(rx, ry, rot_deg, fa, fs, m_current_x+x2, m_current_y+y2);
}

void
Concrete::do_print(std::ostream &out) const {
    out << name << " "
        << m_start_x << "," << m_start_y << " "
        << m_previous_x << "," << m_previous_y << " "
        << m_current_x << "," << m_current_y << ":";
    for (const auto &segment: m_segments) {
        const int i = segment.index;
        switch (static_cast<SegmentType>(segment.type)) {
            case SegmentType::close_path:
                out << " Z";
                break;
            case SegmentType::move_to:
                out << " M "
                    << m_data[i] << "," << m_data[i+1];
                break;
            case SegmentType::line_to:
                out << " L "
                    << m_data[i+2] << "," << m_data[i+3];
                break;
            case SegmentType::quad_to:
                out << " Q "
                    << m_data[i+2] << "," << m_data[i+3] << " "
                    << m_data[i+4] << "," << m_data[i+5];
                break;
            case SegmentType::cubic_to:
                out << " C "
                    << m_data[i+2] << "," << m_data[i+3] << " "
                    << m_data[i+4] << "," << m_data[i+5] << " "
                    << m_data[i+6] << "," << m_data[i+7];
                break;
            case SegmentType::rquad_to:
                out << " A "
                    << m_data[i+2] << "," << m_data[i+3] << "," << m_data[i+4]
                    << ' '
                    << m_data[i+5] << "," << m_data[i+6];
                break;
            case SegmentType::fline_to:
                out << " fL "
                    << m_data[i+2] << "," << m_data[i+3];
                break;
            default:
                // should never be here!
                assert(0);
                break;
        }
    }
}

inline void
Concrete::do_transform(const Transform &xf) {
    // if identity, nothing to do
    if(xf.is_identity())
        return;
    for (auto &segment: m_segments) {
        const int i = segment.index;
        const SegmentType t = static_cast<SegmentType>(segment.type);
        switch (t) {
            case SegmentType::close_path:
                break;
            case SegmentType::move_to: {
                r2::Point p0 = xf * r2::Point{m_data[i], m_data[i+1]};
                m_data[i] = p0.x; m_data[i+1] = p0.y;
                break;
            }
            case SegmentType::fline_to:
            case SegmentType::line_to: {
                r2::Point p0 = xf * r2::Point{m_data[i+2], m_data[i+3]};
                m_data[i+2] = p0.x; m_data[i+3] = p0.y;
                break;
            }
            case SegmentType::quad_to: {
                r2::Point p0 = xf * r2::Point{m_data[i+2], m_data[i+3]};
                r2::Point p1 = xf * r2::Point{m_data[i+4], m_data[i+5]};
                m_data[i+2] = p0.x; m_data[i+3] = p0.y;
                m_data[i+4] = p1.x; m_data[i+5] = p1.y;
                break;
            }
            case SegmentType::cubic_to: {
                r2::Point p0 = xf * r2::Point{m_data[i+2], m_data[i+3]};
                r2::Point p1 = xf * r2::Point{m_data[i+4], m_data[i+5]};
                r2::Point p2 = xf * r2::Point{m_data[i+6], m_data[i+7]};
                m_data[i+2] = p0.x; m_data[i+3] = p0.y;
                m_data[i+4] = p1.x; m_data[i+5] = p1.y;
                m_data[i+6] = p2.x; m_data[i+7] = p2.y;
                break;
            }
            case SegmentType::rquad_to: {
                r3::Point p0 = xf * r3::Point{m_data[i+2], m_data[i+3], m_data[i+4]};
                r2::Point p1 = xf * r2::Point{m_data[i+5], m_data[i+6]};
                m_data[i+2] = p0.x; m_data[i+3] = p0.y; m_data[i+4] = p0.z;
                m_data[i+5] = p1.x; m_data[i+6] = p1.y;
                break;
            }
            default:
                std::cerr << t << '\n';
                // should never be here!
                assert(0);
                break;
        }
    }
}

void
Concrete::do_append_into(Base &that) const {
    for (const auto &segment: m_segments) {
        const int i = segment.index;
        switch (static_cast<SegmentType>(segment.type)) {
            case SegmentType::close_path:
                that.close_path();
                break;
            case SegmentType::move_to:
                that.move_to_abs(m_data[i], m_data[i+1]);
                break;
            case SegmentType::line_to:
                that.line_to_abs(m_data[i+2], m_data[i+3]);
                break;
            case SegmentType::quad_to:
                that.quad_to_abs(m_data[i+2], m_data[i+3],
                        m_data[i+4], m_data[i+5]);
                break;
            case SegmentType::cubic_to:
                that.cubic_to_abs(m_data[i+2], m_data[i+3],
                        m_data[i+4], m_data[i+5], m_data[i+6], m_data[i+7]);
                break;
            case SegmentType::rquad_to:
                that.rquad_to_abs(m_data[i+2], m_data[i+3],
                        m_data[i+4], m_data[i+5], m_data[i+6]);
                break;
            case SegmentType::fline_to: // ignore these internal segments
                break;
            default:
                // should never be here!
                assert(0);
                break;
        }
    }
}

inline void
Concrete::do_append_into(Base &that, const Transform &xf) const {
    for (const auto &segment: m_segments) {
        const int i = segment.index;
        const SegmentType t = static_cast<SegmentType>(segment.type);
        switch (t) {
            case SegmentType::close_path:
                that.close_path();
                break;
            case SegmentType::move_to: {
                r2::Point p0 = xf * r2::Point{m_data[i], m_data[i+1]};
                that.move_to_abs(p0.x, p0.y);
                break;
            }
            case SegmentType::line_to: {
                r2::Point p0 = xf * r2::Point{m_data[i+2], m_data[i+3]};
                that.line_to_abs(p0.x, p0.y);
                break;
            }
            case SegmentType::quad_to: {
                r2::Point p0 = xf * r2::Point{m_data[i+2], m_data[i+3]};
                r2::Point p1 = xf * r2::Point{m_data[i+4], m_data[i+5]};
                that.quad_to_abs(p0.x, p0.y, p1.x, p1.y);
                break;
            }
            case SegmentType::cubic_to: {
                r2::Point p0 = xf * r2::Point{m_data[i+2], m_data[i+3]};
                r2::Point p1 = xf * r2::Point{m_data[i+4], m_data[i+5]};
                r2::Point p2 = xf * r2::Point{m_data[i+6], m_data[i+7]};
                that.cubic_to_abs(p0.x, p0.y, p1.x, p1.y, p2.x, p2.y);
                break;
            }
            case SegmentType::rquad_to: {
                r3::Point p0 = xf * r3::Point{m_data[i+2], m_data[i+3], m_data[i+4]};
                r2::Point p1 = xf * r2::Point{m_data[i+5], m_data[i+6]};
                that.rquad_to_abs(p0.x, p0.y, p0.z, p1.x, p1.y);
                break;
            }
            case SegmentType::fline_to: // ignore these internal segments
                break;
            default:
                // should never be here!
                assert(0);
                break;
        }
    }
}

int
Concrete::append_command(CommandType type, const float *args) {
    switch (type) {
        case CommandType::close_path_abs:
            close_path();
            return 0;
        case CommandType::close_path_rel:
            close_path();
            return 0;
        case CommandType::move_to_abs:
            move_to_abs(args[0], args[1]);
            return 2;
        case CommandType::move_to_rel:
            move_to_rel(args[0], args[1]);
            return 2;
        case CommandType::line_to_abs:
            line_to_abs(args[0], args[1]);
            return 2;
        case CommandType::line_to_rel:
            line_to_rel(args[0], args[1]);
            return 2;
        case CommandType::hline_to_abs:
            hline_to_abs(args[0]);
            return 1;
        case CommandType::hline_to_rel:
            hline_to_rel(args[0]);
            return 1;
        case CommandType::vline_to_abs:
            vline_to_abs(args[0]);
            return 1;
        case CommandType::vline_to_rel:
            vline_to_rel(args[0]);
            return 1;
        case CommandType::quad_to_abs:
            quad_to_abs(args[0], args[1], args[2], args[3]);
            return 4;
        case CommandType::quad_to_rel:
            quad_to_rel(args[0], args[1], args[2], args[3]);
            return 4;
        case CommandType::squad_to_abs:
            squad_to_abs(args[0], args[1]);
            return 2;
        case CommandType::squad_to_rel:
            squad_to_rel(args[0], args[1]);
            return 2;
        case CommandType::cubic_to_abs:
            cubic_to_abs(args[0], args[1], args[2], args[3],
                    args[4], args[5]);
            return 6;
        case CommandType::cubic_to_rel:
            cubic_to_rel(args[0], args[1], args[2], args[3],
                    args[4], args[5]);
            return 6;
        case CommandType::scubic_to_abs:
            scubic_to_abs(args[0], args[1], args[2], args[3]);
            return 4;
        case CommandType::scubic_to_rel:
            scubic_to_rel(args[0], args[1], args[2], args[3]);
            return 4;
        case CommandType::rquad_to_abs:
            rquad_to_abs(args[0], args[1], args[2], args[3], args[4]);
            return 5;
        case CommandType::rquad_to_rel:
            rquad_to_rel(args[0], args[1], args[2], args[3], args[4]);
            return 5;
        case CommandType::svgarc_to_abs: {
            int fa_fs = int(args[3]);
            bool fa = fa_fs & 0x2;
            bool fs = fa_fs & 0x1;
            svgarc_to_abs(args[0], args[1], args[2], fa, fs, args[4], args[5]);
            return 6;
        }
        case CommandType::svgarc_to_rel: {
            int fa_fs = int(args[3]);
            bool fa = fa_fs & 0x2;
            bool fs = fa_fs & 0x1;
            svgarc_to_rel(args[0], args[1], args[2], fa, fs, args[4], args[5]);
            return 6;
        }
        default:
            // reaching here is a runtime error...
            assert(0);
            return 0;
    }
}

}} // namespace rvg::path
