#ifndef RVG_CONCRETE_H
#define RVG_CONCRETE_H

#include <vector>
#include "path.h"

// Note that before actually using the information in a
// path, we have to ensure it is closed at least by an
// fline by calling ensure_closed. Ideally, we should avoid
// an unecessary copy-on-write when doing so! ??
namespace rvg { namespace path {

// In our internal representation of paths, there is always an initial moveto,
// there are no relative segments, hlinetos and vlinetos are transformed
// to linetos, squadtos to quadtos, scubictos to cubictos, and a
// close_path segment simply marks the need for a special
// join connecting the first and last segments when stroking
// An additional lineto is generated to actually close the path
// Abscent a close_path, an fline is generated to "close"
// the path so it can be filled properly without leaking out.
// We also have each segment point to its first point so the
// entire path can be processed in parallel
class Concrete: public Base {
public:

    class internal_const_iterator: public Base::internal_const_iterator {
    public:
        internal_const_iterator(const Concrete *path, int index): m_path(path),
            m_index(index) {}

        internal_const_iterator *clone(void) const {
            return new internal_const_iterator(m_path, m_index);
        }

        void increment(void) {
            const auto &s = m_path->segments();
            // skip over fline_to segments
            // it is never the first segment in a path
            // there is never more than one in a row
            // here seems the right place to do it ??
            ++m_index;
            if (m_index < int(s.size())) {
                SegmentType t = static_cast<SegmentType>(s[m_index].type);
                if (t == SegmentType::fline_to) ++m_index;
            }
        }

        int index(void) const { return m_index; }

        bool equal(const Base::internal_const_iterator *base_iterator) const {
            const internal_const_iterator *other =
                static_cast<const internal_const_iterator *>(base_iterator);
            assert(m_path == other->m_path);
            return m_index == other->m_index;
        }

        Base::internal_const_iterator::reference dereference(void) const {
            const Segment &seg = m_path->segment(m_index);
            const int i = seg.index;
            const int t = seg.type;
            switch (static_cast<SegmentType>(t)) {
                case SegmentType::close_path:
                    return Command{static_cast<CommandType>(t << 1)};
                case SegmentType::move_to:
                    return Command{static_cast<CommandType>(t << 1),
                        m_path->datum(i), m_path->datum(i+1)
                    };
                case SegmentType::line_to:
                    return Command{static_cast<CommandType>(t << 1),
                        m_path->datum(i+2), m_path->datum(i+3)
                    };
                case SegmentType::quad_to:
                    return Command{static_cast<CommandType>(t << 1),
                        m_path->datum(i+2), m_path->datum(i+3),
                        m_path->datum(i+4), m_path->datum(i+5)
                    };
                case SegmentType::rquad_to:
                    return Command{static_cast<CommandType>(t << 1),
                        m_path->datum(i+2), m_path->datum(i+3),
                        m_path->datum(i+4), m_path->datum(i+5),
                        m_path->datum(i+6)
                    };
                case SegmentType::cubic_to:
                    return Command{static_cast<CommandType>(t << 1),
                        m_path->datum(i+2), m_path->datum(i+3),
                        m_path->datum(i+4), m_path->datum(i+5),
                        m_path->datum(i+6), m_path->datum(i+7)
                    };
                default:
                    assert(0); // shouldn't reach here!
                    return Command{static_cast<CommandType>(t << 1)};
            }
        }
    private:
        const Concrete* m_path;
        int m_index;
    };

    static constexpr const char *name = "concrete";
    struct Segment {
        // broken into 16 types, 256M indices.
        // could change this ratio (??)
        unsigned type: 4;        // segment from SegmentType
        unsigned index: 28;      // index into data array
    };

    // internal segment commands
    void fline_to_abs(float x0, float y0) {
        return do_fline_to_abs(x0, y0);
    }

    void fline_to_rel(float x0, float y0) {
        return do_fline_to_rel(x0, y0);
    }

    Concrete(void);

    Concrete(const Concrete &that, const Transform &xf);

    virtual ~Concrete() { ; }

    // restore implicit assignment, move assignment, move constructor 
    // they were deleted because of the stupid virtual destructor. 
    // maybe we could make Concrete final??
    Concrete(const Concrete &) = default;
    Concrete(Concrete &&) = default;
    Concrete &operator=(const Concrete &) = default;
    Concrete &operator=(Concrete &&) = default;

    const std::vector<Segment> &segments(void) const { return m_segments; }

    const Segment &segment(int i) const { return m_segments[i]; }

    const std::vector<float> &data() const { return m_data; }

    float datum(int i) const { return m_data[i]; }

protected:
    std::vector<Segment> m_segments;
    std::vector<float> m_data;
    // current, start, and previous points
public:
    float m_current_x, m_current_y;
    float m_start_x, m_start_y;
    float m_previous_x, m_previous_y;
    // number of closed contours in path (i.e., number of move_to segments)
    int m_ncontours;

    // make sure a moveto doesn't leave an open contour
    void ensure_closed(void);

    IMPLEMENT_CLONEABLE(Concrete);
protected:

    void do_close_path(void) override;
    void do_move_to_abs(float x0, float y0) override;
    void do_move_to_rel(float x0, float y0) override;
    void do_line_to_abs(float x0, float y0) override;
    void do_line_to_rel(float x0, float y0) override;
    void do_hline_to_abs(float x0) override;
    void do_hline_to_rel(float x0) override;
    void do_vline_to_abs(float y0) override;
    void do_vline_to_rel(float y0) override;
    void do_quad_to_abs(float x0, float y0, float x1, float y1) override;
    void do_quad_to_rel(float x0, float y0, float x1, float y1) override;
    void do_squad_to_abs(float x1, float y1) override;
    void do_squad_to_rel(float x1, float y1) override;
    void do_cubic_to_abs(float x0, float y0, float x1, float y1,
        float x2, float y2) override;
    void do_cubic_to_rel(float x0, float y0, float x1, float y1,
        float x2, float y2) override;
    void do_scubic_to_abs(float x1, float y1, float x2, float y2) override;
    void do_scubic_to_rel(float x1, float y1, float x2, float y2) override;
    void do_rquad_to_abs(float x0, float y0, float w0, float x1, float y1) override;
    void do_rquad_to_rel(float x0, float y0, float w0, float x1, float y1) override;
    void do_svgarc_to_abs(float rx, float ry, float rot_ang, bool fa, bool fs,
        float x2, float y2) override;
    void do_svgarc_to_rel(float rx, float ry, float rot_ang, bool fa, bool fs,
        float x2, float y2) override;

    void do_append_into(Base &that) const override;
    void do_append_into(Base &that, const Transform &xf) const override;
    void do_append(const Base &that) override;
    void do_append(const Base &that, const Transform &xf) override;
    void do_append(int ncommands, const char *types,
        const float *arguments) override;
    void do_append(const std::initializer_list<Command> commands) override;
    void do_append(const Command &cmd) override;

    Concrete* do_xform_clone(const Transform &xf) const override;
    virtual Concrete* do_stroke_clone(const Transform &xf, const stroke::Style &style, 
        float atol, float swtol, int maxdiv) const  override; 
    Concrete* do_subdiv_clone(const Transform &xf, 
        float atol, float ptol, int maxdiv) const  override; 
    Concrete* do_flatten_clone(const Transform &xf, 
        float atol, float ptol, int maxdiv) const  override; 

    void do_transform(const Transform &xf) override;

    void do_print(std::ostream &out) const override;

    const_iterator do_begin(void) const override;

    const_iterator do_end(void) const override;

    virtual void do_fline_to_abs(float x0, float y0);
    virtual void do_fline_to_rel(float x0, float y0);


private:
    void set_current(float x, float y) {
        m_current_x = x; m_current_y = y;
    }
    void set_previous(float x, float y) {
        m_previous_x = x; m_previous_y = y;
    }
    void set_start(float x, float y) {
        m_start_x = x; m_start_y = y;
    }
    int append_command(CommandType type, const float *args);

    // make sure segments always have an initial point
    void ensure_non_empty(void);
    // most segments share a point with the previous segment
    // thus the rewind of 2
    void push_segment(SegmentType type, int rewind = 2);

    void push_data(float x0, float y0);
    void push_data(float x0, float y0, float x1, float y1);
    void push_data(float x0, float y0, float w0, float x1, float y1);
    void push_data(float x0, float y0, float x1, float y1, float x2, float y2);

    // do we need an implicit close_fill?
    int m_is_closed;
};

}} // namespace rvg::path

#include "concrete.hpp"

#endif // CONCRETE_H
