namespace rvg { namespace path {

inline
Concrete::Concrete(void) {
    set_start(0.f, 0.f);
    set_previous(0.f, 0.f);
    set_current(0.f, 0.f);
    m_is_closed = 1;
    m_ncontours = 0;
}

inline
Concrete::Concrete(const Concrete &that, const Transform &xf) {
    m_data.reserve(that.m_data.size());
    m_segments.reserve(that.m_segments.size());
    that.append_into(*this, xf);
}

inline void
Concrete::do_close_path(void) {
    // ignore when already closed
    if (!m_is_closed) {
        // generate lineto only if needed
        if (m_current_x != m_start_x || m_current_y != m_start_y) {
            line_to_abs(m_start_x, m_start_y);
        }
        // mark the need for a join when stroking
        push_segment(SegmentType::close_path, 0);
        set_previous(m_start_x, m_start_y);
        set_current(m_start_x, m_start_y);
        m_is_closed = 1;
    }
}

inline
Base::const_iterator
Concrete::do_begin(void) const {
    return Base::const_iterator(new internal_const_iterator(this, 0));
}

inline
Base::const_iterator
Concrete::do_end(void) const {
    return Base::const_iterator(new internal_const_iterator(this,
        m_segments.size()));
}

inline void
Concrete::do_move_to_abs(float x0, float y0) {
    ensure_closed();
    push_segment(SegmentType::move_to, 0);
    push_data(x0, y0);
    set_start(x0, y0);
    set_current(x0, y0);
    set_previous(x0, y0);
    ++m_ncontours;
}

inline void
Concrete::do_move_to_rel(float x0, float y0) {
    move_to_abs(m_current_x+x0, m_current_y+y0);
}

inline void
Concrete::do_line_to_abs(float x0, float y0) {
    if (m_current_x != x0 || m_current_y != y0) { // prevent empty lines
        ensure_non_empty();
        push_segment(SegmentType::line_to);
        push_data(x0, y0);
        set_current(x0, y0);
        set_previous(x0, y0);
    }
}

inline void
Concrete::do_fline_to_abs(float x0, float y0) {
    push_segment(SegmentType::fline_to);
    push_data(x0, y0);
}

inline void
Concrete::do_fline_to_rel(float x0, float y0) {
    fline_to_abs(x0+m_current_x, y0+m_current_y);
}

inline void
Concrete::do_line_to_rel(float x0, float y0) {
    line_to_abs(m_current_x+x0, m_current_y+y0);
}

inline void
Concrete::do_hline_to_abs(float x0) {
    line_to_abs(x0, m_current_y);
}

inline void
Concrete::do_hline_to_rel(float x0) {
    hline_to_abs(m_current_x+x0);
}

inline void
Concrete::do_vline_to_abs(float y0) {
    line_to_abs(m_current_x, y0);
}

inline void
Concrete::do_vline_to_rel(float y0) {
    vline_to_abs(m_current_y+y0);
}

inline void
Concrete::do_quad_to_abs(float x0, float y0, float x1, float y1) {
    ensure_non_empty();
    push_segment(SegmentType::quad_to);
    push_data(x0, y0, x1, y1);
    set_previous(x0, y0);
    set_current(x1, y1);
}

inline void
Concrete::do_quad_to_rel(float x0, float y0, float x1, float y1) {
    quad_to_abs(m_current_x+x0, m_current_y+y0, m_current_x+x1,
            m_current_y+y1);
}

inline void
Concrete::do_cubic_to_abs(float x0, float y0, float x1, float y1,
        float x2, float y2) {
    ensure_non_empty();
    push_segment(SegmentType::cubic_to);
    push_data(x0, y0, x1, y1, x2, y2);
    set_previous(x1, y1);
    set_current(x2, y2);
}

inline void
Concrete::do_cubic_to_rel(float x0, float y0, float x1, float y1,
        float x2, float y2) {
    cubic_to_abs(m_current_x+x0, m_current_y+y0, m_current_x+x1,
            m_current_y+y1, m_current_x+x2, m_current_y+y2);
}

inline void
Concrete::do_squad_to_abs(float x1, float y1) {
    quad_to_abs(2.f*m_current_x-m_previous_x, 2.f*m_current_y-m_previous_y,
            x1, y1);
}

inline void
Concrete::do_squad_to_rel(float x1, float y1) {
    squad_to_abs(m_current_x+x1, m_current_y+y1);
}

inline void
Concrete::do_scubic_to_abs(float x1, float y1, float x2, float y2) {
    cubic_to_abs(2.f*m_current_x-m_previous_x, 2.f*m_current_y-m_previous_y,
            x1, y1, x2, y2);
}

inline void
Concrete::do_scubic_to_rel(float x1, float y1, float x2, float y2) {
    scubic_to_abs(m_current_x+x1, m_current_y+y1, m_current_x+x2,
            m_current_y+y2);
}

inline void
Concrete::do_rquad_to_abs(float x0, float y0, float w0, float x1, float y1) {
    ensure_non_empty();
    push_segment(SegmentType::rquad_to);
    push_data(x0, y0, w0, x1, y1);
    set_previous(x1, y1);
    set_current(x1, y1);
}

inline void
Concrete::do_rquad_to_rel(float x0, float y0, float w0, float x1, float y1) {
    rquad_to_abs(m_current_x*w0+x0, m_current_y*w0+y0, w0,
            m_current_x+x1, m_current_y + y1);
}

inline void
Concrete::do_append(const Base &that) {
    that.append_into(*this);
}

inline void
Concrete::do_append(const Base &that, const Transform &xf) {
    that.append_into(*this, xf);
}

inline void
Concrete::do_append(int ncommands, const char *types, const float *arguments) {
    for (int i = 0; i < ncommands; ++i) {
        arguments +=
            append_command(static_cast<CommandType>(types[i]), arguments);
    }
}

inline void
Concrete::do_append(const std::initializer_list<Command> commands) {
    for (const auto &cmd: commands) {
        append_command(cmd.type, cmd.arguments);
    }
}

inline void
Concrete::do_append(const Command &cmd) {
    append_command(cmd.type, cmd.arguments);
}

inline void
Concrete::ensure_non_empty(void) {
    // ignore for non-empty paths
    if (m_ncontours == 0) {
        push_segment(SegmentType::move_to, 0);
        push_data(m_current_x, m_current_y);
        m_ncontours = 1;
    }
    m_is_closed = 0;
}

// make sure a moveto doesn't leave an open contour
inline void
Concrete::ensure_closed(void) {
    // ignore when already closed
    if (!m_is_closed) {
        // generate xlineto only if path is not already "closed"
        if (m_current_x != m_start_x || m_current_y != m_start_y) {
            ensure_non_empty();
            push_segment(SegmentType::fline_to);
            push_data(m_start_x, m_start_y);
        }
        m_is_closed = 1;
    }
}

inline void
Concrete::push_segment(SegmentType type, int rewind) {
    // most segments share a point with the previous segment
    // thus the rewind of 2
    m_segments.push_back(Segment{
        // casts needed due to narrowing
        // maybe we should add an assert here (??)
        static_cast<unsigned>(type),
        static_cast<unsigned>(m_data.size()-rewind)
    });
}

inline void
Concrete::push_data(float x0, float y0) {
    m_data.push_back(x0); m_data.push_back(y0);
}

inline void
Concrete::push_data(float x0, float y0, float x1, float y1) {
    m_data.push_back(x0); m_data.push_back(y0);
    m_data.push_back(x1); m_data.push_back(y1);
}

inline void
Concrete::push_data(float x0, float y0, float w0, float x1, float y1) {
    m_data.push_back(x0); m_data.push_back(y0); m_data.push_back(w0);
    m_data.push_back(x1); m_data.push_back(y1);
}

inline void
Concrete::push_data(float x0, float y0, float x1, float y1,
    float x2, float y2) {
    // m_data.reserve(size()+6); ??
    m_data.push_back(x0); m_data.push_back(y0);
    m_data.push_back(x1); m_data.push_back(y1);
    m_data.push_back(x2); m_data.push_back(y2);
}

}} // namespace rvg::path
