#include <numeric>

#include "flatten.h"
#include "arc.h"
#include "../math/util.h"
#include "../math/lerp.h"
#include "../math/intersect.h"

// This module flattens all curves in a path according to
// three metrics: the maximum angle between two line segments
// in the polyline (tight), and the maximum distance between
// the curve and the flattened polyline (conservative), and
// the subdivision level (to avoid crashing on unforseen
// situations).
//
// It does not try to generate the minimum number of
// line segments, or even to be smart about it.
//
// It is supposed to run on paths that have already been
// monotonized, and all cubics already subdivided at the
// inflection points. It will only work "most of the time" if
// these conditions are not satisfied. ??
namespace rvg { namespace path {

// subdivision criterion for integral quadratic Bezier
static inline bool shouldsubdiv(const r2::Point &a, const r2::Point &b,
    const r2::Point &c, float ctol, float ptol) {
    auto ba = b-a;
    auto cb = c-b;
    auto ca = c-a;
    auto sq = [](float x) { return x*x; };
    float ca2 = dot(ca,ca);
    return sq(dot(ba,ca)) < sq(ctol)*dot(ba,ba)*ca2 ||
           sq(dot(cb,ca)) < sq(ctol)*dot(cb,cb)*ca2 ||
           sq(dot(cross(ca),ba)) >= 4.f*sq(ptol)*ca2;
}

// subdivision criterion for integral cubic Bezier
static inline bool shouldsubdiv(const r2::Point &a, const r2::Point &b,
    const r2::Point &c, const r2::Point &d, float ctol, float ptol) {
    auto da = d-a;
    auto ba = b-a;
    auto dc = d-c;
    float da2 = dot(da,da);
    auto sq = [](float x) { return x*x; };
    return sq(dot(ba,da)) < sq(ctol)*dot(ba,ba)*da2 ||
           sq(dot(dc,da)) < sq(ctol)*dot(dc,dc)*da2 ||
           std::max(sq(dot(cross(da),ba)), sq(dot(cross(da),dc))) >=
                4.f*sq(ptol)*da2;
}

// subdivision criterion for rational quadratic Bezier
static inline bool shouldsubdiv(const r3::Point &a, const r3::Point &b,
    const r3::Point &c, float ctol, float ptol) {
    const r2::Point r2a{a.x,a.y}, r2b{b.x,b.y}, r2c{c.x,c.y};
    auto ba = c.z*(r2b*a.z-r2a*b.z);
    auto cb = a.z*(r2c*b.z-r2b*c.z);
    auto ca = b.z*(r2c*a.z-r2a*c.z);
    float ca2 = dot(ca,ca);
    auto sq = [](float x) { return x*x; };
    return sq(dot(ba,ca)) < sq(ctol)*dot(ba,ba)*ca2 ||
           sq(dot(cb,ca)) < sq(ctol)*dot(cb,cb)*ca2 ||
           sq(dot(cross(ca),ba)) >= 4.f*sq(ptol*a.z*b.z*c.z)*ca2;
}

static void subdiv(r2::Point a, r2::Point b, r2::Point c, r2::Point d,
        Concrete &path) {
    r2::Point am = math::lerp(a,b,c,d,.0f,.0f,.5f);
    r2::Point bm = math::lerp(a,b,c,d,.0f,.5f,.5f);
    r2::Point m = math::lerp(a,b,c,d,.5f,.5f,.5f);
    r2::Point mc = math::lerp(a,b,c,d,.5f,.5f,1.f);
    r2::Point md = math::lerp(a,b,c,d,.5f,1.f,1.f);
    path.cubic_to_abs(am.x, am.y, bm.x, bm.y, m.x, m.y);
    path.cubic_to_abs(mc.x, mc.y, md.x, md.y, d.x, d.y);
}

static void subdiv(r2::Point a, r2::Point b, r2::Point c, Concrete &path) {
    r2::Point am = math::lerp(a,b,c,0.f,.5f);
    r2::Point m = math::lerp(a,b,c,.5f,.5f);
    r2::Point mc = math::lerp(a,b,c,.5f,1.f);
    path.quad_to_abs(am.x, am.y, m.x, m.y);
    path.quad_to_abs(mc.x, mc.y, c.x, c.y);
}

static void subdiv(r3::Point a, r3::Point b, r3::Point c, Concrete &path) {
    r3::Point am = math::lerp(a,b,c,0.f,.5f);
    r3::Point m = math::lerp(a,b,c,.5f,.5f);
    r3::Point mc = math::lerp(a,b,c,.5f,1.f);
    // Concrete paths only support arcs in canonical form
    // I.e., the first and last points must have w = 1.
    // But the subdivision process generates general arcs,
    // where all control points can have w != 1.
    // Here we convert at each step back into cannonic form.
    // This is silly, because we should perform this
    // conversion only at the last subdivision step. ??
    RationalArc arc1{a, am, m};
    RationalArc arc2{m, mc, c};
    path.rquad_to_abs(arc1.p1().x, arc1.p1().y, arc1.p1().z,
            arc1.p2().x, arc1.p2().y);
    path.rquad_to_abs(arc2.p1().x, arc2.p1().y, arc2.p1().z,
            arc2.p2().x, arc2.p2().y);
}

bool endpoint(SegmentType s, int i,
    const std::vector<float> &data, r2::Point &p) {
    switch (s) {
        case SegmentType::close_path:
        case SegmentType::fline_to:
            return false;
        case SegmentType::move_to:
            p = r2::Point{data[i+0], data[i+1]};
            return false;
        case SegmentType::line_to:
            p = r2::Point{data[i+2], data[i+3]};
            return true;
        case SegmentType::quad_to:
            p = r2::Point{data[i+4], data[i+4]};
            return true;
        case SegmentType::cubic_to:
            p = r2::Point{data[i+6], data[i+7]};
            return true;
        case SegmentType::rquad_to:
            p = r2::Point{data[i+5], data[i+6]};
            return true;
        default:
            std::cerr << s << '\n';
            // should never be here!
            assert(0);
            return false;
    }
}

void mark(const Concrete &path, float ctol, float ptol,
        std::vector<unsigned char> &flags) {
    const auto &data = path.data();
    const auto &segments = path.segments();
    // the flags vector already contained only zeros. grow
    // it with more zeros so we have one entry for each segment
    flags.resize(segments.size(), 0);
    // check subdivision for distance and angle tolerances
    for (int j = 0; j < int(segments.size()); j++) {
        const auto &segment = segments[j];
        const int i = segment.index;
        SegmentType s = static_cast<SegmentType>(segment.type);
        switch (s) {
            case SegmentType::close_path:
            case SegmentType::fline_to:
            case SegmentType::move_to:
            case SegmentType::line_to:
                break;
            case SegmentType::quad_to:
                if (shouldsubdiv(r2::Point{data[i+0], data[i+1]},
                    r2::Point{data[i+2], data[i+3]},
                    r2::Point{data[i+4], data[i+5]},
                    ctol, ptol)) {
                    flags[j] = 1;
                }
                break;
            case SegmentType::cubic_to:
                if (shouldsubdiv(r2::Point{data[i+0], data[i+1]},
                    r2::Point{data[i+2], data[i+3]},
                    r2::Point{data[i+4], data[i+5]},
                    r2::Point{data[i+6], data[i+7]},
                    ctol, ptol)) {
                    flags[j] = 1;
                }
                break;
            case SegmentType::rquad_to:
                if (shouldsubdiv(r3::Point{data[i+0], data[i+1], 1.f},
                    r3::Point{data[i+2], data[i+3], data[i+4]},
                    r3::Point{data[i+5], data[i+6], 1.f},
                    ctol, ptol)) {
                    flags[j] = 1;
                }
                break;
            default:
                std::cerr << s << '\n';
                // should never be here!
                assert(0);
                break;
        }
    }
}

Concrete subdiv(const Concrete &path, std::vector<unsigned char> &flags) {
    Concrete divided;
    const auto &data = path.data();
    const auto &segments = path.segments();
    // check subdivision for distance and angle tolerances
    for (int j = 0; j < int(segments.size()); j++) {
        const auto &segment = segments[j];
        const int i = segment.index;
        SegmentType s = static_cast<SegmentType>(segment.type);
        switch (s) {
            // pass-through segments
            case SegmentType::close_path:
                divided.close_path();
                break;
            case SegmentType::move_to:
                divided.move_to_abs(data[i], data[i+1]);
                break;
            case SegmentType::line_to:
                divided.line_to_abs(data[i+2], data[i+3]);
                break;
            case SegmentType::quad_to:
                if (flags[j]) {
                    subdiv(r2::Point{data[i+0], data[i+1]},
                        r2::Point{data[i+2], data[i+3]},
                        r2::Point{data[i+4], data[i+5]},
                        divided);
                } else {
                    divided.quad_to_abs(data[i+2], data[i+3],
                            data[i+4], data[i+5]);
                }
                break;
            case SegmentType::cubic_to:
                if (flags[j]) {
                    subdiv(r2::Point{data[i+0], data[i+1]},
                        r2::Point{data[i+2], data[i+3]},
                        r2::Point{data[i+4], data[i+5]},
                        r2::Point{data[i+6], data[i+7]},
                        divided);
                } else {
                    divided.cubic_to_abs(data[i+2], data[i+3], data[i+4],
                        data[i+5], data[i+6], data[i+7]);
                }
                break;
            case SegmentType::rquad_to:
                if (flags[j]) {
                    subdiv(r3::Point{data[i+0], data[i+1], 1.f},
                        r3::Point{data[i+2], data[i+3], data[i+4]},
                        r3::Point{data[i+5], data[i+6], 1.f},
                        divided);
                } else {
                    divided.rquad_to_abs(data[i+2], data[i+3], data[i+4],
                        data[i+5], data[i+6]);
                }
                break;
            // ignored segments
            case SegmentType::fline_to:
                break;
            default:
                std::cerr << s << '\n';
                // should never be here!
                assert(0);
                break;
        }
        // clear flag for following breadth-first iterations
        flags[j] = 0;
    }
    return divided;
}

flat::Path toflat(const Concrete &path) {
    flat::Path flat;
    flat::Contour contour;
    const auto &data = path.data();
    for (const auto &segment: path.segments()) {
        const int i = segment.index;
        SegmentType s = static_cast<SegmentType>(segment.type);
        switch (s) {
            case SegmentType::close_path:
                contour.close();
                break;
            case SegmentType::move_to:
                flat.append(std::move(contour));
                contour = flat::Contour{};
                contour.move_to(r2::Point{data[i], data[i+1]});
                break;
            case SegmentType::line_to:
                contour.line_to(r2::Point{data[i+2], data[i+3]});
                break;
            case SegmentType::quad_to:
                contour.line_to(r2::Point{data[i+4], data[i+5]});
                break;
            case SegmentType::cubic_to:
                contour.line_to(r2::Point{data[i+6], data[i+7]});
                break;
            case SegmentType::rquad_to:
                contour.line_to(r2::Point{data[i+5], data[i+6]});
                break;
            case SegmentType::fline_to:
                break; // ignore
            default:
                std::cerr << s << '\n';
                // should never be here!
                assert(0);
                break;
        }
    }
    flat.append(std::move(contour));
    return flat;
}

Concrete subdiv(const Concrete &path, const Transform &xf, 
        float ctol, float ptol, int maxdiv) {
    Concrete concrete(path, xf);
    std::vector<unsigned char> flags(path.segments().size(), 0);
    for (int i = 0; i < maxdiv; i++) {
        mark(concrete, ctol, ptol, flags);
        if (std::accumulate(flags.begin(), flags.end(), 0) == 0)
            break;
        concrete = subdiv(concrete, flags);
    }
    return concrete;
}

flat::Path flatten(const Concrete &path, const Transform &xf, 
    float ctol, float ptol, int maxdiv) {
    return toflat(subdiv(path, xf, ctol, ptol, maxdiv));
}

}} // namespace rvg::path
