#ifndef RVG_PATH_FLATTEN_H
#define RVG_PATH_FLATTEN_H

#include <vector>
#include <cfloat>
#include "concrete.h"
#include <boost/iterator/iterator_facade.hpp>

namespace rvg { namespace path { namespace flat {

struct Segment final {
    r2::Point p1;
    float length;
};

class Contour final {
public:
    explicit Contour(void): m_prev{0.f,0.f}, m_length(0.f), m_closed(false) {
        // first entry holds the move_to point
        // last entry is a sentinel for the reverse iterator
        // neither of them need to be initialized with meaninful values
        m_segments.push_back({m_prev, 0.f});
        m_segments.push_back({m_prev, 0.f});
    }

    bool empty(void) const { return m_segments.size() < 3; }

    void move_to(const r2::Point p) {
        assert(empty());
        m_segments[0].p1 = m_prev = p;
    }

    void line_to(const r2::Point &p) {
        r2::Vector dp = p-m_prev;
        line_to(p, sqrt(dot(dp,dp)));
    }

    void line_to(const r3::Point &p) {
        line_to(r2::Point{p.x/p.z, p.y/p.z});
    }

    void line_to(const r2::Point &p, float length) {
        if (length > FLT_MIN) {
            m_segments.back().p1 = p; // overwrite sentinel
            m_segments.back().length = length;
            m_segments.push_back({p, length});   // allocate new sentinel
            m_length += length;
            m_prev = p;
        }
    }

    bool is_closed() const { return m_closed; }

    void close(void) {
        m_closed = true;
    }

    float legth(void) const { return m_length; }

    typename std::vector<Segment>::const_iterator begin(void) const {
        return m_segments.begin();
    }

    typename std::vector<Segment>::const_iterator end(void) const {
        return m_segments.end()-1; // skip sentinel
    }

    class const_reverse_iterator:
        public boost::iterator_facade<const_reverse_iterator,
            Segment, boost::forward_traversal_tag, Segment> {
    public:
        const_reverse_iterator(std::vector<Segment>::const_iterator iter):
            m_iter(iter) {}

    private:

        void increment(void) {
            --m_iter;
        }

        bool equal(const const_reverse_iterator &other) const {
            return m_iter == other.m_iter;
        }

        Segment dereference(void) const {
            return Segment{m_iter[-1].p1, m_iter[0].length};
        }

        std::vector<Segment>::const_iterator m_iter;

        friend class boost::iterator_core_access;
    };

    const_reverse_iterator rbegin(void) const {
        return const_reverse_iterator(m_segments.end()-1); // skip sentinel
    }

    const_reverse_iterator rend(void) const {
        return const_reverse_iterator(m_segments.begin());
    }

private:
    r2::Point m_prev;  // previous target of a line_to;
    float m_length;
    bool m_closed;
    std::vector<Segment> m_segments;

};

static inline
std::ostream &operator<<(std::ostream &out, const flat::Contour &contour) {
    out << (contour.is_closed()? "closed": "open");
    for (const flat::Segment &s: contour) {
        out << ' ' << s.p1 << " [" << s.length << ']';
    }
    return out;
}

class Path final {
public:
    Path(): m_closed(false) { }

    operator Concrete() const {
        Concrete path;
        for (const auto &contour: m_contours) {
            if (!contour.empty()) {
                auto s = contour.begin();
                path.move_to_abs(s->p1.x, s->p1.y);
                for (++s; s != contour.end(); ++s) {
                    path.line_to_abs(s->p1.x, s->p1.y);
                }
                if (contour.is_closed()) path.close_path();
            }
        }
        return path;
    }

    void append(const Contour &contour) {
        if (!contour.empty())
            m_contours.emplace_back(contour);
    }

    void append(Contour &&contour) {
        if (!contour.empty())
            m_contours.emplace_back(std::move(contour));
    }

    const std::vector<Contour> &contours() const { return m_contours; }

    bool is_closed(void) const { return m_closed; }

    void close(void) { m_closed = true; }

private:
    bool m_closed;
    std::vector<Contour> m_contours;
};

static inline
std::ostream &operator<<(std::ostream &out, const flat::Path &path) {
    int i = 0;
    out << (path.is_closed()? "closed\n": "open\n");
    for (const auto &c: path.contours()) {
        out << "  contour " << i++ << ' ' << c << '\n';
    }
    return out;
}

#if 0
static inline
std::ostream &operator<<(std::ostream &out, const Path &path) {
    int i = 0;
    for (const auto &contour: path.contours()) {
        out << "contour " << i++ << " (" << contour.is_closed() << ") " <<
            contour;
    }
    return out;
}
#endif



} // namespace rvg::path::flat

flat::Path flatten(const Concrete &path, const Transform &xf, 
    float ctol, float ptol, int max = 10);

Concrete subdiv(const Concrete &path, const Transform &xf, 
    float ctol, float ptol, int max = 10);

}} // namespace rvg::path

#endif
