#ifndef RVG_PATH_H
#define RVG_PATH_H

#include <iosfwd>
#include <initializer_list>

#include "../util/refcnt.h"
#include "../xform/transformable.h"
#include "../constants.h"
#include "../iterator/base.h"
#include "command.h"
#include "base.h"

namespace rvg { namespace path {

// This is the Path object actually returned to the user
// The indirection to Base is vital for the copy-on-write semantics
class Path final: public Transformable<Base> {
public:

    Path(): Transformable<Base>(nullptr, Transform{}) {}

    using const_iterator = Base::const_iterator;

    explicit Path(Base *base, const Transform &xf = Transform{}):
        Transformable<Base>(base, xf) {}

    explicit Path(const Base &base, const Transform &xf = Transform{}):
        Transformable<Base>(base, xf) {}

    Path &append(int ncommands, const char *types, const float *arguments) {
        rw()->append(ncommands, types, arguments);
        return *this;
    }

    Path &append(const Command &cmd) {
        rw()->append(cmd);
        return *this;
    }

    Path &append(const std::initializer_list<Command> commands) {
        rw()->append(commands);
        return *this;
    }

    Path &append(const Path &other) {
        rw()->append(*other.ro(), other.xf());
        return *this;
    }

    // atol is the algular tolerance for subdivision, in degrees
    // swtol is the distance tolerance, in units of stroke width
    Path &stroke(const stroke::Style &style,
        float atol, float swtol, int maxdiv) {
        reset(ro()->stroke_clone(xf(), style, atol, swtol, maxdiv));
        set_xf({});
        return *this;
    }

    // ptol is the distance tolerance
    Path &flatten(float atol, float ptol, int maxdiv) {
        reset(ro()->flatten_clone(xf(), atol, ptol, maxdiv));
        set_xf({});
        return *this;
    }

    Path &subdiv(float atol, float ptol, int maxdiv) {
        reset(ro()->subdiv_clone(xf(), atol, ptol, maxdiv));
        set_xf({});
        return *this;
    }

    Path &transform(const Transform &xf) {
        Transformable<Base>::transform(xf);
        return *this;
    }

    const_iterator begin(void) const {
        return ro()->begin();
    }

    const_iterator end(void) const {
        return ro()->end();
    }

    friend std::ostream &operator<<(std::ostream &out, const Path &path);
};

} // namespace rvg::path

using path::Path;

} // namespace rvg

#include "path.hpp"

#endif // RVG_PATH_H
