#include <vector>
#include <iostream>
#include <numeric>

#define STATS 0

#include "../math/util.h"
#include "../math/lerp.h"
#include "../math/intersect.h"
#include "arc.h"
#include "flatten.h"
#include "stroke.h"

namespace rvg { namespace path {

class Dashed {
public:
    operator Concrete() const {
        Concrete path;
        for (const flat::Path &flat: m_paths) {
            path.append(static_cast<Concrete>(flat));
        }
        return path;
    }

    void append(const flat::Path &path) {
        if (!path.contours().empty()) {
            m_paths.emplace_back(path);
        }
    }

    void append(flat::Path &&path) {
        if (!path.contours().empty()) {
            m_paths.emplace_back(std::move(path));
        }
    }

    const std::vector<flat::Path> &paths(void) const { return m_paths; }

private:

    std::vector<flat::Path> m_paths;
};

static inline
std::ostream &operator<<(std::ostream &out, const Dashed &dashed) {
    int i = 0;
    for (const auto &p: dashed.paths()) {
        out << "path " << i++ << ' ' << p;
    }
    return out;
}


static std::vector<float> adjust_dash_array(const stroke::Style style,
    bool &i_on, float &i_phase, int &i_dash) {
    std::vector<float> array;
    for (float d: style.dash.array) {
        if (d < FLT_MIN) return array;
    }
    if (style.dash.array.empty()) return array;
    // odd-sized dash array must be doubled to be even-sized
    // eliminate degenerate dashes from array
    array = style.dash.array;
    if (array.size() % 2 != 0) {
        array.reserve(2*array.size());
        array.insert(array.end(), array.begin(), array.end());
    }
    // compute initial phase, on state, and dash index
    float pattern_length = std::accumulate(array.begin(), array.end(), 0.f);
    i_phase = fmod(style.dash.initial_phase, pattern_length);
    i_dash = 0;
    i_on = true;
    // advance dashes until we are within the dash of the initial phase
    while (i_phase >= array[i_dash]) {
        i_phase -= array[i_dash];
        i_dash = (i_dash+1) % array.size();
        i_on = !i_on;
    }
    return array;
}

// We eliminate degenerate line segments here.
// Ideally, we wouldn't really eliminate them, but rather still
// add the caps. ??
Dashed dash(const flat::Path &path, const stroke::Style &style) {
    Dashed dashed;
    // initial phase, on state, and dash index
    float i_phase = 0.f;
    bool i_on = true;
    int i_dash = 0;
    std::vector<float> array = adjust_dash_array(style, i_on, i_phase, i_dash);
    if (!array.empty()) {
        // each contour will become a new path
        // each dash in a contour becomes a contour in this new path.
        // current phase, on state, and dash index
        float c_phase = i_phase;
        bool c_on = i_on;
        int c_dash = i_dash;
        flat::Path dashed_contour;
        flat::Contour dash{};
        for (const auto &contour: path.contours()) {
            bool s_on = c_on;
            if (contour.empty()) continue;
            auto s = contour.begin();
            r2::Point p0 = s->p1;
            for (++s; s != contour.end(); ++s) {
                const r2::Point p1 = s->p1;
                const float len = s->length;
                float remaining = len;
                // if segment is long enough to finish current dash
                float needed = array[c_dash]-c_phase;
                if (needed <= len) {
                    // first dash is partial
                    if (c_on) {
                        r2::Point p = math::lerp(p0, p1, needed/len);
                        if (dash.empty()) dash.move_to(p0);
                        if (needed > FLT_MIN) dash.line_to(p, needed);
                        dashed_contour.append(std::move(dash));
                        dash = flat::Contour{};
                    }
                    remaining -= needed;
                    c_dash = (c_dash+1) % array.size();
                    c_on = !c_on;
                    // intermediate dashes are full
                    while (remaining > array[c_dash]) {
                        if (c_on) {
                            float s0 = remaining/len;
                            float s1 = (remaining-array[c_dash])/len;
                            r2::Point d0 = math::lerp(p1, p0, s0);
                            r2::Point d1 = math::lerp(p1, p0, s1);
                            dash.move_to(d0);
                            dash.line_to(d1, array[c_dash]);
                            dashed_contour.append(std::move(dash));
                            dash = flat::Contour{};
                        }
                        remaining -= array[c_dash];
                        c_dash = (c_dash+1) % array.size();
                        c_on = !c_on;
                    }
                    // last dash is also partial
                    if (c_on) {
                        r2::Point d0 = math::lerp(p1, p0, remaining/len);
                        if (remaining > FLT_MIN) {
                            dash.move_to(d0);
                            dash.line_to(p1, remaining);
                        } else {
                            dash.move_to(p1);
                        }
                    }
                    c_phase = remaining;
                // segment only partially completes dash
                } else {
                    if (c_on) {
                        if (dash.empty()) dash.move_to(p0);
                        if (len > FLT_MIN) dash.line_to(p1, len);
                    }
                    c_phase += len;
                }
                p0 = p1;
            }
            // last dash (could be empty. if so, it will be ignored)
            dashed_contour.append(std::move(dash));
            dash = flat::Contour{};
            // if state was on when we started the first
            // dash and when we finished the last dash,
            // close the path
            if (s_on && c_on && contour.is_closed()) dashed_contour.close();
            dashed.append(std::move(dashed_contour));
            dashed_contour = flat::Path{};
            // potentially reset state between contours
            if (style.dash.phase_reset) {
                c_phase = i_phase;
                c_on = i_on;
                c_dash = i_dash;
            }
        }
    } else {
        // a solid stroke is like a stroke with a single dash
        // each contour becomes a path with a single contour
        flat::Path solid_contour;
        for (const auto &contour: path.contours()) {
            solid_contour.append(contour);
            if (contour.is_closed()) solid_contour.close();
            dashed.append(std::move(solid_contour));
            solid_contour = flat::Path{};
        }
    }
    return dashed;
}

static void joinmiter(const r2::Point &p00o, const r2::Point &p01o,
    const r2::Point &p10o, const r2::Point &p11o, const r2::Point &p,
    const stroke::Style &style, Concrete &output) {
    auto sq = [](float x) { return x*x; };
    r3::Point v = math::intersect(p00o, p01o, p10o, p11o);
    // if miter is within limit, use intersection
    if (sq(v.x-p.x*v.z)+sq(v.y-p.y*v.z) <= sq(style.miter_limit*v.z)) {
        output.line_to_abs(v.x/v.z, v.y/v.z);
        output.line_to_abs(p10o.x, p10o.y);
    // else, limit
    } else {
    // OpenXPS Specification and Reference Guide suggests a
    // complicated approach that clips the miter parallel to the bevel,
    // at a distance from the shared corner that is equal to the miter limit.
#if 0
        r2::Vector vd = r2::Point{v.x, v.y}-v.z*p; // direction from p to v
        float d = .5f*style.width*style.miter_limit/sqrt(dot(vd, vd));
        // point at the miter_limit distance away from p in direction of v
        r2::Point m = p+d*vd;
        r3::Point c0 = math::intersect(p00o, p01o, m, m+cross(vd));
        r3::Point c1 = math::intersect(p10o, p11o, m, m+cross(vd));
        output.line_to_abs(c0.x/c0.z, c0.y/c0.z);
        output.line_to_abs(c1.x/c1.z, c1.y/c1.z);
        output.line_to_abs(p10o.x, p10o.y);
#endif
    // SVG seems to simply convert to a bevel when the limit
    // is exceeded. This is much easier.
        output.line_to_abs(p10o.x, p10o.y);
    }
}

static void joinround(const r2::Point &p01o, const r2::Point &p10o,
    const stroke::Style &style, Concrete &output) {
    // This can be optimized. We know this
    // is a circle. We know the center. No
    // need to use the (conveinent) generic function. ??
    path::RationalArc a(p01o, .5f*style.width, .5f*style.width,
            1.f, 0.f, false, false, p10o);
    const r3::Point m = a.p1();
    output.rquad_to_abs(m.x, m.y, m.z, p10o.x, p10o.y);
    //offset.line_to_abs(p10o.x, p10o.y);
}

template <bool MOVETO>
void join(const r2::Point &p00, const r2::Point &p,
        const r2::Point &p11, float inv_len0, float inv_len1,
        const stroke::Style &style, float ctol, Concrete &output) {
    const r2::Vector dp0 = cross(p-p00);
    const r2::Point p01p = p+inv_len0*dp0; // p01"plus"
    const r2::Point p00p = p00+inv_len0*dp0; // p00"plus"
    const r2::Point p00m = p00-inv_len0*dp0; // p00"minus"
    const r2::Vector dp1 = cross(p11-p);
    const r2::Point p10p = p+inv_len1*dp1;
    const r2::Point p11p = p11+inv_len1*dp1;
    const r2::Point p11m = p11-inv_len1*dp1;
    // if we are starting a new path, move to starting point
    if (MOVETO) output.move_to_abs(p01p.x, p01p.y);
    // curving away from offset segment?
    if (dot(cross(dp0), dp1) < 0.f) {
        // if curve is sharp enough, need a fancy join
        if (dot(dp0,dp1)*inv_len0*inv_len1 <= ctol) {
            switch (style.join) {
                case JoinStyle::round:
                    joinround(p01p, p10p, style, output);
                    break;
                case JoinStyle::bevel:
                    output.line_to_abs(p10p.x, p10p.y);
                    break;
                case JoinStyle::miter:
                    joinmiter(p00p, p01p, p10p, p11p, p, style, output);
                    break;
            }
        } else {
            // bevel join is fine
            output.line_to_abs(p10p.x, p10p.y);
        }
    // curving into the offset segment
    } else {
        // the simple join may need additional points
        // to prevent self-intersections
#if 0
        if (math::intersects(p01p, p10p, p00p, p00m) ||
            math::intersects(p01p, p10p, p11p, p11m)) {
            output.line_to_abs(p.x, p.y);
        }
#else
        if (math::intersects(p01p, p10p, p00p, p00m) ||
            math::intersects(p01p, p10p, p11p, p11m)) {
            output.line_to_abs(p01p.x, p01p.y);
            output.line_to_abs(p.x, p.y);
            if (det(r3::Point{p,1.f},
                    r3::Point{p11p,1.f},
                    r3::Point{p00p,1.f}) < 0.f) {
                output.line_to_abs(p11p.x, p11p.y);
                output.line_to_abs(p00p.x, p00p.y);
            } else {
                output.line_to_abs(p00p.x, p00p.y);
                output.line_to_abs(p11p.x, p11p.y);
            }
            output.line_to_abs(p.x, p.y);
        }
#endif
        output.line_to_abs(p10p.x, p10p.y);
    }
}


template <bool MOVETO, class ITER0, class ITER1>
void join(ITER0 begin0, ITER0 end0, ITER1 begin1, ITER1 end1,
        const stroke::Style &style, float ctol, Concrete &output) {
    if (begin0 == end0) return; // not even one point
    const r2::Point p01 = begin0->p1;
    ++begin0;
    if (begin0 == end0) return; // not even one segment
    const r2::Point p00 = begin0->p1;
    float len0 = begin0->length;
    float inv_len0 = .5f*style.width/len0;
    if (begin1 == end1) return; // not even one point
    const r2::Point p10 = begin1->p1;
    ++begin1;
    if (begin1 == end1) return; // not even one segment
    const r2::Point p11 = begin1->p1;
    float len1 = begin1->length;
    float inv_len1 = .5f*style.width/len1;
    if (p01 != p10) {
        std::cerr << "p01 != p10 ??? " << p01 << ' ' << p10 << '\n';
    }
    assert(p01 == p10);
    const r2::Point p = p01;
    join<MOVETO>(p00, p, p11, inv_len0, inv_len1, style, ctol, output);
}

template <class ITER>
static void side(ITER begin, ITER end, const stroke::Style &style, float ctol,
    Concrete &output) {
    if (begin == end) return; // not a single point, ignore
    r2::Point p00 = begin->p1;
    ++begin;
    if (begin == end) return; // not a single segment, ignore
    float len0 = begin->length;
    r2::Point p = begin->p1;
    float inv_len0 = .5f*style.width/len0;
    r2::Point p01p = p+inv_len0*cross(p-p00);
    // first segment
    output.line_to_abs(p01p.x, p01p.y);
    // remaining segments
    while (1) {
        ++begin;
        if (begin == end) break;
        float len1 = begin->length;
        const r2::Point p11 = begin->p1;
        const r2::Vector dp1 = cross(p11-p);
        float inv_len1 = .5f*style.width/len1;
        r2::Point p11p = p11+inv_len1*dp1;
        // join old segment to new segment
        join<false>(p00, p, p11, inv_len0, inv_len1, style, ctol, output);
        // new segment
        output.line_to_abs(p11p.x, p11p.y);
        // save previous values for next iteration
        p00 = p; p = p11;
        inv_len0 = inv_len1;
    }
}

template <bool MOVETO, class ITER>
static void cap(ITER begin, ITER end, const stroke::Style &style,
    Concrete &output) {
    if (begin == end) return; // not a single point, skip
    r2::Point p0 = begin->p1;
    ++begin;
    if (begin == end) return; // not a single segment, skip
    float len = begin->length;
    r2::Point p1 = begin->p1;
    r2::Vector dp = cross(p1-p0);
    float inv_len = .5f*style.width/len;
    r2::Point q0 = p0-inv_len*dp;
    if (MOVETO) output.move_to_abs(q0.x, q0.y);
    r2::Point q1 = p0+inv_len*dp;
    switch (style.cap) {
        case CapStyle::butt: {
            output.line_to_abs(q1.x, q1.y);
            break;
        }
        case CapStyle::round: {
            r2::Vector m = .5f*cross(q1-q0);
            output.rquad_to_abs(m.x, m.y, 0.f, q1.x, q1.y);
            break;
        }
        case CapStyle::square: {
            r2::Point s0 = q0+inv_len*cross(dp);
            r2::Point s1 = q1+inv_len*cross(dp);
            output.line_to_abs(s0.x, s0.y);
            output.line_to_abs(s1.x, s1.y);
            output.line_to_abs(q1.x, q1.y);
            break;
        }
        case CapStyle::triangle: {
            r2::Point m = p0+inv_len*cross(dp);
            output.line_to_abs(m.x, m.y);
            output.line_to_abs(q1.x, q1.y);
            break;
        }
    }
}

Concrete offset(const Dashed &dashed, const stroke::Style &style, float ctol) {
    Concrete output;
    for (const auto &contour: dashed.paths()) {
        const auto &dashes = contour.contours();
        int c = contour.is_closed();
        // do internal dashes first (skip first and last)
        for (int i = c; i < int(dashes.size())-c; i++) {
            const auto &dash = dashes[i];
            cap<true>(dash.begin(), dash.end(), style, output);
            side(dash.begin(), dash.end(), style, ctol, output);
            cap<false>(dash.rbegin(), dash.rend(), style, output);
            side(dash.rbegin(), dash.rend(), style, ctol, output);
            output.close_path();
        }
        // do remaining dashes
        if (c) {
            // do first dash connected to last dash
            if (dashes.size() > 1) {
                const auto &first = dashes.front();
                const auto &last = dashes.back();
                cap<true>(last.begin(), last.end(), style, output);
                side(last.begin(), last.end(), style, ctol, output);
                join<false>(last.rbegin(), last.rend(), first.begin(),
                        first.end(), style, ctol, output);
                side(first.begin(), first.end(), style, ctol, output);
                cap<false>(first.rbegin(), first.rend(), style, output);
                side(first.rbegin(), first.rend(), style, ctol, output);
                join<false>(first.begin(), first.end(), last.rbegin(),
                        last.rend(), style, ctol, output);
                side(last.rbegin(), last.rend(), style, ctol, output);
                output.close_path();
            // stroke is a single closed dash
            } else if (dashes.size() > 0) {
                const auto &single = dashes.front();
                join<true>(single.rbegin(), single.rend(), single.begin(),
                        single.end(), style, ctol, output);
                side(single.begin(), single.end(), style, ctol, output);
                output.close_path();
                join<true>(single.begin(), single.end(), single.rbegin(),
                        single.rend(), style, ctol, output);
                side(single.rbegin(), single.rend(), style, ctol, output);
                output.close_path();
            }
        }
    }
    return output;
}

Concrete stroke(const Concrete &path, const Transform &xf,
    const stroke::Style &style, float ctol, float swtol, int maxdiv) {
    auto flat = flatten(path, xf, ctol, swtol*style.width, maxdiv);
    auto dashed = dash(flat, style);
    auto stroked = offset(dashed, style, ctol);
#if STATS
    auto in = path.segments().size();
    auto out = stroked.segments().size();
    std::cerr << double(out)/double(in) << ' ' << in << ' ' << out << '\n';
    if (out/in > 100) {
        std::cerr << path << '\n';
        std::cerr << style << '\n';
    }
#endif
    return stroked;
}

}} // namespace rvg::path

