#ifndef RVGC_H
#define RVGC_H
/* ----------------------------------------------------------------------
 * RVG vector graphics rendering C API
 * This library is distributed under the MIT License.
 * See notice at the end of this file.
 * ---------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* these are preliminary definitions. don't worry too much
 * about the types or about the values of constants. */

typedef void* rvg_handle;
typedef rvg_handle rvg_context;
typedef rvg_handle rvg_surface;
typedef rvg_handle rvg_path;
typedef rvg_handle rvg_paint;
typedef rvg_handle rvg_stroke_type;
typedef rvg_handle rvg_quadtree;
typedef rvg_handle rvg_color_ramp;
typedef rvg_handle rvg_scene;
typedef rvg_handle rvg_xform;

/* matrices
 * path-to-surface: affine transformation mapping path
 *   coordinates to pixel coordinates in the viewport
 * image-to-surface: projective transformation used by rvg_draw_image
 * paint-to-user: projective transformation from paint coordinates
 *   to user coordinates
 * surface-to-surface: projective transformation mapping
 *   surface coordinates to surface coordinates. the result
 *   is that the rendering behaves as if it was rendered to
 *   an image and then drawn as an image to the viewport via
 *   this transformation. */
typedef enum _rvg_matrix_mode {
  RVG_MATRIX_PATH_TO_SURFACE         = 0x1400, /* affine */
  RVG_MATRIX_IMAGE_TO_SURFACE        = 0x1401, /* projective */
  RVG_MATRIX_PAINT_TO_USER           = 0x1402, /* projective */
  RVG_MATRIX_SURFACE_TO_SURFACE      = 0x1403  /* projective */
} rvg_matrix_mode;

typedef enum _rvg_prefilter_kernel {
  RVG_DIRAC,
  RVG_BOX,
  RVG_CATMULL_ROM,
  RVG_MITCHELL_NETRAVALI,
  RVG_CARDINAL_CUBIC_BSPLINE
} rvg_prefilter_kernel;

/* Balzer, M., Schlomer, T., and Deussen, O. 2009.
 * Capacity-constrained point distributions: A variant of Lloyd’s method.
 * In SIGGRAPH ’09, 86:1–8. */
/* A. Kensler 2013.
 * Correlated Multi-Jittered Sampling
 * Pixar Technical Memo 13-01 */
/* Patterns *can* be tiled, but should really be independent per pixel area */
typedef enum _rvg_sampling_pattern {
  RVG_BALZER,
  RVG_KENSLER,
  RVG_REGULAR,
  RVG_STRATIFIED
} rvg_sampling_pattern;

typedef enum _rvg_sampling_rate {
  RVG_SAMPLING_RATE_POINT,
  RVG_SAMPLING_RATE_LOW,
  RVG_SAMPLING_RATE_MEDIUM,
  RVG_SAMPLING_RATE_HIGH,
  RVG_SAMPLING_RATE_CUSTOM
} rvg_sampling_rate;

/* We need a better definition of device, driver, and context.
 * I think a device is a combination of the type of hardware
 * where the renderer will run and a particular implementation
 * that runs on that hardeware.
 * A context is a combination of the device and whatever internal
 * rendering state it needs.
 * A driver is a class that encapsulates the context. */
typedef enum _rvg_device_type {
  RVG_DEVICE_REFERENCE,
  RVG_DEVICE_CPU,
  RVG_DEVICE_GPU,
  RVG_DEVICE_SVG
} rvg_device_type;

typedef enum _rvg_segment_type {
  RVG_CLOSE_PATH                     = 0,
  RVG_MOVE_TO                        = 1,
  RVG_LINE_TO                        = 2,
  RVG_HLINE_TO                       = 3,
  RVG_VLINE_TO                       = 4,
  RVG_QUAD_TO                        = 5,
  RVG_SQUAD_TO                       = 6,
  RVG_RQUAD_TO                       = 7,
  RVG_CUBIC_TO                       = 8,
  RVG_SCUBIC_TO                      = 9,
  RVG_FLINE_TO                       = 10,
  RVG_SVGARC_TO                      = 11
} rvg_segment_type;

typedef enum _rvg_abs_rel {
  RVG_ABSOLUTE                       = 0,
  RVG_RELATIVE                       = 1
} rvg_abs_rel;

/* the value of the constants is not important, but the fact
 * they are derived from the rvg_segment_type in this particular way
 * (i.e., rivg_segment_type*2 + rvg_abs_rel) is important! */
typedef enum _rvg_command_type {
  RVG_CLOSE_PATH_ABS                 = (RVG_CLOSE_PATH << 1) | RVG_ABSOLUTE,
  RVG_CLOSE_PATH_REL                 = (RVG_CLOSE_PATH << 1) | RVG_RELATIVE,
  RVG_MOVE_TO_ABS                    = (RVG_MOVE_TO << 1)    | RVG_ABSOLUTE,
  RVG_MOVE_TO_REL                    = (RVG_MOVE_TO << 1)    | RVG_RELATIVE,
  RVG_LINE_TO_ABS                    = (RVG_LINE_TO << 1)    | RVG_ABSOLUTE,
  RVG_LINE_TO_REL                    = (RVG_LINE_TO << 1)    | RVG_RELATIVE,
  RVG_HLINE_TO_ABS                   = (RVG_HLINE_TO << 1)   | RVG_ABSOLUTE,
  RVG_HLINE_TO_REL                   = (RVG_HLINE_TO << 1)   | RVG_RELATIVE,
  RVG_VLINE_TO_ABS                   = (RVG_VLINE_TO << 1)   | RVG_ABSOLUTE,
  RVG_VLINE_TO_REL                   = (RVG_VLINE_TO << 1)   | RVG_RELATIVE,
  RVG_QUAD_TO_ABS                    = (RVG_QUAD_TO << 1)    | RVG_ABSOLUTE,
  RVG_QUAD_TO_REL                    = (RVG_QUAD_TO << 1)    | RVG_RELATIVE,
  RVG_SQUAD_TO_ABS                   = (RVG_SQUAD_TO << 1)   | RVG_ABSOLUTE,
  RVG_SQUAD_TO_REL                   = (RVG_SQUAD_TO << 1)   | RVG_RELATIVE,
  RVG_RQUAD_TO_ABS                   = (RVG_RQUAD_TO << 1)   | RVG_ABSOLUTE,
  RVG_RQUAD_TO_REL                   = (RVG_RQUAD_TO << 1)   | RVG_RELATIVE,
  RVG_CUBIC_TO_ABS                   = (RVG_CUBIC_TO << 1)   | RVG_ABSOLUTE,
  RVG_CUBIC_TO_REL                   = (RVG_CUBIC_TO << 1)   | RVG_RELATIVE,
  RVG_SCUBIC_TO_ABS                  = (RVG_SCUBIC_TO << 1)  | RVG_ABSOLUTE,
  RVG_SCUBIC_TO_REL                  = (RVG_SCUBIC_TO << 1)  | RVG_RELATIVE,
  RVG_SVGARC_TO_ABS                  = (RVG_SVGARC_TO << 1)  | RVG_ABSOLUTE,
  RVG_SVGARC_TO_REL                  = (RVG_SVGARC_TO << 1)  | RVG_RELATIVE
} rvg_command_type;

typedef enum _rvg_cap_style {
  RVG_CAP_BUTT,
  RVG_CAP_ROUND,
  RVG_CAP_SQUARE,
  RVG_CAP_TRIANGLE
} rvg_cap_style;

typedef enum _rvg_join_style {
  RVG_JOIN_MITER,
  RVG_JOIN_ROUND,
  RVG_JOIN_BEVEL,
} rvg_join_style;

typedef enum _rvg_fill_rule {
  RVG_EVEN_ODD,
  RVG_NON_ZERO,
} rvg_fill_rule;

typedef enum _rvg_paint_type {
  RVG_PAINT_TYPE_SOLID,
  RVG_PAINT_TYPE_LINEAR_GRADIENT,
  RVG_PAINT_TYPE_RADIAL_GRADIENT,
  RVG_PAINT_TYPE_TEXTURE
} rvg_paint_type;

typedef enum _rvg_spread_method {
  RVG_SPREAD_PAD,
  RVG_SPREAD_REPEAT,
  RVG_SPREAD_REFLECT,
  RVG_SPREAD_TRANSPARENT
} rvg_spread_method;

typedef enum _rvg_blend_mode {
  RVG_BLEND_SRC,
  RVG_BLEND_SRC_OVER,
  RVG_BLEND_DST_OVER,
  RVG_BLEND_SRC_IN,
  RVG_BLEND_DST_IN,
  RVG_BLEND_MULTIPLY,
  RVG_BLEND_SCREEN,
  RVG_BLEND_DARKEN,
  RVG_BLEND_LIGHTEN,
  RVG_BLEND_ADDITIVE
} rvg_blend_mode;

typedef enum _rvg_xform_type {
  RVG_XFORM_IDENTITY,
  RVG_XFORM_TX,
  RVG_XFORM_TY,
  RVG_XFORM_ROTATION,
  RVG_XFORM_TRANSLATION,
  RVG_XFORM_SCALE,
  RVG_XFORM_SHEAR,
  RVG_XFORM_LINEAR,
  RVG_XFORM_AFFINE,
  RVG_XFORM_PROJECTIVE
} rvg_xform_type;

/* creates and returns a new context */
rvg_context rvg_new_context(rvg_device_type device_type);
/* decreases the reference count of the context
 * context is destroyed when no references are left */
void rvg_release_context(rvg_context context);
/* sets the active context
 * increases reference count */
void rvg_set_active_context(rvg_context context);
/* returns active context
 * increases reference count */
rvg_context rvg_get_active_context(void);

/* creates and returns a new surface */
rvg_surface rvg_new_surface(int width, int height);
/* sets the target surface for context
 * increases reference count */
void rvg_set_render_target(rvg_surface surface);
/* decreases the reference count of the surface
 * surface is destroyed when no references are left */
void rvg_release_surface(rvg_surface surface);

/* creates and returns a new path */
rvg_path rvg_new_path(void);
/* appends data to path */
void rvg_append_path_data(rvg_path path, int nsegments,
    const rvg_command_type *commands, const float *arguments);
/* decreases path reference count
 * path is destroyed when no references are left */
void rvg_release_path(rvg_path path);
/* appends a path to the and of another */
void rvg_append_path(rvg_path dest, rvg_path src);
/* transforms a path by the current path-to-surface matrix */
void rvg_transform_path(rvg_path path, rvg_xform xform);
/* interpolates between two paths */
void rvg_interpolate_path(rvg_path dest, rvg_path start, rvg_path end, float t);

/* returns a stroked version of path */
rvg_path rvg_stroke_path(rvg_path path, rvg_stroke_type stroke_type);

/* creates and returns a new scene command buffer */
rvg_scene rvg_new_scene(void);
/* decreases scene reference count
 * scene is destroyed when no references are left */
void rvg_release_scene(rvg_scene scene);

/* adds a path to the scene */
void rvg_fill(rvg_scene scene, rvg_path path);
void rvg_eofill(rvg_scene scene, rvg_path path);
/* adds multiple instances to the scene
 * instance i uses path paths[instances[i]],
 * transformed by xforms+i*xform_size(xforms_type)
 * (before and in addition to the current path-to-surface matrix) */
void rvg_fill_instanced(rvg_scene scene, int npaths, rvg_path *paths,
    int ninstances, const int *instances, rvg_xform_type xforms_type,
    const float *xforms);
void rvg_eofill_instanced(rvg_scene scene, int npaths, rvg_path *paths,
    int ninstances, const int *instances, rvg_xform_type xforms_type,
    const float *xforms);

/* builds and returns a tree for the scene under current viewport */
rvg_quadtree rvg_build_quadtree(rvg_scene scene);
/* decreases the reference count of the tree
 * tree is destroyed when no references are left */
void rvg_release_quadtree(rvg_quadtree tree);
/* renders the tree into the viewport (affected by surface-to-surface matrix) */
void rvg_render(rvg_quadtree tree);
/* draws an image into the viewport (affected by image-to-surface matrix) */
void rvg_draw_image(rvg_surface surface);

/* sets prefilter kernel, sampling pattern, and sampling rate */
void rvg_set_antialiasing(rvg_prefilter_kernel kernel,
    rvg_sampling_pattern pattern, rvg_sampling_rate rate);

/* set rendering viewport in render target
 * [vx0, vx1-1] x [vy0, vy1-1]
 * recall image coordinate convention is "y grows up, x grows right" */
void rvg_set_viewport(int vx0, int vy0, int vx1, int vy1);

/* multiplies current transformation by a matrix that maps a
 * window in user space to a viewport in the render target */
void rvg_window_viewport(float wx0, float wy0, float wx1, float wy1,
    int vx0, int vy0, int vx1, int vy1);

/* set matrix mode */
void rvg_set_matrix_mode(rvg_matrix_mode matrix_mode);
/* multiplies current transformation by ... */
void rvg_translate(float tx, float ty);
void rvg_scale(float sx, float sy);
void rvg_shear(float shx, float shy);
void rvg_rotate(float angle_degs);
/* multiply current matrix by general projection matrix
 * if matrix-mode is path-to-surface, the last row is ignored */
void rvg_multiply_matrix(const float *matrix_column_major);
/* returns the current matrix
 * always writes 9 values */
void rvg_get_matrix(float *matrix_column_major);
/* load identity into current matrix */
void rvg_load_identity(void);


/* resets the current clipping path */
void rvg_reset_clip(void);
/* intersects the current clipping path with a new path */
void rvg_clip(rvg_path path);
void rvg_eoclip(rvg_path path);

/* we could offer this function as well.
 * or perhaps think about how to provide more general Boolean operations */
/* negates the result of inside and outside for a path */
rvg_path rvg_negate_path(rvg_path path);

/* creates and returns a new color ramp */
rvg_color_ramp rvg_new_color_ramp(void);
/* appends stops into color ramp
 * maybe another format is better for specifying the stops */
void rvg_append_color_ramp_data(rvg_color_ramp color_ramp,
    int nstops, const float *stops, const float *r, const float *g,
    const float *b, const float *a);
/* decreases the reference count of the color ramp
 * color ramp is destroyed when no references are left */
void rvg_release_color_ramp(rvg_color_ramp color_ramp);

/* sets current paint */
void rvg_set_paint(rvg_paint paint);
/* creates a new solid paint, colors in sRGB */
rvg_paint rvg_new_color(float r, float g, float b, float a);
/* creates a new radial gradient paint */
rvg_paint rvg_new_radial_gradient(rvg_color_ramp color_ramp,
    rvg_spread_method spread_method, float cx, float cy,
    float fx, float fy, float r);
/* creates a new linear gradient paint */
rvg_paint rvg_new_linear_gradient(rvg_color_ramp color_ramp,
    rvg_spread_method spread_method, float p1x, float p1y, float p2x, float p2y);
/* creates a new texture paint from a surface */
rvg_paint rvg_new_texture(rvg_surface texture, rvg_spread_method spread_method);

/* creates and returns a new stroke type */
rvg_stroke_type rvg_new_stroke_type(float width, rvg_cap_style cap_style,
    rvg_join_style join_style, float miter_limit, int ndash_stops,
    const float *dash_stops, float initial_phase, int reset_phase);

/* clears the current viewport with a given color */
void rvg_clear_viewport(float r, float g, float b, float a);

/* sets the blend mode for operations on viewport */
void rvg_set_blend_mode(rvg_blend_mode blend_mode);

#ifdef __cplusplus
}
#endif

#endif /* RVGC_H */

/* ----------------------------------------------------------------------
 * Copyright (C) 2013 Diego Nehab. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ---------------------------------------------------------------------- */
