#ifndef RVG_SCENE_ELEMENT_H
#define RVG_SCENE_ELEMENT_H

#include "../path/path.h"
#include "../paint/paint.h"

namespace rvg { namespace scene {

class Element {
private:

    enum InternalType {
        // types of element
        _fill = 32,    // filled path
        _clip = 64,    // clip path
        _stack = 128,  // stack operation

        // fill rules
        _odd = 0,      // same as even-odd
        _even = 1,     // complement of even-odd
        _nonzero = 2,  // same as non-zero winding
        _zero = 3,     // complement of non-zero winding

        _push = 8,
        _pop  = 16,
    };

public:

    enum class Type {
        // stack operations
        commit   = _stack | 0,
        pushfill = _stack | _push | 0,
        popfill  = _stack | _pop  | 0,
        pushclip = _stack | _push | 1,
        popclip  = _stack | _pop  | 1,

        // filled paths
        ofill  = _fill | _odd,
        efill  = _fill | _even,
        nzfill = _fill | _nonzero,
        zfill  = _fill | _zero,

        // clip paths
        oclip  = _clip | _odd,
        eclip  = _clip | _even,
        nzclip = _clip | _nonzero,
        zclip  = _clip | _zero,
    };

    bool is_fill(void) const {
        return static_cast<int>(type) & _fill;
    }

    bool is_stack(void) const {
        return static_cast<int>(type) & _stack;
    }

    bool is_clip(void) const {
        return static_cast<int>(type) & _clip;
    }

    bool is_push(void) const {
        return (static_cast<int>(type) & _stack) &&
            (static_cast<int>(type) & _push);
    }

    bool is_pop(void) const {
        return (static_cast<int>(type) & _stack) &&
            (static_cast<int>(type) & _pop);
    }

    bool is_commit(void) const {
        return (static_cast<int>(type) & _stack) &&
            !(static_cast<int>(type) & (_push | _pop));
    }

    Element(void) = delete;

    Element(const Element &that): type(that.type) {
        if (that.is_fill()) {
            new(&fill.path) Path(that.fill.path);
            new(&fill.paint) Paint(that.fill.paint);
        } else if (that.is_clip()) {
            new(&clip.path) Path(that.clip.path);
            clip.depth = that.clip.depth;
        } else { // push,pop,commit
            stack.depth = that.stack.depth;
        }
    }

    Element(Element &&that): type(std::move(that.type)) {
        if (that.is_fill()) {
            new(&fill.path) Path(std::move(that.fill.path));
            new(&fill.paint) Paint(std::move(that.fill.paint));
        } else if (that.is_clip()) {
            new(&clip.path) Path(std::move(that.clip.path));
            clip.depth = std::move(that.clip.depth);
        } else { // push,pop,commit
            stack.depth = std::move(that.stack.depth);
        }
    }

    Element(Type _type, Path _path, Paint _paint): type(_type) {
        assert(static_cast<int>(type) & _fill);
        new(&fill.path) Path(std::move(_path));
        new(&fill.paint) Paint(std::move(_paint));
    }

    Element(Type _type, Path _path, int _depth): type(_type) {
        assert(static_cast<int>(type) & _clip);
        new(&clip.path) Path(std::move(_path));
        clip.depth = std::move(_depth);
    }

    Element(Type _type, int _depth): type(_type) {
        assert(static_cast<int>(type) & _stack);
        stack.depth = std::move(_depth);
    }

    ~Element() {
        if (is_fill()) {
            fill.path.~Path();
            fill.paint.~Paint();
        } else if (is_clip()) {
            clip.path.~Path();
        }
    }

    Element &transform(const Transform &xf) {
        if (is_fill()) {
            fill.path.transform(xf);
            fill.paint.transform(xf);
        } else if (is_clip()) {
            clip.path.transform(xf);
        }
        return *this;
    }

    Type type;
    union {
        struct {
            Path path;
            Paint paint;
        } fill;
        struct {
            Path path;
            int depth;
        } clip;
        struct {
            int depth;
        } stack;
    };

};

static inline
std::ostream &operator<<(std::ostream &out, const Element::Type &t) {
    switch (t) {
        case Element::Type::commit: out << "commit"; break;
        case Element::Type::pushfill: out << "pushfill"; break;
        case Element::Type::popfill: out << "popfill"; break;
        case Element::Type::pushclip: out << "pushclip"; break;
        case Element::Type::popclip: out << "popclip"; break;
        case Element::Type::ofill: out << "ofill"; break;
        case Element::Type::efill: out << "efill"; break;
        case Element::Type::nzfill: out << "nzfill"; break;
        case Element::Type::zfill: out << "zfill"; break;
        case Element::Type::oclip: out << "oclip"; break;
        case Element::Type::eclip: out << "eclip"; break;
        case Element::Type::nzclip: out << "nzclip"; break;
        case Element::Type::zclip: out << "zclip"; break;
    };
    return out;
}

static inline
std::ostream &operator<<(std::ostream &out ,const Element &e) {
    out << "element " << e.type << ' ';
    if (e.is_fill()) {
        out << e.fill.path << ' ' << e.fill.paint;
    } else if (e.is_clip()) {
        out << "depth " << e.clip.depth << ' ' << e.clip.path;
    } else if (e.is_stack()) {
        out << "depth " << e.stack.depth;
    }
    return out;
}

}} // namespace rvg::scene

#endif // RVG_SCENE_ELEMENT_H
