#ifndef RVG_SCENE_H
#define RVG_SCENE_H

#include <cassert>
#include <vector>
#include <stdexcept>

#include "../constants.h"
#include "../util/refcnt.h"
#include "../xform/transformable.h"
#include "../path/path.h"
#include "../paint/paint.h"
#include "element.h"

namespace rvg { namespace scene {

class RefElements final: public RefCounted {
public:
    RefPtr<RefElements> xform_clone(const Transform &xf) const;
    void transform(const Transform &xf);
    std::vector<Element> &elements(void) { return m_elements; }
    const std::vector<Element> &elements(void) const { return m_elements; }
private:
    std::vector<Element> m_elements;
};

class Scene: public Transformable<RefElements> {
public:
    using internal_const_iterator_base = iterator::Internal<const Element>;

    class internal_const_iterator: public internal_const_iterator_base {
    public:
        internal_const_iterator(const Scene *scene, int index): m_scene(scene),
            m_index(index) {}

        internal_const_iterator *clone(void) const {
            return new internal_const_iterator(m_scene, m_index);
        }

        void increment(void) { ++m_index; }

        int index(void) const { return m_index; }

        bool equal(const internal_const_iterator_base *base_iterator) const {
            const internal_const_iterator *other =
                static_cast<const internal_const_iterator *>(base_iterator);
            assert(m_scene == other->m_scene);
            return m_index == other->m_index;
        }

        internal_const_iterator_base::reference dereference(void) const {
            return m_scene->elements()[m_index];
        }

    private:
        const Scene* m_scene;
        int m_index;
    };

    using const_iterator = iterator::Iterator<internal_const_iterator>;

    Scene(void): Transformable<RefElements>(new RefElements),
        m_depth(0), m_uncommitted(0) {}

    Scene &fill(Path path, Paint paint,
        Element::Type type = Element::Type::nzfill);

    Scene &nzfill(const Path &path, const Paint &paint);
    Scene &zfill(const Path &path, const Paint &paint);
    Scene &eofill(const Path &path, const Paint &paint);
    Scene &ofill(const Path &path, const Paint &paint);
    Scene &efill(const Path &path, const Paint &paint);

    Scene &clip(Path path, Element::Type type = Element::Type::nzfill);

    Scene &nzclip(const Path &path);
    Scene &zclip(const Path &path);
    Scene &eoclip(const Path &path);
    Scene &oclip(const Path &path);
    Scene &eclip(const Path &path);

    Scene &push(void);
    Scene &commit(void);
    Scene &pop(void);

    Scene &transform(const Transform &xform) {
        Transformable<RefElements>::transform(xform);
        return *this;
    }

    Scene &append(const Element &element);

    const std::vector<Element> &elements(void) const {
        return ro()->elements();
    }

    // iterator interface
    const_iterator begin(void) const {
        return const_iterator(new internal_const_iterator(this, 0));
    }

    const_iterator end(void) const {
        return const_iterator(new internal_const_iterator(this,
            elements().size()));
    }

private:
    int m_depth;      // stack depth
    int m_uncommitted; // depth of uncommitted clip paths
};

} // namespace rvg::scene

using scene::Scene;

} // namespace rvg

#include "scene.hpp"

#endif // RVG_SCENE_HPP
