#ifndef RVG_COW_H
#define RVG_COW_H

#ifdef TRACE_COW
#include <iostream> // for std::clog
#endif

#include "cloneable.h"
#include "refcnt.h"

namespace rvg {

template<class T, bool CLONEABLE = is_cloneable<T>::value>
struct CloneFunctor
{
    RefPtr<T> operator()(const T &obj) const { return obj.clone(); }
};

template<class T>
struct CloneFunctor<T,false>
{
    RefPtr<T> operator()(const T &obj) const { return new T(obj); }
};

template <class BASE, class CLONE=CloneFunctor<BASE>>
class COW final {
public:
    COW() = default;
    explicit COW(RefPtr<BASE> base, CLONE cloner=CLONE())
        : m_data(std::move(base), std::move(cloner)) {}

    // const method returns read-only data and does not call copy-on-write
    const BASE *ro(void) const { return m_data.ptr.get(); }
    // non-const method returns read-write and calls copy-on-write
    BASE *rw(void) { cow(); return m_data.ptr.get(); }

    explicit operator bool() const { return m_data.ptr ? true : false; }

    void reset(RefPtr<BASE> obj = {})
    {
        m_data.ptr = std::move(obj);
    }

    const BASE *operator->() const { return ro(); }
    const BASE &operator*() const { return *ro(); }

    CLONE &cloner() { return m_data; }
    const CLONE &cloner() const { return m_data; }

    // after the call to cow(), we are the sigle owner of our data
    bool cow(void) {
        if (m_data.ptr && m_data.ptr->ref_count() > 1) {
#ifdef TRACE_COW
            std::clog << "-- cow cloning " << this << std::endl;
#endif

            m_data.ptr = m_data(*m_data.ptr);
            return true;
        } else {
#ifdef TRACE_COW
            std::clog << "-- cow passthrough " << this << std::endl;
#endif
            return false;
        }
    }
private:
    // EBO kicks in if CLONE is empty, saving us 1 byte
    struct Data : CLONE
    {
        Data() = default;
        Data(RefPtr<BASE> p, CLONE c)
            : CLONE(std::move(c))
            , ptr(std::move(p))
        {
        }

        RefPtr<BASE> ptr;
    } m_data;
};

#define IMPLEMENT_COW(BASE, member) \
    public: \
        BASE *rw(void) { return member.rw(); } \
        const BASE *ro(void) const { return member.ro(); }

} // namespace rvg

#endif // RVG_COW_H
