#ifndef RVG_REFCNT_H
#define RVG_REFCNT_H

#include <boost/intrusive_ptr.hpp>
#include <cassert>

#ifdef TRACE_REF
#include <iostream> // for std::clog
#endif

namespace rvg {

class RefCounted {
public:
    // don't copy the refcount, new object must have
    // refcnt==0
    RefCounted(const RefCounted &)
        : m_ref_count(0)
    {
    }

    RefCounted(void): m_ref_count(0) { ; }
    virtual ~RefCounted() {
        assert(m_ref_count == 0);
    }
    void add_ref(void) {
        ++m_ref_count;
#ifdef TRACE_REF
    std::clog << "-- up " << this << ": " << m_ref_count << std::endl;
#endif
    }
    void release(void) {
        if (--m_ref_count == 0) {
#ifdef TRACE_REF
            std::clog << "-- down " << this << ": " << m_ref_count << std::endl;
            std::clog << "-- deleting " << this << std::endl;
#endif
            delete this;
        }
#ifdef TRACE_REF
        else {
            std::clog << "-- down " << this << ": " << m_ref_count << std::endl;
        }
#endif
    }
    int ref_count(void) const {
        return m_ref_count;
    }

    RefCounted &operator=(const RefCounted&)
    {
        // don't copy the refcount, this copy operation
        // doesn't change the objetct's refcount
        return *this;
    }
protected:
    int m_ref_count;
};

inline void intrusive_ptr_add_ref(RefCounted *p) { p->add_ref(); }
inline void intrusive_ptr_release(RefCounted *p) { p->release(); }

template <class T>
using RefPtr = boost::intrusive_ptr<T>;

using boost::dynamic_pointer_cast;
using boost::static_pointer_cast;
using boost::const_pointer_cast;

} // namespace qvg

#endif // RVG_REFCNT_H
