namespace rvg { namespace xform
{

#if TRACE_XFORM
inline Affine::Affine()
{
    std::clog << "-- default ctor " << name << "()" << std::endl;
}
#endif

inline Affine::Affine(float a11, float a12, float a13,
                      float a21, float a22, float a23)
    : m_matrix{ {{{a11,a12,a13}},
                 {{a21,a22,a23}}} }
{
#if TRACE_XFORM
    std::clog << "-- ctor " << *this << std::endl;
#endif
}

inline Affine::Affine(const Identity &)
    : m_matrix{ {{{1.f,0.f,0.f}},
                 {{0.f,1.f,0.f}}} }
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Identity) " << *this << std::endl;
#endif
}

inline Affine::Affine(const Linear &b)
    : m_matrix{ {{{b[0][0],b[0][1],0.f}},
                 {{b[1][0],b[1][1],0.f}}} }
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Linear) " << *this << std::endl;
#endif
}

inline Affine::Affine(const Rotation &b): Affine(Linear(b)) // delegate
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Rotation) " << *this << std::endl;
#endif
}

inline Affine::Affine(const Scale &b): Affine(Linear(b)) // delegate
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Scale) " << *this << std::endl;
#endif
}

inline auto Affine::operator[](int i) 
    -> std::array<float,3> &
{
    return m_matrix[i];
}

inline auto Affine::operator[](int i) const
    -> const std::array<float,3> &
{
    return m_matrix[i];
}

inline Affine &Affine::operator *=(const Translation &that)
{
    auto &a = *this;
    const auto &b = that;

    a[0][2] += a[0][0]*b.dx + a[0][1]*b.dy;
    a[1][2] += a[1][0]*b.dx + a[1][1]*b.dy;

    return *this;
}

inline Affine &Affine::operator *=(const Rotation &that)
{
    auto &a = *this;
    const auto &b = that;

    float a00 = a[0][0];
    a[0][0] =  b.c*a00 + b.s*a[0][1];
    a[0][1] = -b.s*a00 + b.c*a[0][1];

    float a10 = a[1][0];
    a[1][0] =  b.c*a10 + b.s*a[1][1];
    a[1][1] = -b.s*a10 + b.c*a[1][1];

    return *this;
}

inline Affine &Affine::operator *=(const Scale &that)
{
    auto &a = *this;
    const auto &b = that;

    a[0][0] *= b.sx;
    a[0][1] *= b.sy;
    a[1][0] *= b.sx;
    a[1][1] *= b.sy;

    return *this;
}

inline Linear Affine::linear(void) const
{
    auto &a = *this;
    return Linear(a[0][0], a[0][1], a[1][0], a[1][1]);
}

inline Affine inv(const Affine &xf) {
    float a = xf[0][0], b = xf[0][1];
    float c = xf[1][0], d = xf[1][1];
    float tx = xf[0][2], ty = xf[1][2];
    float det = a*d - b*c;
    return Affine( d/det, -b/det, (b*ty - d*tx)/det,
                    -c/det, a/det, (c*tx - a*ty)/det );
}

inline Affine &Affine::operator *=(const Linear &that)
{
    auto &a = *this;
    const auto &b = that;

    float a00 = a[0][0];
    a[0][0] = a00*b[0][0] + a[0][1]*b[1][0];
    a[0][1] = a00*b[0][1] + a[0][1]*b[1][1];

    float a10 = a[1][0];
    a[1][0] = a10*b[0][0] + a[1][1]*b[1][0];
    a[1][1] = a10*b[0][1] + a[1][1]*b[1][1];

    return *this;
}

inline Affine &Affine::operator *=(const Affine &that)
{
    return *this = *this * that;
}

inline Affine operator*(const Affine &a, const Affine &b)
{
    Affine r;

    r[0][0] = a[0][0]*b[0][0]+a[0][1]*b[1][0];
    r[0][1] = a[0][0]*b[0][1]+a[0][1]*b[1][1];
    r[0][2] = a[0][0]*b[0][2]+a[0][1]*b[1][2] + a[0][2];

    r[1][0] = a[1][0]*b[0][0]+a[1][1]*b[1][0];
    r[1][1] = a[1][0]*b[0][1]+a[1][1]*b[1][1];
    r[1][2] = a[1][0]*b[0][2]+a[1][1]*b[1][2] + a[1][2];

    return std::move(r);
}

inline Affine operator*(const Translation &a, const Rotation &b)
{
    return Affine(b.c, -b.s, a.dx,
                  b.s,  b.c, a.dy);
}

inline Affine operator*(const Translation &a, const Scale &b)
{
    return Affine(b.sx,     0, a.dx,
                     0,  b.sy, a.dy);
}

inline Affine operator*(const Translation &a, const Linear &b)
{
    return Affine(b[0][0], b[0][1], a.dx,
                  b[1][0], b[1][1], a.dy);
}

inline Affine operator*(const Translation &a, const Affine &b)
{
    return Affine(b[0][0], b[0][1], b[0][2]+a.dx,
                  b[1][0], b[1][1], b[1][2]+a.dy);
}

inline Affine operator*(const Rotation &a, const Translation &b)
{
    return Affine(a.c, -a.s, a.c*b.dx - a.s*b.dy,
                  a.s,  a.c, a.s*b.dx + a.c*b.dy);
}

inline Affine operator*(const Rotation &a, const Affine &b)
{
    return Affine(a.c*b[0][0] - a.s*b[1][0],
                  a.c*b[0][1] - a.s*b[1][1],
                  a.c*b[0][2] - a.s*b[1][2],

                  a.s*b[0][0] + a.c*b[1][0],
                  a.s*b[0][1] + a.c*b[1][1],
                  a.s*b[0][2] + a.c*b[1][2]);
}

inline Affine operator*(const Scale &a, const Translation &b)
{
    return Affine(a.sx,     0, a.sx*b.dx,
                     0,  a.sy, a.sy*b.dy);
}

inline Affine operator*(const Scale &a, const Affine &b)
{
    return Affine(a.sx*b[0][0], a.sx*b[0][1], a.sx*b[0][2],
                  a.sy*b[1][0], a.sy*b[1][1], a.sy*b[1][2]);
}

inline Affine operator*(const Linear &a, const Translation &b)
{
    return Affine(a[0][0], a[0][1], a[0][0]*b.dx+a[0][1]*b.dy,
                  a[1][0], a[1][1], a[1][0]*b.dx+a[1][1]*b.dy);
}

inline Affine operator*(const Linear &a, const Affine &b)
{
    return Affine(a[0][0]*b[0][0]+a[0][1]*b[1][0],
                  a[0][0]*b[0][1]+a[0][1]*b[1][1],
                  a[0][0]*b[0][2]+a[0][1]*b[1][2],

                  a[1][0]*b[0][0]+a[1][1]*b[1][0],
                  a[1][0]*b[0][1]+a[1][1]*b[1][1],
                  a[1][0]*b[0][2]+a[1][1]*b[1][2]);
}

inline r2::Point Affine::do_multiply(const r2::Point &p) const
{
    auto &A = *this;

    return {A[0][0]*p.x + A[0][1]*p.y + A[0][2],
            A[1][0]*p.x + A[1][1]*p.y + A[1][2]};
}

inline r2::Vector Affine::do_multiply(const r2::Vector &v) const
{
    auto &A = *this;

    return {A[0][0]*v.x + A[0][1]*v.y,
            A[1][0]*v.x + A[1][1]*v.y};
}

inline r3::Point Affine::do_multiply(const r3::Point &p) const
{
    auto &A = *this;

    return {A[0][0]*p.x + A[0][1]*p.y + A[0][2]*p.z,
            A[1][0]*p.x + A[1][1]*p.y + A[1][2]*p.z,
            p.z};
}

}} // namespace rvg::xform
