#ifndef QVG_XFORM_BASE_H
#define QVG_XFORM_BASE_H

#include <iosfwd>
#include <type_traits>
#include <tuple>
#include "../math/point.h"
#include "../util/refcnt.h"
#include "../util/cloneable.h"
#include "visitor.h"

namespace rvg { namespace xform
{

class Visitor;

class Base : public Cloneable<Base>
{
public:
    virtual ~Base() {}

    virtual void accept(Visitor &v) const = 0;

    r2::Point operator*(const r2::Point &p) const;
    r2::Vector operator*(const r2::Vector &v) const;
    r3::Point operator*(const r3::Point &p) const;

    friend std::ostream &operator<<(std::ostream &out, const Base &xf);

private:
    // needs access to compose_into
    friend class Transform;

    // needs access to rev_compose_into
    template <class T> friend class BaseHelper;

    // perform actual composition, has access to Transform internals
    class DoComposition;

    // both need access to compose_into
    friend Transform operator*(const Base &lhs, const Transform &rhs);
    friend Transform &&operator*(const Base &lhs, Transform &&rhs);

    virtual void do_print(std::ostream &out) const = 0;

    virtual r2::Point do_multiply(const r2::Point &p) const = 0;
    virtual r2::Vector do_multiply(const r2::Vector &v) const = 0;
    virtual r3::Point do_multiply(const r3::Point &p) const = 0;

    // virtual template method pattern - base
    template <class T, class EN = typename std::enable_if<std::is_const<T>::value>::type>
    void compose_into(Transform &xf, T &b)
        { do_compose_into(xf, b); }

    template <class T>
    void compose_into(Transform &xf, T &b) const
        { do_compose_into(xf, b); }

    template <class T>
    void rev_compose_into(Transform &xf, T &a)
        { do_rev_compose_into(xf, a); }

    template <class T>
    void rev_compose_into(Transform &xf, T &a) const
        { do_rev_compose_into(xf, a); }

    // Base
    virtual void do_compose_into(Transform &xf, const Base &b) = 0;
    virtual void do_compose_into(Transform &xf, Base &b) const = 0;
    virtual void do_compose_into(Transform &xf, const Base &b) const = 0;

    // Identity
    virtual void do_compose_into(Transform &xf, const Identity &b) = 0;
    virtual void do_compose_into(Transform &xf, Identity &b) const = 0;
    virtual void do_compose_into(Transform &xf, const Identity &b) const = 0;

    virtual void do_rev_compose_into(Transform &xf, const Identity &a) = 0;
    virtual void do_rev_compose_into(Transform &xf, Identity &a) const = 0;
    virtual void do_rev_compose_into(Transform &xf, const Identity &a) const = 0;

    // Scale
    virtual void do_compose_into(Transform &xf, const Scale &b) = 0;
    virtual void do_compose_into(Transform &xf, Scale &b) const = 0;
    virtual void do_compose_into(Transform &xf, const Scale &b) const = 0;

    virtual void do_rev_compose_into(Transform &xf, const Scale &a) = 0;
    virtual void do_rev_compose_into(Transform &xf, Scale &a) const = 0;
    virtual void do_rev_compose_into(Transform &xf, const Scale &a) const = 0;

    // Rotation
    virtual void do_compose_into(Transform &xf, const Rotation &b) = 0;
    virtual void do_compose_into(Transform &xf, Rotation &b) const = 0;
    virtual void do_compose_into(Transform &xf, const Rotation &b) const = 0;

    virtual void do_rev_compose_into(Transform &xf, const Rotation &a) = 0;
    virtual void do_rev_compose_into(Transform &xf, Rotation &a) const = 0;
    virtual void do_rev_compose_into(Transform &xf, const Rotation &a) const = 0;

    // Translation
    virtual void do_compose_into(Transform &xf, const Translation &b) = 0;
    virtual void do_compose_into(Transform &xf, Translation &b) const = 0;
    virtual void do_compose_into(Transform &xf, const Translation &b) const = 0;

    virtual void do_rev_compose_into(Transform &xf, const Translation &a) = 0;
    virtual void do_rev_compose_into(Transform &xf, Translation &a) const = 0;
    virtual void do_rev_compose_into(Transform &xf, const Translation &a) const = 0;

    // Linear
    virtual void do_compose_into(Transform &xf, const Linear &b) = 0;
    virtual void do_compose_into(Transform &xf, Linear &b) const = 0;
    virtual void do_compose_into(Transform &xf, const Linear &b) const = 0;

    virtual void do_rev_compose_into(Transform &xf, const Linear &a) = 0;
    virtual void do_rev_compose_into(Transform &xf, Linear &a) const = 0;
    virtual void do_rev_compose_into(Transform &xf, const Linear &a) const = 0;

    // Affine
    virtual void do_compose_into(Transform &xf, const Affine &b) = 0;
    virtual void do_compose_into(Transform &xf, Affine &b) const = 0;
    virtual void do_compose_into(Transform &xf, const Affine &b) const = 0;

    virtual void do_rev_compose_into(Transform &xf, const Affine &a) = 0;
    virtual void do_rev_compose_into(Transform &xf, Affine &a) const = 0;
    virtual void do_rev_compose_into(Transform &xf, const Affine &a) const = 0;

    // Projective
    virtual void do_compose_into(Transform &xf, const Projective &b) = 0;
    virtual void do_compose_into(Transform &xf, Projective &b) const = 0;
    virtual void do_compose_into(Transform &xf, const Projective &b) const = 0;

    virtual void do_rev_compose_into(Transform &xf, const Projective &a) = 0;
    virtual void do_rev_compose_into(Transform &xf, Projective &a) const = 0;
    virtual void do_rev_compose_into(Transform &xf, const Projective &a) const = 0;
};

// use CRTP to implement some pure methods
template <class T>
class BaseHelper
    : public Base
{
public:
#if TRACE_XFORM
    static void *operator new(std::size_t s);
    static void *operator new(std::size_t s, std::nothrow_t) throw();
    static void *operator new(std::size_t s, void *p);
#endif

    void accept(Visitor &v) const final;

    RefPtr<T> clone() const { return static_pointer_cast<T>(Base::clone()); }

private:
    T &derived();
    const T &derived() const;

    Base *do_clone() const override;

    template <class U> friend class BaseHelper;

    // virtual template method pattern - derived
    template <class U>
    void compose_into(Transform &a, const U &b);

    template <class U>
    void rev_compose_into(Transform &xf, U &a) const;

    template <class U>
    void compose_into(Transform &a, U &b) const;

    template <class U>
    void rev_compose_into(Transform &xf, const U &a);


    // Base
    void do_compose_into(Transform &xf, const Base &b) final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, Base &b) const final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, const Base &b) const final
        { compose_into(xf, b); }

    // Identity
    void do_compose_into(Transform &xf, const Identity &b) final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, Identity &b) const final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, const Identity &b) const final
        { compose_into(xf, b); }

    void do_rev_compose_into(Transform &xf, const Identity &a) final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, Identity &a) const final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, const Identity &a) const final
        { rev_compose_into(xf, a); }

    // Scale
    void do_compose_into(Transform &xf, const Scale &b) final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, Scale &b) const final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, const Scale &b) const final
        { compose_into(xf, b); }

    void do_rev_compose_into(Transform &xf, const Scale &a) final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, Scale &a) const final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, const Scale &a) const final
        { rev_compose_into(xf, a); }

    // Rotation
    void do_compose_into(Transform &xf, const Rotation &b) final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, Rotation &b) const final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, const Rotation &b) const final
        { compose_into(xf, b); }

    void do_rev_compose_into(Transform &xf, const Rotation &a) final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, Rotation &a) const final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, const Rotation &a) const final
        { rev_compose_into(xf, a); }

    // Translation
    void do_compose_into(Transform &xf, const Translation &b) final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, Translation &b) const final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, const Translation &b) const final
        { compose_into(xf, b); }

    void do_rev_compose_into(Transform &xf, const Translation &a) final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, Translation &a) const final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, const Translation &a) const final
        { rev_compose_into(xf, a); }

    // Linear
    void do_compose_into(Transform &xf, const Linear &b) final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, Linear &b) const final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, const Linear &b) const final
        { compose_into(xf, b); }

    void do_rev_compose_into(Transform &xf, const Linear &a) final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, Linear &a) const final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, const Linear &a) const final
        { rev_compose_into(xf, a); }

    // Affine
    void do_compose_into(Transform &xf, const Affine &b) final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, Affine &b) const final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, const Affine &b) const final
        { compose_into(xf, b); }

    void do_rev_compose_into(Transform &xf, const Affine &a) final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, Affine &a) const final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, const Affine &a) const final
        { rev_compose_into(xf, a); }

    // Projective
    void do_compose_into(Transform &xf, const Projective &b) final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, Projective &b) const final
        { compose_into(xf, b); }
    void do_compose_into(Transform &xf, const Projective &b) const final
        { compose_into(xf, b); }

    void do_rev_compose_into(Transform &xf, const Projective &a) final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, Projective &a) const final
        { rev_compose_into(xf, a); }
    void do_rev_compose_into(Transform &xf, const Projective &a) const final
        { rev_compose_into(xf, a); }
};

template<class T>
struct is_xform;

template <class T, class U, class EN=void>
struct can_multiply;

template <class T, class U, class EN=void>
struct can_subtract;

// used only if T::operator*=(U) exists
template <class T, class U>
auto operator*(T a, const U &b)
    -> typename std::enable_if<can_multiply<T,U>::value, T>::type;

// used only if T::operator-=(U) exists
template <class T, class U>
auto operator-(T a, const U &b)
    -> typename std::enable_if<can_subtract<T,U>::value, T>::type;


}

using xform::is_xform;

} // namespace rvg::xform

#include "base.hpp"

#endif
