#if TRACE_XFORM
#   include <iostream> // for std::clog
#endif

namespace rvg { namespace xform
{

inline r2::Point Base::operator*(const r2::Point &p) const
{
    return do_multiply(p);
}

inline r2::Vector Base::operator*(const r2::Vector &v) const
{
    return do_multiply(v);
}

inline r3::Point Base::operator*(const r3::Point &p) const
{
    return do_multiply(p);
}


#if TRACE_XFORM
template <class T>
void *BaseHelper<T>::operator new(std::size_t s)
{
    std::clog << "-- new " << T::name << " in heap" << std::endl;
    return ::operator new(s);
}

template <class T>
void *BaseHelper<T>::operator new(std::size_t s, std::nothrow_t) throw()
{
    std::clog << "-- new " << T::name << " in heap" << std::endl;
    return ::operator new(s, std::nothrow_t());
}

template <class T>
void *BaseHelper<T>::operator new(std::size_t s, void *p)
{
    std::clog << "-- new " << T::name << " in heap" << std::endl;
    return ::operator new(s, p);
}
#endif

template <class T>
T &BaseHelper<T>::derived()
{
    return static_cast<T &>(*this);
}

template <class T>
const T &BaseHelper<T>::derived() const
{
    return static_cast<const T &>(*this);
}

class Base::DoComposition
{
public:
    // called when types are compatible (T *= U exists)
    template <class T, class U, class XF>
    static auto compose_into(XF &xf, T &, U &b)
        -> typename std::enable_if<can_multiply<T,U>::value>::type
    {
        T &a = static_cast<T &>(*xf.m_impl.rw());

#if TRACE_XFORM
        std::clog << "-- inplace compose " << a << " * " << b << std::endl;
#endif
        a *= b;
    }

    // called when types are reverse-compatible
    template <class T, class U, class XF>
    static auto compose_into(XF &xf, T &a, U &)
        -> typename std::enable_if<!can_multiply<T,U>::value &&
                                   can_multiply<U,T>::value>::type
    {
        U &b = static_cast<U &>(*xf.m_impl.rw());

#if TRACE_XFORM
        std::clog << "-- inplace compose " << a << " * " << b << std::endl;
#endif
        b = a*b;
    }

    // called when types aren't compatible either way
    template <class T, class U, class XF>
    static auto compose_into(XF &xf, T &a, U &b)
        -> typename std::enable_if<(std::is_const<T>::value &&
                                   std::is_const<U>::value)
                                || (!can_multiply<T,U>::value &&
                                    !can_multiply<U,T>::value)
                            >::type
    {
        using typeA = typename std::remove_cv<T>::type;
        using typeB = typename std::remove_cv<U>::type;

        using result_type =
            typename std::remove_cv<typename std::remove_reference<
                decltype(std::declval<typeA>()*std::declval<typeB>())
            >::type>::type;

#if TRACE_XFORM
        std::clog << "-- compose " << a << " * " << b << std::endl;
#endif
        xf = Transform(new result_type(a*b));
    }
};

template <class T> template <class U>
void BaseHelper<T>::compose_into(Transform &xf, const U &b)
{
    b.rev_compose_into(xf, derived());
}

template <class T> template <class U>
void BaseHelper<T>::rev_compose_into(Transform &xf, U &a) const
{
    DoComposition::compose_into(xf, a, derived());
}

template <class T> template <class U>
void BaseHelper<T>::compose_into(Transform &xf, U &b) const
{
    b.rev_compose_into(xf, derived());
}

template <class T> template <class U>
void BaseHelper<T>::rev_compose_into(Transform &xf, const U &a)
{
    DoComposition::compose_into(xf, a, derived());
}

template <class T>
void BaseHelper<T>::accept(Visitor &v) const
{
    // gcc will devirtualize this call
    // if this function was reached through
    // non-virtual call due to previous
    // compiler devirtualization
    v.visit(derived());
}

template <class T>
Base *BaseHelper<T>::do_clone() const
{
#if TRACE_XFORM
    std::clog << "-- cloning " << *this << std::endl;
#endif

    return new T(derived());
}

template<class T>
struct is_xform
    : std::integral_constant<bool, std::is_base_of<Base,T>::value ||
                                   std::is_same<T,Transform>::value>
{
};

// strips out reference
template<class T>
struct is_xform<T&> : is_xform<T>
{
};

template <class T, class U, class EN>
struct can_subtract
    : std::integral_constant<bool, false>
{
};

template <class T, class U, class EN>
struct can_multiply
    : std::integral_constant<bool, false>
{
};

// if T::operator*=(U) exists, then T and U are compatible
template <class T, class U>
struct can_multiply<T, U, 
    typename std::enable_if<
        sizeof(std::declval<T>()*=std::declval<U>())!=0
    >::type>
    : std::integral_constant<bool, true>
{
};

// if T::operator-=(U) exists, then T and U are compatible
template <class T, class U>
struct can_subtract<T, U, 
    typename std::enable_if<
        sizeof(std::declval<T>()-=std::declval<U>())!=0
    >::type>
    : std::integral_constant<bool, true>
{
};

// used only if T::operator*=(U) exists
template <class T, class U>
auto operator*(T a, const U &b)
    -> typename std::enable_if<can_multiply<T,U>::value, T>::type
{
    return std::move(a *= b);
}

// used only if T::operator-=(U) exists
template <class T, class U>
auto operator-(T a, const U &b)
    -> typename std::enable_if<can_subtract<T,U>::value, T>::type
{
    return std::move(a -= b);
}

}} // namespace rvg::xform
