namespace rvg { namespace xform
{

#if TRACE_XFORM
inline Linear::Linear()
{
    std::clog << "-- default ctor " << name << "()" << std::endl;
}
#endif

inline Linear::Linear(float a11, float a12,
                      float a21, float a22)
    : m_matrix{ {{{a11,a12}},
                 {{a21,a22}}} }
{
#if TRACE_XFORM
    std::clog << "-- ctor " << *this << std::endl;
#endif
}

inline Linear::Linear(const Identity &)
    : m_matrix{ {{{1.f,0.f}},
                 {{0.f,1.f}}} }
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Identity) " << *this << std::endl;
#endif
}

inline Linear::Linear(const Rotation &b)
    : m_matrix{ {{{b.c,-b.s}},
                 {{b.s, b.c}}} }
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Rotation) " << *this << std::endl;
#endif
}

inline Linear::Linear(const Scale &b)
    : m_matrix{ {{{b.sx,0.f}},
                 {{0.f,b.sy}}} }
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Scale) " << *this << std::endl;
#endif
}

inline auto Linear::operator[](int i)
    -> std::array<float,2> &
{
    return m_matrix[i];
}

inline auto Linear::operator[](int i) const
    -> const std::array<float,2> &
{
    return m_matrix[i];
}

inline r2::Vector Linear::col(int i) const
{
    return r2::Vector{m_matrix[0][i], m_matrix[1][i]};
}

inline r2::Vector Linear::row(int i) const
{
    return r2::Vector{m_matrix[i][0], m_matrix[i][1]};
}

inline Linear transpose(const Linear &a) {
    return Linear(a[0][0], a[1][0], a[0][1], a[1][1]);
}

inline float det(const Linear &a) {
    return a[0][0]*a[1][1]-a[1][0]*a[0][1];
}

inline Linear cofactors(const Linear &a) {
    return Linear{a[1][1], -a[1][0], -a[0][1], a[0][0]};
}

inline Linear star(const Linear &a) {
    return transpose(cofactors(a));
}

inline Linear inv(const Linear &a) {
    return (1.f/det(a))*star(a);
}

inline float tr(const Linear &a) { // trace
    return a[0][0]+a[1][1];
}

inline float frobenius(const Linear &a) {
    auto sq = [](float x) { return x*x; };
    return sq(a[0][0]) + sq(a[0][0]) + sq(a[1][0]) + sq(a[1][1]);
}

inline Linear &Linear::operator *=(const Linear &that)
{
    return *this = *this * that;
}

// Shouldn't we instead define the operator*=() and leave
// the operator*() for the template in base.h??
inline Linear operator *(const Linear &a, const Linear &b)
{
    Linear r;

    r[0][0] = a[0][0]*b[0][0] + a[0][1]*b[1][0];
    r[0][1] = a[0][0]*b[0][1] + a[0][1]*b[1][1];
    r[1][0] = a[1][0]*b[0][0] + a[1][1]*b[1][0];
    r[1][1] = a[1][0]*b[0][1] + a[1][1]*b[1][1];

    return std::move(r);
}

inline Linear &Linear::operator *=(float b)
{
    auto &a = *this;

    a[0][0] *= b;
    a[0][1] *= b;
    a[1][0] *= b;
    a[1][1] *= b;

    return *this;
}


inline Linear &Linear::operator *=(const Scale &that)
{
    auto &a = *this;
    const auto &b = that;

    a[0][0] *= b.sx;
    a[0][1] *= b.sy;
    a[1][0] *= b.sx;
    a[1][1] *= b.sy;

    return *this;
}

inline Linear &Linear::operator *=(const Rotation &that)
{
    auto &a = *this;
    const auto &b = that;

    float a00 = a[0][0];
    a[0][0] =  b.c*a00 + b.s*a[0][1];
    a[0][1] = -b.s*a00 + b.c*a[0][1];

    float a10 = a[1][0];
    a[1][0] =  b.c*a10 + b.s*a[1][1];
    a[1][1] = -b.s*a10 + b.c*a[1][1];

    return *this;
}

inline Linear &Linear::operator -=(const Linear &that)
{
    auto &a = *this;
    const auto &b = that;

    a[0][0] -= b[0][0];
    a[0][1] -= b[0][1];
    a[1][0] -= b[1][0];
    a[1][1] -= b[1][1];

    return *this;
}


inline Linear operator*(const Rotation &a, const Scale &b)
{
    return Linear(a.c*b.sx, -a.s*b.sy,
                  a.s*b.sx,  a.c*b.sy);
}

inline Linear operator*(const Rotation &a, const Linear &b)
{
    return Linear(a.c*b[0][0] - a.s*b[1][0], a.c*b[0][1] - a.s*b[1][1],
                  a.s*b[0][0] + a.c*b[1][0], a.s*b[0][1] + a.c*b[1][1]);
}

inline Linear operator*(const Scale &a, const Rotation &b)
{
    return Linear(b.c*a.sx, -b.s*a.sx,
                  b.s*a.sy,  b.c*a.sy);
}

inline Linear operator*(float a, const Linear &b)
{
    return Linear(a*b[0][0], a*b[0][1],
                  a*b[1][0], a*b[1][1]);
}

inline Linear operator*(const Scale &a, const Linear &b)
{
    return Linear(a.sx*b[0][0], a.sx*b[0][1],
                  a.sy*b[1][0], a.sy*b[1][1]);
}

inline r2::Point Linear::do_multiply(const r2::Point &p) const
{
    auto &L = *this;

    return {L[0][0]*p.x + L[0][1]*p.y, L[1][0]*p.x + L[1][1]*p.y};
}

inline r2::Vector Linear::do_multiply(const r2::Vector &v) const
{
    auto &L = *this;

    return {L[0][0]*v.x + L[0][1]*v.y, L[1][0]*v.x + L[1][1]*v.y};
}

inline r3::Point Linear::do_multiply(const r3::Point &p) const
{
    auto &L = *this;

    return {L[0][0]*p.x + L[0][1]*p.y, L[1][0]*p.x + L[1][1]*p.y, p.z};
}

}} // namespace rvg::xform
