namespace rvg { namespace xform
{

#if TRACE_XFORM
inline Projective::Projective()
{
    std::clog << "-- default ctor " << name << "()" << std::endl;
}
#endif

inline Projective::Projective(float a11, float a12, float a13,
                              float a21, float a22, float a23,
                              float a31, float a32, float a33)
    : m_matrix{ {{{a11,a12,a13}},
                 {{a21,a22,a23}},
                 {{a31,a32,a33}}} }
{
#if TRACE_XFORM
    std::clog << "-- ctor " << *this << std::endl;
#endif
}

inline Projective::Projective(const Identity &)
    : m_matrix{ {{{1.f,0.f,0.f}},
                 {{0.f,1.f,0.f}},
                 {{0.f,0.f,1.f}}} }
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Identity) " << *this << std::endl;
#endif
}

inline Projective::Projective(const Scale &b): Projective(Affine(b)) // delegate
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Scale) " << *this << std::endl;
#endif
}

inline Projective::Projective(const Rotation &b): Projective(Affine(b)) // delegate
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Rotation) " << *this << std::endl;
#endif
}

inline Projective::Projective(const Linear &b): Projective(Affine(b)) // delegate
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Linear) " << *this << std::endl;
#endif
}

inline Projective::Projective(const Affine &b)
    : m_matrix{ {{{b[0][0],b[0][1],b[0][2]}},
                 {{b[1][0],b[1][1],b[1][2]}},
                 {{0.f,0.f,1.f}}} }
{
#if TRACE_XFORM
    std::clog << "-- copy ctor (Affine) " << *this << std::endl;
#endif
}

inline auto Projective::operator[](int i) 
    -> std::array<float,3> &
{
    return m_matrix[i];
}

inline auto Projective::operator[](int i) const
    -> const std::array<float,3> &
{
    return m_matrix[i];
}

inline Projective transpose(const Projective &a) {
    return Projective(
        a[0][0], a[1][0], a[2][0],
        a[0][1], a[1][1], a[2][1],
        a[0][2], a[1][2], a[2][2]
    );
}

inline Linear Projective::linear(void) const
{
    auto &a = *this;
    return Linear(a[0][0], a[0][1], a[1][0], a[1][1]);
}

inline Affine Projective::affine(void) const
{
    auto &a = *this;
    return Affine(a[0][0], a[0][1], a[0][2], a[1][0], a[1][1], a[1][2]);
}

inline Projective &Projective::operator *=(const Translation &that)
{
    auto &a = *this;
    const auto &b = that;

    a[0][2] += a[0][0]*b.dx + a[0][1]*b.dy;
    a[1][2] += a[1][0]*b.dx + a[1][1]*b.dy;
    a[2][2] += a[2][0]*b.dx + a[2][1]*b.dy;

    return *this;
}

inline Projective &Projective::operator *=(const Rotation &that)
{
    auto &a = *this;
    const auto &b = that;

    float a00 = a[0][0];
    a[0][0] =  b.c*a00 + b.s*a[0][1];
    a[0][1] = -b.s*a00 + b.c*a[0][1];

    float a10 = a[1][0];
    a[1][0] =  b.c*a10 + b.s*a[1][1];
    a[1][1] = -b.s*a10 + b.c*a[1][1];

    float a20 = a[2][0];
    a[2][0] =  b.c*a20 + b.s*a[2][1];
    a[2][1] = -b.s*a20 + b.c*a[2][1];

    return *this;
}

inline Projective &Projective::operator *=(const Scale &that)
{
    auto &a = *this;
    const auto &b = that;

    a[0][0] *= b.sx;
    a[0][1] *= b.sy;
    a[1][0] *= b.sx;
    a[1][1] *= b.sy;
    a[2][0] *= b.sx;
    a[2][1] *= b.sy;

    return *this;
}

inline Projective &Projective::operator *=(const Linear &that)
{
    auto &a = *this;
    const auto &b = that;

    float a00 = a[0][0];
    a[0][0] = a00*b[0][0] + a[0][1]*b[1][0];
    a[0][1] = a00*b[0][1] + a[0][1]*b[1][1];

    float a10 = a[1][0];
    a[1][0] = a10*b[0][0] + a[1][1]*b[1][0];
    a[1][1] = a10*b[0][1] + a[1][1]*b[1][1];

    float a20 = a[2][0];
    a[2][0] = a20*b[0][0] + a[2][1]*b[1][0];
    a[2][1] = a20*b[0][1] + a[2][1]*b[1][1];

    return *this;
}

inline Projective &Projective::operator *=(const Affine &that)
{
    auto &a = *this;
    const auto &b = that;

    float a00 = a[0][0];
    a[0][0] =  a00*b[0][0]+a[0][1]*b[1][0];
    a[0][2] += a00*b[0][2]+a[0][1]*b[1][2];
    a[0][1] =  a00*b[0][1]+a[0][1]*b[1][1];

    float a10 = a[1][0];
    a[1][0] =  a10*b[0][0]+a[1][1]*b[1][0];
    a[1][2] += a10*b[0][2]+a[1][1]*b[1][2];
    a[1][1] =  a10*b[0][1]+a[1][1]*b[1][1];

    float a20 = a[2][0];
    a[2][0] =  a20*b[0][0]+a[2][1]*b[1][0];
    a[2][2] += a20*b[0][2]+a[2][1]*b[1][2];
    a[2][1] =  a20*b[0][1]+a[2][1]*b[1][1];

    return *this;
}

inline Projective &Projective::operator *=(const Projective &that)
{
    return *this = *this * that;
}

inline Projective operator*(const Projective &a, const Projective &b)
{
    Projective r;

    r[0][0] = a[0][0]*b[0][0]+a[0][1]*b[1][0]+a[0][2]*b[2][0];
    r[0][1] = a[0][0]*b[0][1]+a[0][1]*b[1][1]+a[0][2]*b[2][1];
    r[0][2] = a[0][0]*b[0][2]+a[0][1]*b[1][2]+a[0][2]*b[2][2];

    r[1][0] = a[1][0]*b[0][0]+a[1][1]*b[1][0]+a[1][2]*b[2][0];
    r[1][1] = a[1][0]*b[0][1]+a[1][1]*b[1][1]+a[1][2]*b[2][1];
    r[1][2] = a[1][0]*b[0][2]+a[1][1]*b[1][2]+a[1][2]*b[2][2];

    r[2][0] = a[2][0]*b[0][0]+a[2][1]*b[1][0]+a[2][2]*b[2][0];
    r[2][1] = a[2][0]*b[0][1]+a[2][1]*b[1][1]+a[2][2]*b[2][1];
    r[2][2] = a[2][0]*b[0][2]+a[2][1]*b[1][2]+a[2][2]*b[2][2];

    return std::move(r);
}

inline Projective operator*(const Rotation &a, const Projective &b)
{
    return Projective(a.c*b[0][0] - a.s*b[1][0],
                      a.c*b[0][1] - a.s*b[1][1],
                      a.c*b[0][2] - a.s*b[1][2],

                      a.s*b[0][0] + a.c*b[1][0],
                      a.s*b[0][1] + a.c*b[1][1],
                      a.s*b[0][2] + a.c*b[1][2],

                                      b[2][0],
                                      b[2][1],
                                      b[2][2]);
}

inline Projective operator*(const Translation &a, const Projective &b)
{
    return Projective(b[0][0]+b[2][0]*a.dx,
                      b[0][1]+b[2][1]*a.dx,
                      b[0][2]+b[2][2]*a.dx,

                      b[1][0]+b[2][0]*a.dy,
                      b[1][1]+b[2][1]*a.dy,
                      b[1][2]+b[2][2]*a.dy,

                                   b[2][0],
                                   b[2][1],
                                   b[2][2]);
}

inline Projective operator*(const Scale &a, const Projective &b)
{
    return Projective(a.sx*b[0][0], a.sx*b[0][1], a.sx*b[0][2],
                      a.sy*b[1][0], a.sy*b[1][1], a.sy*b[1][2],
                           b[2][0],      b[2][1],      b[2][2]);
}

inline Projective operator*(const Linear &a, const Projective &b)
{
    return Projective(a[0][0]*b[0][0]+a[0][1]*b[1][0],
                      a[0][0]*b[0][1]+a[0][1]*b[1][1],
                      a[0][0]*b[0][2]+a[0][1]*b[1][2],

                      a[1][0]*b[0][0]+a[1][1]*b[1][0],
                      a[1][0]*b[0][1]+a[1][1]*b[1][1],
                      a[1][0]*b[0][2]+a[1][1]*b[1][2],

                                              b[2][0],
                                              b[2][1],
                                              b[2][2]);
}

inline Projective operator*(const Affine &a, const Projective &b)
{
    return Projective(a[0][0]*b[0][0]+a[0][1]*b[1][0]+a[0][2]*b[2][0],
                      a[0][0]*b[0][1]+a[0][1]*b[1][1]+a[0][2]*b[2][1],
                      a[0][0]*b[0][2]+a[0][1]*b[1][2]+a[0][2]*b[2][2],

                      a[1][0]*b[0][0]+a[1][1]*b[1][0]+a[1][2]*b[2][0],
                      a[1][0]*b[0][1]+a[1][1]*b[1][1]+a[1][2]*b[2][1],
                      a[1][0]*b[0][2]+a[1][1]*b[1][2]+a[1][2]*b[2][2],

                                                              b[2][0],
                                                              b[2][1],
                                                              b[2][2]);
}

inline r2::Point Projective::do_multiply(const r2::Point &p) const
{
    auto &A = *this;

    return {A[0][0]*p.x + A[0][1]*p.y + A[0][2],
            A[1][0]*p.x + A[1][1]*p.y + A[1][2]};
}

inline r2::Vector Projective::do_multiply(const r2::Vector &v) const
{
    auto &A = *this;

    return {A[0][0]*v.x + A[0][1]*v.y,
            A[1][0]*v.x + A[1][1]*v.y};
}

inline r3::Point Projective::do_multiply(const r3::Point &p) const
{
    auto &A = *this;

    return {A[0][0]*p.x + A[0][1]*p.y + A[0][2]*p.z,
            A[1][0]*p.x + A[1][1]*p.y + A[1][2]*p.z,
            A[2][0]*p.x + A[2][1]*p.y + A[2][2]*p.z};
}

}} // namespace rvg::xform
